/*****************************************************************************
  $Id: chooseNK.mu,v 1.1.1.2 2005/01/26 13:42:24 nthiery Exp $
  File:		lib/COMBINAT/chooseNK.mu
  Authors:	Florent Hivert <florent.hivert@univ-mlv.fr>
            	Xavier Molinero <molinero@lsi.upc.es>
  Created:	08/01/2004
  License:	LGPL
*****************************************************************************/

domain combinat::chooseNK
    inherits Dom::BaseDomain;
    category Cat::CombinatorialClass;
    axiom    Ax::systemRep;

    info_str := "Library 'combinat::chooseNK': utilities for choosing k numbers in the set {1..n}";

    isA  := 
    proc(obj : Type::AnyType, 
	 n = null()  : Type::Union(DOM_NULL, Type::NonNegInt), 
	 k = null()  : Type::Union(DOM_NULL, Type::NonNegInt))
	local maxi;
    begin
	if not testtype(obj, Type::ListOf(Type::PosInt))
	   then return(FALSE)
	end_if;
	maxi := max(0,op(obj)); // 0 to make this work with the empty subset
	if args(0) > 1 and maxi > n then return(FALSE) end_if;
	if args(0) > 2 and nops(obj) <> k then return(FALSE) end_if;
	return(TRUE);
    end_proc;

    count := 
    proc(n: Type::NonNegInt, k: Type::NonNegInt)
    begin
	combinat::binomialRecursive(n, k);
    end_proc;

    generator :=
    proc(n: Type::NonNegInt, k: Type::NonNegInt)
        : combinat::generators
        option escape;
        local subword, first_time, dif, i;
        /* 
           generator(n,k)
		generator for the subwords of length k of the word [1..n]

           The subwords are generated lexicographically. Algorithm:
           find in the current subword the biggest element i that can
           be increased without collision, increase it, and reset all
           the bigger elements to be the immediate successors i+1,
           i+2, ... of i. For example: with n=9 and l=[1,3,7,8,9], we
           find that i is 3, and so the next subword is [1,4,5,6,7].

           (that's essentially Kemp's algorithm)

           Complexity:
	    - Constant time amortized
	    - worst case linear for each iteration

           The algorithm is actually slightly more general. We
           generate all words of length k of [1..n] such that
           subwords[i]+dif <= subwords[i+1].

	   Setting dif=1, this generates all the subwords, and setting
	   dif=0, all the subwords with repetitions.
	*/

    begin
        if testargs() then
            if args(0) > 3 or (args(0)=3 and args(3) <> Repeat) then
                error("Wrong number of arguments");
            end_if ;
        end_if;

        if args(0)=3 then // repetitions are allowed (better to pass the diff parameter!)
            dif := 0;
        else
            dif := 1;
        end_if;
        
        if k=0 then
            return(combinat::generators::repeater([],1));
        end_if;
        if n<1+(k-1)*dif then
            return(combinat::generators::empty);
        else
            subword:=[ 1+i*dif $ i=0..k-1];
        end_if;
        
        first_time := TRUE;
        return(proc()
                   local i,j;
               begin
                   if first_time then
                       first_time:=FALSE;
                       return(subword);
                   end_if;
                   
                   // Find the biggest element that can be increased
                   if subword[k]<n then
                       subword[k]:=subword[k]+1;
                       return(subword);
                   end_if;
                   for i from k-1 downto 1 do
                       if subword[i]+dif<subword[i+1] then
                           subword[i]:=subword[i]+1;
                           // reset the bigger elements
                           for j from 1 to k-i do
                               subword[i+j]:=subword[i]+j*dif;
                           end_for;
                           return(subword);
                       end_if;
                   end_for;
                   return(FAIL);
               end_proc)
    end_proc;

    // Floyd's algorithm whose complexity is O(k + k*log(k))
    random :=
    proc(n : Type::NonNegInt, k : Type::NonNegInt)
	local l, i, r, rnd, lRes;
    begin
        l:=[$1..n];
        for i from n downto n-k+1 do
            rnd        := random(i);
            r          := rnd() + 1;
            lRes[n-i+1]:= l[r];
            if r < i then l[r]:= l[i] end_if
        end_for;
        sort([ lRes[i] $ i=1..k ]);
    end_proc;
    
    // complexity: O(k*log(n) + k*log(k))
    randomSlow :=
    proc(n : Type::NonNegInt,
	 k : Type::NonNegInt)
	local i, rem, res, posRes, iChoice, choice,
	      randomGenerator;
    begin
	randomGenerator := random(n);
	rem    := [$1..n];
	res    := [0$k];
	posRes := 0;
	for i from k downto 1 do
	    iChoice := randomGenerator()+1;
	    choice  := op(rem, iChoice);
	    delete(rem[iChoice]);
	    n       := n - 1;
	    res[(posRes := posRes+1)] := choice;
	end_for;
	sort(res);
    end;

    // complexity: O(k*log(n))
    // It uses a dichotomic search and the Lexicographic order.
    // The rank starts from 1
    unrank :=
    proc(r: Type::PosInt, n: Type::NonNegInt, k: Type::NonNegInt)
         local rr, ii, kk, res;
    begin
        if r > combinat::binomialRecursive(n,k) then return(FAIL) end_if;
        rr    := r;
        ii    := 0;
        res[0]:= 0:
        for kk from k downto 1 do
            ii     := ii+1;
            // Values r, n, k have to be reorganized for each iteration
            res[ii]:= combinat::chooseNK::DichSearch(
	                        rr, n-res[ii-1], kk, kk-1, n-res[ii-1]-1);
            rr     := rr - combinat::chooseNK::sumSearch(
	                             kk-1, n-res[ii-1]-res[ii]+1, n-res[ii-1]-1);
            res[ii]:= res[ii] + res[ii-1]
        end_for;
        [ res[kk] $ kk=1..k ]
    end_proc;
    // Partial sum "sum(binomial(ll,k) $ ll=n0..nf))"
    sumSearch :=
    proc(k: Type::NonNegInt, n0: Type::NonNegInt, nf: Type::NonNegInt)
      : Type::NonNegInt
    begin
        combinat::binomialRecursive(nf+1, k+1) -
        combinat::binomialRecursive(n0+1, k+1) +
        combinat::binomialRecursive(n0, k)
    end_proc;
    // It finds "nm" such that
    // "sumSearch(k,nm+1,nf) >= i >sumSearch(k,nm,nf)",
    // where n0<=nm<=nf 
    DichSearch :=
    proc(i: Type::PosInt, n: Type::NonNegInt, k: Type::NonNegInt,
         n0: Type::NonNegInt, nf: Type::NonNegInt)
        local nm;
    begin
        if ( i = 1 ) then return(1) end_if;
        if (nf = n0) then return(n-n0) end_if;
        if (nf-n0 = 1) then
           if i <= combinat::chooseNK::sumSearch(k-1, nf, n-1) then return(n-nf)
           else return(n-n0)
           end_if
        end_if;
        nm:= floor((n0+nf) / 2);
        if i <= combinat::chooseNK::sumSearch(k-1, nm, n-1) then
           combinat::chooseNK::DichSearch(i, n, k, nm, nf)
        else combinat::chooseNK::DichSearch(i, n, k, n0, nm)
        end_if
    end_proc;

    // complexity: O(k*log(n))
    // It uses a dichotomic search, but not the Lexicographic order.
    // The rank starts from 1
    unrankNoLexi :=
    proc(r:Type::NonNegInt, n:Type::NonNegInt, k:Type::NonNegInt) //: Type::ListOf(Type::NonNegInt)
      local res, i;
    begin
      if r > combinat::binomialRecursive(n,k) then return(FAIL) end_if;
      [(res:= combinat::chooseNK::unrankDichSearchNoLexi(r, k, 1, n);
        r:= r - combinat::binomialRecursive(res-1, k); k:=k-1; res)
        $ i=1..k]
    end_proc;

    unrankDichSearchNoLexi :=
    proc(i:Type::NonNegInt, k:Type::NonNegInt,
	 k0:Type::NonNegInt, ke:Type::NonNegInt) : Type::NonNegInt
    local km;
    begin
      if i=0 then return(k0) end_if;
      if ke-k0 = 0 then return(k0)
      elif ke-k0 = 1 then if i <= combinat::binomialRecursive(k0,k) then return(k0);
			  else return(ke);
                          end_if;
      else km:= ceil((ke+k0)/2);
	if i <= combinat::binomialRecursive(km,k) then
           combinat::chooseNK::unrankDichSearchNoLexi(i, k, k0, km);
	else combinat::chooseNK::unrankDichSearchNoLexi(i, k, km, ke);
	end_if;
      end_if;
    end_proc;

    // complexity: O(k*n)
    // It does not uses a dichotomic search and the Lexicographic order.
    // The rank starts from 1
    unrankSlow := 
    proc(r : Type::PosInt, n: Type::NonNegInt, k: Type::NonNegInt)
	local res, newchoice, lastchoice, bin, posRes;
    begin
	if r > combinat::binomialRecursive(n,k) then return(FAIL); end_if;
	res    := [0$k];
	posRes := 0;
	lastchoice := 0;
	r := r-1; // the rank start from 1...
	while (k > 0) do
	    k := k-1;
	    for newchoice from n-1 downto k do
		bin := combinat::binomialRecursive(newchoice, k);
		if r >= bin then r := r - bin;
		else break;
		end_if;
	    end_for;
	    lastchoice := lastchoice + n - newchoice;
	    n := newchoice;
	    res[(posRes := posRes+1)] := lastchoice;
	end_while;
	res;
    end;

end_domain:
