/*****************************************************************************
  $Id: compositions.mu,v 1.1.1.6 2005/01/26 13:42:24 nthiery Exp $
  File:		lib/COMBINAT/compositions.mu
  Authors:	Nicolas M. Thiry <nthiery@users.sourceforge.net>
                Florent Hivert <florent.hivert@univ-mlv.fr>
  Created:	16/02/2001
  License:	LGPL
*****************************************************************************/

// Optimization for MuPAD > 2.0.0: we use RD_INF which is faster than infinity
%if traperror(RD_INF+RD_INF) = 0 then
    alias(infinity  = RD_INF);
end_if:

domain combinat::compositions
    inherits Dom::BaseDomain;
    category Cat::GradedCombinatorialClass, Cat::IntegerListsLexClass;
    axiom    Ax::systemRep;

    info_str := "Library 'combinat::compositions': utilities for compositions";

    interfaceAdd :=
		{ hold(finer), hold(fatter), hold(isFiner), 
		  hold(finerVectors),		  
		  hold(conjugate), 
		  hold(descents), hold(descentsSet), hold(fromDescents),
		  hold(majorIndex),
		  hold(peaks), hold(peaksComposition),
		  hold(toSkewPartition),
                  hold(printPretty), hold(printCompact),
		  hold(peaks), hold(peaksComposition), hold(fromPeaks)
		};

    parseOptions :=
    proc()
    begin
        dom::parseOptionsIntegerListsLexClass
        (table(MinLength = 0,
               MaxLength = infinity,
               Length	 = NIL,
               MinPart	 = 1,
               MaxPart	 = infinity,
               Inner	 = NIL,
               Outer	 = NIL,
	       MinSlope	 = -infinity,
	       MaxSlope	 = infinity
              ),
         args()
         ):
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // Pretty-print
    //////////////////////////////////////////////////////////////////////////
 
    printPretty :=
        combinat::skewPartitions::printPretty@dom::toSkewPartition;
 
    printCompact :=
        combinat::skewPartitions::printCompact@dom::toSkewPartition;
 
    // For backward compatibility
    prettyPrint := dom::printPretty;
 
    //////////////////////////////////////////////////////////////////////////
    // count
    //////////////////////////////////////////////////////////////////////////
    /// count(n, <Options>)	number of compositions of n
    count :=
    proc(n: Type::Integer) : Type::NonNegInt
    begin
        if args(0) > 1 then
            // There is no efficient counter yet when there are
            // further constraints.
            return(dom::count_integerListsLexClass(args()));
        end_if;

        if n >= 1 then
            2^(n-1);
        elif n=0 then
            1
        else
            0;
        end_if;
    end_proc;

    fastList :=
    proc(n : Type::PosInt) : Type::ListOf(dom)
	local i;
    begin
	if n = 0 then return([[]]); end_if;
	_concat(map(eval(procname(n-i)), z->_concat([i],z)) $ i=1..n);
    end_proc;

    makeFastList2 :=
    proc() 
	option escape;
	local bloodyFastListRec;
    begin 
	bloodyFastListRec := 
        proc(n : Type::PosInt) : Type::ListOf(dom)
	    local i;
	begin
	    if n = 0 then return([[]]); end_if;
	    _concat(map(bloodyFastListRec(n-i), z->_concat([i],z)) $ i=1..n);
	end_proc;
    end_proc;
  fastList2 := dom::makeFastList2();

/// Refinement order of compositions
    isFiner := 
    proc(c1 : combinat::integerVectors, c2 : dom) : DOM_BOOL
	local i1, sum1, i2, sum2;
    begin
	if testargs() and 
	   _plus(op(c1)) <> _plus(op(c2))
	    then 
	    error("Sum mismatch : ".expr2text(c1)." and ".expr2text(c2));
	end_if;
	if (c1=[]) and (c2=[]) then
 	    TRUE;
	end_if;
	sum1 := 0; sum2 := 0;
	i1 := 0;
	for i2 from 1 to nops(c2) do
	    sum2 := sum2 + c2[i2];
 	    while sum1 < sum2 do
		i1 := i1 + 1;
 		sum1 := sum1 + c1[i1];
 	    end_while;
 	    if sum1 > sum2 then return(FALSE);
 	    end_if;
 	end_for;
 	TRUE;   
    end_proc;

    // Stolen from MuEC...
    conjugate := 
    proc(comp : dom) : dom
	local n, coofcp, cocjg, i, j;
    begin
	if (comp = []) then
	    [];
	else
	    n := nops(comp);
	    coofcp := [_plus(op(comp,1..j))-j+1 $ j=1..n]; 
	    cocjg  := [(i $ k=1..(coofcp[n-i+1]-coofcp[n-i])) $ i=1..n-1,
			 n $ j=1..coofcp[1]];
	    append([cocjg[1]], (cocjg[i]-cocjg[i-1]+1 $ i=2..nops(cocjg)));
	end_if;
    end_proc;

    // Stolen from MuEC...
    majorIndex := 
    proc(c : dom) : Type::NonNegInt
	local lv, i;
    begin
	lv := nops(c);
	if (lv=1) then 0; else _plus((lv-i)*op(c,i) $ i=1..lv); end_if;
    end_proc;


    //    d       : descent set
    //    args[2] : total sum
    fromDescents := 
    proc(d   : Type::Union(Type::ListOf(Type::PosInt),
			   Type::SetOf(Type::PosInt)),
	 nps = null() : Type::Union(Type::NonNegInt, DOM_NULL)) : dom
	local i;
    begin
	d := sort([op(d)]);
	if d=[] then 
	    if nps = 0 then return ([]);
	    else            return([nps]);
	    end_if;
	end_if;
	if nps <> null() then 
	    if testargs() and nps <= max(op(d)) then
		error(expr2text(d).
		      " is not included in ".
		      expr2text([$1..nps-1]));
	    end_if;
	    d := append(d, nps);
	end_if;
        [d[1], d[i+1]-d[i] $ i=1..nops(d)-1];
    end_proc;
 

/**
    descents(compo)
    Returns the descent list of the composition compo

    the option FinalDescent adds a descent at the end
**/
    descents := 
    proc(compo : Type::ListOf(Type::NonNegInt)) : Type::ListOf(Type::NonNegInt)
	local s, res, i;
    begin
	s := 0;
	res := [(s := compo[i]+s) $ i=1..nops(compo)-1];
	if args(0) > 1 and args(2) = FinalDescent and compo <> [] then
    // the empty composition has no final descent!!!???
		res := append(res, _plus(op(compo)));
	end_if;
	res;
    end_proc;
  
    descentsSet := DOM_SET@op@dom::descents;
  
/**
    peaks(compo)
    Returns the peak list of the composition compo
**/
  peaks := 
    proc(compo : dom) : dom
	local des, n;
    begin
	n := _plus(op(compo));
	des := dom::descentsSet(compo);
	sort([op(map(({$1..n} minus des), _plus, 1) intersect des)]);
    end_proc;

/**
    peaksComposition(compo)
    Returns the composition associated to peak set of the composition compo
**/
  peaksComposition := 
    proc(compo : dom) : dom
	local des, n;
    begin
	n := _plus(op(compo));
	des := dom::descentsSet(compo);
	des := sort([op(map(({$1..n} minus des), _plus, 1) intersect des)]);
	dom::fromDescents(des, n);
    end_proc; 

/**
    fromPeaks(peaks)
    Returns the compositions whose peaks set is compo
**/
  fromPeaks := 
    proc(ps   : Type::Union(Type::ListOf(Type::PosInt),
			   Type::SetOf(Type::PosInt)),
	 nps = null() : Type::Union(Type::NonNegInt, DOM_NULL)) : Type::ListOf(dom)
	local diffs, i, n, res;
    begin
	ps := sort([op(ps)]);
	if ps=[] and (nps = 0 or nps=null()) then return ([[]]); end_if;
	n := nops(ps);
	if nps=null() then 
	    nps := ps[n]; 
	    ps  := [op(ps, 1..n-1)];
	    n   := n-1;
	elif testargs() and (ps <> []) and
	     (nps <= max(op(ps)) or min(op(ps)) < 2) then
	    error(expr2text(ps).
		  " is not included in ".
		  expr2text([$2..nps-1]));
	end_if;
	if ps <> [] then
	    diffs := [ps[1], ps[i+1]-ps[i] $ i=1..n-1];
	    res   := combinat::cartesianProduct(op(map(diffs, z-> [$0..z-2])), [$0..nps-ps[n]-1]);
	    diffs := append(diffs, nps-ps[n]);
	else 
	    diffs := [nps];
	    res   := [[i]$i=0..nps-1];
	end_if;
	map(res,
	    proc(hooks) 
		local j;
	    begin
		[ (1$hooks[j], diffs[j]-hooks[j]) $ j=1..n+1];
	    end_proc);
    end_proc; 


    toSkewPartition :=
    proc(compo : dom) : combinat::skewPartitions 
	local resOuter, resInner, p, sumOuter;
    begin
	resOuter := [compo[1]];
	sumOuter := compo[1];
	resInner := [];
	for p in compo do
	    resInner := [sumOuter - 1].resInner;
	    sumOuter := sumOuter - 1 + p;
	    resOuter := [sumOuter].resOuter;
	end_for;
	[resOuter, resInner];
    end_proc;


end_domain:


domain combinat::compositions::finer
    inherits Dom::BaseDomain;
    category Cat::CombinatorialClass;
    axiom    Ax::systemRep;

    info      := "Library 'combinat::compositions::finer': utilities for compositions finer than a given one";

    domtype   := DOM_LIST;
    typeCompositions      := combinat::compositions;
    interface := dom::interfaceCombinatorialClass;

    // generator for finer composition
    generator :=
    proc(c : dom::typeCompositions)
    begin
	combinat::generators::map(
 	    combinat::generators::cartesian(
 		combinat::compositions::generator(x) $ x in c),
 	    map, op);
    end_proc;
begin
end_domain:

domain combinat::compositions::fatter
    inherits Dom::BaseDomain;
    category Cat::CombinatorialClass;
    axiom    Ax::systemRep;

    info      := "Library 'combinat::compositions::fatter': utilities for compositions fatter than a given one";

    domtype := DOM_LIST;
    typeCompositions    := combinat::compositions;

    generator :=
    proc(c : dom::typeCompositions)
	option escape;
    begin
	combinat::generators::map(
 	    combinat::compositions::generator(nops(c)),
            proc(ref)
		name "fatterGenerators";
		local psum, j, res;
	    begin 
		psum   := 0;
		res := ref;
		for j from 1 to nops(ref) do 
		    res[j] := _plus(c[k] $ k = psum + 1..psum + ref[j]);
		    psum := psum + ref[j];
		end_for;
		res;
	    end_proc);	  
    end_proc;
    
begin
end_domain:

// Backward compatibility aliases
combinat::compositions::fatters := combinat::compositions::fatter:
combinat::compositions::finers := combinat::compositions::finer:

domain combinat::compositions::finerVectors
    inherits Dom::BaseDomain;
    category Cat::CombinatorialClass;
    axiom    Ax::systemRep;

    info      := "Library 'combinat::compositions::finerVectors': utilities for vectors finer than a given composition";
    typeVectors      := combinat::integerVectors;

    // generate the integers vectors of length len whose 
    // associated composition is finer than c
    generator := 
    proc(c : dom::typeVectors, len : Type::NonNegInt)
	option escape;
	local nc;
    begin
	if (nc := nops(c)) = 0 then 
	    return(combinat::generators::fromList([[0$len]]));
	end_if;    
	combinat::generators::pipe(
    // Choose in how many part of c each part is broken 
	    combinat::compositions::generator(len, Length = nc),
    // compute the breaking of all parts
            brk -> combinat::generators::map(
 	             combinat::generators::cartesian(
    // To avoid repetition only the first part is allowed 
    // to start with a zero.
		         combinat::integerVectors::generator(c[1], brk[1]),
		         combinat::integerVectors::generator(c[i], brk[i], 
							     Inner = [1])
						     $ i = 2..nc),
    // concatenate all the breaking of the part
		   map, op));
    end_proc;
    
begin
end_domain:
    
    
