/*****************************************************************************
  $Id: dyckWords.mu,v 1.1.1.5 2005/01/26 13:42:24 nthiery Exp $
  File:		lib/COMBINAT/dyckWords.mu
  Authors:	Nicolas M. Thiry <nthiery@users.sourceforge.net>
  Created:	16/02/2001
  License:	LGPL
*****************************************************************************/

domain combinat::dyckWords
    inherits Dom::BaseDomain;
    category Cat::GradedCombinatorialClass, Cat::DecomposableClass;
    axiom    Ax::systemRep;

    info_str := "Library 'combinat::dyckWords': utilities for dyck Words";

/**
 A Dyck word of size n is a word with n zeroes and n ones such that in
 any prefix there is more zeroes than ones. If a one is interpreted
 as an opening parenthesis and a zero as a closing parenthesis, one gets
 a word of well balanced parenthesis. Usually, a Dyck word is
 interpreted as a discrete path in the plane NxN, going from (0,0) to
 (2n,0), with right-up steps (zeroes) and right-down steps (ones),
 without going below the x-axis.

 Examples:
  - [1,1,1,0,0,0]		((()))
  - [1,0,1,0,1,0]		()()()
  - [1,0,0,1,1,0]		())(() // Invalid!
  - [1,1,1,1,0,0]		(((()) // Invalid! but Prefix

 Should we store dyck words as strings \"(()())\" instead ?

 - generator(n)		generator for the dyckWords of size n
 - generator(n1,n2)	generator for the dyckWords with n1 '(' and n2 ')'

 - toNonCrossingPartition(w)	Bijection of Biane with non crossing partitions
 - fromNonCrossingPartition(p)	(see combinat::nonCrossingPartitions)
 - toString(w)		Conversion with strings of well balanced parenthesis
 - fromString(s)	
**/

    interfaceAdd :=
        {hold(typePrefix), hold(isAPrefix),
         hold(associatedParenthesis),
         hold(toNonCrossingPartition), hold(fromNonCrossingPartition),
         hold(toString), hold(fromString),
	 hold(printPretty), hold(size)
         };

    domtype := DOM_LIST;

    typePrefix := Type::Predicate("DyckWordPrefix", dom::isAPrefix);

    /// Symbol used for representing an opening parenthesis "("
    openSymbol   := 1;	// sym
    /// Symbol used for representing a closing parenthesis "("
    closeSymbol := 0;

    isAPrefix := 
    proc(obj : Type::AnyType, 
	 k1 = FAIL : Type::Union(Type::Integer, DOM_FAIL), 
	 k2 = FAIL : Type::Union(Type::Integer, DOM_FAIL))
	local p, nOpens, nCloses;
    begin
        if args(0) > 3 then error("Wrong number of arguments"); end_if; 
        if args(0) = 2 then k2:= k1; end_if; 
        if k1 <> FAIL and k1 < k2 then 
            error(expr2text(k1)." must be >= than ".expr2text(k2)); 
        end_if;
	if not testtype(obj, Type::ListOf(Type::Integer)) then 
	    return(FALSE);
	end_if;
	nOpens := 0; nCloses := 0;
	for p in obj do
	    if   p = dom::openSymbol  then nOpens  := nOpens  + 1;
	    elif p = dom::closeSymbol then nCloses := nCloses + 1;
	    else return(FALSE);
	    end_if;
	    if nOpens < nCloses then return(FALSE); end_if;
	end_for;
	case args(0) // nCloses <= nOpens
	    of 1 do return(TRUE); 
	    of 2 do return(bool(nOpens = k1));
	    otherwise return(bool((nOpens = k1) and (nCloses = k2)));
	end_case;
    end_proc;

    isA := 
    proc(obj : Type::AnyType, 
	 k1 = FAIL : Type::Union(Type::Integer, DOM_FAIL), 
	 k2 = FAIL : Type::Union(Type::Integer, DOM_FAIL))
	local p, nOpens, nCloses;
    begin
        if args(0) > 3 then error("Wrong number of arguments"); end_if;
        if args(0) = 2 then k2:= k1; end_if;
        if k1 <> FAIL and k1 < k2 then 
            error(expr2text(k1)." must be >= than ".expr2text(k2));
        end_if;
	if not testtype(obj, Type::ListOf(Type::Integer)) then 
	    return(FALSE);
	end_if;
	nOpens := 0; nCloses := 0;
	for p in obj do
	    if   p = dom::openSymbol  then nOpens  := nOpens  + 1;
	    elif p = dom::closeSymbol then nCloses := nCloses + 1;
	    else return(FALSE);
	    end_if;
	    if nOpens < nCloses then return(FALSE); end_if;
	end_for;
	case args(0)
	    of 1 do return(bool(nOpens = nCloses)); 
	    of 2 do return(bool((nOpens = nCloses) and (nOpens = k1)));
	    otherwise return(bool((nOpens = k1) and (nCloses = k2)));
	end_case;
    end_proc;

    size := proc(w : dom) begin nops(w)/2; end_proc;

    //////////////////////////////////////////////////////////////////////////
    // count
    //////////////////////////////////////////////////////////////////////////

    count := combinat::catalan;

    list := 
    proc(k1: Type::NonNegInt, k2=0: Type::NonNegInt) : Type::ListOf(dom::typePrefix)
        option remember;
        local dycks, d, d1, d2, i;
    /*
       Returns the list of all dyck words with k1 "(" and k2 ")". 
       They are listed in the order: ["()()()", ... "((()))"]
       corresponding to the inductive generation.

       This is no longer True, because of the revert(sort())...
       The documentation says that they are listed in lexicographic order 
    */
    begin
        if args(0) > 2 then error("Wrong number of arguments"); end_if;
        if args(0) = 1 then k2:= k1; end_if;
        if k1 < k2 then error(expr2text(k1)." must be >= than ".expr2text(k2)); end_if;
        if k2 = 0 then return([[dom::openSymbol $ k1]]); end_if;
        if k1 = 1 then return([[dom::openSymbol,dom::closeSymbol]]); end_if;
        if k1 = 0 then return([[]]); end_if;
        dycks:=[];
        if k1 > k2 then
            dycks:=dycks.[[dom::openSymbol]. d $ d in dom::list(k1-1,k2)];
        end_if;
        for i from 0 to k2-1 do
            for d1 in dom::list(i,i) do
                for d2 in dom::list(k1-i-1, k2-i-1) do
                    dycks:=dycks.[[dom::openSymbol].d1.[dom::closeSymbol].d2];
                end_for;
            end_for;
        end_for;
        revert(sort(dycks));
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // Description of the grammar
    //////////////////////////////////////////////////////////////////////////

    specification :=
    ([DyckWord = Alias(DyckSequence, DOM_LIST),
      DyckSequence = Union(Epsilon(),
			   Alias(Prod(Atom   (dom::openSymbol),
				      DyckSequence,
				      Epsilon(dom::closeSymbol),
				      DyckSequence),
				 op))
     ]);

    decomposableClassOptions := [Boustrophedonic = FALSE];

    //////////////////////////////////////////////////////////////////////////
    // Methods
    //////////////////////////////////////////////////////////////////////////

    height :=
    // works also for no complete dyck words with other symbols
    // e.g.: Motzkin words
    proc(word: combinat::words)
	local letter, height, heightMax;
    begin
	height    := 0;
	heightMax := 0;
	for letter in word do
	    if letter = dom::openSymbol then
		height := height + 1;
		heightMax := max(height, heightMax);
	    elif letter = dom::closeSymbol then
		height := height - 1;
	    end_if;
	end_for;
	heightMax
    end_proc;
	
    associatedParenthesis :=
    // works also for no complete dyck words with other symbols
    // e.g.: Motzkin words
    proc(word: combinat::words, pos: Type::PosInt) : 
	Type::Union(DOM_FAIL, Type::PosInt)
	local dir, height;
    begin
	if testargs() and pos > nops(word) then
	    error("Error: Invalid index");
	end_if;
	if   word[pos] = dom::openSymbol then 
	    dir := +1; height := +1;
	elif word[pos] = dom::closeSymbol then 
	    dir := -1; height := -1;
	else error("Unknown symbol")
	end_if;
	while (height <> 0) do
	    pos    := pos + dir;
	    if pos = 0 or pos > nops(word) then return(FAIL);
	    end_if;
	    if word[pos] = dom::openSymbol then 
		height := height + 1;
	    elif word[pos] = dom::closeSymbol then 
		height := height - 1;
	    end_if;
	end_while;
	pos;
    end_proc;
	
    //////////////////////////////////////////////////////////////////////////
    // Conversions and bijections
    //////////////////////////////////////////////////////////////////////////

    /// Bijection of Biane from dyck words to non crossing partitions
    //
    // Rfrence: Philippe Biane, Permutations suivant le type
    // d'excedances et le nombre d'inversions et interpretation
    // combinatoire d'une fraction continue de Heine
    // Europ. J. Comb., 14(4): 277-284, 1993
    //
    // Thanks to Mathieu Dutour for describing the bijection
    toNonCrossingPartition :=
    proc(l: dom)
        : combinat::nonCrossingPartitions
        local partition, stack,
              i, j, p, nz;
    begin
        partition := {};
        stack := [];
        i := 1;
        p := 1;
        // Invariants:
        //  - l[i]=0
        //  - p is the number of the opening parenthesis at position i
        while i<=nops(l) do
            stack:=append(stack,p);
            j := i+1;
            while j<=nops(l) and l[j]=dom::closeSymbol do
                j := j+1;
            end_while;
            // Now j points to the next 1, or past the end of l
            nz := j-(i+1); // number of )'s between i and j
            if nz > 0 then
                // Remove the nz last elements of stack, and make a
                // new part in partition.
                if nz>nops(stack) then
                    error("Incorrect dyck word");
                end_if;
                partition := partition 
                            union {[op(stack, nops(stack)-nz+1..nops(stack))]};
                stack := [op(stack, 1..nops(stack)-nz)];
            end_if;
            i := j;
            p := p+1;
        end_while;
        if nops(stack)>0 then
            error("Incorrect dyck word");
        end_if;
        partition;
    end_proc;

    /// Bijection of Biane from non crossing partitions to dyck words
    fromNonCrossingPartition :=
    proc(partition: combinat::nonCrossingPartitions)
        : dom
        local l, v, i;
    begin
        l:=[0 $ _plus(nops(v) $ v in partition)];
        for v in partition do
            l[v[nops(v)]] := nops(v);
        end_for;
        [ (dom::openSymbol, dom::closeSymbol $ i) $ i in l ]:
    end_proc;

    /// conversion from dyck words to well-balanced strings of parenthesis
    toString :=
    proc(word: combinat::words) : DOM_STRING
	local c;
    begin
        _concat(if c=dom::openSymbol then
                    "("
                else
                    ")"
                end_if 
                $ c in word);
    end_proc;

    /// conversion from well-balanced strings of parenthesis to dyck words
    fromString :=
    proc(string: DOM_STRING) : combinat::words
	local i;
    begin
	%if "12"[1]="1" then // work around a change in strings indexed 
	                     // access from MuPAD 3.0.0 on
        [ if string[i]="(" then 
              dom::openSymbol;
          else
              dom::closeSymbol; 
          end_if
         $ i = 1..length(string) ];
	else
	            [ if string[i]="(" then 
              dom::openSymbol;
          else
              dom::closeSymbol; 
          end_if
         $ i = 0..length(string)-1 ];
	end_if;
    end_proc;

    printPretty :=
    proc(word: dom::typePrefix) : output::asciiArt
	local drawing, x, y, letter;
    begin
	drawing := output::asciiArt();
	x:=0; y:=0;
	for letter in word do
	    x := x+1;
	    if letter = dom::openSymbol then
		y := y + 1;
		drawing[x,y]:="/";
	    else
		drawing[x,y]:="\\";
		y := y - 1;
	    end_if;
	end_for;
	output::asciiArt::setAlignment(drawing,Bottom);
    end_proc;

    // For backward compatibility
    prettyPrint := dom::printPretty;

begin
end_domain:

