/*****************************************************************************
  $Id $
  File:		lib/COMBINAT/freeTreesLabelled.mu
  Authors:	Nicolas M. Thiry <nthiery@users.sourceforge.net>
  Purpose:	Source code for the domain combinat::freeTreesLabelled
  Created:	Mon Jan 12 15:54:36 2004
  License:	LGPL
  Project:	MuPAD-Combinat, http://mupad-combinat.sf.net
*****************************************************************************/


alias(intSet = (%if domtype(experimental) = DOM_DOMAIN and
		    experimental::intSetSort <> FAIL then
		    experimental::intSetSort
		else
		    DOM_SET
		end_if)):

domain combinat::freeTreesLabelled
    inherits Dom::BaseDomain;
    category Cat::GradedCombinatorialClass;
    axiom Ax::systemRep;

    interfaceAdd := { hold(fromListPrufer), hold(toListPrufer) };

    info_str := "combinat::freeTreesLabelled -- the combinatorial class of labelled free trees (i.e. connected acyclic graphs)";

    /* Prfer correspondence between lists of n-2 numbers in 1..n,
     * and labelled free trees on n nodes.
     *
     * For each number i in list in turn, we will add an edge
     * joining i to some node j. The rule for choosing j is
     * that j should be the smallest index not appearing in
     * the remainder of list, and not used yet (so no other edge
     * touching j will ever be added).
     *
     * Knuth, the art of computer programming, volume 1, enumeration of trees
     *
     * The implementation below is in O(n^2) using a plain DOM_SET
     * for the set of available nodes, and O(n log(n)) if using a sorted
     * container instead (experimental::intSetSort)
     *
     * I have heard that there exists an O(n) version, but don't have
     * a reference.
     *
     * Some timings on 1GHz Athlon laptop

>> for n from 1 to 2001 step 200 do
       r := random(n):
       l := [ r()+1 $ i = 1..n]:
       fprint(Unquoted, 0, n,"\t",time(combinat::freeTreesLabelled::fromListPrufer(l))/n^2*1.0)
   end_for ;
1	1.0
201	0.001089081953
401	0.0008395470177
601	0.0008416366511
801	0.0008447617756
1001	0.0008932126814
1201	0.0008624508718
1401	0.0009654570178
1601	0.0009433516919
1801	0.00102355376
2001	0.001011488259
*/

    fromListPrufer :=
    proc(list: Type::ListOf(Type::PosInt))
	: Type::ListOf(Type::SetOf(Type::PosInt))
	local n, i, j,
	      edges, nbOccurences, available, p;
    begin
	n := nops(list) + 2;
	if testargs() and nops(list) > 0 and not max(op(list)) <= n then
	    error("The input is not a list of n-2 positive integers bounded by n".expr2text(list))
	end_if;

	p := poly(hold(x));

	nbOccurences := [ 0 $ n ];
	for i in list do
	    nbOccurences[i] := nbOccurences[i] + 1;
	end_for;
        // Invariants:
        // - nbOccurences[i]: number of occurences of i in the
	//   remainder of list
	// - available: nodes with zero occurences that have not
	//   yet been used
	available := intSet(op(select({$1..n}, i -> nbOccurences[i] = 0)));
	edges := [];
	for i in list do
	    j := min(available);
	    edges := append(edges, {i,j});
	    available := available minus {j};
	    if (nbOccurences[i] := nbOccurences[i] - 1) = 0 then
		available := available union {i};
	    end_if;
	end_for;
	assert(nops(available)=2);
	append(edges, expr(available));
    end_proc;

    toListPrufer :=
    proc(edges: Type::ListOf(Type::SetOf(Type::PosInt)))
	: Type::ListOf(Type::PosInt)
	local n, e, i, j,
	      list, neighbours, leaves;
    begin
	n := nops(edges) + 1;

        // Use sets of neighbours instead of list of edges to
        // represent the data structure of the tree
	neighbours := [ {} $ n];
	for e in edges do
	    neighbours[op(e,1)] := neighbours[op(e,1)] union {op(e,2)};
	    neighbours[op(e,2)] := neighbours[op(e,2)] union {op(e,1)};
	end_for;
	if testargs() and contains(map(neighbours, nops), 0) > 0 then
	    error("Not a tree ".expr2text(args));
	end_if;

        // Invariant: leaves contains all the leaves of the tree
	leaves := select({$1..n}, i->nops(neighbours[i]) = 1);
	list := [];
	while nops(list) < n-2 do
	    j := min(op(leaves));
	    leaves := leaves minus {j};
	    i := op(neighbours[j], 1);
	    list := append(list,i);
	    // delete the edge from the tree
	    // neighbours[j] := {}; // not really needed
	    if nops((neighbours[i] := neighbours[i] minus {j})) = 1 then
		leaves := leaves union {i};
	    end_if;
	end_while;
	list
    end_proc;

end_domain:
