/*****************************************************************************
  $Id: generators.mu,v 1.1.1.6 2005/01/26 13:42:24 nthiery Exp $
  File:		lib/COMBINAT/generators.mu
  Authors:	Nicolas M. Thiry <nthiery@users.sourceforge.net>
  Created:	25/02/2001
  License:	LGPL
*****************************************************************************/

/* TODO:
 - generators::listUniq(g): returns the list of the elements generated
   by g, with duplicates removed
 */

/*****************************************************************************
 A generator is a device that allow for iterating over a large
 quantity of objects without having to produce and store all of them
 at once. This typically is very handy in combinatorics.

 This package provides tools to produce and manipulate generators. The
 convention is that a generator is any MuPAD object g such that the
 function call g() is defined. Each call g() returns returns either a
 new object, or FAIL if no more objects are available.

 Example:
 >> g:=combinat::generators::fromList([a,a,c]):
 >> g();
					a
 >> g();
					a
 >> g();
					c
 >> g();
                                        FAIL

 The effect of calling again a generator g after the FAIL value has
 been returned is undefined (i.ee.g. possibly disastrous).

 Currently, the easiest way to define a generator in MuPAD is to use
 the so called closure-effect. Here is for example how to construct a
 generator for the sequence 1^2, 2^2, 2^2, ...

 >> power_generator:=
 >> proc()
	local i;
        option escape;
    begin
        i:=0;
	return(proc()
	       begin
		   i:=i+1;
		   return(i^2)
	       end_proc)
    end:
 >> g:=power_generator(2):
 >> g();
					1
 >> g();
					4
 >> g();
					9

 What happens here is that the variable i in the procedure g returned
 by power_generator is lexically bound to the context of
 power_generator. Hence, it is not reseted inbetween calls of g.

 The same generator can be also constructed alternatively as follow:
 >> g:=combinat::generators::fromNext(1, n->n+1);
 >> h:=combinat::generators::map(g, n->n^2);

 Generators cannot be constructed with most MuPAD elements, since
 there is no reference effect. The only exceptions to this rule are
 closures, domains (not domains elements!), and objects defined in
 dynamic modules.

 Generators have a reference effect! Assume g is defined as above:
 >> g();
					1
 >> h:=g: h();
					4
 >> g();
					9

 - toList(generator)		list of the elements generated by generator
 - count(generator)		number of elements generated by generator

 - empty			the trivial generator
 - repeater(value, n)		generator repeating n times value
 - repeater(value)		generator repeating value forever
 - fromRange(i..j)		generator for the elements of the range i..j
 - fromList(list)		generator for elements of list
 - fromNext(first, next)	generator from first value and next operation
 - compose(f, g)		composition of the generator g with the function f
 - pipe(g1, g2)			pipe of the generators g1, g2, ...
 - concat(g1, g2, ...)		concatenation of the generators g1, g2, ...
 - cartesian(g1, g2, ...)	cartesian product of the generators g1, g2, ...
				(inelegant implementation in MuPAD 2.0.0)
 - select(g, predicate <,opts>) generator for the elements e generated by g
                                satisfying predicate(e, opts)
*****************************************************************************/

domain combinat::generators
    inherits Dom::BaseDomain;
    category Cat::BaseCategory;
    axiom Ax::systemRep;

    info_str := "Library 'combinat::generators': utilities for generators";

    interfaceAdd := {hold(copy),
		  hold(toList), hold(count), hold(accumulate),
                  hold(empty), hold(repeater),
		  hold(fromList), hold(fromNext),
		  hold(fromRange), hold(fromUnrank),
                  hold(concat), hold(map), hold(compose), hold(select),
                  hold(pipe),
		  hold(cartesian), hold(cartesianPower),
		  hold(multiset), hold(powerset)
		};

    expr := id;
    convert := id;

    typeFunction := Type::AnyType;
    testtype  := (x) -> testtype(x, Type::AnyType);

    typeConstructor := Type::Function;

    copy :=
    %if version()=[2,0,0] then
	proc()
	begin
	    error("Copy of generator does not work with MuPAD 2.0.0");
	end_proc:
    else
	proc(g: dom) : dom
	begin
	    if domtype(g) = DOM_FUNC_ENV and g::copy <> FAIL then
		g::copy(g);
	    elif domtype(g) = DOM_PROC then
		copyClosure(g);
	    else
		g
	    end_if;
	end_proc;
    end_if;
    generatorWithCopy :=
    proc(generator: DOM_PROC, copy: DOM_PROC)
    begin
	/* Note: at first, generator and copy have different
	   DOM_PROC_ENV's (which should refer to the same lexical
	   context. When I do the copy, I make a copyClosure of
	   g, and assign the DOM_PROC_ENV of g to copy. This seems to
	   work. But is it safe?
	 */
	funcenv
	(generator,
	 NIL,
	 table("copyState" = copy,
	       "copy" =
	       proc(x)
		   name generatorWithCopy::copy;
		   local generator, environment;
	       begin
		   generator := copyClosure(extop(x,1));
		   environment := extop(generator,12);
		   x := subsop(x,
			       1=generator,
			       3=map(extop(x,3),
				     f -> if domtype(f) = DOM_PROC then
					      extsubsop(f, 12=environment)
					  else
					      f
					  end_if)
			      );
		   x::copyState();
		   x;
	       end_proc));
    end_proc;

    checkCopyable := 
    proc(generator : dom, nLoop = 10) : DOM_BOOL
	local res, i, g, g1;
    begin
	g := dom::copy(generator);
	if (res := g()) <> generator() then return(FALSE); end_if;
	i := 0;
	while (i < nLoop) and res <> FAIL do
	    g1 := g;
	    g := dom::copy(g);
	    g1(); // should not affect g
	    res := g();
	    if res  <> generator() then return(FALSE); end_if;
	    i := i+1;
	end_while;
	TRUE;
    end_proc;
    
    empty := () -> FAIL;

    singleton :=
    proc()
	option escape;
	local firstTime, object;
    begin
	object := args();
	firstTime := TRUE;
	proc()
	    name singletonGenerator;
	    option noDebug;
	begin
	    if firstTime then
		firstTime := FALSE;
		object
	    else
		FAIL
	    end_if;
	end_proc;
    end_proc;

    repeater :=
    proc(value: Type::AnyType, n=FAIL: Type::Union(DOM_FAIL,Type::NonNegInt)) : dom
        option escape;
    begin
        if args(0) = 1 then
	    // Returning value directly is not safe in case
	    // value is a function!
	    ()->value;
        else
	    proc()
		name repeaterGenerator;
		option noDebug;
	    begin
		if n>0 then
		    n:=n-1;
		    return(value);
		else
		    return(FAIL);
		end_if;
	    end_proc
        end_if;
    end_proc;

    /**
       fromRange(range)		Generator iterating over range
    **/
    fromRange :=
    proc(range: "_range") : dom
        local i;
        option escape, noDebug;
    begin
        i:=op(range,1)-1;	// Avoids using a first_time
	return(proc(): Type::Union(Type::Integer, DOM_FAIL)
		   name fromRangeGenerator;
		   option noDebug;
               begin
                   i:=i+1;
                   if i<=op(range,2) then
                       return(i);
                   else
                       return(FAIL);
                   end_if;
               end_proc
              );
    end_proc;

    /** fromList(values)	Generator iterating over the elements of values
    **/
    fromList :=
    proc(list: Type::Union(DOM_LIST, DOM_SET)) : dom
        local index;
        option escape;
    begin
        index:=0;
	return(proc()
		   name fromListGenerator;
		   option noDebug;
               begin
                   index:=index+1;
                   if index<=nops(list) then
                       return(op(list,index));
                   else
                       return(FAIL);
                   end_if;
               end_proc
              );
    end_proc;

    /** toList(generator)	List of the elements generated by generator
    **/
    
    toList := dom::toList_doubleSize;
    
    toList_append := // faster for short lists (length < 1000)
    proc(generator: dom) : DOM_LIST
        local result, value;
    begin
	result := [];
	while((value:=generator())<>FAIL) do
	    result := append(result, value);
	end_while;
	result;
    end_proc;
    
    toList_doubleSize := // faster for list of length > 1000
    proc(generator: dom, len) : DOM_LIST
        local result, value, curLen, i, j;
	//option noDebug;
    begin
	if args(0) = 1 then
	    curLen := 16;
	    result := [FAIL $ curLen];
	    i   := 0;
	    while((value:=generator())<>FAIL) do
		if (i := i+1) = curLen then
		    result := append(result, FAIL $ curLen);
		    curLen := curLen + curLen;
		end_if;
		result[i]:=value;
	    end_while;
	    result := [result[j] $ j=1..i];
	else
	    result := [generator()$i=1..len];
	    assert(contains(result, FAIL) = 0);
	    assert(generator() = FAIL);
	end_if;
	result;
    end_proc;

    // add the toList_Table from Walter

    // tentative toList_table by Florent seems to be slow
    toList_Table :=
    proc(generator: dom) : DOM_LIST
	local result, i, value;
    begin
	i := 1;
	result := table();
	while (value:=generator()) <> FAIL do
	    result[i] := value;
	    i := i+1;
	end_while;
	map([op(result)], op, 2);
    end_proc;

    toList_better2 := // slightly faster but tricky
    proc(generator: dom, len) : DOM_LIST
        local result, value, curLen, errcode, i, j;
    begin
	if args(0) = 1 then
	    curLen := 64;
	    i   := 1;
	    result:=[FAIL $ curLen]; 
	    repeat
	// the trick is that there is no test in this loop,
	// the test is done by the kernel during the sysassign.	   
		errcode :=  traperror((while ((value:=generator())<>FAIL) do
					   sysassign(result[i], value);
					   i := i+1;
				       end_while));
		if  errcode = 1017 then // Invalid index
		    result := append(result, FAIL $ curLen);
		    curLen := curLen + curLen;
		    result[i]:=value;
		    i := i+1;
		elif errcode <> 0 then lasterror();
		end_if;
	    until value = FAIL end_repeat;
	    result := [result[j] $ j=1..i-1];
	    assert(contains(result, FAIL) = 0);
	else
	    result := [generator()$i=1..len];
	    assert(contains(result, FAIL) = 0);
	    assert(generator() = FAIL);
	end_if;
	result;
    end_proc;

    /** count(generator) Count the number of elements generated by generator
    **/
    count :=
    proc(generator: dom) : Type::NonNegInt
        local n;
	option noDebug;
    begin
        n:=0;
        while(generator()<>FAIL) do
            n:=n+1;
        end_while;
        return(n);
    end_proc;

    /** accumulate(generator)
	_plus(0, x $ x in g):                        accumulate(g)
	_mult(1, x $ x in g):                        accumulate(g, 1, _mult)
	_plus(0, f(x) $ x in g):                     accumulate(map(g,f))
	_plus(0, f(x, ...) $ x in g):                accumulate(map(g,f,...))
	r := 0;	for x in g do r := f(x, r, ...) end: accumulate(g, 0, f)
    **/
    accumulate :=
    proc(generator: dom,
	 start = 0,
	 plus = _plus : dom::typeFunction) : Type::AnyType
        local value, options;
	option noDebug;
    begin
	options := args(4..args(0));
        while((value := generator())<>FAIL) do
	    start := plus(start, value, options);
        end_while;
        return(start);
    end_proc;

    /** consume(generator) Consumes the generator
    **/
    consume :=
    proc(generator: dom) : DOM_NULL
	option noDebug;
    begin
        while(generator()<>FAIL) do
        end_while;
	null();
    end_proc;

    /** concat(generator1, generator2, ...)

        The concatenation of the generators generator1, generator2, ...
        That is the generator returning successively the elements
        generated by generator1, generator2 and so on.
    **/
    concat :=
    proc()
        local gens, i;
        option escape, noDebug;
    begin
        gens := [args()];
        if nops(gens)=0 then
            return(dom::empty);
        end_if;
        i:=1; // invariant: gens[i] is a valid generator
        dom::generatorWithCopy
	(proc()
	     local x;
	     option noDebug;
	     name concatGenerator;
	 begin
	     while (x:=gens[i]()) = FAIL do
		 gens[i] := FAIL; // discard the terminated generator.
		 i:=i+1;
		 if i>nops(gens) then
		     return(FAIL);
		 end_if;
	     end_while;
	     x;
	 end_proc,
	 proc()
	     name concatGenerator::copy;
             local k;
	 begin
	     for k from i to nops(gens) do
		 gens[k] := combinat::generators::copy(gens[k]);
	     end_for;
	 end_proc)
    end_proc;

    /** map(g, f)	Map a function f on the elements generated by g

	A generator for f(x_1), f(x_2), ..., where x_1, x_2, ...
        are the elements generated by generator.
    **/
    map :=
    proc(generator: dom, f: dom::typeFunction)
        local options;
        option escape, noDebug;
    begin
        options := args(3..args(0));
	dom::generatorWithCopy
	(proc()
	     name mapGenerator;
	     option noDebug;
             local x;
	 begin
	     if (x:=generator()) = FAIL then
		 generator := FAIL; // discard the terminated generator.
		 FAIL
	     else
		 f(x, options)
	     end_if;
	 end_proc,
	 proc()
	     name mapGenerator::copy;
	 begin
	     generator := combinat::generators::copy(generator);
	 end_proc)
    end_proc;

    /** compose(f, g)	Composition of the generator g by the function f

	A generator for f(x_1), f(x_2), ..., where x_1, x_2, ...
        are the elements generated by generator.
	deprecated
    **/
    compose :=
    proc(f: Type::Function, generator: dom)
	option noDebug;
    begin
	dom::map(generator, f, args(3..args(0)));
    end_proc;

    /**	   pipe(generator, generator_constructor <, options>)

	Concatenation of the elements generated successively by
	the generators generator_constructor(generator(), args);
    **/
    pipe :=
    proc(generator: dom, generator_constructor: dom)
        local options, gen;
        option escape, noDebug;
    begin
        options := args(3..args(0));
        gen := generator();
        if gen=FAIL then
            return(dom::empty);
        end_if;
        gen := generator_constructor(gen, options);
        
	dom::generatorWithCopy
	(proc()
	     name pipeGenerator;
             local x;
	     option noDebug;
	 begin
	     while TRUE do
		 if (x:=gen()) <> FAIL then
		     return(x)
		 end_if;
		 if (gen:=generator()) = FAIL then
		     generator := FAIL; // discard the terminated generator.
		     return(FAIL);
		 end_if;
		 gen := generator_constructor(gen, options)
	     end_while;
	 end_proc,
	 proc()
	     name pipeGenerator::copy;
	 begin
	     gen       := combinat::generators::copy(gen);
	     generator := combinat::generators::copy(generator);
	 end_proc);
    end_proc;

    fromNext :=
    proc(first, Next: Type::Function)
        local first_time, arguments;
        option escape;
    begin
        first_time := TRUE;
        arguments := args(3..args(0));
        proc()
	    option noDebug;
	    name fromNextGenerator;
        begin
            if first_time then
                first_time:=FALSE;
            else
                first:=Next(first, arguments);
            end_if;
            return(first);
        end_proc;
    end_proc;

    fromUnrank :=
    proc(unrankFun : Type::Function)
	local r, options;
	option escape, noDebug;
    begin
	r := 0;
	options := args(2..args(0));
        () -> (r:=r+1; unrankFun(r, options));
    end_proc;

    cartesian :=
     if domtype(copyClosure) = DOM_FUNC_ENV then
//    if FALSE then 
              // copyClosure does _not_ copy recursively. Embeded generators
              // are not copied. 
	proc()
            option escape, noDebug;
            local generators,
                  gen_copies,
                  current,
                  first_time,
                  n;
        begin
            n := args(0);
            generators := [args()];
            gen_copies := map(generators, combinat::generators::copy);
            first_time := TRUE;
	    dom::generatorWithCopy(
            proc()
		name cartesianGenerator;
		option noDebug;
                local i,j;
            begin
                if first_time then
                    current := [ FAIL $ n ];
                    for i from 1 to n do
                        if (current[i]:=generators[i]()) = FAIL then
			    generators := FAIL; gen_copies := FAIL; // discard the terminated generators. 
                            return(FAIL);
                        end_if;
                    end_for;
                    first_time := FALSE;
                    return(current);
                else
		    // Invariants:
                    //  - n is the number of generators
	            //  - gen_copies contains a copy of the generators in
	            //	  their original state
	            //  - generators contains the generators in their 
	            //	  current state
	            //  - The generators are non empty
	            //  - current contains the last generated element
                    for i from n downto 1 do
                        if (current[i]:=generators[i]()) <> FAIL then
                            for j from i+1 to n do
                                generators[j]:=combinat::generators::copy(gen_copies[j]);
                                current[j] := generators[j]();
                            end_for;
                            return(current);
                        end_if;
                    end_for;
		    generators := FAIL; gen_copies := FAIL; // discard the terminated generators. 
                    return(FAIL);
                end_if;
            end_proc,
	    proc()
		name cartesianGenerator::copy;
                local i;
	    begin
		for i from 1 to n do
		    generators[i] := combinat::generators::copy(generators[i]);
		end_for;
	    end_proc); // generatorWithCopy
	end_proc;
// Old implementation that does not use a first_time...
//         proc()
//             option escape;
//             local gen_copies, 
//                   generators, 
//                   current,
//                   i;
//         begin
//             generators := [args()];
//             gen_copies := map(generators, copyClosure);
//             current := [ FAIL $ args(0) ];
//             for i from 1 to nops(generators)-1 do
//                 if (current[i]:=generators[i]()) = FAIL then
//                     return(dom::empty);
//                 end_if;
//             end_for;
// 	    // Invariants (n is the number of generators):
//             //  - The n-1 first terms of current are "valid"
//             //  - gen_copies contains a copy of the generators in their
//             //	  original state
//             //  - generators contains the generators in their 
//             //	  current state
//             //  - The first n-1 generators are non empty
//             proc()
//             begin
//                 for i from nops(generators) downto 1 do
//                     if current[i]:=generators[i]() <> FAIL then
//                         for j from i+1 to nops(generators) do
//                             generators[j]:=copyClosure(gen_copies[j]);
//                             current[j] := generators[j]();
//                         end_for;
//                         if current[nops(current)] = FAIL then
//                             return(FAIL);
//                         end_if;
//                         return(current);
//                     end_if;
//                 end_for;
//                 return(FAIL);
//             end_proc;
//         end_proc;
    else
	proc(/* sequence of dom */ ) 
            // Temporary inelegant implementation when copyClosure
	    // is not available (mupad 2.0.0)
	    local l;
	    option noDebug;
        begin
            combinat::cartesianProduct::generator
            (dom::toList(l) $ l in [args()]);
        end_proc;
    end_if;

    cartesianPower :=
    proc(g: dom, k: Type::NonNegInt) : dom
	local i;
    begin
	if k=0 then
	    return(dom::singleton([]));
	end_if;
	dom::cartesian(g, dom::copy(g) $ i=1..k-1) // don't remove the i!
    end_proc;

    multiset :=
    proc(g: dom, k: Type::NonNegInt, shift=0: Type::NonNegInt)
	option escape, noDebug;
	local firstTime, generators, multiset, x, count, ranks, i;
    begin
	// shift means that the rank of an element in the list
	// should be at least the rank of the previous element + shift
	// Well, there are essentially only two useful cases:
	//  - shift=0: multisets
	//  - shift=1: powersets

	// Complexity (C(g): complexity of one iteration of g)
	//  - Constant Time Amortized * C(g) (theoretical)
	//  - O(k*C(g)) time worst case
	// Complexity in memory: O(k*(1+C(g)))
	//  - if shift > 0, one may need to take into account
	//    O(k) arithmetic operations on big numbers
	// Special case: when no elements is produced and shift > 0,
	// there is still a cost of k*C(g) just for detecting it.

	// Treat base and special cases
	if k=0 then
	    return(dom::singleton([]));
	elif k=1 then // Not needed; just an optimization
	    return(dom::map(g, DOM_LIST));
	elif (x := g()) = FAIL then
	    return(dom::empty);
	end_if;

	// We use a table instead of a list to avoid flattening, just in case
	// one of the generator returns an expression sequence
	// The overhead of table indexed accesses should be negligible
	multiset   := table(i=NIL $ i=1..k);
	generators := [ g, NIL $ k-1];
	if shift>0 then
	    ranks  := [ 0 $ k ];
	end_if;
	// Invariants:
	// - firstTime is true or:
	//   - multiset contains the previously generated element
	//   - generators contains the corresponding generators ready to give the next element
        //   - ranks[i] returns the rank of the current i-th element of multiset
	//     we do not need to maintain it when shift=0
	//   - ranks[i] + shift <= ranks[i+1]
	//   - once the generator for the last element has been exhausted once, count
	//     contains the number of elements in g, and ranks[k] <= count. Then,
	//     ranks[i] + (k-i) * shift <= count
	//
	firstTime := TRUE;
        prog::bless(
	proc()
	    name multisetGenerator;
	    option noDebug;
	    local i, j, l;
	begin
	    if not firstTime then
		// look for the position where we can increase an element
		for i from k downto 1 do
		    g := generators[i];
		    if shift > 0 and i < k and ranks[i] + 1 + (k-i) * shift > count then
			next;
		    end_if;
		    if (x := g()) <> FAIL then
			break
		    else
			if shift > 0 then
			    assert(i=k);
			    count := ranks[k];
			end_if;
		    end_if;
		end_for;
		if i=0 then
		    return(FAIL);
		end_if;
	    else
		firstTime := FALSE;
		i := 1;
	    end_if;
	    // Now we need to fill multiset between i and n
	    // and set properly the corresponding generators
	    multiset[i] := x;
	    if shift > 0 then
		ranks   [i] := ranks[i] + 1;
	    end_if;
	    for j from i+1 to k do
		g := dom::copy(g);
		for l from 1 to shift do
		    if (x := g()) = FAIL then
			// That should never happen, except for the first time
			// when g produces less than k elements.
			assert(multiset[j]=NIL);
			return(FAIL)
		    end_if;
		end_for;
		multiset  [j] := x;
		generators[j] := g;
		if shift > 0 then
		    ranks[j]    := ranks[j-1] + shift;
		end_if;
	    end_for;
	    // Transform back the table into a list
	    return([multiset[i] $ i=1..k]);
	end_proc, dom)
    end_proc;

    powerset :=
    proc(g: dom, k: Type::NonNegInt)
    begin
	dom::multiset(g, k, 1);
    end_proc;

    /** select(g, predicate <,opts>)	selects the elements generated by g satisfying predicate

	A generator for the elements e generated by g satisfying
	predicate(e, opts).
    **/
    select := // Don't use Type::Function because they disallow TRUE for example
    proc(generator: dom, predicate: dom::typeFunction)
        option escape, noDebug;
        local options;
    begin
	if predicate = TRUE then // Optimization
	    return(generator);
	elif predicate = FALSE then
	    return(dom::empty);
	end_if;
        options := args(3..args(0));
	dom::generatorWithCopy
	(proc()
	     name selectGenerator;
	     option noDebug;
             local x;
	 begin
	     while (x:=generator()) <> FAIL do
		 if predicate(x, options) then
		     return(x)
		 end_if;
	     end_while;
	     generator := FAIL; // discard the terminated generator. 
	     return(FAIL);
	 end_proc,
	 proc()
	     name selectGenerator::copy;
	 begin
	     generator := combinat::generators::copy(generator);
	 end_proc);

    end_proc;

end_domain:
