/*****************************************************************************
  $Id: linearExtensions.mu,v 1.1.1.6 2005/01/26 13:42:24 nthiery Exp $
  File:		lib/COMBINAT/linearExtensions.mu
  Authors:	Nicolas M. Thiry <nthiery@users.sourceforge.net>
  Created:	25/02/2001
  License:	LGPL
*****************************************************************************/

/*************************************************************/
/* lexicographic enumeration of linear extensions of a graph */
/*************************************************************/
/* The algorithm goes as follows:

TODO

During the algorithm the following loop invariant holds
 - The triple (placeds, minimums, remaining) is a set partition of the 
   vertices of the graph;
 - the element of placeds form a lower ideal of the graphs, they are listed in 
   an order compatible with the graph;
 - inDegree is the table of the number of predecessor in the subgraph
   with vertices (minimums, remaining);
 - minimums is the ordered list of the minimums of the subgraph, that is the 
   orderd lists of vertices v such that inDegree[v]=0. 
It is implemented by the following asserts:
       assert({op(vertices)} = {op(placeds), op(minimums), op(remainings)});
       assert({op(minimums)} = map({select(inDegree, z->(op(z,2)=0))}, op, 1));
       assert(remaining = map({select(inDegree, z->(op(z,2)<>0))}, op, 1));
******************************************************************************/

domain combinat::linearExtensions
    inherits Dom::BaseDomain;
    category Cat::CombinatorialClass;
    axiom    Ax::systemRep;

    info_str := "Library 'combinat::linearExtensions': utilities for linearExtensions of graphs";

    interfaceAdd :=
        {hold(first), hold(Next)
         };

    domtype   := DOM_LIST;
    typeOrder := Type::Union(DOM_PROC, DOM_FUNC_ENV);
    typeGraph := Type::Intersection(Graph, Type::Predicate(Graph::isDirected));

    isA := 
    proc(obj          : Type::AnyType,
	 graph = FAIL : Type::Union(DOM_FAIL, dom::typeGraph)) : DOM_BOOL
	local vertices, edges, ed, pSrc, pDst;
    begin
	if   domtype(obj) <> DOM_LIST then FALSE
	elif args(0) > 1 then
	    vertices := Graph::getVertices(graph);
	    if {op(vertices)} <> {op(obj)} then return(FALSE) end_if;
	    edges    := Graph::getEdges(graph);
	    for ed in edges do
		pSrc := contains(obj, ed[1]);
		pDst := contains(obj, ed[2]);
		if _lazy_or(pSrc = 0, pDst = 0, pSrc >= pDst) then
		    return(FALSE);
		end_if;
	    end_for;
	end_if;
	TRUE;
    end_proc;

    first :=
    proc(graph    : dom::typeGraph, 
	 order    = sysorder : dom::typeOrder)
    // Compute the smallest lexicographical linear extension w.r.t. the 
    // lexcicographical order based on the given order.
    local vertices, 
	  inDegree,       
	  placeds,    // vertices that are already placeds 
	  minimums,   // vertices that can be placeds
	  remainings; // remainings vertices
    begin
	vertices      := Graph::getVertices(graph);
	inDegree      := Graph::inDegree(graph);
	[minimums, remainings, placeds] := split(vertices, z -> (inDegree[z]=0));
	// placeds is set to the empty lists []
	minimums      := sort(minimums, order);
// check for the precondition
//assert({op(vertices)} = {op(placeds), op(minimums), op(remainings)});
//assert({op(minimums)} = map({select(inDegree, z->(op(z,2)=0))}, op, 1));
//assert(remaining = map({select(inDegree, z->(op(z,2)<>0))}, op, 1));
	dom::first_internal(graph, order, inDegree, 
			    placeds, minimums, 1, {op(remainings)});
    end_proc;


    first_internal := 
/* Compute the smallest linear extension L of graph w.r.t. the lexorder based 
on order such that:
 - L is of the form [op(placed), minimums[imin], ...]
*/
    proc(graph, order, 
	 inDegree,   //  
	 placeds,    // vertices that are already placeds
	 minimums,   // ordered list of vertices that can be placeds   
	 imin, 
	 remainings) // set of remainings vertices          
    local vertices, edgesLeaving,
	  newMins,   // new minimums
	  v, vChild, // current working vertex and child
	  c;  
    begin
	vertices      := Graph::getVertices(graph);
	edgesLeaving  := Graph::getEdgesLeaving(graph);
	
	while minimums <> [] do
// check for the loop invariant
//assert({op(vertices)} = {op(placeds), op(minimums), op(remainings)});
//assert({op(minimums)} = map({select(inDegree, z->(op(z,2)=0))}, op, 1));
//assert(remaining = map({select(inDegree, z->(op(z,2)<>0))}, op, 1));
	   
	    v        := minimums[imin];
	    delete(minimums[imin]):
	    vChild   := edgesLeaving[v];
	    newMins  := [];
	    for c in vChild do 
		inDegree[c] := inDegree[c] - 1;
		if inDegree[c] = 0 then
		    delete(inDegree[c]);
		    newMins := append(newMins,c);
		end_if;
	    end_for;
	    placeds    := append(placeds, v);
	    minimums  := listlib::merge(minimums, 
				       sort(newMins, order), order);
	    remainings := remainings minus {op(newMins)};
	    imin := 1; // for now we search the smallest linear extension.
	end_while;
	if remainings = {} then placeds else FAIL end_if;
    end_proc;


    Next := 
    // Compute the next linear extension of a graph. 
    proc(linExt   : dom,
	 graph    : dom::typeGraph, 
	 order    = sysorder : dom::typeOrder)
    local vertices, edgesLeaving, inDegree,
	  placeds,    // vertices that are already placeds 
	  minimums,   // vertices that can be placeds
	  remainings, // remainings vertices
	  v, vChild,  // current working vertex and child
	  c, pc, cc;  
    begin
	vertices     := Graph::getVertices(graph);
	edgesLeaving := Graph::getEdgesLeaving(graph);
	inDegree     := table();
	placeds      := linExt;
	minimums     := [];
	remainings   := {};
	// backtracking of last placeds elements
	for c from nops(placeds) downto 1 do
// check for the loop invariant
//assert({op(vertices)} = {op(placeds), op(minimums), op(remainings)});
//assert({op(minimums)} = map({select(inDegree, z->(op(z,2)=0))}, op, 1));
//assert(remaining = map({select(inDegree, z->(op(z,2)<>0))}, op, 1));
	
	    v           := placeds[c];
	    inDegree[v] := 0;
	    vChild     := edgesLeaving[v];
	    for cc in vChild do 
		inDegree[cc] := inDegree[cc]+1; 
	    end_for;
	    placeds    := [op(placeds, 1..nops(placeds)-1)];
	    minimums   := listlib::setDifference(minimums, vChild);
	    minimums   := sort(append(minimums, v), order);
	    remainings := remainings union {op(vChild)};
	    pc         :=  contains(minimums, v);
	    if pc < nops(minimums) then
		return(dom::first_internal(graph, order, inDegree, 
					   placeds, minimums, pc+1, remainings));
	    end_if;
	end_for;
	FAIL;    
    end_proc;


    generator :=
    proc(graph    : dom::typeGraph, 
	 order    = sysorder : dom::typeOrder)
        : combinat::generators
    begin
	combinat::generators::fromNext(dom::first(graph, order),
				       dom::Next, graph, order);
    end_proc;

end_domain:

