/*****************************************************************************
  $Id: multisetsNK.mu,v 1.1.1.2 2005/01/26 13:42:24 nthiery Exp $
  File:		lib/COMBINAT/multisetsNK.mu
  Authors:	Xavier Molinero <molinero@lsi.upc.es>
  Created:	03/02/2004
  License:	LGPL
*****************************************************************************/

domain combinat::multisetsNK
    inherits Dom::BaseDomain;
    category Cat::CombinatorialClass;
    axiom    Ax::systemRep;

    info_str := "Library 'combinat::multisetsNK': utilities for choosing k numbers with repetitions in the set {1..n}";

    isA  := 
    proc(obj : Type::AnyType, 
	 n = null()  : Type::Union(DOM_NULL, Type::NonNegInt), 
	 k = null()  : Type::Union(DOM_NULL, Type::NonNegInt))
	local maxi;
    begin
	if not testtype(obj, Type::ListOf(Type::PosInt))
	   then return(FALSE)
	end_if;
	maxi := max(0,op(obj)); // 0 to make this work with the empty subset
	if args(0) > 1 and maxi > n then return(FALSE) end_if;
	if args(0) > 2 and nops(obj) <> k then return(FALSE) end_if;
	return(TRUE);
    end_proc;

    count := 
    proc(n: Type::NonNegInt, k: Type::NonNegInt)
    begin
	combinat::binomialRecursive(n+k-1, k);
    end_proc;

    ////////////////////////////////////////////////////
    // generator

    ////////////////////////////////////////////////////
    // Floyd's algorithm whose complexity is O(k)
    random :=
    proc(n : Type::NonNegInt, k : Type::NonNegInt)
	local i, rnd, lRes;
    begin
        for i from 1 to k do
            rnd    := random(n);
            lRes[i] := rnd() + 1;
        end_for;
        [ lRes[i] $ i=1..k ];
    end_proc;

    // complexity: O(k*log(n))
    // It uses a dichotomic search and the Lexicographic order.
    // The rank starts from 1
    unrank :=
    proc(r: Type::PosInt, n: Type::NonNegInt, k: Type::NonNegInt)
         local rr, ii, kk, res;
    begin
        if (r > combinat::binomialRecursive(n+k-1,k)) then
          return(FAIL)
        end_if;
        rr    := r;
        ii    := 0;
        res[0]:= 1:
        for kk from k downto 1 do
            ii     := ii+1;
            // Values r, n, k have to be reorganized for each iteration
            res[ii]:= combinat::multisetsNK::DichSearch(
	               rr, n-res[ii-1]+1, kk, 1, n-res[ii-1]+1);
            rr     := rr - combinat::multisetsNK::sumSearch(
	                    kk-1, n-res[ii]+2-res[ii-1]+1, n-res[ii-1]+1);
            res[ii]:= res[ii] + res[ii-1] - 1
        end_for;
        [ res[kk] $ kk=1..k ]
    end_proc;

    // Partial sum "sum(binomial(ll+k-1,k) $ ll=n0..nf))"
    sumSearch :=
    proc(k: Type::NonNegInt, n0: Type::NonNegInt, nf: Type::NonNegInt)
      : Type::NonNegInt
    begin
        combinat::binomialRecursive(nf+k, k+1) -
        combinat::binomialRecursive(n0+k-1, k+1)
    end_proc;
    // It finds "nm" such that
    // "sumSearch(k,nm+1,nf) >= i >sumSearch(k,nm,nf)",
    // where n0<=nm<=nf 
    DichSearch :=
    proc(i: Type::PosInt, n: Type::NonNegInt, k: Type::NonNegInt,
         n0: Type::NonNegInt, nf: Type::NonNegInt)
        local nm;
    begin
        if ( i = 1 ) then return(1) end_if;
        if (nf = n0) then return(n-n0+1) end_if;
        if (nf-n0 = 1) then
           if i <= combinat::multisetsNK::sumSearch(k-1, nf, n) then return(n-nf+1)
           else return(n-n0+1)
           end_if
        end_if;
        nm:= floor((n0+nf) / 2);
        if i <= combinat::multisetsNK::sumSearch(k-1, nm, n) then
           combinat::multisetsNK::DichSearch(i, n, k, nm, nf)
        else combinat::multisetsNK::DichSearch(i, n, k, n0, nm)
        end_if
    end_proc;

end_domain:
