/*****************************************************************************
  $Id: partitions.mu,v 1.1.1.6 2005/01/26 13:42:24 nthiery Exp $
  File:		lib/COMBINAT/partitions.mu
  Authors:	Nicolas M. Thiry <nthiery@users.sourceforge.net>
                Florent Hivert <florent.hivert@univ-mlv.fr>
		Authors of the combinat package of MuPAD 2.0.0
		Authors of mu-EC
  Created:	16/02/2001
  License:	LGPL
*****************************************************************************/

// Optimization for MuPAD > 2.0.0: we use RD_INF which is faster than infinity
%if traperror(RD_INF+RD_INF) = 0 then
    alias(infinity  = RD_INF);
end_if:

export(output, Alias, asciiArt):

// Hack to work around the bug in combinat::subClass which does
// not accept domains with several direct super categories.
category combinat::partitionsTemporaryHack
    category Cat::GradedCombinatorialClass, 
             Cat::IntegerListsLexClass,
             Cat::CombinatorialClassWith2DBoxedRepresentation;
end_category:

domain combinat::partitions
    inherits Dom::BaseDomain;
    category combinat::partitionsTemporaryHack;
    axiom    Ax::systemRep;

    info_str := "Library 'combinat::partitions': utilities for integer partitions";

//  A partition of n is a decreasing list of positive integers which sum
//  up to n.

//  Partitions are naturally ordered lexicographically.
//  Several functions of this library are directly inherited from
//  Cat::IntegerListsLexClass, and accept generic
//  constraints:

//   - length of the partition:	MinLength = , MaxLength = , Length =
//   - size of parts: 		MinPart = , MaxPart =
//   - slope:			MinSlope = , MaxSlope =
//   - containment:		Inner = , Outer =

//  Note that specifying MinPart=0 allows for 0 parts. This can be
//  useful in conjunction with the option Length.

//  See also Cat::IntegerListsLexClass, Cat::CombinatorialClass.

    interfaceAdd :=
        {hold(conjugate), hold(hooklengths),
	 hold(toExp), hold(fromExp), 
	 hold(centralizerSize),
	 hold(conjugacyClassSize),
	 hold(printPretty), hold(printCompact),
         hold(corners),
         hold(printPrettyCorners), hold(printCompactCorners),
	 hold(dominate),
	 hold(rCore), hold(rQuotient), hold(fromCoreAndQuotient),
	 hold(removeRibbonsAndSpin), hold(addRibbonsAndSpin)
         };


// type and isA are now inherited. 
//////
// but several functions allows trailing zeroes in the partition
    isAWithZeroes := proc(compo : Type::AnyType, summ) 
	       local startOptions;
	   begin
	       if not testtype(compo, Type::ListOf(Type::NonNegInt)) then return(FALSE) end_if;
	       if args(0) > 1 and domtype(summ) = DOM_INT then
		   startOptions := 3;
		   if _plus(op(compo)) <> summ then return(FALSE); end_if;
	       else 
		   startOptions := 2;
	       end_if;
	       combinat::integerListsLexTools::isCorrect
	           (compo, dom::parseOptions(MinPart=0, 
					     args(startOptions..args(0))));
	   end_proc;
    typeWithZeroes := Type::Predicate(expr2text(dom)."WithZeroes", 
				      dom::isAWithZeroes);


    //////////////////////////////////////////////////////////////////////////
    // Pretty-print
    //////////////////////////////////////////////////////////////////////////
    toArray :=
    proc(p: dom) : DOM_ARRAY
	local col, row;
    begin
	p := revert(p);
	// Note the max(1, instead of max(0, to please array
	// See Cat::CombinatorialClassWith2DBoxedRepresentation
	array(1..max(1, nops(p)), 1..max(1, op(p)),
	      (row, col) = ""
	      $ col = 1..p[row]
	      $ row=1..nops(p));
    end_proc;

    toArrayVerticalAlignment := -1; // alignement on the bottom row of the array

    printCompact :=
    proc(p: dom) : output::asciiArt
	local i;
    begin
	output::asciiArt([_concat("#" $ i) $ i in p])
    end_proc;

    // For backward compatibility
    prettyPrint := dom::printPretty;

    //////////////////////////////////////////////////////////////////////////
    // generator & co from with combinat::IntegerListsLexClass
    //////////////////////////////////////////////////////////////////////////

    parseOptions :=
    proc()
    begin
        dom::parseOptionsIntegerListsLexClass
        (table(MinLength = 0,
               MaxLength = infinity,
               Length	 = NIL,
               MinPart	 = 1,
               MaxPart	 = infinity,
               Inner	 = NIL,
               Outer	 = NIL,
	       MinSlope	 = -infinity,
	       MaxSlope	 = 0
              ),
         args()
         ):
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // count
    //////////////////////////////////////////////////////////////////////////

    /**
...
    **/

    count :=
    proc(n: Type::NonNegInt) : Type::NonNegInt
        local k;
    begin
        if args(0) > 1 then
            // There is no efficient counter yet when there are
            // further constraints.
	    return(dom::count_integerListsLexClass(args()));
	end_if;

	if n<100 then // use Euler's pentagonal formula
            // the following avoids stacks problems under MACOS
            for k from n mod 25 to n step 25 do
                dom::count_Euler(k)
            end_for
        else // use Hardy-Ramanujan-Rademacher formula
            dom::count_HRR(n)
        end_if
    end_proc;

    /** number of partitions of n, using Euler's pentagonal formula

      p(n) + sum((-1)^k*(p(n-w(k))+p(n-w(-k))),k=1..infinity) = 0

      where w(k)=(3*k^2+k)/2. The asymptotic expansion of p(n) is:

                             1/2    /    / 2 n \1/2 \
                            3    exp| PI | --- |    |
                                    \    \  3  /    /
                     p(n) ~ -------------------------
                                      12 n

      Reference: G. Andrews, The Theory of Partitions, Addison-Wesley, 1976.
    **/
    count_Euler :=
    proc(n: Type::NonNegInt) : Type::NonNegInt
        local s,t,k;
        option remember;
    begin
        if n=0 then 
            return(1);
        end_if;
        s := 0;
        for k from 1 to n do
            t := (3*k^2-k)/2;
            if n < t then return(abs(s)) end_if;
            s := dom::count_Euler(n-t)-s;
            t := t+k;
            if n < t then return(abs(s)) end_if;
            s := dom::count_Euler(n-t)+s
        end_for;
    end_proc;

    /** Number of partitions of n using Hardy-Ramanujan-Rademacher formula.
	It is more efficient than Euler's pentagonal formula for n>=100.
    **/
    count_HRR := 
    proc(n: Type::NonNegInt) : Type::NonNegInt
        local s,k,t,f,x,X,C,zeroes,Akn,r,rt,ok;
        save DIGITS;
    begin
        // first compute an approximation to set DIGITS
        s:=float(1/(4*n*sqrt(3))*exp(PI*sqrt(2*n/3)));
        s:=ln(s)/ln(10.0);
        DIGITS:=ceil(s+ln(s)+10);
        C:= float(1/PI/sqrt(2));
        X:= genident();
        x:= genident();
        f:= subs( diff(sinh(PI/X*(2/3*(x-1/24))^(1/2))/(x-1/24)^(1/2),x),x=n);

        s:=0; zeroes:=1; r:=0.0; ok:=0;
        k:=1; 
        repeat
            if igcd(zeroes,k)=1 then
                Akn:=dom::HRR_A(k,n);
                if Akn=0 then
                    zeroes:=zeroes*k
                else
                    t:=r+C*Re(float( Akn*k^(1/2)*eval(subs(f,X=k)) ));
                    rt:=round(t); s:=s+rt; r:=t-rt;
                    DIGITS:=ceil(0.4343*ln(abs(t))+10);
                    userinfo(1,"iteration ".k.
                             " last term=".expr2text(rt).
                             " rest=".expr2text(r));
                    if rt=0 and abs(r)<0.1 then 
                        ok:=ok+1 
                    else
                        ok:=0
                    end_if
                end_if
            end_if;
            k:=k+1;
        until ok=3 end_repeat;
        s
    end_proc;

    HRR_A := 
    proc(k,n)
        local s,h;
    begin
        s:=0;
        for h from 0 to k-1 do
            if igcd(h,k)=1 then
                dom::HRR_omega(h,k)/2-n*h/k;
                s:=s+(if type(%)=DOM_INT then 
                          1 
                      else
                          cos(2*PI*mods(op(%))/op(%,2)) 
                      end_if);
            end_if
        end_for;
        s
    end_proc;

    HRR_omega := 
    proc(h,k) 
        local mu;
    begin
        _plus((frac(mu/k)-1/2)*(frac(h*mu/k)-1/2)$mu=1..k-1)
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // Former implementation of first
    //////////////////////////////////////////////////////////////////////////

    first_old := 
    proc(n: Type::NonNegInt,
         k=FAIL: Type::Union(Type::NonNegInt, DOM_FAIL), 
         m=FAIL: Type::Union(Type::NonNegInt, DOM_FAIL))
        : Type::Union(dom, DOM_FAIL)
    begin
        if testargs() and args(0) > 3 then
            error("Wrong number of arguments")
        end_if;
        if n=0 then
            return([]);
        end_if;
        if k=0 then
            return(FAIL);
        end_if;
        if m=FAIL then
            return([n])
        end_if;
        if k <> FAIL and m*k < n then
            return(FAIL);
        end_if;
        return([m $ n div m, n mod m]);
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // Former implementation of Next
    //////////////////////////////////////////////////////////////////////////

    Next_old :=
    proc(p: dom,
         k=FAIL: Type::Union(Type::NonNegInt, DOM_FAIL), 
         m=FAIL: Type::Union(Type::NonNegInt, DOM_FAIL))
        : Type::Union(dom, DOM_FAIL)
        local n, r, index, i;
    begin
        if testargs() and args(0) > 3 then
            error("Wrong number of arguments")
        end_if;

        // Actually, m is never used, since the partition p is
        // supposed to already satisfy the conditions, and the
        // operation only decreases the sizes of the part. 
        // m is recycled as a local variable.
        
        // Algo: find the first index where it's possible to decrease
        // an entry, and redistribute n:=l[index]+l[index+1]+... as
        // the smallest partition of n with parts of size at most
        // l[index]-1
        n:=0;
        for index from nops(p) downto 1 do
            n:=n+p[index];
            if p[index]>1 then
                m := p[index]-1;	// maximal value of the parts
                if k <> FAIL and m*(k-index+1) < n then
                    next;
                end_if;
                r := n mod m;		// size of the last part
                return([p[i] $ i=1..index-1,
                        m $ n div m,
                        (if r>0 then r else null() end_if)]);
            end_if;
        end_for;
        return(FAIL);
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // generator
    //////////////////////////////////////////////////////////////////////////

    generator_old :=
    proc(n: Type::NonNegInt,
         k=FAIL: Type::Union(Type::NonNegInt, DOM_FAIL), 
         m=FAIL: Type::Union(Type::NonNegInt, DOM_FAIL))
        : combinat::generators
    begin
        combinat::generators::fromNext(dom::first_old(args()),
                                       dom::Next_old, args(2..args(0)));
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // methods
    //////////////////////////////////////////////////////////////////////////

    /// The conjugate partition of p
    // Taken from mu-EC
    conjugate :=
    proc(p: dom::typeWithZeroes) : dom
        local l, i;
    begin
        if (p = []) then
            [];
        else
            l := nops(p);
            [l $ p[l], l-i $ p[l-i]-p[l-i+1] $ i=1..l-1];
        end_if;
    end_proc;

    /// The list of the hook lengths of p
    // Should this list be sorted ?
    hooklengths :=
    proc(p: dom) : Type::ListOf(Type::PosInt)
        local pconj, i, j;
    begin
        pconj := dom::conjugate(p);
        [ p[i]-i + pconj[j]-j+1 $ j=1..p[i] $ i=1..nops(p) ];
    end_proc;

    /// Multiplicities of the parts of a partition
    //	Cf. the documentation from PARTPart2Exp.help in mu-EC
    toExp :=
    proc(p: dom, k=0: Type::Union(Type::NonNegInt, DOM_FAIL))
        : Type::ListOf(Type::NonNegInt)
        local a, i;
    begin
	if nops(p)>0 then
	    k := max(k, p[1]);
        end_if;
        a := [0 $ k];
        ( a[i] := a[i]+1 ) $ i in p;
        a;
    end_proc;

    /// partition from the multiplicities of its parts
    //  Cf. the documentation from PARTExp2Part.help
    fromExp :=
    proc(l: Type::ListOf(Type::NonNegInt)) : dom
	local i;
    begin
        revert([ i $ l[i] $ i=1..nops(l) ]);
    end_proc;

    /**
    centralizerSize(p) -> Type::PosInt

    The size of the centralizer of the conjugacy class of the
    symmetric group indexed by the partition p.

    Ref: A.~Kerber, "Algebraic Combinatoric Via Finite Group Action", 1.3 p 24
    **/
    centralizerSize :=
    proc(p: dom) : Type::PosInt
        local a, i;
    begin
        a:=combinat::partitions::toExp(p);
        _mult(i^a[i]*fact(a[i]) $ i=1..nops(a))
    end_proc;

    /**
    combinat::conjugacyClassSize(p) -> Type::PosInt

    The size of the conjugacy class of the symmetric group
    indexed by the partition p.

    Ref: A.~Kerber, "Algebraic Combinatoric Via Finite Group Action", 1.3 p 24
    **/
    conjugacyClassSize :=
    proc(p: dom) : Type::PosInt
    begin
        _plus(op(p))! / dom::centralizerSize(p);
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // corners
    //////////////////////////////////////////////////////////////////////////
 
    corners:=
    // corners of a partitions returns list of lists !!
    // should return list of pairs !!
    proc(p: dom) : Type::ListOf(Type::ListOf(Type::NonNegInt))
        local nblcors, lcors, nn, i, pi;
    begin
        if p = [] then return([]); end_if;
        lcors:= [[1, op(p,1)]];
        nn:= nops(p);
        if nn = 1 then return(lcors); end_if;
        nblcors:= 1;
        for i from 2 to nn do
            pi:= op(p,i);
            if pi = op(p,i-1) then
                lcors[nblcors][1]:= lcors[nblcors][1] + 1;
            else
                lcors:= lcors.[[i,pi]];
                nblcors:= nblcors + 1;
            end_if;
        end_for;
        return(lcors);
    end_proc;
 
    //////////////////////////////////////////////////////////////////////////
    // Pretty-print for corners
    //////////////////////////////////////////////////////////////////////////
 
    printPrettyCorners :=
    proc(p: dom) : output::asciiArt
        local i;
    begin
	combinat::tableaux::printPretty([["" $ i-1, "*"] $ i in revert(p)]);
    end_proc;
 
    printCompactCorners :=
    proc(p: dom) : output::asciiArt
        local i, lasci, lcors;
    begin
        lasci:= [_concat("#" $ i) $ i in p];
        lcors:= dom::corners(p);
        %if "12"[1]="1" then // work around a change in strings indexed
                             // access from MuPAD 3.0.0 on
            (lasci[lcors[i][1]][lcors[i][2]]:= "*";) $ i = 1..nops(lcors);
        else
            (lasci[lcors[i][1]][lcors[i][2]-1]:= "*";) $ i = 1..nops(lcors);
        end_if;
        output::asciiArt(lasci);
    end_proc;

    dominate := // dominance order of partitions
    proc(p1 : dom, p2 : dom)
	local sp1, sp2, i;
    begin
	if testargs() and _plus(op(p1)) <> _plus(op(p2)) then
	    error("Sum mismatch : ".expr2text(p1)." and ".expr2text(p2));
	end_if;
	sp1 := 0;
	sp2 := 0;
	for i from 1 to min(nops(p1), nops(p2)) do
	    sp1 := sp1 + p1[i];
	    sp2 := sp2 + p2[i];
	    if sp1 < sp2 then return(FALSE); end_if;
	end_for;
	TRUE;
    end_proc;
   
//////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////

    /////////////////////////////////////////////////////////////////////
    ///    r-core and r-quotient bijection for partition             ////
    /////////////////////////////////////////////////////////////////////

    //Algorithm in order to find the r-core of a partition
    /////////////////////////////////////////////////////////////////////
 
    rCore := 
    proc (part:combinat::partitions, length: Type::NonNegInt) : combinat::partitions
        local k, i, e;
    begin
        //lenght partition normalization
	if (nops(part) mod length <> 0) then 
	    part := append(part, 0 $ length-irem(nops(part),length));
	end_if;
        //adding canonical vector to the partition 
	part := [part[i] + nops(part)-i $ i=1..nops(part)];
        //Reduced modulo length with succesive representing
	for e from 0 to length-1 do
	    k := e;
	        for i from nops(part) downto 1 do
	            if part[i] mod length = e then
		        part[i] := k; 
		        k       := k+length;
		    end_if;
	        end_for;
	end_for;
	part := revert(sort(part));
        //removing canonical vector after operations
	part := [part[i]-nops(part)+i $ i=1..nops(part)];
        //selecting the r-core
	part := select(part, _not@iszero);
        return(part);
    end_proc;

    //Algorithm in order to find the r-quotient of a partition
    ///////////////////////////////////////////////////////////////////

    rQuotient := 
    proc (part:combinat::partitions, length: Type::NonNegInt) : Type::ListOf(combinat::partitions)
	local k, i, tmp, res, e;
    begin
        //length partition normalization
	if (nops(part) mod length <> 0) then 
	    part := append(part, 0 $ length-irem(nops(part), length));
	end_if;
        //adding canonical vector
	part := [part[i] + nops(part)-i $ i=1..nops(part)];
	res  := [FAIL $ length];
        //reducing vector 
	for e from 0 to length - 1 do
	    k   := e;
	    tmp := [];
	    for i from nops(part) downto 1 do
		if part[i] mod length = e then
		    tmp := append(tmp, (part[i]-k)/length);
		    k   := k + length;
		end_if;
	    end_for;
	    res[e+1] := revert(select(tmp, _not@iszero));
	end_for;
	return(res);
    end_proc;

    //Algorithm in order to find partition from its r-Core and r-Quotient
    /////////////////////////////////////////////////////////////////////

    fromCoreAndQuotient :=  
    proc(core: combinat::partitions, quotient : Type::ListOf(combinat::partitions)) : combinat::partitions
	local len, i, j, k, v, w;
    begin
	len := nops(quotient);
	k   := len*max(ceil(nops(core)/len), 
		       nops(core), op(map(quotient,nops))) + len;
	v   := [core[i]-i+1 $ i=1..nops(core), -i+1 $ i=nops(core)+1..k];
	w   := [select(v, proc(vl) begin bool(vl-i mod len=0) end) $ i=1..len];
	w   := [(op(w,[i,j]) + len*op(quotient,[i,j]) $ j=1..nops(quotient[i]), 
		 op(w,[i,j]) $ j=nops(quotient[i])+1..nops(w[i])
		) $ i=1..len];
	w   := revert(sort(w));
	select([w[i]+i-1 $ i=1..nops(w)], _not@iszero);
    end_proc;

    ////////////////////////////////////////////////////////////////////////////////////
    //    Functions which add/remove ribbon strip to/from a partition and computing
    //    the spin of the corresponding strip
    ////////////////////////////////////////////////////////////////////////////////////

    addRibbonsAndSpin := 
    proc (part: Type::ListOf(Type::NonNegInt), plant: Type::ListOf(Type::NonNegInt))
	local res, resinv, delta, i;
    begin
        part   := [op(part),0$i=1..nops(plant)-nops(part)];
        delta  := revert([i $ i=0..nops(part)-1]);
        part   := zip(part, delta, _plus);
        if nops(part) = 0 then return(FAIL); end_if;
        //planting ribbons to the partitions
	res    := zip(part, plant, _plus);
        if (nops({op(res)}) <> nops(res)) then
	    return(FAIL);
	end_if;
        resinv := res;
	res    := revert(sort(res));
        res    := zip(res,delta,_subtract);
	if (res[nops(res)] < 0 ) 
            then return(FAIL); 
	else 
	    res := select(res, _not@has, 0);  
            return([res,
		    combinat::permutations::inversionsNumber(
                    combinat::words::standard(revert(resinv)))]
                  ); 
        end_if;
    end_proc;   
    
    removeRibbonsAndSpin := 
    proc (part: Type::ListOf(Type::NonNegInt), plant: Type::ListOf(Type::NonNegInt))
	local res, resinv, delta, i;
    begin
        part   := [op(part),0$i=1..nops(plant)-nops(part)];
        delta  := revert([i $ i=0..nops(part)-1]);
        part   := zip(part, delta, _plus);
        if nops(part) = 0 then return(FAIL); end_if;
        //Unplanting ribbons to the partitions
	res    := zip(part, plant, _subtract);
        if (nops({op(res)}) <> nops(res)) then
	    return(FAIL);
	end_if;
        resinv := res;
	res    := revert(sort(res));
        res    := zip(res,delta,_subtract);
	if (res[nops(res)] < 0 ) 
            then return(FAIL); 
	else
	    res := select(res, _not@has, 0);  
	    return([res,
		    combinat::permutations::inversionsNumber(
                    combinat::words::standard(revert(resinv)))]
                  ); 
        end_if;
    end_proc;   

end_domain:

