/*****************************************************************************
  $Id: permutations.mu,v 1.1.1.8 2005/01/26 13:42:24 nthiery Exp $
  File:		lib/COMBINAT/permutations.mu
  Authors:	Nicolas M. Thiry <nthiery@users.sourceforge.net>
  		Florent Hivert <florent.hivert@univ-mlv.fr>
                Xavier Molinero <molinero@lsi.upc.es>
                Vincent Prosper (some things are reused from mu-EC)
  Created:	16/02/2001
  License:	LGPL
*****************************************************************************/

export(combinat, Alias,
       generators, linearExtensions):
alias(permutations=combinat::permutations):

domain combinat::permutations
    inherits combinat::words;
    category Cat::GradedCombinatorialClass;
    axiom    Ax::systemRep;

    info_str := "Library 'combinat::permutations': utilities for permutations";

    /*
 Note: there may be problems with lists of elements that do not have a
 proper internal order in the system kernel. Is it always guaranted
 that equal elements are contiguous in sort(l) ?
    */

    interfaceAdd :=
        {hold(typeStandard), hold(isStandard), hold(isAStandard),
	 hold(first), hold(last), 
	 hold(Next), hold(previous),
	 hold(rank), hold(unrank),
	 hold(random),
	 hold(toMatrix), 
     // Algebraic structure
	 hold(_mult), hold(mult2), hold(inverse), hold(multInverse),
     // Inversion, code and reduced words, 
	 hold(code), hold(coCode), hold(fromCode),
	 hold(inversions), hold(inversionsNumber), hold(sign), hold(isEven),
	 hold(reducedWord), hold(fromReducedWord),
	 hold(reducedWords),
    // Cycles	 
	 hold(fixedPoints), hold(fixedPointsNumber),
	 hold(cycles), hold(cycleType), hold(fromCycles),
    // Bruhat Order
	 hold(leequalBruhat), 
	 hold(bruhatInversions),
	 hold(predBruhat),    hold(succBruhat),
	 hold(smallerBruhat), hold(greaterBruhat),
    // Permutohedron Order
	 hold(leequalPermutohedron), 
	 hold(predPermutohedron),    hold(succPermutohedron),
	 hold(smallerPermutohedron), hold(greaterPermutohedron),
    // Descents
	 hold(majorIndex), hold(majorCode), hold(fromMajorCode),
	 hold(descentPolynomial),
	 hold(descentsToMin), hold(descentsCompositionToMin),
	 hold(descentsToMax), hold(descentsCompositionToMax),
	 hold(fromDescents), hold(fromDescentsComposition),
    // Peaks and saillances
	 hold(peaks), hold(peaksNumber),
	 hold(saillances), hold(saillancesNumber),
    // Patterns
	 hold(hasPattern), hold(patternPositions),
    // Connected Shape and Young Subgroup
//	 hold(connectedShape), hold(isConnected), // => permutations::connected::isA ?
//	 hold(fromConnectedShape), // TODO (Florent) Steal from Connexe
//       hold(connected),          // will be added later...
	 hold(modYoungSubgroup)
	};

    interfaceAddIfExists :=
        {// those are domains and might not exists in subdomains
	 hold(youngSubgroup), hold(fromSetPartition)
	};

    //////////////////////////////////////////////////////////////////////////
    // typeChecking
    //////////////////////////////////////////////////////////////////////////

    typeStandard := Type::Predicate("StandardPermutation", dom::isAStandard);

    typeDirection := Type::Union(Type::Singleton(Right),
				 Type::Singleton(Left));
    isAStandard := 
	(p) -> _lazy_and(
	      testtype(p, Type::ListOf(Type::PosInt)),
	      sort(p)=[$1..nops(p)]);
    isStandard :=
    proc(p: dom)
    begin
	bool(sort(p)=[$1..nops(p)]);
    end_proc;

    isA := 
    proc(obj         : Type::AnyType,
	 word = FAIL : Type::Union(Type::NonNegInt, DOM_LIST, DOM_FAIL)) : DOM_BOOL
    begin
	if domtype(obj) <> DOM_LIST then return(FALSE); end_if;
	if args(0) = 1 then return(TRUE); end_if;
	if domtype(word) = DOM_LIST then
	    return(bool(sort(obj) = sort(word)));
	else
	    return(bool(sort(obj)=[$1..word]));
	end_if;
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // count
    //////////////////////////////////////////////////////////////////////////

    size := proc(p : dom) begin nops(p); end_proc;

    count :=
    proc(n: Type::Union(Type::NonNegInt, DOM_LIST))
        local list, r, s, i;
    begin

        if testargs() then
            if args(0) > 2 then
                error("wrong number of arguments!");
            end_if;
            if args(0) = 2 and args(2) <> Duplicate then
                error("Unknown option '".expr2text(args(2))."'") ;
            end_if;
        end_if;
  
        if domtype(n)=DOM_LIST then
            list:=n;
        else
            return(n!);
        end_if;

        if args(0) = 2 then
            return(nops(list)!);
        else
            // returns n!/i1!i2!...ik!, where i1, i2, ..., ik are the
            // numbers of elements in the list of each given value
            r:=nops(list)!;
            list:=sort(list);
            s:=1;			// number of successive equal elements
            for i from 2 to nops(list) do
                if list[i]=list[i-1] then
                    s:=s+1;
                    r:=r/s;
                else
                    s:=1;
                end_if;
            end_for;
            return(r);
        end_if ;
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // first
    //////////////////////////////////////////////////////////////////////////

    first :=
    proc(l: Type::Union(Type::NonNegInt, DOM_LIST)) : dom
    begin
        if testargs() then
            if args(0) > 2 then
                error("wrong number of arguments!");
            end_if;
            if args(0) = 2 and args(2) <> Duplicate then
                error("Unknown option '".expr2text(args(2))."'") ;
            end_if;
        end_if;

        if domtype(l)=DOM_LIST then
	    return(sort(l));
        else
            return([$1..l]);
        end_if
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // last
    //////////////////////////////////////////////////////////////////////////

    last :=
    proc(l: Type::Union(Type::NonNegInt,DOM_LIST)) : dom
    begin
        revert(dom::first(args()));
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // next
    //////////////////////////////////////////////////////////////////////////

    Next :=
    proc(p: dom)
        : Type::Union(dom, DOM_FAIL)
        local i, j;
    begin
        if testargs() and args(0) > 3 then
            error("Wrong number of arguments")
        end_if;

        // Look for the last strict climb p[i]<p[i+1] in the
        // permutation. Obviously, on the right of i+1, the elements are
        // sorted decreasingly. We look for the last element there
        // which is strictly greater than p[i], and exchange it with
        // p[i]. We also reverse the order of the other elements on the
        // right of p. In the final list q, there is no strict climb
        // between i and i+1, and only weak climbs on the right of i.
        for i from nops(p)-1 downto 1 do
            if p[i]<p[i+1] then
                for j from i+1 to nops(p)-1 do
                    if p[j+1]<=p[i] then break; end_if;
                end_for;
		return([op(p, 1..i-1)].	 [p[j]].
		       revert([op(p, j+1..nops(p))]). [p[i]].
		       revert([op(p, i+1..j-1)]));
            end_if;
        end_for;
        return(FAIL); // There is no climb, so p was the last permutation
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // previous
    //////////////////////////////////////////////////////////////////////////

    previous :=
    proc(p: dom)
        : Type::Union(dom, DOM_FAIL)
        local i, j;
    begin
        if testargs() and args(0) > 3 then
            error("Wrong number of arguments")
        end_if;

        // Look for the first strict descent p[i]>p[i+1] in the
        // permutation. Obviously, on the left of i, the elements are
        // sorted increasingly. We look for the last position j such
        // that p[j]<p[i], and exchange p[i] and p[j]. We
        // also reverse the order the other elements on the right of
        // p. In the final list q, there is no strict descent between
        // i and i+1, and only weak descents on the right of i.
        for i from nops(p)-1 downto 1 do
            if p[i]>p[i+1] then
                for j from i+1 to nops(p)-1 do
                    if p[j+1]>=p[i] then break; end_if;
                end_for;
		return([op(p, 1..i-1)].	 [p[j]].
		       revert([op(p, j+1..nops(p))]). [p[i]].
		       revert([op(p, i+1..j-1)]));
            end_if;
        end_for;
        return(FAIL); // There is no descent, so p was the first permutation
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // generator
    //////////////////////////////////////////////////////////////////////////

    generator :=
    proc(list: Type::Union(Type::NonNegInt,DOM_LIST)) 
        : combinat::generators
        option escape;
        local lindex, i, j, n;
    begin
        if testargs() then
            if args(0) > 2 then
                error("wrong number of arguments!");
            end_if;
            if args(0) = 2 and args(2) <> Duplicate then
                error("Unknown option '".expr2text(args(2))."'") ;
            end_if;
        end_if;
        if domtype(list)=DOM_LIST and 
           ((args(0) = 2) or not testtype(list, Type::ListOf(Type::Real))) then
            // Next and previous cannot deal with duplications, nor
            // with lists of elements that cannot be ordered by <. So,
            // we create a generator for the permutations of a list
            // lindex of integers corresponding the indices of the
            // elements of l.
            list:=sort(list);
            n:=nops(list);
            if args(0) = 2 then
                // With the option duplicate, the integers are simply 1..n
                lindex:=[$1..n];
            else
                // Without the option duplicate, we mimic the
                // repetitions of an element in l by repeating the
                // index of the first element having this value. E.g.
                // [a,b,b,b,c,d,d] -> [1,2,2,2,5,6,6]
                lindex:=[1$n];
                j:=1;
                for i from 2 to n do
                    if list[i-1] <> list[i] then
                        j:=i;
                    end_if;
                    lindex[i]:=j;
                end_for;
            end_if;

            return(combinat::generators::map
                   (combinat::generators::fromNext(lindex, dom::Next),
		    p->[list[i] $ i in p]));
        end_if;
        return(combinat::generators::fromNext(dom::first(args()),
                                              dom::Next));
    end_proc;
  
  
  // Compute the rank of a permutations that is the position of 
  // this permutations in permutations::list 
    rank := 
    proc(p : dom::typeStandard)
	local code, res, nbr, i; 
    begin
	nbr := 1;
	res := 1;
	code := revert(dom::code(p));
	for i from 2 to nops(p) do
	    res := res + nbr*code[i];
	    nbr := nbr*i;
	end_for;
	res;
    end_proc;
  
    unrank := 
    proc(r : Type::NonNegInt, n : Type::NonNegInt)
	local code, i, nbr;
    begin
	nbr := n!;
	if r > n! then return(FAIL) end_if;
	r := r - 1;
	code := [0 $ n];
	for i from n downto 2 do 
	    nbr := nbr div i;
	    code[i] := r div nbr;
	    r := r mod nbr;
	end_for;
	dom::fromCode(revert(code));
    end_proc;
  
    /// random(n) - random permutation of [1..n]
    /// random(l) - random permutation of the list l
    // Floyd's algorithm whose complexity is O(n))
    random :=
    proc(l: Type::Union(Type::NonNegInt,DOM_LIST)) : dom
	local n, i, rnd, r, lRes;
    begin
        if domtype(l)<>DOM_LIST then
	    l:=[$1..l];
        end_if;
        n := nops(l);
        for i from n downto 1 do
            rnd    := random(i);
            r      := rnd() + 1;
            lRes[i]:= l[r];
            if r < i then l[r]:= l[i] end_if
        end_for;
        [ lRes[i] $ i=1..n ];
    end_proc;

    /// random(n) - random permutation of [1..n]
    /// random(l) - random permutation of the list l
    // Cost: O(n*log(n))
    randomSlow :=
    proc(l: Type::Union(Type::NonNegInt,DOM_LIST)) : dom
	local beenthere, rnd, i, n;
    begin
        if domtype(l)<>DOM_LIST then
	    l:=[$1..l];
        end_if;
        n := nops(l);
	if n=0 then // special case since n=0 is an invalid argument for random
            return([]);
        end_if;
	beenthere:= [FALSE $ n];
	rnd:=random(n);	   // random generator for 0..n-1
        [ (repeat i:=rnd()+1 until not beenthere[i] end_repeat;
           beenthere[i]:=TRUE;
           l[i];
          ) $ j=1..n
        ];
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // Methods
    //////////////////////////////////////////////////////////////////////////

    /// toMatrix(sp) - permutation matrix of the standard permutation sp
    toMatrix :=
    proc(sigma: dom::typeStandard) : matrix
        local M, m, i;
    begin
        m:=nops(sigma);
        M:=matrix(m,m);
        for i from 1 to m do
            M[sigma[i],i]:=1;
        end_for;
        M;
    end_proc;


    //////////////////////////////////////////////////////////////////////////
    // Arithmetic operations
    //////////////////////////////////////////////////////////////////////////

    /// inverse of the permutation
    inverse :=
    proc(x: dom::typeStandard) : dom::typeStandard
	local result, i;
    begin
	result:=x; // just to create a list of same size
	for i from 1 to nops(x) do
	    result[x[i]]:=i;
	end_for;
        result;
    end_proc;

    /// binary multiplication of permutations
    // should it accept n permutations ?
    _mult := dom::mult2;
    mult2 :=
    proc(x: dom, y: dom::typeStandard): dom
	local nx, ny, i;
    begin
	nx := nops(x); ny := nops(y);
	// complete with fixed points if the permutations are not
	// of the same length
	x := [op(x), $nx+1..ny];
	y := [op(y), $ny+1..nx];
	[ x[i] $ i in y ];
    end_proc;

    /// multiplication of a permutation by the inverse of another
    multInverse :=
    proc(x: dom, y: dom::typeStandard) : dom
	local result, nx, ny, i;
    begin
	nx := nops(x); ny := nops(y);
	// complete with fixed points if the permutations are not
	// of the same length
	x := [op(x), $nx+1..ny];
	y := [op(y), $ny+1..nx];
	result:=x; // just to create a list of same size
	for i from 1 to nops(x) do
	    result[y[i]] := x[i];
	end_for;
        result;
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // Inversions and Lehmer code
    //////////////////////////////////////////////////////////////////////////


    rotheDiagram :=
    proc(p: dom)
	local
	      i,   /* var for loop...                      */
	      j,   /* old row...                           */
	      s,   /* the next elementary transposition... */
	      inv, /* the set of inversions...             */
	      d,   /* the diagram : a matrix (n,n)...      */
	      n;   /* degree of the symmetric group...     */
    begin
	n := nops(p);
	d := array(1..n,1..n, [["."$n]$n]);
	inv := dom::inversions(p);
        /* traduction in terms of diagram... */
	for i from 1 to nops(inv) do
	    d[op(inv[i],1),p[op(inv[i],2)]] := 1;
	end_for;
        /* we put a x to denote the position of the values 1, 2, ... */
        /* and values to exhibit a reduced decomposition... */
	for i from 1 to n do
	    d[i,p[i]] := "x";
	    s := i;
	    for j from 1 to n do
		if (d[i,j]=1) then
		    d[i,j] := s;
		    s := s+1;
		end_if;
	    end_for;
	end_for;
	output::asciiArt::fromArray(d, Compact, ColumnSeparator=" ");
    end_proc;

    /// list of the inversions of a permutation
    /// O(n^3) because of the append. Should be O(n^2).
    // Precompute the number of inversions to preallocate
    // the list; 
    inversions := 
    proc(p: dom): Type::ListOf(DOM_SET)
        local i,j,result, len; 
    begin
        len := nops(p);
        result := [];
        for i from 1 to len-1 do
            for j from i+1 to len do
                if (p[i]>p[j]) then
                    result := append(result,{i,j});
                end_if;
            end_for;
        end_for;
        result;
    end_proc;


    /// number of inversions of a permutation
    /// O(n^2), should be O(n log n)
    inversionsNumber := 
    proc(p: dom): Type::NonNegInt
        local i,j,result, 
              len; 
    begin
        len := nops(p);
        result := 0;
        for i from 1 to len-1 do
            for j from i+1 to len do
                if (p[i]>p[j]) then
                    result := result + 1;
                end_if;
            end_for;
        end_for;
        result;
    end_proc;


    /// number of inversions of a permutation in O(n log n).
    //  Fast version using a modified merge sort algorithm.
    //  faster that inversionsNumber for permutations of size >= 17.
    //////
    //  The number of inversion is incremented in res durign the merge sort. 
    //  For small permutations (size <= sizeBlock) the trivial version 
    //  is used resulting in faster computations. 
    //////
    inversionsNumberMerge := 
    proc(l : dom, 
	 sizeBlock = 16) // Experimentally tuned (see experimental/mergeSort.mu).
	local res, mergeBlocRec;
    begin
	mergeBlocRec := 
	// return src sorted and as a side effect increment res
	// by the number of inversions of src.
	proc(src)
	    local n, cut,          // size of src and cut point for merging. 
                  src1, i1, i1End, // two source lists, current and end pointers. 
                  src2, i2, i2End, 
                  dst, iDst;       // the destination list and an pointer.
	begin
	    n := nops(src);
	    if n <= sizeBlock then 
		res := res + combinat::permutations::inversionsNumber(src);
		return(sort(src));
	    end_if;
	    cut := n div 2;
	    i1 := 1; 
	    i2 := 1;
	    i1End := cut + 1;
	    i2End := n - cut + 1;
	    src1 := mergeBlocRec([op(src, 1..cut)]);
	    src2 := mergeBlocRec([op(src, i1End..n)]);
	    dst  := src1.src2; // The start of src1 and the end of src2 are in place. 

	    // trivial case and start of src1 :
	    if i2 >= i2End then return(dst); end_if;
	    while src1[i1] <= src2[i2] do 
		i1 := i1 + 1;
		if i1 >= i1End then return(dst); end_if;
	    end_while;
	    iDst := i1;
	    while TRUE do
		while src2[i2] < src1[i1] do
		    dst[iDst] := src2[i2];
		    iDst := iDst + 1;
		    i2   := i2   + 1;
		    if i2 >= i2End then 
			res := res + (i1End-i1) * (i2-1);
			return([op(dst, 1..iDst-1), op(src1, i1..i1End-1)]);
		    end_if;
		end_while;
		while src1[i1] <= src2[i2] do
		    dst[iDst] := src1[i1];
		    iDst := iDst + 1;
		    i1   := i1   + 1;
		    res  := res  + i2 - 1;
		    if i1 >= i1End then // The rest of dst[i2] is already in place. 
			return(dst); 
		    end_if;
		end_while;
	    end_while;
	end_proc;
	res := 0;
	mergeBlocRec(l); // At the last stage the sorted list is not needed this
	                 // should probably be optimized. 
	res;
    end_proc;


    /// sign of a permutation
    /// O(n^2), should be O(n log n)
    signFromInversionsNumber :=
    proc(p: dom): Type::Integer
    begin
        if dom::inversionsNumber(p) mod 2 = 0 then 1 else -1; end_if;
    end_proc;


    /// O(n) optimal, slower for n=0,1,2,3, with a factor at most 2
    /// O(2n) in memory
    isEven :=
    proc(p: dom::typeStandard): DOM_BOOL
	local isEven, known, i, k;
    begin
	// Invariant: known[i] is TRUE iff i is already involved in a cycle
	known := [ FALSE $ nops(p) ];
	isEven := bool(nops(p) mod 2 = 0);
	for i from 1 to nops(p) do
	    if not known[i] then
		isEven := not isEven;
		k := i;
		repeat
		    known[k] := TRUE;
		    k := p[k];
		until i = k end_repeat;
	    end_if;
	end_for;
	isEven;
    end_proc;

    sign := (b->if b then 1 else -1 end_if) @ dom::isEven;

    /// Code of a permutation
    /// O(n^2), should be O(n log n)
    code := 
    proc(p: dom): Type::ListOf(Type::NonNegInt)
        local i, j, code, n; 
    begin
        n := nops(p);
        code := [0 $ n];
        for i from 1 to n-1 do
            for j from i+1 to n do
                if (p[i]>p[j]) then
		    code[i] := code[i] + 1;
                end_if;
            end_for;
        end_for;
        code;
    end_proc;


    /// standard permutation from its code
    /// O(n), optimal
    fromCode :=
    proc(code: Type::ListOf(Type::NonNegInt), n) : dom::typeStandard
	local p, remaining, i;
    begin
	if args(0)=1 then
	    n := max(0, (i+code[i]) $ i=1..nops(code));
	else
	    if testargs() and n<max(0, (i+code[i]) $ i=1..nops(code)) then
		error("Incompatible code and size ".expr2text(code).", ".expr2text(n));
	    end_if;
	end_if;
	code := append(code, 0 $ n-nops(code));
	remaining := [$1..n]; // values in [1..n] that have not yet been used
	p := [NIL $ n];
	for i from 1 to n do
	    p[i] := remaining[code[i]+1];
	    delete(remaining[code[i]+1]);
	end_for;
	p;
    end_proc;

    /// cocode of a permutation
    /// O(n^2), should be O(n log n)
    coCode :=
    proc(p: dom): Type::ListOf(Type::NonNegInt)
        local i, j, code, n;
    begin
        n := nops(p);
        code := [0 $ n];
        for i from 2 to n do
            for j from 1 to i-1 do
                if (p[j]>p[i]) then
		    code[i] := code[i] + 1;
                end_if;
            end_for;
        end_for;
        code;
    end_proc;

    /// reduced word of a permutation
    reducedWord :=
    proc(p: dom): Type::ListOf(Type::PosInt)
	local c, i, j;
    begin
	c := dom::code(p);
	[ (i+(c[i]-j) $ j=1..c[i]) $ i=1..nops(c) ];
    end_proc;

    /// lexicographically minimal reduced word of a permutation
    reducedWordLexMin :=
    proc(p: dom): Type::ListOf(Type::PosInt)
	local c, i;
    begin
	c := dom::coCode(dom::inverse(p));
	[op(revert([$i-c[i]..i-1])) $ i=2..nops(c)];
    end_proc;

    reducedWords := 
    proc(p: dom): Type::ListOf(Type::ListOf(Type::PosInt))
	local descs, rdws;
    begin
	rdws  := dom::reducedWords;
	descs := dom::descents(p);
	if nops(descs) = 0 then [[]];
	else
	    map(descs, 
		proc(d)
		    local pp, n; 
		begin
		    n := nops(p);
		    pp := [op(p, 1..d-1), p[d+1], p[d], op(p, d+2..n)];
		    op(map(rdws(pp), append, d)); 
		end_proc);
	end_if;
    end_proc;


    /// product of elementary transpositions
    /// product on the right of a permutation by elementary transpositions
    fromReducedWord :=
    proc(word: Type::ListOf(Type::PosInt), 
	 p=FAIL: Type::Union(Type::NonNegInt, dom, DOM_FAIL) ) : dom
	local i, tmp, maxi;
    begin
	maxi := max(-1, i $ i in word);
	if args(0)=1 then
	    p:=[$1..maxi+1];
	elif domtype(p)=DOM_INT then
	    if testargs() and p <= maxi then
		error("The transposition ("
		      .expr2text(maxi,maxi+1)
		      .") is not in the symmetric group S"		
		      .expr2text(p));
	    end_if;
	    p:=[$1..p];
	elif domtype(p)=DOM_LIST then
	    if testargs() and nops(p) <= maxi then
		error("The transposition ("
		      .expr2text(maxi,maxi+1)
		      .") is not in the symmetric group S"		
		      .expr2text(p));
	    end_if;
	end_if;
	for i in word do
	    tmp := p[i];
	    p[i]:=p[i+1];
	    p[i+1]:=tmp;
	end_for;
	p;
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // Cycle decomposition of standard permutations
    //////////////////////////////////////////////////////////////////////////

    /// cycle decomposition of a standard permutation
    cycles :=
    proc(p: dom::typeStandard)
	local i, k, cycle, result, known;
    begin
	known := [ FALSE $ nops(p) ];
	// Invariant: l[i] is TRUE iff i is already involved in a cycle
	cycle:=[];
	result:=[];
	for i from 1 to nops(p) do	
	    if not known[i] then
		k := i; 
		repeat
		    known[k]:=TRUE;
		    cycle := append(cycle,k);
		    k := p[k];
		until i = k end_repeat;
		result:=append(result,cycle);
		cycle := [];
	    end_if;
	end_for;
	result
    end_proc;

    /// cycle type of a standard permutation
    cycleType :=
    proc(p: dom::typeStandard)
    begin
	revert(sort(map(dom::cycles(p), nops)))
    end_proc;

    /// standard permutation from its cycle decomposition
    fromCycles :=
    proc(cycles: Type::ListOf(Type::ListOf(Type::PosInt)),
	 n=0: Type::NonNegInt) : dom::typeStandard
	local c, i, p, known;
    begin
	if args(0)=1 then
	    if nops(cycles)>0 then
		n:=max(op(map(cycles, op)));
	    else
		n:=0;
	    end_if;
	end_if;
	if testargs() then
	    known:=[FALSE $ n];
	    for c in cycles do
		for i in c do
		    if i>n then
			error("Cycles should consists of integers in {1..n}")
		    end_if;
		    if known[i] then
			error("Cycles are not disjoint");
		    end_if;
		    known[i]:=TRUE;
		end_for;
	    end_for;
	end_if;
	p:=[$1..n];
	for c in cycles do
	    for i from 1 to nops(c)-1 do
		p[c[i]]:=c[i+1];
	    end_for;
	    p[c[nops(c)]]:=c[1];
	end_for;
	p;
    end_proc;

    /// fixed points of a standard permutation
    fixedPoints :=
    proc(p: dom::typeStandard) : Type::ListOf(Type::PosInt)
    begin
	select([$1..nops(p)], i ->p[i]=i);
    end_proc;

    /// number of fixed points of a standard permutation
    fixedPointsNumber :=
    proc(p: dom::typeStandard) : Type::NonNegInt
	local result, i;
    begin
	result := 0;
	for i from 1 to nops(p) do
	    if p[i]=i then result := result + 1; end_if
	end_for;
	result;
    end_proc;

/**
    majorIndex(p)
    Returns the major index of the permutation p (sum of the descents)
    References: Mac-Mahon (1915) reprint 1960
**/

    majorIndex :=
    proc(p: dom):  Type::NonNegInt
    begin
        _plus(op(dom::descents(args())));
    end_proc;

    majorCode := 
    // it is possible to do this in n*log(n)
    proc(p : dom::typeStandard): Type::ListOf(Type::NonNegInt)
	local res, i;
    begin
	if p = [] then return([]) end_if;
	res := [dom::majorIndex(select(p, not _less, i )) $ i=1..nops(p)];
	[res[i]-res[i+1] $ i=1..nops(res)-1, 0];
    end_proc;

    fromMajorCode := 
    proc(code: Type::ListOf(Type::NonNegInt), n) : dom::typeStandard
	local res, i, desc, rise, ins;
    begin
	if args(0)=1 then
	    n := max(0, (i+code[i]) $ i=1..nops(code));
	else
	    if testargs() and n<max(0, (i+code[i]) $ i=1..nops(code)) then
		error("Incompatible code and size ".expr2text(code).", ".expr2text(n));
	    end_if;
	end_if;
	if code = [] then return([]) end_if;
	code := append(code, 0 $ n-nops(code));
	res  := [n];
	for i from n-1 downto 1 do
	    desc := revert(dom::descents(res));
	    rise := sort([op({$0..n-i} minus {op(desc)})]);
	    if code[i] < nops(desc) then
		ins  := desc[code[i] + 1] + 1;
	    else
		ins  := rise[code[i] - nops(desc) + 1] + 1;
	    end_if;
		res  := listlib::insertAt(res, i, ins);
	end_for;
	res;
    end_proc;

    reculs := dom::descents @ dom::inverse;

/**
    descentPolynomial(p, z)
 
    Returns the descent polynomial of the permutation p. That is the
    product of all the (z[p[1]]..z[p[i]]), where i are descents of p.
 
    References: Garsia and Stanton 1984
**/
 
    descentPolynomial :=
    proc(p: dom, z=hold(z): DOM_IDENT)
        local result, Pol, i;
    begin
        result:=1;
        Pol:=1;
        for i from 1 to nops(p)-1 do
            Pol:=Pol*z[p[i]];
            if p[i]>p[i+1] then
                result:=result*Pol;
            end_if
        end_for;
        result;
    end_proc;
  
///    Returns the smallest permutation with prescribed descents
  descentsCompositionToMin :=
    proc(dc : combinat::compositions) 
	local cpl,
	      sum, // partial sum 
	      i, j, res;
    begin
	cpl := revert(combinat::compositions::conjugate(dc));
	res := null();
	sum := 0;
	for i from 1 to nops(cpl) do
	    res := res, (sum + cpl[i]+1-j $ j=1..cpl[i]); // column filling
	    sum := sum + cpl[i];
	end_for;
	[res];
    end_proc;
  
  descentsToMin := dom::descentsCompositionToMin @ 
		  combinat::compositions::fromDescents;
  
///    Returns the biggest permutation with prescribed descents
  descentsCompositionToMax :=
    proc(dc : combinat::compositions)
	local size, sum, 
	      i, res;
    begin
	size := nops(dc);
	sum := 0;
	res := null();
	for i from size downto 1 do
	    res := $(sum+1)..(sum+dc[i]), res; // remplissage colonne inverse
	    sum := sum + dc[i];
	    
	    end_for;
	[res];	
    end_proc;
  
  descentsToMax := dom::descentsCompositionToMax @ 
		  combinat::compositions::fromDescents;
 
  
  // taken from	muEC::SG::Desc2ListPerm_rec	     
  fromDescents_rec := 
    proc(compo, perm)
	local
	      n, /* nops(perm)...  */
	      i, /* var for seq... */
	      DOM;	// Standard DOM hack
	option remember;
    begin
	DOM := dom;
	n := nops(perm);
	if (nops(compo)<=1) then
	    [perm];
	elif ({op(compo)}={1}) then
	    [[perm[-i] $i=-n..-1]];
	elif (compo[1]=1) then
	    map(
		map([[perm[i], [op(perm, 1..i-1), op(perm, i+1..n)]] $i = 2..n],
                    proc(x,y) begin [x[1], DOM::fromDescents_rec(y, x[2])] end_proc, 
	            [op(compo,2..nops(compo))]),
                proc(x) begin
			    op(map(x[2],
                               proc(y,z) 
			       begin
				   if (y[1]>z) then null(); else [z, op(y)] end_if;
			       end_proc,
				   x[1]))
			end_proc
	       );
	else
	    map(
		map([[perm[i], [op(perm,1..i-1), op(perm, i+1..n)]]$i = 1..n-1],
                proc(x,y) begin [x[1], DOM::fromDescents_rec(y, x[2])] end_proc,
		    [compo[1]-1, op(compo,2..nops(compo))]),
                proc(x) begin
			    op(map(x[2], 
                               proc(y,z) begin
					     if (y[1]<z) then null(); else [z, op(y)] end_if;
					 end_proc, 
				   x[1]))
			end_proc
	       );
	end_if;
    end_proc;
  
  fromDescentsComposition :=
    proc(desc : combinat::compositions)
	local res;
    begin
	res := dom::fromDescents_rec(desc, [$1.._plus(op(desc))]);
	// I'd rather use extsubsop than subsop to clearly indicate
	// that we are using a hack; however, extsubsop does not work
	// on kernel domains with MuPAD 2.0.0
	dom::fromDescents_rec := subsop(dom::fromDescents_rec, 5=NIL);
	res;
    end_proc;
 
  fromDescents := dom::fromDescentsComposition @
		  combinat::compositions::fromDescents;
  

    // youngSubgroup

    /*
     * Given a permutation p, and a parabolic (young) subgroup S_I
     * of the symmetric group, returns a canonical representative of p
     * modulo the right action of S_I on the positions.
     * This simply amount to sorting p in each interval defined by I.
    **/
    modYoungSubgroup :=
    proc(p: dom, c: combinat::integerVectors) : dom
	local start, i;
    begin
	assert(_plus(op(c))=nops(p));
	start := 1;
	[ op(sort([op(p, start..(start:=start+i)-1)])) $ i in c]
    end_proc;

    // See also muEC::SG::FactorPerm

  connectedShape := 
    proc(per : dom::typeStandard)
	local i,j, mx,  // maximum
	      res; // result
    begin
	j := 0;
	mx := 0;
	res := [];
	for i from 1 to nops(per) do
	    mx := max(mx,per[i]);
	    j  := j+1;
	    if mx = i then
		res := append(res,j);
		j := 0;
	    end_if;
	end_for;
	res;
    end_proc;

  isConnected := 
    proc(per : dom::typeStandard)
    begin
        bool(nops(dom::connectedShape(per))=1);
    end_proc;
  
    //////////////////////////////////////////////////////////////////////////
    // Peaks
    //////////////////////////////////////////////////////////////////////////

/**
    peaks(p)
    Returns the list of peaks of the permutation p
 
    That is the list of the i for which p[i-1]<= p[i] >p[i+1]
**/
    peaks :=
    proc(p: dom):  Type::ListOf(Type::NonNegInt)
        local result, i;
    begin
        result:=[];
        for i from 2 to nops(p)-1 do
            if p[i]>p[i+1] and p[i]>=p[i-1]then
                result:=append(result,i);
            end_if
	end_for;
	result;
    end_proc;

    peaksNumber :=
    proc(p: dom):  Type::NonNegInt
    begin
	nops(dom::peaks(p));
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // Patterns
    //////////////////////////////////////////////////////////////////////////
 
    patternPositions := 
    proc (perm : dom, patt : dom::typeStandard) 
    begin
	combinat::generators::toList(
	     combinat::generators::select(
		  combinat::subwords::generator([$1..nops(perm)], nops(patt)),
		     pos -> (combinat::permutations::standard(map(pos, z-> perm[z]))			     = patt)));
    end_proc;

    hasPattern := 
    proc(perm : dom, patt : dom::typeStandard)
	local n, l, gen, pos, res;
    begin
	n := nops(perm);
	l := nops(patt);
	res := null();
	gen := combinat::subwords::generator([$1..n], l);
	while (pos := gen()) <> FAIL do
	    if combinat::permutations::standard(map(pos, z-> perm[z])) = patt
		then 
		return(TRUE);
	    end_if;
	end_while;
	FALSE;
    end_proc;

  
    //////////////////////////////////////////////////////////////////////////
    // Saillances
    //////////////////////////////////////////////////////////////////////////
    
    saillances :=
    proc(p: dom): Type::ListOf(Type::PosInt)
        local result, i, j;
    begin
        result:=[];
        for i from 1 to nops(p) do
	    if _and(p[i]>p[j] $ j=i+1..nops(p)) then
                result:=append(result,i);
            end_if
	end_for;
	result;
    end_proc;

    saillancesNumber :=
    proc(p: dom): Type::NonNegInt
    begin
	nops(dom::saillances(p));
    end_proc;

    has1324 :=
    proc(p: dom): DOM_BOOL
	local saillances, m, i, j;
    begin
	if nops(p)=0 then
	    return(FALSE);
	end_if;
	saillances := {op(dom::saillances(p))}; // set of the saillances of p
	m := p[1]; // m: minimal value in p[1]..p[i-1]
	for i from 2 to nops(p) do
	    if p[i]<m then
		m:=p[i];
		next;
	    end_if;
	    if i in saillances then
		next;
	    end_if;
	    for j from i+1 to nops(p) do
		if j in saillances then
		    if p[j]<p[i] then
			break;
		    else
			next;
		    end_if;
		end_if;
		if m < p[j] and p[i]>p[j] then
		    return(TRUE);
		end_if;
	    end_for;
	end_for;
	FALSE;
    end_proc;
  
    //////////////////////////////////////////////////////////////////////////
    // Bruhat Order
    //////////////////////////////////////////////////////////////////////////
  
  leequalBruhat :=
    proc(perm1 : dom::typeStandard, perm2 : dom::typeStandard)
    // Stolen from MuEC
	local i, j, c, n1, n2;
    begin
	n1 := nops(perm1);
	n2 := nops(perm2);
	if testargs() and (n1<>n2) then
	    error("Length mismatch : ".expr2text(perm1)." and ".expr2text(perm2));
	end_if;
	if n1=0 then return(TRUE); end_if;
	if ((perm1[1] > perm2[1]) or (perm1[n1] < perm2[n1])) then
	    return(FALSE);
	end_if;
	for i from 1 to n1 do
	    c := 0;
	    for j from 1 to n1 do
		if (perm2[j] > i) then
		    c := c+1;
		end_if;
		if (perm1[j] > i) then
		    c := c-1;
		end_if;
		if (c<0) then
		    return(FALSE);
		end_if;
	    end_for;
	end_for;
	TRUE;
    end_proc;
  
  bruhatInversions :=
    proc(p : dom::typeStandard)
        local i, j, k, result, len, ok; 
    begin
        len := nops(p);
        result := [];
        for i from 1 to len-1 do
            for j from i+1 to len do
		if (p[i]>p[j]) then
		    ok := TRUE;
		    for k from i+1 to j-1 do
			if (p[i]>p[k]) and (p[k]>p[j]) then
			    ok := FALSE; break;
			end_if;
		    end_for;
		    if ok then 
			result := append(result,{i,j});
		    end_if;
                end_if;
            end_for;
        end_for;
        result;
    end_proc;
  
  succBruhat :=
    proc(perm : dom::typeStandard)
	local z, n; 
    begin
	n := nops(perm);
	[subsop(perm, z[1]=perm[z[2]], z[2]=perm[z[1]]) 
	 $ z in dom::bruhatInversions(map(perm, z-> n+1-z))];
    end_proc;
  
  predBruhat :=
    proc(perm : dom::typeStandard)
	local z; 
    begin
	[subsop(perm, z[1]=perm[z[2]], z[2]=perm[z[1]]) 
	 $ z in dom::bruhatInversions(perm)];
    end_proc;
  
  smallerBruhat :=
    proc(perm : dom::typeStandard)
    begin
	combinat::transitiveIdeal(dom::predBruhat, perm);	
    end_proc;

  greaterBruhat :=
    proc(perm : dom::typeStandard)
    begin
	combinat::transitiveIdeal(dom::succBruhat, perm);	
    end_proc;
  
    //////////////////////////////////////////////////////////////////////////
    // Permutohedron Order
    //////////////////////////////////////////////////////////////////////////
  
  predPermutohedron :=
    proc(perm : dom::typeStandard, dir = Right : dom::typeDirection )
	local i, recoil, des, n;
    begin
	n := nops(perm);
	if dir = Right then
	    des := dom::descents;
	    [subsop(perm, i=perm[i+1], i+1=perm[i]) $ i in des(perm)];
	else
	    recoil := (perm) -> select([$1..n-1], 
				i -> contains(perm, i) > contains(perm, i+1));
	    [subs(perm, {i=i+1, i+1=i}) $ i in recoil(perm)];
	end_if;
    end_proc;
  
  smallerPermutohedron :=
    proc(perm : dom::typeStandard, dir = Right : dom::typeDirection )
    begin
	combinat::transitiveIdeal(
	     prog::bless(z -> dom::predPermutohedron(z, dir), dom), perm);    
	end_proc;
  
  succPermutohedron :=
    proc(perm : dom::typeStandard, dir = Right : dom::typeDirection )
	local i, rise, advance, n;
    begin
	n := nops(perm);
	if dir = Right then
	    rise    := (perm) -> select([$1..n-1], i->perm[i] < perm[i+1]);
	    [subsop(perm, i=perm[i+1], i+1=perm[i]) $ i in rise(perm)];
	else
	    advance := (perm) -> select([$1..n-1], 
				 i -> contains(perm, i) < contains(perm, i+1));
	    [subs(perm, {i=i+1, i+1=i}) $ i in advance(perm)];
	end_if;
    end_proc;
  
  greaterPermutohedron :=
    proc(perm : dom::typeStandard, dir = Right : dom::typeDirection )
    begin
	combinat::transitiveIdeal(
	    prog::bless(z -> dom::succPermutohedron(z, dir), dom), perm);
    end_proc;

  leequalPermutohedron :=
    proc(perm1 : dom::typeStandard, perm2 : dom::typeStandard,
	 dir = Right : dom::typeDirection)
	local n1, l1, n2, l2, prod;
    begin
	n1 := nops(perm1);
	n2 := nops(perm2);
	if testargs() and (n1<>n2) then
	    error("Length mismatch : ".expr2text(perm1)." and ".expr2text(perm2));
	end_if;
	l1 := dom::inversionsNumber(perm1);
	l2 := dom::inversionsNumber(perm2);
	if (l1 > l2) then return(FALSE); end_if;
	if dir = Right then 
	    prod := dom::_mult(dom::inverse(perm2), perm1);
	else
	    prod := dom::_mult(perm1, dom::inverse(perm2));
	end_if;
	bool(dom::inversionsNumber(prod) = l2 - l1); 
    end_proc;

    // Backward compatibility aliases

    leequalPermutoedron := dom::leequalPermutohedron;
    predPermutoedron := dom::predPermutohedron;
    succPermutoedron := dom::succPermutohedron;
    smallerPermutoedron := dom::smallerPermutohedron;
    greaterPermutoedron := dom::greaterPermutohedron;

end_domain:

// should probably go into combinat::setPartitions when this one is created
domain combinat::permutations::fromSetPartition
    inherits combinat::permutations;
    category Cat::CombinatorialClass;
    axiom    Ax::systemRep;

    info_str := "The combinatorial class of the permutations stabilizing a set partition of {1,...,n}";

    typeStandardSetPartition :=
	Type::Intersection
	(Type::ListOf(Type::Union(Type::SetOf(Type::PosInt),
				  Type::ListOf(Type::PosInt))),
	 Type::Predicate(l->bool({ op(x) $ x in l } =
				 { $ 1.._plus(nops(x) $ x in l) })));

    generator :=
    proc(setPartition: dom::typeStandardSetPartition)
	local n, x, component;
	option escape;
    begin
	n := _plus(nops(x) $ x in setPartition);
	generators::map
	(generators::cartesian
	 (permutations::generator(nops(component))
	  $ component in setPartition),
	 proc(l: Type::ListOf(permutations))
	     local result, component, p, i, j;
	 begin
	     // Given the permutations on each component of the set partition
	     // (given in standard form), builds the corresponding
	     // permutation of [1..n]
	     result := [ $ 1..n ];
	     for i from 1 to nops(setPartition) do
		 component := setPartition[i];
		 p         := l[i];
		 for j from 1 to nops(component) do
		     result[op(component,j)] := op(component, p[j]);
		 end_for;
	     end_for;
	     result;
	 end_proc)
    end_proc;

    count :=
    proc(setPartition: dom::typeStandardSetPartition)
	local component;
    begin
	_mult(nops(component)! $ component in setPartition);
    end_proc;
end_domain:


// should probably go into combinat::setPartitions when this one is created
domain combinat::permutations::youngSubgroup
    inherits combinat::permutations;
    category Cat::CombinatorialClass;
    axiom    Ax::systemRep;

    info_str := "The combinatorial class for young subgroups";

    generator :=
    proc(c: combinat::integerVectors) : combinat::generators
	local start;
    begin
	if nops(c) = 0 then // Otherwise _concat below complains
	    return(combinat::generators::singleton([]));
	end_if;
	n := _plus(op(c));
	start := 1;
	components := [ [ $ start .. (start := start + part) -1] $ part in c];
	generators::map
	(generators::cartesian
	 (permutations::generator(component)
	  $ component in components),
	 _concat@op)
    end_proc;

    count :=
    proc(c: combinat::integerVectors) : Type::PosInt
	local i;
    begin
	_mult(i! $i in c);
    end_proc;
end_domain:


/////////////////////////////////////////////////////////////////////////
//                                                                     //
//                    Connected permutations                           //
//                                                                     //
/////////////////////////////////////////////////////////////////////////
domain combinat::permutations::connected
    inherits combinat::permutations;
    category Cat::CombinatorialClass;
    axiom    Ax::systemRep;

    info_str := "The combinatorial class of connected permutations";

    list :=
    proc(n : Type::NonNegInt)
    begin
        dom::connectedSup([$1..n]);
    end_proc;

    count :=
    proc(n : Type::NonNegInt)
    begin
        dom::connectedSupNumber([1$1..n]);
    end_proc;

    isA :=
    proc(p : Type::AnyType,
         n=null() : Type::Union(DOM_NULL,Type::NonNegInt)
        )
    begin
        if not(testtype(p,combinat::permutations::typeStandard)) then
           return(FALSE);
        end_if;
        if args(0)=2 then
            if (nops(p)<>n) then return(FALSE);
            end_if;
        end_if;
        return(combinat::permutations::isConnected(p));
    end_proc;
    ////////////////////////////////////
    // Connected Cuts of a permutation
    //////
    // FirstConnectedCut returns the last index of the first block
    // NextConnectedCut  returns the last index of the next block
    //////
    // per : permutation
    // pos : last index of a block
    firstConnectedCut := 
    proc(per : combinat::permutations) 
    begin 
        dom::nextConnectedCut(per, 0);
    end_proc;
    //////
    nextConnectedCut := 
    proc(per : combinat::permutations, pos : Type::NonNegInt)
        local i, mx;  // maximum
    begin
	if per = [] then return(0); end_if;
        mx  := per[pos+1];
        i   := pos + 1;
        while mx<>i do
	    i := i+1;
	    mx := max(mx, per[i]);
        end_while;
        i;
    end_proc;
    
    ///////////////////////////////////////////////////////////
    // List of the connected permutations sigma greater than 
    // leftw.rightw and such that:
    //  - leftw   is a subword of sigma
    //  - leftw.a is not a subword of sigma (a: first letter of rightw)
    //////
    //  - leftw is a permutation
    //  - rightw is not empty and does'nt split into connected blocks
    //////
    // Note : ConnectedShuffle compute the end-point of the recursion 
    //        of the function ConnectedSup
    connectedShuffle := 
    proc(leftw, rightw)
        local i, shu, cut, leftRes, reset;
    begin    
        cut := dom::firstConnectedCut(leftw);
        
        res := [];
        for i from 1 to cut do // position where a is inserted in leftw
	    leftRes := [op(leftw, 1..i-1), rightw[1]];
	    shu     := combinat::words::shuffle(
                         [op(leftw,i..nops(leftw))],
			 [op(rightw,2..nops(rightw))]);
	    res     := res.map(shu, z -> leftRes.z);
	end_for;
        res;
    end_proc;

    /////////////////////////////
    // number of such permutation
    //////
    //  - rightC has only one part
    connectedShuffleNumber := 
    proc(leftC, rightC)
	local leftS, rightS, i;
    begin
	leftS  := _plus(op(leftC));
	rightS := rightC[1];
	_plus(binomial(rightS+leftS-i,leftS-i+1) $ i=1..leftC[1]);
    end_proc;

    ///////////////////////////////////////////////////////////
    // List of the connected permutation sigma greater than 
    // leftw.rightw and such that:
    //  - leftw is a subword of sigma
    //////
    //  - leftw, leftw.rightw are permutations
    //  - rightw is NOT empty (=> leftw.rightw is NOT connected)
    //////
    // Note : ConnectedSupRec compute the running step of the recursion 
    //        of the function ConnectedSup
    connectedSupRec := 
    proc(leftw, rightw)
	local n, lright, // total and right length
	cut,      // position of the cut = lleft
    	leftComp, // connected block composition of leftw
    	rightCut, // position of the new cut in the total 
    	remw,     // and remaining part
    	shuf, res;
    begin
        cut    := nops(leftw);
        lright := nops(rightw);
        n      := cut + lright;
        leftComp := combinat::permutations::connectedShape(leftw);
        
        rightCut := dom::nextConnectedCut(leftw.rightw, cut);
        if rightCut = n then // There are no blocks left : Go to end-point
	    return(dom::connectedShuffle(leftw, rightw)) 
	end_if;
	
        res := dom::connectedSupRec( // We transfert a block and
                                    // call ConnectedSupRec recursively
			    [op(leftw), op(rightw, 1..rightCut-cut)],
			    [op(rightw,rightCut-cut+1..nops(rightw))]);
    
        // Non trivial shuffle of the left word with the middle-block
        shuf     := combinat::words::shuffle(leftw,
                       [op(rightw,1..rightCut-cut)]);
        shuf     := [op(shuf,2..nops(shuf))]; // removes the concatenation
        remw     := [op(rightw, rightCut-cut+1..lright)];
        res.map(shuf, op @ dom::connectedSupRec, remw);
    end;

    /////////////////////////////
    // number of such permutation
    connectedSupNumberRec := 
    proc(leftC, rightC)
	local res, i, nShuffleIc, Ic, shBlock, shBlock1, block, remC;
    begin
        if nops(rightC) = 1 then // There are no blocks left : Go to end-point
	    return(dom::connectedShuffleNumber(leftC, rightC));
	end_if;
	
        block   := rightC[1];
        remC    := [op(rightC,2..nops(rightC))];
        res := dom::connectedSupNumberRec([op(leftC), block], remC);
    	
        shBlock := block;
        // ConnectedShape of non trivial shuffle
	for i from nops(leftC) downto 1 do
	    shBlock1  := shBlock;
	    shBlock   := shBlock + leftC[i];
	    Ic         := [op(leftC, 1..i-1), shBlock];
	    nShuffleIc := binomial(shBlock, block) 
		       - binomial(shBlock1, block);
	    res       := res + nShuffleIc * dom::connectedSupNumberRec(Ic, remC);
	end_for;
        res;
    end_proc;
    
    /////////////////////////////////////////////////////
    // List of the connected permutation greater than per
    //////
    // per : permutation
    connectedSup := 
    proc(per)
	local fb, n;
    begin
	n  := nops(per);
	fb := dom::firstConnectedCut(per);
	if fb = n then [per]; // per is connected
	else
	    dom::connectedSupRec([op(per,1..fb)],[op(per,fb+1..n)]);
	end_if;
    end_proc;

    ///////////////////////////////////////////////////
    // Number of connected permutation greater than any 
    // permutation of connected composition compo
    //////
    // compo : composition
    connectedSupNumber := 
    proc(compo)
    begin
	if nops(compo) = 1 then 1
	else dom::connectedSupNumberRec([op(compo,1)],
                  [op(compo,2..nops(compo))]);
	end_if;
    end_proc;

    // TODO : compute by the means of the generating series
    // See NCSF6.
    count := 
    proc(n)
    begin
	if n = 0 then 0
	else
	    dom::connectedSupNumber([1$n]);
	end_if;
    end_proc;

    listRec := 
    proc(rightW, maxRightW, remainings)
	local res, i, start, lenRightW, nRem;
    begin
	if nops(remainings) = 1 then return([rightW.remainings]);
	end_if;
	lenRightW := nops(rightW) + 1;
	nRem := nops(remainings);
	if maxRightW > lenRightW then
	    res := _concat(dom::listRec(append(rightW, remainings[i]),
					max(maxRightW, remainings[i]),
					[op(remainings, 1..i-1), op(remainings, i+1..nRem)])
			   $ i = 1..nRem);
	else
	    start := 1;
	    while remainings[start] <= lenRightW do 
		start := start+1;
	    end_while;
	    _concat(dom::listRec(append(rightW, remainings[i]),
				 remainings[i],
				 [op(remainings, 1..i-1), op(remainings, i+1..nRem)])
		    $ i = start..nRem);	    
	end_if;
    end_proc;

    list := 
    proc(n)
    begin
	if n = 0 then [] 
	else
	    dom::listRec([], 0, [$1..n]);
	end_if;
    end_proc;
end_domain:

/////////////////////////////////////////////////////////////////////////
//                                                                     //
//         Right cosets Blah: Don't know what name to give	       //
//                                                                     //
// Given a word w, let l be the smallest index such that w[l]... is    //
// decreasing.                                                         //
// rightCosetBlah(w) returns a list of permutations p such that        //
// mult(w,p) goes through all the permutations of w which fix          //
// [w[1],...,w[l-1]]                                                   //
/////////////////////////////////////////////////////////////////////////

domain permutations::rightCosetsBlah
    inherits permutations;
    category Cat::CombinatorialClass;
    axiom    Ax::systemRep;

    info_str := "The combinatorial class of right cosets of permutations";

    // Returns the smallest l such that, w[l]..w[nops(bla)] is decreasing
    // Base case: 1 is returned if w is empty.
    startIndex :=
    proc(w: dom) : Type::PosInt
	local l, i;
    begin
	l := 1:
	for i from 2 to nops(w) do
	    if w[i-1]<w[i] then l := i; end;
	end_for;
	l;
    end_proc;
    
    count :=
    proc(w: dom) : Type::NonNegInt
    begin
	permutations::count([op(w, dom::startIndex(w)..nops(w))]);
    end_proc;
    
    generator :=
    proc(w) : generators
	local vertices, edges, i, l;
    begin
	l := dom::startIndex(w);
	vertices := [$ 1..nops(w)];
	if l=1 then // w is directly a partition
	    edges := [];
	else
	    edges := [ [i-1,i] $ i=2..l-1, [l-1,i] $i=l..nops(w) ];
	end_if;
	for i from l+1 to nops(w) do
	    if w[i-1]=w[i] then
		edges := append(edges,[i-1,i]);
	    end_if;
	end_for;
	linearExtensions::generator(Graph(vertices, edges, hold(Directed)));
    end_proc;
end_domain:
