/*****************************************************************************
  $Id: setPartitions.mu,v 1.1.1.4 2005/01/26 13:42:24 nthiery Exp $
  File:		lib/COMBINAT/setPartitions.mu
  Authors:	Florent Hivert <florent.hivert@univ-mlv.fr>,
                Jean-Christophe Novelli <novelli@univ-mlv.fr>
  Created:	16/02/2001
  License:	LGPL
*****************************************************************************/

domain combinat::setPartitions
    inherits Dom::BaseDomain;
    category Cat::GradedCombinatorialClass;
    axiom    Ax::systemRep;

    info_str := "Library 'combinat::setPartitions': utilities for set partitions";

    interfaceAdd :=
       { hold(_less), hold(inf), hold(sup)
       };

    domtype := DOM_SET;


    //////////////////////////////////////////////////////////////////////////
    // isA
    //////////////////////////////////////////////////////////////////////////

    isA := 
    proc(obj          : Type::AnyType,
         set = FAIL : Type::Union(DOM_FAIL, DOM_SET, DOM_LIST,
                                  Type::NonNegInt),
         part = FAIL  : Type::Union(DOM_FAIL, combinat::partitions,
                                    Type::NonNegInt)
	 ) : DOM_BOOL
    local sizes;
    begin
	if not testtype(obj, Type::SetOf(DOM_SET)) then return(FALSE); end_if;
	sizes := map([op(obj)],nops);
        if contains(sizes,0) <> 0 then return(FALSE); end_if;
        if nops(_union(op(obj)))<>_plus(op(sizes)) then return(FALSE); end_if;
	if args(0) = 1 then return(TRUE); end_if;

	if domtype(set)=DOM_INT then 
	    set := {$1..set};
	else
	    set := {op(set)};
	end_if;
	if _union(op(obj)) <> set then return(FALSE); end_if;
	if args(0) = 2 then return(TRUE); end_if;

        if domtype(part)=DOM_INT
          then return(bool(nops(obj)=part));
        end_if;
        if revert(sort([op(sizes)])) <> part then return(FALSE); end_if;
        TRUE;
    end_proc;

    count :=
    proc(n           : Type::Union(DOM_SET, DOM_LIST, Type::NonNegInt),
	 part = FAIL : Type::Union(DOM_FAIL, combinat::partitions,
                                   Type::NonNegInt)
	) : DOM_INT
	local res, expo, i;
    begin
	if domtype(n)<> DOM_INT then n := nops(n); end_if;
	if args(0) = 1 then combinat::bell(n)
	elif domtype(part) = DOM_INT then combinat::stirling2(n, part);
	elif testargs() and n <> _plus(op(part)) then
	    error("Size mismatch: ".expr2text(n).", ".expr2text(part));
	else
	    res := n!;
	    expo := combinat::partitions::toExp(part);
	    for i from 1 to nops(expo) do
		// there are expo blocs of size i
		if expo[i]<>0 then 
		    res := res / expo[i]! / (i!)^expo[i];
		end_if;
	    end_for;
	    res;
	end_if;
    end_proc;

    list :=
    proc(obj         : Type::Union(DOM_SET, DOM_LIST, Type::NonNegInt),
	 part = FAIL : Type::Union(DOM_FAIL, combinat::partitions,
                                   Type::NonNegInt)
	) : DOM_LIST
/*
Here is the algorithm used to compute unordered set partitions described
on an example.
When one asks about the list of set partitions of 1..15  in a partition of
shape [3,3,2,2,2,2,1], the program first computes an _ordered_ set partition
of lengths [3+3,2+2+2+2,1], that is, first regroups the parts of the partition
in equal components (see the line
$ blocs in combinat::setPartitionsOrdered::list(obj,taillesblocs)];
).

Given an ordered set partition of [6,8,1], it then computes all the unordered
partitions of the 6 elements in two blocks of size 3, then the unordered
partitions of the 8 elements in four blocks of size 2, and the unordered
partitions of the element in one block of size 1.
(that is the role of listbloc)
*/
	local expo, taillesblocs, blocs, i, nonzero, res;
    begin
	if domtype(obj) = DOM_INT then obj := [$1..obj];
	else obj := [op(obj)];
	end_if;
	
	if args(0) = 1 then
          return(map(combinat::partitions::list(nops(obj)),
		     prog::bless(z-> op(dom::list(obj,z)), dom)));
	elif domtype(part)=DOM_INT then
          return(map(combinat::partitions::list(nops(obj),Length=part),
                     prog::bless(z-> op(dom::list(obj,z)), dom)));
        end_if;
	
	if testargs() and nops(obj) <> _plus(op(part)) then
	    error("Size mismatch: ".expr2text(obj).", ".expr2text(part));
	end_if;

	nonzero := [];
	expo := combinat::partitions::toExp(part);
	for i from 1 to nops(expo) do
	    if expo[i]<>0 then nonzero := append(nonzero, [i,expo[i]]); end_if;
	end_for;
	
	taillesblocs := [nonzero[i][1]*nonzero[i][2] $i=1..nops(nonzero)];
	
	res:=[op(combinat::cartesianProduct::list(
	        (dom::listbloc(op(nonzero[i]), [op(blocs[i])])
                $i=1..nops(nonzero))))
	     $ blocs in combinat::setPartitionsOrdered::list(obj,taillesblocs)];
	
	res := map(res, map, op); // set(set_bloc(set)) ->set(set) 
	res := map(res, DOM_SET@op);
	return(res);
    end_proc;

    listbloc :=
    proc(n        : Type::NonNegInt,
	 nbrepets : Type::NonNegInt,
	 listint = FAIL : Type::Union(FAIL, DOM_LIST, Type::NonNegInt)
	) : DOM_LIST
/*
listbloc decomposes a set of n integers (the list listint) in _nbrepets_ parts
of size n/nbrepets. Please notice that this function is not accessible to the
user and so, does not check anything of complatibility between n and listint
and between n and nbrepets...
*/
	local smallest, ssens;
    begin
	if nbrepets=1 then return([{{op(listint)}}]); end_if;
	
        if args(0)=2 or domtype(listint)=DOM_INT
          then listint:=[$1..n];
        end_if;

	smallest := listint[1];
	listint  := [op(listint,2..nops(listint))];
	
	[(op(map(dom::listbloc(n, nbrepets-1,
			       listlib::setDifference(listint,ssens)),
		 (z,ssens) -> {_union(ssens,{smallest}),op(z)},ssens)))
	 $ ssens in combinat::subsets::list({op(listint)},n-1)];	
    end_proc;
  
    //////////////////////////////////////////////////////////////////////////
    // Methods
    //////////////////////////////////////////////////////////////////////////

    _less := 
    proc(part1 : dom, part2 : dom)
	local p;
    begin
	if testargs() and _union(op(part1)) <> _union(op(part2)) then
	    error("Cannot compare partitions of different sets");
	end_if;
	if part1 = part2 then return(FALSE); end_if;
	for p in part1 do
	    if nops(select(part2, z -> _intersect(z, p) <> {})) <> 1 then 
		return(FALSE);
	    end_if;
	end_for;
	TRUE;
    end_proc;

    inf := 
    proc(part1 : dom, part2 : dom)
	local p1, p2;
    begin
	if testargs() and _union(op(part1)) <> _union(op(part2)) then
	    error("Cannot compare partitions of different sets");
	end_if;
	{_intersect(p1, p2) $ p1 in part1 $ p2 in part2} minus {{}}
    end_proc;

    sup := 
    proc(part1 : dom, part2 : dom)
	local res, p, inters;
    begin
	if testargs() and _union(op(part1)) <> _union(op(part2)) then
	    error("Cannot compare partitions of different sets");
	end_if;
	res := part1;
	for p in part2 do
	    inters := select(res, z -> _intersect(z, p) <> {});
	    res := res minus inters union {_union(op(inters))};
	end_for;
	res;
    end_proc;

    supOfSets := 
    proc()
	local res, p, inters;
    begin
	res := map(_union(args()), DOM_SET);
	for p in [args()] do
	    inters := select(res, z -> _intersect(z, p) <> {});
	    res := res minus inters union {_union(op(inters))};
	end_for;
	res;
    end_proc;

begin
end_domain:
