/*****************************************************************************
  $Id: skewTableaux.mu,v 1.1.1.2 2005/01/26 13:42:24 nthiery Exp $
  File:		lib/COMBINAT/skewTableaux.mu
  Authors:	Francois Descouens < francois.descouens@wanadoo.fr >
                Teresa Gomez-Diaz <Teresa.Gomez-Diaz@univ-mlv.fr>
  Created:	24/09/2003
  License:	LGPL
*****************************************************************************/

export(output, Alias, asciiArt):

domain combinat::skewTableaux
    inherits Dom::BaseDomain;
    category Cat::CombinatorialClassWith2DBoxedRepresentation;
    axiom    Ax::canonicalRep;         // objects of this domain are not typed
    
    info_str := "Library 'combinat::skewTableaux': utilities for skew tableaux";
    
    interfaceAdd :=
        { hold(fromShapeAndWord), hold(shape), hold(toWord), hold(toChain),
	  hold(listSemiStandard),
	  hold(isASkewTableau), hold(isAStandardSkewTableau)
	};
  
// the type of objects in this domain is [innerP, filling] where
// innerP is the inner partition of a skew partition than can be built as
// [[zip(map(filling, nops), innerP, _plus, 0)], innerP]
// and the filling is a list of lists of NNI (in the case "standard"), to fill the
// rows of the skewPartition following the rules of standard tableaux
// NOTE: innerP  is listed from down to up
//       filling is listed from up to down

   // checks only list of list laying over a inner partition 
   typeSkewTabloid  := Type::Predicate("SkewTabloid", dom::isASkewTabloid);

   // checks typeTabloid plus shape
   typeSkewFilling  := Type::Predicate("SkewFilling", dom::isASkewFilling);

   // checks typeFilling plus entries from 1 to the size
   typeStandardSkewFilling := Type::Predicate("StandardSkewTableauFilling",  
					      dom::isAStandardSkewFilling);
  
   // checks typeFilling plus order constraints
   typeSkewTableau := Type::Predicate("SkewTableau", dom::isASkewTableau);

   // checks typeStandardFilling plus order constraints
   typeStandardSkewTableau := Type::Predicate("StandardSkewTableau", dom::isAStandardSkewTableau);
  
  
 isASkewTabloid:=
   proc(x : Type::AnyType)
   begin
      if x=[] then return(TRUE); end_if;
      if testtype(x,dom) then return(TRUE);
      elif nops(x) <> 2 then return(FALSE);
      //for canonical representation of a given skew partition
      elif nops(x[1]) > nops(x[2]) then return(FALSE); end_if;
      //first operand must be a partition (the inner one) and the second one must be a list of list.
      if not testtype(x[1], combinat::partitions) or not
   testtype(x[2], Type::ListOf(DOM_LIST)) or x = [[]] then return(FALSE); end_if;
      if contains(x[1], []) <> 0 then return(FALSE);  end_if;
      return(TRUE);
   end_proc;
  
 isASkewFilling:=
   proc(x : Type::AnyType)  
   begin
     if testtype(x,dom) or x = [] then return(TRUE); end_if;
       if dom::isASkewTabloid(x) then  
           return(testtype(dom::shape(x)[2], combinat::partitions));
       end_if;
       return(FALSE);
   end_proc;
  
  isAStandardSkewFilling:=
  proc(x : Type::AnyType)  
     local i,j; 
  begin
    if x = [] then return(TRUE); end_if;
    _lazy_and(dom::isASkewFilling(x), is({op(dom::toWord(x))} = {j$j=1.._plus(nops(x[2][i])$i=1..nops(x[2]) )} ));
  end_proc; 
 
 isASkewTableau :=
    proc (skt: Type::AnyType, shp = FAIL : Type::Union(DOM_FAIL, combinat::skewPartitions))
        local part, tab, size, length, k, start, i; 
    begin
        if skt = [] then return(TRUE); end_if;	
        if not testtype(skt, dom::typeSkewFilling) then return(FALSE); end_if;
        part := skt[1];
        tab  := skt[2];
        size := _plus(op(map(tab, nops))); // size of the skp
        if size = 0 then return(TRUE); end_if;
        length := nops(tab);
        part:= part.[0 $ length-nops(part)];
        // 2. test "tableaux" rules (order constraints)
        for k from 1 to length do
            if not(testtype(combinat::tableaux::new([tab[k]]), combinat::tableaux))
                then return(FALSE);
            end_if;
        end_for;
        for k from 1 to length-1 do
            start := part[length-k] - part[length-k+1];
            for i from 1 to min( part[length-k] + nops(tab[k]), part[length-k+1] + nops(tab[k+1]) ) do
                if i+start <= nops(tab[k]) and tab[k+1][i] >= tab[k][i+start] then
                    return(FALSE);
                end_if;
            end_for;
        end_for;
        return(TRUE);
    end_proc;
   
    isAStandardSkewTableau :=
    proc (skt: Type::AnyType, shp = FAIL : Type::Union(DOM_FAIL, combinat::skewPartitions))
        local part, tab, outerP, size, length, k, start, i; 
    begin
        if domtype(skt) <> DOM_LIST then return(FALSE); end_if;
        if nops(skt) <> 2 then return(FALSE); end_if;
        [part, tab]:= skt;   // NOTE: part is listed from down to up
                             //       tab  is listed from up to down
        // construction of the outer partition
        outerP:= zip(map(revert(tab), nops), part, _plus, 0);
        if outerP = [0] then outerP:= []; end_if;
        // 1. test if the shape is a skewPartition
        if not(testtype([outerP, part], combinat::skewPartitions)) then
            return(FALSE);
        end_if;
        size:= _plus(op(map(tab, nops))); // size of the skp
        if size = 0 then return(TRUE); end_if;
        // 2. test if the filling is a "standard" word
        if {op(map(tab,op))} <> {$1..size} then
            return(FALSE);
        end_if;
        length := nops(tab);
        part:= part.[0 $ length-nops(part)];
        // 3. test "tableaux" rules (order constraints)
        for k from 1 to length do
            if not(testtype(combinat::tableaux::new([tab[k]]), combinat::tableaux))
                then return(FALSE);
            end_if;
        end_for;
        for k from 1 to length-1 do
            start := part[length-k] - part[length-k+1];
            for i from 1 to min( part[length-k] + nops(tab[k]), part[length-k+1] + nops(tab[k+1]) ) do
                if i+start <= nops(tab[k]) and tab[k+1][i] >= tab[k][i+start] then
                    return(FALSE);
                end_if;
            end_for;
        end_for;
        return(TRUE);
    end_proc;
  
    //////////////////////////////////////////////////////////////////////////
    // typed objects specific functions
    /////////////////////////////////////////////////////////////////////////
    fromExpr:= x -> new(dom, op(x));
    
    new :=  
    proc(x: Type::Union(combinat::skewPartitions, dom::typeSkewFilling))
    begin  	   
       if testtype(x, combinat::skewPartitions) then
           if( x = [[],[]]) then return([]) ;	
           else return(dom::list(x)); end_if;	
       elif testtype(x, dom::typeSkewFilling) then
           return(dom::fromExpr(x));
       else
           error("illegal arguments");
       end_if;
   end_proc;
  
   op := 
   proc(x : dom, i = FAIL)
   begin	
     if i = FAIL then return(expr(x)[2][i] $i=1..nops(expr(x)[2])) ; end_if; 
     if i = 0 then 
       return(dom);
     else   
       return(expr(x)[2][i]); 
     end_if;
     return(FAIL);
   end_proc;
  
  nops := 
  proc(x:dom)
  begin
    return( nops(expr(x)[2]) );
  end_proc;
  
  convert :=
  proc(x: Type::Union(dom, dom::typeSkewFilling))
  begin  
    if domtype(x) = dom then return(x); 
    else return(dom::fromExpr(x)); end_if;  
  end_proc;
  
  expr  := [extop];
  _index := op; 
  
  set_index :=
  proc(t: dom, i: Type::PosInt, value)
  begin
    extsubsop(t, i=value);
  end_proc;

  expr2text :=
  proc(t: dom)
  begin  
    "new(".expr2text(dom).
    if extnops(t)>0 then
      ",".expr2text(extop(t));
    else
      ""
    end_if.
    ")";
  end_proc;
    ///////////////////////////////////////////////////////////////////////////
    //   shape function and toWord function
    //////////////////////////////////////////////////////////////////////////
  
    // function which returns the inner and the outer partition of a 
    // skewTableau
    shape := 
    proc( skt : Type::Union(dom, dom::typeSkewTabloid))
      local i,tmp;
    begin
      skt := dom::expr(skt);
      if skt[2] = [[]] then return([skt[1],skt[1]]); end_if;
      tmp := skt[1][i] + nops(skt[2][nops(skt[2])-i+1]) $i=1..nops(skt[1]);
      tmp := [tmp, nops(skt[2][nops(skt[2])-i+1])$i=nops(skt[1])+1..nops(skt[2])];
      return([skt[1], tmp]);
    end_proc;
  
    eval :=
    proc( skt: Type::Union(dom, dom::typeSkewTabloid))
      local word,i; 
    begin
      word := dom::toWord(skt);
      [nops(select(word, has, i)) $i=1..nops({op(word)})];
    end_proc;
	    
    //function which convert a skew tableau to a chain of increasing partitions
    toChain :=
    proc( skt : Type::Union(dom,dom::typeSkewTableau))
      local temp, inner, eval, res, i, k;
    begin
      inner := dom::shape(skt)[1];
      eval  := {op(dom::toWord(skt))};
      k     := 0;
      res := null();
      temp  := map(expr(skt)[2] ,x->select(x, has, [i$i=1..k]));
      while (temp <> [] and k <= nops(eval)) do
	inner := [op(inner),0$i=1..nops(temp)-nops(inner)];
	temp  := zip(inner, [ nops(temp[nops(temp)-i+1])$i=1..nops(temp) ],_plus );
	res   := res, select(temp,_not@has,0);
	k     := k + 1;
	temp  := map(expr(skt)[2] ,x->select(x, has, [i$i=1..k]));
      end_while;
      return([res]);
    end_proc;
  
    // function which returns the word corresponding to a given tableau 
    // with reading from top to floor
    toWord :=
    proc( skt : Type::Union(dom, dom::typeSkewTabloid))
    local i;  
    begin
      return([op(expr(skt)[2][i])$i=1..nops(expr(skt)[2])]);
    end_proc;
  
    //////////////////////////////////////////////////////////////////////////
    // fromShapeAndWord
    //////////////////////////////////////////////////////////////////////////

    fromShapeAndWord :=
    // type of result is not "dom" because the filling can be shuch that it does not
    // verify "standard tableaux" rules
    proc(skp: combinat::skewPartitions, word: combinat::words)
        local longs, i, j;
    begin     
	if word = [] then return([]); end_if;
        longs:= revert(combinat::skewPartitions::rowLenghts(skp));
        i:=0;
        combinat::skewTableaux([skp[2], [[op(word, i+1..(i:= i+j))] $ j in longs]]);
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // generator
    //////////////////////////////////////////////////////////////////////////
 
    toPoset :=
    // auxiliary function to generator
    // toPoset(rempl), where rempl is the list [part, tab] and part is
    // the inner partition of a skewPartition and tab is a list of
    // lists of NNI to fill the rows of the skewPartition.
    // This filling does NOT verify standard tableaux rules, and corresponds
    // to a box enumeration starting in top row and first column.
    // This function returns a poset (partial ordered set) [vertex, [edges]]
    // where vertex is a list with the box numbers
    // and [edges] gives the order relation between boxes
    // for example :
    //                     +---+
    //                     | 1 |
    //                     +---+---+---+---+
    //                         | 2 | 3 | 4 |
    //                         +---+---+---+---+
    //                             | 5 | 6 | 7 |
    //                             +---+---+---+
    // toPoset([2,1], [[1],[2,3,4],[5,6,7]])
    // gives : [[1,2,3,4,5,6,7], [[2,3],[3,4],[5,6],[6,7],[5,3],[6,4]]]
    // Note: The box enumeration can also be done in other order, because
    // the important part is the box order relationship (rather than the filling).
    proc(rempl)
        local start, vertex, edges, tab, part, i, k, length;
    begin
        [part, tab] := [extop(rempl)];
        length:= nops(tab);
        part:= part.[0 $ length - nops(part)];
        vertex:= map(tab, op);
        i:= 1;
        edges:= ([tab[i][k], tab[i][k+1]] $ k = 1..nops(tab[i])-1) $ i = 1..nops(tab);
        for k from 1 to length-1 do
	    start:= part[length-k] - part[length-k+1];
	    for i from 1 to min(part[length-k] + nops(tab[k]), part[length-k+1] + nops(tab[k+1])) do 
	        if i+start <= nops(tab[k]) then
	            edges:= edges, [tab[k+1][i], tab[k][i+start]];
	        end_if;
	    end_for;
        end_for;
        [vertex, [edges]];
    end_proc;

    generator :=
    proc(skp : combinat::skewPartitions) : combinat::generators
        local si, word, poset, gr, DOM;
        option escape;
    begin
        si:= combinat::skewPartitions::size(skp);
        if si = 0 then
           return(combinat::generators::singleton(new(dom)));
       end_if;
        word:= [$ 1..si];
        poset:= op(dom::toPoset(dom::fromShapeAndWord(skp, word)));
	gr:= Graph(poset, hold(Directed));
        DOM:= dom;
        combinat::generators::map(
            combinat::linearExtensions::generator(gr),
            w -> DOM::fromShapeAndWord(skp, combinat::permutations::inverse(w)));   
    end_proc;
  
    //function for listing semi-standard young skew tableaux with a specialization 
    //of the ribbons tableaux list algorithm at k=1
  
    listSemiStandard := 
    proc(part, evaluation: combinat::compositions)
	local a, b, c, retire, perms, tmp, selection, work, h, k, i, j, tab, inner, outer, inner2, partp;
    begin
        outer := part[1];
        inner := part[2];
      
        if(_plus(outer[i] $i = 1..nops(outer)) - _plus(inner[i] $i = 1..nops(inner)) <> _plus(evaluation[i] $i = 1..nops(evaluation))  )
	  then return([]); end_if;
        if testtype(part, combinat::skewPartitions) = FALSE then return([]); end_if;
        //Initialization of the first empty tableaux and the working vector
        tab   := [ [0 $ i=1..outer[nops(outer)-j+1]] $ j=1..nops(outer)];
        work  := [[tab,outer]]; 
      
          for k from nops(evaluation) downto 1 do
            tmp:=null();
	    for i from 1 to nops(work) do
	        partp     := combinat::partitions::conjugate(work[i][2]);
                selection := null();
                perms     := combinat::permutations::list(
                             [0 $ (nops(partp)-evaluation[k]), 1 $ evaluation[k]]);
                //selection of valide permutation vectors
                for j from 1 to nops(perms) do
		    retire := zip(partp, perms[j], _subtract);
		    //without ending zeros
		    a      := nops(retire);
		    while a >=1 and retire[a] = 0  do a := a - 1; end_while;
		    retire := [retire[b]$b=1..a];
		    inner2 := [op(inner) , 0$b=1..nops(retire)-nops(combinat::partitions::conjugate(inner))];
		    retire := [op(retire), 0$b=1..nops(combinat::partitions::conjugate(inner))-nops(retire)];
		  if testtype(retire, combinat::partitions)=TRUE and  
		     [select((op(zip(retire,combinat::partitions::conjugate(inner2),_subtract))), testtype, Type::Negative)] = []
		    then 
		     selection := selection, [retire,perms[j]];
      	            end_if;
		end_for;
                selection := [selection];
                //we list all valid tableau at this step (incomplete evaluation)          
	        for h from 1 to nops(selection) do
		    c   := combinat::ribbonsTableaux::insertionTableau(partp,selection[h][2], k, work[i][1]);
		    if c <> FAIL then 
		      tmp := tmp, [c, combinat::partitions::conjugate(selection[h][1])];  
		    end_if;
                end_for;
            end_for;
	    work := [tmp];
	  end_for;
      for i from 1 to nops(work) do
	work[i][1] := map(work[i][1], x->select(x,_not@has, 0));
      end_for;
    return(map([ [inner,work[i][1]] $ i=1..nops(work) ], dom::new ));
    end_proc;
    
  
    //////////////////////////////////////////////////////////////////////////
    // Pretty-print
    //////////////////////////////////////////////////////////////////////////

    toArray :=
    proc(skt : Type::Union(dom, dom::typeSkewFilling)
	) : DOM_ARRAY
        local row, col, q, fil, width, height;
    begin  
        [q, fil] := [extop(skt)];
	// Expand the inner partition with zeroes, and put it in the
	// same order as p
        q := revert(q.[0 $ nops(fil)-nops(q)]);
        // Compute the height and width of the skew tableau
	// Could go in separate methods?
	// Note the max(1, instead of max(0, to please array
	// See Cat::CombinatorialClassWith2DBoxedRepresentation
	height := max(1, nops(fil));
        width  := max(1, op(zip(map(fil, nops), q, _plus)));
        array(1..height, 1..width,
	      (row, col+q[row]) = fil[row][col]
	      $ col = 1..nops(fil[row])
	      $ row = 1..nops(fil));
    end_proc;

    //toArrayVerticalAlignment := -1; also in the definition of combinatorialClassWith2DboxedRepresentation

end_domain:

