/*****************************************************************************
  $Id: subsets.mu,v 1.1.1.7 2005/01/26 16:56:06 nthiery Exp $
  File:		lib/COMBINAT/subsets.mu
  Authors:	Nicolas M. Thiry <nthiery@users.sourceforge.net>
  Created:	16/02/2001
  License:	LGPL
*****************************************************************************/

/**
 A subset of a set s is, well, a subset. This package serves both for
 subsets of a fixed size, or subsets of any size (i.e. the powerset).

 Right now, sets are generated lexicographically w.r.t. the internal
 order of the set provided as argument. Should we specify the order in
 which sets are returned ? Should we use the natural order of the
 elements instead ?


 Powerset of a set:

 - count(n)		number of subsets of {1..n}
 - count(s)		number of subsets of the set s
 - list(n)		list of all subsets of {1..n}
 - list(s)		list of all subsets of the set s
 - generator(n)		generator for the subsets of {1..n}
 - generator(s)		generator for the subsets of the set s

 - first(n)		first subset of {1..n}
 - first(s)		first subset of the set s
 - last(n)		last subset of {1..n}
 - last(s)		last subset of the set s

 Subsets of size k of a set:

 - count(n,k)		number of subsets of size k of {1..n}
 - count(s,k)		number of subsets of size k of the set s
 - list(n,k)		list of all subsets of size k of {1..n}
 - list(s,k)		list of all subsets of size k of the set s
 - generator(n,k)	generator for the subsets of size k of {1..n}
 - generator(s,k)	generator for the subsets of size k of the set s

 - first(n,k)		first subset of size k of {1..n}
 - first(s,k)		first subset of size k of the set s
 - last(n,l)		last subset of size k of {1..n}
 - last(s,k)		last subset of size k of the set s

**/

domain combinat::subsets
    inherits Dom::BaseDomain;
    category Cat::GradedCombinatorialClass;
    axiom    Ax::systemRep;

    info_str := "Library 'combinat::subsets': utilities for subsets of a set";

    interfaceAdd :=
        {hold(first), hold(last)};

    domtype := DOM_SET;
    typeSet  := Type::Union(Type::NonNegInt, DOM_SET);
    typeSize := Type::Union(Type::NonNegInt, DOM_NULL);

    isA      := 
    proc(sub             : Type::AnyType, 
	 set    = null() : Type::Union(DOM_NULL, dom::typeSet), 
	 size   = null() : dom::typeSize) 
    begin
	if domtype(sub) <> DOM_SET then return(FALSE) end_if;
	if args(0) = 1 then return(TRUE) end_if;
	if args(0) > 2 and nops(sub) <> size then return(FALSE); end_if;
	if domtype(set) = DOM_INT then set := {$1..set}; end_if;
	if sub minus set = {} then TRUE else FALSE; end_if;
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // count
    //////////////////////////////////////////////////////////////////////////

    count :=
    proc(set: dom::typeSet, k=null(): dom::typeSize)
        : Type::NonNegInt
        local n;
    begin
        if testargs() then
            if args(0) > 2 then
                error("Wrong number of arguments") ;
            end_if ;
        end_if;
        if not testtype(set, DOM_SET) then
            n:=set;
        else
            n:=nops(set);
        end_if;
        if k=null() then
            2^n
        else
            binomial(n,k);
        end_if;
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // first
    //////////////////////////////////////////////////////////////////////////
    
    first :=
    proc(set: dom::typeSet, k=null(): dom::typeSize)
        : Type::Union(dom, DOM_FAIL)
        local w, i;
    begin
        if testargs() then
            if args(0) > 2 then
                error("Wrong number of arguments") ;
            end_if ;
        end_if ;

        if domtype(set)<>DOM_SET then
            set:={$1..set}
        end_if;

        w:=combinat::subwords::first([$1..nops(set)], args(2..args(0)));
        if w=FAIL then
            FAIL
        else
	    {op(set, i) $ i in w};
        end_if;
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // last
    //////////////////////////////////////////////////////////////////////////
    
    last :=
    proc(set: dom::typeSet, k=null(): dom::typeSize)
        : Type::Union(dom, DOM_FAIL)
        local w, i;
    begin
        if testargs() then
            if args(0) > 2 then
                error("Wrong number of arguments") ;
            end_if ;
        end_if ;

        if domtype(set)<>DOM_SET then
            set:={$1..set}
        end_if;

        w:=combinat::subwords::last([$1..nops(set)], args(2..args(0)));
        if w=FAIL then
            FAIL
        else
            {op(set, i) $ i in w};
        end_if;
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // generator
    //////////////////////////////////////////////////////////////////////////

    generator :=
    proc(set: dom::typeSet, k=null(): dom::typeSize)
        : combinat::generators
	option escape;
	local i;
    begin
        if testargs() then
            if args(0) > 2 then
                error("Wrong number of arguments") ;
            end_if ;
        end_if;

        if domtype(set)<>DOM_SET then
            set:={$1..set}
        end_if;
        // We construct a generator for the subwords of [1..nops(n)]
        // and map each of them into a subset of set
        return(combinat::generators::map
               (combinat::subwords::generator([$1..nops(set)],
                                              args(2..args(0))),
		indexList->{op(set, i) $ i in indexList}));
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // unrank
    //////////////////////////////////////////////////////////////////////////

    unrank := 
    proc(r : Type::PosInt, set: dom::typeSet, k=null(): dom::typeSize)
	local i, n, res, bin;
    begin
	if testargs() then
	    if args(0) > 3 then
                error("Wrong number of arguments") ;
            end_if ;
        end_if ;

        if domtype(set)<>DOM_SET then
            set:={$1..set}
        end_if;
	
	n := nops(set);
	if k=null() then
	    for k from 0 to n do
		bin := combinat::binomialRecursive(n,k);
		if r > bin then r := r - bin 
		else return({op(set, i) $ i in combinat::chooseNK::unrank(r, n, k)});
		end_if;
	    end_for;
	    return(FAIL);
	end_if;
	
	res := combinat::chooseNK::unrank(r, n, k); 
	if res <> FAIL then {op(set, i) $ i in res}; else FAIL; end_if;
    end;

    //////////////////////////////////////////////////////////////////////////
    // random
    //////////////////////////////////////////////////////////////////////////
    
    random :=
    proc(set: dom::typeSet, k=null(): dom::typeSize)
        : Type::Union(dom, DOM_FAIL)
        local res, n, i;
    begin
        if testargs() then
            if args(0) > 2 then
                error("Wrong number of arguments") ;
            end_if ;
        end_if ;

        if domtype(set)<>DOM_SET then
            set:={$1..set}
        end_if;

	if k=null() then
	    return(select(set, iszero@random(2)));
// The previous version is shorter and 3 % faster...	    
//	    r := random(2);
//	    return(map(set, z -> if r() = 0 then z else null(); end_if));
	end_if;
	
	n := nops(set);
	res := combinat::chooseNK::random(n, k);
        if res <> FAIL then {op(set, i) $ i in res}; else FAIL end_if;
    end_proc;
    
    
begin
end_domain:

