/*****************************************************************************
  $Id: subwords.mu,v 1.1.1.5 2005/01/26 16:56:05 nthiery Exp $
  File:		lib/COMBINAT/subwords.mu
  Authors:	Nicolas M. Thiry <nthiery@users.sourceforge.net>
                Florent Hivert <florent.hivert@univ-mlv.fr>
  Created:	16/02/2001
  License:	LGPL
*****************************************************************************/

domain combinat::subwords
    inherits Dom::BaseDomain;
    category Cat::CombinatorialClass;
    axiom    Ax::systemRep;

    info_str := "Library 'combinat::subwords': utilities for subwords of a word";

/** 
    A subword of a word w is a word obtaining by deleting the letters at
    some of the positions in w.

    Example:
     - [b,n,j,r] is a subword of [b,o,n,j,o,u,r]
     - [n,r] is a subword of [b,o,n,j,o,u,r]
     - [r,n] is not a subword of [b,o,n,j,o,u,r]
     - [n,n,u] is not a subword of [b,o,n,j,o,u,r]
     - [n,n,u] is a subword with repetitions of [b,o,n,j,o,u,r]
**/

    interfaceAdd :=
        {hold(first), hold(last), hold(smallestPositions)
         };

    domtype  := DOM_LIST;
    typeSize := Type::Union(Type::NonNegInt, DOM_NULL);
    isA      := 
    proc(subword       : Type::AnyType, 
	 word = null() : Type::Union(DOM_LIST, DOM_NULL),
	 len  = null() : Type::Union(Type::NonNegInt, DOM_NULL)) 
    begin
	if domtype(subword) <> DOM_LIST then return(FALSE) end_if;
	if args(0) = 1 then return(TRUE) end_if;
	if args(0) > 2 and nops(subword) <> args(3) then return(FALSE); end_if;
	if dom::smallestPositions(word, subword) <> FAIL then TRUE else FALSE; end_if;
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // count
    //////////////////////////////////////////////////////////////////////////

    count :=
    proc(w: dom, k=null(): dom::typeSize)
        : Type::NonNegInt
    begin
        if testargs() then
            if args(0) > 2 then
                error("Wrong number of arguments") ;
            end_if ;
        end_if;
        if k=null() then
            2^(nops(w))
        else
            binomial(nops(w),k);
        end_if;
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // first
    //////////////////////////////////////////////////////////////////////////
    
    first :=
    proc(w: dom, k=null(): dom::typeSize)
        : Type::Union(dom, DOM_FAIL)
    begin
        if testargs() then
            if args(0) > 2 then
                error("Wrong number of arguments") ;
            end_if ;
        end_if ;

        if k=null() then
            return([]);
        elif k>nops(w) then
            return(FAIL);
        end_if;
        return([op(w, 1..k)]);
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // last
    //////////////////////////////////////////////////////////////////////////
    
    last :=
    proc(w: dom, k=null(): dom::typeSize)
        : Type::Union(dom, DOM_FAIL)
    begin
        if testargs() then
            if args(0) > 2 then
                error("Wrong number of arguments") ;
            end_if ;
        end_if;

        if k=null() then
            return(w);
        elif k>nops(w) then
            return(FAIL);
        end_if;
        return([op(w, nops(w)-k+1..nops(w))]);
    end_proc;
    
    //////////////////////////////////////////////////////////////////////////
    // generator
    //////////////////////////////////////////////////////////////////////////
    
    generator :=
    proc(w: dom, k=null(): dom::typeSize)
        : combinat::generators
        option escape;
        local generator, i;
    begin
        if testargs() then
            if args(0) > 3 or (args(0)=3 and args(3) <> Repeat) then
                error("Wrong number of arguments") ;
            end_if ;
        end_if;

        // Case 1: build recursively a generator for all subwords
        if k=null() then
            return(combinat::generators::concat(dom::generator(w,k) $k=0..nops(w)));
        end_if;

        // Case 2: build a generator for the subwords of length k
        // We construct a generator for the subwords of [1..nops(n)]
        // and map each of them into a subword of w
        if args(0)<3 then
            generator:=combinat::chooseNK::generator(nops(w),k);
        else
            generator:=combinat::chooseNK::generator(nops(w),k, Repeat);
        end_if;
        return(combinat::generators::map
	       (generator,
		subword -> [op(w,i) $ i in subword]));
    end_proc;


    //////////////////////////////////////////////////////////////////////////
    // smallestPositions
    // if subword appear in word starting with position pos,
    // returns the smallest positions for where subword appear in word.
    // -> should perhaps be a new combinatorial class (positions::first)
    // -> to be compared with word::splitting. 

    smallestPositions := 
    proc(word : dom, subword : dom, pos = 1 : Type::NonNegInt)
	local i;
    begin
	pos := pos - 1;
	for i from 1 to nops(subword) do
	    pos := contains(word, subword[i], pos + 1);
	    if pos=0 then return(FAIL) end_if;
	    subword[i] := pos;
	end_for;
	subword;
    end_proc;


    //////////////////////////////////////////////////////////////////////////
    // random
    //////////////////////////////////////////////////////////////////////////

    random :=
    proc(w: dom, k=null(): dom::typeSize) : dom
	local n, i, res;	
    begin
	if testargs() and args(0) > 3 then
		error("Wrong number of arguments") ;
        end_if;
	if k=null() then
	    return(select(w, iszero@random(2)));
//	    r := random(2);
//	    return(map(w, z -> if r() = 0 then z else null(); end_if));
	end_if;
	
	n := nops(w);
	res := combinat::chooseNK::random(n, k); 
	if res <> FAIL then [op(w, i) $ i in res]; else FAIL; end_if;
    end_proc;


    //////////////////////////////////////////////////////////////////////////
    // unrank
    //////////////////////////////////////////////////////////////////////////

    unrank := 
    proc(r : Type::PosInt, w: dom, k = null() : dom::typeSize)
	local bin, n, i, res;
    begin
        if testargs() and args(0) > 3 then
	    error("Wrong number of arguments") ;
        end_if;
	n := nops(w);
	if k=null() then
	    for k from 0 to n do
		bin := combinat::binomialRecursive(n,k);
		if r > bin then r := r - bin 
		else return([op(w, i) $ i in combinat::chooseNK::unrank(r, n, k)]);
		end_if;
	    end_for;
	    return(FAIL);
	end_if;
	
	res := combinat::chooseNK::unrank(r, n, k); 
	if res <> FAIL then [op(w, i) $ i in res ]; else FAIL; end_if;
    end_proc;

begin
end_domain:

