/*****************************************************************************
 $Id: tableaux.mu,v 1.1.1.7 2005/01/26 13:42:24 nthiery Exp $
 File: lib/COMBINAT/tableaux.mu
 Authors: Nicolas M. Thiry <nthiery@users.sourceforge.net>
  Florent Hivert <florent.hivert@univ-mlv.fr>
 Created: 31/10/2001
 License: LGPL
*****************************************************************************/

export(combinat, Alias,
       permutations, partitions, skewPartitions, compositions):
export(output, Alias, asciiArt):

protect(Symmetrica): // To silence prog::check when this file is
		     // included in the main MuPAD library without
		     // the full package being loaded

// Note :
// The empty partition is forbiden as argument to construct a tableau.

domain combinat::tableaux
   inherits Dom::BaseDomain;
   category Cat::CombinatorialClassWith2DBoxedRepresentation;
   axiom    Ax::canonicalRep;

   info_str := "Library 'combinat::tableaux': utilities for tableaux";

   interfaceAdd :=
       {
	hold(isA), hold(isAStandardTableau),
        hold(new), hold(convert), hold(expr),
        hold(canonical), hold(rowCanonical), hold(columnCanonical),
        hold(indexFilling), hold(cocharge), hold(kostkaPol),
	hold(conjugate), hold(toWord), hold(fromWord), hold(shape), hold(toChain),
	hold(schensted), hold(invSchensted),
	hold(insertWord), hold(fromShapeAndWord), hold(toPoset),
	hold(rowStabilizer), hold(columnStabilizer),
	hold(listSemiStandard)
        };

    interfaceAddIfExists :=
       {
	hold(kostka)
       };

   // checks only list of list
   typeTabloid  := Type::Predicate("Tabloid", dom::isATabloid);

   // checks typeTabloid plus shape
   typeFilling  := Type::Predicate("TableauFilling", dom::isAFilling);

   // checks typeFilling plus entries from 1 to the size
   typeStandardFilling := Type::Predicate("StandardTableauFilling",  
                                                        dom::isAStandardFilling);
   // checks typeStandardFilling plus order constraints
   typeStandardTableau := Type::Predicate("StandardTableau", dom::isAStandardTableau);

   // checks typeFilling plus order constraints
   typeTableau := Type::Predicate("Tableau", dom::isATableau);

   isATabloid:=
   proc(x : Type::AnyType)
   begin
      if not testtype(x, Type::ListOf(DOM_LIST)) then return(FALSE); end_if;
      if contains(x, []) <> 0 then return(FALSE); end_if;
      return(TRUE);
   end_proc;

   shape :=
   proc(x: Type::Union(dom::typeTabloid, dom))
   begin
       // expr(x) is a list of non-empty lists
       revert(map(expr(x), nops));
   end_proc;
  
   toChain := 
   proc (tab: Type::Union(dom, Type::ListOf(DOM_LIST) ))
   begin
     combinat::skewTableaux::toChain([[],expr(tab)])
   end_proc;
  
   isAFilling:=
   proc(x : Type::AnyType)
   begin
       // if x of type dom, it has the right shape
       if testtype(x, dom) then return(TRUE); end_if;
       if dom::isATabloid(x) then
           return(testtype(dom::shape(x), combinat::partitions));
       end_if;
       return(FALSE);
   end_proc;

   // entries from 1 to the size
   // hasConsecutiveIntegerEntries? hasIntegerEntries?
   numberEntries:=
   proc(x : dom::typeFilling)
       local numbers, shap;
   begin
       numbers := map(expr(x), op);
       shap:= dom::shape(x);
       if {op(numbers)} <> {$1.._plus(op(shap))} then return(FALSE); end_if;
       return(TRUE);
   end_proc;

   isAStandardFilling:=
   proc(x : Type::AnyType)
   begin
       _lazy_and(dom::isAFilling(x), dom::numberEntries(x));
   end_proc;

   isAStandardTableau :=
   proc(x : Type::AnyType,
	shp = FAIL : Type::Union(DOM_FAIL, combinat::partitions))
   begin
     _lazy_and(dom::isAStandardFilling(x), dom::isATableau(x, shp));
   end_proc;

  isATableau :=
  proc(x : Type::AnyType,
       shp = FAIL : Type::Union(DOM_FAIL, combinat::partitions))
    local shap, line, col, fill;
  begin
    if not testtype(x, dom::typeFilling) then return(FALSE); end_if;
    shap := dom::shape(x);
    if args(0) >= 2 and (shap <> shp) and shp<>FAIL then 
      return(FALSE); end_if;
    fill:= revert(shap);
    for line from 1 to nops(fill)-1 do
      for col from 1 to fill[line]-1 do
	if    (x[line][col] > x[line][col+1])
	   or (x[line][col] <= x[line+1][col])
	  then return(FALSE);
	end_if;
      end_for;
      if x[line][col] <= x[line+1][col]
	then return(FALSE);
      end_if;
    end_for;
    if fill <> [] then
      for col from 1 to fill[line]-1 do
	if x[line][col] > x[line][col+1]
	  then return(FALSE);
	end_if;
      end_for; end_if;
    return(TRUE);
  end_proc;
  
  isA := dom::isATableau;

   //////////////////////////////////////////////////////////////////////////
   // no Ax::systemRep functions
   //////////////////////////////////////////////////////////////////////////

   // to be used internally, x of type typeFilling,
   // doesn't deal with [] nor [[]]
  fromExpr:= x -> new(dom, op(x));

   // override the shortcut from Cat::CombinatorialClass
   new :=  
   // argument is either a partition (so it calls list)
   // or a internal representation, ie a filling
   // and so new of [3],[2,4] constructs an element
   // of the domain and pretty-prints it
   proc(x: Type::Union(combinat::partitions, dom::typeFilling))
   begin  
       if x = [] then return(new(dom)); end_if;
       if testtype(x, combinat::partitions) then
           return(dom::list(x));
       elif testtype(x, dom::typeFilling) then
           return(dom::fromExpr(x));
       else
           error("illegal arguments");
       end_if;
   end_proc;

   convert :=
   proc(x: Type::Union(dom, dom::typeFilling))
   begin  
       if x = [] then return(new(dom)); end_if;
       case domtype(x)
       of dom do
           return(x);
       of dom::typeFilling do
           return(dom::fromExpr(x));
       end_case;
       FAIL;
   end_proc;

   expr:= [extop];

   // other interesting functions

   // to be able to say t[1] for t tableau
   _index:= extop;

   // to be able to say t[1]:= for t tableau
   set_index :=
   proc(t: dom, i: Type::PosInt, value)
   begin
       extsubsop(t, i=value);
       //extsubsop(t, 1=subsop(extop(t,1), i=value));
   end_proc;

   expr2text :=
   proc(t: dom)
   begin  
  "new(".expr2text(dom).
   if extnops(t)>0 then
              ",".expr2text(extop(t));
           else
              ""
           end_if.
           ")";
   end_proc;

   map :=
   proc(t: dom, f)
   begin
       dom::fromExpr(map(expr(t), f, args(3..args(0))));
   end_proc;

   //////////////////////////////////////////////////////////////////////////
   // Pretty-print
   //////////////////////////////////////////////////////////////////////////

   toArray :=
   proc(p: dom::typeFilling) : DOM_ARRAY
       local col, row;
   begin
       // Note the max(1, instead of max(0, to make array happy
       // See Cat::CombinatorialClassWith2DBoxedRepresentation
       array(1..max(1, nops(p)), 1..max(1, op(map(p, nops))),
             (row, col) = p[row][col]
             $ col=1..nops(p[row])
             $ row = 1..nops(p));
   end_proc;

   toArrayVerticalAlignment := -1; // alignement on the bottom row of the array

   //////////////////////////////////////////////////////////////////////////
   // count using the hook formula
   //////////////////////////////////////////////////////////////////////////
   count :=
  proc(lambda: combinat::partitions)
    : Type::PosInt
    local lambdaConj, n, i, j;
  begin
    n := _plus(op(lambda));
    lambdaConj := combinat::partitions::conjugate(lambda);
    n! / _mult(lambda[i] + lambdaConj[j] - i - j + 1 $ j = 1..lambda[i] $ i = 1..nops(lambda))
  end_proc;

   //////////////////////////////////////////////////////////////////////////
   // generator
   //////////////////////////////////////////////////////////////////////////

   generator :=
   proc(lambda: combinat::partitions)
       : combinat::generators
       option escape;
       local n, corow, row, slambda, gen, DOM;
   begin
       if nops(lambda)=0 then
           return(combinat::generators::singleton(new(dom)));
       end_if;

       DOM := dom; // Usual hack since dom is not lexically bound
              // inside subfunctions.

       n:=_plus(op(lambda));
       row:=1;
       corow:=nops(lambda);
       slambda := [op(lambda, 1..nops(lambda)-1)];
       if (lambda[nops(lambda)])>1 then
           slambda:=append(slambda,lambda[nops(lambda)]-1);
       end_if;
       gen:=dom::generator(slambda);
       // Invariant:
       //  - gen is a valid generator for the tableaux with one less
       //    box on the row row; those are of shape slambda
       //  - corow = nops(lambda)-row+1
       proc() : Type::Union(dom, DOM_FAIL)
           local t;
       begin
           while (t:=gen())=FAIL do
               repeat
                   row:=row+1;
                   corow:=corow-1;
                   if corow=0 then
                       return(FAIL);
                   end_if;
               until lambda[corow]>lambda[corow+1] end;
               slambda := lambda;
               slambda[corow] := slambda[corow]-1;
               gen := DOM::generator(slambda);
           end_while;
           if nops(slambda) < nops(lambda) then // better test ?
               if op(t) = [] then t:= dom::fromExpr([[n]]);
               else t:= DOM::fromExpr([[n],op(t)]);
               end_if;
           else
               t[row]:= append(t[row], n);
           end_if;
           return(t);
       end_proc;
   end_proc;
 
  listSemiStandard :=
  proc(lambda, mu = FAIL : Type::Union(combinat::compositions ,DOM_FAIL))
    local liste;
  begin
    if mu = FAIL then 
      combinat::tableaux::list(lambda);
    else
      liste := combinat::skewTableaux::listSemiStandard([lambda,[]],mu);
    end_if;
    map(liste, x->combinat::tableaux::new([op(x)]));
  end_proc;

    /**
     * Computation of kostka coeffs from Littlewood-Richardson coeffs
     *
     * Thanks to Frdric Toumazet for the trick:
     *
     * K_lambda^mu = C^{tau(mu)}_{lambda, sigma(mu)}, where
     * sigma(mu) and tau(mu) are the inner and outer shape of the
     * skew partition built of disconnected rows of length mu[1],\dots,m[k]
     *
     * Example: K_{(3,2)}^{(2,2,1)} = C^{(5,3,1)}_{(3,2),(3,1)}
     **/

    kostkaFromLrcoef :=
    proc(lambda: partitions, mu: compositions) : Type::NonNegInt
	local tau, sigma, i;
    begin
	// tau: partial sums of mu, from the right
	// sigma: tau without its first part
	tau   := [ NIL $ nops(mu) ];
	if nops(mu) > 0 then
	    tau[nops(mu)] := mu[nops(mu)];
	    for i from nops(mu)-1 downto 1 do
		tau[i] := tau[i+1] + mu[i];
		sigma[i] := tau[i+1];
	    end_for;
	end_if;
	sigma := [op(tau, 2..nops(tau))];
	combinat::lrcalc::lrcoef(tau, lambda, sigma);
    end_proc;

    kostkaFromSymmetrica :=
    proc(lambda: Type::Union(partitions, skewPartitions), mu: compositions) : Type::NonNegInt
    begin
	if nops(mu) = 0 or nops(lambda) = 0 then
	    // Work around: Symmetrica::KostkaNumber does not deal well
	    // with empty partitions
	    // This should be fixed in Symmetrica instead ...
	    if nops(lambda) = nops(mu) then 1 else 0; end_if;
	else
	    Symmetrica::KostkaNumber(lambda,mu)
	end_if;
    end_proc;

    kostkaFromListNonStandardTableaux :=
    proc(lambda: partitions, mu: compositions) : Type::NonNegInt
    begin
	nops(dom::listSemiStandard(lambda, mu));
    end_proc;

    kostka :=
    if combinat::lrcalc <> FAIL then
	dom::kostkaFromLrcoef;
    elif domtype(Symmetrica) = DOM_DOMAIN then
	dom::kostkaFromSymmetrica;
    else
	dom::kostkaFromListNonStandardTableaux;
    end_if;
    
  kostkaPol :=
  proc(lambda, mu, t)
    local tab,i;
  begin
    tab := map(dom::listSemiStandard(lambda, mu), combinat::words::charge @ revert @ dom::toWord);
    return(_plus(t^(tab[i]) $i=1..nops(tab)));
  end_proc;

  canonical := dom::rowCanonical;
  
  rowCanonical :=
  proc(lambda: combinat::partitions): dom::typeStandardTableau
       // Canonical tableau of shape lambda, filled with 1..n from
       // the bottom row to the upper row.
       //
       // Example: canonical([3,2]) = [[4,5], [1,2,3]]
   local result, i, j;
   begin
       if nops(lambda) = 0 then return(new(dom)); end_if;
       result := [];
       i:=_plus(op(lambda));
       for j in revert(lambda) do
           result := append(result, [ $ i-j+1..i ] );
           i:=i-j;
       end_for;
       dom::fromExpr(result);
   end_proc;

   columnCanonical :=
   proc(lambda: combinat::partitions): dom::typeStandardTableau
       // column-canonical standard tableau of shape lambda, filled
       // with 1..n from the leftmost column to the rightmost column
       //
       // Example: columnCanonical([3,2]) = [[2,4], [1,3,5]]
   local result, i, row, col;
   begin
       if nops(lambda) = 0 then return(new(dom)); end_if;
       // Allocates directly the space required for the resulting tableau
       // That's probably faster than appending the elements to the rows
       // one by one
       result := [[NIL$i] $ i in revert(lambda)];
       i:=1;
       for col from 1 to lambda[1] do
           for row from nops(result) downto 1 do
               if col>nops(result[row]) then
                   break;
               end_if;
               result[row][col]:=i;
               i:=i+1;
           end_for;
       end_for;
       dom::fromExpr(result);
   end_proc;

   indexFilling :=
   proc(S: dom::typeStandardTableau) : dom::typeFilling
       // Returns the index filling I(S) of the standard tableau S.
       // Reference: Higher Specht Polynomials for the symmetric group
       // and the wreath product, S. Ariki, T. Terasoma, H. Yamada.
       //
       // Example:
       //   indexFilling([[3,5],[1,2,4]]) = [[1, 2], [0, 0, 1]]
       local n, result, row_indexes, previous_col, row, col, i, index;
   begin
       if testtype(S, dom::typeStandardTableau) then S:= dom::expr(S); end_if;
       n := _plus(op(map(S, nops)));
       result := S; // Tableau of same shape at S
       
       row_indexes := [1 $ nops(S)-1,2];
       previous_col :=1;
       index:=0;
       result[nops(result)][1]:=index;
       for i from 2 to n do
           // Invariants:
           //  - previous_col is the column of the box containing i-1 in S
           //  - the boxs [r,c] of S with c<row_indexes[r] contains 1..i-1
   //  - result is properly filled in the boxs containing 1..i-1 in S

           // Look for the position of i in S
           for row from nops(S) downto 1 do
               col := row_indexes[row];
               if col <= nops(S[row]) and
                  S[row][col] = i then
                   row_indexes[row]:=col+1;
                   break;
               end_if;
           end_for;
           // Since S is standard, we know that S[row][col]=i;
           if S[row][col]<>i then
               error("the tableau is not standard");
           end_if;
           if col <= previous_col then
               index:=index+1;
           end_if;
           result[row][col] := index;
           previous_col := col;
       end_for;
       result;
   end_proc;

   cocharge :=
   proc(S: dom::typeStandardTableau) : Type::NonNegInt
       // The cocharge of the tableau S (or the charge ???)
       // Reference:
   begin
       _plus(op(map(dom::indexFilling(S), op)));
   end_proc;

   fromShapeAndWord :=
  proc(shap: combinat::partitions, word: combinat::words) : dom
    local i, j;
  begin
    if _lazy_and(testargs(), _plus(op(shap))<>nops(word)) then
      error("Length mismatch".expr2text(shap)." and ".expr2text(word));
    end_if;
    if nops(shap) = 0 then return(new(dom)); end_if;
    i:=0;
    dom::fromExpr([[op(word,i+1..(i:=i+j))] $ j in revert(shap)]);
  end_proc;

  fromWord :=
   proc(p: combinat::words) : dom
       local T, i;
   begin
       if nops(p)=0 then return(new(dom)); end_if;
       T:=[[p[1]]];
       for i from 2 to nops(p) do
           if p[i-1]>p[i] then
               T:=append(T,[p[i]]);
           else
               T[nops(T)]:=append(T[nops(T)], p[i]);
           end_if;
       end_for;
       dom::fromExpr(T);
   end_proc;

   toWord :=
   proc(T: Type::Union(dom::typeFilling, dom)) : combinat::words
   begin
       if testtype(T, dom) then map(dom::expr(T), op);
       else map(T, op); end_if;
   end_proc;

// other possible design: insert(tab, _exprseq)...
// ask on mailing list...
  insertLetter :=
  proc(T: dom::typeFilling, letter) : dom
    local result, lenline, nline, i, exT;
  begin
    exT:= dom::expr(T);
    if exT=[] then return(dom::fromExpr([[letter]])); end_if;
    result := [];
    for nline from nops(exT) downto 1 do
      lenline := nops(exT[nline]);
      if letter >= exT[nline][lenline] then
	return(dom::fromExpr([op(exT,1..nline-1),
			      append(exT[nline], letter),
			      op(result)]));
      else
	for i from 1 to lenline do
	  if exT[nline][i] > letter then
	    result := append([subsop(exT[nline], i=letter)],
			     op(result));
	    letter := exT[nline][i];
	    break;
	  end_if;
	end_for;
      end_if;
    end_for;
    dom::fromExpr([[letter], op(result)]);
  end_proc;

   insertWord :=
  proc(T: dom::typeFilling,
       w :combinat::words) : dom
    local i;
  begin
    for i from 1 to nops(w) do
      T := dom::insertLetter(T, w[i]);
    end_for;
    T;
  end_proc;

   invSchenstedExtractLetter :=
   // extract the letter at the end of the i-th line in the tableau P
   // the lines are counted from the top, starting from 1.
  proc(P : Type::ListOf(Type::ListOf(Type::PosInt)),
               // dom is too strong because of the
               // remaining [] when there is an empty line
       iLine : Type::PosInt)
    local letter, newLetter, i;
  begin
    letter := P[iLine][nops(P[iLine])];
    P      := subsop(P, iLine = [op(P[iLine], 1..nops(P[iLine])-1)]);
    while(iLine < nops(P)) do
      iLine := iLine + 1;
      i := nops(P[iLine]);
      while P[iLine][i] >= letter do
	i := i - 1;
      end_while;
      newLetter := P[iLine][i];
      P := subsop(P, iLine = subsop(P[iLine], i=letter));
      letter := newLetter;
    end_while;
    P, letter;
  end_proc;

  schensted :=
  proc(w : combinat::words) :
    Type::ListProduct(dom, dom::typeStandardTableau)
  begin
    [ dom::insertWord(dom::fromExpr([]), w),
     dom::insertWord(dom::fromExpr([]),
		     combinat::permutations::inverse(combinat::words::standard(w)))];
  end_proc;
  
  invSchensted :=
  proc(P : dom::typeTableau, Q : dom::typeStandardTableau)
    local n,      // number of remaining cases
	  iLine,   // index of the line containing n
	  letter, // extracted letter
	  res, i;    // result
  begin
    if testargs() and dom::shape(P) <> dom::shape(Q) then
      error("shape mismatch ".expr2text(P)." and ".expr2text(Q));
    end_if;
    P:= dom::expr(P); Q:= dom::expr(Q);
    res := [];
    for n from _plus(nops(i) $ i in Q) downto 1 do
      iLine := 1;
      while contains(Q[iLine], n) = 0 do
	iLine := iLine + 1;
      end_while; // n is at the end of Q[line]
      [P, letter] := [dom::invSchenstedExtractLetter(P, iLine)];
      Q   := subsop(Q, iLine = [op(Q[iLine], 1..nops(Q[iLine])-1)]);
      res := [letter, op(res)];
    end_for;
    res;
  end_proc;
  
  conjugate :=
  proc(T: dom::typeFilling) : dom::typeFilling
     local result, col, row, Row;
  begin
    if nops(T)=0 then
      return([]);
    end_if;
    result := [];
    for col from nops(T[nops(T)]) downto 1 do
      row := nops(T);
      Row := [];
      while row >=1 and col <= nops(T[row]) do
	Row := append(Row, T[row][col]);
	row := row - 1;
      end_while;
      result := append(result, Row);
    end_for;
    result;
  end_proc;

   toPoset :=
   proc(T: dom::typeFilling) : Graph
     local vertex, edges, lin, col;
   begin
     vertex := dom::toWord(T);
     edges  := [ ([T[lin+1][col], T[lin][col]]
		  $ col = 1..nops(T[lin])) $ lin=1..nops(T)-1,
		([T[lin][col],T[lin][col+1]]
		 $ col = 1..nops(T[lin])-1) $ lin=1..nops(T)];
     Graph(vertex, edges, hold(Directed));
   end_proc;

   rowStabilizer := combinat::permutations::fromSetPartition;

   columnStabilizer := combinat::permutations::fromSetPartition @ dom::conjugate;

end_domain:












