/*****************************************************************************
  $Id: trees.mu,v 1.1.1.6 2005/01/26 13:42:24 nthiery Exp $
  File:		lib/COMBINAT/trees.mu
  Authors:	Nicolas M. Thiry <nthiery@users.sourceforge.net>
                Florent Hivert <florent.hivert@univ-mlv.fr>
  Created:	18/09/2002
  License:	LGPL
*****************************************************************************/
/**
    Data structure:

    Empty tree: dom()

    Otherwise: dom(label, first subtree1, ..., last subtree)

    A subtree may be empty. This allow for representing incomplete binary trees.

    Unlabelled trees are represented as labelled trees.


    Examples:

    dom(1):				o

    dom(1, dom(1)):	                o
					|
					o

    dom(1, dom(),dom(1)):	        o
					 \
					  o

    dom(1, dom(1), dom(1)):             o
				       / \
				      o   o

Current listed interface (A * means that the function is documented): 
{one*, zero*, dummyLabel*, type*, testtype
   size*, list*, count*, unrank*, generator*, random*, print*, 
   op*, subsop*, nops*,
   expr*, convert*
   fromShapeAndWord*, revert*, toWord*, toPoset*}
_index and set_index are documented but not in the interface.
inherited methods : new, testtype.
**/

export(output, Alias, asciiArt):

domain combinat::trees
    inherits Dom::BaseDomain;
    category Cat::GradedCombinatorialClass, Cat::TreesClass();
    axiom    Ax::normalRep, Ax::canonicalRep;

    info_str := "Domain 'combinat::trees': utilities for trees";

    interfaceAdd :=
                 {hold(dummyLabel),
		  hold(zero), hold(one),
		  hold(size),
		  hold(nops), hold(op), hold(subsop),
		  hold(_index), hold(set_index),
		  hold(expr), hold(print), hold(convert),
		  hold(revert),
		  hold(shape),
		  hold(toWord),
		  hold(fromShapeAndWord),
		  hold(toPoset)
		 };
	
    typeLabelled   := dom;
    typeUnlabelled := dom;

    zero := new(dom);

    one  := new(dom, dom::dummyLabel);

    dummyLabel := NIL; // See initDomain below before changing!

    //////////////////////////////////////////////////////////////////////////
    // Conversions
    //////////////////////////////////////////////////////////////////////////

    // override the new shortcut from Cat::CombinatorialClass
    new :=
    proc(x)
    begin
	if domtype(x)=DOM_INT then
	    dom::list(x);
	else
	    dom::convert(args());
	    if % = FAIL then error("illegal arguments") else % end_if
	end_if;
    end_proc;

    convert :=
    proc(x)
	local i;
    begin
	case domtype(x)
	of dom do
	    return(x);
	of DOM_LIST do
	    if nops(x)=0 then
		return(dom::zero);
	    else
		return(new(dom, 
			   x[1], 
			   dom::convert(x[i]) $i=2..nops(x)));
	    end_if;
	end_case;
	FAIL;
    end_proc;
    
    expr :=
    proc(t: dom)
    begin
	map([extop(t)], expr);
    end_proc;

    expr2text :=
    proc(t: dom)
    begin
    	"new(".expr2text(dom).
	    if extnops(t)>0 then
		",".expr2text(extop(t));
	    else
		""
	    end_if.
	    ")";
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // Operands access
    //////////////////////////////////////////////////////////////////////////

    nops :=
    proc(t: dom) : Type::NonNegInt
    begin
	if t = dom::zero then
	    0
	else 
	    extnops(t)-1;
	end_if;
    end_proc;
    
    op :=
    proc(t: dom, i)
    begin
	if args(0)=1 then
	    if extnops(t)>=2 then
		return(extop(t,2..extnops(t)));
	    else
		return(null())
	    end_if;
	elif domtype(i)=DOM_INT then
	    return(extop(t,i+1));
	elif type(i)="_range" then
	    return(extop(t,lhs(i)+1..rhs(i)+1));
	end_if;
	extop(t,i);
    end_proc;

    subsop :=
    proc(t: dom)
	local eq;
    begin
	extsubsop(t, op(eq,1)+1=op(eq,2) $ eq in [args(2..args(0))]);
    end_proc;
    
    _index := dom::op;

    set_index :=
    proc(t: dom, i: Type::NonNegInt, value)
    begin
	extsubsop(t, i+1=value);
    end_proc;

    // Maps f on all operands of t, that is all the childs of t
    // FIXME: document and test!
    map :=
    proc(t: dom, f)
    begin
	new(dom, extop(t,1), op(map([extop(t, 2..extnops(t))], f, args(3..args(0)))));
    end_proc;

    size := 
    proc(t: dom) : Type::NonNegInt
	local c;
    begin
	if t = dom::zero then
	    0
	else
	    1 + _plus(dom::size(c) $ c in t);
	end_if;
    end_proc;

    height := 
    proc(t: dom) : Type::NonNegInt
	local c;
    begin
	if t = dom::zero then
	    0
	else
	    1 + max(dom::size(c) $ c in t);
	end_if;
    end_proc;

    isLeaf :=
    proc(t: dom): DOM_BOOL
    begin
	bool(nops(t)=0);
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // Description of the grammar
    //////////////////////////////////////////////////////////////////////////
    
/*    specification := ([Tree  = Alias(Prod(Label, Sequence(Tree)),
				     prog::bless(x -> dom([ op(x,1), op(op(x,2))]),
						 dom)),
		       Label = Atom(dom::dummyLabel)
		      ]);

    decomposableClassOptions := [Boustrophedonic = FALSE];

    fromGrammarOld :=
    proc(tree) : dom
    begin
	dom(eval(subs(tree, [Label=dom::dummyLabel,
			     Leaf=dom::one,
			     Epsilon=null(), LeafNil=dom::zero, 
			     Prod=DOM_LIST@id,
			     Sequence=id
			    ])));
    end_proc;*/

    //////////////////////////////////////////////////////////////////////////
    // shape of a tree: its unlabelled version
    //////////////////////////////////////////////////////////////////////////
  
    shape :=
    proc(t: dom::typeLabelled): dom::typeUnlabelled
	local t1;
    begin
	if t=dom::typeLabelled::zero then
	    dom::typeUnlabelled::zero;
	else
	    new(dom::typeUnlabelled, 
		dom::dummyLabel,
		dom::shape(t1) $ t1 in [op(t)]);
	end_if;
    end_proc;
  
    //////////////////////////////////////////////////////////////////////////
    // Walks through the tree
    //////////////////////////////////////////////////////////////////////////

    readingOrders := table(
			   Prefix       = [Prefix, FALSE],
			   LeftPrefix   = [Prefix, FALSE],
			   RightPrefix  = [Postfix, TRUE],
			   
                           Infix        = [Infix, FALSE],
			   LeftInfix    = [Infix, FALSE],
			   RightInfix   = [Infix, TRUE],
			   
                           Postfix      = [Postfix, FALSE],
			   LeftPostfix  = [Postfix, FALSE],
			   RightPostfix = [Prefix, TRUE]
			  );
    defaultOrder := Prefix;

    toWord := 
    proc(t : dom, order = dom::defaultOrder)
	option escape;
	local rec, DOM, reverse, result;
    begin
	DOM   := dom;
	if testargs() and not(contains(dom::readingOrders, order)) then
	    error("Unknown reading order: ".expr2text(order));
	end_if;
	[order,reverse] := dom::readingOrders[order];
	result:=[];
	rec :=
        proc(t: dom)
	    name combinat::trees::toWordRec;
	    local s, i;
	begin
	    if t<>DOM::zero then
		case order
		  of Prefix do
		      result := append(result, op(t,0));
		      rec(s) $ s in t;
		      return();
		  of Postfix do
		      rec(s) $ s in t;
		      result := append(result, op(t,0));
		      return();
		  of Infix do
		      if nops(t)>=1 then
			  rec(t[1]);
		      end_if;
		      result := append(result, op(t,0));
		      rec(t[i]) $ i=2..nops(t);
		end_case;
	    end_if;
	end_proc;
	rec(t);
	if reverse then
	    revert(result)
	else 
	    result;
	end_if;
    end_proc;

    fromShapeAndWord := 
    proc(t: Type::Union(dom::typeLabelled, dom::typeUnlabelled), 
	 word: combinat::words,
	 order = dom::defaultOrder) : dom::typeLabelled
	option escape;
	local i, rec, DOM, reverse;
    begin
	if domtype(t)<>dom::typeUnlabelled then
	    t := t::dom::shape(t);
	end_if;
	DOM   := dom;
	if testargs() then
	    if not(contains(dom::readingOrders, order)) then
		error("Unknown reading order: ".expr2text(order));
	    end_if;
	    if nops(word) <> t::dom::size(t) then
		error("Length mismatch");
	    end_if;
	end_if;
	[order,reverse] := dom::readingOrders[order];
	if reverse then
	    word:=revert(word)
	end_if;
	
	i:=0;
	rec :=
	proc(t: dom::typeUnlabelled)
	    local tmp, s, j;
	begin
	    if t=t::dom::zero then
		DOM::zero;
	    else
		case order
		  of Prefix do
		      return(new(DOM, word[(i:=i+1)], rec(s) $ s in t));
		  of Postfix do
		      tmp:=rec(s) $ s in t;
		      return(new(DOM, word[(i:=i+1)], tmp));
		  of Infix do
		      if nops(t)>=1 then
			  tmp:=rec(t[1]);
		      else
			  tmp:=null();
		      end_if;
		      return(new(DOM, word[(i:=i+1)], 
				 tmp, (rec(t[j]) $ j=2..nops(t))));
		end_case;
	    end_if;
	end_proc;
	rec(t);
    end_proc;

    // Buggy version: there seems to be a conflict between the local
    // variable i and the dummy variable i in i=2. To be investigated.
    // Try: 
    // t:=BT([1, [], [1, [], []]]);
    // BT::fromShapeAndWordBuggy(t, [1,2],Infix);
    fromShapeAndWordBuggy := 
    proc(t: dom, 
	 word: combinat::words,
	 order = dom::defaultOrder)
	option escape;
	local i, rec, DOM, reverse;
    begin
	DOM   := dom;
	if testargs() then
	    if not(contains(dom::readingOrders, order)) then
		error("Unknown reading order: ".expr2text(order));
	    end_if;
	    if nops(word) <> dom::size(t) then
		error("Length mismatch");
	    end_if;
	end_if;
	[order,reverse] := dom::readingOrders[order];
	if reverse then
	    word:=revert(word)
	end_if;
	
	i:=0;
	rec :=
	proc(t: dom)
	    local tmp, s, i;
	begin
	    if t=DOM::zero then
		t;
	    else
		case order
		  of Prefix do
		      return(new(DOM, word[(i:=i+1)], rec(s) $ s in t));
		  of Postfix do
		      tmp:=rec(s) $ s in t;
		      return(new(DOM, word[(i:=i+1)], tmp));
		  of Infix do
		      if nops(t)>=1 then
			  tmp:=rec(t[1]);
		      else
			  tmp:=null();
		      end_if;
		      return(new(DOM, word[(i:=i+1)], 
				 tmp, (rec(t[i]) $ i=2..nops(t))));
		end_case;
	    end_if;
	end_proc;
	rec(t);
    end_proc;


    //////////////////////////////////////////////////////////////////////////
    // print
    //////////////////////////////////////////////////////////////////////////
    
    printold :=
    proc(t: dom)
	local drawing, x, DOM, f;
    begin
	if t=dom::zero then
	    return(output::asciiArt(["."]));
	end_if;
	DOM := dom;
	drawing := output::asciiArt();
	x := 1;
	f := 
	proc(t: dom, y: Type::PosInt, c: DOM_STRING) : Type::PosInt
	    // Draw the subtree t on the drawing, below y (included) and to
	    // the right of x (included)
	    // Update x
	    // Returns the horizontal position of the root
	    local n, xc, i;
	begin
	    if t=DOM::zero then  // empty tree
		return(x);
	    end_if;
	    n := nops(t);
	    for i from 1 to n div 2 do
		f(t[i], y+1, "/");
	    end_for;
	    if n mod 2 = 1 then
		xc := f(t[i], y+1, "|");
		i := i+1;
	    else
		xc := x;
		x := x+1; 
	    end_if;
	    drawing[xc,y]:=c;
	    for i from i to n  do
		f(t[i], y+1, "\\");
	    end_for;
	    xc;
	end_proc:
	f(t,1,"o");
	drawing::dom::flipy(drawing);
    end_proc;

    slashn := "\n"; // ""or "\n"

    TeX :=
    proc(t: dom)
	local drawing, x, DOM, f, maxx, maxy;
    begin
	if t=dom::zero then
	    return("$\\emptyset$");
	end_if;
	DOM := dom;
	drawing := "";
	x    := 1;
	maxx := dom::size(t);
	maxy := dom::height(t);
	f := 
	proc(t: dom, y) //: Type::NonNegInt) : Type::NonNegInt
	    // Draw the subtree t on the drawing, below y (included) and to
	    // the right of x (included)
	    // Update x
	    // Returns the horizontal position of the root
	    local n, xc, i, posChild;
	begin
	    if t=DOM::zero then  // empty tree
		return(x);
	    end_if;
	    n := nops(t);
	    posChild := [f(t[i], y+1) $ i=1..n div 2];
	    i := n div 2;
	    if n mod 2 = 1 then
		xc := f(t[i], y+1);
		posChild := append(posChild, xc);
		i := i+1;
	    else
		xc := x;
		x := x+1; 
	    end_if;
	    drawing := drawing."\\put(".xc.",".(maxy-y)."){\\circle*{0.7}}".DOM::slashn;
	    posChild := append(posChild, f(t[i], y+1) $ i = i+1..n);
	    for i from 1 to n do
		if t[i] <> DOM::zero then
		    drawing := drawing."\\put(".xc.",".(maxy-y)."){\\Line(".(posChild[i]-xc).",".-1.")}".DOM::slashn;
		end_if;
	    end_for;
	    xc;
	end_proc:
	x := f(t,0);
	drawing := drawing."\\put(".x.",".maxy."){\\circle*{1}}".DOM::slashn;

	"{\\begin{picture}(".(maxx).",".(maxy+1).")".(DOM::slashn).drawing."\\end{picture}}";
    end_proc;

    print :=
    proc(t: dom, printUnlabelledNodes = FALSE) : asciiArt
	local f;
    begin
	f :=
	prog::bless(
	proc(t: dom, link = "": DOM_STRING) : Type::ListProduct(asciiArt, Type::NonNegInt)
	    local n, i, h, result,
		  leftSide, leftPos, rightSide, rightPos, maxRemove, position, drawing, label;
	begin
	    if t = dom::zero then
		if args(0) = 1 then // at the top level
		    return([asciiArt(["."]), 0]);
		else
		    return([asciiArt(), 0]);
		end_if;
	    end_if;
	    label := op(t,0);
	    if label = dom::dummyLabel then
		if args(0) = 1 or // at the top level
		     printUnlabelledNodes then
		    label := hold(o);
		end_if;
	    end_if;

	    n := nops(t);

 	    // Drawing of the left childs of the tree
	    leftSide := asciiArt();
	    leftPos  := 0;
	    // Invariant:
	    // leftPos = 0 iff leftSide is empty iff there is no non trivial left child so far
	    for i from 1 to n div 2 do
		if t[i] = dom::zero then next; end_if;
		[drawing, leftPos] := f(t[i], "/");
		leftSide := asciiArt::concat(leftSide, drawing, Align=Top, Packed);
		leftPos := asciiArt::width(leftSide) - asciiArt::width(drawing) + leftPos;
	    end_for;

	    // Drawing of the right childs of the tree
	    rightSide := asciiArt();
	    rightPos  := 0;
	    for i from n downto n + 1 - (n div 2) do
		if t[i] = dom::zero then next; end_if;
		[drawing, rightPos] := f(t[i], "\\");
		rightSide := asciiArt::concat(drawing, rightSide, Align=Top, Packed);
	    end_for;
	    // Invariant:
	    // rightPos = 0 iff rightSide is empty iff there is no non trivial left child

	    // Take care of the central subtree tree if there is one, and
	    // determin the horizontal position of the root node of this tree
	    if n mod 2 = 1 and t[n mod 2] <> dom::zero then
		// there is a non trivial central child
		[drawing, position] := f(t[i], "|");
		result := asciiArt::concat(leftSide, drawing, Align=Top, Packed);
		position := asciiArt::width(result) - asciiArt::width(drawing) + position;
		result := asciiArt::concat(result, rightSide, Align=Top, Packed);
	    elif leftPos = 0 and rightPos = 0 then
		// there are no non trivial childs
		result := asciiArt();
		position := 1;
	    elif rightPos = 0 then
		// there are no non trivial right childs
		// there are non trivial left childs
		result   := leftSide();
		position := leftPos;
		if not (link = "\\" and label = dom::dummyLabel) then
		    position := position + 1;
		end_if;
	    elif leftPos = 0 then
		// there are no non trivial left childs
		// there are non trivial right childs
		result   := rightSide();
		position := rightPos;
		if not (link = "/" and label = dom::dummyLabel) then
		    if position > 1 then
			position := position - 1;
		    else
			result := asciiArt::concat(asciiArt([" "]), result, Align=Top);
		    end_if;
		end_if;
	    else
		// there are non trivial left and right childs
		// ensures a minimum distance between the roots of left and
		// right childs
		if (link="/" or link="\\") and label = dom::dummyLabel then
		    maxRemove := infinity;
		else
		    maxRemove := asciiArt::width(leftSide) - leftPos + rightPos - 1;
		    if maxRemove = 0 then
			leftSide := asciiArt::concat(leftSide, asciiArt([" "]));
		    else
			maxRemove := maxRemove - 1;
		    end_if;
		end_if;
		result := asciiArt::concat(leftSide, rightSide, Align=Top, Packed,
					   MaxRemove = maxRemove);
		rightPos := asciiArt::width(result) - asciiArt::width(rightSide) + rightPos;
		if rightPos = leftPos + 1 then
		    assert((link="/" or link="\\") and label = dom::dummyLabel);
		    if link="/" then
			position := rightPos;
		    else
			position := leftPos;
		    end_if;
		else
		    position := (leftPos+rightPos) div 2;
		end_if;
	    end_if;
	    assert(position >= 1);
	    // print the label if needed
	    h := asciiArt::height(result)+1;
	    if label <> dom::dummyLabel then
		result[position,h] := expr2text(label);
		h := h + 1;
	    end_if;
	    // print the link if needed (/, |, \)
	    if args(0) >= 2 then
		result[position,h] := link;
	    end_if;
	    [result, position];
	end_proc,
	dom);
	asciiArt::setAlignment(f(t)[1], Top);
    end_proc;

    revert := 
    proc(t: dom) : dom
	// revert recursively the order of the branches of t
    begin
	// could be optimized
	if t=dom::zero then
	    t
	else
	    new(dom, t[0], op(map(revert([op(t)]),revert)));
	end_if;
    end_proc;


    toPoset := 
    proc(T: dom) : Graph
	local toPosetRec, DOM;
    begin
	DOM := dom;
	toPosetRec := 
	proc(t)
	    local ress, edges, vertices;
	begin
	    if t=DOM::zero then
		[[],[]]
	    else
		ress := map([op(t)], toPosetRec); 
		vertices := [t[0], op(map(ress, t -> op(op(t,1))))]; 
		if nops({op(vertices)}) <> nops(vertices) then 
		    error("The labels of the vertices must be distict!")
		end_if; 
		edges := map(ress, t -> op(op(t,2))); 
		[vertices,
		 [op(edges), 
		  op(map([op(t)], 
                   proc(t1) 
		   begin 
		       if t1 <> DOM::zero 
			   then [t[0], t1[0]] 
		       else null() 
		       end_if;
		   end_proc))]];
	    end_if; 
	end_proc;
	Graph(op(toPosetRec(T)), 
		   hold(Directed)); // not yet declared as an option.
    end_proc;

    initDomain :=
    proc()
    begin
	// Force the insertion of NIL as value for dummyLabel
	// (the entry above is actually ignored)
	dom::dummyLabel := NIL;
    end_proc;
end_domain:
