/*****************************************************************************
  $Id: CombinatorialClass.mu,v 1.1.1.5 2005/01/26 13:42:22 nthiery Exp $
  File:		lib/COMBINAT/CombinatorialClass.mu
  Authors:	Nicolas M. Thiry <nthiery@users.sourceforge.net>
                Florent Hivert <florent.hivert@univ-mlv.fr>
  Created:	16/02/2001
  License:	LGPL
*****************************************************************************/

category Cat::CombinatorialClass
    category Cat::BaseCategory;

    info_str := "Category 'Cat::CombinatorialClass': generic utilities for combinatorial classs";

/*****************************************************************************
Many operations, like counting or generating, are common to all
combinatorial classs. This category defines a standardized
interface, and generic tools.

 Standardized entries (not all of them need to be implemented):
// - type				MuPAD type for the class
 - domtype                      MuPAD domain for the class
 - testtype(p)			test if p has the correct type
 - isA(p, size, <constraints>)  test if p if of size and verify te constraints
 - size(s)			size of the objet s
 - count(<size>)		number of objects of size <size>
 - generatingSeries(<ident>)	generating series by size
 - generator(<size>)		generator for all the objects of size <size>
 - list(<size>)			list of all the objects of size <size>
 - first(<size>)		first object of size <size>, or FAIL
 - last(<size>)			last object of size <size>, or FAIL
 - Next(s)			next object of size <size>, or FAIL
 - previous(s)			previous object of size <size>, or FAIL
 - rank(s)			rank of the object s
 - unrank(n, <size>)		n-th object of size <size>, or FAIL
 - random(<size>)		random object of size <size>
 - _less(s1,s2)			comparison of the objects s1 and s2

 Moreover, calling directly combinat::myObjects(<size>) is a
 shortcut for combinat::mymyObjects::list(<size>). See
 combinat::partitions for an exception to this rule.

 Required entries:
 - type, isA
 - generator or list

   Warning: if none of them is provided, the default implementations
   that are provided will call each other, and cause an infinite loop!

 Default implementations are provided for:
 - count
 - list
 - new
 - generator
*****************************************************************************/

    interfaceAdd :=
    { hold(isA),
      hold(count), hold(list), hold(generator)
    };

    //////////////////////////////////////////////////////////////////////////
    // Type checking
    //////////////////////////////////////////////////////////////////////////

    
    /// the domtype of the elements of this domain (for facade domains)
    domtype  :=
    if dom::hasProp(Ax::systemRep) then
    /* Don't force a combinatorial class to define domtype; there are cases
       where this can't be defined
	toBeDefined; */
    /* Don't define domtype as dom; this raises an infinite loop for
       some reason I could not debug (02/09/2003 NT)
    else
        dom; */
    end;

    /// default implementation of testtype for facade domains using dom::isA
    testtype := 
    if dom::hasProp(Ax::systemRep) then
	proc(p,T)
	    option noDebug;
            local res;
	begin
	    if T = dom then
		//temporary for not breaking all the type checks
		//testtype(p, dom::type);
		res := dom::isA(p);
		if res = UNKNOWN then 
		    res := TRUE;
		end_if;
		res;
	    else
		FAIL
	    end_if;
	end_proc:
    end_if;

    /// Test if an object of the class is of the specified "size"
    isA := if dom::hasProp(Ax::systemRep) then
	       () -> UNKNOWN;
	   else
	       proc(obj : Type::AnyType, size)
	       begin
		   if not testtype(obj, dom) then FALSE
		   elif args(0) > 1 then
		       bool(dom::size(obj) = size);
		   else TRUE;
		   end_if;
	       end_proc;
	   end_if;


    /**
     * Syntactic sugar: when there is no conflict with the construction
     * of domain elements (typically for facade domain), calling the
     * domain constructor is a shortcut for list. This way one can write
     * combinat::compositions(4) to get the list of the compositions of 4
     *
     * For non facade domains, this is the same as Cat::BaseCategory::new
     * except that dom::list is called if the conversion returns FAIL.
    **/
    new :=
    if dom::hasProp(Ax::systemRep) then
	dom::list;
    else
	proc()
	begin
	    dom::convert(args());
	    if % = FAIL then
		dom::list(args());
	    else
		%
	    end_if
	end_proc;

    end_if;

    convert :=
    if dom::hasProp(Ax::systemRep) then
	x -> if domtype(x)=dom::domtype then x else FAIL; end_if;
    end_if;

    expr :=
    if dom::hasProp(Ax::systemRep) then
	expr;
    end_if;


    //////////////////////////////////////////////////////////////////////////
    //

    count     := dom::countFromGenerator;
    list      := dom::listFromGenerator;

    generator := (if dom::first <> FAIL and dom::Next <> FAIL then
                      dom::generatorFromNext;
                  elif dom::last <> FAIL and dom::previous <> FAIL then
                      dom::generatorFromPrevious;
                  elif dom::unrank <> FAIL then
                      dom::generatorFromUnrank;
                  else
		      if dom::generatorFromListLoopGard = TRUE then
			  ()->error(expr2text(dom).
				    "::generator called but not implemented");
		      else
			  dom::generatorFromListLoopGard := TRUE;
			  dom::generatorFromList;
		      end_if;
                  end_if);

    //////////////////////////////////////////////////////////////////////////
    // Default implementations
    //////////////////////////////////////////////////////////////////////////
    // With Pref::typeCkeck(Interactive) the use of ... @ ... ensures that
    // the type checking is done 
    // (cf mailing list, Christopher Creutzig, 2002 june 27) 
    ///////
    countFromGenerator := combinat::generators::count  @ dom::generator;
    listFromGenerator  := combinat::generators::toList @ dom::generator;
    
    generatorFromList  := combinat::generators::fromList @ dom::list;
    //////////////////////////////////////////////////////////////////////////
    // One has to use auxiliary function:
    // generatorFromNext := (z->[...]::fromNext(z, dom::Next)) @ dom::first;
    // does not work.
    // (cf mailing list, Christopher Creutzig, 2002 june 27 
    ///////
    generatorFromNextAux  := 
        z -> combinat::generators::fromNext(z, dom::Next); 
    generatorFromPreviousAux := 
        z -> combinat::generators::fromNext(z, dom::previous);    
    generatorFromNext     := dom::generatorFromNextAux @ dom::first;
    generatorFromPrevious := dom::generatorFromPreviousAux @ dom::last;

    generatorFromUnrank   := 
    proc()
	local DOM, r, options;
	option escape;
    begin
	DOM := dom;
	r := 0;
	options := args();
        () -> (r:=r+1; DOM::unrank(r, options));
    end_proc;

    listFromUnrankAndCount :=
    proc()
	local r;
    begin
	[ dom::unrank(r, args()) $ r = 1..dom::count(args()) ]
    end_proc;

    unrankFromGenerator :=
    proc(r: Type::PosInt)
	local g, i, p;
    begin
	g := dom::generator(args(2..args(0)));
	for i from 1 to r do
	    if (p := g()) = FAIL then
		return(FAIL)
	    end_if;
	end_for;
	return(p);
    end_proc;

end_category:

