/* $Id: FiniteGroupInvariantRing.mu,v 1.1.1.3 2004/10/04 17:25:13 nthiery Exp $ */

/*++
Cat::FiniteGroupInvariantRing -- the category of invariant rings of finite
    groups of matrices.

Required entries:
 - group, with entries:
    - group::size()
    - group::allElements()
    - group::charPolynomial()
 - primaryInvariantsDegrees()

Define entries:
 - Poly
 - Hilbert series, ...
 - mu

++*/

category Cat::FiniteGroupInvariantRing(R: DOM_DOMAIN)
    category Cat::Algebra(R);

    info_str := "The category of invariant rings of finite groups of matrices over a field R.";

    interfaceAdd := {
		  hold(variables), hold(Poly),
		  hold(group),
		  hold(dimen),
		  hold(isModular), hold(isCohenMacaulay),
		  hold(primaryInvariantsDegrees),
		  hold(primaryInvariantsRingSeries),
		  hold(HilbertSeries),
		  hold(HilbertSeries_Molien),
		  hold(HilbertSeries_FromSecondary),
		  hold(secondaryInvariantsSeries_FromHilbert),
		  hold(secondaryInvariantsSeries),
		  hold(mu)
		 };

/*--
    Manipulations in the polynomial ring over the free-module.
--*/

    variables;
    group;
    dimen       := nops(dom::variables);

    Poly	:= polylib::Poly([op(dom::variables)], R);


    isModular :=
    if dom::characteristic>0 then
	bool(dom::group::size mod dom::characteristic=0);
    else
	FALSE;
    end_if;
    
    // isCohenMacaulay: TRUE/FALSE/UNKNOWN
    //
    // Whether the invariant ring is Cohen-Macaulay. More precisely,
    // whether the invariant ring is free module over
    // dom::primaryInvariants(). Theoretically this does not depends on
    // the choice of a system of primary invariants, but you may want
    // to force this to UNKNOWN, if you are not sure whether the invariants
    // provided in dom::primaryInvariants() are valid.
    //
    // Default value: TRUE in the non modular case and UNKNOWN else.

    isCohenMacaulay :=
    if not dom::isModular then
	TRUE;
    else
	UNKNOWN;
    end_if;

    // primaryInvariantsDegrees() -> Type::ListOf(Type::PosInt)
    // List of the degrees of the primary invariants

    primaryInvariantsDegrees;

    // Generating series of the primary invariants

    primaryInvariantsSeries :=
    proc(z = hold(z): DOM_IDENT)
	local d;
    begin
	_plus(poly(z^d, [z], Dom::Integer) $
	      d in dom::primaryInvariantsDegrees())
    end_proc;

    // Generating series of the ring generated by the primary invariants

    primaryInvariantsRingSeries :=
    proc(z=hold(z): DOM_IDENT)
	local degrees;
    begin
	if args(0)>=2 then
	    degrees:=args(2);
	else
	    degrees:=dom::primaryInvariantsDegrees();
	end_if;
	combinat::integerVectorsWeighted::generatingSeries(degrees, z)
    end_proc;
    
/*--
  HilbertSeries_Molien(z:DOM_IDENT) -> rational function in z: Compute
  the Hilbert series of the invariant ring under the action of the
  group, using Molien formula.

  Ref: Sturmfels "Algorithms in Invariant Theory, p 29 / p 72".

  Does it mean anything in non-zero characteristic, for non
  permutation groups ?
--*/

    HilbertSeries_Molien :=
    proc(z = hold(z) : DOM_IDENT)
	option remember;
	local m;
    begin
        if dom::characteristic>0 then
            userinfo(1, "Warning: Molien formula may be non-sensical in non-zero characteristic");
        end_if;
	1/dom::group::size()
	* _plus(1/expr(dom::group::charPolynomial(m, z))
                 $ m in dom::group::allElements())
    end_proc;

    HilbertSeries_FromSecondary :=
    proc(z=hold(z):DOM_IDENT)
    begin
	expr(dom::secondaryInvariantsSeries(args())) *
	dom::primaryInvariantsRingSeries(args());
    end_proc;

    HilbertSeries := dom::HilbertSeries_Molien;

    secondaryInvariantsSeries_FromHilbert :=
    proc(z=hold(z):DOM_IDENT)
    begin
	poly(normal(dom::HilbertSeries(z)
		    / dom::primaryInvariantsRingSeries(args())),
	     [z], Dom::Integer);
    end_proc;

    secondaryInvariantsSeries := dom::secondaryInvariantsSeries_FromHilbert;
	
/*--
    mu()
    
    Returns the lowest degree of an invariant relative to the linear
    caracter det^{-1}. This is also the difference between the highest
    degree of a secondary and the bound given by the degrees of the
    primary invariants.
    
    Ref: Stanley "Invariants of Finite Groups and Their Applications to
    Combinatorics" (1979), p. 485.
--*/

    mu :=
    proc()
    begin
	_plus(op(dom::primaryInvariantsDegrees()))-dom::dimen
	-degree(dom::secondaryInvariantsSeries(hold(z)));
    end_proc;
  
begin
end_category:
