/*****************************************************************************
  $Id $
  File:		lib/DOMAINS/CATEGORY/ModuleWithSeveralBases.mu
  Authors:	Florent Hivert <florent.hivert@univ-mlv.fr>
  Purpose:	Source code for the category Cat::ModuleWithSeveralBases
  Created:	Fri Jan 30 18:03:00 2004
  License:	LGPL
  Project:	MuPAD-Combinat, http://mupad-combinat.sf.net
*****************************************************************************/

alias(declareConversion = operators::overloaded::declareConversion):
alias(declareSignature  = operators::overloaded::declareSignature):

category Cat::ModuleWithSeveralBases(CoeffRing : DOM_DOMAIN) // only for type checking.
/////////////////////////////////////////////////////////////////////////
//  This category ought to be the standard placeholder for the interface
//  domain of a module with several bases. The goal is to provide an
//  automatic declaration of the changes of basis, and default
//  implementations. It provide also some tools for scalar product and
//  dual basis. 
//  The different bases are supposed of course to share the same 
//  coefficients ring, but not necessarily the same combinatorial class 
//  of indices. 
//  Moreover, new bases and basis change can be added at run time. 
/////////////////////////////////////////////////////////////////////////

    category Cat::UseOverloading, Cat::Module(CoeffRing);

    info_str := "Cat::ModuleWithSeveralBases -- The category of free modules with several bases";

    interfaceAdd :=
       {
	hold(print), hold(expr), // should be in Cat::BaseCategory?

	hold(coeffRing), // should be in Cat::Module
	hold(zero), hold(iszero),
	hold(_plus), hold(_negate),
	hold(_mult), hold(mult2), hold(_invert),

	hold(bases),
	// hold(declareBasis),
	// hold(declareBasisChange),
	// hold(basisChanges), // or basesChanges?
	// hold(basisChangesBasis),
	hold(autoDefineBasisChanges)
	// hold(dualBasesPairs), // or basesDualPairs?
	// hold(scalar),
	// hold(declareDualBases),
       };

/////////////////////////////////////////////////////////////////////////
//  Basic information on the module
//  These are used for type checking and are automatically set during the 
//  initialisation of the module or the adjonction of a new basis. 
    coeffRing      := CoeffRing;  
    bases          := {}; // The set of the domains of the differents bases
/////////////////////////////////////////////////////////////////////////


/////////////////////////////////////////////////////////////////////////
// Basis change control
//////
    basisChangesBasis := table();
//  basisChangesBasis is an hash table of associations of the type
//      (basSource, basTarget) = basSourceTobasTargetChangeBasis
//  where basSourceTobasTargetChangeBasis is of the type
//      proc(x : basSource::basisIndices) : basTarget
//  and returns the image of the element basSource(x) expressed in basTarget.
//  * WARNING: to quickly take care of linearity, x is an *index* whereas
//    basSourceTobasTargetChange(x) is and *element* of the domain basTarget. 
    basisChanges := table();
//  basisChanges is an hash table of associations of the type
//      (basSource, basTarget) = basSourceTobasTargetChange
//  where basSourceTobasTargetChange is of the type
//      proc(x : basSource) : basTarget
//  * Only the explicit change of basis must be declared here, 
//    those who are computed by transitivity are being take care of 
//    by the overloading mechanism.
//  * These functions can be built using the following default implementation
//    dom::invertBasisChange      // invert a change of basis
//    dom::transposeBasisChange   // transpose a change of basis
//////
    autoDefineBasisChanges := FALSE;
//  If autoDefineBasisChanges is TRUE, the system define automatically the
//  inverse basis changes of each given basis changes and the transpose of
//  all basis changes for dual bases. 
/////////////////////////////////////////////////////////////////////////



/////////////////////////////////////////////////////////////////////////
//  dom is facade domain. Define standard procedure according to.
/////////////////////////////////////////////////////////////////////////

    standardBasis := dom::bases[1];
    zero  := (dom::standardBasis)::zero;

    _plus :=
    proc()
        // try to make dom::zero the zero of all the bases
        local ar;
    begin
        ar := select([args()], not _equal, dom::zero);
        if nops(ar) = 0 then dom::zero
        else operators::_plus(op(ar));
        end_if;
    end_proc;

    _invert := _invert;
    _negate := _negate;
    expr    := expr;
    iszero  := iszero;
    print   := print;
    _mult   := operators::_mult;
    mult2   := operators::mult2;

    testtype :=
    proc(e, T)
    begin
        _lazy_and(T = dom,
		  contains(dom::bases, domtype(e)));
    end_proc;
    
    convert := () -> 
	error("Cannot create an element of ".expr2text(dom).". You must specify the basis.");

/*  ToDo : convert automatically an expression by substitution. 
    convert :=
    proc(el)
        local res;
    begin
        res := eval(subs(el,
			 map(dom::bases, z-> z::domainWrapper
                        [hold(m)  = dom::m::domainWrapper,
                         hold(F)  = dom::F::domainWrapper]));
    end;
*/
/////////////////////////////////////////////////////////////////////////
//  Changes of basis implementations
/////////////////////////////////////////////////////////////////////////

    basisChangeMatrix := 
    // Compute the matrix of the change of bases basSource => basTarget. 
    // lBas : list of the basisIndices in the order which the matrix is computed.
    //        If not given basSource::
    // rank : rank of the element of the target. If not given, this is the same 
    //        as the rank in the list lBas.
    proc(basSource     : DOM_DOMAIN,
	 basTarget     : DOM_DOMAIN,
	 lBas = null() : Type::Union(DOM_NULL, DOM_LIST),
	 rank = null() : Type::Union(DOM_NULL, DOM_LIST, combinat::rankers::typeRanker))
	local p;
    begin
	if args(0) = 2 then
	    lBas := basSource::basisIndices::list();
	end_if;
	if args(0) <= 3 then 
	    rank := combinat::rankers::rankFromList(lBas);
        elif domtype(rank) = DOM_LIST then
            rank := combinat::rankers::rankFromList(rank);
        end_if;
	_concat(basTarget::sparseVector(basTarget(basSource(p)), rank) $ p in lBas);
    end_proc;


    basisChangeByMatrixOperation :=
    // Compute the change of basis basSourceResult => basTargetResult 
    // by doing the operation matrixOperation on the matrix of the 
    // change of basis basSourceKnown => basTargetKnown.
    // If basSource::basisIndices::size
    proc(basSource       : DOM_DOMAIN,
	 basTarget       : DOM_DOMAIN,
	 basSourceKnown  : DOM_DOMAIN,
	 basTargetKnown  : DOM_DOMAIN,
	 matrixOperation : Type::Function,
	 sizeProc=null() : Type::Union(DOM_NULL, Type::Function))
	option escape;
        local res, indicesDoms, xDom;
    begin
	// If the basisIndices of the different domain are the same 
	// (which is almost always the case), we avoid to recompute the lists.
	indicesDoms := map({args(1..4)}, xDom -> xDom::basisIndices);
	if sizeProc = null() then
	    sizeProc := basSource::basisIndices::size;
	end_if;
	res :=
	prog::bless(
	proc(el) // : basSource::basisIndices)
	         // put back the type check when the hack below is solved. 
	    local size, lists, rank, i, lSource, lTarget, mat;
	begin
	    if sizeProc <> FAIL then
		size  := sizeProc(el);
		lists := table(xDom = xDom::list(size) $ xDom in indicesDoms);
		userinfo(Text, 1, "Computing the basis change ", basSource, "==", basTarget,
			 " in size ", size);
	    else 
		lists := table(xDom = xDom::list()     $ xDom in indicesDoms);
		userinfo(Text, 1, "Computing the basis change ", basSource, "==>", basTarget);
	    end_if;
	    rank := combinat::rankers::rankFromList(lists[basTargetKnown::basisIndices]);
	    mat := dom::basisChangeMatrix(basSourceKnown, basTargetKnown, 
					  lists[basSourceKnown::basisIndices], rank);
	    mat  := matrixOperation(mat);
	    // We now fill the remember table of the proc.
	    lSource := lists[basSource::basisIndices];
	    lTarget := lists[basTarget::basisIndices];
	    for i from 1 to nops(lSource) do
		res(lSource[i]) := basTarget::fromSparseVector(linalg::col(mat, i), lTarget);
	    end_for;
	    res(el);
	end_proc, dom);
	// At some point res is copied so that it does not rememer its result.
	// This is *not* done by the automatic prog::bless of BasisChange.
	// I encapsulate it into yet an other closure...
	// TODO: can this hack be avoided ? 
	proc(el : basSource::basisIndices)
	begin
	    res(el);
	end;
    end_proc;

    invertBasisChange :=
    // Compute the change of basis basSource => basTarget
    // by inverting the matrix of the reverse change. 
    proc(basTarget : DOM_DOMAIN, 
	 basSource : DOM_DOMAIN)
    begin
	dom::basisChangeByMatrixOperation(basSource, basTarget, 
					  basTarget, basSource, _invert);
    end_proc;

    transposeBasisChange :=
    // Compute the change of basis basSource => basTarget
    // by transposing the matrix of the dual change. 
    proc(basSourceDual : DOM_DOMAIN,
	 basTargetDual : DOM_DOMAIN,
	 basSource : DOM_DOMAIN,
	 basTarget : DOM_DOMAIN)
    begin
	dom::basisChangeByMatrixOperation(basSource,     basTarget, 
					  basSourceDual, basTargetDual, linalg::transpose);
    end_proc;

/////////////////////////////////////////////////////////////////////////
//  Initialisations and declarations
/////////////////////////////////////////////////////////////////////////
	
    declareBasis := 
    // Declare a basis for this module. 
    // It does not harm to redeclare a basis
    proc(basisDom : DOM_DOMAIN)
    begin
	if CoeffRing <> basisDom::coeffRing then
	    error("Canot declare basis with different coefficients rings :".
		  expr2text(CoeffRing <> basisDom::coeffRing));
	end_if;
	dom::bases := dom::bases union {basisDom};
    end;	

    basisChangeIsKnown := 
    proc(basSource   : DOM_DOMAIN,
	 basTarget   : DOM_DOMAIN)
    begin
	contains(dom::basisChanges, (basSource, basTarget));
    end_proc;

    declareBasisChangeBasis := 
    // Declare a basis change for this module. Is is possible to update a 
    // previously declared change for example with a faster implementation. 
    proc(basSource   : DOM_DOMAIN,
	 basTarget   : DOM_DOMAIN,
	 basisChange : Type::Function)
    begin
	userinfo(Text, 1, "Declaring basis change on basis: ", basSource, " to ", basTarget);
	// Ensure that the bases are known. 
	dom::declareBasis(basSource);
	dom::declareBasis(basTarget);
	// If basisChange is not assigned to a particular basis, assign it to dom. 
	if extop(basisChange, 13) = NIL then 
	    userinfo(Text, 3, "blessing basisChangeBasis: ", basSource, " to ", basTarget);
	    basisChange := prog::bless(basisChange, dom);
	end_if;
	dom::basisChangesBasis[basSource, basTarget] := basisChange;	    
	if not contains(dom::basisChanges, (basSource, basTarget)) then
	    dom::declareBasisChange(basSource, basTarget,
				    operators::makeLinear(basisChange, 
							  Source  = basSource, 
							  ImageSet= basTarget)
				   );
	end_if;
	if dom::autoDefineBasisChanges and
	   not dom::basisChangeIsKnown(basTarget, basSource) then
	    dom::declareBasisChangeBasis(basTarget, basSource, 
				dom::invertBasisChange(basSource, basTarget));
	end_if;
	null();
    end;

    declareBasisChange := 
    // Declare a basis change for this module. Is is possible to update a 
    // previously declared change for example with a faster implementation. 
    proc(basSource   : DOM_DOMAIN,
	 basTarget   : DOM_DOMAIN,
	 basisChange : Type::Function)
	option escape; 
    begin
	userinfo(Text, 1, "Declaring basis change: ", basSource, " to ", basTarget);
	// Ensure that the bases are known. 
	dom::declareBasis(basSource);
	dom::declareBasis(basTarget);
	// If basisChange is not assigned to a particular basis, assign it to dom. 
	if extop(basisChange, 13) = NIL then 
	    userinfo(Text, 3, "blessing basisChange: ", basSource, " to ", basTarget);
	    basisChange := prog::bless(basisChange, dom);
	end_if;
	dom::basisChanges[basSource, basTarget] := basisChange;	    
	userinfo(Text, 1, "Declaring conversion: ", basSource, " to ", basTarget);
	declareConversion(basSource, basTarget, basisChange);
	if not contains(dom::basisChangesBasis, (basSource, basTarget)) then
	    dom::declareBasisChangeBasis(basSource, basTarget,
					 proc(b : basSource::basisIndices)
					 begin
					     basisChange(basSource::term(b));
					 end_proc);
	end_if;
	null();
    end;
	
    initDomainAdd := dom::initDomainModuleWithSeveralBases;
    initDomainModuleWithSeveralBases :=
    proc()
	local chngt, basSource, basTarget, basisChange, pairs;
    begin
	for chngt in dom::basisChangesBasis do 
	    // (basSource, basTarget) = basSourceTobasTargetChange;
	    basSource   := op(chngt, [1, 1]);
	    basTarget   := op(chngt, [1, 2]);
	    basisChange := op(chngt, 2);

	    dom::declareBasisChangeBasis(basSource, basTarget, basisChange);
	end_for;
	for pairs in dom::dualBasesPairs do
	    dom::declareDualBases(op(pairs));
	end_for;
	assert(map({op(dom::basisChanges)}, DOM_LIST@op ,1) = 
	       map({op(dom::basisChangesBasis)}, DOM_LIST@op ,1));
    end_proc;
/////////////////////////////////////////////////////////////////////////
//  Dual bases and scalar product
//  Note : Should this be elsewhere, so that the dual bases can be in 
//         another different module ? Example : (QSym, NCSF)
/////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////
//  The dual module
//  This is used to transpose automatically the change of bases. 
//    dual := ???;
    dualBasesPairs := {};
    scalar := operators::scalar;

    declareDualBases := 	
    proc(basDom : DOM_DOMAIN, basDual : DOM_DOMAIN) 
	local pair;
    begin
	// Ensure that the bases are known. 
	if dom::dual = FAIL then
	    error("You must set ".expr2text(hold(slot)(dom, "dual"))." before declaring dual bases.");
	end_if;
	userinfo(Text, 1, "declaring pairing: ", basDom, " with ", basDual);
	dom::declareBasis(basDom);
	dom::dualBasesPairs := {op(dom::dualBasesPairs)} union {[basDom, basDual]};
        // set the bases information for duality.
        basDom::dual   := basDual;
        basDom::scalar := operators::scalar;
        basDom::dualityScalarProduct := 
            DomainConstructor::getEntryFromCategory(dom, Cat::ModuleWithBasis(CoeffRing), 
						    "dualityScalarProduct");
        // declare the scalar product to the overload system. 
	declareSignature
        (operators::scalar, [basDom, basDual], basDom::dualityScalarProduct);
        
	if dom::autoDefineBasisChanges then 
	    for pair in [op(dom::dual::basisChangesBasis)] do
		pair := [op(pair,1)];
		if pair[2] = basDom::dual         and
		   (pair[1])::dual in dom::bases  and
		   not contains(dom::basisChangesBasis, (basDom, (pair[1])::dual))
		    then
		    dom::declareBasisChangeBasis(basDom, (pair[1])::dual,
						 dom::transposeBasisChange(op(pair), 
									   basDom, (pair[1])::dual));		    
		end_if;
	    end_for;
	end_if;
	
        // Do the same for the dual basis, if not already done.
	if basDual::dual <> basDom then
	    dom::dual::declareDualBases(basDual, basDom);
	end_if;
    end_proc;

end_category:
