/* $Id: PermutationGroupInvariantRing.mu,v 1.1.1.3 2004/10/04 17:25:13 nthiery Exp $ */

/*++
Cat::PermutationGroupInvariantRing(S) -- the category of invariant
    rings of permutation groups over S.

Such modules arises when a group acts (by permutation) on a finite set
X, and one considers the induced representation on the functions from
X to another set S. Otherwise said, consider the free module over S of
basis X. m will be the size of X.

This is a specialisation of Cat::FiniteGroupInvariantRing. A lot of
things can be computed much more efficiently than for general groups.
(Hilbert Series, ...)


Definitions / Notations:

 - $m$: size of X
 - $S_m$: group of all the permutations of X

 - Cycle Type:

    Given a permutation $\sigma$ of $X$, we define its cycle type to
    be $(a_i)_{i=1\dots m}$, where $a_i$ is the number of cycles of length
    i in $\sigma$.

    Sometimes instead, its easier to use the dual notion by
    considering the partition of $m$ $(l_i)_{i=1\dots k}$ where the
    $l_i$ are the length of the cycles ordered decreasingly.

    Two permutations are conjugate if they have the same cycle type.
    So, the conjugacy classes of Sn are indexed by the partitions p of
    n. This will be used for example in GrapheModule to speed up
    computations of summations over the group by working by conjugacy
    class, instead of permutation by permutation. (ex: for S_10 there
    are 42 conjugacy class instead of 10! permutations !).

References:
 - A.~Kerber, "Algebraic Combinatoric Via Finite Group Action"
 - B.~Sturmfels, "Algorithms in Invariant Theory"

Methods:
++*/

category Cat::PermutationGroupInvariantRing(R: DOM_DOMAIN)
    category Cat::FiniteGroupInvariantRing(R),
	Cat::GradedAlgebraWithBasis(R);

    interfaceAdd := {
		  hold(elementary),
		  hold(elementaries),
		  hold(powersum),
		  hold(powersums),
		  hold(symmetricMonomial),
		  hold(primaryInvariants),
		  hold(primaryInvariantsDegrees),
		  hold(primaryInvariantsNames),
		  hold(setPrimaryInvariants),
		  hold(setPrimaryInvariantsDegrees),
		  hold(HilbertSeries_Polya),
		  hold(secondaryInvariantsSeries_FromCycleTypes),
		  hold(numberOfSecondaryInvariants),
		  hold(secondaryInvariantsMaxDegree),
		  hold(fineChainSeries),
		  hold(chainSeries),
		  hold(fineSecondaryInvariantsTable),
		  hold(fineSecondaryInvariantsSeries),
		  hold(Invar), hold(Invar2), hold(Singular), hold(magma),
		  hold(magmaMinimalGeneratingSet),
		  hold(magmaMinimalGeneratingSet_HomogeneousModuleTestBasis),
		  hold(magmaMinimalGeneratingSet_FundamentalInvariants)
		 };


    info_str := "The category of invariant rings of permutation groups over a field R.";

    options := table();

/*--
    Some classical symmetric and thus invariant polynoms.

    powersum: (k:non negative integer) -> dom::Poly
    Returns the k th symetric powersum: $\sum x_i^k$

    elementary: (k:non negative integer) -> dom::Poly
    Returns the k th symetric elementary polynom.
--*/

    powersum;

    powersums := 
    proc()
        option remember;
	local i;
    begin
        [dom::powersum(i) $ i=1..dom::dimen];
    end_proc;

    elementary;

    elementaries := 
    proc()
        option remember;
	local i;
    begin
        [dom::elementary(i) $ i=1..dom::dimen];
    end_proc;

    symmetricMonomial;

/*--
    primaryInvariants:
    
    List containing the primary invariants.


    In the default implementation, the primary invariants are the
    dimen first symetric powersum, which is often suboptimal.
    This can be changed by the user with setPrimaryInvariants.
    If a sub domain wants overrides this definition, it should also
    override primaryInvariantsDegrees.


    primaryInvariantsDegrees()->Type::ListOf(Type::PosInt):

    List of the degrees of the primary invariants.


    setPrimaryInvariants(l):

	Sets the primary invariants to l.

    setPrimaryInvariantsDegrees(l):

	Sets the primary invariants degrees to l.

    A lot of information depends on the primary invariants, or at
    least from their degrees. Those two methods try to keep the
    integrity of the data structure by updating, or clearing from
    remember tables, such informations. However, it's safer to only
    call setPrimaryInvariants right after the creation of the domain.
    
--*/

    primaryInvariantsDegrees :=
    proc()
        option remember;
    begin
        [$ 1..dom::dimen];
    end_proc;

    primaryInvariantsNames := 
    proc()
        option remember;
	local i;
    begin
        [ hold(s).i $ i=1..dom::dimen ];
    end_proc;

    primaryInvariants := 
    proc()
        option remember;
    begin
        dom::powersums();
    end_proc;

    setPrimaryInvariants :=
    proc(l: Type::ListOf(dom))
	local degs;
    begin
	dom::primaryInvariants():=l;
	degs:=map(l,degree);
	if {op(degs)}<>{op(dom::primaryInvariantsDegrees())} then
	    dom::setPrimaryInvariantsDegrees(degs);
	end_if;
    end_proc;

    setPrimaryInvariantsDegrees :=
    proc(l: Type::ListOf(Dom::Integer))
    begin
	dom::primaryInvariantsDegrees():=l;
    end_proc;

/*--
    HilbertSeries(z:DOM_IDENT) -> polylib::Poly([z]):

    Overload the default method for computing the Hilbert Series of
    the module. For a permutation group, its faster to compute it from
    the cycle indicator polynomial, in particular when the computation
    of the cycle types (dom::group::cycleTypes()) is easy.
--*/

    HilbertSeries := dom::HilbertSeries_Polya;

    HilbertSeries_Polya :=
    dom::basisIndices::generatingSeries_Polya;

    secondaryInvariantsSeries_FromCycleTypes :=
    proc(z=hold(z): DOM_IDENT)
        option remember;
    local p, p1, primaryP, PolyI, PolyR, equa, ct, d, i;
    begin
	// Do the intermediate polynomial have integers or rational coeffs ?
	PolyR:=polylib::Poly([z], Dom::Rational):
	PolyI:=polylib::Poly([z], Dom::Integer):
	userinfo(3, "secondaryInvariantsSeries: Computation from cycle types");
	primaryP:=PolyR(_mult((1-z^d) $ d in dom::primaryInvariantsDegrees())):
	p:=PolyR(0);
	for equa in dom::group::cycleTypes() do
	    userinfo(5, "secondaryInvariantsSeries: working on cycletype",
		     op(equa,1));
	    ct:=op(equa,1); /* cycle type */
	    /* op(equa,2): #elements of the group with this cycle type */
	    p1:= primaryP /
		 PolyR(_mult((1-z^i)^ct[i] $ i=1..dom::dimen));
	    if (p1=FAIL) then
		error("Oops, the algorithm for secondaryInvariantsSeries_FromCycleTypes is incorrect.\
The characteristic polynomial of some element of the group does not divide the
primary Invariants Series.\
Please send a description of your group to the author <nthiery@users.sourceforge.net>.");
	    end_if;
	    p:=p+multcoeffs(p1, op(equa,2));
	end_for;
	PolyI(multcoeffs(p, 1/dom::group::size));
    end_proc;

    secondaryInvariantsSeries := dom::secondaryInvariantsSeries_FromCycleTypes;

    numberOfSecondaryInvariants :=
    proc()
	local b, d, MoebiusInv;
    begin
	if args(0)=0 then 
	    return(_mult(op(dom::primaryInvariantsDegrees()))/dom::group::size);
	end_if;
	d:=args(1);
	if not testtype(d, Type::Union(Type::NonNegInt, dom::TypeShapeSet))
	   then
	    error("Type mismatch for first argument");
	end_if;
	if domtype(d)=DOM_INT then
	    coeff(dom::secondaryInvariantsSeries(hold(z)),d);
	elif dom::chainProduct then
	    MoebiusInv:=0;
	    for b in combinat::subsets::list(d) do
		MoebiusInv :=
		   MoebiusInv
		  +dom::basisIndices::count(dom::shapeSet2List(b))
		  *(-1)^(nops(d)-nops(b));
	    end_for;
	end_if
    end_proc;

    secondaryInvariantsMaxDegree :=
    proc()
        option remember;
	local d;
    begin
        if dom::isGorenstein then
            _plus(d-1 $ d in dom::primaryInvariantsDegrees());
        else
            degree(dom::secondaryInvariantsSeries(hold(z)));
        end_if;
    end_proc;

/*--
    Generating series for the Stanley-Reisner Ring.

    Mathematical background:

    One can consider a monomial as a chain of subsets of {1..Dimen}
    ordered by inclusion. This is done by spliting the monomial 
    in horizontal layers. Eg:
    
		x1^2 x2^3 x3  <-> {1,2,3} > {1,2} > {2}
    
    In fact its usualy a multichain since you may have the same layer
    several times

		x1^3 x2^4 x3  <-> {1,2,3} > {1,2} > {1,2} > {2},

    one can give a weight to such monomials by counting the layers by
    size. The previous monomial will have weight:

		t[1] t[2]^2 t[3],

    since there are 1 layer of size 1, 2 layers of size 2 and 1 of
    size 3.

    Moreover the empty layer and the complete layers are not counted.
    That is they have both weight 1. This weight is preserved by the
    action of the group, so that this gives a fine graduation of the
    invariant ring. The usual product don't preserve it. However its
    possible to define another product that will preserve it. The ring
    of invariants with this product is called the Stanley-Reisner
    Ring. The elementary symmetric polynomial are a system of
    parameters for this ring. Its possible to compute the generating
    function the chains and the secondaries from the cycle indicator
    polynomial.

    Ref: Garsia/Stanton: Group Actions on Stanley-Reisner Rings and
    Invariants of Permutation Groups, 1994.

    fineChainSeries(t)

    Returns a generating function for the chains (not the
    multichains), counted by size of the layers. The coeff of
    t[1].t[4].t[5] is the number of chains with layers of sizes 1, 4, 5.

    This is computed from the cycle indicator polynomial. Don't ask me
    exactly how. I have forgoten, but it works and thats all I want :-)

--*/

    fineChainSeries :=
    proc(t:DOM_IDENT)
	option remember;
	local _list2Part, list2Part, i;
    begin
    
	/*
	    list2part(m)

	    Let p be a set partition of {1..Dimen}. Let m[i] count the
	    number of parts of size i of p. The length of m is Dimen.
  
	    list2part returns the generating polynomial for all chains
	    obtained by adding successively parts of p. Multiple parts
	    can be added at the same time.
	    
	    Eg: Let p be {{1,2},{3},{4}} you have among others the
	    following chains:

		{1,2} < {1,2,3}	  < {1,2,3,4}
		{4}   < {1,2,3,4}
		{1,2}
		{3}   < {1,2,3}	  < {1,2,3,4}
	*/

	_list2Part :=
	proc(m: DOM_LIST, Pol, k:DOM_INT)
	    local P, k2, b, choices;
	begin
	    P:=Pol(1);
	    choices:=combinat::subcompositions::list(m);
	    for b in choices do
		if b<>m and b<>[0$nops(m)] then
		    k2:=k+_plus(i*b[i] $ i=1..nops(b));
		    P:= P + Pol(_mult( binomial(m[i], b[i]) $ i=1..nops(m)) * t[k2] )
		       * _list2Part(zip(m, b, _subtract), Pol, k2);
		end_if;
	    end_for;
	    P;
	end_proc;

	list2Part :=
	proc(m: DOM_LIST)
	local Pol;
	begin
	    userinfo(5, "Counting parts of", m); /* C'est l que ca rame !!! */
	    Pol:=polylib::Poly([t[i] $ i=1..nops(m)-1], Dom::Rational);
	    _list2Part(m, Pol, 0);
	end_proc;

	_plus(op(map(poly2list(dom::group::cycleIndicatorPoly(hold(z))),
		     x->multcoeffs(list2Part(op(x,2)),
				   op(x,1)))));
    end_proc;

    fineChainTable :=
    proc()
	option remember;
	local list2Set, i, x;
    begin
	// should go in combinat ?
	list2Set :=
	proc(l: DOM_LIST)
	begin
	    { (if l[i]=1 then i else null() end_if) $ i=1..nops(l) }
	end_proc:
	table(list2Set(x[2])=x[1]
	      $ x in poly2list(dom::fineChainSeries(hold(t))));
    end_proc;

/*--
    chainSeries
--*/

    chainSeries :=
    proc(z= hold(z): DOM_IDENT)
	local i;
    begin
	poly(subs(expr(dom::fineChainSeries(hold(t))),
		  (hold(t)[i]=z^i $ i=1..dom::dimen-1)),
	     [z], Dom::Rational);
    end_proc;

/*--
    fineSecondaryInvariantsSeries()

    Returns a generating function for the secondary invariants
    counted by the size of the layers. The primaries are here
    the elementary symmetric polynomials.
    
    Let S = {s1..sk} be a subset of {1..n}. Let a[S] be the number of
    chains whose layers have size s1..sk. In the same way, let b[S] be
    the number of secondary Invariants whose layers are s1..sk.
    
    You have the formula:
    
	a_[T] = sum_{S\subset T} b[S]
    
    Thus one can compute the b[S] by Mbius inversion of the a_[T]:
    
	b[S] = sum_{T\supset S} (-1)^(|T|-|S|) a[S]
    
--*/	

    fineSecondaryInvariantsTable :=
    proc()
	local T, invT, inv, a, s, b;
	option remember;
    begin
	T:=dom::fineChainTable();
	invT:=table();
	for a in T do
	    a:=op(a,1);
	    s:=nops(a);
	    inv:=0;
	    for b in combinat::subsets::list(a) do
		if contains(T,b) then
		    inv:=inv+T[b]*(-1)^(s-nops(b));
		end_if;
	    end_for;
	    if inv<>0 then
		invT[a]:=inv;
	    end_if;
	end_for;
	invT;
    end_proc;

    fineSecondaryInvariantsSeries :=
    proc(t:DOM_IDENT)
	option remember;
	local Pol, set2List, i;
    begin
	set2List :=
    proc(S: DOM_SET)
	begin
	    [ (if contains(S,i) then 1 else 0; end_if)
	      $ i=1..dom::dimen ];
	end_proc:

	Pol:=polylib::Poly([t[i] $ i=1..dom::dimen-1], Dom::Integer);
	Pol(map([op(dom::fineSecondaryInvariantsTable())],
		x->[set2List(op(x,2)), op(x,1)]));
    end_proc;


/*--
    Invar(s: DOM_STRING)
    
    Write a file named 's' to be read by maple. It contains everything
    that is needed to compute the invariant ring using the Invar
    package from G. Kemper.
--*/

    Invar :=
    proc(s)
    begin
	fprint(Unquoted, Text, s,
"kernelopts(printbytes=false);
with(share):
readshare(invar,algebra):
with(invar):
".s.":=".dom::group::Invar().";
invring(".s.");
");
    end_proc;

/*--
    Invar2(s: DOM_STRING)
    
    Write a file named 's' to be read by maple. It contains everything
    that is needed to compute the invariant ring using the second version
    of the Invar package from G. Kemper.
--*/
    
    Invar2 :=
    proc(s:DOM_STRING)
    begin
	fprint(Unquoted, Text, s,
"kernelopts(printbytes=false):

read(invar):
with(invar):
".s.":=".dom::group::Invar2().":
".s."[vars]:=".expr2text(dom::variables).":
".s."[prims]:=".expr2text(map(dom::primary, expr)).":
".s."[primdegs]:=".expr2text(dom::primaryInvariantsDegrees()).":
".s."[primnames]:=".expr2text(dom::primaryInvariantsNames()).":

protocol_level:=10:

#no_GB_in_secondaries:=true:
#no_derive_in_secondaries:=true:
#GB_in_derive:=true:
no_transfer:=false:	# Use Reynolds operator to generate invariants

invring(".s."):
");
    end_proc;

/*--
    Singular(s: DOM_STRING)
    
    Write a file named 's' to be read by Singular. It contains
    everything that is needed to compute the invariant ring using the
    finvar package.
--*/

    Singular :=
    proc(file: DOM_STRING)
	local fd, i, j, p, group, s;
    begin
	fd:=fopen(Text,file,Write);
	fprint(Unquoted, fd, "LIB \"finvar.lib\";");
	fprint(Unquoted, fd, "ring R=0,(".
	       dom::variables[2].
	       (",".dom::variables[i] $ i=2..dom::dimen)."),dp;");
	for i from 1 to nops(dom::group::generators) do
	    fprint(Unquoted, fd,
		   "matrix A".i."[".dom::dimen."][".dom::dimen."]=");
	    p:=dom::group::generators[i];
	    s:="";
	    for j from 1 to dom::dimen do
		if j>1 then s:=s.", "; end_if;
		s:=_concat(s, ("0," $ p(j)-1), "1", (",0" $ dom::dimen-p(j)));
	    end_for;
	    fprint(Unquoted, fd, s, ";");
	end_for;
	group:=_concat("A".i.", " $ i=1..nops(dom::group::generators));
	fprint(Unquoted, fd, "
intvec flags = 0,0,1;
matrix REY,M = reynolds_molien(".group." flags);
matrix P     = primary_char0_random(REY,M,1, flags[3]);
secondary_char0(P, REY, M, flags[3]);
");
    end_proc;

/*--
    magma()
    
    Returns a string containing Magma code to define the invariant ring
--*/

    magma :=
    proc()
    begin
        _concat(
                "R:=InvariantRing(", dom::group::magma(), ", ",
                (if dom::characteristic=0 then
                     "Rationals()"
                 else
                     "FiniteField(".dom::characteristic.")"
                 end_if),
                ");\n"
               );
    end_proc;

/*--
    magmaMinimalGeneratingSet()
    
    Returns a string containing Magma code to compute a minimal
    generating set.

    Several implementations:
     - FundamentalInvariants
     - HomogeneousModuleTestBasis
--*/

    magmaMinimalGeneratingSet :=
	magmaMinimalGeneratingSet_FundamentalInvariants;

    magmaMinimalGeneratingSet_FundamentalInvariants :=
    proc()
    begin
        _concat(
                dom::magma(),
                "SetVerbose(\"Invariants\", 4);\n",
                "FundamentalInvariants(R);"
               );
    end_proc;

    magmaMinimalGeneratingSet_HomogeneousModuleTestBasis :=
    proc()
    begin
        _concat(
                dom::magma(),"
SetVerbose(\"Invariants\", 4);
Gen:=InvariantsOfDegree(R,1);
for d in { 1 .. ", dom::degreeBound(), "} do
    Basis:=InvariantsOfDegree(R,d);
    Gen:=Gen cat [Basis[i] : i in HomogeneousModuleTestBasis(Gen, [R!1], Basis)];
end for;
Gen;
"
               );
    end_proc;

begin
end_category:
