/*****************************************************************************
  $Id: BooleanSemiRing.mu,v 1.1.1.3 2004/10/16 15:15:03 nthiery Exp $
  File:		lib/DOMAINS/DOMAIN/BooleanSemiRing.mu
  Authors:	Eric Laugerotte <Eric.Laugerotte@univ-rouen.fr>
  Created:	2002/10/24
  License:	LGPL
*****************************************************************************/

/*++

Dom::BooleanSemiRing --  the domain of booleans 

Domains: 
Dom::BaseDomain

Categories:
Cat::UseOverloading
Cat::SemiRing

Methods:
zero         	    - returns the neutral element for +
one	    	    - returns the neutral element for *
convert_to(x,T)	    - converts x to an element of type T
_plus(x1,x2,...)    - returns the sum of x1+x2+...
_mult(x1,x2,...)    - returns the product of x1*x2*...
_power(x,i)	    - returns x*x*...*x (i times) for a nonzero integer i
star(x)		    - returns the star of x

++*/

unalias(SRB):
alias(SRB = Dom::Matrix(Dom::BooleanSemiRing)):

domain Dom::BooleanSemiRing
    local SRB;
    inherits Dom::BaseDomain;
    category Cat::UseOverloading,Cat::SemiRing;
    axiom Ax::canonicalRep,Ax::normalRep;

    interfaceAdd := { hold(star) };

    /////////////////////////////////////////////////////////////////////////
    // Constructors and constants
    /////////////////////////////////////////////////////////////////////////

    initDomain := 
    proc()
  	option escape;
	local DOM;
    begin
	DOM := dom;
	operators::overloaded::declareSignature(operators::plus2,
	    [dom,Dom::WeightedAutomaton(dom)],
	        (x,aut) -> 
		    (Dom::WeightedAutomaton(dom))::plus(
                        (Dom::WeightedAutomaton(dom))::new(1,[],SRB(1,1,[x]),
			    table(),SRB(1,1,[dom::one])),aut));
	operators::overloaded::declareSignature(operators::plus2,
	    [dom,dom],
		(x1,x2) -> DOM::plus(x1,x2));
	operators::overloaded::declareSignature(operators::mult2,
            [dom,Dom::WeightedAutomaton(dom)],
		proc(x,aut)
		    local a,d,i,j; 
		begin
		  a:=aut;
		  d:=(Dom::WeightedAutomaton(DOM))::dimension(a);
		  i:=(Dom::WeightedAutomaton(DOM))::initial(a);
		  for j from 1 to d do
		      a:=(Dom::WeightedAutomaton(DOM))::
					subsInitialWeight(a,j,x*i[1,j])
		  end_for;
		  return(a)
		end_proc);
	operators::overloaded::declareSignature(operators::mult2,
	    [dom,dom],
		(x1,x2) -> DOM::mult(x1,x2));
    end_proc;

    new :=
    proc(n : Type::Predicate("Semiring",
         n->_lazy_or(testtype(extop(n,1),stdlib::Infinity),
		     testtype(n,Type::Union(DOM_INT,DOM_RAT,DOM_FLOAT,
					    DOM_COMPLEX)), 
		     testtype(n,stdlib::Infinity),
		     n::dom::hasProp(Cat::SemiRing))))
    begin
      if testtype(extop(n,1),stdlib::Infinity) then
	    if n=n::dom::zero then
	    	return(new(dom,0))
	    else
	    	return(new(dom,1))
	    end_if
      elif testtype(n,Type::Union(DOM_INT,DOM_RAT,DOM_FLOAT,DOM_COMPLEX)) then
	    if extop(n,1)=0 then
	    	return(new(dom,0))
	    else
	    	return(new(dom,1))
	    end_if
      elif n=n::dom::zero then
	    return(new(dom,0))
	else
	    return(new(dom,1))
	end_if
    end_proc;

    convert :=
    proc(n : Type::Predicate("Semiring",
         n->_lazy_or(testtype(extop(n,1),stdlib::Infinity),
	             testtype(n,Type::Union(DOM_INT,DOM_RAT,DOM_FLOAT,
					    DOM_COMPLEX)),
	             testtype(n,stdlib::Infinity),
	             n::dom::hasProp(Cat::SemiRing))))
   begin
      if testtype(extop(n,1),stdlib::Infinity) then
	    if n=n::dom::zero then
	    	return(new(dom,0))
	    else
	    	return(new(dom,1))
	    end_if
      elif testtype(n,Type::Union(DOM_INT,DOM_RAT,DOM_FLOAT,DOM_COMPLEX)) then
	    if extop(n,1)=0 then
	    	return(new(dom,0))
	    else
	    	return(new(dom,1))
	    end_if
      elif n=n::dom::zero then
	    return(new(dom,0))
      else
	    return(new(dom,1))
      end_if
    end_proc;

    convert_to :=
    proc(x : dom,
	 T : DOM_DOMAIN)
    begin
	if contains({DOM_INT,DOM_RAT,DOM_FLOAT,DOM_COMPLEX},T) then
	    if x=dom::zero then
		return(0)
	    else
		return(1)
	    end_if
	elif T::hasProp(Cat::SemiRing) then
	    if x=dom::zero then
	   	return(T::zero)
	    elif x=dom::one then
	     	return(T::one)
	    end_if
	else
	    return(FAIL)
	end_if	
    end_proc;

    print :=
    proc(x : dom)
    begin
	return(extop(x,1))
    end_proc;

    expr := 
    proc(x : dom)
    begin
	return(extop(x,1))
    end_proc;

    zero := dom::new(0);
       
    one := dom::new(1);

    /////////////////////////////////////////////////////////////////////////
    // Arithmetic operations
    /////////////////////////////////////////////////////////////////////////

    _plus := operators::_plus;

    plus :=
    proc()
    begin
	if contains([args()],dom::one)=0 then 
	    return(dom::zero)
	end_if; 
	return(dom::one)
    end_proc;

    _mult := operators::_mult;

    mult := 
    proc()
    begin
	if args(0)=0 then
	    return(dom::zero)
	end_if;
	if contains([args()],dom::zero)=0 then 
	    return(dom::one)
	end_if; 
	return(dom::zero)
    end_proc;	

    _invert := 
    proc(x : dom)
    begin
    	if x=dom::one then
	    return(dom::one)
	end_if;
	return(FAIL)
    end_proc;

    _power :=
    proc(x : dom, i : Dom::Integer)
    begin
	if i=0 then
	    return(dom::one)
	else
	    return(x)
	end_if
    end_proc;

    star := 
    proc(x : dom)
    begin
	return(dom::one)
    end_proc;

begin
end_domain:


