/*****************************************************************************
  $Id: FreeModuleList.mu,v 1.1.1.3 2004/10/04 17:25:12 nthiery Exp $
  File:		lib/DOMAINS/DOMAIN/FreeModuleList.mu
  Authors:	Stefan Wehmeier, Uni Paderborn
	        Werner M. Seiler, Uni Karlsruhe/Mannheim
		(wms@ira.uka.de, seiler@euler.math.uni-mannheim.de)
		Nicolas M. Thiry <nthiery@users.sourceforge.net>
  Created:	30/06/2003 (taken from Dom::FreeModule of MuPAD 3.0.0)
  License:	LGPL
*****************************************************************************/
/* 
Changes:
 - removed characteristic
 - removed all conversions stuff
 - removed all algebra stuff (one, ...)
 - generate1 -> monomial
 - lcoeff(dom::zero) returns FAIL
 - lterm(dom::zero) returns dom::zero
 - nthterm(dom::zero, 1) returns FAIL instead of raising an error
 - coeff does not do coercion on the support
 - does not work with a finite list anymore
*/


/*++

Old documentation:

Dom::FreeModule(R,S)  -- free module over the ring R with (finite) support
        in the combinatorial class S

An element is represented as a domain element with operands:

0 - domain Dom::FreeModule(R,S)
1 - a list [[r1,s1], ...,[rn,sn]] of lists of one element of R and one
    element of S, with s1 > s2 > ... > sn

Elements are generated exclusively with the two methods generate and generate1.
The first one expects as argument a list as described above; generate1 takes as
arguments one element of R and one of S. Domains based on Dom::FreeModule which
want to extend the representation by further operands (computable from the
list) can thus simply overwrite these two methods and then still inherit all
other methods implemented here.

The provided methods are essentially those of a Cat::FiniteMonoidRing. Some
methods like _mult, _divide, convert, convert_to or expr are implemented under
different names, so that the implementation can be more easily inherited by
other domains build on Dom::FreeModule. For example, scalar multiplication is
called smult and _mult is identical to smult. If now an algebra is constructed
upon Dom::FreeModule, it will have to redefine _mult. But it can still make use
of the here implemented scalar multiplication by calling smult.

If S has a method degree or degreevec, then they are lifted to Dom::FreeModule.
If there exists methods has and subs in S (or if S is a list), then a subs
method is implemented for Dom::FreeModule. S::has must return TRUE whenever
S::subs can perform a substitution. The provided implementation of subs
supports all calls of the standard MuPAD function subs. However, the option
Unsimplified has no effect and only substitutions in/for elements of S are
possible.


Authors: Stefan Wehmeier, Uni Paderborn
         Werner M. Seiler, Uni Karlsruhe/Mannheim
         (wms@ira.uka.de, seiler@euler.math.uni-mannheim.de)
Created: June 30th, 1997
Last update: 	Tue Feb 15 14:04:09 2000
++*/
// 13.10.2000


alias(REP(x)=extop(x,1),
      TYPE_ERR=error("wrong type of argument"),
      INT_ERR=error("invalid index")):

domain Dom::FreeModuleList(R,S)

    local Rzero:R, Riszero:DOM_PROC, Rconvto:DOM_PROC, Rone:R, Risone:DOM_PROC,
	  Rcoerce:DOM_PROC, Rexpr:DOM_PROC, Rplus:DOM_PROC, Rmult:DOM_PROC, 
	  Rnegate:DOM_PROC, Rdivide:DOM_PROC, Rlength:DOM_PROC,
	  Sone, Sisone, Sexpr, Sconvto, Sdegree, Sdegreevec, Smult, Slength,
	  maxr:DOM_INT, finite:DOM_BOOL;
    inherits Dom::BaseDomain;
    category Cat::ModuleWithBasis(R);


    //////////////////////////////////////////////////////////////////////////
    // Conversions and overloading
    //////////////////////////////////////////////////////////////////////////

    _plus := operators::_plus;
    _mult := operators::_mult;

    /////////////////////////////////////////////////////////////////////////
    // Basic information on the free module
    /////////////////////////////////////////////////////////////////////////

    basisIndices       := S;

    /* =============== */
    /* local functions */
    /* =============== */

    Srandom :=
        if finite then
	    proc()
	    local r:DOM_PROC;
	    begin
	        r := random(1..nops(S));
	        S[r()];
	    end_proc;
	else 
	    proc():S
	    begin S::random() end_proc;
	end_if;


    Scoerce :=
        if finite then
	    proc(x)
	    begin if contains(S,x)=0 then FAIL else x end_if end_proc;
	else 
	    proc()
	    begin S::coerce(args()) end_proc; 
	end_if;


    Shas :=
        if finite then _equal 
	else 
	    proc():DOM_BOOL
	    begin S::has(args()) end_proc;
	end_if;


    Ssubs :=
        if finite then
            proc(x,y)
            begin if x=op(y,1) then op(y,2) else x end_if end_proc;
        else 
	    proc():S
	    begin S::subs(args()) end_proc;
	end_if;


    greater2 :=
        if finite then
	    proc(x:DOM_LIST,y:DOM_LIST):DOM_BOOL
	    begin bool(contains(S,x[2])<contains(S,y[2])) end_proc;
	else
	    proc(x:DOM_LIST,y:DOM_LIST):DOM_BOOL
	    begin bool(S::_less(y[2],x[2])) end_proc;
	end_if;


    // removes multiple occurrence of generators
    //    x - list of lists of the form [r,e]
    normalizeRep :=
        proc(x:DOM_LIST):DOM_LIST
        local i:DOM_INT;
        begin
            if x=[] then return([]) end_if;
            x := sort(x,dom::greater2);
            i := 1;
            while i<nops(x) do
	        if x[i][2]=x[i+1][2] then
		    x[i][1] := Rplus(x[i][1],x[i+1][1]);
		    delete x[i+1];
	        else i:=i+1 end_if;
            end_while;
	    // ZERO CHECK
            select(x, ((l) -> not(Riszero(l[1]))));
        end_proc;


    coerceSummand :=
        proc(l:DOM_LIST):DOM_LIST
        begin [Rcoerce(l[1]),dom::Scoerce(l[2])] end_proc;


    /* ========================= */
    /* representation, in/output */
    /* ========================= */

    generate :=
        proc(x:DOM_LIST):dom
        begin new(dom,x) end_proc;


    monomial :=
    proc(x: R, y: dom::basisIndices) : dom
    begin
	new(dom,[[x,y]])
    end_proc;

    // For backward compatibility
    generate1 := dom::monomial;

    /* ====================== */
    /* "polynomial" functions */
    /* ====================== */

    toList :=
    proc(x: dom) : Type::ListOf(Type::ListProduct(dom::coeffRing, dom::basisIndices))
    begin
	REP(x);
    end_proc;

    support :=
    proc(x: dom): Type::ListOf(dom::basisIndices)
    begin
	map(REP(x),op,2)
    end_proc;

    terms :=
    proc(x:dom): Type::ListOf(dom)
    begin
	map(REP(x), dom::term@op,2) 
    end_proc;

    coeff :=
    proc(x:dom, s) : Type::Union(dom::coeffRing, Type::SequenceOf(dom::coeffRing))
	local pos:DOM_INT;
    begin
	if args(0)=1 then op(map(REP(x),op,1))
	else
	    if testargs() then
		if not testtype(s, dom::basisIndices) then TYPE_ERR end_if;
	    end_if;
	    pos := contains(dom::support(x),s);
	    if pos = 0 then Rzero else op(op(REP(x),pos),1) end_if;
	end_if;
    end_proc;

    coeffs :=
    proc(x:dom) : Type::ListOf(dom::coeffRing)
    begin
	map(REP(x),op,1);
    end_proc;

    lsupport :=
        proc(x:dom) : Type::Union(dom::basisIndices, DOM_FAIL)
        begin
            if dom::iszero(x) then FAIL else op(REP(x),[1,2]) end_if;
        end_proc;

    lterm :=
        proc(x:dom) : dom
        begin
            if dom::iszero(x) then dom::zero else dom::term(op(REP(x),[1,2])) end_if;
        end_proc;

    lcoeff :=
        proc(x:dom): dom::coeffRing
        begin
            if dom::iszero(x) then Rzero else op(REP(x),[1,1]) end_if;
        end_proc;

    lmonomial :=
        proc(x:dom) : dom
        begin
            if dom::iszero(x) then dom::zero
            else dom::generate1(op(REP(x)[1])) end_if;
        end_proc;


    tcoeff :=
        proc(x:dom):R
        local n:DOM_INT;
        begin
            n := dom::nterms(x);
            if n = 0 then Rzero else op(REP(x),[n,1]) end_if;
        end_proc;


    tterm :=
        proc(x:dom)
        local n:DOM_INT;
        begin
            n := dom::nterms(x);
	    if n = 0 then dom::zero else dom::term(op(REP(x),[n,2])) end_if;
	end_proc;


    tmonomial :=
        proc(x:dom):dom
        local n:DOM_INT;
        begin
            n := dom::nterms(x);
            if n = 0 then x else dom::generate1(op(REP(x)[n])) end_if;
        end_proc;


    nthmonomial :=
        proc(x:dom, i:Type::PosInt) : Type::Union(dom, DOM_FAIL)
        begin
	    x := REP(x);
            if i > nops(x) then
		FAIL;
	    else
		dom::monomial(op(x[i]));
	    end_if;
        end_proc;

    nthsupport :=
        proc(x:dom, i: Type::PosInt) : Type::Union(dom::basisIndices, DOM_FAIL)
        begin
	    x := REP(x);
            if i > nops(x) then
		FAIL;
	    else
		x[i][2];
	    end_if;
        end_proc;

    nthterm :=
        proc(x:dom, i: Type::PosInt) : Type::Union(dom, DOM_FAIL)
        begin
	    x := REP(x);
            if i > nops(x) then
		FAIL;
	    else
		dom::term(x[i][2]);
	    end_if;
        end_proc;


    nthcoeff :=
        proc(x:dom,i:DOM_INT) : Type::Union(dom::coeffRing, DOM_FAIL)
        begin
	    x := REP(x);
            if i > nops(x) then
		FAIL;
	    else
		x[i][1];
	    end_if;
        end_proc;

    reductum :=
        proc(x:dom) : dom
	local l:DOM_LIST;
	begin
	    case dom::nterms(x)
	    of 0 do error("not defined for zero");
	    of 1 do return(dom::zero);
	    otherwise
		l := REP(x);
		delete l[1];
		return(dom::generate(l));
	    end_case;
	end_proc;

    nterms :=
        proc(x:dom) : Type::NonNegInt
        begin nops(REP(x)) end_proc;


    /* =========== */
    /* arithmetics */
    /* =========== */

    zero := dom::generate([]);

    iszero :=
        proc(x:dom):DOM_BOOL
        begin bool(REP(x)=[]) end_proc;

    plus2 :=
        proc(l1: dom, l2: dom) : dom
	local i:DOM_INT, tmp:R, res:DOM_LIST;
	begin
	    l1 := REP(l1);
	    l2 := REP(l2);
            res := [ 0 $ nops(l1)+nops(l2) ];
	    i := 0;
            while (l1<>[]) and (l2<>[]) do
		if dom::greater2(l1[1],l2[1]) then
		    i := i+1;
                    res[i] := l1[1];
                    delete l1[1];
                elif l1[1][2]=l2[1][2] then
                    tmp := Rplus(l1[1][1],l2[1][1]);
		    // ZERO CHECK
                    if not Riszero(tmp) then
			i := i+1;
                        res[i] := [tmp,l1[1][2]];
                    end_if;
                    delete l1[1];
                    delete l2[1];
                else
		    i := i+1;
                    res[i] := l2[1];
                    delete l2[1];
                end_if;
            end_while;
	    dom::generate(
		if   (l1<>[]) then [op(res,1..i)] . l1
		elif (l2<>[]) then [op(res,1..i)] . l2
		else               [op(res,1..i)] 
		end_if);
	end_proc;

    // Added _subtract and _negate to work with the double inheritance
    // problem with Cat:AbelianGroup see mailing list (23 Feb 2004)
    _subtract := proc(x,y) begin dom::_plus(x, _negate(y)) end_proc;
			       
    _negate := // much better that de default version of Cat::ModuleWithBasis. 
    proc(x: dom)
    begin
	dom::mapcoeffs(x, Rnegate);
    end_proc;
    
    
    smult := dom::multcoeffsLeft; // for backward compatibility

    sdivide :=
        if Rdivide<>FAIL then
	    proc(x,lambda)
	    begin
		x := dom::convert(x);
		lambda := Rcoerce(lambda);
                if testargs() then
                    if x=FAIL then TYPE_ERR end_if;
                    if lambda=FAIL then TYPE_ERR end_if;
                end_if;
	        //if dom::iszero(x) or Risone(lambda) then return(x) end_if;
 	        dom::mapcoeffs(x,Rdivide,lambda);
	    end_proc;
        else NIL end_if;


    _divide := dom::sdivide;


    _invert := proc() begin FAIL end_proc;
    
    
    /* ============= */
    /* miscellaneous */
    /* ============= */

    mapcoeffs :=
        proc(x:dom,f):dom
        begin
            if dom::iszero(x) then return(dom::zero) end_if;
	    dom::generate(
			  // ZERO CHECK
		select(map(REP(x),
			   ((t,f) -> subsop(t,1=f(t[1],args(3..args(0))))),
			   args(2..args(0))),
		       ((t) -> not(Riszero(t[1])))));
	end_proc;


    // monomialSubs(m,s,fm) performs the substitution s=fm in a
    // monomial m=[r,s] where r is an element of R and s is an element
    // of S and fm of the free module; returns a sequence of tuples [ri,si]
    monomialSubs :=
        proc(m:DOM_LIST,s,fm:dom)
        begin
            if m[2]=s then
                op(map(REP(fm), 
		       ((l,m) -> subsop(l,1=Rmult(m[1],l[1])), m)))
            else m end_if;
        end_proc;


    // allows substitutions only for elements of S
    subs :=
        proc(x):dom
        local i:DOM_INT, eqs:DOM_LIST, eq, res:DOM_LIST, DOM:DOM_DOMAIN;
        begin
            eqs := [];
            for i from 2 to args(0) do
                case type(args(i))
                of "_equal" do
                    eqs := append(eqs,args(i));
                    break;
                of DOM_LIST do
                    eqs := eqs . args(i);
                    break;
                of DOM_SET do
                of DOM_TABLE do
                    eqs := eqs . [op(args(i))];
                    break;
                of DOM_IDENT do
                    if args(i)=hold(Unsimplified) then break end_if;
                otherwise TYPE_ERR;
                end_case;
            end_for;

	    DOM := dom;
	    x := dom::coerce(x);
            if testargs() then
                if x=FAIL then TYPE_ERR end_if;
                for i from 1 to nops(eqs) do
                    if type(eqs[i])<>"_equal" then TYPE_ERR end_if;
                    eqs[i] := subsop(eqs[i],1=DOM::Scoerce(op(eqs[i],1)),
                                            2=DOM::coerce(op(eqs[i],2)));
                    if contains([op(eqs[i])],FAIL)>0 then TYPE_ERR end_if;
                end_for;
            end_if;

            res := REP(x);
            for eq in eqs do
                res := map(res,DOM::monomialSubs,op(eq,1),op(eq,2));
            end_for;
            dom::generate(dom::normalizeRep(res));
        end_proc;
	
	
    length :=
	proc(x:dom):DOM_INT
	local l:DOM_LIST;
	begin
	    l := REP(x);
	    if l=[] then return(0) end_if;
	    1+_plus(op(map(l,(ll->Rlength(ll[1])+Slength(ll[2])+1))));
	end_proc;


    random :=
        if R::random<>FAIL and dom::Srandom<>FAIL then
	    proc(N=maxr:DOM_INT):dom
	    local r:DOM_PROC, n:DOM_INT, i:DOM_INT;
	    begin
	        r := random(1..N);
	        n := r();
	        dom::generate(
	             dom::normalizeRep([[R::random(),
					 dom::Srandom()] $ i=1..n]));
	    end_proc; 
	else NIL end_if;


    generatorSet := dom::basisIndices; // for backward compatibility

begin
    finite := FALSE;
    Rzero := R::zero;
    Riszero := R::iszero;
    Rplus   := R::_plus;
    Rdivide := R::_divide;
    Rnegate := R::_negate;
    maxr := 10;        // one should check for a finite set!
   // TODO
/*   if args(0)<>2 then error("wrong number of arguments") end_if;
     if (domtype(R)=DOM_DOMAIN) and (R::hasProp<>FAIL) and
	R::hasProp(Cat::Rng) then
     	 Riszero := R::iszero;
	 Rone := R::one;
	 Risone := R::isone;
    	 Rcoerce := R::coerce;
	 Rconvto := R::convert_to;
     	 Rexpr := R::expr;
	 Rplus := R::_plus;
	 Rmult := R::_mult;
	 Rnegate := R::_negate;
	 Rlength := R::length;
	 if Rlength=FAIL then Rlength := length end_if;
     else error("first argument must be Cat::Rng") end_if;
     if domtype(S)=DOM_LIST then
         finite := TRUE;
	 maxr := nops(S);
	 Sexpr := id;
	 Sconvto := ((s,T) -> s);
	 Sone := FAIL;
	 Sisone := FAIL;
         Sdegree := ((s) -> (1));
         Sdegreevec := FAIL;
	 Smult := FAIL;
	 Slength := length;
     elif domtype(S)=DOM_DOMAIN and (S::hasProp<>FAIL) and
	  S::hasProp(Cat::OrderedSet) then
         finite := FALSE;
	 maxr := 10;        // one should check for a finite set!
     	 Sexpr := S::expr;
	 Sconvto := S::convert_to;
     	 Sone := (if S::hasProp(Cat::Monoid) then S::one
	          elif S::hasProp(Cat::AbelianMonoid) then S::zero
	          else S::one;           // will probably be FAIL !
	          end_if);
     	 Sisone := (if S::hasProp(Cat::Monoid) then S::isone
	           elif S::hasProp(Cat::AbelianMonoid) then S::iszero
	           else S::isone;        // will probably be FAIL !
	           end_if);
	 Smult := (if S::hasProp(Cat::Monoid) then S::_mult
		   else FAIL end_if);
         Sdegree := S::degree;
         Sdegreevec := S::degreevec;
	 Slength := S::length;
	 if Slength=FAIL then Slength := length end_if;
     else error("second argument must be list or Cat::OrderedSet") end_if;
*/
end_domain:


unalias(REP,TYPE_ERR,INT_ERR):
