/*****************************************************************************
  $Id: FreeModuleTable.mu,v 1.1.1.3 2005/01/26 13:42:22 nthiery Exp $
  File:		lib/DOMAINS/DOMAIN/FreeModuleTable.mu
  Authors:	Nicolas M. Thiry <nthiery@users.sourceforge.net>
  Created:	16/02/2001
  License:	LGPL
*****************************************************************************/

/**
    Dom::FreeModuleTable(R,S) -- R-Free module over the basis S

    R: ring
    S: set, type, ...

    For now, R needs to be commutative (still right?)

    This is an alternative implementation of Dom::FreeModule using tables.
**/

unalias(REP,TYPE_ERR):
alias(REP(x)=extop(x,1),
      TYPE_ERR=error("wrong type of argument")
      ):

domain Dom::FreeModuleTable
    (R: DOM_DOMAIN,
     BasisIndices: DOM_DOMAIN
    )
    local Options;
    inherits Dom::BaseDomain;
    category Cat::ModuleWithBasis(R);

    info_str := "Dom::FreeModuleTable: domain for free modules, implemented with tables";

    interfaceAdd :=
        {
         hold(generate)
         };

    //////////////////////////////////////////////////////////////////////////
    // Conversions and overloading
    //////////////////////////////////////////////////////////////////////////

    _plus := operators::_plus;
    _mult := operators::_mult;

    /////////////////////////////////////////////////////////////////////////
    // Basic information on the free module
    /////////////////////////////////////////////////////////////////////////

    basisIndices       := BasisIndices;

    /////////////////////////////////////////////////////////////////////////
    // Constructors and constants
    /////////////////////////////////////////////////////////////////////////

    // Unused ... For compatibility with standard mupad domains ?????
    generate := 
    proc(x: Type::TableOfIndex(dom::basisIndices)) : dom
    begin 
	new(dom, x); 
    end_proc;

    monomial :=
    proc(c: R, support: dom::basisIndices) : dom
    begin
	new(dom,table(support=c))
    end_proc;

    term :=
    proc(support: dom::basisIndices) : dom
    begin
	new(dom,table(support=R::one))
    end_proc;

    zero := dom::generate(table());

    iszero :=
    proc(x: dom) : DOM_BOOL
    begin
	bool(nops(REP(x))=0);
    end_proc;

    /////////////////////////////////////////////////////////////////////////
    // Access to terms, leading terms, monomials, coefficients, ...
    /////////////////////////////////////////////////////////////////////////

    nterms :=
    proc(x: dom)
    begin
	nops(REP(x));
    end_proc;

    support :=
    proc(x: dom)
	local eq;
    begin
	[ op(eq,1) $ eq in REP(x) ];
    end_proc;

    toList :=
    proc(x: dom)
	local eq;
    begin
	[[op(eq,2), op(eq,1)] $ eq in REP(x)];
    end_proc;

    coeff :=
    proc(x: dom, s)
	local t, eq;
    begin
	t:=REP(x);
	if args(0)=2 then
	    if contains(t,s) then t[s];
	    else R::zero;
	    end_if;
	else
	    op(eq,2) $ eq in t;
	end_if;
    end_proc;

    /////////////////////////////////////////////////////////////////////////
    // Low level manipulations
    /////////////////////////////////////////////////////////////////////////

    addToRep :=
    if Options[ZeroCheck] then
    proc(t: DOM_TABLE, s: dom::basisIndices, r: R)
    begin
	if contains(t, s) then
	    // ZERO CHECK
	    if iszero((t[s]:=t[s]+r)) then
		delete(t[s]);
	    end_if;
	else
	    // ZERO CHECK
	    if not iszero(r) then
		t[s]:=r;
	    end_if;
	end_if;
	t;
    end_proc;
    else
    proc(t: DOM_TABLE, s: dom::basisIndices, r: R)
    begin
	if contains(t, s) then
	    t[s]:=t[s]+r;
	else
	    t[s]:=r;
	end_if;
	t;
    end_proc;
    end_if;

/*--
    mapcoeffs ( x, f, [args] )

    map the function f on all the coefficients of x. f must return a
    valid coefficient! The only check is for removing monomials with
    zero coefficient. As with the usual map function, [args] are
    passed as further arguments for f.
--*/

    mapcoeffs :=
    if Options[ZeroCheck] then
    proc(x: dom, f: Type::AnyType)
	local t;
    begin
	t:=REP(x);
	t:=select(map(t,f, args(3..args(0))), not iszero@op, 2); // ZERO CHECK
	new(dom,t);
    end_proc;
    else
    proc(x: dom, f: Type::AnyType)
    begin
	new(dom, map(REP(x),f, args(3..args(0))));
    end_proc;
    end_if;

/*--
    multcoeffs (x, r)

    Multiply the coefficients of the terms of x by r.
--*/

    multcoeffs :=
    if R::hasProp(Ax::noZeroDivisors) and Options[ZeroCheck] then
	proc(x: dom, r: R)
	begin
	    // ZERO CHECK
	    if iszero(r) then
		dom::zero
	    else
		new(dom, map(REP(x), R::_mult, r));
	    end_if;
	end_proc;
    end_if;

    multcoeffsLeft :=
    if R::hasProp(Ax::noZeroDivisors) and Options[ZeroCheck] then
	proc(c: R, x: dom)
	begin
	    if iszero(c) then
	        // ZERO CHECK
		dom::zero
	    else
		new(dom, map(REP(x), c1 -> c*c1));
	    end_if;
	end_proc;
    end_if;

/*--
    mapsupport ( x, f, [args] )

    Map the function f on all the terms of x. The function f must
    always return a valid term! no check is done. As with the usual
    map function, [args] are passed as further arguments for f.

    TODO: if testargs() is TRUE, do the testing!
--*/

    mapsupport :=
    proc(x: dom, f)
	local m;
    begin
	new(dom,table(f(op(m,1),args(3..args(0))) = op(m,2) $ m in REP(x)));
    end_proc;

    /////////////////////////////////////////////////////////////////////////
    // Arithmetic operations
    /////////////////////////////////////////////////////////////////////////

    plus2 :=
    proc(x: dom, y: dom)
	local s;
    begin
	x := REP(x);
	y := REP(y);
	for s in y do
	    x := dom::addToRep(x,op(s,1),op(s,2));
	end_for;
	new(dom, x);
    end_proc;

    // Added _subtract and _negate to work with the double inheritance
    // problem with Cat:AbelianGroup see mailing list (23 Feb 2004)
    _subtract := proc(x,y) begin dom::_plus(x, _negate(y)) end_proc;
			       
    _negate := // much better that de default version of Cat::ModuleWithBasis. 
    proc(x: dom)
    begin
	dom::mapcoeffs(x, R::_negate);
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // Specialized tools to build operators
    // Can probably be thrown away after some benchmarking
    //////////////////////////////////////////////////////////////////////////

    makeLinearFunction :=
    proc(f)
    /**
        f is a function from dom::basisIndices to some module over
	dom::coeffRing, where the multiplication by scalars can
	be done via multcoeffs
	
	Returns the linear extension of f to dom;

	Extra arguments of the constructed operator are passed down to f.
    **/
	option escape;
    begin
        proc(x: dom)
	    local s;
	begin
	    _plus(multcoeffs(f(op(s,1), args(2..args(0))),
			     op(s,2))
		  $ s in REP(x));
	end_proc;
    end_proc;

    makeBilinearFunction :=
    proc(f)
    /**
        f is a function from dom::basisIndices x dom::basisIndices to some module over
	dom::coeffRing, where the multiplication by scalars can
	be done via multcoeffs
	
	Returns the linear extension of f to dom;

	Extra arguments of the constructed operator are passed down to f.
    **/
	option escape;
    begin
        proc(x1: dom, x2: dom)
	    local s1, s2;
	begin
	    _plus(multcoeffs(f(op(s1,1), op(s2,1), args(3..args(0))),
			     op(s1,2)*op(s2,2))
		  $ s1 in REP(x1)
		  $ s2 in REP(x2));
	end_proc;
    end_proc;

    makeBilinearProduct_Table :=
    proc(product)
        // product returns the product of two basis elements of F, as a table
        // table(term1 = coeff1, ...) or a list [term1=coeff1, ...]
	local DOM;
	option escape;
    begin
	DOM := dom;
	proc(x1: DOM, x2: DOM): DOM
	    local t, s1, s2, r, s;
	begin
	    t := table();
	    for s1 in REP(x1) do
		for s2 in REP(x2) do
		    r := op(s1,2)*op(s2,2);
		    for s in product(op(s1,1), op(s2,1)) do
			t:=DOM::addToRep(t,op(s,1), r*op(s,2)); // R ?
		    end_for;
		end_for
	    end_for;
	    return(new(dom, t));
	end_proc;
    end_proc;

    makeBilinearProduct := dom::makeBilinearFunction;

begin
    Options := combinat::getOptions([ZeroCheck = TRUE], [ZeroCheck = DOM_BOOL], 2, Error);
end_domain:

unalias(REP,TYPE_ERR):
