/*++
Dom::PermutationGroupInvariantRing(R,G, [options]) --
    R: ring
    G: permutation group
    
    Domain for the invariant ring of the permutation group G,
    over the base ring R
    
Inherits representation and bugs from InvariantMonoidAlgebra.

Author: Nicolas M. Thiry <nthiery@users.sourceforge.net>
Created: 24/10/1998
Last update: $Id: PermutationGroupInvariantRing.mu,v 1.1.1.7 2005/01/26 13:42:22 nthiery Exp $

WARNING:

Many domain expressions depends on the primary invariants!
If the primary invariants are changed, those need to be reset/recalculated!
Shall we use methods instead of domain expressions?

Should the elements of basisIndices be typed ?
Advantage:
 - use overloading for _plus, subtract, _less, ...

Timings: operand access to a domain element compared to a list is not
seriously more expensive (<10%)

x0:=[1,2,3]:
domain bla1 end:
x1:=new(bla1, 1,2,3):
domain bla2 op := extop; end:
x2:=new(bla2, 1,2,3):
domain bla3 op := (x,i) -> op(extop(x),i); end;
x3:=new(bla2, [1,2,3]):

>> time(for i from 1 to 1000000 do end)
                                   1190

>> time(for i from 1 to 1000000 do op(x0,1); end) - 1190
                                   1490
>> time(for i from 1 to 1000000 do op(x1,1); end) - 1190
                                   1600
>> time(for i from 1 to 1000000 do op(x2,1); end) - 1190
                                   2660
>> time(for i from 1 to 1000000 do op(x3,1); end) - 1190
                                   2650

Option 1 is not unreasonnable, except that we can't zip directly on it, which makes all the zip operations expensive.

>> time(for i from 1 to 1000000 do new(bla1, op(zip([extop(x1)],[extop(x1)],_plus))) end)$

                                   17380
>> x0:=[1,2,3]
                                 [1, 2, 3]
>> time(for i from 1 to 1000000 do zip(x0,x0,_plus) end) - 1190
                                   8610


++*/

domain Dom::PermutationGroupInvariantRing(R: DOM_DOMAIN,
					  G: DOM_DOMAIN)
    local options;
    inherits Dom::FreeModulePoly(R, combinat::integerVectorsModPermutationGroup(G), hold(Variables)=options[hold(Variables)], MultirankerBasis = [id,id]);
    category Cat::PermutationGroupInvariantRing(R);

    interfaceAdd := {
		  hold(isGorenstein),
		  hold(degreeBound),
		  hold(degreeBound_GarsiaStantonGoebel),
		  hold(degreeBound_NoetherFleishmannSchmidDomokosHegedusSezer),
		  hold(isPrime),
		  hold(divisors),
		  hold(smallPrimeDivisors),
		  hold(primes),
		  hold(SAGBI),
		  hold(normalfModNonPrimes),
		  hold(criticalDivisors),

		  hold(minimalGeneratingSet),
		  hold(secondaryInvariantsShape),
		  hold(decomposablesModNonPrimesBasis),
		  hold(normalfModDecomposables),
		  hold(isDecomposable),
		  hold(isGeneratingSet),

		  hold(secondaryInvariants),
		  hold(irreducibleSecondaryInvariants)
		 };

    options := options;

    chainProduct := options[hold(ChainProduct)];
    useLinBox := options[hold(UseLinBox)];
    useALP := options[hold(UseALP)];
    variables := options[Variables];
    basisName := hold(o);

    /////////////////////////////////////////////////////////////////////////
    // product rule
    /////////////////////////////////////////////////////////////////////////
/* Generic form:
    mult2Basis :=
    proc(v1: dom::basisIndices, v2: dom::basisIndices) : dom
    begin
	o2 := classSize(v2);
	dom::fromList([ [o2/classSize(v), v]
                        $ v in map(class(v1), operation, v2) ]);
    end_proc;
*/
    oneBasis := [ 0 $ dom::dimen ];

    straightenBasis := dom::term @ dom::basisIndices::straighten;

    _divide := () -> FAIL;

    mult2Basis :=
    proc(v1: dom::basisIndices, v2: dom::basisIndices) : dom
	local o2, v, shape,
	      isok; // only the terms satifying the predicate isok will be kept
    begin
	o2 := dom::basisIndices::orbitSize(v2);
	if dom::chainProduct then
	    // The sum of the shapes
	    // Small optimization: dom::shape is inlined, and the shapes
	    // are sorted increasingly and not decreasingly
	    shape:=zip(sort(v1),sort(v2), _plus);
	    // Predicate: the shape of the sum is the sum of the shape
	    isok := v -> bool(sort(v)=shape);
	else
	    isok := TRUE;
	end_if;
	dom::fromList([ [o2/dom::basisIndices::orbitSize(v),
			 dom::basisIndices::straighten(v) ]
		       $ v in select(map(dom::basisIndices::orbitSet(v1),
					 zip, v2, _plus),
				     isok)]);
    end_proc;

    degreeBasis := _plus@op;

    /////////////////////////////////////////////////////////////////////////
    // Utilities for converting with shapes
    /////////////////////////////////////////////////////////////////////////

    // shapeSet2List(S: dom::TypeShapeSet) -> dom::TypeShapeList

    shapeSet2List :=
    proc(S: dom::TypeShapeSet)
	local l, s, j;
    begin
	l:=[ 0 $ dom::dimen ];
	for s in S do
	    for j from 1 to s do
		l[j]:=l[j]+1;
	    end_for;
	end_for;
	l;
    end_proc;

    // shapesByDegree(d: Type::NonNegInt)
    //	-> list of dom::TypeShapeSet
    //
    // Returns all the shapes of degree d, eg all subsets of {1..dom::dimen-1}
    // of total sum d
    shapesByDegree :=
    proc(d: Type::NonNegInt) : Type::ListOf(dom::TypeShapeSet)
    begin
	map(combinat::partitions::list(d, MaxPart=dom::dimen-1, MaxSlope=-1),
	    DOM_SET@op)
    end_proc;

    /////////////////////////////////////////////////////////////////////////
    // Usefull types
    /////////////////////////////////////////////////////////////////////////
    
    // Shall we specify that the elements of a TypeRowReducedList
    // should be sorted w.r.t. their leading term ?
    TypeRowReducedList	 := Type::ListOf(dom);
    TypeShapeSet	 := Type::SetOf(Type::PosInt);
    TypeShapeList	 := combinat::integerVectors;
    TypeThisList	 :=
    Type::Union(Type::ListOf(dom),
		Type::TableOfEntry(Type::ListOf(dom)),
		Type::TableOfEntry(Type::TableOfEntry(Type::ListOf(dom))));

    /////////////////////////////////////////////////////////////////////////
    // Basic information on the invariant ring
    /////////////////////////////////////////////////////////////////////////

    dimen := G::card;
    group := G;

    /////////////////////////////////////////////////////////////////////////
    // Conversions
    /////////////////////////////////////////////////////////////////////////

    polyBasis :=
    proc(t: dom::basisIndices)
	local t2;
    begin
	dom::Poly([[ R::one, t2 ]
		   $ t2 in dom::basisIndices::orbitSet(t)]);
    end_proc;

    poly :=
    operators::makeLinear(dom::polyBasis,
			  Source=dom,
			  Zero=dom::Poly::zero);

    fromPoly :=
    proc(P//: dom::Poly
	 ) : dom
    begin
	dom::fromList(select(poly2list(P),
			     dom::basisIndices::isStraight @ op, 2));
    end_proc;

    /////////////////////////////////////////////////////////////////////////
    // Mathematical information on the invariant ring
    /////////////////////////////////////////////////////////////////////////

    // isModular: TRUE/FALSE
    //
    // Whether the invariant ring is modular, ie the size |G| of the
    // group is invertible in the base field. In the non-modular case,
    // one can define the reynolds operator:
    //			p -> \frac{1}{|G|} \sum_$g\in G$ g.p,
    // This yields nice properties, including the Cohen-Macaulay
    // property.

    
    // isGorenstein: TRUE/FALSE
    //
    // Whether the invariant ring is Gorenstein. A necessary and
    // sufficient condition is that the group of is a subgroup of
    // SL(dimen), otherwise said that all the permutations in the
    // group are even.

    isGorenstein := bool(_and( sign(g)=1 $ g in dom::group::generators));
    
    //////////////////////////////////////////////////////////////////////////
    // Degree bound
    //////////////////////////////////////////////////////////////////////////

    // degreeBound()
    //
    // Returns the best known upper bound on the maximal degree of
    // the invariants in a minimal generating set. By default, the best
    // of the following bounds is taken:
    //  - degreeBound_NoetherFleishmannSchmidDomokosHegedusSezer
    //    general degree bound for finite groups in the non modular case.
    //	- degreeBound_GarsiaStantonGoebel
    //    general degree bound for permutation groups
    //	- max(degs primary invariants, degs secondary invariants)
    //	  If the invariant ring is Cohen-Macaulay.
    //	- If moreover the invariant ring is Gorenstein, and there are
    //	  strictly more than 2 secondary invariants, the previous
    //	  bound can be improved. Indeed, the secondary invariant of
    //	  highest degree Dmax is not irreducible (see Chapter 21 of
    //	  "Commutative Algebra ..." by Eisenbud, in particular,
    //	  Exercise 21.9 on page 551-552).
    //	  We don't know if this is still valid with the chain product.
    //
    // If you know a better degree bound, you can set it manually with
    // myRing::degreeBound():=13
    // 
    // This entry is updated by some computations like minimal
    // generating sets, secondary Invariants which yield better degree
    // bounds as a by-product.

    // degreeBound_APriori() -> Type::NonNegInt
    //
    // Same as degreeBound(), except that it does not get updated.

    // degreeLowerBound() -> Type::NonNegInt
    //
    // Best lower bound available on the maximal degree of the
    // invariants in a minimal generating set. Also updated by
    // certain computations.

    // degreeBoundIsOptimal() -> DOM_BOOL
    //
    // Whether the degree bounds are tight, e.g. whether any minimal
    // generating set contains an invariant of degree degreeBound.
    // Unless explicitely set, it's TRUE iff
    // degreeBound=DegreeLowerBound, and UNKNOWN otherwise.

    // degreeBound_GarsiaStantonGoebel() -> Type::NonNegInt
    //
    // Degree bound of Garsia & Stanton 1984 in coprime characteristic
    // Enhanced in Smith PIF 1995 for non transitive permutation groups
    // Generalized in Goebel.1996 for any characteristic
    //
    // If G acts transitively on the vector space basis:
    //   max(dim, binomial dim,2)
    // If not, max(bound) over each orbit

    degreeBound_GarsiaStantonGoebel :=
    proc()
        option remember;
        local bound, m, dim;
    begin
        bound:=1;
        for m in dom::basisIndices(1) do
            dim  :=dom::basisIndices::orbitSize(m);
            bound:=max(bound,dim, binomial(dim,2));
        end_for;
        bound;
    end_proc;

    degreeBound_NoetherFleishmannSchmidDomokosHegedusSezer :=
    proc()
    //  - degreeBound_NoetherFleishmannSchmidDomokosHegedusSezer()
    //
    //    General degree bound for finite groups in the non-modular case:
    //	  Noether 1916: |G| (valid for p=0 and p>|G|)
    //    Extended to p<|G| by Fleishmann 1999
    //    Enhanced by Schmid (1991) to: |G|-1 
    //    Enhanced for non cyclic group by Domokos and Hegedus (2000) to:
    //     - 3/4.|G| if |G| even
    //     - 5/8.|G| if |G| odd
    //    Extended to p<|G| by Sezer (2001)
    //
    //	  We don't know if this is still valid with the chain product.

    begin
        if dom::isModular or dom::chainProduct then
            return(FAIL);
        end_if;
        if dom::group::isCyclic() then
            dom::group::size;
        elif testtype(dom::group::size, Type::Even) then
            floor(dom::group::size*3/4);
        else
            floor(dom::group::size*5/8);
        end_if;
    end_proc;

    degreeBound_APriori :=
    proc()
	option remember;
	local b, b2, p;
    begin
	b:=dom::degreeBound_GarsiaStantonGoebel();
        userinfo(3, "degreeBound: from the dimension of the space: ".b);
	if dom::isCohenMacaulay=TRUE then
	    b2 := max(op(dom::primaryInvariantsDegrees()),
                      (if dom::isGorenstein and not dom::chainProduct and
			 dom::numberOfSecondaryInvariants()>2 then
			   p:=dom::secondaryInvariantsSeries(hold(z));
			   degree(p-lmonomial(p));
		       else
			   dom::secondaryInvariantsMaxDegree();
		       end_if));
	    if b2 < b then
	    userinfo(3, "degreeBound: enhancement from the degrees of the primary invariants: ". b2);
		b:=b2;
	    end_if;
	end_if;
	if not dom::isModular and not dom::chainProduct then
            b2:=dom::degreeBound_NoetherFleishmannSchmidDomokosHegedusSezer();
            if b2 < b then
                userinfo(3, "degreeBound: enhancement from the size of the group: ".b2);
                b := b2;
            end_if;
	end_if;
	b;
    end_proc;

    degreeBound := () -> dom::degreeBound_APriori();

    degreeLowerBound :=
    proc()
	option remember;
    begin;
	1;
    end_proc;

    degreeBoundIsOptimal :=
    proc()
    begin
	if dom::degreeBound()=dom::degreeLowerBound() then
	    TRUE;
	else
	    UNKNOWN;
	end_if
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // Vector space basis of the invariant ring
    //////////////////////////////////////////////////////////////////////////

/*    // basisIndices (d: Type::NonNegInt) -> Type::ListOf(dom::basisIndices)
    //
    // Elements of degree d of dom::basisIndices, sorted w.r.t. the term ordering

    basisIndices :=
    proc(d: Type::Union(Type::NonNegInt, dom::TypeShapeList))
	option remember;
    begin
	if domtype(d)=DOM_INT then
	    sort([dom::basisIndices(dom::group::canonic(m))
                  $ m in dom::group::canonicCompositions(d) ],
                 dom::basisIndices::_less);
	else
	    sort([dom::basisIndices(dom::group::canonic(m))
		  $ m in dom::group::canonicListsByShape(d) ],
		 dom::basisIndices::_less);
	end_if
    end_proc; */

    // Specification: if two monomials m1 and m2 have same degree and m1<m2,
    // then index(m1)<index(m2)

    // This is not fully specified yet. Should we number each
    // homogeneous componnents independently ? Or should we make it so
    // that for any two monomials (possibly of different degree) such
    // that m1<m2 (w.r.t. the term ordering), index(m1)<index(m2) ?

    basisElementIndex :=
    proc(m: dom::basisIndices)
        local i, m1, found;
    begin
        i:=1; found:=FALSE;
        for m1 in dom::basisIndices(_plus(op(m))) do
            dom::basisElementIndex(m1):=i;
            i:=i+1;
            if not found and m=m1 then found:=TRUE; end_if;
        end_for;
        if not found then 
            error(m, "not found in the basis elements"); 
        end_if;
        return(dom::basisElementIndex(m));
    end_proc;

    // sparseVector(p: dom) : Type::ListOf(DOM_LIST)
    //
    // Returns a sparse vector representing the homogeneous element p
    // This is a list of the form [[n1, c1], [n2,c2], ...],
    // Where the ni's uniquely represents the elements of the basis of
    // the invariant ring at the degree of p, and the ci's are the
    // corresponding coefficients of p.
    //
    // This representation can be fed directly into a sparse matrix,
    // to make linear algebra computations between elements of *same*
    // degree.
    //
    // This is not cleanly specified yet: the problem is that most
    // sparse matrix systems make row echelon forms w.r.t increasing
    // indexes, whereas for algebras, the natural way is the converse,
    // in order to be consistent with the term ordering.

    sparseVectorOld :=
    proc(p: dom)
        local vect, m;
    begin
        vect:=[];
        for m in dom::support(p) do
            vect:=append(vect, [
                                nops(dom::basisIndices(_plus(op(m)))) -
                                dom::basisElementIndex(m), coeff(p,m)]);
        end_for;
        sort(vect,(x,y)->op(x,1)<op(y,1));
    end_proc;

    /////////////////////////////////////////////////////////////////////////
    // It would be reasonable to put extra stuff about polynomials here

    /////////////////////////////////////////////////////////////////////////
    // Some classical symmetric polynomials
    /////////////////////////////////////////////////////////////////////////

    powersum :=
    proc(i: Type::NonNegInt)
	local v;
    begin
	_plus(dom::term(v) $ v in dom::basisIndices([i, 0 $ dom::dimen-1]));
    end_proc;

    elementary :=
    proc(d: Type::NonNegInt)
	local v;
    begin
	_plus(dom::term(v) $ v in dom::basisIndices([1$d, 0 $ dom::dimen-d]));
    end_proc;

    symmetricMonomial :=
    proc(shape: dom::TypeShapeList)
    begin
	_plus(op(dom::basis(shape)));
    end_proc;

    /////////////////////////////////////////////////////////////////////////
    // Operations on the initial poset; SAGBI basis
    /////////////////////////////////////////////////////////////////////////

    /**
     * shape(x: dom::basisIndices)
     * Return the list of the valuations in x, sorted decreasingly
     * eg: shape([5,2,3,6,2,0]) -> [6,5,3,2,2,0]
     **/
    shape := revert@sort;

    divides :=
    proc(m1: dom::basisIndices, m2: dom::basisIndices) : DOM_BOOL
	local m;
    begin
	if not _lazy_and(op(zip(m1, m2, _leequal))) then
	    return(FALSE);
	end_if;
	m:=zip(m2,m1,_subtract);
	if dom::chainProduct and sort(m2) <> zip(sort(m1), sort(m), _plus) then
	    return(FALSE);
	end_if;
	dom::basisIndices::isStraight(m);
    end_proc;

    // aDivisor(m dom::basisIndices) -> dom::basisIndices
    //
    // Returns the smallest (for the term ordering) prime divisor of
    // m, or FAIL if there is none
    
    aDivisor :=
    proc(m: dom::basisIndices) : Type::Union(dom::basisIndices, DOM_FAIL)
	local d1, m1;
    begin
	for d1 from 1 to floor(_plus(op(m))/2) do
	    for m1 in dom::primes(d1) do
		if dom::divides(m1,m) then
		    return(m1);
		end_if
	    end_for
	end_for;
	FAIL;
    end_proc;

    // isPrime(m: dom::basisIndices) -> DOM_BOOL
    //
    // Return true if m is prime
    
    isPrime :=
    proc(m: dom::basisIndices)
    begin
	return(bool(dom::aDivisor(m)=FAIL));
    end_proc;

    // divisors(m: dom::basisIndices) -> Type::ListOf(dom::basisIndices)
    //
    // Return the list of prime elements which divides strictly m

    divisors :=
    proc(m: dom::basisIndices) : Type::ListOf(dom::basisIndices)
	local divisors, m1, d1;
    begin
	divisors:=[];
	for d1 from 1 to _plus(op(m)) - 1 do
	    for m1 in dom::primes(d1) do
		if dom::divides(m1,m) then
		    divisors:=append(divisors,m1);
		end_if
	    end_for;
	    if d1=floor(_plus(op(m))/2) and nops(divisors)=0 then
		return(divisors);
	    end_if
	end_for;
	divisors;
    end_proc;

    // smallPrimeDivisors(m: dom::basisIndices) -> Type::ListOf(dom::basisIndices)
    //
    // Return the list of prime elements of degree <= degree(m)/2
    // which divides strictly m. Exception: if m=s1+s2 with s1 and s2
    // prime, and s1<s2 then only s1 will be in the list.

    smallPrimeDivisors :=
    proc(m: dom::basisIndices)
	local divisors, d1, m1;
    begin
	divisors:=[];
	for d1 from 1 to floor(dom::basisDegree(m)/2) do
	    for m1 in dom::primes(d1) do
		if dom::divides(m1,m) and
		   (2*d1<dom::basisDegree(m) or contains(divisors,dom::basisSubtract(m,m1))=0) then
		    divisors:=append(divisors,m1);
		end_if
	    end_for
	end_for;
	return(divisors);
    end_proc;

    // primes(d: Type::NonNegInt) -> Type::ListOf(dom::basisIndices)
    //
    // Returns the prime elements of degree d sorted w.r.t. the term ordering
    //
    // There are two implementations: primes_FromTop, and
    // primes_EratostheneSieve. The second one avoid calls to divide
    // and thus to isCanonic, and is a little bit faster.
    //
    // By default, primes=primes_EratostheneSieve.

    primes := dom::primes_EratostheneSieve;
    
    primes_FromTop :=
    proc(d: Type::NonNegInt)
	option remember;
    begin
	select(dom::basisIndices(d), dom::isPrime);
    end_proc;

    primes_EratostheneSieve :=
    proc(d: Type::NonNegInt)
	option remember;
	local d1, m1, m2, t;
    begin
	t:=table();
	for d1 from 1 to floor(d/2) do
	    for m1 in dom::primes(d1) do
		for m2 in dom::basisIndices(d-d1) do
		    t[zip(m1,m2,_plus)] := TRUE;
		end_for
	    end_for
	end_for;
	select(dom::basisIndices(d), m -> not contains(t,m));
    end_proc;

    // constructInitialPoset(d: Type::NonNegInt) -> DOM_NULL;
    //
    // Construct and store the structure of the initial poset. This
    // takes up some memory, but computations with
    // minimalGeneratingSet or normalfModNonPrimes are much faster
    // afterward.
    // 
    // Side effects: compute and store in remember tables the values of
    //	- primes(d)
    //	- smallPrimeDivisors(m), for m in dom::basisIndices(d)
    //	- aDivisor(m),		 for m in dom::basisIndices(d)

    constructInitialPoset := () -> dom::constructInitialPoset_Divisors(args());
    // constructInitialPoset := dom::constructInitialPoset_Divisors ought
    // to be enough, but there seems to be yet another bug with remember 
    // tables and domain elements.

    // constructInitialPoset_Divisors(d: Type::NonNegInt) -> null();

    constructInitialPoset_Divisors :=
    proc(d: Type::NonNegInt)
	option remember;
	local primes, smallDivisors, d1, m, m1, m2;
    begin
	// Ensure all the information is already computed for d1<d
	for d1 from 1 to d-1 do
	    dom::constructInitialPoset_Divisors(d1, args(2..args(0)));
	end_for;
	userinfo(3, "Constructing the initial poset at degree ". d);
	smallDivisors:=table();
	for d1 from 1 to d-1 do
	    for m1 in dom::primes(d1) do
		for m2 in dom::basisIndices(d-d1) do
		    m := zip(m1, m2, _plus);
		    if contains(smallDivisors, m) then
			if   2*d1<d or
			    (2*d1=d and contains(smallDivisors[m],m2)=0) then
			    smallDivisors[m]:=append(smallDivisors[m],m1);
			end_if;
		    else
			smallDivisors[m]:=[m1];
		    end_if
		end_for
	    end_for
	end_for;
	primes:=[];
	for m in dom::basisIndices(d) do
	    if contains(smallDivisors,m) then
		dom::smallPrimeDivisors(m):=smallDivisors[m];
		dom::aDivisor(m):=smallDivisors[m][1];
	    else
		dom::aDivisor(m):=FAIL;
		dom::smallPrimeDivisors(m):=[];
		primes:=append(primes,m);
	    end_if
	end_for;
	dom::primes(d):=primes;
	null():
    end_proc;

    // constructInitialPoset_LCMS(d: Type::NonNegInt) -> null();
    //
    // Alternative implementation which also updates:
    //	- LCMS(s1,s2), isLCM(s1,s2,m),
    //		for s1, s2 prime, m in dom::basisIndices(d)

    constructInitialPoset_LCMS :=
    proc(d: Type::NonNegInt)
	option remember;
	local primes, divisors, smallDivisors, d1, m, m1, m2, m3;
    begin
	// Ensure all the information is already computed for d1<d, in
	// particular for isLCM/LCMS.
	for d1 from 1 to d-1 do
	    dom::constructInitialPoset_LCMS(d1, args(2..args(0)));
	end_for;
	userinfo(3, "Constructing the initial poset at degree ". d);
        divisors:=table();
	smallDivisors:=table();
	for d1 from 1 to d-1 do
	    for m1 in dom::primes(d1) do
		for m2 in dom::basisIndices(d-d1) do
		    m := zip(m1, m2, _plus);
		    if contains(divisors, m) then
			for m3 in divisors[m] do
			    // Assertion: m3<m1.
			    // This ensures storage and lookup in the
			    // remember table is always done with the
			    // two first arguments in the same order.
			    dom::isLCM(m3,m1,m);
			end_for;
			divisors[m]:=append(divisors[m],m1);
			if   2*d1<d or
			    (2*d1=d and contains(smallDivisors[m],m2)=0) then
			    smallDivisors[m]:=append(smallDivisors[m],m1);
			end_if;
		    else
			divisors [m]:=[m1];
			smallDivisors[m]:=[m1];
		    end_if
		end_for
	    end_for
	end_for;
	primes:=[];
	for m in dom::basisIndices(d) do
	    if contains(divisors,m) then
		// dom::divisors(m):=divisors[m]; // ?
		dom::smallPrimeDivisors(m):=smallDivisors[m];
		dom::aDivisor(m):=divisors[m][1];
	    else
		dom::aDivisor(m):=FAIL;
		primes:=append(primes,m);
	    end_if
	end_for;
	dom::primes(d):=primes;
	null():
    end_proc;

    // SAGBI(d: Type::NonNegInt) -> Type::ListOf(dom)
    //
    // Returns the elements of degree d of the minimal simple SAGBI
    // basis
    
    SAGBI :=
    proc(d: Type::NonNegInt)
    begin
	map(dom::primes(d), dom);
    end_proc;

    // normalfModNonPrimes(p: dom) -> dom
    //
    // Preconditions: p is homogeneous of degree d
    //
    // Compute the remainder of p on invariant division of p by the
    // elements of the SAGBI basis of degree < d. The result is a
    // linear combination of prime invariants.
    // This defines a normal form of p modulo the subspace generated
    // by non prime invariants of degree d.

    normalfModNonPrimes :=
    proc(p: dom)
    local m,c,lt,m1,result;
    begin
	result:=dom::zero;
	while(not iszero(p)) do
	    m:=dom::lsupport(p);
	    c:=coeff(p,m);
	    lt:=dom::monomial(c,m);
	    if ((m1:=dom::aDivisor(m))<>FAIL) then
		p :=p-c*dom(m1)*dom(m-m1);
		// Remark: we use here the fact that the leading
		// coefficient of the product of two orbit sums is 1.
	    else
		p:=p-lt;
		result:=result+lt;
	    end_if
	end_while;
	return(result);
    end_proc;

    // isLCM(m1, m2, m) -> DOM_BOOL
    //
    // Precondition: m is a common multiple of m1 and m2.
    // Guaranted to return TRUE if m is a least common multiple of m1
    // and m2. Depending on the history, it may return TRUE even if m
    // is not a least common multiple.
    //
    // More precisely, it returns TRUE if and only if isLCM(m1,m2,m')
    // has not been called before for a common multiple m' of m1 and
    // m2 which divides m. One way to ensure that isLCM only detects
    // real least common multiples is to call isLCM(m1,m2,m) for all
    // common multiples m of m1 and m2 by ascending degree.
    //
    // This uses the remember table of the module function LCMS(m1,m2).
    
    isLCM :=
    proc(m1: dom::basisIndices, m2: dom::basisIndices, m: dom::basisIndices)
	local m3;
    begin
	for m3 in dom::LCMS(m1,m2) do
	    if _plus(op(m3)) = _plus(op(m)) then
		break;
	    end_if;
	    if dom::divides(m3,m) then
		return(FALSE);
	    end_if
	end_for;
	dom::LCMS(m1,m2):=append(dom::LCMS(m1,m2),m);
	return(TRUE);
    end_proc;

    LCMS :=
    proc(m1: dom::basisIndices, m2: dom::basisIndices)
    begin
	return([]);
    end_proc;

    // LCMPartition((divisors: Type::ListOf(dom::basisIndices), m: dom::basisIndices)
    //	-> Type::ListOf(Type::SetOf(dom::basisIndices))
    //
    // Precondition: the elements of divisors are all divisors of m
    //
    // Returns the set partition of the divisors defined as follow:
    //
    // Consider the relation R between divisors of m defined by
    //		m1 R m2 iff m is not a LCM of m1 and m2.
    // The equivalence relation R* obtained by transitive closure of
    // R defines the wanted set partition of the divisors.
    //
    // It's guaranted that the first element of divisors will be in
    // the first subset of the list. This is needed by
    // minimalGeneratingSetDegree, to ensure the smallest divisor will
    // be in the first subset.

    LCMPartition :=
    proc(divisors: Type::ListOf(dom::basisIndices), m: dom::basisIndices)
	local partition, newPartition, linkedToPart, m1, m2, part, m1Part;
    begin
	partition:=[];
	for m1 in divisors do
	    newPartition:=[];
	    m1Part:=0; // Part to which m1 belongs; 0 if not known yet
	    for part in partition do
		linkedToPart:=FALSE;
		for m2 in part do
		    // Assertion: m2<m1.
		    // This ensures storage and lookup in the
		    // remember table is always done with the
		    // two first arguments in the same order.
		    if not dom::isLCM(m2,m1,m) then
			linkedToPart:=TRUE;
			break;
		    end_if;
		end_for;
		if linkedToPart then
		    if m1Part=0 then
			newPartition:=append(newPartition,
					     part union {m1});
			m1Part:=nops(newPartition);
		    else
			newPartition[m1Part]:=
			    newPartition[m1Part] union part;
		    end_if;
		else
		    newPartition:=append(newPartition,part);
		end_if
	    end_for;
	    if m1Part=0 then
		newPartition:=append(newPartition, {m1});
	    end_if;
	    partition:=newPartition;
	end_for;
	partition;
    end_proc;

    // criticalDivisors(m: dom::basisIndices) : Type::ListOf(dom::basisIndices)
    //
    // Let Decomposables(d) be the space of decomposable invariants of
    // degree d, that is the subspace of the homogeneous componnent
    // I(G)_d spanned by all products homogeneous pq with degree(p)<d,
    // degree(q)<d, and degree(pq)=d.
    // 
    // Let m be a monomial of degree d. Let Decomposables(d, <=m) be
    // the subspace of Decomposables(d) spanned by the above products,
    // with the additional constraint that lt(pq)<=m. Similarily, let
    // Decomposables(d, <m) be the subspace of Decomposables(d)
    // spanned by the above products, with the additional constraint
    // that lt(pq)<m.
    //
    // criticalDivisors(m) returns a list [s1,...,sk] of divisors of m,
    // such that the products si^*(m/si)^* together with
    // Decomposables(d, <m) span Decomposables(d, <=m).
    //
    // Obviously, returning all divisors of m would do the job, but we
    // can do much better with the following remark, which is
    // essentially the Buchberger criterions revisited:
    //
    // If s1s2=m then their s-polynomial s1^*(m/s1)^*-s2^*(m/s2)^* is zero.
    //
    // If s1/m1, s2/m1 and m1/m (that is, if m is not a least common
    // multiple of s1 and s2), then their s-polynomial is already in
    // Decomposables(d, <m).
    //
    // In both case, we can safely ignore the critical pair. Actually,
    // we can improve on this. If m is not a LCM of s1 and s2, and
    // m is not a LCM of s2 and s3, m might still be a LCM of s1 and
    // s3, but the s-polynomial of s1 and s3 will still be in
    // Decomposables(d, <m). That is we can take the transitive
    // closure of the relation "m is not a LCM of ....", and ignore
    // all the corresponding s-polynomials.
    //
    // One can prove that considering only the small prime divisors
    // does not change the obtained critical divisors.

    criticalDivisors := dom::criticalDivisors_Divisors;

    // criticalDivisors_LCMS is a straightforward implementation of this.

    criticalDivisors_LCMS :=
    proc(m: dom::basisIndices) : Type::ListOf(dom::basisIndices)
    begin
        map(dom::LCMPartition(dom::smallPrimeDivisors(m), m), op, 1);
    end_proc;

    basisDegree   := _plus@op;
    basisPlus2    := (x,y)->zip(x,y,_plus);
    basisSubtract := (x,y)->zip(x,y,_subtract);
    basisMin      :=
    proc() : dom::basisIndices
	local x, i;
    begin
	x := args(1);
	for i from 2 to args(0) do
	    if combinat::words::lexLess(args(i), x) then
		x := args(i);
	    end_if;
	end_for;
	x;
    end_proc;
	
    // Computing LCMs is (with the current implementation) costly, and
    // we can give another formulation of the criterions above that is
    // also trivial, but that can be implemented using less memory and
    // computation time.
    //
    // Let S1 be the set containing s1 and all the prime divisors of
    // m/s1. Let S2 be the set containing s2 and all the prime
    // divisors of m/s2. If the intersection of S1 and S2 is non
    // empty, then the s-polynomial of s1 and s2 is in
    // Decomposables(d, <m). Once again, one can take the transitive
    // closure of this relation.
    //
    // Here also, one can prove that considering only the small prime
    // divisors does not change the obtained critical divisors.
    //
    // criticalDivisors_Divisors is a straightforward implementation of this.

    criticalDivisors_Divisors :=
    proc(m: dom::basisIndices) : Type::ListOf(dom::basisIndices)
        local divisors, partition, newPartition, m1, m1Part, m1Pos, part;
    begin
        divisors:=dom::smallPrimeDivisors(m);
	partition:=[];
	for m1 in divisors do
	    newPartition:=[];
            m1Part:={m1} union {op(dom::smallPrimeDivisors(dom::basisSubtract(m,m1)))};
	    m1Pos:=0; // Part to which m1 belongs; 0 if not known yet
	    for part in partition do
                if nops(part intersect m1Part)>0 then
		    if m1Pos=0 then
			newPartition:=append(newPartition,
					     part union {m1});
			m1Pos:=nops(newPartition);
		    else
			newPartition[m1Pos]:=
			    newPartition[m1Pos] union part;
		    end_if;
		else
		    newPartition:=append(newPartition,part);
		end_if
	    end_for;
	    if m1Pos=0 then
		newPartition:=append(newPartition, {m1});
	    end_if;
	    partition:=newPartition;
	end_for;
	map(partition, dom::basisMin@op);
    end_proc;

    // SPair(m1, m2, m) -> dom
    //
    // Precondition: m is a common multiple of m1 and m2.
    // Returns the S-pair relative to m of the orbit sums of m1 and m2
    //
    // Remark: no need to worry about the leading coefficients. Indeed,
    // the leading coefficient of the product of any number of orbit
    // sums is always one.

    SPair :=
    proc(m1, m2, m)
    begin
	dom(m2)*dom(m-m2)-dom(m1)*dom(m-m1);
    end_proc;

    /////////////////////////////////////////////////////////////////////////
    // Computations a minimal generating sets
    /////////////////////////////////////////////////////////////////////////
    //
    // minimalGeneratingSet([options]) -> Type::TypeThisList
    // 
    // Computes a minimal homogeneous generating set of the invariant ring.
    //
    // It's returned as a table t, where t[d] is the list of the
    // generators of degre d, or is unassigned if there is none.
    //
    // Options:
    //	 - maxDegree: Type::NonNegInt
    //	   only runs the computation up to degree maxDegree
    //
    // Side effects:
    //	- updates dom::degreeBound(), and dom::degreeLowerBound().
    //	- compute and store information on the initial poset, including:
    //	  - prime elements,
    //	  - partial SAGBI basis,
    //	  - LCMS, isLCM,
    //	  - divisors
    //	- sets decomposablesModNonPrimes(...)

    // minimalGeneratingSetInit([options])
    // minimalGeneratingSet(d)		  -> Type::ListOf(dom)
    //
    // If only the (irreducible) secondary invariants of some degree d
    // are needed, one can call directly minimalGeneratingSet(d).
    // minimalGeneratingSetInit([options]) should be called beforehand.

    minimalGeneratingSet :=
    proc()
	local d,t,mgs,l;
	option remember;
    begin
	
	// Calls with just one integer argument
	
	if args(0)=1 and not(testtype(args(1), Type::Option)) then
	    return(dom::minimalGeneratingSetDegree(args(1)));
	end_if;

	// Calls with no arguments, or just options
	
	experimental::computationHeader(
	"Computation of a minimal generating set using SAGBI+linear algebra.");
	t     :=time();
	mgs   :=table();
	dom::minimalGeneratingSetInit(args());

	// Main loop

	for d from 0 to dom::minimalGeneratingSetOptions[hold(maxDegree)] do
	    l:=dom::minimalGeneratingSet(d);
	    if nops(l)>0 then
		mgs[d]:=l;
	    end_if;
	end_for:

	// Update of the degree bound

	if d >= dom::degreeBound() then
	    for d from dom::degreeBound() downto 0 do
		if contains(mgs,d) then
		    break;
		end_if
	    end_for;
	    if d<dom::degreeLowerBound() then
		userinfo(0,"The degree lower bound was incorrect!");
	    end_if;
	    if d<dom::degreeBound() then    
		userinfo(3, "Updating the degreeBound to ".d." which is optimal!");
		dom::degreeBound() := d;
	    end_if;
	end_if;

	// User information

	userinfo(3, "Total time: ".experimental::time::format(time()-t));

        // Bugware for MuPAD 2.0.0: the remember table of
        // dom::minimalGeneratingSet does not get updated properly
        // because we have modified it manually from inside.

        dom::minimalGeneratingSet(args()):=mgs;

	return(mgs);
    end_proc;

    // decomposablesModNonPrimes(d: Type::NonNegInt)
    //	-> Type::ListOf(dom) or hold(allPrimes)
    //
    // Return a row reduced basis of the projection by
    // normalfModNonPrimes on the primes invariants of the subspace
    // of the homogeneous decomposable invariants of degree d.
    // As a special case, if every homogeneous invariant of degree d
    // is decomposable, the value hold(allPrimes) is returned instead

    decomposablesModNonPrimesBasis :=
    proc(d: Type::NonNegInt)
    begin
	dom::minimalGeneratingSetDegree(d);
	dom::decomposablesModNonPrimesBasis(d);
    end_proc;

    minimalGeneratingSetDefaultOptions :=
    proc()
    begin
        table(hold(maxDegree) = dom::degreeBound(),
	      hold(Algo) = hold(F4),
              hold(LinearAlgebra) = hold(DynamicModule)
             );
    end_proc;

    minimalGeneratingSetInit :=
    proc()
    begin
        // Bugware against a bug of MuPAD 2.0.0 with remember tables:
        
        // When a method bla with remember option is called for the
        // first time, that is before being actually inserted in the
        // domain the result is not inserted in its remember table.
        // Forcing manually the insertion of the method in the domain
        // is enough to solve the problem.
        // So we force manually the insertion of all entries of the domain
        // This does not work for the method who has called this one!
        // See end of minimalGeneratingSet;

        %if version()=[2,0,0] then
        dom::allEntries();
        dom::group::allEntries();
        dom::constructInitialPoset_Divisors;
        dom::constructInitialPoset;
        end_if;
	dom::minimalGeneratingSetOptions :=
	   combinat::getOptions(dom::minimalGeneratingSetDefaultOptions(),
				0, Error);
	null():
    end_proc;

    // decomposableMatrix(d: Type::NonNegInt [, Skip])
    //
    // Returns a sparse matrix whose columns represent a family of
    // invariants (not necessarily linearly independent) spanning the
    // homogeneous subspace of decomposable invariants of degree d.
    //
    // The rows are indexed by the elements of dom::basisIndices(d)

    decomposableMatrix :=
    proc(d: Type::NonNegInt)
        local skip, skipped, needed, M, initials, m, m1, m2, p, divisors,
	      rank, unrank;
    begin
	if dom::minimalGeneratingSetOptions=FAIL then
	    dom::minimalGeneratingSetInit();
	end_if;
	initials:=dom::basisIndices(d);
	[rank, unrank] := combinat::rankers::fromList(initials);
        dom::constructInitialPoset(d);
        if args(0)=2 then
            skip:=bool(args(2)=hold(Skip));
        else
            skip:=FALSE;
        end_if;
        userinfo(5, "Generating the full matrix of decomposable invariants by iterating through ".
                 nops(initials). " initial monomials");
        if dom::useALP then
            M := Dom::Matrix_ALP(dom::coeffRing)();
        else
            M := Dom::Matrix(dom::coeffRing)(0,0);
        end_if;
        if skip then
            skipped:=[];
            needed:=table();
        end_if;
        for m in revert(initials) do
            userinfo(hold(NoNL), 5, ".");
            divisors:=dom::criticalDivisors(m);
            if nops(divisors)>0 then
                if nops(divisors)>1 or not skip or needed[m]=TRUE then
                    for m2 in divisors do
                        userinfo(hold(NoNL), 5, "p");
                        p:=dom::term(m2)*dom::term(dom::basisSubtract(m,m2));
                        if skip then
                            for m1 in dom::support(p) do
                                needed[m1]:=TRUE;
                            end_for;
                        end_if;
                        M := M.dom::sparseVector(p, rank);
                    end_for;
                else
                    skipped:=append(skipped, rank(m));
                end_if;
            end_if;
        end_for;
        userinfo(hold(NoNL), 5, "\n");
        if skip then
            return([M,skipped]);
        else
            return(M);
        end_if;
    end;

    // minimalGeneratingSetDegree(d) -> Type::ListOf(dom)
    //
    // Private subroutine that actually does the computation of the
    // minimal generating set at degree d.
    //
    // Side effects:
    //	- updates the degree lower bound
    //	- sets decomposablesModNonPrimesBasis(d)

    minimalGeneratingSetDegree :=
    proc(d: Type::NonNegInt)
	// option remember;
	local stat, lbasis, initials, divisors, mgs,
	    m, m1, m2, spair, i, t, s, M,
            j, r, skipped, char_row, dim, p, counter,
	    infile, outfile, rank, det, predictedmgs,
	      cols, k, rows;
    begin
	if dom::minimalGeneratingSetOptions=FAIL then
	    dom::minimalGeneratingSetInit();
	end_if;
	userinfo(3, "Constructing a minimal generating set at degree ". d);
	t:=time();
 
	// Special cases

	if d=0 or d>dom::degreeBound() or
	    (dom::isCohenMacaulay=TRUE and
	     coeff(dom::primaryInvariantsSeries(hold(z)),d)=0 and
	     coeff(dom::secondaryInvariantsSeries(hold(z)),d)=0) then
	    userinfo(1, "No generators at degree ".d);
	    dom::decomposablesModNonPrimesBasis(d):=hold(allPrimes);
	    return([]);
	end_if;

	// Initialization

	// This speeds up the following computation of normalfModNonPrimes.
	// (trade -25% time + 5% memory for IGn(5) at degree 8)
	// It also ensure isLCM has been called properly for lower degrees.

        dom::constructInitialPoset(d);
        initials:=dom::basisIndices(d);
        
        if dom::minimalGeneratingSetOptions[hold(Algo)]=hold(Buchberger) then

            // Main loop, with Buchberger algorithm

            counter:=nops(initials);// number of initial monomials to treat
            stat:=table(hold(spairs)  =0, // #spairs which did not reduce to 0
		hold(spairszg)=0, // #spairs which reduce to 0 by normalf
		hold(spairszr)=0, // #spairs which reduce to 0 by lbasis
		hold(spairszl)=0  // #spairs which were predicted to
				  // reduce to 0 due to non-lcm trick
            ); // Some statistics

            userinfo(5, "Looking for S-pairs by iterating through all ".
                     counter. " initial monomials");
            lbasis:=[];
            for m in initials do
                userinfo(hold(NoNL), 5, ".");
                divisors:=dom::criticalDivisors(m);
                if nops(divisors)>=2 then
                    // construct a basis of the S-Pairs anhilating m
                    userinfo(hold(NoNL), 5, "".counter.":");
                    stat[hold(spairszl)]:=stat[hold(spairszl)]+
                        nops(dom::smallPrimeDivisors(m))-nops(divisors);
                    m1:=divisors[1];
                    for i from 2 to nops(divisors) do
                        m2:=divisors[i];
                        userinfo(hold(NoNL), 5, "S");
                        spair:=dom::SPair(m1,m2,m);

                        userinfo(hold(NoNL), 5, "n");
                        spair:=dom::normalfModNonPrimes(spair);
                        if iszero(spair) then
                            userinfo(hold(NoNL), 5, "*");
                            stat[hold(spairszg)]:=stat[hold(spairszg)]+1;
                        else
                            userinfo(hold(NoNL), 5, "r");
                            spair:=dom::GaussNormalf(spair,lbasis);
                            if iszero(spair) then
                                stat[hold(spairszr)]:=stat[hold(spairszr)]+1;
                                userinfo(hold(NoNL), 5, "*");
                            else
                                stat[hold(spairs)]:=stat[hold(spairs)]+1;
                                userinfo(hold(NoNL), 5, "+");
                                lbasis:=dom::GaussInsert(spair,lbasis);
                            end_if
                        end_if;
                    end_for
                end_if;
                counter:=counter-1;
            end_for;
            userinfo(hold(NoNL), 5, "\n");

            // The minimal generating system is composed of the prime
            // monomials that are not leading terms in lbasis.
            mgs :=
            [dom::monomial(R::one, p)
             $ p in select(dom::primes(d),
                           not _in,
                           {op(map(lbasis, dom::lsupport))}) ];

            // Data structure updates

            if nops(mgs)>0 then
                dom::decomposablesModNonPrimesBasis(d):=lbasis;
            else
                dom::decomposablesModNonPrimesBasis(d):=hold(allPrimes);
            end_if;

            userinfo(3, "  S-Pairs: ". stat[hold(spairs)]. " (+ ".
                     stat[hold(spairszg)]. " normalf->0 + ".
                     stat[hold(spairszr)]. " reduction->0 + ".
                     stat[hold(spairszl)]. " non-lcm->0)");
        else
	    
            // Main loop, with pure linear algebra " la Faugre 4"

            [M, skipped]:=dom::decomposableMatrix(d, hold(Skip));
            dim := nops(initials);
	    [rows,cols]:=linalg::matdim(M);
            userinfo(5, "Decomposable matrix of size: ". rows."x".cols.", actually ".rows."x".(dim-nops(skipped)));
            userinfo(5, "  Generation of ".nops(skipped)." rows/cols avoided.");

	    if dom::useLinBox then
		userinfo(5, "Computing the rank with LinBox");
		infile :="dom-matrix".d.".txt2";
		outfile:="dom-matrix".d.".rank";
		M::dom::write_sparse2(M, infile);
//		system("linbox ". infile." 10 10 -4 4 5 1 1 >". outfile);
		system("linbox ". infile." 10 10 -4 4 65521 1 1 >". outfile);
		rank := read(outfile, Quiet);
		userinfo(5, "  Got:".rank);
		// The size of the minimal generating set, as
		// predicted by linbox. This should always be an
		// upper bound
		predictedmgs := dim - (rank+nops(skipped));
		if predictedmgs = 0 then
		    userinfo(5, "  No generators at degree ".d);
		    return([]);
		else
		    userinfo(5, "  There should be ".predictedmgs." generators at degree ".d);
		end_if;
	    end_if;
            userinfo(5, "Applying Gauss Elimination");
            // M::dom::dump(M, "matrix-data".d);
	    // Flip the matrix vertically so that gaussElim returns
	    // an upper-triangular matrix
	    M := extsubsop(M, 1=max(extop(M,1), dim+1), 3=map(extop(M,3), p->poly(map(poly2list(p), m->[m[1],dim+1-m[2]]), extop(p,2..3))));
            [M, rank, det, char_row] := M::dom::gaussElim(M, ColumnElimination);
	    // Flip back char-row
	    char_row := map(char_row, c->dim+1-c);
	    char_row := revert(sort([op(char_row)]));

	    //print(M);
	    //print(dom::fromSparseVector(M::dom::col(M,i))) $i=1..rank;
            // M::dom::dump(M, "matrix-result".d);
	    /*if M::dom::lcolumns = FAIL then
	    end_if;
            if dom::useALP then
                char_row:=M::dom::lcolumns(M);
            else
                char_row:=
                map(M::dom::rows(M),
                    l->(if nops(l)>0 then op(l,[1,1]) else NIL; end_if));
            end_if;*/
	    //char_row:=map(char_row, c->dim-c);

            userinfo(5, "Extracting initial monomials that are not hit");

            // Put in mgs the initial monomials that are neither
            // referenced in skipped nor in char_row
            mgs:=[];
            i := 1;
            r := nops(char_row);
            s := nops(skipped);
            while(r>0 or s>0) do
                if r=0 or (s>0 and skipped[s] < char_row[r]) then
                    j:=skipped[s];
                    s:=s-1;
                else
                    j:=char_row[r];
                    r:=r-1;
                end_if;
                mgs := append(mgs, initials[k] $ k=i..j-1);
                i := j+1;
            end_while;
            mgs := append(mgs, initials[k] $ k=i..dim);
	    assert(dim = nops(skipped) + nops(char_row) + nops(mgs));
            userinfo(0, "Dim: ". dim . " Skipped: ". nops(skipped) . "
            Char Col: ". nops(char_row) . " Mgs: " . nops(mgs));
	    if dom::useLinBox then
	        if nops(mgs) > predictedmgs then
		    error("Bad prediction of LinBox!");
	        elif nops(mgs) < predictedmgs then
		    userinfo(1, "Got ".nops(mgs)."generators, while Linbox predicted ".predictedmgs."!");
		end_if;
	    end_if;
        end_if;

	// User information and structure update

        if nops(mgs)>0 then
            dom::degreeLowerBound():=max(dom::degreeLowerBound(), d);
        end_if;
	userinfo(3, "".d."\t".(time()-t)."\t".op([bytes()],2).
		 "\t#Stats: degree,time(ms),mem(bytes)");
	userinfo(1, "Statistics for degree ".d.":");
	userinfo(hold(NoNL), 3, (system("echo '	 Date: '`date`");""));
	userinfo(3, "  Time: ".experimental::time::format(time()-t));
	userinfo(3, "  Memory: ". (op([bytes()],2) div 1000)." ko.");
	userinfo(1, "  Minimal generators: ". nops(mgs));
	if nops(mgs)>0 then
	    userinfo(3,"  --------------------------------------");
	    for s in mgs do
		userinfo(3, "	 ". expr2text(s));
	    end_for;
	    userinfo(3,"  --------------------------------------");
	end_if;
	mgs:=map(mgs, dom);
    end_proc;

    // normalfModDecomposables(p: dom) -> dom
    //
    // Preconditions: p is homogeneous of degree d.
    //
    // Compute a normal form of p modulo the subspace of homogeneous
    // decomposable invariants of degree d. The result is a linear
    // combination of elements of dom::minimalGeneratingSet(d).
    //
    // Note: requires the computation of minimalGeneratingSet(d).

    normalfModDecomposables :=
    proc(p: dom)
    begin
	dom::GaussNormalf(dom::normalfModNonPrimes(p),
			   dom::decomposablesModNonPrimesBasis(degree(p)));
    end_proc;

    // isDecomposable(p: dom) -> DOM_BOOL
    //
    // Preconditions: p is homogeneous of degree d.
    //
    // Returns true iff p is decomposable, i.e. can be expressed using
    // sums and products of invariants of degree <d.
    //
    // Note: requires the computation of minimalGeneratingSet(d).

    isDecomposable :=
    proc(p: dom)
    begin
	return(iszero(dom::normalfModDecomposables(p)));
    end_proc;

    // isGeneratingSet(gs: dom::TypeThisList) -> DOM_BOOL
    //
    // Precondition: gs is a set of homogeneous invariants sorted by degree
    // in the same format as the result of dom::minimalGeneratingSet.
    //
    // Returns true if and only if gs is a generating set.
    //
    // Note: requires the computation of minimalGeneratingSet().
    
    isGeneratingSet :=
    proc(gs: dom::TypeThisList)
	local d, gen;
    begin
	for d from 1 to dom::degreeBound() do
	    if contains(gs, d) then
		gen:=gs[d];
	    else
		gen:=[];
	    end_if;
	    if not dom::isGeneratingSetDegree(gen,d) then
		return(FALSE);
	    end_if;
	end_for;
	return(TRUE);
    end_proc;

    // isGeneratingSetDegree(gen: Type::ListOf(dom), d: Type::NonNegInt)
    //	-> DOM_BOOL
    //
    // Precondition: gen is a list of homogeneous invariants of degree d.
    //
    // Returns true iff gen is a partial generating set for degree d.
    // Otherwise said, iff gen together with the decomposable
    // invariants generate as a vector space all the invariants of degree d.
    //
    // Note: requires the computation of minimalGeneratingSet(d).
    
    isGeneratingSetDegree :=
    proc(gen: Type::ListOf(dom), d: Type::NonNegInt)
	local lbasis, dim, g, ngen;
    begin
	dim:=nops(dom::minimalGeneratingSet(d));
	if dim=0 then
	    return(TRUE);
	end_if;
	ngen:=nops(gen); // number of remaining generators;
	lbasis:=[];
	for g in gen do
	    if ngen<dim then
		return(FALSE)
	    end_if;
	    g:=dom::GaussNormalf(dom::normalfModDecomposables(g),
				  lbasis);
	    if iszero(g) then
		userinfo(hold(NoNL), 5, "*");
	    else
		dim:=dim-1;
		if dim=0 then
		    return(TRUE);
		end_if;
		lbasis:=dom::GaussInsert(g, lbasis);
	    end_if;
	    ngen:=ngen-1;
	end_for;
	return(FALSE);
    end_proc;

    /////////////////////////////////////////////////////////////////////////
    // Computation of Hironaka Decompositions
    /////////////////////////////////////////////////////////////////////////
    
    // secondaryInvariants([options]) -> dom::TypeThisList
    //
    // Computes a secondary invariants for the invariant ring.
    //
    // They are returned as a table t, where t[d] is the list of all
    // the secondary invariants of degre d, or is unassigned if there
    // is none.
    //
    // Options:
    //	 - useHilbertSeries: DOM_BOOL
    //	   If set to false, secondaryInvariants won't use the Hilbert
    //	   Series as a stopping condition. This is mostly usefull
    //	   for debugging, and double checking the computations.
    //	   Default value: TRUE.
    //	 - maxDegree: Type::NonNegInt
    //	   only runs the computation up to degree maxDegree
    //
    // Side effects:
    //	- updates dom::degreeBound() and dom::degreeLowerBound()
    //	- sets dom::irreducibleSecondaryInvariants

    // irreducibleSecondaryInvariants([options]):
    // 
    // Same as secondaryInvariants, but only returns the irreducible
    // secondary invariants. Also, By default the computation is only
    // run up to dom::degreeBound(), since there is no irreducible
    // secondary invariants of higher degree.

    // secondaryInvariantsInit([options])
    // secondaryInvariants(d)		 -> Type::ListOf(dom)
    // irreducibleSecondaryInvariants(d) -> Type::ListOf(dom)
    //
    // If only the (irreducible) secondary invariants of some degree d
    // are needed, one can call directly secondaryInvariants(d) or
    // irreducibleSecondaryInvariants(d).
    // secondaryInvariantsInit([options]) should be called beforehand.

    secondaryInvariants :=
    proc()
	local d, t, l, secondaries;
	option remember;
    begin
	
	// Calls with just one argument
	
	if args(0)=1 and not(testtype(args(1), Type::Option)) then
	    if domtype(args(1))=DOM_INT then
		return(dom::secondaryInvariantsDegree(args(1)));
	    elif dom::chainProduct then
		return(dom::secondaryInvariantsShape(args(1)));
	    end_if;
	end_if;

	// Calls with no arguments, or just options
	
	experimental::computationHeader(
	    "Computation of (irreducible) secondary invariants.",
	    "Algorithm: linear algebra".
	    (if dom::chainProduct then " + chain product." else "."; end_if));
	t:=time();
	secondaries:=table();
	dom::secondaryInvariantsInit(args());

	// Main loop

	for d from 0 to dom::secondaryInvariantsOptions[hold(maxDegree)] do
	    l:=dom::secondaryInvariants(d);
	    if nops(l)>0 then
		secondaries[d]:=l;
	    end_if;
	end_for;

	// Update of the degree bound

	if d >= dom::degreeBound() then
	    for d from dom::degreeBound() downto 0 do
		if dom::irreducibleSecondaryInvariants(d)<>table() then
		    break;
		end_if
	    end_for;
	    d:=max(d,op(dom::primaryInvariantsDegrees()));
	    if d<dom::degreeBound() then
		userinfo(3, "Updating degreeBound to ".d);
		dom::degreeBound()	   := d;
	    end_if;
	end_if;

	// User information

	userinfo(3, "Total time: ".experimental::time::format(time()-t));

	return(secondaries);
    end_proc;

    irreducibleSecondaryInvariants :=
    proc()
	local d, l, result, options;
	option remember;
    begin
	
	// Calls with just one integer or shape argument
	
	if args(0)=1 and not(testtype(args(1), Type::Option)) then
	    if domtype(args(1))=DOM_INT and args(1)>dom::degreeBound() then
		return([]);
	    end_if;
	    dom::secondaryInvariants(args(1));
	    return(dom::irreducibleSecondaryInvariants(args(1)));
	end_if;

	options:=table(args());
	if not contains(options, hold(maxDegree)) then
	    options[hold(maxDegree)]:=dom::degreeBound();
	end_if;

	// All the job is actually done by dom::secondaryInvariants

	dom::secondaryInvariants(op(options));
	
	// Build the result table

	for d from 0 to options[hold(maxDegree)] do
	    l:=dom::irreducibleSecondaryInvariants(d);
	    if nops(l)>0 then
		result[d]:=l;
	    end_if;
	end_for;

	return(result);
    end_proc;

    secondaryInvariantsInit :=
    proc()
    begin
	dom::secondaryInvariantsOptions :=
	    combinat::getOptions(dom::secondaryInvariantsDefaultOptions(),
				 0, Error);
	null():
    end_proc;

    secondaryInvariantsDefaultOptions :=
    proc()
    begin
        table(hold(useHilbertSeries)     = TRUE,
	      hold(maxDegree)	     = dom::secondaryInvariantsMaxDegree()
             );
    end_proc;

    // set the irreducible secondary invariants
    // Side effects:
    //	- Update the lower degree bound

    setIrreducibleSecondaryInvariants :=
    proc(d: Type::Union(Type::NonNegInt, dom::TypeShapeSet),
	 irreducibles: dom::TypeThisList)
    begin
	dom::irreducibleSecondaryInvariants(d):=irreducibles;
	if domtype(d)<>DOM_INT then
	    d:=_plus(op(d));
	end_if;
	if nops(irreducibles)>0 then
	    dom::degreeLowerBound():=max(dom::degreeLowerBound(),d);
	end_if;
    end_proc;

    // Display statistics for the user

    secondaryInvariantsShowStats :=
    proc(d: Type::Union(Type::NonNegInt, dom::TypeShapeSet),
	 t: Type::NonNegInt,
	 secondaries:  dom::TypeThisList,
	 irreducibles: dom::TypeThisList)
	local irr, nirr, nsec;
    begin
	if domtype(d)=DOM_INT then
	    userinfo(3, "".d."\t".(time()-t)."\t".op([bytes()],2).
		     "\t#Stats: degree,time(ms),mem(bytes)");
	    userinfo(1, "Statistics for degree ".d.":");
	else
	    userinfo(1, "Statistics for shape ".expr2text(d).":");
	end_if;
	if domtype(secondaries)=DOM_TABLE then
	    nsec:=_plus(op(map([op(secondaries)], nops@op, 2)));
	else
	    nsec:=nops(secondaries);
	end_if;
	if domtype(irreducible)=DOM_TABLE then
	    nirr:=_plus(op(map([op(irreducibles)], nops@op, 2)));
	else
	    nirr:=nops(secondaries);
	end_if;

	if dom::isCohenMacaulay=TRUE and
	   nsec<>dom::numberOfSecondaryInvariants(d) then
	    userinfo(0,"WARNING: inconsistent number of secondary invariants found!");
	    userinfo(0,"Is the invariant ring really Cohen-Macaulay?");
	end_if;
	userinfo(hold(NoNL), 3, (system("echo '	 Date: '`date`");""));
	userinfo(3, "  Time: ".experimental::time::format(t));
	userinfo(3, "  Memory: ". (op([bytes()],2) div 1000)." ko.");
	userinfo(1, "  Secondary invariants: ". nsec);
	userinfo(1, "  Irreducible secondary invariants: ". nirr);
	if nops(irreducibles)>0 then
	    userinfo(3,"  --------------------------------------");
	    for irr in irreducibles do
		userinfo(3, "	 ". expr2text(irr));
	    end_for;
	    userinfo(3,"  --------------------------------------");
	end_if;
    end_proc;

    // secondaryInvariantsDegree(d) -> dom::TypeThisList
    //
    // Private subroutine that actually does the computation of the
    // secondary invariants at degree d.
    //
    // Side effects:
    //	- updates the irreducible secondary invariants
    //	- updates the degree lower bound
    
    secondaryInvariantsDegree :=
    if dom::chainProduct then
	dom::secondaryInvariantsDegree_ChainProduct
    else
	dom::secondaryInvariantsDegree_LinearAlgebra
    end_if;

    secondaryInvariantsDegree_LinearAlgebra :=
    proc(d: Type::NonNegInt)
	local irreducibles, secondaries, lbasis, t, dim, p, charRows, i;
    begin
	if dom::secondaryInvariantsOptions=FAIL then
	    error("Please initialize the computation with dom::secondaryInvariantsInit()");
	end_if;

	t:=time();

	if d=0 then
	    secondaries :=[dom::one];
	    irreducibles:=[];
	elif (dim:=dom::numberOfSecondaryInvariants(d);
	      dim=0 and dom::isCohenMacaulay=TRUE) then
	    userinfo(1, "Degree ".d. ": no secondary invariants");
	    secondaries :=[];
	    irreducibles:=[];
	else
	    userinfo(1, "Constructing ".dim.
		     " secondary invariants of degree ".d);
	    // Compute a basis of the subspace of decomposable invariants
	    ([lbasis, charRows, secondaries]) :=
		dom::decomposablesBasis_LinearAlgebra(d);

	    // Construct a set of irreducible invariants from the
	    // leading monomials that are not leading terms in lbasis.
	    for p in dom::basisIndices(d) do
		// Just makes sure that the ranker has been
		// initialized properly
		dom::rankBasis(p);
	    end_for;
	    irreducibles:=
		[dom::term(dom::unrankBasis(i))
		 $ i in { $ 1..dom::basisIndices::count(d) } minus charRows];
	    secondaries:=secondaries.irreducibles;
	end_if;

	dom::setIrreducibleSecondaryInvariants(d, irreducibles);
	dom::secondaryInvariantsShowStats(d, time()-t,
					   secondaries, irreducibles);
	secondaries;
    end_proc;

    // idealBasis_LinearAlgebra(d: Type::NonNegInt)
    //
    // Subroutine to compute a column-echelon vector space basis of the
    // homogeneous component of degree d of the ideal generated by the
    // invariants of degree <d, as a matrix of sparse column vectors.

    idealBasis_LinearAlgebra :=
    proc(d: Type::NonNegInt)
	: Type::ListProduct(dom::sparseMatricesDomain,
			    Type::SetOf(Type::PosInt))
    /*option remember;*/
	local lbasis, primaries, primariesDegs,
	p, pr, s, d2, degs, t,
	c, charRows, det, rank;
    begin
	t:=time();
	userinfo(3, "Computing the ideal generated by ".
		 "the primary invariants at the degree ".d);

	// primaries:	  primary invariants
	// primariesDegs: degrees of the primary invariants.
	primaries    :=dom::primaryInvariants();
	primariesDegs:=map(primaries, degree);

	// matrix whose column vectors will span the ideal at degree d
	lbasis := dom::sparseMatricesDomain(0,0);

	for d2 from 0 to d-1 do
	    if nops(dom::secondaryInvariants(d2))=0 then next; end_if;
	    userinfo(3, "Secondary invariants of degree ".d2.": ".
		    (combinat::integerVectorsWeighted::count(d-d2, primariesDegs)
		     *nops(dom::secondaryInvariants(d2))) .
		    " products.");
	    for degs in combinat::integerVectorsWeighted::list(d-d2, primariesDegs) do
		userinfo(hold(NoNL), 5, "P");
		// compute _mult(primaries[i]^(degs[i]), i)
		// That's way suboptimal!!! products gets recomputed
		// over and over
		p:=_mult(op(select(zip(primaries, degs, _power),
				   _unequal, dom::one)));
		for s in dom::secondaryInvariants(d2) do
		    userinfo(hold(NoNL), 5,"p");
		    if d2=0 then  // s is dom::one; save a multiplication
			pr:=p;
		    else
			pr:=dom::mult2(p,s);
		    end_if;
		    lbasis := lbasis.dom::sparseVector(pr);
		end_for;
	    end_for;
	    userinfo(hold(NoNL), 5, "\n");
	end_for;
	userinfo(3, "Running gauss elimination");
	[lbasis, rank, det, charRows] :=
	lbasis::dom::gaussElim(lbasis, ColumnElimination);
	if rank < linalg::ncols(lbasis) then
	    userinfo(0, if dom::isCohenMacaulay then "WARNING: " else ""; end.
		     "the invariant ring is not a free module over the primary invariants:");
	    dom::isCohenMacaulay:=FALSE;
	    for c from linalg::ncols(lbasis) downto rank + 1 do
		lbasis := lbasis::dom::delCol(lbasis, c);
	    end_for;
	end_if;
	userinfo(3, "Done in ".experimental::time::format( time()-t ));
	[lbasis, charRows];
    end_proc;

    // decomposablesBasis_LinearAlgebra(d: Type::NonNegInt)
    //	-> [ ... ]
    //
    // An invariant of degree d is called decomposable if it can be
    // expressed using sums and products of invariants of degree <d.
    //
    // This subroutine computes a row-reduced basis of the vector
    // space of homogeneous decomposable invariants of degree d. As a
    // by-product, it also returns invariants which, together with the
    // irreducible secondary invariants, will form secondary
    // invariants of degree d for the invariant ring.

    decomposablesBasis_LinearAlgebra :=
    proc(d: Type::NonNegInt)
	: Type::ListProduct(dom::sparseMatricesDomain,
			    Type::SetOf(Type::PosInt),
			    Type::ListOf(dom))
	local secondaries, lbasis, charRows, t, d2, p, q, s, useHS, dim,
	      det, lbasisnew, rank, totaldim;
    begin
	[lbasis, charRows] := dom::idealBasis_LinearAlgebra(d);
	userinfo(3, "Generating secondaries by products ".
		 "of previous secondaries. ");
	t:=time();
	secondaries:=[];
	dim := linalg::ncols(lbasis);
	useHS := dom::secondaryInvariantsOptions[hold(useHilbertSeries)];
	if useHS then
	    totaldim := dom::basisIndices::count(d);
	end_if;
	for d2 from 1 to d-1 do
	    for q in dom::irreducibleSecondaryInvariants(d2) do
		for s in dom::secondaryInvariants(d-d2) do
		    if useHS and dim=totaldim then
			userinfo(hold(NoNL), 5, "\n");
			userinfo(3, "Done in ".experimental::time::format(time()-t));
			return([lbasis, secondaries, charRows]);
		    end_if;
		    userinfo(hold(NoNL), 5, "p");
		    p:=dom::mult2(s,q);
		    userinfo(hold(NoNL), 5, "r");
		    lbasis := lbasis.dom::sparseVector(p);
		    // Reduces the new column with respect to the previous
		    // ones. Doing this by running gauss elimination on the
		    // full matrix wastes a little bit of time; but that's
		    // probably not so much (no extra linear
		    // combination of columns is done).
		    // The copy should not be too much of a bother either, since
		    // all the columns but the last one should be shared
		    // in memory between them.
		    [lbasisnew, rank, det, charRows] :=
		    lbasis::dom::gaussElim(lbasis, ColumnElimination);
		    if rank=dim then
			userinfo(hold(NoNL), 5, "*");
		    else
			lbasis := lbasisnew;
 			dim := rank; // = dim + 1
			userinfo(hold(NoNL), 5, "+[".dim."]");
			secondaries:=append(secondaries,p);
		    end_if;
		end_for;
	    end_for;
	end_for;
	userinfo(hold(NoNL), 5, "\n");
	userinfo(3, "Done in ".experimental::time::format(time()-t));
	[lbasis, charRows, secondaries];
    end_proc;
    
    //////////////////////////////////////////////////////////////////////////
    // Implementation of secondary invariants with the chain product
    //////////////////////////////////////////////////////////////////////////
    
    secondaryInvariantsDegree_ChainProduct :=
    proc(d: Type::NonNegInt)
	local t, l, shape, secondaries, irreducibles;
    begin
	t:=time();
	secondaries  := table();
	irreducibles := table();
	for shape in dom::shapesByDegree(d) do
	    l:=dom::secondaryInvariants(shape);
	    if nops(l)>0 then
		secondaries[shape]:=l;
	    end_if;
	    l:=dom::irreducibleSecondaryInvariants(shape);
	    if nops(l)>0 then
		irreducibles[shape]:=l;
	    end_if;
	end_for;

	dom::setIrreducibleSecondaryInvariants(d, irreducibles);
	dom::secondaryInvariantsShowStats(d, time()-t,
					   secondaries, irreducibles);
	secondaries;
    end_proc;

    secondaryInvariantsShape :=
    proc(shape: Type::SetOf(Type::NonNegInt))
	local lbasis, dim, cshape, t, irreducibles, secondaries, p;
    begin
	if dom::secondaryInvariantsOptions=FAIL then
	    error("Please initialize the computation with dom::secondaryInvariantsInit()");
	end_if;

	t:=time();

	if shape={} then
	    secondaries :=[dom::one];
	    irreducibles:=[];
	elif (dim:=dom::numberOfSecondaryInvariants(shape);
	      dim=0 and dom::isCohenMacaulay=TRUE) then
	    userinfo(1, "Shape ".expr2text(shape).
		     ": no secondary invariants");
	    secondaries :=[];
	    irreducibles:=[];
	elif (cshape:=dom::shapeComplement(shape);
	      contains(op(dom::secondaryInvariants,5), cshape)) then
	    // If the computations has already been done for the
	    // complement shape, reuse its result by complementation.
	    // We use a hugly (non portable?) hack to check in the
	    // remember table of dom::secondaryInvariants
	    userinfo(1, "Shape ".expr2text(shape).
		     ": Reusing the results of shape ".expr2text(cshape).
		     "by complementation");
	    secondaries :=map(dom::secondaryInvariants(cshape),
			      dom::complement);
	    irreducibles:=map(dom::irreducibleSecondaryInvariants(cshape),
			      dom::complement);
	else
	    userinfo(1, "Constructing ".dim.
		     " secondary invariants of shape ".expr2text(shape).
		     " (degree "._plus(op(shape)).")");
	    ([lbasis, secondaries]) :=
		dom::decomposablesBasis_ChainProduct(shape);

	    // Construct a set of irreducible invariants from
	    // leading monomials that are not leading terms in lbasis.
	    irreducibles:=
		[dom::monomial(R::one, p)
		 $ p in select(dom::basisIndices(dom::shapeSet2List(shape)),
			       not _in,
			       {op(map(lbasis, dom::lsupport))})];
	    secondaries:=secondaries.irreducibles;
	end_if;

	dom::setIrreducibleSecondaryInvariants(shape, irreducibles);
	dom::secondaryInvariantsShowStats(shape, time()-t,
					   secondaries, irreducibles);
	secondaries;
    end_proc;


    // idealBasis_ChainProduct(shape: Type::SetOf(Type::NonNegInt))
    //
    // Subroutine to compute a row-reduced vector space basis of the
    // homogeneous component of shape shape of the ideal generated by the
    // invariants of shape < shape.

    idealBasis_ChainProduct :=
    proc(shape: Type::SetOf(Type::NonNegInt))
    /*option remember;*/
     local p, pr, s, shape2, t, lbasis;
    begin
	t:=time();
	userinfo(3, "Computing the ideal generated by ".
		 "primaries for the shape ".expr2text(shape));
	lbasis:=[];
	for shape2 in combinat::subsets::list(shape) do
	    if shape2=shape
		then next;
	    end_if;
	    if nops(dom::secondaryInvariants(shape2))=0 then next; end_if;
	    userinfo(10, "Expanding ", nops(dom::secondaryInvariants(shape2)),
		     "secondaries of shape ".expr2text(shape2));
	
	    userinfo(5..9, hold(NoNL),"E".expr2text(shape2));
	    p:=dom::symmetricMonomial(dom::shapeSet2List(shape minus shape2));
	    for s in dom::secondaryInvariants(shape2) do
		userinfo(hold(NoNL), 5,"p");
		if shape={} then
		    pr:=p;	// s is dom::one; save a multiplication
		else
		    pr:=dom::mult2(p,s);
		end_if;
		userinfo(hold(NoNL), 5, "r");
		pr:=dom::GaussNormalf(pr, lbasis);
		if iszero(pr) then
		    userinfo(0, "WARNING: The invariant ring is not a free module over the primary invariants!");
		    userinfo(hold(NoNL), 5, "*");
		else
		    userinfo(hold(NoNL), 5, "+");
		    lbasis:=dom::GaussInsert(pr, lbasis);
		end_if;
	    end_for;
	end_for;
	userinfo(hold(NoNL), 5, "\n");
	userinfo(3, "Done in ".experimental::time::format( time()-t ));
	lbasis;
    end_proc;

    // decomposablesBasis_ChainProduct(shape: dom::TypeShapeSet)
    //	-> [lbasis: Type::ListOf(dom), secondaries: Type::ListOf(dom)]
    //
    // An invariant of degree d is called decomposable if it can be
    // expressed with sums and products of invariants of degree <d).
    //
    // This subroutine computes a row-reduced basis of the vector
    // space of homogeneous decomposable invariants of shape shape. As
    // a by-product, it also returns invariants which, together with
    // the irreducible secondary invariants, will form secondary
    // invariants of shape shape for the invariant ring.


    decomposablesBasis_ChainProduct :=
    proc(shape: dom::TypeShapeSet)
	local lbasis, secondaries, t, shape2, pr, p, q, s, useHS, dim;
    begin
	lbasis:=dom::idealBasis_ChainProduct(shape);
	userinfo(3, "Generating secondaries by products of previous secondaries. ");
	t:=time();
	useHS:=dom::secondaryInvariantsOptions[hold(useHilbertSeries)];
	if useHS then
	    dim:=dom::basisIndices::count(dom::shapeSet2List(shape))
		 -nops(lbasis);
	end_if;
	secondaries:=[];
	for shape2 in combinat::subsets::list(shape) do
	    if shape2=shape or shape2={} then next; end_if;
	    for q in dom::irreducibleSecondaryInvariants(shape2) do
		for s in dom::secondaryInvariants(shape minus shape2) do
		    userinfo(hold(NoNL), 5, "p");
		    p:=dom::mult2(s,q);
		    userinfo(hold(NoNL), 5, "r");
		    pr:=dom::GaussNormalf(p, lbasis);
		    if iszero(pr) then
			userinfo(hold(NoNL), 5, "*");
		    else
			secondaries:=append(secondaries,p);
			lbasis:=dom::GaussInsert(pr, lbasis);
			if useHS then
			    dim:=dim-1;
			    userinfo(hold(NoNL), 5, "+[".dim."]");
			    if dim=0 then
				userinfo(hold(NoNL), 5, "\n");
				userinfo(3, "Done in ".experimental::time::format(time()-t));
				return([lbasis, secondaries]);
			    end_if;
			else
			    userinfo(hold(NoNL), 5, "+");
			end_if;
		    end_if;
		end_for;
	    end_for;
	end_for;
	userinfo(hold(NoNL), 5, "\n");
	userinfo(3, "Done in ".experimental::time::format(time()-t));
	[lbasis, secondaries];
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // Routines to compute primitive invariants
    //////////////////////////////////////////////////////////////////////////

    primitiveInvariant :=
    proc(g: dom::group)
        local S, d, m;
    begin
        // Should test that g is not in the group ...
        S := dom::basisIndices;
        for d from 1 to binomial(dom::dimen,2) do
            userinfo(1, "Searching at degree ".d);
            for m in dom::basisIndices(d) do
                userinfo(hold(NoNL), 5, ".");
                if S::canonic(S::act(g,m))<>m then
                    return(dom(m));
                end_if;
            end_for;
        end_for;
        error("There should be a primitive invariant!")
    end_proc;

begin
    if not R::hasProp(Cat::Ring) then
	error("first argument must be a Cat::Ring")
    end_if;
    options := combinat::getOptions(table(hold(ChainProduct)=FALSE,
					  hold(Standard)=NIL,
					  hold(Variables)=NIL,
					  hold(CompactVariables)=TRUE,
					  hold(UseALP) = FALSE,
					  hold(UseLinBox) = FALSE
					 ),
				    table(hold(Variables) = Type::Union(DOM_NIL, Type::ListOf(DOM_IDENT)),
					  hold(ChainProduct) = DOM_BOOL,
					  hold(CompactVariables) = DOM_BOOL,
					  hold(UseALP) = DOM_BOOL,
					  hold(UseLinBox) = DOM_BOOL),
				    2, Error);

    if options[hold(Variables)] = NIL then
	options[hold(Variables)] :=
	map(G::set,
	    proc(e)
		local v,s;
	    begin
		if options[hold(CompactVariables)] then
		    s:=G::compactName(e);
		    if traperror((v:=_concat(hold(x),s)))=0 then
			return(v);
		    else userinfo(10, "Warning: could not create compact variable name. Ignoring CompactVariables option");
		    end_if;
		end_if;
		hold(x)[e];
	    end_proc);
    end_if;
end_domain:
