/*****************************************************************************
  $Id: PermutationGroupOnSets.mu,v 1.1.1.2 2003/09/19 15:11:40 nthiery Exp $
  File:		lib/DOMAINS/CATEGORY/PermutationGroup.mu
  Authors:	Nicolas M. Thiry <nthiery@users.sourceforge.net>
  Taken from:	PerMuVAR, 02/12/2001
  Created:	08/08/1999
  License:	LGPL
*****************************************************************************/

/*++
Dom::PermutationGroupOnSets - Permutation group acting on subsets of a set
Dom::PermutationGroupOnSets(G, K, [options]) -- Given the action of a
	group G on a finite set S, construct its induced action on the
	set of all the subsets of size K of S.

 - G:	Group acting on a finite set (Cat::PermutationGroup).
 - K:	Size of the subsets on which the group will act.
	The rest of the arguments will be converted in a table
	containing optional arguments

To be added: axiom (or option): oriented
	idem for loops ? depends if it's usefull.

++*/

alias(REP(x)=extop(x,1)):

domain Dom::PermutationGroupOnSets(Group: DOM_DOMAIN,
				   K: Type::Union(Type::NonNegInt,
						  Type::ListOf(Type::NonNegInt))
				  )
    local Opts: Type::TableOfIndex(DOM_IDENT),
	  Card: Type::NonNegInt;
    inherits Dom::PermutationGroup(Card, toBeDefined);
    category Cat::PermutationGroup;

    options := Opts;

    underlyingGroup := Group;

    size := dom::underlyingGroup::size;

/*--
    set

    All the subsets of the set on which the Group acts, returned as a list
    of sets.
--*/

    set :=
    proc(thedom: DOM_DOMAIN): Type::ListOf(DOM_SET)
	local k, uset;
    begin
	uset := thedom::underlyingGroup::set;
	if domtype(uset) <> DOM_LIST then
	    uset := [op(uset)];
	end_if;
	// Using subwords on a list ensure that the order is always the same
	// The revert's ensures that we get the subsets in the same order
	// as before; that's not necessarily a required feature.
	revert(map(_concat(combinat::subwords::list(revert(uset),k)
			   $ k in K), DOM_SET@op));
    end_proc(dom);

    compactName :=
    proc(e)
	option remember;
    begin
	_concat(op(sort([op(map(e,dom::underlyingGroup::compactName))])));
    end_proc;

/*--
--*/

    induced :=
    proc(p:dom::underlyingGroup)
	local pind, x;
    begin
	pind:=[ dom::invertSet[map(x, p)] $ x in dom::set ];
	new(dom, pind);
    end_proc;

    allElements :=
    proc()
    begin
	map(dom::underlyingGroup::allElements(), dom::induced);
    end_proc;

    generators := map(dom::underlyingGroup::generators, dom::induced);

    act := 
    proc(g:dom, e)
    begin
	if testargs() and not contains(dom::invertSet,e) then
		error("The 2nd argument must belong to the set on which the permutation acts");
	end_if;
	dom::set[op(extop(g,1),dom::invertSet[e])]
    end_proc;


    func_call := 
    proc(g:dom, element: Type::AnyType)
    begin
	element:=context(element);
	if testtype(element, DOM_INT) or testtype(element, DOM_LIST) then
	// There was a bug here due to dynamic variable scoping.
	// Fixed with MuPAD 2 ?
	    dom::func_call_inherits(g,element);
	else
	    dom::act(g,element);
	end_if;
    end_proc;

/*--
   cycleTypes:

   Count the elements of the group by cycle type.
   (Cf Cat::PermutationGroupModule).

   If g is the permutation induced by a permutation h, the cycle
   type of g depends only on the cycle type of h. This allow to
   compute cycleTypes for this group from cycleTypes of the
   underlying group.

   The basic algorithm is simply to build any permutation h with
   a given cycle type, to build the induced permutation g, and to
   compute the cycle type of g.

   For subsets of size 2, we can shortcut this computation, using the
   following fact:

   Two cycles of h of length p_i and p_j induce cycles in g of length:

    - p_i odd	-> (p_i-1)/2  cycles of length p_i
    - p_i even	-> p_i/2-1    cycles of length p_i + 1 cycle of length p_i/2
    - p_i, p_j	-> gcd(p_i,p_j) cycles of length lcm(p_i,p_j)

   Ref: A. Kerber, Algebraic Combinatoric Via Finite Group Action p 43
--*/

    cycleTypes :=
    if K=[2] then
	proc()	// Implementation for subsets of size 2
	    option remember;
	    local t,cycleType,ct,a,i,j;
	begin
	    userinfo(3, "cycleTypes: starting computation (optimized version for subsets of size 2)");
	    t:=table();
	    
	    for cycleType in dom::underlyingGroup::cycleTypes() do
		ct:=op(cycleType,1);
		userinfo(5, "working on cycle type", ct);
		/* compute the cycle type of an induced permutation */
		a:=[0 $ dom::card];
		for i from 1 to dom::underlyingGroup::card do
		    if ct[i]>0 then
			if i mod 2=0 then
			    a[i]  :=a[i]  +ct[i]*(i/2-1);
			    a[i/2]:=a[i/2]+ct[i];
			else
			    a[i]  :=a[i]  +ct[i]*(i-1)/2;
			end_if;
			a[i]:=a[i]+binomial(ct[i],2)*i;
			for j from i+1 to dom::underlyingGroup::card do
			    if ct[j]>0 then
				a[lcm(i,j)]:=a[lcm(i,j)]+ct[i]*ct[j]*gcd(i,j);
			    end_if;
			end_for;
		    end_if;
		end_for;
		if contains(t,a) then
		    t[a]:=t[a]+op(cycleType,2);
		else
		    t[a]:=op(cycleType,2);
		end_if;
	    end_for;
	    t;
	end_proc
    else		// Default implementation for subsets of sizes<>2
	proc()
	    option remember;
	    local t,cycleType,ct,a,i,j;
	begin
	    userinfo(3, "cycleTypes: starting computation");
	    t:=table();

	    for cycleType in dom::underlyingGroup::cycleTypes() do
		ct:=op(cycleType,1);
		userinfo(5, "working on cycle type", ct);
		/* compute the cycle type of an induced permutation */
		a:=dom::cycleType(dom::induced(combinat::cycleType2perm(ct)));
		if contains(t,a) then
		    t[a]:=t[a]+op(cycleType,2);
		else
		    t[a]:=op(cycleType,2);
		end_if;
	    end_for;
	    t;
	end_proc
    end_if;

begin
    if testargs() then
	if not Group::hasProp(Cat::PermutationGroup) then
	    error("Group must be a Cat::PermutationGroup");
	end_if;
    end_if;
    if testtype(K, DOM_INT) then
	K:=[K];
    end_if;
		      
    Opts:=table(args(3..args(0)));

    // To be updated when loops and arrows will be added.
    Card:=_plus(binomial(Group::card,k) $ k in K);
end_domain:
