/*****************************************************************************
  $Id: SemiRing.mu,v 1.1.1.2 2004/10/16 15:15:00 nthiery Exp $
  File:		lib/DOMAINS/DOMAIN/WeightedAutomata.mu
  Authors: 	Eric Laugerotte 	<Eric.Laugerotte@univ-rouen.fr>
		Houda Abbad 		<houda.abbad@lycos.com>
  Created:	2004/06/17
  License:	LGPL
*****************************************************************************/

/*++

Dom::SemiRing --  the constructor domain which creates semirings 

Methods: 
zero         	        - returns the neutral element for +
one	    	        - returns the neutral element for *
changeCoeffRing(A,T)	- converts x to an element of type T
_plus(x1,x2,...)        - returns the sum of x1+x2+...
_mult(x1,x2,...)        - returns the product of x1*x2*...
_power(x,i)	        - returns x*x*...*x (i times) for a nonzero integer i
star(x)			- returns the star of x

++*/


/* we could overload star by a::dom::star(a) or preferably a^* */

domain Dom::SemiRing()
    local Options;
    inherits Dom::BaseDomain;
    category 
	Cat::SemiRing, Cat::UseOverloading, 
	if contains(Options,hold(Categ)) then
	    Options[hold(Categ)]
	end_if;
    axiom Ax::canonicalRep,Ax::normalRep;

    interface := 
        {star};

    /////////////////////////////////////////////////////////////////////////
    // Constructors and constants
    /////////////////////////////////////////////////////////////////////////

    initDomain := 
    proc()
  	option escape;
	local DOM;
    begin
	DOM := dom;
	operators::overloaded::declareSignature(operators::plus2,
	    [dom,dom],
	    (x1,x2) -> DOM::plus(x1,x2));
	operators::overloaded::declareSignature(operators::mult2,
	    [dom,dom],
	    (x1,x2) -> DOM::mult(x1,x2));
	operators::overloaded::declareSignature(operators::mult2,
	    [dom,Dom::Integer],
	    (x1,x2) -> DOM::intmult(x2,x1));
	operators::overloaded::declareSignature(operators::mult2,
	    [Dom::Integer,dom],
	    (x1,x2)->DOM::intmult(x1,x2));
	operators::overloaded::declareSignature(operators::mult2,
	    [dom,Dom::Matrix(dom)],
	    (x,M) -> map(M,(x,y)->DOM::mult(y,x),x)) ;
      	operators::overloaded::declareSignature(operators::mult2,
	    [Dom::Matrix(dom),dom],
	    (M,x) -> map(M,(x,y)->DOM::mult(x,y),x)) ;
    end_proc;

    new :=              
    proc(x : Type::Predicate("SemiRingElement",
		e->if testtype(Options[hold(Coefficient)],DOM_LIST) then
		       return(bool(contains(Options[hold(Coefficient)],e)>0))
		   elif testtype(Options[hold(Coefficient)],DOM_SET) then
		       return(contains(Options[hold(Coefficient)],e))
		   elif not(contains(Options,hold(Prop))) then
		       return(testtype(e,Options[hold(Coefficient)]))
		   else
		       return(_lazy_and(testtype(e,Options[hold(Prop)]),
			      _lazy_or(testtype(e,Options[hold(Coefficient)]),
					testtype(e,Options[hold(Prop)]))))
		   end_if))    
    begin
      	return(new(dom,x))
    end_proc;

    changeCoeffRing :=
    proc(x : dom,
	 SR : DOM_DOMAIN)
	local y;
    begin
	if x=dom::zero then
	    return(SR::zero)
	elif x=dom::one then
	    return(SR::one)
	end_if;
	Pref::typeCheck(Always);
	y:=SR(extop(x,1));
	Pref::typeCheck(Interactive);
	return(y)
    end_proc;

    print := x -> extop(x,1);

    expr := 
    proc(x : dom)
    begin
	return(extop(x,1))
    end_proc;

    zero :=dom::new(Options[hold(Zero)]);

    one :=dom::new(Options[hold(One)]);
    
    /////////////////////////////////////////////////////////////////////////
    // Arithmetic operations
    /////////////////////////////////////////////////////////////////////////

    _plus := operators::_plus;

    plus := proc()
    begin
      return(dom::new(Options[hold(Plus)](extop(args(1),1),
					  extop(args(2),1))))
    end_proc;

    _mult := operators::_mult;

    mult :=
    proc()
    begin
      return(dom::new(Options[hold(Mult)](extop(args(1),1),
					  extop(args(2),1))))
    end_proc;

    _negate :=
	if not(contains(Options,hold(Negate))) then
	    FAIL
	else  
	    (x -> dom::new(Options[hold(Negate)](extop(x,1))))
	end_if;
		       
    _invert :=
	if not(contains(Options,hold(Invert))) then
	    FAIL
	else  
	    (x -> dom::new(Options[hold(Invert)](extop(x,1))))
	end_if;

    intmult :=
    proc()
    begin
        return(domains::repeatedSquaring(dom,"_plus")(args(2),args(1)))
    end_proc;

    _power := domains::repeatedSquaring(dom,"_mult");

    star :=
	if not(contains(Options,hold(Star))) then 
            FAIL
	else
    	    (x -> dom::new(Options[hold(Star)](extop(x,1))))
	end_if;

begin
    if not(testtype(args(0),Type::Interval([1,10],Type::PosInt))) then 
    	error("Wrong number of arguments") 
    end_if;
    Options := combinat::getOptions
	     ([hold(Coefficient), hold(Prop),
	       hold(Plus), hold(Zero), hold(Negate),
	       hold(Mult), hold(One), hold(Invert),
	       hold(Star), hold(Categ)],
	      [hold(Coefficient) = Type::Union(DOM_DOMAIN,DOM_LIST,DOM_SET),
	       hold(Prop) 	 = Type,
	       hold(Plus) 	 = Type::Function,
	       hold(Negate)      = Type::Function,
	       hold(Mult)        = Type::Function,
	       hold(Invert)      = Type::Function, 
	       hold(Star) 	 = Type::Function, 
	       hold(Categ)    	 = Category],
	      0, Error);
    if not(contains(Options,hold(Coefficient))) then 
	error("Coefficients are expected")
    end_if;
    if (testtype(Options[hold(Coefficient)],DOM_DOMAIN) and 
       ((Options[hold(Coefficient)])::hasProp(Cat::SemiRing)=FAIL)) or 
       testtype(Options[hold(Coefficient)],Type::Union(DOM_LIST,DOM_SET)) then
    	if not(contains(Options,hold(Plus))) then 
	    error("Addition is expected")
    	elif not(contains(Options,hold(Zero))) then 
	    error("Neutral element for addition is expected")
        elif not(contains(Options,hold(Mult))) then 
	    error("Multiplication is expected")
    	elif not(contains(Options,hold(One))) then 
	    error("Neutral element for multiplication is expected")
      	end_if
    else
        if not(contains(Options,hold(Plus))) then
	    if contains(Options,hold(Zero)) then
	        error("Addition is expected")
	    else
		Options[hold(Zero)]:=(Options[hold(Coefficient)])::zero
	    end_if;
	    if contains(Options,hold(Negate)) then
	        error("Addition is expected")
	    elif (Options[hold(Coefficient)])::hasProp(Cat::Ring) then 
	        Options[hold(Negate)]:=(Options[hold(Coefficient)])::_negate
	    end_if;
	    Options[hold(Plus)]:=(Options[hold(Coefficient)])::_plus
	elif not(contains(Options,hold(Zero))) then
	    error("Zero is expected")
	end_if;
        if not(contains(Options,hold(Mult))) then
	    if contains(Options,hold(One)) then
	        error("Multiplication is expected")
	    else
		Options[hold(One)]:=(Options[hold(Coefficient)])::one
	    end_if;
	    if contains(Options,hold(Invert)) then
	        error("Multiplication is expected")
	    elif (Options[hold(Coefficient)])::hasProp(Cat::SkewField) then
	       Options[hold(Invert)]:=(Options[hold(Coefficient)])::_invert
	    end_if;
	    Options[hold(Mult)]:=(Options[hold(Coefficient)])::_mult;
	end_if;
    end_if;
    if testtype(Options[hold(Coefficient)],DOM_LIST) then   
    	if contains(Options[hold(Coefficient)],Options[hold(Zero)])=0 then 
     	    error("Bad neutral element for addition")
    	end_if
    elif testtype(Options[hold(Coefficient)],DOM_SET) then 
    	if not(contains(Options[hold(Coefficient)],Options[hold(Zero)])) then
     	    error("Bad neutral element for addition")
    	end_if
    elif contains(Options,hold(Prop)) then
    	if not(testtype(Options[hold(Zero)],Options[hold(Prop)])) then
      	    error("Bad neutral element for addition")
    	end_if
    elif not(testtype(Options[hold(Zero)],Options[hold(Coefficient)])) then
    	error("Bad neutral element for addition")
    end_if;
    if testtype(Options[hold(Coefficient)],DOM_LIST) then   
    	if contains(Options[hold(Coefficient)],Options[hold(One)])=0 then 
     	    error("Bad neutral element for multiplication")
   	end_if
    elif  testtype(Options[hold(Coefficient)],DOM_SET) then 
    	if not(contains(Options[hold(Coefficient)],Options[hold(One)])) then
     	    error(" Bad neutral element for multiplication")
    	end_if
    elif contains(Options,hold(Prop)) then
    	if not(testtype(Options[hold(One)],Options[hold(Prop)])) then
            error("Bad neutral element for multiplication")
    	end_if
    elif not(testtype(Options[hold(One)],Options[hold(Coefficient)])) then
    	error("Bad neutral element for multiplication")
    end_if;
end_domain:

