/*****************************************************************************
   $Id: SparseMatrix_LinBox.mu,v 1.1.1.3 2005/01/26 13:42:22 nthiery Exp $
   File:	lib/DOMAINS/DOMAIN/Matrix_ALP.mu
   Authors:	Nicolas M. Thiry <nthiery@users.sourceforge.net>
   Created:	16/02/2001
   License:	LGPL
*****************************************************************************/

/*
smithform_gap Filename, Hom level, Tim level, Method, Mode, coeff, <numerator, denominator>

Hom level, Tim level: verbosity level

Mthode:
  3:	elimination
  4:	blackbox
 -3/-4:	just the rank

Mode:
SMITHFORM       0
INTEGRANK       1
TORSIONRANK     2  rang modulo coeff
PRIMERANK       4  rang modulo coeff=p
PRIMEPOWERRANK  5  rang modulo coeff=q=p^k
PRIMEPOWERRANK  6  rang modulo coeff=q=p^k

Numerator, denominator: probabilistic quality
*/

// stringlib::split is missing in MuPAD < 3.0.0
%if stringlib::split=FAIL then
    stringlib::split :=
    proc(s)
	local i, j, result;
    begin
	result := [];
	i := 0;
	for j from 0 to length(s) do
	    if j = length(s) or s[j] = " " then
		if j-i > 0 then
		    result := append(result, substring(s,i..j-1));
		end_if;
		i := j+1;
	    end_if;
	end_for;
	result;
    end_proc;
end_if:

// When MuPAD is called with the -R option, "system" function calls
// are delegated to the interface and are not synchronized. That
// should be fixed in future versions of MuPAD
//
// Work around: emulate system by using shell::system
// Caveats:
//  - when the command returns a non zero exit status,
//    shell::system raises an error instead of returning the status
//  - redirecting the output does not work readily with MuPAD <=3.0.0
//    17/06/2003; so we embed the call in () to run it inside a subshell.

%if version()[1]<3 then
alias(system=
      proc(cmd)
      begin
	  if traperror(shell::system("(".cmd.")")) <> 0 then
	      1;
	  else
	      0;
	  end_if;
      end_proc):
else
alias(system=shell::system); // does not fix the zero exit status bug.
end_if:

domain Dom::SparseMatrix_LinBox(R: DOM_DOMAIN)
    inherits Dom::Matrix(R);
    category Cat::Matrix(R);
    axiom NIL;

    info_str := "Dom::SparseMatrix_LinBox -- Domain for sparse matrices implemented via the LinBox C++ Library";

    linbox := "smithform_gap";

    //////////////////////////////////////////////////////////////////////////
    // Utilities
    //////////////////////////////////////////////////////////////////////////

    tmpfilename :=
    proc()
	local i;
	option escape;
    begin
	i:=0;
        proc()
	begin
	    i:=i+1;
	    return("/tmp/SparseMatrix_LinBox".i);
	end_proc;
    end_proc();

    //////////////////////////////////////////////////////////////////////////
    // Accessors
    //////////////////////////////////////////////////////////////////////////

    matdim :=
    proc(mat: dom)
    begin
	[extop(mat,1)[hold(NRows)], extop(mat,1)[hold(NCols)]];
    end_proc;

    file :=
    proc(mat: dom)
    begin
	extop(mat,1)[hold(FileName)];
    end_proc;

    //newByRow :=		Module::newByRow;

    _delete :=
    proc(mat: dom)
    begin
	system("unlink ".extop(mat,1)[hold(FileName)]);
    end_proc;

    // Access to size

    appendRow :=
    proc(mat: dom, row: Type::ListOf(Type::ListProduct(Dom::Integer, R)))
	local t, value, column;
    begin
	t := extop(mat,1);
	if not contains(t, hold(FileHandle)) then
	    t[hold(NRows)] := fopen(t[hold(FileName)], Append);
	end_if;
	t[hold(NRows)] := t[hold(NRows)] + 1;
	for value in row do
	    [column, value] := value;
	    fprint(Unquoted, t[hold(FileHandle)], 
		   t[hold(NRows)], " ", column, " ", expr(value));
	    if column>t[hold(NCols)] then
		t[hold(NCols)] := column;
	    end_if;
	end_for;
	new(dom, t);
    end_proc;

    // Linear algebra operations

    rank :=
    // compute the rank of the matrix with linbox, using blackbox method
    proc(mat: dom, char=R::characteristic) : Type::NonNegInt
	local tmpfilename, command;
    begin
	tmpfilename := dom::tmpfilename();
	command := dom::linbox. " ".extop(mat, 1)[hold(FileName)].
	       " 10 10 -4 4 ".expr2text(char)." 1 1 >". tmpfilename;
	userinfo(3, "Calling: ".command);
	system(command);
	read(tmpfilename, Quiet);
    end_proc;

    rankGauss :=
    proc(mat: dom, char=R::characteristic) : Type::NonNegInt
	local tmpfilename, command, fd, rank;
    begin
	tmpfilename := dom::tmpfilename();
	command := dom::linbox. " ".extop(mat, 1)[hold(FileName)].
	       " 10 10 -3 6 ".expr2text(char)." 1 1 >". tmpfilename;
	userinfo(3, "Calling: ".command);
	system(command);
	fd:=fopen(tmpfilename);
	ftextinput(fd);
	rank := ftextinput(fd);
	fclose(fd);
	text2expr(rank);
    end_proc;

    characteristicColumns :=
    proc(mat: dom, char=R::characteristic)
	local tmpfilename, command, fd, cols;
    begin
	tmpfilename := dom::tmpfilename();
	command := dom::linbox. " ".extop(mat, 1)[hold(FileName)].
	       " 10 10 -3 6 ".expr2text(char)." 1 1 >". tmpfilename;
	userinfo(3, "Calling: ".command);
	system(command);
	fd:=fopen(tmpfilename);
	cols := ftextinput(fd);
	fclose(fd);
	map(stringlib::split(cols), text2expr);
    end_proc;

//     gaussElim :=
//     proc(e: dom)
//     begin
//         dom::triang_Gauss(e);
//         return([e, dom::rank(e), FAIL, {op(dom::lcolumns(e))}]);
//     end_proc;

    //write_sparse1 :=		Module::write_sparse1;
    //write_sparse2 :=		Module::write_sparse2;
    //write :=		Module::write;
    //write_ps :=		Module::write_ps;
    //write_dense :=	Module::write_dense;

    /*************************************************************************
       
    *************************************************************************/

    new := 
    proc()
	local filename;
	local fd;
    begin
        if args(0)=0 then
	    filename := "/tmp/matrix";
	    if (fd := fopen(Text, filename, Write)) = FAIL then
		error("could not open ".filename);
	    end_if;
	    fprint(Unquoted, fd, "0 0 M");
	    new(dom, table(hold(FileHandle)=fd,
			   hold(FileName)=filename,
			   hold(NRows)=0,
			   hold(NCols)=0)):
        else
	    FAIL
            //return(dom::convert(Dom::Matrix(R)(args(1..args(0)))));
        end_if;
    end_proc;

    fix := 
    proc(m: dom)
	local t, fd;
    begin
	t := extop(m,1);
	if contains(t, hold(FileHandle)) then
	    fd := t[hold(FileHandle)];
	    fprint(Unquoted, fd, "0 0 0");
	    fclose(fd);
	    delete(t[hold(FileHandle)]);
	end_if;
	new(dom, t);
    end_proc;

/*--
   convert -- convert an expression to a matrix 
--*/

    //convert :=

/*--
    convert_to  --  convert matrices 
--*/
    //convert_to :=

    // lapack<double>

    print :=
    proc(mat: dom)
        local n,m;
    begin
        [n,m]:=dom::matdim(mat);
	return("<LinBox matrix, file=\"".extop(mat, 1)[hold(FileName)]."\", size=".n."x".m." >");
    end_proc;

begin
    if shell=FAIL then
	warning("Dom::SparseMatrix_LinBox requires the dynamic module shell");
    end_if;
end_domain:
