/*****************************************************************************
  $Id: WeightedAutomaton.mu,v 1.1.1.4 2004/10/16 15:15:12 nthiery Exp $
  File:		lib/DOMAINS/DOMAIN/WeightedAutomata.mu
  Authors: 	Eric Laugerotte 	<Eric.Laugerotte@univ-rouen.fr>
		Houda Abbad 		<houda.abbad@lycos.com>
		Hatem Hadj Kacem	<hatem.hadj-kacem@univ-rouen.fr>
  Created:	2002/10/18
  License:	LGPL
*****************************************************************************/

/*++

Dom::WeightedAutomaton --  the constructor domain of weighted automata

Methods:
weight(A,u)          	 - returns the weight of the word u for A
proper(A)	     	 - returns the proper automaton of A
dimension(A)	     	 - returns the dimension of A
alphabet(A)	     	 - returns the alphabet of A
initial(A)	     	 - returns the initial vector of A
transition(A)        	 - returns the transitions matrices of A
transition(A,a)      	 - returns the transition matrix for the letter a of A
final(A)	     	 - returns the final vector of A
addLetter(A,l,t)     	 - adds the letter l with transtion matrix t to the 
		       	   automaton A
delLetter(A,l)		 - removes the letter from the automaton A
addState(A)		 - adds a state to the automaton A 
addState(A,it,ot,i,o)    - adds a state to the automaton A, its ingoing 
			   transitions it, its outgoing transitions with i and 
			   o as initial and final weight respectively
delState(A,s)	     	 - removes the state s from the automaton A 
subsTransition(A,t)	 - puts the weight SR::one for the transition t in the 
			   automaton A
subsTransition(A,t,w)	 - puts the weight w for the transition t in the 
			   automaton A
subsInitialWeight(A,s) 	 - puts the initial weight SR::one for the state s in 
			   the automaton A
subsInitialWeight(A,s,w) - puts the initial weight w for the state s in the
			   automaton A
subsFinalWeight(A,s)	 - puts the final weight SR::one for the state s in the
			   automaton A
subsFinalWeight(A,s,w)	 - puts the final weight w for the state s in the
			   automaton A
changeCoeffRing(A,T)  	 - converts the automaton A to an automaton over the 
		       	   semiring T
_plus(A1,A2,...)     	 - returns the addition of A1+A2+...
_mult(A1,A2,...)     	 - returns the Cauchy product of A1*A2*...
_power(A,i)	     	 - returns A*A*...*A (i times) for a positive integer i
hadamard(A1,A2)      	 - returns the Hadamard product of A1 and A2
shuffle(A1,A2)	     	 - returns the shuffle product of A1 and A2
infiltration(A1,A2)  	 - returns the infiltration product of A1 and A2
star(A)		     	 - returns the star operation of A	
complementary(A)     	 - returns the complementary of A
difference(A,B)	     	 - returns the boolean difference A-B
transpose(A)	     	 - returns the transpose automaton of A    
elimination(A,a)     	 - removes the letter a in A
isAccessible(A)	     	 - checks whether the automaton A is accessible and 
			   returns the accessible state
isAccessible(A,q)    	 - checks whether the state q is accessible for the 
		           automaton A
isAccessible(A,S)    	 - checks whether the set S is composed with 
			   accessible states of the automaton A
accessible(A)	     	 - returns the accessible automaton equivalent to A
isCoaccessible(A)    	 - checks whether the automaton A is coaccessible and 
		       	   returns the coaccessible state
isCoaccessible(A,q)  	 - checks whether the state q is coaccessible for the 
		           automaton A
isCoaccessible(A,S)  	 - checks whether the set S is composed with 
			   coaccessible states of the automaton A
coaccessible(A)	     	 - returns the coaccessible automaton equivalent to A
trim(A)		     	 - returns the trimed automaton equivalent to A
isComplete(A)	     	 - checks whether the automaton A is complete
complete(A)	     	 - returns the complete automaton of A
isDeterministic(A)   	 - checks whether the automaton A is deterministic
determinization(A)   	 - returns the deterministic automaton of A
minimization(A)	     	 - returns the minimal automaton of A
areEquivalent(A1,A2) 	 - checks whether the automata A1 and A2 are 
			   equivalent 

++*/

unalias(DIMENSION,ALPHABET,INPUT,TRANSITION,OUTPUT,TYPE_ERR):
alias(DIMENSION(x)  = extop(args(x),1),
      ALPHABET(x)   = extop(args(x),2),
      INPUT(x)	    = extop(args(x),3),
      TRANSITION(x) = extop(args(x),4),
      OUTPUT(x)	    = extop(args(x),5)
      ):

domain Dom::WeightedAutomaton(SR: DOM_DOMAIN)
    local SRM;
    inherits Dom::BaseDomain;
    category Cat::UseOverloading;

    interfaceAdd := 
	{hold(weight),hold(proper),hold(dimension),hold(alphabet),
	 hold(initial),hold(transition),hold(final),hold(changeCoeffRing),
	 hold(addLetter),hold(delLetter),hold(addState),hold(delState),
	 hold(subsTransition),hold(subsInitialWeight),hold(subsFinalWeight),
	 hold(hadamard),hold(shuffle),hold(infiltration),hold(star),
	 hold(complementary),hold(difference),hold(transpose),
	 hold(elimination),hold(isAccessible),hold(accessible),
	 hold(isCoaccessible),hold(trim),hold(coaccessible),
	 hold(isComplete),hold(complete),hold(isDeterministic),
	 hold(determinization),hold(minimization),hold(areEquivalent)};

    /////////////////////////////////////////////////////////////////////////
    // Constructors and constants
    /////////////////////////////////////////////////////////////////////////

    new :=
    proc(dim	    : Type::NonNegInt, 
	 alphabet   : DOM_LIST, 
         input	    : SRM,
	 transition : Type::TableOfEntry(Dom::Matrix(SR)),
	 output	    : SRM)
	local i,no;
    begin
	if SRM::matdim(input)<>[1,dim] then
	    error("Initial vector has incorrect dimension")
	elif SRM::matdim(output)<>[dim,1] then
	    error("Final vector has incorrect dimension")
	else
            no:=extnops(alphabet);
	    if no<>extnops([op(transition)]) then
		error("Wrong number of transition matrices")
	    end_if;
	    i:=0;
	    while (i<no) do
		i:=i+1;
		if testtype(transition[alphabet[i]],DOM_EXPR) then
		    error("Undefined transition matrices")
		elif SRM::matdim(transition[alphabet[i]])<>[dim,dim] then
		    error("Transition matrices have incorrect dimension")
		end_if
	    end_while
	end_if;
	return(new(dom,dim,alphabet,input,transition,output))
    end_proc;

    print :=
    proc()
    begin
	return(extop(args())[3],extop(extop(args())[4]),extop(args())[5])
    end;

    expr := dom::print;

    initDomain :=
    if SR<>Dom::Integer then
	proc()
            option escape;
    	begin
	    operators::overloaded::declareConversion(SR, dom,
		x->dom::new(1,[],SRM(1,1,[x]),table(),SRM(1,1,[SR::one])),
	   	Explicit);
 	    operators::overloaded::declareSignature(operators::plus2,
            	[dom,dom],
       	    	(aut1,aut2) -> dom::plus(aut1,aut2));
	    operators::overloaded::declareSignature(operators::plus2,
	    	[SR,dom],
	    	(x,aut) -> _plus(dom::new(1,[],
					  SRM(1,1,[x]),table(),
					  SRM(1,1,[SR::one])),aut));
	    operators::overloaded::declareSignature(operators::plus2,
	    	[dom,SR],
	    	(aut,x) -> dom::plus(aut,dom::new(1,[],SRM(1,1,[x]),table(),
					SRM(1,1,[SR::one]))));
	    operators::overloaded::declareSignature(operators::mult2,
            	[dom,dom],
       	    	(aut1,aut2) -> dom::mult(aut1,aut2));
	    operators::overloaded::declareSignature(operators::mult2,
	    	[SR,dom],
		proc(x,aut)
		    local a,d,i,j; 
		begin
		  a:=aut;
		  d:=dom::dimension(a);
		  i:=dom::initial(a);
		  for j from 1 to d do
		      a:=dom::subsInitialWeight(a,j,x*i[1,j])
		  end_for;
		  return(a)
		end_proc);
	    operators::overloaded::declareSignature(operators::mult2,
	    	[dom,SR],
	    	proc(aut,x)
		    local i,g;
		begin
		    g:=OUTPUT(1);
		    for i from 1 to DIMENSION(1) do
			g[i,1]:=x*g[i,1]
		    end_for;
		    return(dom::new(DIMENSION(1),ALPHABET(1),INPUT(1),
				    TRANSITION(1),g))
		end_proc);
	    operators::overloaded::declareSignature(operators::plus2,
	    	[Dom::Integer,dom],
		proc(x,aut)
		    local a,pp;
		begin
		    pp:=domains::repeatedSquaring(dom,"_plus");
		    if (x<0) then
			if not(SR::hasProp(Cat::Ring)) then
			    error("Substraction is not defined")
			else
			    a:=pp(dom::new(1,[],SRM(1,1,[SR::one]),table(),
					SRM(1,1,[SR::one])),-x);
			    return(-a+aut)
			end_if
		    else
			a:=pp(dom::new(1,[],SRM(1,1,[SR::one]),table(),
					SRM(1,1,[SR::one])),x);
			return(a+aut)
		    end_if;
		end_proc);
	    operators::overloaded::declareSignature(operators::plus2,
	    	[dom,Dom::Integer],
		proc(aut,x)
		    local a,pp;
		begin
		    pp:=domains::repeatedSquaring(dom,"_plus");
		    if (x<0) then
			if not(SR::hasProp(Cat::Ring)) then
			    error("Substraction is not defined")
			else
			    a:=pp(dom::new(1,[],SRM(1,1,[SR::one]),table(),
					SRM(1,1,[SR::one])),-x);
			    return(aut-a)
			end_if
		    else
			a:=pp(dom::new(1,[],SRM(1,1,[SR::one]),table(),
					SRM(1,1,[SR::one])),x);
			return(aut+a)
		    end_if; 
		end_proc);
	    operators::overloaded::declareSignature(operators::mult2,
	    	[Dom::Integer,dom],
	    	proc(x,aut)
		    local a,pp;
		begin
		    pp:=domains::repeatedSquaring(dom,"_plus");
		    if (x<0) then
			if not(SR::hasProp(Cat::Ring)) then
			    error("Substraction is not defined")
			else
			    a:=pp(aut,-x);
			    return(dom::new(1,[],SRM(1,1,[-SR::one]),table(),
					SRM(1,1,[SR::one]))*a)
			end_if
		    else
			return(pp(aut,x))
		    end_if
		end_proc);
	    operators::overloaded::declareSignature(operators::mult2,
	    	[dom,Dom::Integer],
	    	proc(aut,x)
		    local a,pp;
		begin
		    pp:=domains::repeatedSquaring(dom,"_plus");
		    if (x<0) then
			if not(SR::hasProp(Cat::Ring)) then
			    error("Substraction is not defined")
			else
			    a:=pp(aut,-x);
			    return(a*
				dom::new(1,[],SRM(1,1,[-SR::one]),table(),
						    	SRM(1,1,[SR::one])))
			end_if
		    else
			return(pp(aut,x))
		    end_if   
		end_proc)
    	end_proc
    else
	proc()
            option escape;
    	begin
	    operators::overloaded::declareConversion(SR, dom,
		x->new(dom,1,[],SRM(1,1,[x]),table(),SRM(1,1,[SR::one])),
	    	Explicit);
 	    operators::overloaded::declareSignature(operators::plus2,
            	[dom,dom],
       	    	(aut1,aut2) -> dom::plus(aut1,aut2));
	    operators::overloaded::declareSignature(operators::plus2,
	    	[SR,dom],
	    	(x,aut) -> dom::plus(dom::new(1,[],
						SRM(1,1,[x]),table(),
						SRM(1,1,[SR::one])),aut));
	    operators::overloaded::declareSignature(operators::plus2,
	    	[dom,SR],
	    	(aut,x) -> dom::plus(aut,dom::new(1,[],SRM(1,1,[x]),table(),
					SRM(1,1,[SR::one]))));
	    operators::overloaded::declareSignature(operators::mult2,
            	[dom,dom],
       	    	(aut1,aut2) -> dom::mult(aut1,aut2));
	    operators::overloaded::declareSignature(operators::mult2,
	    	[SR,dom],
		proc(x,aut)
		    local i,l;
		begin
		    l:=INPUT(2);
		    for i from 1 to DIMENSION(2) do
			l[1,i]:=x*l[1,i]
		    end_for;
		    return(dom::new(DIMENSION(2),ALPHABET(2),l,TRANSITION(2),
				    OUTPUT(2)))
		end_proc);
	    operators::overloaded::declareSignature(operators::mult2,
	    	[dom,SR],
	    	proc(aut,x)
		    local i,g;
		begin
		    g:=OUTPUT(1);
		    for i from 1 to DIMENSION(1) do
			g[i,1]:=x*g[i,1]
		    end_for;
		    return(dom::new(DIMENSION(1),ALPHABET(1),INPUT(1),
				    TRANSITION(1),g))
		end_proc);
	end_proc
    end_if;

    /////////////////////////////////////////////////////////////////////////
    // Weight of a word
    /////////////////////////////////////////////////////////////////////////

    weight :=
    proc(automaton : dom,
	 word	   : DOM_LIST)
	local i,sword,w;
    begin
	w:=INPUT(1);
	sword:=extnops(word);
	i:=1;
	while (i<=sword) and contains(TRANSITION(1),word[i]) do
	      w:=w*TRANSITION(1)[word[i]];
	      i:=i+1
	end_while;
	if i=sword+1 then 
	    return((w*OUTPUT(1))[1,1])
	else
	    return(SR::zero)
	end_if
    end_proc;

    /////////////////////////////////////////////////////////////////////////
    // Proper Automaton 
    /////////////////////////////////////////////////////////////////////////

    proper := 
    proc(automaton : dom)
	local co,i,inp,mu,out,sa,sm,tra;
    begin
	sa:=extnops(ALPHABET(1));
	sm:=SRM::stackMatrix;
	co:=SRM::_concat;
	mu:=SRM::_mult;
	inp:=SRM(1,DIMENSION(1)+1,[SR(1)]);
	out:=sm(SRM(1,1),OUTPUT(1));
	tra:=table();
	for i from 1 to sa do
            tra[ALPHABET(1)[i]]:=co(SRM(DIMENSION(1)+1,1),
				    sm(mu(INPUT(1),
					  TRANSITION(1)[ALPHABET(1)[i]]),
					  TRANSITION(1)[ALPHABET(1)[i]]))
	end_for;
	return(dom::new(DIMENSION(1)+1,ALPHABET(1),inp,tra,out))
    end_proc;


    /////////////////////////////////////////////////////////////////////////
    // Dimension of an automaton 
    /////////////////////////////////////////////////////////////////////////

    dimension := 
    proc(automaton : dom)
    begin
	return(DIMENSION(1))
    end_proc;

    /////////////////////////////////////////////////////////////////////////
    // Alphabet of an automaton
    /////////////////////////////////////////////////////////////////////////

    alphabet :=
    proc(automaton : dom)
    begin
	return(ALPHABET(1))
    end_proc;

    
    /////////////////////////////////////////////////////////////////////////
    // Initial vector
    /////////////////////////////////////////////////////////////////////////

    initial :=
    proc(automaton : dom)
    begin
	return(INPUT(1))
    end_proc;

    /////////////////////////////////////////////////////////////////////////
    // Transition matrix
    /////////////////////////////////////////////////////////////////////////

    transition :=
    proc()
    begin
	if args(0)=1 then
	    return(dom::automattrans1(args()))
	else
	    return(dom::automattrans2(args()))
	end_if
    end_proc;

    automattrans1 :=
    proc(automaton : dom)
    begin
	return(extop(TRANSITION(1)))
    end_proc;

    automattrans2 :=
    proc(automaton : dom,
	 letter)
    begin
	if (contains(ALPHABET(1),letter)<>0) then 
	    return(TRANSITION(1)[letter])
	else 
	    return(FAIL)
	end_if
    end_proc;

    /////////////////////////////////////////////////////////////////////////
    // Final vector
    /////////////////////////////////////////////////////////////////////////

    final :=
    proc(automaton : dom)
    begin
	return(OUTPUT(1))
    end_proc;

    /////////////////////////////////////////////////////////////////////////
    // Matrix operation 
    /////////////////////////////////////////////////////////////////////////

    tensor := 
    proc(smatrix1 : SRM,
	 smatrix2 : SRM)
        local co,d1,d2,i,j,l,m,md,sm; 
    begin
	md:=SRM::matdim;
	d1:=md(smatrix1);
	d2:=md(smatrix2);
	m:=SRM(0,d1[2]*d2[2]);
	co:=SRM::_concat;
	sm:=SRM::stackMatrix;
	for i from 1 to d1[1] do
	    l:=SRM(d2[1],0);
	    for j from 1 to d1[2] do
		l:=co(l,SRM::_mult(smatrix1[i,j],smatrix2))
	    end_for;
	    m:=sm(m,l)
	end_for;
	return(m)
    end_proc;

    mstar :=
    proc(smatrix : SRM)
	local dim,dim1,m,subm1,subm2,sm;
    begin
	dim:=SRM::matdim(smatrix)[1];
	if dim=0 then
	    return(smatrix)
	elif dim=1 then
	    return(SRM(1,1,SR::star(smatrix[1,1])))
	end_if; 
	if dim mod 2=0 then 
	    dim1:=dim/2
	else
	    dim1:=dim-1
	end_if;
	sm:=SRM::stackMatrix;
	subm1:=dom::mstar(smatrix[1..dim1,1..dim1]);
	subm2:=dom::mstar(smatrix[dim1+1..dim,dim1+1..dim]);
	m:=dom::mstar(smatrix[1..dim1,1..dim1]
		+smatrix[1..dim1,dim1+1..dim]*subm2
		*smatrix[dim1+1..dim,1..dim1]);
	m:=sm(m,subm2*smatrix[dim1+1..dim,1..dim1]*m);
	subm2:=dom::mstar(smatrix[dim1+1..dim,dim1+1..dim]
		    +smatrix[dim1+1..dim,1..dim1]*
		    subm1*smatrix[1..dim1,dim1+1..dim]);
	return(SRM::_concat(m,sm(subm1*smatrix[1..dim1,dim1+1..dim]*subm2,
				subm2)))
    end_proc;
    
    /////////////////////////////////////////////////////////////////////////
    // Type convertion
    /////////////////////////////////////////////////////////////////////////

    changeCoeffRing :=
    proc(automaton : dom,
	 T : DOM_DOMAIN)
	local c,inp,l,lp,letter,out,r,t,tra,TM;
    begin
	TM:=Dom::Matrix(T);
	lp:=extop(INPUT(1))[3];
	t:=table();
	// Note from Nicolas: we use here heavily our knowledge
	// of the internal representation of matrices; is there any
	// way to avoid this?
	for c from 1 to DIMENSION(1) do
	    t[1,c]:=T(extop(coeff(lp[c],hold(_X),1),1));
	end_for;
	inp:=TM(1,DIMENSION(1),t);
	l:=poly2list(extop(OUTPUT(1))[3][1]);
	t:=table();
	for r from 1 to extnops(l) do
	    t[l[r][2],1]:=T(extop(l[r][1],1))
	end_for;
	out:=TM(DIMENSION(1),1,t);
	tra:=table();
	for letter in ALPHABET(1) do
	    t:=table();
	    lp:=extop(TRANSITION(1)[letter])[3];
	    for c from 1 to DIMENSION(1) do
		l:=poly2list(lp[c]);
		for r from 1 to extnops(l) do
		    t[l[r][2],c]:=T(extop(l[r][1],1))
		end_for
	    end_for;
	    tra[letter]:=TM(DIMENSION(1),DIMENSION(1),t)
	end_for;
	return(Dom::WeightedAutomaton(T)(DIMENSION(1),ALPHABET(1),
		   inp,tra,out))
    end_proc;


    /////////////////////////////////////////////////////////////////////////
    // Operations on weighted automaton
    /////////////////////////////////////////////////////////////////////////

    addLetter := 
    proc(automaton : dom, 
	 letter, 
	 transition : SRM)
    	local trans;
    begin
    	if contains(ALPHABET(1),letter)>0 then 
	    error("Letter to add already appears in the weighted automaton") 
	end_if;
    	if SRM::matdim(transition)<>[DIMENSION(1),DIMENSION(1)] then 
	    error("Incorrect dimension of the transition matrix") 
	end_if;
	trans:=TRANSITION(1);
	trans[letter]:=transition;
	return(dom::new(DIMENSION(1),append(ALPHABET(1),letter),INPUT(1),trans
								,OUTPUT(1)));
    end_proc;

    delLetter := 
    proc(automaton : dom, 
	 letter)
	local alph,i,trans;
    begin
	i:=contains(ALPHABET(1),letter);
	if i>0 then
	    alph:=ALPHABET(1);
	    delete alph[i]; 
	    trans:=TRANSITION(1);
	    delete trans[letter];
 	    return(dom::new(DIMENSION(1),alph,INPUT(1),trans,OUTPUT(1)))
	end_if;
	return(automaton)
    end_proc;

    addState := 
    proc()
    begin
	if args(0)=1 then
	    dom::addState1(args())
	elif args(0)=5 then 
	    dom::addState5(args())
	else
	    error("wrong number of arguments")
	end_if
    end_proc;

    addState1 := 
    proc(automaton : dom)
	local letter,sC,sM,trans;
    begin
	sC:=SRM::_concat;
    	sM:=SRM::stackMatrix;
	trans:=table();
    	for letter in ALPHABET(1) do 
	    trans[letter]:=sM(sC(TRANSITION(1)[letter],SRM(DIMENSION(1),1)),
			      SRM(1,DIMENSION(1)+1))
    	end_for;
    	return(dom::new(DIMENSION(1)+1,ALPHABET(1),sC(INPUT(1),SRM(1,1)),trans,
       			sM(OUTPUT(1),SRM(1,1))))
    end_proc;

    addState5 := 
    proc(automaton : dom,
	 itrans : SRM,
	 otrans :SRM,
         iweight : SR,
	 fweight : SR)
    	local ix,md,sC,sM,trans,letter;
    begin
	sC:=SRM::_concat;
    	sM:=SRM::stackMatrix;
	trans:=table();
	md:=SRM::matdim;
	if md(itrans)[2]<>DIMENSION(1)+1 then
	    error("Wrong number of ingoing transitions")
	end_if;
	if md(otrans)[1]<>DIMENSION(1)+1 then
	    error("Wrong number of outgoing transitions")
	end_if;
	if itrans[1,DIMENSION(1)+1]<>otrans[DIMENSION(1)+1,1] then 
	    warning("Different weights are given for the cycle")
	end_if;
	ix:=SRM::_index;
	for letter in ALPHABET(1) do
	    trans[letter]:=sM(sC(TRANSITION(1)[letter],
			       	 ix(otrans,1..DIMENSION(1),1..1)),
			      itrans)
    	end_for;
    	return(dom::new(DIMENSION(1)+1,ALPHABET(1),
			sC(INPUT(1),SRM(1,1,[iweight])),trans,
       			sM(OUTPUT(1),SRM(1,1,[fweight]))))
    end_proc;

    delState :=
    proc(automaton : dom, 
	 state : Type::PosInt)
    	local dC,dR,letter,trans;
    begin
    	if state>DIMENSION(1) then 
	    error("Incorrect state to delete")
	end_if;
    	dC:=SRM::delCol;
    	dR:=SRM::delRow;
	for letter in ALPHABET(1) do
	     trans[letter]:=dC(dR(TRANSITION(1)[letter],state),state)
	end_for;
    	return(dom::new(DIMENSION(1)-1,ALPHABET(1),dC(INPUT(1),state),trans,
                     dR(OUTPUT(1),state)))
    end_proc;

    subsTransition := 
    proc()
    begin
	if args(0)=2 then
	    return(dom::subsTransition3(args(),SR::one))
	else
	    return(dom::subsTransition3(args()))
	end_if
    end_proc;

    subsTransition3 :=
    proc(automaton : dom,
	 t : Type::Predicate("Transition",l->_lazy_and(testtype(l,DOM_LIST),
				     	     testtype(l[1],DOM_INT),
				             testtype(l[3],DOM_INT))),
	 weight : SR) 
	local m;
    begin
    	if contains(ALPHABET(1),t[2])=0 then 
	    error("Wrong letter")
	end_if;
	if (t[1]>DIMENSION(1)) or (t[3]>DIMENSION(1)) then 
	    error("Wrong state")
	end_if;
	m:=TRANSITION(1);
	m[t[2]][t[1],t[3]]:=weight;
    	return(dom::new(DIMENSION(1),ALPHABET(1),INPUT(1),m,OUTPUT(1)))
    end_proc;

   subsInitialWeight := 
    proc()
    begin
	if args(0)=2 then
	    return(dom::subsInitialWeight3(args(),SR::one))
	else
	    return(dom::subsInitialWeight3(args()))
	end_if
    end_proc;

    subsInitialWeight3 :=
    proc(automaton : dom,
	 state : Type::PosInt,
	 weight : SR) 
	local m;
    begin
	if (state>DIMENSION(1)) then 
	    error("Wrong state")
	end_if;
	m:=INPUT(1);
	m[1,state]:=weight;
    	return(dom::new(DIMENSION(1),ALPHABET(1),m,TRANSITION(1),OUTPUT(1)))
    end_proc;

   subsFinalWeight := 
    proc()
    begin
	if args(0)=2 then
	    return(dom::subsFinalWeight3(args(),SR::one))
	else
	    return(dom::subsFinalWeight3(args()))
	end_if
    end_proc;

    subsFinalWeight3 :=
    proc(automaton : dom,
	 state : Type::PosInt,
	 weight : SR) 
	local m;
    begin
	if (state>DIMENSION(1)) then 
	    error("Wrong state")
	end_if;
	m:=OUTPUT(1);
	m[state,1]:=weight;
    	return(dom::new(DIMENSION(1),ALPHABET(1),INPUT(1),TRANSITION(1),m))
    end_proc;



    /////////////////////////////////////////////////////////////////////////
    // Arithmetic operations
    /////////////////////////////////////////////////////////////////////////

    _plus := operators::_plus;

    plus :=
    proc(automaton1 : dom,
	 automaton2 : dom)
	local a,a2,co,i,inp,j,out,sa,sm,tra,z11,z12,z21,z22;
    begin
        a:=ALPHABET(1);
        sa:=extnops(a);
	a2:=ALPHABET(2);
	co:=SRM::_concat;
	sm:=SRM::stackMatrix;
	inp:=co(INPUT(1),INPUT(2));
	out:=sm(OUTPUT(1),OUTPUT(2));
        tra:=table();
	z21:=SRM(DIMENSION(2),DIMENSION(1));
	z12:=SRM(DIMENSION(1),DIMENSION(2));
	z22:=SRM(DIMENSION(2),DIMENSION(2));
	z11:=SRM(DIMENSION(1),DIMENSION(1));
	for i from 1 to sa do
	    j:=contains(a2,a[i]);
	    if j<>0 then
		tra[a[i]]:=co(sm(TRANSITION(1)[a[i]],z21),
			      sm(z12,TRANSITION(2)[a[i]]));
		delete a2[j]
	    else
		tra[a[i]]:=co(sm(TRANSITION(1)[a[i]],z21),
			      sm(z12,z22))
	    end_if
	end_for;
	a:=a.a2;
	sa:=extnops(a);
	for i from extnops(ALPHABET(1))+1 to sa do 
	    tra[a[i]]:=co(sm(z11,z21),
			  sm(z12,TRANSITION(2)[a[i]]))
	end_for;
	return(dom::new(DIMENSION(1)+DIMENSION(2),a,inp,tra,out))
    end_proc;

    _mult := operators::_mult;

    mult := 
    proc(automaton1 : dom,
	 automaton2 : dom)
	local a,a2,co,i,j,inp,mu,out,sa,sm,tra,z11,z12,z21,z22;
    begin
	a:=ALPHABET(1);
        sa:=extnops(a);
	a2:=ALPHABET(2);
	co:=SRM::_concat;
	sm:=SRM::stackMatrix;
	mu:=SRM::_mult;
	inp:=co(INPUT(1),SRM(1,DIMENSION(2)));
	out:=sm(mu(OUTPUT(1),INPUT(2),OUTPUT(2)),OUTPUT(2));
	tra:=table();
	z21:=SRM(DIMENSION(2),DIMENSION(1));
	z12:=SRM(DIMENSION(1),DIMENSION(2));
	z22:=SRM(DIMENSION(2),DIMENSION(2));
	z11:=SRM(DIMENSION(1),DIMENSION(1));
	for i from 1 to sa do
	    j:=contains(a2,a[i]);	
	    if j<>0 then
		tra[a[i]]:=co(sm(TRANSITION(1)[a[i]],z21),
			      sm(mu(OUTPUT(1),INPUT(2),TRANSITION(2)[a[i]]),
				 TRANSITION(2)[a[i]]));
		delete a2[j]
	    else
		tra[a[i]]:=co(sm(TRANSITION(1)[a[i]],z21),
			      sm(z12,z22))
	    end_if
	end_for;
	a:=a.a2;
	sa:=extnops(a);
	for i from extnops(ALPHABET(1))+1 to sa do 
	    tra[a[i]]:=co(sm(z11,z21),
		      	  sm(mu(OUTPUT(1),INPUT(2),TRANSITION(2)[a[i]]),
			     TRANSITION(2)[a[i]]))
	end_for;
	return(dom::new(DIMENSION(1)+DIMENSION(2),a,inp,tra,out))
    end_proc; 

    _power := domains::repeatedSquaring(dom,"_mult");

    hadamard := 
    proc(automaton1 : dom,
	 automaton2 : dom)
	local a,a2,d,i,inp,j,out,sa,tra;
    begin
	a:=ALPHABET(1);
        sa:=extnops(a);
	a2:=ALPHABET(2);
	inp:=dom::tensor(INPUT(1),INPUT(2));
	out:=dom::tensor(OUTPUT(1),OUTPUT(2));
	tra:=table();
	d:=DIMENSION(1)*DIMENSION(2);
	for i from 1 to sa do
	    j:=contains(a2,a[i]);	
	    if j<>0 then
		tra[a[i]]:=dom::tensor(TRANSITION(1)[a[i]],
				       TRANSITION(2)[a[i]]);
		delete a2[j]
	    else
		delete a[i]
	    end_if
	end_for;
	return(dom::new(d,a,inp,tra,out))
    end_proc; 		
	 
    shuffle := 
    proc(automaton1 : dom,
	 automaton2 : dom)
        local a,a2,i,i1,i2,inp,j,out,p,sa,tra;
    begin
	a:=ALPHABET(1);
        sa:=extnops(a);
	a2:=ALPHABET(2);
	inp:=dom::tensor(INPUT(1),INPUT(2));
	out:=dom::tensor(OUTPUT(1),OUTPUT(2));
	i1:=SRM(DIMENSION(1),DIMENSION(1),i->SR(1),Diagonal);
	i2:=SRM(DIMENSION(2),DIMENSION(2),i->SR(1),Diagonal);
	p:=SRM::_plus;
	tra:=table();
	for i from 1 to sa do
	    j:=contains(a2,a[i]);	
	    if j<>0 then
		tra[a[i]]:=p(
		   dom::tensor(TRANSITION(1)[a[i]],i2),
		   dom::tensor(i1,TRANSITION(2)[a[i]]));
		   delete a2[j]
	    else
		tra[a[i]]:=
		   dom::tensor(TRANSITION(1)[a[i]],i2)
	    end_if
	end_for;
	a:=a.a2;
	sa:=extnops(a);
	for i from extnops(ALPHABET(1))+1 to sa do 
	    tra[a[i]]:=dom::tensor(i1,TRANSITION(2)[a[i]])
	end_for;
	return(dom::new(DIMENSION(1)*DIMENSION(2),a,inp,tra,out))
    end_proc; 	

    infiltration := 
    proc(automaton1 : dom,
	 automaton2 : dom)
	local a,i,p,sa,sautomaton,tra;

    begin
	sautomaton:=dom::shuffle(automaton1,automaton2);
	a:=dom::automatalph(sautomaton);
        sa:=extnops(a);
	p:=SRM::_plus;
	tra:=table();
	for i from 1 to sa do
	    if (contains(ALPHABET(1),a[i])<>0) and  
	       (contains(ALPHABET(2),a[i])<>0) then 
		tra[a[i]]:=p(dom::automattrans(sautomaton,a[i]),
		   	 dom::tensor(TRANSITION(1)[a[i]],TRANSITION(2)[a[i]]))
	    else
		tra[a[i]]:=dom::automattrans(sautomaton,a[i])
	    end_if
	end_for;
	return(dom::new(dom::automatdim(sautomaton),
	       		a,
	       		dom::automatinit(sautomaton),
			tra,
			dom::automatfin(sautomaton)))
    end_proc; 	

    star := 
    proc(automaton : dom)
	local co,i,inp,mu,out,pl,sa,sm,tra,w;
    begin
	w:=SR::star(dom::weight(automaton,[]));
	if w=FAIL then
	    return(FAIL)
	end_if;
	sa:=extnops(ALPHABET(1));
	sm:=SRM::stackMatrix;
	co:=SRM::_concat;
	pl:=SRM::_plus;
	mu:=SRM::_mult;
	inp:=SRM(1,DIMENSION(1)+1,[w]);
	out:=sm(SRM(1,1,[SR::one]),OUTPUT(1)*w);
	tra:=table();
	for i from 1 to sa do
            tra[ALPHABET(1)[i]]:=
		co(SRM(DIMENSION(1)+1,1),
		   sm(mu(SRM(1,1,[w]),INPUT(1),TRANSITION(1)[ALPHABET(1)[i]]),
		      pl(TRANSITION(1)[ALPHABET(1)[i]],
			 mu(OUTPUT(1),
			    INPUT(1),
			    TRANSITION(1)[ALPHABET(1)[i]])*w)))
	end_for;
	return(dom::new(DIMENSION(1)+1,ALPHABET(1),inp,tra,out))
    end_proc;

    /////////////////////////////////////////////////////////////////////////
    // Algebraic operations
    /////////////////////////////////////////////////////////////////////////

    transpose :=
    proc(automaton : dom)
	local letter,trans;
    begin
	trans:=table(); 
	for letter in ALPHABET(1) do
	    trans[letter]:=SRM::transpose(TRANSITION(1)[letter])
	end_for;
	return(dom::new(DIMENSION(1),ALPHABET(1),SRM::transpose(OUTPUT(1)),
				trans,SRM::transpose(INPUT(1))))
    end_proc;

    elimination :=
    proc(automaton : dom,
	 letter)
	local trans, alph,i,m,n,p;
    begin
	p:=contains(ALPHABET(1),letter);
	if p=0 then
	    return(automaton)
	end_if;
	alph:=ALPHABET(1);
	delete(alph[p]);
	m:=dom::mstar(TRANSITION(1)[letter]);
	trans:=table();
	n:=extnops(alph);
	for i from 1 to n do
	    trans[alph[i]]:=m*TRANSITION(1)[alph[i]]
	end_for;
        return(dom::new(DIMENSION(1),alph,INPUT(1),trans,m*OUTPUT(1)))
    end_proc;

    /////////////////////////////////////////////////////////////////////////
    // Boolean operations
    /////////////////////////////////////////////////////////////////////////

    complementary := 
    if SR=Dom::BooleanSemiRing then
    	proc(automaton : dom)
	    local automaton1,dim,i,out;
    	begin
 	    automaton1:=dom::complete(dom::determinization(automaton));
	    dim:=dom::dimension(automaton1);
	    out:=dom::final(automaton1);
	    for i from 1 to dim do
		if out[i,1]=SR::zero then 
		    out[i,1]:=SR::one
                else 
		    out[i,1]:=SR::zero 
		end_if;
	    end_for;
	    return(new(dom,dim,ALPHABET(1),dom::initial(automaton1),
			table(dom::transition(automaton1)),out))
    	end_proc
    else
	proc(automaton : dom)
	begin
	    return(FAIL)
	end_proc
    end_if;

    difference :=
    if SR=Dom::BooleanSemiRing then
    	proc(automaton1, automaton2 : dom)
    	begin    
    	    return(dom::hadamard(automaton1,dom::complementary(automaton2)))
    	end_proc
    else
	proc(automaton1, automaton2 :dom)
	begin
	    return(FAIL)
	end_proc
    end_if;

    /////////////////////////////////////////////////////////////////////////
    // Complete automaton
    /////////////////////////////////////////////////////////////////////////

    isComplete := 
    proc(automaton : dom)
	local i,j,k;
    begin
    	i:=1;
	while (i<=DIMENSION(1)) and (INPUT(1)[i]=SR::zero) do
    	    i:=i+1
	end_while;
	if i>DIMENSION(1) then
	    return(FALSE)
	end_if;
	i:=1;
	while (i<=extnops(ALPHABET(1))) do
	    j:=1;
	    while (j<=DIMENSION(1)) do
		k:=1;
	        while _lazy_and(k<=DIMENSION(1),
			(TRANSITION(1)[ALPHABET(1)[i]])[j,k]=SR::zero) do
		    k:=k+1
		end_while;
		if k>DIMENSION(1) then
		    return(FALSE)
		end_if;
		j:=j+1;
	    end_while;
	    i:=i+1;
	end_while;
    return(TRUE)
    end_proc;


    complete := 
    proc(automaton : dom)
	local conc,i,j,k,m,sM,v;
    begin
    sM:=SRM::stackMatrix;
    conc:=SRM::_concat;
    if dom::isComplete(automaton) then
	return(automaton)
    end_if;
    i:=1;
    while (i<=DIMENSION(1)) and (INPUT(1)[i]=SR::zero) do
        i:=i+1
    end_while;
    if i>DIMENSION(1) then
        return(dom::new(0,ALPHABET(1),SRM(1,1,[SR::one]),
	       table(ALPHABET(1)[i]=SRM(1,1,[SR::one])
				$i=1..extnops(ALPHABET(1))),SRM(1,1)))
    end_if;
    m:=table();
    for i from 1 to extnops(ALPHABET(1)) do
    	v:=SRM(DIMENSION(1),1);
    	for j from 1 to DIMENSION(1) do
	    k:=1;
            while (k<=DIMENSION(1)) and 
			((TRANSITION(1)[ALPHABET(1)[i]])[j,k]=SR::zero) do
	    	k:=k+1
	    end_while;
	    if k>DIMENSION(1) then
	    	v[j,1]:=SR::one
	    end_if;
    	end_for;
	m[ALPHABET(1)[i]]:=TRANSITION(1)[ALPHABET(1)[i]];
	m[ALPHABET(1)[i]]:=sM(conc(m[ALPHABET(1)[i]],v),SRM(1,DIMENSION(1)+1));
    	(m[ALPHABET(1)[i]])[DIMENSION(1)+1,DIMENSION(1)+1]:=SR::one;
    end_for;
    return(dom::new(DIMENSION(1)+1,ALPHABET(1),conc(INPUT(1),SRM(1,1)),
					m,sM(OUTPUT(1),SRM(1,1))))
    end_proc;

    /////////////////////////////////////////////////////////////////////////
    // Accessibility
    /////////////////////////////////////////////////////////////////////////

    isAccessible :=
    proc()
    begin
	if args(0)=1 then
	    return(dom::isAccessible1(args()))
	elif testtype(args(2),Type::PosInt) then
	    return(dom::isAccessible21(args()))
	else 
	    return(dom::isAccessible22(args()))
	end_if
    end_proc;

    isAccessible1 :=
    proc(automaton : dom)
	local BM,BSR,i,letter,m,s,S;
    begin
	if DIMENSION(1)=0 then
	    return(TRUE)
	end_if;
	BSR:=Dom::BooleanSemiRing;
	BM:=Dom::Matrix(BSR);
	m:=BM(DIMENSION(1),DIMENSION(1));
	for letter in ALPHABET(1) do
	    m:=m+SRM::convert_to(TRANSITION(1)[letter],BM)
	end_for;
       	m:=SRM::convert_to(INPUT(1),BM)
				*(Dom::WeightedAutomaton(BSR))::mstar(m);
	s:=0;
	S:={};
	for i from 1 to DIMENSION(1) do
	    if m[1,i]=BSR::one then 
		s:=s+1;
		S:=S union {i}
	    end_if
	end_for;
	if bool(s=DIMENSION(1)) then
	    return(TRUE)
	else
	    return(FALSE,S)
	end_if
    end_proc;

    isAccessible21 :=
    proc(automaton : dom,
	 state : Type::PosInt)
	local L;
    begin
	if state>DIMENSION(1) then
	    return(FALSE)
	else
	    L:=[dom::isAccessible1(automaton)];
	    if extnops(L)=1 then
	    	return(TRUE)
	    else
	    	return(contains(L[2],state))
	    end_if
	end_if
    end_proc;

    isAccessible22 :=
    proc(automaton : dom,
	 S : Type::SetOf(Type::PosInt))
	local L;
    begin
	L:=[dom::isAccessible1(automaton)];
	if extnops(L)=1 then
	    return(bool((S minus {$1..DIMENSION(1)})={}))
	else
	    return(bool((S minus L[2])={}))
	end_if
    end_proc;

    accessible := 
    proc(automaton : dom)
	local dc,dr,i,inp,L,letter,nb,outp,S,state,trans;
    begin
	L:=dom::isAccessible(automaton);
	if extnops(L)=1 then
	    return(automaton)
	elif extnops(L[2])=0 then
	    return(dom::new(0,ALPHABET(1),SRM(1,0),
		   table(ALPHABET(1)[i]=SRM(0,0)$i=1..extnops(ALPHABET(1))),
		   SRM(0,1)))
	else
	    dc:=SRM::delCol;
	    dr:=SRM::delRow;
	    inp:=INPUT(1);
	    outp:=OUTPUT(1);
	    trans:=TRANSITION(1);
	    nb:=0;
	    S:=sort([extop({$1..DIMENSION(1)} minus L[2])]);
	    for state in S do
		inp:=dc(inp,state-nb);
		outp:=dr(outp,state-nb);
		for letter in ALPHABET(1) do 
		    trans[letter]:=dc(dr(trans[letter],state-nb),state-nb)
		end_for;
		nb:=nb+1
	    end_for;
	    return(dom::new(DIMENSION(1)-nb,ALPHABET(1),inp,trans,outp))
	end_if
    end_proc;

    isCoaccessible :=
    proc()
    begin
	if args(0)=1 then
	    return(dom::isAccessible1(dom::transpose(args(1))))
	elif testtype(args(2),Type::PosInt) then
	    return(dom::isAccessible21(dom::transpose(args(1)),args(2)))
	else 
	    return(dom::isAccessible22(dom::transpose(args(1)),args(2)))
	end_if
    end_proc;

    coaccessible :=
    proc(automaton : dom)
    begin
	return(dom::transpose(dom::accessible(dom::transpose(automaton))))
    end_proc;

    trim := 
    proc(automaton : dom)
    begin
	return(dom::coaccessible(dom::accessible(automaton)))
    end_proc;

    /////////////////////////////////////////////////////////////////////////
    // Determinization
    /////////////////////////////////////////////////////////////////////////

    determinization := 
    if SR=Dom::BooleanSemiRing then
    	proc(automaton : dom)
    	begin
	    if dom::isDeterministic(automaton) then
	    	return(automaton)
	    end_if;
	    return(dom::boolean_determinization(automaton))
    	end_proc
    else
	proc(automaton : dom)
	begin
	    if dom::isDeterministic(automaton) then
	       	return(automaton)
	    end_if;
	    return(FAIL)
	end_proc
    end_if;

    isDeterministic :=
    proc(automaton : dom)
	local a,c,ind,r;
    begin
	if DIMENSION(1)=0 then
	    return(FALSE)
	end_if;
	ind:=0;
	for c from 1 to DIMENSION(1) do 
	    if INPUT(1)[1,c]<>SR::zero then
		ind:=ind+1;
		if ind=2 then
		    return(FALSE)
		end_if
	    end_if
	end_for;
	if ind=0 then
	    return(FALSE)
	end_if;
	for a in ALPHABET(1) do
	    for r from 1 to DIMENSION(1) do 
		  ind:=0;
		  for c from 1 to DIMENSION(1) do
		      if TRANSITION(1)[a][r,c]<>SR::zero then
			  ind:=ind+1;
		      	  if ind=2 then 
			      return(FALSE)
		      	  end_if
		      end_if
		  end_for
	    end_for
	end_for;
	return(TRUE)
    end_proc;		  

    /////////////////////////////////////////////////////////////////////////
    // Boolean Determinization
    /////////////////////////////////////////////////////////////////////////

    follow := 
    proc(automaton : dom,
	 v : SRM,
	 letter) 
    begin
	return(v*TRANSITION(1)[letter])
    end_proc;

    boolean_determinization :=
    proc(automaton : dom)
   	local c,conc,final,i,inde,indice,j,letter,lst,mark,nalpha,r,sM,trans;
    begin
	conc:=SRM::_concat;
	sM:=SRM::stackMatrix;
	inde:=SRM::_index;
	if DIMENSION(1)=0 then
	    trans:=table();
	    for letter in ALPHABET(1) do
	   	trans[letter]:=SRM(1,1)
	    end_for;
	    return(new(dom,1,ALPHABET(1),SRM(1,1,[1]),trans,SRM(1,1,[0])))
	end_if;
	if INPUT(1)=SRM(1,DIMENSION(1)) then
	    trans:=table();
	    for letter in ALPHABET(1) do
	   	trans[letter]:=SRM(1,1)
	    end_for;
	    return(new(dom,1,ALPHABET(1),SRM(1,1,[1]),trans,SRM(1,1,[0])))
	end_if;
	r:=INPUT(1);
	indice:=2;
	lst:=[r];
	mark:=table();
	mark[SRM(1,DIMENSION(1))]:=1;
	mark[r]:=2;
	trans:=table();
	for letter in ALPHABET(1) do
	    trans[letter]:=table();
	    trans[letter][1,1]:=SR::one
	end_for;
	final:=table();
	repeat
	    r:=lst[1];
	    delete lst[1];
	    if (r*OUTPUT(1))[1,1]<>SR::zero then
		final[mark[r]]:=SR::one
	    end_if;
	    for letter in ALPHABET(1) do
	    	c:=dom::follow(automaton,r,letter);
		if not(testtype(mark[c],DOM_INT)) then
		    indice:=indice+1;
		    mark[c]:=indice;
		    lst:=append(lst,c);
		end_if;
		trans[letter][mark[r],mark[c]]:=SR::one
	    end_for
	until lst=[] end_repeat;
	for letter in ALPHABET(1) do 
	    trans[letter]:=SRM(indice,indice,trans[letter])
	end_for;
	nalpha:=extnops(ALPHABET(1));
	i:=0;
	mark:=TRUE;
	while (i<nalpha) and mark do
	    i:=i+1;
	    j:=2;
	    while (j<=indice) and mark do
	    	if trans[ALPHABET(1)[i]][j,1]<>SR::zero then
		    mark:=FALSE
		end_if;
		j:=j+1
	    end_while
	end_while;
	if mark then 
	    for letter in ALPHABET(1) do
		trans[letter]:=inde(trans[letter],2..indice,2..indice)
	    end_for;
	    return(new(dom,indice-1,ALPHABET(1),
		   SRM(1,indice-1,table(1=SR::one)),trans,
				inde(SRM(indice,1,final),2..indice,1..1)))
	else
	    return(new(dom,indice,ALPHABET(1),
		   SRM(1,indice,table(2=SR::one)),trans,SRM(indice,1,final)))
	end_if
    end_proc;

    /////////////////////////////////////////////////////////////////////////
    // Minimization
    /////////////////////////////////////////////////////////////////////////

    minimization :=
    if SR::hasProp(Cat::EuclideanDomain) then
    	proc(automaton : dom)
    	begin
	    return(dom::EuclideanMinimization(automaton))
	end_proc
    elif SR=Dom::BooleanSemiRing then
   	proc(automaton : dom)
    	begin
	    return(dom::BooleanMinimization(automaton))
	end_proc
    else
   	proc(automaton : dom)
    	begin
            return(FAIL)
    	end_proc
    end_if;

    /////////////////////////////////////////////////////////////////////////
    // Boolean Minimization
    /////////////////////////////////////////////////////////////////////////

    BooleanMinimization :=
    proc(automaton : dom)
    begin
	if not(dom::isDeterministic(automaton)) then
	    automaton:=dom::determinization(automaton)
	else
	    automaton:=dom::accessible(automaton);
	    if not(dom::isComplete(automaton)) then
	    	automaton:=dom::complete(automaton)
	    end_if
	end_if;
        return(dom::det_to_weight(dom::Moore(dom::weight_to_det(automaton))))
    end_proc;

    weight_to_det := 
    proc(automaton : dom)
	local i,initial,j,letter,trans;
    begin
	i:=1;
	while INPUT(1)[1,i]=SR::zero do 
	    i:=i+1
	end_while;
	initial:=[i,INPUT(1)[1,i]];
	trans:=table();
	for letter in ALPHABET(1) do
	    for i from 1 to DIMENSION(1) do
		j:=1;
		while (j<=DIMENSION(1)) and 
				(TRANSITION(1)[letter][i,j]=SR::zero) do
		    j:=j+1
		end_while;
		if j<=DIMENSION(1) then 
		    trans[i,letter]:=[j,TRANSITION(1)[letter][i,j]]
		end_if
    	    end_for
	end_for;
	return([DIMENSION(1),ALPHABET(1),initial,trans,OUTPUT(1)])
    end_proc;

    det_to_weight :=
    proc(detautomaton : DOM_LIST)
	local i,letter,trans;
    begin
	trans:=table();
	for letter in detautomaton[2] do 
	    trans[letter]:=SRM(detautomaton[1],detautomaton[1]);
	end_for;
	for i from 1 to detautomaton[1] do
	    for letter in detautomaton[2] do
		if testtype(detautomaton[4][i,letter],DOM_LIST) then
		    trans[letter][i,detautomaton[4][i,letter][1]]:=
						detautomaton[4][i,letter][2]
		end_if
	    end_for
	end_for;
	return(dom::new(detautomaton[1],detautomaton[2],
		SRM(1,detautomaton[1],
			table((1,detautomaton[3][1])=detautomaton[3][2])),
		trans,detautomaton[5]))
    end_proc;

    Moore := 
    proc(detautomaton : DOM_LIST)
	local Eq,final,i,initial,j,k,letter,L,LL,mark,nalpha,p,q,trans;
    begin
	Eq:=table();
	L:=table();
	for i from 2 to detautomaton[1] do
	    for j from 1 to i-1 do
		if bool((detautomaton[5][i,1]=SR::one) =
				(detautomaton[5][j,1]=SR::one)) then
		    Eq[i,j]:=TRUE;
		    L[i,j]:=[];
		end_if
	    end_for
	end_for;
	nalpha:=extnops(detautomaton[2]);
	for i from 2 to detautomaton[1] do
	    for j from 1 to i-1 do
		if testtype(Eq[i,j],DOM_BOOL) then
		    mark:=TRUE;
		    k:=0;
		    while (k<nalpha) and mark do
			k:=k+1; 
			p:=detautomaton[4][i,detautomaton[2][k]][1];
			q:=detautomaton[4][j,detautomaton[2][k]][1];
			if p<q then
			    p:=q-p;
			    q:=q-p;
			    p:=q+p
			end_if;
			if not((p=i) and (q=j) or (p=q) or 
				testtype(Eq[p,q],DOM_BOOL)) then
			    mark:=FALSE;
			end_if;
		    end_while;
		    if mark then
			for letter in detautomaton[2] do
			    p:=detautomaton[4][i,letter][1];
			    q:=detautomaton[4][j,letter][1];
			    if p<q then
			        p:=q-p;
			    	q:=q-p;
			    	p:=q+p
			    end_if;
			    if ((p<>i) or (q<>j)) and (p<>q) then
			    	L[p,q]:=append(L[p,q],[i,j]);
			    end_if
			end_for
		    else
			delete Eq[i,j];
			LL:=L[i,j];
			delete L[i,j];
			while LL<>[] do
			    if testtype(Eq[extop(LL[1])],DOM_BOOL) then
				delete Eq[extop(LL[1])];
				LL:=LL.L[extop(LL[1])];
				delete L[extop(LL[1])]
			    end_if;
			    delete LL[1]
			end_while
		    end_if
		end_if
	    end_for
	end_for;
	final:=table();
	L:=table();
	LL:=table();
	L[1]:=1;
	LL[1]:=1;
	final[1,1]:=detautomaton[5][1];
	k:=1;
	for i from 2 to detautomaton[1] do
	    j:=0;
	    mark:=TRUE;
	    while (j<i-1) and mark do
		j:=j+1;
		if testtype(Eq[i,j],DOM_BOOL) then
		    mark:=FALSE;
		    L[i]:=L[j];
		end_if;
	    end_while;
	    if mark then
		k:=k+1;
		L[i]:=k;
		LL[k]:=i;
		final[k,1]:=detautomaton[5][i]
	    end_if;
	end_for;
	initial:=[L[detautomaton[3][1]],SR::one];
	trans:=table();
	for i from 1 to k do 
	    for letter in detautomaton[2] do
		if testtype(L[detautomaton[4][LL[i],letter][1]],DOM_INT) then
		    trans[i,letter]:=[L[detautomaton[4][LL[i],letter][1]],
				      SR::one]
		end_if
	    end_for
	end_for;
	return([k,detautomaton[2],initial,trans,SRM(k,1,final)])
    end_proc;

    /////////////////////////////////////////////////////////////////////////
    // Euclidean and Skew Field Minimization
    /////////////////////////////////////////////////////////////////////////

    EuclideanMinimization :=
    proc(automaton : dom)
    begin
        if extnops(ALPHABET(1))=0 then
	    return(dom::new(1,[],SRM(1,1,[SR::one]),table(),
						INPUT(1)*OUTPUT(1)))
        else
            return(dom::transpose(dom::reduction(dom::transpose(
						dom::reduction(automaton)))))
        end_if	
    end_proc;

    reduction := 
    proc(automaton : dom)
        local alphauto,cMx,cpt,dimc,FM,i,L,letter,swC,T1,Tt; 
    begin
        swC:=SRM::swapCol;
        cMx:=SRM::_concat;
    	dimc:=DIMENSION(1);
    	i:=1;
    	while (i<=dimc) and INPUT(1)[1,i]=SR::zero do
            i:=i+1
    	end_while;
    	if i=dimc+1 then
	    return(dom::new(0,ALPHABET(1),SRM(1,0),
		table(ALPHABET(1)[i]=SRM(0,0)$i=1..extnops(ALPHABET(1))),
								SRM(0,1)))
    	else
    	    L:=dom::prefixe(automaton);
    	    cpt:=0;
    	    T1:=SRM(L[1],0);
	    Tt:=TRANSITION(1);
    	    alphauto:=ALPHABET(1);
    	    for i from 1 to L[1] do
	        cpt:=cpt+1;
            	while L[3][i,cpt]=SR::zero do cpt:=cpt+1 end_while;
	    	T1:=cMx(T1,L[3][1..L[1],cpt..cpt]);
	    	for letter in alphauto do
	            Tt[letter]:=swC(Tt[letter],i,cpt)
	    	end_for
            end_for;
    	    for letter in alphauto do
	        Tt[letter]:=(L[3]*Tt[letter])[1..L[1],1..L[1]];
            end_for;
    	    if SR::hasProp(Cat::SkewField) then
	    	T1:=SRM::_invert(T1);
	    	for letter in alphauto do
	            Tt[letter]:=Tt[letter]*T1
	    	end_for;
	    	return(dom::new(L[1],ALPHABET(1),L[2],Tt,L[3]*OUTPUT(1)))
    	    else
	    	FM:=Dom::Matrix(Dom::Fraction(SR));
	    	T1:=SRM::convert_to(T1,FM);
	    	T1:=FM::_invert(T1);
	    	for letter in alphauto do
	            Tt[letter]:=FM::convert_to(SRM::convert_to(Tt[letter],FM)
								      *T1,SRM)
	    	end_for;
	    	return(dom::new(L[1],ALPHABET(1),L[2],Tt,L[3]*OUTPUT(1)))
    	    end_if
        end_if
    end_proc;

    prefixe :=
    proc(automaton : dom)
    	local cMx,dimc,dimr,dR,l,L,M,nalpha,nl,sMx,T,Y,Z;
    begin
    	dimc:=DIMENSION(1);
    	nalpha:=extnops(ALPHABET(1));
    	l:=SRM(1,1,[SR::one]);
    	dimr:=1;
    	M:=INPUT(1);
    	T:=M;
    	Y:=[FALSE];
    	Z:=SRM(0,dimc);
    	nl:=1;
    	sMx:=SRM::stackMatrix;
    	cMx:=SRM::concatMatrix;
    	dR:=SRM::delRow;
    	repeat
	    if Y[1] then
	    	T:=sMx(T,Z[1..1,1..dimc]*TRANSITION(1)[ALPHABET(1)[nl]])
	    else
	    	T:=sMx(T,M[1..1,1..dimc]*TRANSITION(1)[ALPHABET(1)[nl]])
	    end_if;
	    L:=dom::triang(dimr+1,dimc,cMx(l,SRM(1,1)),T);
	    if L[1]=0 then
	    	dimr:=dimr+1;
	    	M:=sMx(M,T[dimr..dimr,1..dimc]);
	    	l:=L[2];
	    	T:=L[3];
	    	Y:=[extop(Y),FALSE]
	    elif L[1]=1 then
	      	T:=dR(T,dimr+1);
       	    else
	    	Z:=sMx(Z,T[dimr+1..dimr+1,1..dimc]);
	    	T:=dR(L[3],dimr+1);
	    	l:=L[2][1..1,1..dimr];
       	    	Y:=[extop(Y),TRUE]
	    end_if;
            nl:=nl+1;
	    if nl=nalpha+1 then
	        if Y[1] then 
		    Z:=dR(Z,1)
	    	else
		    M:=dR(M,1)
	        end_if;	
	        delete Y[1];	
	    	nl:=1
	    end_if;
        until (Y=[])  end_repeat;
    	return([dimr,l,T])
    end_proc;

    triang :=
    proc(dimr : DOM_INT,
     	 dimc : DOM_INT,
         l	  : SRM,
         T 	  : SRM)
    	local cas,cdiag,ccMx,L,nr,posu,posv,seRMx,stMx,x;
    begin
    	seRMx:=SRM::setRow;
    	stMx:=SRM::stackMatrix;
    	ccMx:=SRM::_concat;
    	cas:=0;
    	cdiag:=FALSE;
    	posv:=1;
    	posu:=1;
    	nr:=1;
    	repeat
            while (posv<=dimc) and (T[dimr,posv]=SR::zero) do 
            	posv:=posv+1
	    end_while;
    	    if (posv<=dimc) then
            	while (nr<dimr) and (posu<posv) do
	            if T[nr,posu]=SR::zero then
	            	posu:=posu+1
	   	    else
	           	nr:=nr+1
	   	    end_if
	    	end_while;
	    	if (nr=dimr) and (posu<posv) then
		    cas:=0;
	    	elif T[nr,posu]=SR::zero then
		      if nr=1 then
		          l:=ccMx(l[1..1,dimr..dimr],l[1..1,1..dimr-1]);
		          T:=stMx(T[dimr..dimr,1..dimc],T[1..dimr-1,1..dimc])
		      else
		      	  l:=ccMx(l[1..1,1..nr-1],l[1..1,dimr..dimr],
						l[1..1,nr..dimr-1]);
		          T:=stMx(T[1..nr-1,1..dimc],T[dimr..dimr,1..dimc],
							T[nr..dimr-1,1..dimc])
		      end_if;
		      cas:=0;
		      nr:=dimr;
	        else
	      	      L:=SR::gcdex(T[nr,posv],T[dimr,posv]);
	      	      if L[1]<>T[nr,posv] then 
		          cdiag:=TRUE
	      	      end_if;
		      x:=l[1,nr];
		      l[1,nr]:=x*T[nr,posv]/L[1]+l[1,dimr]*T[dimr,posv]/L[1];
		      l[1,dimr]:=x*(-L[3])+l[1,dimr]*L[2];
		      x:=T[nr..nr,1..dimc];
		      T:=seRMx(T,nr,L[2]*x+L[3]*T[dimr..dimr,1..dimc]);
		      T:=seRMx(T,dimr,-T[dimr,posv]/L[1]*x
				+x[1,posv]/L[1]*T[dimr..dimr,1..dimc]);
	        end_if;
	        posv:=posv+1
	    elif cdiag then
	        cas:=2;
	      	nr:=dimr
	    else
	        cas:=1;
	        nr:=dimr
	    end_if;
        until (nr=dimr) end_repeat;
        return([cas,l,T])
    end_proc;

    /////////////////////////////////////////////////////////////////////////
    // Equivalence
    /////////////////////////////////////////////////////////////////////////

    areEquivalent := 
    if SR=Dom::BooleanSemiRing then
    	proc(automaton1 : dom,
	     automaton2 : dom)
    	begin
	    return(dom::areEquivalentBoolean(automaton1,automaton2))
	end_proc
    elif SR::hasProp(Cat::SkewField) then
    	proc(automaton1 : dom,
	     automaton2 : dom)
    	begin
	    return(dom::areEquivalentSF(automaton1,automaton2))
	end_proc
    elif SR::hasProp(Cat::EuclideanDomain) then
    	proc(automaton1 : dom,
	     automaton2 : dom)
    	begin
	    return(dom::areEquivalentED(automaton1,automaton2))
	end_proc
    else
    	proc(automaton1 : dom,
	     automaton2 : dom)
    	begin
	    return(FAIL)
    	end_proc
    end_if;

    areEquivalentED :=
    proc(automaton1 : dom,
	 automaton2 : dom)
	local F,W;
    begin
	F:=Dom::Fraction(SR);
	W:=Dom::WeightedAutomaton(F);
	return(W::areEquivalentSF(dom::changeCoeffRing(automaton1,F),
				dom::changeCoeffRing(automaton2,F)))
    end_proc;

    areEquivalentSF :=
    proc(automaton1 : dom,
	 automaton2 : dom)
	local letter,sa1,sa2;
    begin
	sa1:={extop(ALPHABET(1))};
	sa2:={extop(ALPHABET(2))};
	for letter in sa1 minus sa2 do
	    automaton2:=dom::addLetter(automaton2,letter,
					SRM(DIMENSION(2),DIMENSION(2)))
	end_for;
	for letter in sa2 minus sa1 do
	    automaton1:=dom::addLetter(automaton1,letter,
					SRM(DIMENSION(1),DIMENSION(1)))
	end_for;
	if dom::dimension(dom::minimization(automaton1-automaton2))=0 then
	    return(TRUE)
	else
	    return(FALSE)
	end_if;
    end_proc;

    areEquivalentBoolean :=
    proc(automaton1 : dom,
	 automaton2 : dom)
   	local detaut1,detaut2,iso,letter,sa1,sa2,state1,state2,stlist;
    begin
	sa1:={extop(ALPHABET(1))};
	sa2:={extop(ALPHABET(2))};
	for letter in sa1 minus sa2 do
	    automaton2:=dom::addLetter(automaton2,letter,
					SRM(DIMENSION(2),DIMENSION(2)))
	end_for;
	for letter in sa2 minus sa1 do
	    automaton1:=dom::addLetter(automaton1,letter,
					SRM(DIMENSION(1),DIMENSION(1)))
	end_for;
	detaut1:=dom::weight_to_det(dom::BooleanMinimization(automaton1));
	detaut2:=dom::weight_to_det(dom::BooleanMinimization(automaton2));
	if detaut1[1]<>detaut2[1] then
	    return(FALSE)
	else
	    sa1:={extop(ALPHABET(1))} union {extop(ALPHABET(2))}; 
	    iso[detaut1[3]]:=detaut2[3];
	    stlist:=[detaut1[3][1]];		
	    while stlist<>[] do
		for letter in sa1 do
		    state1:=detaut1[4][stlist[1],letter][1];
		    state2:=detaut2[4][stlist[1],letter][1];
		    if testtype(iso[state1],DOM_INT) then
			if iso[state1]<>state2 then
			    return(FALSE)
			end_if
		    else
			iso[state1]:=state2;
			stlist:=append(stlist,state2)
		    end_if
		end_for;
		delete stlist[1]
	    end_while;
	    return(TRUE)
	end_if
    end_proc;

begin
SRM:=Dom::Matrix(SR)
end_domain:

unalias(DIMENSION,ALPHABET,INPUT,TRANSITION,OUTPUT):
