/*****************************************************************************
  $Id: HeckeSymmetricGroupAlgebra.mu,v 1.1.1.5 2005/01/26 13:42:23 nthiery Exp $
  File:         lib/EXPERIMENTAL/HeckeClifford/Hecke0.mu
  Authors:      Nicolas M. Thiry <nthiery@users.sf.net>
  Created:      05/06/2003
  License:      LGPL
*****************************************************************************/

/*

Let V be the vector space spanned by permutations. Consider the
elementary transpositions \sigma_i seen as operators in GL(V) by
acting on the positions. Similarwise, consider the operators $\pi_i$
of the Hecke algebra that act on V by sorting between positions i and
i+1.

This file implements the algebra A generated by the operators
\sigma_i's and \pi_i's together, with respect to composition. Here is
a description of the relations between the \sigma_i's and \pi_i's

\pi1 \sigma1 = \pi1

(sigma1 + id) pi1 = sigma1 + id

Note: $\pi_i(1+s_i)=2\pi_i$.

sigma1 pi1 = sigma1 - pi1 + id
sigma1 pi2 = pi_12  + pi_21  sigma 21  - pi_212 // Using reduced word notation
sigma1 pi2 = pi_231 + pi_312 sigma 312 - pi_321 // Using permutation notation
sigma2 pi1 = pi_312 + pi_231 sigma 231 - pi_321 // Using permutation notation

Plus "braid relations": s_1 s_2 p_1 = p_2 s_1 s_2 (or s_1 p_2 s_1 = s_2 p_1 s_2)

sigma 21


Conjecture: From those relations, this algebra is spanned by the
products \pi_u \sigma_v, where u and v are permutations such that the
descents of u are disjoint from the reculs of v. Those products
actually form a basis of A.

*/

/*
   This does not work when operators has been loaded before the Combinat
   package because plus2, ... are not in the interface of operators.
   Anyway, none of those aliases is used below.

export(operators, Alias,
       plus2, mult2, makeLinear):
*/
export(operators::overloaded, Alias,
       declareDomain, declareConversion, declareSignature):
export(combinat, Alias,
       generators, permutations):
alias(rankers = combinat::rankers): // not yet in the interface of combinat
export(combinat::permutations, Alias,
       reducedWord, fromReducedWord):

alias(HSGA=experimental::HeckeSymmetricGroupAlgebra):
alias(HSGAT=experimental::HeckeSymmetricGroupAlgebraTools):

//////////////////////////////////////////////////////////////////////////////
domain HSGAT
    info_str := "Helper tools for the domain experimental::HeckeSymmetricGroupAlgebra";

    interface := {hold(makePi), hold(makePih), hold(makeTranspo),

		  hold(getIndeterminates),
		  hold(genericElement),
		  hold(genericElement2Basis),
		  hold(commutant),
		  hold(commutantRem),
		  hold(commutantByEigenvalue),
		  hold(commutantByEigenvalueRem),
		  hold(colinearEquations),
		  hold(colinearEquationsRandom),
		  hold(minimalIdempotents),
		  hold(selectMinimalIdempotents),
		  hold(isMinimalIdempotent),
		  hold(subsVariables),
		  hold(changeRootsSquareDifferences),
		  hold(SubalgebraFromRowReducedBasis),
		  hold(powersums2poly),
		  hold(poly2powersums),
		  hold(annotate),
		  hold(hamiltonian),
		  hold(sortedRoots)
		 };
    makePi :=
    proc(i)
        option escape;
    begin
	proc(l: combinat::words) : combinat::words
	begin
	    if l[i]<l[i+1] then
		subsop(l, i=l[i+1], (i+1)=l[i]);
	    end_if;
	    l;
	end_proc:
    end_proc;

    makePih :=
    proc(i)
	option escape;
    begin
	proc(l: combinat::words) : combinat::words
	begin
	    if l[i]>l[i+1] then
		subsop(l, i=l[i+1], (i+1)=l[i]);
	    end_if;
	    l;
	end_proc:
    end_proc;

    makeTranspo :=
    proc(i)
	option escape;
    begin
	proc(l: combinat::words) : combinat::words
	begin
	    subsop(l, i=l[i+1], (i+1)=l[i]);
	end_proc:
    end_proc;

    /**
     * P is a polynomial with real solutions which are sufficiently
     * distinct (i.e. their floating point approximations are distinct)
     * and which are not given as RootOf's
     * Returns the sorted list of the roots of P
     **/
    sortedRoots :=
    proc(P) : Type::ListOf(Type::AlgebraicConstant)
    begin
	sort([op(map(solve(P), op, [1,2]))],(x,y)->_less(float(x),float(y)));
    end_proc;


    /**
     * Returns the list of the values of the symmetric powersums
     * p_{1}..p_{maxd} of the roots of the polynomial p.
     * By default maxd is the degree of p.
     * Algorithm: standard recursive application of Newtowns identities.
    **/
    poly2powersums :=
    proc(p, maxd=FAIL : Type::Union(DOM_FAIL, Type::NonNegInt)) : DOM_LIST
	local d, i, k, result;
    begin
	d := degree(p);
	// make p a monic polynomial (doesn't work on a ring!)
	p := multcoeffs(p, 1/lcoeff(p));
	if args(0)=1 then maxd:=d; end;
	result := [NIL $ maxd];
	for k from 1 to maxd do
	    result[k] := -_plus(if k<=d then k*coeff(p, d-k) else null() end,
				result[i]*coeff(p,d-(k-i))
				$ i=max(1,k-d)..k-1):
	end;
	result
    end;

    /**
     * Returns a monic polynomial p of degree dsuch that Pk[1]..Pk[d]
     * are the the symmetric powersums of its roots
     * Algorithm: standard recursive application of Newtowns identities.
    **/
    powersums2poly :=
    proc(Pk: DOM_LIST, P=polylib::Poly([x]): DOM_DOMAIN)
	local coefficients, l, d, i;
    begin
	d := nops(Pk);
	coefficients := table(d=1);
	for l from d-1 downto 0 do
	    coefficients[l] := - 1/(d-l) *
			     _plus(Pk[i-l]*coefficients[i] $ i=l+1..d):
	end;
	P([[coefficients[i],i] $ i=0..d]);
    end;

    /**
     * Returns a polynomial whose roots are the square of the differences
     * of the roots of the univariate polynomial p.
     *
     * Decompose p as a*\prod_i (X-x_i) in some algebraic extension of
     * the field which contains all its roots. Then, the result is
     * the expansion of P=\prod_{i<j} (X- (x_i-x_j)^2).
     **/
    changeRootsSquareDifferences :=
    proc(p: DOM_POLY) : DOM_POLY
	local SymF, dp, dP, Pol, X, k, t, l, pkp, pkP;
    begin
	SymF := examples::SymmetricFunctions(polylib::coeffRing(p));
	dp  := degree(p):
	dP  := binomial(degree(p),2); // degree of P
	Pol := polylib::constructor(p);
        X   := polylib::variables(p);
	if nops(X) <> 1 then
	    error("univariate polynomial expected, got ". expr2text(p)):
	end_if;
	X := X[1];
	// Computes the value of the symmetric powersums in the roots of p
	pkp := dom::poly2powersums(p, 2*dP);
	// Procedure that computes the value the the k-th symmetric powersum
	// in the roots of the result, using the formula:
        // pP_k = n p_(2k) + 1/2 \sum_{l=1}^2k (-1)^l \binom{2k}{l} p(2k-l,l)
        // where $n$ is the number of roots of p.
	pkP :=
	proc(k: Type::NonNegInt)
	    option remember;
	begin
	    _plus(dp*pkp[2*k],
		  // optimization by using the symmetry 2*k-l <-> l
		  (-1)^l * binomial(2*k,l) * pkp[2*k-l] * pkp[l]
		  $ l=1..k-1,
		  (-1)^k * binomial(2*k,k) * pkp[k]^2/2
		 );
	end_proc;
	_plus(multcoeffs(Pol(X^k),
			 (-1)^(dP-k) *
			 _plus(_mult(t[1],
				     pkP(k) $ k in t[2])
			       $ t in SymF::p::toList(SymF::p(SymF::e([dP-k])))))
	      $ k=0..dP);
    end_proc;

    /**
    First correct but ****slow**** implementation of
    changeRootsSquareDifferencesPk2m
    Left here just in case some stuff can be reused.
    **/

    changeRootsSquareDifferencesPk2m :=
    // Express the k-th symmetric powersums in the roots of P in terms
    // of the monomial symmetric functions in the roots of p
    proc(S: DOM_DOMAIN, k: Type::NonNegInt, d: Type::NonNegInt)
	local result, l;
    begin
	userinfo(5, "Computing the value of P".k." in the m"):
	result := S::m::plus(max(0,(d-1))             * S::m([2*k    ]),
		   (-1)^l * binomial(2*k,l) * S::m([2*k-l,l])  $ l = 1 .. k);
	userinfo(5, "Done: ". expr2text(result)):
	result;
    end;

    changeRootsSquareDifferencesPk2e :=
    // Express the k-th symmetric powersums in the roots of P in terms
    // of the elementary symmetric functions in the roots of p
    proc(S: DOM_DOMAIN, k: Type::NonNegInt, d: Type::NonNegInt)
	option remember;
	local result;
    begin
	// Converts the result of Pk2m into the e basis, and remember the result
	result := S::e(dom::changeRootsSquareDifferencesPk2m(args()));
	userinfo(5, "Done: ". expr2text(result)):
	result;
    end;

    changeRootsSquareDifferencesPk2ereduced :=
    // Express the k-th symmetric powersums in the roots of P in terms
    // of the elementary symmetric functions in the roots of p
    proc(S: DOM_DOMAIN, k: Type::NonNegInt, d: Type::NonNegInt)
	option remember;
    begin
	// Converts the result of Pk2m into the e basis, and
	// kill the terms that contain e_i for i>d
	S::e::fromList
	(select(S::e::toList(dom::changeRootsSquareDifferencesPk2e(args())),
		t -> _and(i<=d $ i in t[2])));
    end;

    changeRootsSquareDifferencesPk2value :=
    // Express the k-th symmetric powersums in the roots of P in terms
    // of the elementary symmetric functions of the roots of p
    proc(S: DOM_DOMAIN, k: Type::NonNegInt, elems: DOM_TABLE)
	option remember;
    begin
	userinfo(5, "Computing the value of P".k):
	_plus(_mult(t[1],
		    if contains(elems, i) then elems[i] else 0 end_if
		    $ i in t[2])
	      $ t in S::e::toList(dom::changeRootsSquareDifferencesPk2e
				  (S, k, nops(elems)-1)))
    end;

    // S: dommain for symmetric functions
    // SF: symmetric function in (S::e or other base) in the roots of p; 
    // d: degre of p
    // Express SF in terms of the elementary symmetric
    // functions of the roots of p as given in elems
    changeRootsSquareDifferencesE2e :=
    proc(S: DOM_DOMAIN, SF, d: Type::NonNegInt)
	local k, t;
    begin
	S::e::plus(t[1] *
		   S::e::mult(dom::changeRootsSquareDifferencesPk2e(S, k, d)
			      $ k in t[2])
		   $ t in poly2list(S::p(SF)));
    end;

    changeRootsSquareDifferencesE2value :=
    // S: dommain for symmetric functions
    // SF: symmetric function in (S::e or other base) in the roots of p; 
    // d: degre of p
    // Express SF in terms of the values of the elementary symmetric functions
    // of the roots of p as given by elems
    proc(S: DOM_DOMAIN, SF, elems: DOM_TABLE)
	option remember;
    begin
	_plus(t[1] *
	      _mult(dom::changeRootsSquareDifferencesPk2value(S, k, elems)
		    $ k in t[2])
	      $ t in poly2list(S::p(SF)));
    end;


    changeRootsSquareDifferencesOld :=
    proc(p: DOM_POLY) : DOM_POLY
	local result, SymF, dp, dP, Pol, X, k, t;
    begin
	// make p a monic polynomial (doesn't work on a ring!)
	p := multcoeffs(p, 1/lcoeff(p)): 
	SymF := examples::SymmetricFunctions(polylib::coeffRing(p));
	dp  := degree(p):
	dP  := binomial(degree(p),2); // degree of P
	Pol := polylib::constructor(p);
	X   := polylib::variables(p);
	if nops(X) <> 1 then
	    error("univariate polynomial expected, got ". expr2text(p)):
	end_if;
	X := X[1];
	elems := table( i = (-1)^i * coeff(p, dp-i) $ i=0..dp );
	_plus(multcoeffs(Pol(X^k),
			 (-1)^(dP-k) *
			 _plus(dom::changeRootsSquareDifferencesE2value
			       (SymF, SymF::e([dP-k]), elems)))
	      $ k=0..dP);
    end_proc;

    /**
     * Returns the hamiltonian, as an element of the
     * algebra of the symmetric group
     **/
    hamiltonian :=
    proc(H: DOM_DOMAIN, coeffs)
	local i, j, K, roots;
    begin
	K := H::coeffRing;
	if args(0)=1 then
	    roots := dom::sortedRoots(poly(orthpoly::hermite(H::N, hold(x)), [hold(x)], K));
	    coeffs := table((i,j) = 1/(roots[i]-roots[j])^2
				  $ i=1..j-1 $ j=1..H::N);
	    if H::coeffRing=Dom::ExpressionField() then
		coeffs := map(coeffs, radsimp@radsimp);
	    end_if;
	end_if;
	_plus(H::monomial(coeffs[i,j],
			  permutations::fromCycles([[i,j]], H::N))
	      $ i=1..j-1 $ j=1..H::N);
    end_proc;

    /**
     * Returns the list of the values of the remember table of the procedure f
     **/
    valuesOfRememberTable :=
    proc(f: DOM_PROC)
    begin
	map([op(extop(f,5))], op, 2):
    end_proc;

    /**
     * Here we call generic element a list containing an object together
     * with a list of its parameters
     **/

    typeGenericElement := Type::ListProduct(Type::AnyType, DOM_LIST);

    /**
     * Utility to build a generic element out of various things
     **/
    genericElement :=
    proc(basis, makeUnknown) : dom::typeGenericElement
	local makeUnknownRemember, gen, i;
    begin
	if testtype(basis, dom::typeGenericElement) then
	    return(basis);
	end_if;
	if args(0)=1 then
	    makeUnknown := ()->hold(a)[args()];
	end_if;
	makeUnknownRemember :=
	proc()
	    option remember;
	begin
	    makeUnknown(args());
	end_proc:
	if domtype(basis) = DOM_DOMAIN then
	    gen := basis::genericElement(makeUnknownRemember);
	elif domtype(basis) = DOM_LIST then
	    gen := _plus(basis[i]*makeUnknownRemember(i) $ i = 1..nops(basis));
	else
	    error("Incorrect parameters");
	end_if;
	[gen, dom::valuesOfRememberTable(makeUnknownRemember)];
    end_proc;

    genericElement2Basis :=
    proc(f: dom::typeGenericElement) : DOM_LIST
	local substitution, x;
    begin
	substitution := map(f[2], _equal, 0);
	[ mapcoeffs(f[1], eval@subs, x=1, substitution)
	  $ x in f[2] ];
    end_proc;

    /**
     * Apply a substitution of some of the variables of the polynomial p
     * by polynomials of the same domain
     *
     * If p is a polynomial expression, use the standard subs instead,
     * and evaluate the result
     **/
    subsVariables :=
    proc(p, substitution: Type::ListOf(Type::Equation(Type::Indeterminate)))
	local P, t, variables, K;
    begin
	if p::dom::variables = FAIL then
	    return(eval(subs(p, args(2..args(0)))));
	end_if;
	t := table(op(substitution));
	variables := polylib::variables(p);
	K := polylib::constructor(p);
	p(op(map(variables,
		 v -> if contains(t,v) then
			  t[v];
		      else
			  P(v);
		      end_if)));
    end_proc;

    /**
     * f should be a generic element of some affine space
     * Find the commutant of h in this affine space
    **/
    commutantGeneric :=
    proc(h, generic: dom::typeGenericElement) : dom::typeGenericElement
	local f, indeterminates, equations, result, parameters, eq;
    begin
	[f, indeterminates] := generic;
	equations := {coeff(f*h-h*f)};
	result   := op(solve(equations, indeterminates), 1);
	parameters := select(indeterminates, not _in,  {op(eq, 1) $ eq in result} );
	[ mapcoeffs(f, dom::subsVariables, result), parameters ];
    end_proc;

    /**
     * Computes a basis of the commutant of h in the vector space
     * spanned by basis.
     * If a and b are specified, then one actually looks for elements
     * f such that h * f = a * f * (h+b). By default, a=1 and b=0.
     *
     * Application: find operators that send eigenspaces of h:
     * to other eigenspaces for related eigenvalues:
     * If h(x) = lambda x, then h(f(x)) = (a lambda + mu) f(x).
    **/
    commutant :=
    proc(h, basis: DOM_LIST, lambda=FAIL, mu=FAIL) : DOM_LIST
	local M, M2, b, i, kernel, v, w;
    begin
	if lambda=FAIL then
	    lambda := h::dom::one;
	end_if;
	if mu=FAIL then
	    mu := h::dom::zero;
	end_if;
	// Build the matrix
	M := _concat(h::dom::sparseVector(h * b - lambda * b * h) $ b in basis);
	M2 := M;
        kernel := linalg::nullspace(M);
	// Convert back the kernel elements in their original domain
	assert(iszero(w));
	[ _plus(basis[i] * v[i] $ i=1..nops(basis))
	  $ v in kernel ];
    end_proc;

    commutantRem :=
    operators::persistentRemember(dom::commutant);

    /**
     * f should be a generic element of some affine space V
     * Find the elements f of V such that h*f = f*h = lambda f
    **/
    commutantByEigenvalueGeneric :=
    proc(h, generic: dom::typeGenericElement, lambda) : dom::typeGenericElement
	local f, indeterminates, equations, result, parameters, eq, fh, b, i, v;
    begin
	[f, indeterminates] := generic;
	fh := f*h;
	equations := {coeff(fh-h*f), coeff(fh-f*lambda)};
	result   := linsolve(equations, indeterminates);
	parameters := select(indeterminates, not _in,  {op(eq, 1) $ eq in result} );
	[ mapcoeffs(f, eval@subs, result), parameters ];
    end_proc;

    /**
     * Find the elements in the vector space spanned by basis which
     * commutes with h and is a h eigenvector: for the eigenvalue lambda:
     * h*f = f*h = lambda f
    **/
    commutantByEigenvalue :=
    proc(h, basis: DOM_LIST, lambda) : DOM_LIST
	local M, M1, M2, w, kernel, b, i, v;
    begin
	// Build the matrix for f*(h-lambda) = 0
	M1 := _concat(h::dom::sparseVector(b * (h-lambda)) $ b in basis);
	// Build the matrix for (h-lambda) * f = 0
	M2 := _concat(h::dom::sparseVector((h-lambda) * b) $ b in basis);
	// Stack the two matrices
	M := _concat(M1::dom::stackMatrix(M1,M2));
	// Solve the system
	userinfo(5, "solving linear system with ".linalg::nrows(M)." equations and ".linalg::ncols(M)." unknowns");
        kernel := linalg::nullspace(M);
	// Convert back the kernel elements in their original domain
	assert(iszero(w));
	[ _plus(basis[i] * v[i] $ i=1..nops(basis))
	  $ v in kernel ];
    end_proc;

    commutantByEigenvalueRem :=
    operators::persistentRemember(dom::commutantByEigenvalue);

    /**
     *  x and y should be vectors of some finite dimensional free module
     *  Returns all the determinants of the 2x2 minors of x and y
     *  They are non-zero iff x and y are colinear
     **/
    colinearEquations :=
    proc(x,y)
	local b;
    begin
	[ coeff(x, b[1]) * coeff(y,b[2]) - coeff(x,b[2])*coeff(y,b[1])
	  $ b in combinat::subwords(x::dom::basisIndices(), 2) ]
    end_proc;

    areColinear :=
    proc(x,y)
    begin
	_and(op(map(colinearEquations(x,y), iszero)));
    end_proc;

    /**
     *  x and y should be vectors of some finite dimensional free module
     *  Returns the determinant of n random 2x2 minors of x and y
     **/
    colinearEquationsRandom :=
    proc(x, y, n)
	local basis, r, i, j;
    begin
	basis := x::dom::basisIndices();
	if args(0)=2 then
	    n := nops(basis) 
	end_if;
	r := random(nops(basis));
	[ (i := r()+1; j := r()+1;
	   coeff(x, basis[i]) * coeff(y, basis[j]) 
	 - coeff(x, basis[j]) * coeff(y, basis[i]))
	 $ n ];
    end_proc;

    // Check that x is a dimension 1 idempotent in the algebra
    // spanned linearly by basis
    isMinimalIdempotent :=
    proc(x, basis: DOM_LIST) : DOM_BOOL
	local y;
    begin
	if not iszero(x^2-x) then
	    return(FALSE)
	end_if;
	for y in basis do
	    if not dom::areColinear(x,x*y) then
		return(FALSE);
	    end_if;
	end_for;
	TRUE;
    end;

    /**
     * Returns the list of the minimal idempotents in the subalgebra
     * spanned by the generic element generic.
     *
     * Algorithm:
     * - build the equations from the coefficients of f^2-f, and throw
     * in a few extra equations stating that f*x is colinear to f for
     * any element x of the basis of the algebra.
     * - solve this quadradic system (this involves a Grbner basis calculation)
     * - plug back the result to obtain the idempotents
     * - remove the (hopefuly few) extraneous non minimal idempotents
     **/
    minimalIdempotents :=
    proc(generic: dom::typeGenericElement)
	local idempotents, f, unknowns, basis, equations, x;
	option remember;
    begin
	[f, unknowns] := generic;
	if nops(unknowns)=0 then // algebra of dimension 0
	    return([]);
	end_if;
	basis := dom::genericElement2Basis(generic);
	//setuserinfo(groebner,1):
	equations := {coeff(f^2-f),
		      op(dom::colinearEquationsRandom(f,f*x)) $ x in basis}:
	idempotents := [op(map(map(solve(equations, unknowns), map, radsimp),
			       eq -> mapcoeffs(f, subs, eq)))]:
	dom::selectMinimalIdempotents(idempotents);
    end_proc;

    /**
     * Select out a bunch of idempotents those that are minimal
     * w.r.t. the order x<=y if x*y=x.
     *
     * If the number of minimal idempotents is known in advance
     * it can be provided as second argument `size'; the process
     * stops as soon as `size' independant idempotents are found.
     *
     * This is really a lattice algorithm, where we assume that
     * the lattice contains no other antichain of size `size' than
     * its minimal elements.
     **/
    selectMinimalIdempotents :=
    proc(idempotents: DOM_LIST,
	 size = infinity: Type::Union(Type::NonNegInt,
				      Type::Singleton(infinity))) : DOM_LIST
	local minimalIdempotents, x, y, p;
    begin
	minimalIdempotents := {}:
	for x in idempotents do
	    if iszero(x) or contains(minimalIdempotents, x) then next; end_if;
	    for y in minimalIdempotents do
		p := mapcoeffs(x*y, normal);
		if not iszero(p) then
		    if p <> y then // y is not minimal
			minimalIdempotents := minimalIdempotents minus {y};
		    end_if;
		    if p <> x then // x is not minimal
			x := p;  // we continue with p which is smaller than x
			if contains(minimalIdempotents, x) then
			    break;
			end_if;
		    end_if;
		end_if;
	    end_for;
	    minimalIdempotents := minimalIdempotents union {x};
	    if nops(minimalIdempotents) = size then
		break;
	    end_if;
	end_for;
	minimalIdempotents;
    end_proc;

    // given an expression f, returns the list of all the symbolic calls
    // of the form a(bla) that it contains
    getIndeterminates :=
    proc(e, a: DOM_IDENT)
	local f, eq;
    begin
	f := proc() option remember; begin 1 end_proc;
	eval(subs(e, a=f));
	a(op(eq,1)) $ eq in extop(f,5);
    end;

    // Annotate a matrix with row and column indexes
    annotate :=
    proc(m, rowIndex: DOM_LIST, colIndex: DOM_LIST, title=`\`) : DOM_ARRAY
	local r, c, i, j;
    begin
	r := nops(rowIndex);
	c := nops(colIndex);
	array(0..r,0..c,
	      (0,0) = title,
	      (i,j) = m[i,j] $i=1..r $ j=1..c,
	      (i,0) = rowIndex[i] $i=1..r,
	      (0,j) = colIndex[j] $j=1..c)
    end_proc;

end_domain:

domain HSGAT::permutationsOf(n: Type::NonNegInt)
    inherits combinat::subClass(permutations, Parameters=n);
    category Cat::CombinatorialClass;
    axiom    Ax::systemRep;

    one := [$1..n];
end_domain:


/**
 * Let A be a finite dimensional combinatorial algebra
 * Let B be a subalgebra of A
 * Let basis be a vector space basis of B in row echelon form with
 * respect to the term order
 *
 * This implements the subalgebra spanned by B in A
 *
 * TODO: merge with Dom::SubFreeModule from Florent
 **/
domain HSGAT::SubalgebraFromRowReducedBasis(basisOfB: DOM_LIST)
    local Basis, A;
    inherits Dom::FreeModule(A::coeffRing, Basis);
    category Cat::FiniteDimensionalAlgebraWithBasis(A::coeffRing);

    ranker := rankers::fromList(dom::basisIndices::list());
    rankBasis := dom::ranker[1];
    unrankBasis := dom::ranker[2];

    one := toBeDefined;

    fromA :=
    proc(x: A) : dom
	local b;
    begin
	dom::fromList([[coeff(x, b),b] $ b in dom::basisIndices()]);
    end_proc;

    toABasis :=
    proc(s: dom::basisIndices)
    begin
	basisOfB[dom::rankBasis(s)];
    end_proc;

    toA :=
    operators::makeLinear(dom::toABasis,
			  Source=dom, ImageSet=A);

    one := dom::fromA(A::one);

    multBasis :=
    proc() : dom
	// The computations in the big algebra will typically be expensive
	option remember;
	local i;
    begin
	dom::fromA(A::mult(dom::toABasis(args(i)) $ i=1..args(0)));
    end_proc;

begin
    if nops(basisOfB)=0 then
	error("An empty basis can't span a subalgebra!")
    end_if;
    Basis := combinat::finiteClass
	   (map(basisOfB, x -> x::dom::lsupport(x)));
    A := (basisOfB[1])::dom;
    Basis::domtype := A::basisIndices::domtype;
end_domain:



//////////////////////////////////////////////////////////////////////////////
domain HSGAT::reducedPairsOfPermutationsOf(n: Type::NonNegInt)
    inherits Dom::BaseDomain;
    category Cat::CombinatorialClass;
    axiom    Ax::systemRep;

    info_str := "The combinatorial class of pairs [p1,p2] of permutations such that the descents of p1 and the reculs of p2 are disjoint";
  
    domtype := DOM_LIST;

    permutationsOfN           := HSGAT::permutationsOf(n);
    typePairOfPermutationsOfN := Type::ListProduct(dom::permutationsOfN $ 2);
    type := Type::Intersection(dom::typePairOfPermutationsOfN,
			       Type::Predicate(dom::isReduced));

    isReduced :=
    proc(l: dom::typePairOfPermutationsOfN)
    begin
	bool(0 = nops({op(permutations::descents(l[1]))} intersect
		      {op(permutations::reculs(l[2]))}));
    end_proc;

    // Brute force generator by selection
    generator :=
    proc()
    begin
	generators::select
	(generators::cartesian(dom::permutationsOfN::generator() $ i=1..2),
	 dom::isReduced)
    end_proc;

    list :=
    proc()
	option remember;
    begin
	generators::toList(dom::generator());
    end_proc;

    /**
      * straighten([pi, sigma])
      *
      * Remove from sigma the reculs that are descents of pi
     **/
    straighten :=
    proc(l: dom::typePairOfPermutationsOfN) : dom
	local pi, sigma, si, i, j;
    begin
	[pi, sigma] := l;
	// This is best seen on the inverse of sigma: if there is a
	// descent \pi[i]>\pi[i+1] in pi, we are allowed to remove a
	// descent at position i in \sigma^-1
	sigma := permutations::inverse(sigma);
	// We do this by bubble sort; that's suboptimal in asymptotic
        // complexity, but we will deal/ mostly with small permutations;
	// so this is probably not worth the trouble.
	for i from n-1 downto 1 do
	    if sigma[i] > sigma[i+1] and pi[i] > pi[i+1] then
		// there is a common descent here!
		si := sigma[i];
		j  := i;
		repeat
		       sigma[j] := sigma[j+1];
		       j := j + 1;
		until not (j<n and si>sigma[j+1] and pi[j] > pi[j+1])
		end_repeat;
		sigma[j] := si;
	    end_if;
	end_for;
	[pi, permutations::inverse(sigma)];
    end_proc;

end_domain:

domain HSGA(N : Type::NonNegInt,
	    R = Dom::ExpressionField())

    inherits Dom::FreeModule(R, HSGAT::reducedPairsOfPermutationsOf(N));
    category Cat::FiniteDimensionalAlgebraWithBasis(R);

    interfaceAdd := { hold(printMatrix), hold(printMatrixAnnotated),
		  hold(fromMatrix), hold(toMatrix), hold(toMatrixAnnotated),
		   hold(perm)."utations", hold(permutationsRanker), hold(permutationsPrint),
		   hold(fromPermutation), hold(straightenBasis),
		   hold(heckeMult2Basis),
		   hold(onListBasis), //hold(onList),
		   hold(matrixDomain),
		   hold(printTermReducedWords),
		   hold(mult2BasisBraid), hold(mult2BasisOriginal), hold(mult2BasisLocal),

		   hold(isHecke), hold(isHeckeEquations),
		   hold(isSG), hold(isSGEquations),
		   hold(isHeckeSG), hold(isHeckeSGEquations)
		 };

    initDomainAdd :=
    proc()
    begin
	operators::overloaded::declareConversion
	(examples::SymmetricGroupAlgebra(N, R), dom,
	 operators::makeLinear
	 (prog::bless
	  (p -> dom::term([[$1..dom::N], p]),
	   dom),
	  Source=examples::SymmetricGroupAlgebra(N, R), ImageSet=dom));
    end_proc;

    // Makes sure that the ranking of basis element (typically when
    // building sparse vectors) is coherent with the order in the basis

    rankerBasis := rankers::fromList(dom::basisIndices());
    rankBasis   := dom::rankerBasis[1];
    unrankBasis := dom::rankerBasis[2];

    pi :=
    proc(i: Type::PosInt) : dom
    begin
	dom::term([permutations::fromReducedWord([i], dom::N), [$1..dom::N]]);
    end_proc;

    ti :=
    proc(i: Type::PosInt) : dom
    begin
	dom::term([[$1..dom::N], permutations::fromReducedWord([i], dom::N)]);
    end_proc;

    fromPermutation := 
    proc(p: permutations)
    begin
	dom::term([[$1..dom::N],p])
    end_proc;

    N := N;

    printTerm :=
    proc(l: dom::basisIndices)
    begin
	_concat(hold(p), op(l[1]), hold(s), op(l[2]));
    end_proc;

    printTermReducedWords :=
    proc(l)
    begin
	hold(p).op(permutations::reducedWord(l[1])) .
	hold(s).op(permutations::reducedWord(l[2]))
    end;

    one := dom::term([ [$1..N] $ 2 ]);

    algebraGenerators :=
    proc() local i;
    begin
	[dom::ti(i) $ i=1..N-1, dom::pi(i)  $ i=1..N-1];
    end_proc;

    straightenBasis :=
    proc(l: dom::basisIndices::typePairOfPermutationsOfN) : dom
    begin
	// We apply the formula \pi_1 \sigma_1 = \pi_1 to remove
	// the reculs of sigma that are descents of pi.
	dom::term(dom::basisIndices::straighten(l));
    end_proc;

	     
    // Multiplication of p_u p_v for u,v two permutations
    heckeMult2Basis :=
    proc(p1: combinat::permutations, p2: combinat::permutations)
	: combinat::permutations
	local i;
    begin
	for i in permutations::reducedWord(p2) do
	    if p1[i] < p1[i+1] then
		p1 := subsop(p1, i=p1[i+1], (i+1)=p1[i]);
	    end_if;
	end_for;
	p1;
    end_proc;

//    mult2Basis := dom::mult2BasisOriginal;
    mult2Basis := dom::mult2BasisBraid;

    mult2BasisOriginal :=
    proc(l1: dom::basisIndices, l2: dom::basisIndices) : dom
	local u, v, u1, v1, i, j, ti, tj, tij, tji, tjij, s,
	      udesc, vrec;
	option remember;
    begin
	s := dom::basisIndices::straighten;
	u := l1[2]; v := l2[1];
	if u = dom::basisIndices::permutationsOfN::one then
	    return(dom::straightenBasis([dom::heckeMult2Basis(l1[1],l2[1]),
					 l2[2]]));
	elif v = dom::basisIndices::permutationsOfN::one then
	    return(dom::straightenBasis([l1[1],
					 permutations::_mult(l1[2],l2[2])]));
	end_if;
	// Take some i such that u1 := u (i,i+1) < u
	// Take some j such that v1 := (j,j+1) v < v
	// Heuristic: try to select i and j as distant as possible
	udesc := permutations::descents(u);
	vrec  := permutations::reculs(v);
	if vrec[1] + 1 < udesc[nops(udesc)] then
	    i := udesc[nops(udesc)];
	    j := vrec[1];
	elif udesc[1] + 1 < vrec[nops(vrec)] then
	    i := udesc[1];
	    j := vrec[nops(vrec)];
	elif nops({ op(udesc) } intersect { op(vrec) }) > 0 then
	    i := j := op({ op(udesc) } intersect { op(vrec) }, 1);
	else
	    i := udesc[1];
	    j := vrec[1];
	end_if;
	assert(contains(udesc,i)>0 and contains(vrec,j)>0);
	//i  := permutations::descents(u)[1];
	//j  := permutations::reculs  (v)[1];
	ti := permutations::fromReducedWord([i], nops(u));
	u1 := permutations::fromReducedWord([i], u);
	tj := permutations::fromReducedWord([j], nops(v));
	v1 := permutations::_mult(tj, v);
	if i<j-1 or i>j+1 then
	    // commutation s_i pi_j --> pi_j s_i
	    dom::mult2Basis([l1[1], u1], [tj, ti]) *
	    dom::term([v1, l2[2]]);
	elif i=j then
	    // commutation si pi_j = s_i - pi_i + id
	      dom::mult2Basis([l1[1], u ], [v1, l2[2]]) //   s_i
	    - dom::mult2Basis([l1[1], u1], [v,  l2[2]]) // - pi_i
	    + dom::mult2Basis([l1[1], u1], [v1, l2[2]]) // + id
	else
	    // sigma_i pi_j = pi_ij + pi_ji sigma_ji - pi_jij
	    tij  := permutations::fromReducedWord([i,j],   nops(u));
	    tji  := permutations::fromReducedWord([j,i],   nops(u));
	    tjij := permutations::fromReducedWord([j,i,j], nops(u));
	      dom::mult2Basis([l1[1], u1],			 //  pi_ij
			      s([dom::heckeMult2Basis(tij, v1), l2[2]]))
	    - dom::mult2Basis([l1[1], u1],                       // -pi_jij
			      s([dom::heckeMult2Basis(tjij, v1), l2[2]]))
	    + dom::mult2Basis([l1[1], u1], [tji, tji]) *	 // pi_ji s_ji
	      dom::term([v1, l2[2]])
	end_if;
    end_proc;

    mult2BasisBraid := // Version using the commutation s_1 s_2 p_1 = p_2 s_1 s_2
    proc(l1: dom::basisIndices, l2: dom::basisIndices) : dom
	option remember;
	local v, vred;
    begin
	if l1[2] = [ $1..N ] then
	    return(dom::straightenBasis([dom::heckeMult2Basis(l1[1],l2[1]),
					 l2[2]]));
	end_if;
	v := l2[1]; vred := reducedWord(v);
	if nops(vred) = 0  then
	    return(dom::straightenBasis([l1[1],
					 permutations::_mult(l1[2],l2[2])]));
	end_if;

	// l2 starts with pi_j; we compute (l1 * pi_j) * restOfL2
	  dom::mult2BasisBraidPi(l1, vred[1]) 
	* dom::term([fromReducedWord([op(vred, 2..nops(vred))], N), l2[2]]);
    end_proc;

    // Multiplication on the right of an element of the basis by pi_j
    mult2BasisBraidPi :=
    proc(l1, j)
	local u, u1, u3, i, tij, tji, tjij,
	      ured, term, sterm, k;
	option remember;
    begin
	term  := dom::term;
	sterm := dom::straightenBasis;
	u := l1[2]; ured := reducedWord(u);
	for i from nops(ured) downto 1 do
	    // Invariant: the value of s_{u_1...u_i} pi_j u_{i+1...}
	    // Note that j may change, and i may decrease by 2 at once
	    if ured[i] < j-1 or ured[i] > j+1 then
	        // commutation s_i pi_j --> pi_j s_i
		next;
	    end_if;
	    if ured[i] = j then
	        // commutation s_i pi_i = s_i + id - pi_i 
		u1 := fromReducedWord([ured[k] $k=1..i-1], N);
		u3 := fromReducedWord([ured[k] $k=i+1..nops(ured)], N);
		return(   term([l1[1], u ]) // s_i
		       + sterm([l1[1], permutations::mult2(u1,u3)]) // + id
		       -  term([l1[1], u1]) * sterm([fromReducedWord([j], N), u3]) // - pi_j
		      );
	    end_if;
	    // ured[i]=j+1 or ured[i]=j-1
	    // Check whether s_j can be factored out on the right of s_{ured[1]...ured[i-1]}
	    u1 := reducedWord(fromReducedWord([ured[k] $ k=1..i-1,j]));
	    assert(nops(u1) = i-2 or nops(u1) = i);
	    if nops(u1) = i-2 then
		// Yes we can; let's do it:
		for k from 1 to i-2 do 
		    ured[k] := u1[k];
		end_for;
		ured[i-1] := j;
		// Now, we can apply the commutation s_j s_{j+1} p_j = p_{j+1} s_j s_{j+1}
		j := ured[i]; // p_j is transformed into p_{j+1} or p_{j-1}
		i := i-1;  // we pass through two letters at once
		next;
	    end_if;
	    // We have no choice but to apply the remaining large commutation rule
	    // sigma_i pi_j = pi_ij + pi_ji sigma_ji - pi_jij
	    u1   := fromReducedWord([ured[k] $k=1..i-1], N);
	    u3   := fromReducedWord([ured[k] $k=i+1..nops(ured)], N);
	    tij  := fromReducedWord([ured[i],j],   N);
	    tji  := fromReducedWord([j,ured[i]],   N);
	    tjij := fromReducedWord([j,ured[i],j], N);

	    return(   term([l1[1], u1])
		   * (sterm([tij,  u3])
		      + sterm([tji,  permutations::mult2(tji,u3)])
		      - sterm([tjij, u3])
		     ));
	end_for;
	sterm([dom::heckeMult2Basis(l1[1], fromReducedWord([j], N)), u]);
    end_proc;

    straightenSigmaPi :=
    proc(u: permutations, v: permutations) : dom
	local u1, v1, i, j, ti, tj, tij, tji, tjij, bla1, bla2, bla3, s, udesc, v2, vrec;
	option remember;
    begin
	if u = dom::basisIndices::permutationsOfN::one or v = dom::basisIndices::permutationsOfN::one then
	    return(dom::straightenBasis([v,u]));
	end_if;
	s := dom::basisIndices::straighten;
	// Take some i such that u1 := u (i,i+1) < u
	// Take some j such that v1 := (j,j+1) v < v
	// Heuristic: try to select i and j as distant as possible
	udesc := permutations::descents(u);
	vrec  := permutations::reculs(v);
	if vrec[1] + 1 < udesc[nops(udesc)] then
	    i := udesc[nops(udesc)];
	    j := vrec[1];
	elif udesc[1] + 1 < vrec[nops(vrec)] then
	    i := udesc[1];
	    j := vrec[nops(vrec)];
	elif nops({ op(udesc) } intersect { op(vrec) }) > 0 then
	    i := j := op({ op(udesc) } intersect { op(vrec) }, 1);
	else
	    i := udesc[1];
	    j := vrec[1];
	end_if;
	assert(contains(udesc,i)>0 and contains(vrec,j)>0);
	//i  := permutations::descents(u)[1];
	//j  := permutations::reculs  (v)[1];
	ti := permutations::fromReducedWord([i], nops(u));
	u1 := permutations::fromReducedWord([i], u);
	tj := permutations::fromReducedWord([j], nops(v));
	v1 := permutations::_mult(tj, v);
	if i<j-1 or i>j+1 then
	    // commutation s_i pi_j --> pi_j s_i
	    dom::straightenSigmaPi(u1,tj) * dom::straightenSigmaPi(ti, v1);
	elif i=j then
	    // commutation si pi_j = s_i - pi_i + id
	      dom::straightenSigmaPi(u,  v1)
	    - dom::straightenSigmaPi(u1, v )
	    + dom::straightenSigmaPi(u1, v1)
	else
	    // sigma_i pi_j = pi_ij + pi_ji sigma_ji - pi_jij
	    tij  := permutations::fromReducedWord([i,j],   nops(u));
	    tji  := permutations::fromReducedWord([j,i],   nops(u));
	    tjij := permutations::fromReducedWord([j,i,j], nops(u));
	    v2 := dom::heckeMult2Basis(tij, v1);
	    bla1 := dom::straightenSigmaPi(u1, v2);
	    bla2 := dom::straightenSigmaPi(u1, dom::heckeMult2Basis(tjij, v1));
	    //bla3 := dom::straightenSigmaPi(u1,tji) * dom::straightenSigmaPi(tji, v1);
	    bla3 := (dom::straightenSigmaPi(u1,tji) * dom::term([dom::basisIndices::permutationsOfN::one,tji])) * dom::term([v1, dom::basisIndices::permutationsOfN::one]);
	    bla1-bla2+bla3;
	    /*
	      dom::straightenSigmaPi(u1, dom::heckeMult2Basis(tij, v1))         //   p_i_ij
	    - dom::straightenSigmaPi(u1, dom::heckeMult2Basis(tjij, v1))        // - pi_jij
	    + dom::straightenSigmaPi(u1,tji) * dom::straightenSigmaPi(tji, v1); //   pi_ji s_ji
            */
	end_if;
    end_proc;

    mult2BasisLocal :=
    proc(l1: dom::basisIndices, l2: dom::basisIndices) : dom
	local u, v;
//	option remember;
    begin
	u := l1[2]; v := l2[1];
	if u = dom::basisIndices::permutationsOfN::one then
	    return(dom::straightenBasis([dom::heckeMult2Basis(l1[1],l2[1]),
					 l2[2]]));
	elif v = dom::basisIndices::permutationsOfN::one then
	    return(dom::straightenBasis([l1[1],
					 permutations::_mult(l1[2],l2[2])]));
	end_if;
	dom::term([l1[1], dom::basisIndices::permutationsOfN::one]) *
	dom::straightenSigmaPi(u,v) *
	dom::term([dom::basisIndices::permutationsOfN::one, l2[2]]);
    end_proc;

    onListBasis :=
    proc(x: dom::basisIndices, l: combinat::words) : combinat::words
	local i;
    begin
	if testargs() and nops(l) <> N then
	    error("Incoherent list size");
	end_if;
	l := permutations::_mult(l, permutations::inverse(x[2]));
	for i in revert(permutations::reducedWord(x[1])) do
	    if l[i]<l[i+1] then
		l := subsop(l, i=l[i+1], (i+1)=l[i]);
	    end_if;
	end_for;
	l;
    end_proc;

    signatureBasis :=
    proc(x: dom::basisIndices)
	local l;
    begin
	table(l=dom::onListBasis(x,l) $ l in combinat::permutations(N));
    end_proc;

    matrixDomain := Dom::SquareMatrix(N!,R);

    permutations :=
    proc() : Type::ListOf(permutations)
	option remember;
    begin
	permutations::list(N);
    end_proc;

    permutationsByInversions :=
    proc() : Type::ListOf(permutations)
	option remember;
	local l, p, i;
    begin
	if N=3 then
	    return([[1,2,3],[1,3,2],[2,3,1],[2,1,3],[3,1,2],[3,2,1]]);
	end_if;
	l := [ [] $ binomial(N,2)+1 ];
	for p in permutations(N) do
	    i := permutations::inversionsNumber(p) + 1;
	    l[i] := append(l[i], p);
	end_for;
	_concat(op(l));
    end_proc;

    permutationsPrint := l->_concat(hold(p),op(permutations::reducedWord(l)));

    permutationsRanker :=
    proc()
	option remember;
    begin
	combinat::rankers::fromList(dom::permutations());
    end_proc;

    toMatrixFunction :=
    proc(f: Type::Function)  : dom::matrixDomain
        local i, rank, unrank;
    begin
	[rank, unrank] := dom::permutationsRanker();
	dom::matrixDomain(N!, N!,
			  [(rank(f(unrank(i))), i) = 1
			   $ i=1..N!]):
    end_proc;

    toMatrixBasis :=
    proc(x: dom::basisIndices) : dom::matrixDomain
    begin
	dom::toMatrixFunction(dom::onListBasis @ (p->(x, p)));
    end_proc;

    toMatrix := operators::makeLinear
	      (dom::toMatrixBasis,
	       Source=dom, ImageSet=dom::matrixDomain,
	       %if traperror(multcoeffs(matrix(2,2),2))<>0 then
	           // temporary workaround for MuPAD 3.0.0's bug with multcoeff
		   Multcoeffs = ((m,c) -> m*c)
	       else
		   null();
	       end_if);

    // Annotated matrix of the action on the space of permutations
    toMatrixAnnotated := dom::annotateMatrix @ dom::toMatrix;

    // Annotated matrix of the action on the space of permutations
    annotateMatrix :=
    proc(M: dom::matrixDomain)
	local P;
    begin
	P := map(dom::permutations(), dom::permutationsPrint);
	HSGAT::annotate(M, P, P);
    end_proc;

    matrixBasisLU :=
    proc()
	option remember;
	local t;
    begin
	linalg::factorLU(_concat(dom::matrixDomain::sparseVector(dom::toMatrixBasis(t))
				 $ t in dom::basisIndices::list()));
    end_proc;

    fromMatrix :=
    proc(m: dom::matrixDomain)
	option remember;
	local P, i, j, result, t;
    begin
	result := dom::zero;
	P := dom::permutations();
	for i from 1 to nops(P) do
	    for j from 1 to nops(P) do
		if m[i,j] <> 0 then
		    t := [permutations::inverse(P[i]), P[j]];
		    result := result + m[i,j]*dom::term(t);
		    m := m - m[i,j]*dom::toMatrixBasis(t);
		end_if;
	    end_for;
	end_for;
	result;
    end_proc;

    // Matrix pretty print of the coefficients of an element of h
    printMatrix :=
    proc(x: dom)
	local P, i, j;
    begin
	P:=dom::permutations();
	Dom::Matrix(R)(N!,N!,[(i,j)=coeff(x,[P[i],P[j]]) $i=1..N! $ j=1..N!]);
    end_proc;

    // Matrix pretty print of the coefficients of an element of h
    printMatrixAnnotated :=
    proc(x: dom)
	local P;
    begin
	P := map(dom::permutations(), dom::permutationsPrint);
	HSGAT::annotate(dom::printMatrix(x), P, P, hold(`p \ s`));
    end_proc;

    // The matrix of the left action (on values) of an elementary transposition s_i on permutations
    sLeft := i->dom::toMatrixFunction(p -> subs(p, [i=i+1, i+1=i]));
    // The matrix of the left action (on values) of pi_i on permutations
    pLeft := i->dom::toMatrixFunction(p -> permutations::inverse(permutations::swapDecrease(permutations::inverse(p),i),p));

    // Test if a matrix is the matrix of an operator in the Hecke algebra
    isHeckeEquations :=
    proc(M: dom::matrixDomain) : DOM_SET
	local i;
    begin
	{op(dom::pLeft(i)*M-M*dom::pLeft(i)) $ i =1..N-1}
    end_proc;

    isHecke :=
    proc(M: dom::matrixDomain) : DOM_BOOL
    begin
	bool(dom::isHeckeEquations(M) = {dom::coeffRing::zero});
    end_proc;

    // Test if a matrix is the matrix of an operator in the symmetric group algebra
    isSGEquations :=
    proc(M: dom::matrixDomain) : DOM_SET
	local i;
    begin
	{op(dom::sLeft(i)*M-M*dom::sLeft(i)) $ i =1..N-1}
    end_proc;

    isSG :=
    proc(M: dom::matrixDomain) : DOM_BOOL
    begin
	bool(dom::isSGEquations(M) = {dom::coeffRing::zero});
    end_proc;

    // Test if a matrix is the matrix of an operator in the symmetric group algebra
    isHeckeSGEquations :=
    proc(M: dom::matrixDomain) : DOM_SET
	local i;
    begin
	{op((dom::sLeft(i) + 1) * M * (dom::sLeft(i) - 1)) $ i =1..N-1}
    end_proc;

    isHeckeSG :=
    proc(M: dom::matrixDomain) : DOM_BOOL
    begin
	bool(dom::isHeckeSGEquations(M) = {dom::coeffRing::zero});
    end_proc;

end_domain:

