/*****************************************************************************
  $Id: Steenrod.mu,v 1.1.1.3 2005/01/26 13:42:23 nthiery Exp $
  File:		experimental/Steenrod.mu
  Authors:	Florent Hivert <florent.hivert@univ-mlv.fr>
		Nicolas M. Thiry <nthiery@users.sourceforge.net>
  Created:	06/02/2003
  Purpose:	Research tools around the Steenrod Algebra
  License:	LGPL
*****************************************************************************/

export(combinat, Alias,
       permutations):
export(experimental::tables, Alias,
       keys, values):
alias(lsupport(x) = x::dom::lsupport(x)):

alias(Steen=experimental::SteenrodAlgebra):
alias(SteenDual=experimental::SteenrodAlgebraDual):
alias(SteenT=experimental::SteenrodAlgebraTools):
alias(declareConversion = operators::overloaded::declareConversion):
alias(declareSignature  = operators::overloaded::declareSignature):
alias(scalar = operators::scalar):
alias(coproduct = operators::coproduct):

//////////////////////////////////////////////////////////////////////////////
domain SteenT
    info_str := "Helper tools for the domain experimental::SteenrodAlgebra";
end_domain:

//////////////////////////////////////////////////////////////////////////////
domain SteenT::powersum(R: DOM_DOMAIN, q)
    inherits Dom::FreeModule(R, combinat::partitions);
    category Cat::GradedHopfAlgebraWithBasis(R);

    info_str := "Domain for the q-Steenrod algebra on the powersum basis";

    interfaceAdd := {
		   hold(toWeyl)
		 };

    dual := SteenT::powersumDual(R, q);

    oneBasis := [];
    degreeBasis := _plus@op;

    //////////////////////////////////////////////////////////////////////////
    // Conversions and overloading
    //////////////////////////////////////////////////////////////////////////

    basisName := hold(p);
    exprTerm  := dom::exprTermIndex;

    // Express an element p of the steenrod algebra in the Weyl algebra W
    toWeyl :=
    proc(W: DOM_DOMAIN, p: dom)
	local d, m, x;
    begin
	if W::coeffRing <> dom::coeffRing then
	    error("Not the same coefficient ring!");
	end_if;
	_plus(
	      dom::coeff(p,m) *
	      _mult(W(_plus(x^d*(1+q*x*W::DX(x)) $ x in W::XVariables))
		    $ d in m)
	      $ m in dom::support(p));
    end_proc;

    // Express an element p of the original steenrod algebra in the
    // Weyl algebra W
    toWeylOriginal :=
    proc(W: DOM_DOMAIN, p: dom)
	local d, m, x;
    begin
	_plus(
	      dom::coeff(p,m) *
	      _mult(W(_plus(x^d*(x*W::DX(x)) $ x in W::XVariables))
		    $ d in m)
	      $ m in dom::support(p));
    end_proc;

    toWeylStar :=
    proc(W: DOM_DOMAIN, p: dom)
	local d, m, x;
    begin
	if W::coeffRing <> dom::coeffRing then
	    error("Not the same coefficient ring!");
	end_if;
	_plus(
	      dom::coeff(p, m) *
	      _mult(W(_plus((1+q*x*W::DX(x))*W::DX(x)^d $ x in W::XVariables))
		    $ d in revert(m))
	      $ m in dom::support(p));
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // Commutation rules and product
    //////////////////////////////////////////////////////////////////////////

    straightenBasis :=
    proc(l: combinat::compositions) // Should p[0] be disallowed ?
	local i;
    begin
	for i from 1 to nops(l)-1 do
	    if l[i] >= l[i+1] then 
		next;
	    end_if;
	    return(                dom::straightenBasis([op(l,1..i-1),l[i+1],l[i],op(l,i+2..nops(l))]) +
		   dom::multcoeffs(dom::straightenBasis([op(l,1..i-1),l[i]+l[i+1],op(l,i+2..nops(l))]),
				   q*(l[i+1]-l[i]))
		  );
	end_for;
	dom::term(l);
    end_proc;

    mult2Basis :=
    proc(l1: dom::basisIndices, l2: dom::basisIndices)
    begin
	dom::straightenBasis(l1.l2);
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // Antipode
    //////////////////////////////////////////////////////////////////////////

    antipodeBasis :=
    proc(partition: dom::basisIndices)
    begin
	(-1)^nops(partition)*dom::straightenBasis(revert(partition));
    end_proc;

    // Coproduct defined by multiplicativity from the primitive elements
    coproductBasis :=
    proc(partition: dom::basisIndices) : dom::tensorSquare
	local k;
    begin
	if nops(partition)=0 then
	    dom::tensorSquare::term([[], []]);
	else
	    _mult(dom::tensorSquare::term([[k], []]) +
		  dom::tensorSquare::term([[], [k]])
		  $ k in partition);
	end_if;
    end_proc;
end_domain:

//////////////////////////////////////////////////////////////////////////////
domain SteenT::powersumDual(R: DOM_DOMAIN, q)
    inherits (examples::SymmetricFunctions(R))::powersum;
    category Cat::GradedHopfAlgebraWithBasis(R);

    info_str := "Domain for the dual of the q-Steenrod algebra on the powersum basis";

    oneBasis := [];

    basisName := hold(Dp);

    dual := SteenT::powersum(R, q);

    antipodeBasis := ()->error("not implemented");

    coproductBasis :=
    proc(partition: dom::basisIndices) : dom::tensorSquare
	option remember;
	local c, d, partition1, i, partition2, t;
    begin
	d := _plus(op(partition));
	// Compute all the coproducts of basis elements of degree d by
	// transposing the dual product
	for partition1 in dom::basisIndices::list(d) do
	    dom::coproductBasis(partition1) := dom::tensorSquare::zero;
	end_for;
	// Compute all the products p[partition1]p[partition2] in the dual
	for i from 0 to d div 2 do
	    for partition1 in dom::basisIndices::list(i) do
		for partition2 in dom::basisIndices::list(d-i) do
		    // Renormalization factor due to the fact
		    // that the bases are only orthogonal
		    c :=
		    dom::scalarBasis(partition,partition) /
		    dom::scalarBasis(partition1,partition1) /
		    dom::scalarBasis(partition2,partition2);
		    for t in poly2list(dom::dual::mult2Basis(partition1, partition2)) do
			dom::coproductBasis(op(t,2)) :=
			    dom::coproductBasis(op(t,2)) +
			    dom::tensorSquare::term([partition1,partition2]) *
			    c * op(t,1);
		    end_for;
		    if 2*i < d then
			for t in poly2list(dom::dual::mult2Basis(partition2, partition1)) do
			    dom::coproductBasis(op(t,2)) :=
			        dom::coproductBasis(op(t,2)) +
			        dom::tensorSquare::term([partition2,partition1]) *
			        c * op(t,1);
			end_for;
		    end_if;
		end_for;
	    end_for;
	end_for;
	dom::coproductBasis(partition);
    end_proc;

    scalarBasis :=
    proc(partition1: dom::basisIndices, partition2: dom::dual::basisIndices)
	option remember;
	local t;
    begin
	// Hypothesis:
	// dom is the free commutative algebra generated by Dp[1],Dp[2],...
	// Dp[partition1] belongs to the basis B of dom
	// p[partition2]  belongs to the basis Bdual of dom::dual
	// B and Bdual are orthogonal bases
	// p[] is the one of dom
	if partition1<>partition2 then
	    // The basis are orthogonal
	    return(R::zero);
	end_if;
	if nops(partition1) = 0 then
	    // Dp[] is the one of dom
	    return(R::one);
	end_if;
	if nops(partition1) = 1 then
	    // We have some freedom on the coefficients for the algebra
	    // generators of dom
	    return(hold(f)(op(partition1)));
	end_if;
	// Use the coproduct of Bdual to obtain the coefficients
	// scalar(p[4,2,1],           Bdual(partition2)) =
	// scalar(p[4] tensor p[2,1], coproduct(Bdual(partition2)))
	// = \sum scalar(p1, f)*scalar(p42, g)
	// where the f,g are such that
	// coproduct(Bdual(partition2)) = \sum f \tensor g
	_plus(op(t,1) *
	      dom::scalarBasis([op(partition1,1)],                   op(t,[2,1])) *
	      dom::scalarBasis([op(partition1,2..nops(partition1))], op(t,[2,2]))
	      $ t in poly2list(coproduct(dom::dual(partition2))));
    end_proc;

    initDomainAdd :=
    proc()
    begin
	declareSignature
	(scalar, [dom, dom::dual],
	 operators::makeBilinear	// Should be makeSesquilinear
	 (dom::scalarBasis,
	  Source1=dom, Source2=dom::dual, ImageSet=R));
    end_proc;
end_domain:

//////////////////////////////////////////////////////////////////////////////
domain SteenT::homogeneous(R: DOM_DOMAIN, q)
    inherits Dom::FreeModule(R, combinat::partitions);
    category Cat::AlgebraWithBasis(R);

    info_str := "Domain for the q-Steenrod algebra on the homogeneous basis";

    oneBasis := [];
    degreeBasis := _plus@op;

    //////////////////////////////////////////////////////////////////////////
    // Conversions and overloading
    //////////////////////////////////////////////////////////////////////////

    basisName := hold(h);
    exprTerm  := dom::exprTermIndex;
end_domain:

//////////////////////////////////////////////////////////////////////////////
domain Steen(R=Dom::ExpressionField(): DOM_DOMAIN, q=hold(q))

    inherits Dom::BaseDomain;
    category  Cat::HopfAlgebraWithSeveralBases(R);

    info_str := "Library for the rational Steenrod algebra";

    interfaceAdd:={
		hold(p), hold(powersum)
	       };

    powersum       := SteenT::powersum(R, q);
    p              := dom::powersum;
    Tp             := dom::powersum::tensorSquare;
    homogeneous    := SteenT::homogeneous(R, q);
    h              := dom::homogeneous;

    bases := { dom::homogeneous,
	       dom::powersum
	     };

    dual := SteenDual(args());
    dualBasesPairs := { [dom::powersum, dom::dual::powersumDual] };

    initDomainAdd :=
    proc()
	option remember;
    begin
	declareConversion
	(dom::h, dom::p,
	 operators::makeLinear
	 (prog::bless
	  (proc(part: dom::p::basisIndices)
	   begin
	       if part=[] then
		   dom::p::one;
	       else
		   FAIL;
	       end_if;
	   end_proc, dom),
	  Source=dom::h, ImageSet=dom::p)):
     end_proc;

end_domain:

domain SteenDual(R=Dom::ExpressionField(): DOM_DOMAIN, q=hold(q))

    inherits Dom::BaseDomain;
    category  Cat::HopfAlgebraWithSeveralBases(R);

    info_str := "The dual of the rational Steenrod algebra";

    interfaceAdd:={
		hold(powersumDual)
	       };

    powersumDual   := SteenT::powersumDual(R, q);

    bases := {
	       dom::powersumDual
	     };

end_domain:

/*
reset():
S:=experimental::SteenrodAlgebra(): 
P:=S::p::domainWrapper:
P[1]*P[2];
*/

domain experimental::Steenrod

    interface := {hold(ORExt), hold(OREtx), hold(ORExs),
		  hold(qPk), hold(qPPk),
		  hold(pretty),
		  hold(homogeneousRightIdealBasis),
		  hold(homogeneousRightIdealGeneric),
		  hold(homogeneousMultiples),
		  hold(genericHomogeneousMultiple),
		  hold(genericHigherDerivation),
		  hold(antipode),
		  hold(toVector),
		  hold(truncate), hold(truncateOrder),
		  hold(nclog),  hold(ncexp),
		  hold(valuesOfRememberTable),
		  hold(SteenrodConjugated),
		  hold(irreducibleSteenrodConjugatedCoeff),
		  hold(solveInSteenrodRightIdeal),
		  hold(WeylVariables),
		  hold(YoungSubgroupOrbit),
		  hold(YoungSubgroupWeylOrbitsum),
		  hold(genericWeylElementWithSymmetries),
		  hold(bilan),
		  hold(spair),
		  hold(PolynomialsWithSkewThetaCoefficients),
		  hold(PolynomialsWithSkewQThetaCoefficients)
		 };

    antipode := operators::antipode;

    /**
     * Variables of a Weyl algebra
     *
     * @param W	A Weyl OREtx, or ORExs algebra
     * @return	The list of the variables of the algebra
    **/
    WeylVariables :=
    proc(W)
    begin
	if W::constructor = dom::OREtx then
	    W::DXVariables;
	else
	    W::XVariables;
	end_if;
    end_proc;

    // q-Steenrod powersum
    qPk :=
    proc(W: DOM_DOMAIN, k: Type::NonNegInt, X)
	local x;
    begin
	if args(0)=2 then
	    X:= dom::WeylVariables(W);
	end_if;
	_plus(W(x)^k*(1+hold(q)*W::euler(x)) $ x in X);
    end_proc;

    // 1-Steenrod powersum
    Pk :=
    proc(W: DOM_DOMAIN, k: Type::NonNegInt, X)
	local x;
    begin
	if args(0)=2 then
	    X:= dom::WeylVariables(W);
	end_if;
	_plus(W(x)^k*(1+W::euler(x)) $ x in X);
    end_proc;

    // q-Steenrod powersum conjugated by (1+q*t) P_{k,q} (1+q*t)^{-1}
    qPPk :=
    proc(W: DOM_DOMAIN, k: Type::NonNegInt, X)
	local x;
    begin
	if args(0)=2 then
	    X:= dom::WeylVariables(W);
	end_if;
	_plus((1+hold(q)*W::euler(x))*W(x^k) $ x in X);
    end_proc;

    pretty :=
    proc(p)
	local ind;
    begin
	p:=expr(p):
	ind :=sort([op(select(indets(expr(p)),
			      x->(expr2text(x)[	%if "12"[1]="1" then
						    1
						else
						    0
						end ] in {"x", "y"})))]);
	if nops(ind)=0 then
	    p
	else
	    mapcoeffs(Dom::MultivariatePolynomial(ind, Dom::ExpressionField())(expr(p)), factor);
	end_if;
    end_proc;

    // Precondition: idealGenerators and algebraGenerators are lists
    // of homogeneous elements of some graded algebra; the
    // algebraGenerators are of positive degree.
    // Returns a list of homogeneous vectors that span the homogeneous
    // component of degree d of the ideal generated by <idealGenerators>
    // w.r.t. the ring generated by <algebraGenerators>
    homogeneousRightIdealBasis :=
    proc(idealGenerators:   DOM_LIST,
	 algebraGenerators: DOM_LIST,
	 d: Type::NonNegInt)
	local degrees, degvec, p;
    begin
	idealGenerators := select(idealGenerators, p->degree(p)<=d);
	degrees := map(algebraGenerators, degree):
	[ _mult(p, op(zip(algebraGenerators, degvec, _power)))
	 $ degvec in combinat::integerVectorsWeighted(d-degree(p), degrees)
         $ p in idealGenerators ]
    end_proc;

    // Precondition: idealGenerators and algebraGenerators are lists
    // of homogeneous elements of some graded algebra; the
    // algebraGenerators are of positive degree.
    // Returns a generic element in the homogeneous
    // component of degree d of the ideal generated by <idealGenerators>
    // w.r.t. the ring generated by <algebraGenerators>
    homogeneousRightIdealGeneric :=
    proc(idealGenerators:   DOM_LIST,
	 algebraGenerators: DOM_LIST,
	 d: Type::NonNegInt,
	 variable=(x,y->hold(a)[x,y]))
	local degrees, degvec, i;
    begin
	idealGenerators := select(idealGenerators, p->degree(p)<=d);
	degrees := map(algebraGenerators, degree):
	_plus(variable(i,degvec) * 
	      _mult(idealGenerators[i], op(zip(algebraGenerators, degvec, _power)))
	      $ degvec in combinat::integerVectorsWeighted(d-degree(idealGenerators[i]), degrees)
	      $ i = 1..nops(idealGenerators) )
    end_proc;

    homogeneousMultiples :=
    proc(elements: DOM_LIST,
	 d: Type::NonNegInt,
	 criteria=TRUE)
	local degvecs, degvec;
    begin
	degvecs :=
	select(combinat::integerVectorsWeighted(d, map(elements, degree)),
	       criteria);
	[_mult(op(zip(elements, degvec, _power))) $ degvec in degvecs ]
    end_proc;

    genericHomogeneousMultiple :=
    proc(elements: DOM_LIST, d: Type::NonNegInt,
	 variable=hold(a),
	 criteria=TRUE)
	local degvecs, degvec;
    begin
	degvecs := 
	select(combinat::integerVectorsWeighted(d, map(elements, degree)),
	       criteria);
	[_plus(_mult(variable[op(degvec)], op(zip(elements, degvec, _power)))
	       $ degvec in degvecs),
	 [variable[op(degvec)]
	  $ degvec in degvecs]
	]
    end_proc;

    // Builds a generic higher derivation of degree d in the algebra A
    // which satisfies the coproduct relations with the previous ones
    genericHigherDerivation :=
    proc(A, d)
	option remember;
	local f, gen, equations, solutions, i;
    begin
	if d=0 then return(A::one); end:
	f := p->hold(a)[op(p)]; // coefficient names
	gen := A::genericElement(f, d);
	equations := [coeff(operators::coproduct(gen) -
			    operators::tensor(gen, A::one) -
			    operators::tensor(A::one, gen) -
			    _plus(operators::tensor(eval(procname)(A, i),
						    eval(procname)(A, d-i))
			      $ i=1..d-1))];
	solutions := linsolve(equations, A::genericCoefficients(f, d));
	if solutions = FAIL then
	    return(FAIL);
	else
	    return(mapcoeffs(gen, subs, solutions));
	end_if;
    end_proc;

    // Builds a sparse column vector from a polynomial/free module element
    toVector:=
    proc(p, rank=combinat::rankers::fromType(Type::AnyType)[1])
	local t, eq;
    begin
	t := table(rank(op(t,2))=op(t,1) $ t in poly2list(p));
	Dom::Matrix(polylib::coeffRing(p))(max(0,op(eq,1) $ eq in t),1,t)
    end_proc;

    // Anihilates all the terms in p of degree >= ORDER
    truncate :=
    proc(p, order)
	local t;
    begin
	if args(0)<2 then order := ORDER; end;
	_plus(t $ t in select(polylib::monomials(p), t->degree(t)<order))
    end;

    // Anihilates all the terms in p of order >= ORDER
    truncateOrder :=
    proc(p, order)
	local t;
    begin
	if args(0)<2 then order := ORDER; end;
	_plus(t $ t in select(polylib::monomials(p), t->t::dom::order(t)<order))
    end;

    // Truncated Non Commutative Exponnential function for a graded algebra
    // Truncated Non Commutative Logarithm function for a graded algebra
    ncexp := x->x::dom::expTruncated(args()); // cf. GradedAlgebraWithBasis
    nclog := x->x::dom::logTruncated(args()); // cf. GradedAlgebraWithBasis

    g:=p->mapcoeffs(poly(expr(p),[hold(x3),hold(x2),hold(x1)]), factor);

    /**
     * Returns the list of the values of the remember table of the procedure f
    **/
    valuesOfRememberTable :=
    proc(f: DOM_PROC)
    begin
	map([op(extop(f,5))], op, 2):
    end_proc;

    /**
     * Consider the Young subgroup S_{c_1}\times \dots\times S_{c_k}
     * acting by permutation on lists of length c_1+\dots+c_k
     *
     * Returns the orbit of l, as a list
    **/
    YoungSubgroupOrbit :=
    proc(l: DOM_LIST, c: combinat::compositions) : Type::ListOf(DOM_LIST)
	local r, i;
    begin
	r := 1;
	combinat::generators::toList
	(combinat::generators::map
	 (combinat::generators::cartesian
	  (combinat::permutations::generator([op(l, r..(r:=r+i)-1)])
	   $ i in c),
	  _concat@op))
    end;

    /**
     * Consider the subalgebra A of the Weyl algebra W generated by
     * the variables x, and the corresponding euler operators tx=x
     * d/dx.  Consider furthermore a Young subgroup
     * $G=S_{c_1}\times\dots\times S_{c_k}$ with $c_1+\dots+c_k = n$
     * acting by simultaneous permutations of the x's and tx's.
     *
     * YoungSubgroupWeylOrbitsum: returns an orbitsum of A w.r.t. G
     *
     * genericWeylElementWithSymmetries: returns a generic homogeneous
     * G-invariant element of A with prescribed degree and order.
     *
     * The coefficients are computed via f.
    **/
    YoungSubgroupWeylOrbitsum :=
    proc(W: DOM_DOMAIN, m: combinat::integerMatrices, c)
	local m1;
    begin
	_plus(_mult(op(zip(map(dom::WeylVariables(W), W       ),
			   op(m1,1), _power)),
		    op(zip(map(dom::WeylVariables(W), W::euler),
			   op(m1,2), _power)))
	      $ m1 in map(dom::YoungSubgroupOrbit
			  (combinat::integerMatrices::transpose(m),c),
			  combinat::integerMatrices::transpose))
    end_proc;

    genericWeylElementWithSymmetries :=
    proc(W: DOM_DOMAIN,
	 degree: Type::NonNegInt, order: Type::NonNegInt,
	 c: combinat::compositions,
	 f: Type::AnyType)
	local m;
    begin
	assert(_plus(op(c))=W::dimen);
	_plus(f(m, args(6..args(0))) * dom::YoungSubgroupWeylOrbitsum(W, m, c)
	      $ m in combinat::integerMatrices::fromNonWeakDescentsComposition([degree, order],c))
    end_proc;

    /**
     * Try to find the specializations of the generic element F which lie
     * in the right ideal generated by the Steenrod algebra
     * Note: the order of F should be at least 2 more than order.
     **/
    solveInSteenrodRightIdeal :=
    proc(W, F, unknownsF, order)
	local options, generic, S, P1, P2, G1, G2, G3, makeUnknownsG1, makeUnknownsG2, makeUnknownsG3, constraints, eqns, o, pivot, result, unknowns, vars;
    begin
	vars := dom::WeylVariables(W);
	options := combinat::getOptions({hold(Divisible)=FALSE,hold(Q)=hold(q)}, {}, 4, Error):
	S:=experimental::SteenrodAlgebra(Dom::ExpressionField(),
					 options[hold(Q)]):
	generic:=dom::genericWeylElementWithSymmetries;
	P1 := S::p::toWeyl(W, S::p([1]));
	P2 := S::p::toWeyl(W, S::p([2]));
	makeUnknownsG1 := proc(m) option remember; begin hold(g1)(m) end_proc:
	makeUnknownsG2 := proc(m) option remember; begin hold(g2)(m) end_proc:
	makeUnknownsG3 := proc(m) option remember; begin hold(g3)(m) end_proc:
	if options[hold(Divisible)] then
	    // Assumption: G1 is left-divisible by x
	    G1 :=W(hold(x))*_plus(generic(W, 0,o,[1,nops(vars)-1],makeUnknownsG1)
			    $ o=0..order-1):
	else
	    G1 :=_plus(generic(W, 1, o, [1,nops(vars)-1], makeUnknownsG1)
		       $ o=0..order-1):
	end_if;
	G2 :=_plus(generic(W, 1, o, [1,nops(vars)-1], makeUnknownsG2)
		   $ o=0..order-1):
	G3 :=_plus(generic(W, 2, o, [1,nops(vars)-1], makeUnknownsG3)
		   $ o=0..order-1):
	userinfo(1, "Building equations");
	eqns := map({F*P1 - P1*G1, F*P2 - P2*G2 - P1*G3},
		    coeff@dom::truncateOrder, order):
	userinfo(1, "Building unknowns"):
	unknowns := map([makeUnknownsG1, makeUnknownsG2, makeUnknownsG3],
			op@dom::valuesOfRememberTable). unknownsF:
	userinfo(1, "Solving system with ". nops(eqns)." equations and "
		 . nops(unknowns) . " unknowns"):
	[result, constraints, pivot] := linsolve(eqns,unknowns, ShowAssumptions):
	userinfo(1, "assumptions: ". expr2text(constraints). " ".expr2text(map(pivot, factor@op, 1))):
	select(result, not has, [hold(g1),hold(g2),hold(g3)])
    end_proc;

    /**
     * Returns a coefficient c_m of the steenrod-conjugated operator
     * of the form x_i * (\sum_m c_m t^m)
     * See the notes: this operator is not steenrod-conjugated for
     * more than 4 variables, or kills any variable of degree >2
    **/
    irreducibleSteenrodConjugatedCoeff :=
    proc(m: combinat::integerVectors, i=1: Type::NonNegInt)
	local n, k;
    begin
	n := _plus(op(m));
	(-1)^n / _mult(k! $ k in m) / 2 * (2*op(m,i)*hold(q)+n-1) * (n-2)
    end_proc;

    /**
     * Returns the steenrod-conjugated operator of the form
     * x_i * (\sum_m c_m t^m), truncated up to order order.
     * See the notes: this operator is not steenrod-conjugated for
     * more than 4 variables, or kills any variable of degree >2
     **/
    SteenrodConjugated :=
    proc(W: DOM_DOMAIN, order: Type::NonNegInt, monom=1: Type::Union(Type::NonNegInt, combinat::integerVectors))
	option remember;
	local i, d, m;
    begin
	if domtype(monom)=DOM_INT then
	    i := monom;
	    _plus(dom::irreducibleSteenrodConjugatedCoeff(m, i) *
		  W::term(_concat(zip([0 $ i-1,1,0 $ W::dimen-i],m,_plus),m))
		  $ m in combinat::integerVectors(d, W::dimen)
			 $ d = 0..order-1)
	else
	    dom::truncateOrder(_mult(dom::SteenrodConjugated(W, order, i) ^ monom[i]
				     $ i=1..nops(monom)),
			       order);
	end_if;
    end_proc;

    /**
     * Tools for hand-driven Gr"obner basis computations
     **/
    
    bilan :=
    proc(p)
	local i;
    begin
	for i from 1 to nops(p) do
	    fprint(Unquoted, 0, i, "\t",lterm(p[i]),"\t", lcoeff(p[i]));
	end;
    end_proc;

    spair := (p,q) -> (userinfo(1, "dividing by ". expr2text(numer(lcoeff(p))). " and ".expr2text(numer(lcoeff(q)))): mapcoeffs(p/lcoeff(p)-q/lcoeff(q), factor));

    set :=
    proc(i,j,m)
    begin
	r1[i] := r1[j]*m;
	r2[i] := r2[j]*m;
	p [i] := p [j]*m;
    end_proc;

end_domain:

// Ore-Algebra generated by variables x of degree 1 and corresponding
// "euler" operators tx (typically tx=x d/dx) of degree 0, with
// straightening of monomials to tx x by the commutation rule:
//	x^k tx = (tx-k a) x^k =  tx x^k - a k x^k
domain experimental::Steenrod::OREeulerx(X: DOM_LIST,
					 R = Dom::ExpressionField(),
					 a=1)
    inherits Dom::WeylAlgebra_Table([tx.x $x in X], R);
    category Cat::AlgebraWithBasis(R);

    DX		:= x -> hold(t).x;
    euler	:= x -> dom(dom::DX(x));
    XVariables  := [ dom::DX(x) $ x in X ];
    DXVariables := X;

    straightenOneVariable :=
    // computes the normal form of c t^t1[1] x^t1[2] t^t2[1] x^t2[2]
    // as a sum of monomials c t^t[1] x^t[1] + ..., which is as a list
    // [ t = c, ... ]
    proc(t1: Type::ListProduct(Type::NonNegInt $ 2),
	 t2: Type::ListProduct(Type::NonNegInt $ 2),
	 c = dom::coeffRing::one : dom::coeffRing)
    begin
	if t1[2]=0 or t2[1]=0 then
	    [[t1[1]+t2[1], t1[2]+t2[2]] = c]
	else
	    dom::straightenOneVariable([t1[1]+1, t1[2]], [t2[1]-1, t2[2]], c      ) .
	    dom::straightenOneVariable([t1[1],   t1[2]], [t2[1]-1, t2[2]], -c*a*t1[2]);
	end_if;
    end_proc;

    degreeBasis :=
    proc(x: dom::basisIndices)
    begin
	_plus(op(x,dom::dimen+1..2*dom::dimen));
    end_proc;
end_domain:

// Ore-Algebra generated by variables x and corresponding euler operators
// tx=x d/dx, with straightening of monomials to t*x by the
// commutation rule: x^k t = (t-k) x^k =  t*x^k - k*x^k

domain experimental::Steenrod::OREtx(X: DOM_LIST, R = Dom::ExpressionField())
    inherits experimental::Steenrod::OREeulerx(X, R);
    category Cat::AlgebraWithBasis(R);
end_domain:

// Ore-Algebra generated by variables x of degree 1 and corresponding
// "euler" operators tx (typically tx=x d/dx) of degree 0, with
// straightening of monomials to x*tx by the commutation rule:
//	tx x^k = x^k (tx+k*a) = x^k*tx + a*k*x^k

domain experimental::Steenrod::ORExeuler(X: DOM_LIST,
					 R = Dom::ExpressionField(),
					 a=1)
    inherits Dom::WeylAlgebra_Table(X, R);
    category Cat::AlgebraWithBasis(R);

    DX		:= x -> hold(t).x;
    euler	:= x -> dom(dom::DX(x));
    XVariables  := X;

    straightenOneVariable :=
    proc(t1: Type::ListProduct(Type::NonNegInt $ 2),
	 t2: Type::ListProduct(Type::NonNegInt $ 2),
	 c = dom::coeffRing::one : dom::coeffRing)
    // computes the normal form of a monomial c x^t1[1] s^t1[2] x^t2[1] s^t2[2]
    // as a sum of monomials c x^t[1] s^t[1] + ..., which is as a list
    // [ t = c, ... ]
    begin
	if t1[2]=0 or t2[1]=0 then
	    [[t1[1]+t2[1], t1[2]+t2[2]] = c]
	else
	    dom::straightenOneVariable([t1[1], t1[2]-1], [t2[1], t2[2]+1], c      ) .
	    dom::straightenOneVariable([t1[1], t1[2]-1], [t2[1], t2[2]], +c*a*t2[1]);
	end_if;
    end_proc;

    degreeBasis :=
    proc(t: dom::basisIndices)
    begin
	_plus(op(t,1..dom::dimen))
    end_proc;

    /**
      * Returns a generic element with degree xdegree in the x variables
      * and degree tDegree in the t variables, which is symmetric w.r.t.
      * the permutation action of the Young subgroup
      * $S_{c_1}\times\dots\times S_{c_k}$
      *
      * One should have c_1+\dots+c_k=n
      *
      * The coefficients are computed via f.
    **/
    genericElementWithSymmetries :=
    proc(xDegree: Type::NonNegInt, tDegree: Type::NonNegInt,
	 c: combinat::compositions,
	 f: Type::AnyType)
    begin
	assert(_plus(op(c))=dom::dimen);
	_plus(f(m) *
	      _plus(dom::term(map(m1,op))
		    $ m1 in map(combinat::permutations
				(combinat::integerMatrices::transpose(m)),
				combinat::integerMatrices::transpose))
	      $ m in combinat::integerMatrices::fromNonWeakDescentsComposition([xDegree, tDegree],c))
    end_proc;
end_domain:

// Ore-Algebra generated by variables x and corresponding operators
// sx = 1 + q x d/dx, with straightening of monomials to x sx by the
// commutation rule: sx x^k = x^k (sx + q*k)
domain experimental::Steenrod::ORExs(X: DOM_LIST, R = Dom::ExpressionField())
    inherits experimental::Steenrod::ORExeuler(X, R, hold(q));
    category Cat::AlgebraWithBasis(R);

    DX		:= x -> hold(s).x;
    euler	:= x -> dom(dom::DX(x)-1)/q;

    initDomainAdd :=
    proc()
	option escape;
	local DOM, Otx;
    begin
	DOM := dom;
	Otx := experimental::Steenrod::OREtx(X, R);
	operators::overloaded::declareConversion
	(dom, Otx,
	 proc(x:dom)
	     local p;
	 begin
	     p:=poly(expr(x), DOM::XVariables);
	     _plus(
		   Otx(expr(nthterm (p,j))) *
		   Otx(subs(nthcoeff(p,j), hold(s).v=(1+q*hold(t).v) $v in X)) 
		   $j=1..nterms(p));
	 end_proc)
    end_proc;
end_domain:

// Ore-Algebra generated by variables x and corresponding euler operators
// tx = x d/dx, with straightening of monomials to x tx
domain experimental::Steenrod::ORExt(X: DOM_LIST, R = Dom::ExpressionField())
    inherits experimental::Steenrod::ORExeuler(X, R);
    category Cat::AlgebraWithBasis(R);

    DX		:= x -> hold(t).x;
    DXVariables := [ dom::DX(x) $ x in X ];

    initDomainAdd :=
    proc()
	option escape;
	local S;
    begin
	S := experimental::SteenrodAlgebra(R);
	operators::overloaded::declareConversion
	(S::p, dom,
	 operators::makeLinear
	 (prog::bless
	  (proc(p: combinat::partitions)
	   begin
	       _mult(_plus(dom(x^k)*dom(1+q*dom::euler(x))
			   $ x in dom::XVariables)
		     $ k in p);
	   end_proc,
	   dom),
	  Source=S::p, ImageSet=dom));
    end_proc;
end_domain:

domain experimental::Steenrod::integerVectorsLexOrderedMonoid(n: Type::NonNegInt)

    inherits combinat::integerVectors;
    category Cat::CombinatorialClass;
    axiom Ax::systemRep;

    interface := combinat::integerVectors::interface minus { hold(shuffle), hold(fromNonWeakDescentsComposition), hold(fromStandard) };

    iszero := (x) -> bool(x=dom::zero);

    zero := [ 0 $ n];

    _divide :=
    proc(v1,v2) : dom
	local result, i;
    begin
        result := zip(v1,v2,_subtract);
	for i in result do
	    if i<0 then return(FAIL);
	    end_if;
	end_for;
	result;
    end;

    divides :=
    proc(v1,v2) : DOM_BOOL
	local i;
    begin
	for i from 1 to nops(v1) do
	    if v1[i] < v2[i] then
		return(FALSE);
	    end_if;
	end_for;
	TRUE;
    end;

    gcd := (x,y) -> zip(x, y, min);
    lcm := (x,y) -> zip(x, y, max);

    degree := _plus@op;

    relativelyPrime := (x,y) -> dom::gcd(x,y) = dom::zero;
    relativelyPrimeStrong := (x,y) -> dom::gcd(x,y) = dom::zero and x <> dom::zero and y <> dom::zero;
    
    // lcm   := ...
    _less := combinat::words::lexLess;
end_domain:


// Polynomials in the variables vars=[x,y,z], over a ring R containing
// the variables Tvars=[tx,ty,tz], so that tx x = x (tx+1)
domain experimental::Steenrod::PolynomialsWithSkewCoefficientsRight(Variables, TVariables, R)
    inherits examples::MultivariatePolynomials(Variables, R);
    category Cat::GradedAlgebraWithBasis(R);

    interface := (examples::MultivariatePolynomials(Variables, R))::interface union
	       {
		hold(rem), hold(remSymmetric),
		hold(quo),
		hold(leadingMonoid), hold(ringActing),
		hold(symmetricGroupAction)
	       };

    XVariables := Variables;

    // There is some extra work to do because the
    // basis elements *do not* commute with the coefficients

    exprMonomial :=
    proc(x : dom)
    begin
	if dom::lcoeff(x) = R::one then
	    dom::exprTerm(dom::lsupport(x))
	else
	    experimental::nonCommutativeMult(dom::exprTerm(dom::lsupport(x)),
					     expr(dom::lcoeff(x)))
	end_if;
    end_proc;

    printMonomial :=
    proc(x : dom)
    begin
	if dom::lcoeff(x) = R::one then
	    dom::printTerm(dom::lsupport(x))
	else
	    experimental::nonCommutativeMult(dom::printTerm(dom::lsupport(x)),
					     expr(dom::lcoeff(x)))
	end_if;
    end_proc;

    straightenCoefficientRight := id;

    bla :=
    operators::makeLinear(prog::bless(
			  proc(m: dom::basisIndices, c)
			  begin
			      dom::monomial(dom::straightenCoefficientRight(m,c), m)
			  end, dom));

    multcoeffsLeft :=
    proc(c: R, p: dom)
    begin
	dom::bla(p,c)
    end;

    mult2 :=
    proc(x1: dom, x2:dom)
	local s1, s2;
    begin
	dom::plus(dom::multcoeffs(dom::mult2Basis(op(s1,2), op(s2,2)),
				  op(s2,1))
		  $ s2 in poly2list(dom::multcoeffsLeft(op(s1,1), x2))
		  $ s1 in poly2list(x1)
		 )
    end_proc;

    degreeBasis := _plus@op;

    initDomainAdd :=
    proc()
    begin
	declareConversion(DOM_IDENT, dom, dom::fromExpr, Explicit)
    end_proc;

    leadingMonoid := experimental::Steenrod::integerVectorsLexOrderedMonoid(nops(Variables));

    // The initial ring of the ring which acts on this module:
    leadingRing := dom;
    ringActing := dom;
    ringActing := dom;

    /*
     * Computes the monomial f so that p = q*f + terms which are
     * smaller than the leading term of p (not the leading term of p!)
     *
     * Could be generalized to return a coefficient by which multiplying
     * p, in cases one wants to do fraction-free/... elimination
     */
    quo :=
    proc(p, q) : Type::Union(DOM_FAIL, dom::ringActing)
	local lt, f, lm;
    begin
	lt := dom::leadingMonoid::_divide(dom::lsupport(p), dom::lsupport(q));
	if lt=FAIL then
	    return(FAIL);
	end_if;
        // Hypothesis:
        //  - The multiplication by scalars on the right is straight
	//  - The term order is compatible with *
	f := dom::term(lt);
	lm:= lmonomial(q)*f;
	f*normal(lcoeff(p)/lcoeff(lm));
    end_proc;

    remOneStep :=
    proc(p, q) : dom
	local f;
    begin
	f := dom::quo(p,q);
	if f = FAIL then return(p) end_if;
	userinfo(0, "factor: ".expr2text(expr(f)));
        p := p - q * f;
    end_proc;

    rem :=
    proc(p, q) : dom
	local f;
    begin
	while not iszero(p) and (f := dom::quo(p,q)) <> FAIL do
	    userinfo(2, "factor: ".expr2text(expr(f)));
	    p := p - q * f;
	end_while;
	p;
    end_proc;

    // Useless attempt ?
    remSymmetric :=
    proc(p: dom, q: dom) : dom
	local f, ls;
    begin
	while not iszero(p) and (f := dom::quo(p,q)) <> FAIL do
	    ls := dom::lsupport(f);
	/*
	perms := select(permutations::list(nops(dom::variables)),
	                p -> (p := permutations::multInverse(lt,p);
	                      p=lt or dom::leadingMonoid::_less(p, lt)));
        */
	    f := _plus(dom::symmetricGroupAction(permutations::inverse(p),f)
		       $ p in permutations::rightCosetsBlah(ls));
	    p := p - q * f;
	    userinfo(0, "factor: ".expr2text(expr(lterm(f))). " * c + ...");
	end_while;
	p;
    end_proc;

    symmetricGroupAction :=
    proc(p: permutations, x: dom) : dom
	local substitution, i;
    begin
	// computes the substitution on the coefficients
	substitution :=	[ TVariables[i] = TVariables[p[i]] $ i = 1..nops(Variables) ];
	dom::fromList(map(dom::toList(x),
			  m -> [ subs(m[1], substitution),
				 permutations::multInverse(m[2],p)]));
    end_proc;

    typeGBElement := Type::ListProduct(Type::PosInt, dom::ringActing::leadingRing, dom);
    typeGBKey := Type::ListProduct(Type::PosInt, dom::ringActing::leadingRing::basisIndices);
    typeGB := Type::ListOf(dom::typeGBElement);
    typeSPair := Type::ListProduct(dom::typeGBKey $ 2);

    normalForm :=
    proc(f: dom, gb: dom::typeGB) : dom
	local done, G, q;
    begin
	while not iszero(f) do
	    done := TRUE;
	    for G in gb do
		if (q := dom::quo(f, G[3])) <> FAIL then
		    f := f - G[3]*q;
		    done := FALSE;
		end_if;
	    end_for;
	    if done then break; end_if;
	end_while;
	f
    end_proc;

    // Returns -1,0,1 depending whether G is freely strictly
    // smaller, comparable or larger than F
    compareFls :=
    proc(flsF: dom::typeGBElement, flsG: dom::typeGBElement)
	local ltF, ltG, iF, iG;
    begin
	[iF, ltF] := flsF;
	[iG, ltG] := flsG;
	if iF<iG then return(-1); end;
	if iF>iG then return(1); end;

	if ltF = ltG then return(0); end_if;
	if dom::ringActing::leadingMonoid::_less(ltF,ltG) then
	    return(-1);
	else
	    return(1);
	end_if;
    end_proc;

    // Returns -1,0,1 depending whether G is freely strictly
    // smaller, comparable or larger than F
    compare :=
    proc(F: dom::typeGBElement, G: dom::typeGBElement)
	local ltf, ltg;
    begin
	if F[1]<G[1] then return(-1); end;
	if F[1]>G[1] then return(1); end;
	ltf := lsupport(F[2]);
	ltg := lsupport(G[2]);
	
	if ltf = ltg then return(0); end_if;
	if dom::ringActing::leadingMonoid::_less(ltf,ltg) then
	    return(-1);
	else
	    return(1);
	end_if;
    end_proc;

    quoFree :=
    proc(F: dom::typeGBElement, G: dom::typeGBElement)
	local q;
    begin
	if F[1] < G[1] then
	    return(FAIL);
	end_if;
	if (q := dom::quo(F[3], G[3])) = FAIL then
	    return(FAIL);
	end_if;
	if F[1] = G[1] then
	    G[2] := G[2] * q;
	    if dom::compare(F, G) < 0 then // F < G
		return(FAIL);
	    end_if;
	end_if;
	return(q);
    end_proc;

    // Returns true if F can't be free-reduced by G
    isNormalFree := // to be optimized
    proc(F: dom::typeGBElement, G: dom::typeGBElement) : DOM_BOOL
    begin
	bool(dom::quoFree(F,G) = FAIL)
    end_proc;

    // Same as normal form, but only reduce modulo elements that
    // are freely-smaller
    normalFormFree :=
    proc(F: dom::typeGBElement, gb) : dom::typeGBElement
	local done, G, q;
    begin
	while not iszero(F[3]) do
	    done := TRUE;
	    for G in values(gb) do
		if (q := dom::quoFree(F, G)) <> FAIL then
		    F[3] := F[3] - G[3]*q;
		    done := FALSE;
		    break;
		end_if;
	    end_for;
	    if done then break; end_if;
	end_while;
	F
    end_proc;

    // Returns u and v such that the leading monomials of f u and g v coincide
    // This is associated to the leading ring of the ring acting
    // Assumptions: the ground ring is commutative
    //              the multiplication by coefficients on the right is straight
    //
    // Could be extended to detect when f and g are relatively prime for the
    // commutative case?
    spairCoeffs :=
    proc(f: dom, g: dom): Type::ListProduct(dom::ringActing::leadingRing $ 2)
	local lcm, ltu, ltv, ltf, ltg;
    begin
	ltf := lsupport(f);
	ltg := lsupport(g);
	lcm := zip(ltf, ltg, max);
	ltu := zip(lcm, ltf, _subtract);
	ltv := zip(lcm, ltg, _subtract);
        // Better implementation: use a method that divides
	// two initial monomials in this ring, and returns an element
	// of dom::ringActing::leadingRing
	// (coact?)
	[dom::monomial(1 / lcoeff(lmonomial(f) * dom::term(ltu)), ltu),
	 dom::monomial(1 / lcoeff(lmonomial(g) * dom::term(ltv)), ltv)];
    end_proc;

    normalizeFree :=
    proc(F: dom::typeGBElement)
	local c;
    begin
	c := 1/dom::lcoeff(F[3]);
	[F[1], F[2]*c, F[3]*c];
    end_proc;

    spairSelectionStrategy := dom::spairSelectionStrategyDegree;

    spairSelectionStrategyDegree :=
    proc(gb, spairs: Type::SetOf(dom::typeSPair)) : dom::typeSPair
	local s, d, lt1, lt2, s1, d1, deg, DOM;
    begin
	DOM := dom;
	assert(nops(spairs) >= 1);
	deg :=
	proc(spair)
	begin
	    lt1 := lsupport(gb[op(spair,1)][3]);
	    lt2 := lsupport(gb[op(spair,2)][3]);
	    DOM::leadingMonoid::degree
		       (DOM::leadingMonoid::lcm(lt1,lt2))
	end_proc;
	d := infinity;
	for s1 in spairs do
	    d1 := deg(s1);
	    userinfo(NoNL, 1, d1);
	    if d1 < d then
		s := s1;
		d := d1;
	    end_if;
	end_for;
	s;
    end_proc;

    groebner :=
    proc(system: Type::ListOf(dom)) : dom::typeGB
	local gb, spairs, insert, addSpair, nextSpair, treatSpair, lessSpair, isSyzygyFls, i, DOM, printfls, getfls, spair, syzygyFlsGenerators, AOneLs, degrees, f;
    begin
	DOM := dom;

	// Definitions:
	// A: ring which is acting, ltA: leading ring of A, lsA: leading monoid of A
	// In the graded case: d: current maximal degree
	// Free leading term of an element p of the module:
	// flt(p) = [i(p), lt(p)] such that p is of the form
	//     system[i(p)] * lt(p) + free-smaller terms;
	// fls(p): [i(p), ls(p)] where ls(p) is the support of lt(p)
	// I    : module spanned by system
	//   In the graded case: I := I_<=d)
	// I_<=fls: subspace spanned by the elements p of I such that fls(p)<=fls
	//   In the graded case: I_<=fls := I_<=d,<=fls
	// The set of flt's/fls's has a structure of module over ltA/lsA

	// The leading support of the unit of A
	// Question: is it guaranteed that the unit of A has a single term?
	AOneLs := lsupport(dom::ringActing::leadingRing::one);

	degrees := map(system, degree);
	// The order on fls
	if _and(dom::isHomogeneous(f) $ f in system) then
	    // graded case  NOT QUITE CORRECT!
	    lessSpair := (s1, s2) -> bool(s1[1]<s2[1] or s1[1]=s2[1] and
					  DOM::compareFls(s1[2], s2[2])<0);
	else
	    lessSpair := (s1, s2) -> bool(DOM::compareFls(s1[2], s2[2])<0);
	end_if;

	// Data structure for gb / gbout
	// table ( fls(p) = [ i(p), lt(p), p], ... )
	gb   := table();
	// Generators for the module of fls's of relations
	syzygyFlsGenerators := {};
	//  - Test if fls is in syzygyFls, the ltA-module spanned by syzygyFlsGenerators:
	//    { [i, lt * t], [i,lt] in syzygyFlsGenerators, t in ltA }
	isSyzygyFls :=
	proc(fls) : DOM_BOOL
	    local v;
	begin
	    for v in syzygyFlsGenerators do
		if v[1]=fls[1] and dom::leadingMonoid::_divide(fls[2], v[2]) <> FAIL then
		    return(TRUE);
		end_if;
	    end_for;
	    return(FALSE);
	end_proc;

	// Elements f*q (obtained from spairs of elements of gb
	// with fls {fls1, fls2}) that remain to be inserted in gb
	// for gb to be a full grobner basis.
	// { [{fls1, fls2}, fls(f*q), f, q], ... }
	spairs := {};

	// Invariants:
	//  - For any p in gb, fls(p) <= maxfls
	//  - gb form a groebner basis for I_<maxfls
	//    That is any p in I_<maxfls reduces to zero w.r.t. gb
	//  - Any p with fls(p) in syzygyFls reduces to zero w.r.t. I_<fls(p)
	//  - In particular, the following elements reduces to zero w.r.t. gb
	//     - p with fls(p)<maxfls
	//     - p with fls(p)=maxfls and p in syzygyFls
	//     - p-flt(p) with fls(p) <= maxfls
	//  - The elements of gb are mutually reduced (if not fully reduced)
	//    within I_<maxfls
	//  - gb and syzygyFls have no fls in common.
	//  - For each [ {fls1, fls2}, fls(f*q), f, q appearing in spairs:
	//     - there are elements gb[fls1], gb[fls2]
	//     - fls(f*q) > maxfls
	//  - The elements in gb may be removed at some point, but are never
	//    modified

	// Utilities for manipulating F := [ i(p), lt(p), p ]
	getfls   := F -> [F[1], lsupport(F[2])];
	printfls := F -> [F[1], expr(lterm(F[2]))];
	/**
         * Insert F, with fls(F) = maxfls, into gb, doing the necessary
	 * reductions to maintain the invariants
	 **/
	insert :=
	proc(F: dom::typeGBElement)
	    local G, modified, fls, q;
	begin
	    //userinfo(NoNL, 1, expr(F), "\n");
	    fls := getfls(F);
	    // First step: reduce F w.r.t. I_<maxfls
	    modified := TRUE;
	    while not iszero(F[3]) and modified do
		modified := FALSE;
		for G in values(gb) do
		    if (q := dom::quo(F[3], G[3])) <> FAIL then
			if DOM::compareFls(fls, [G[1], lsupport(G[2]*q)]) > 0 then
			    F[3] := F[3] - G[3]*q;
			    modified := TRUE;
			    break;
			end_if;
		    end_if;
		end_for;
	    end_while;

	    if iszero(F[3]) then
		userinfo(NoNL, 1, "*");
		// We have detected a relation in the module.
		syzygyFlsGenerators := syzygyFlsGenerators union {fls};
		return();
	    end_if;
	    userinfo(NoNL, 1, "+");
	    // Renormalize F to have leading coefficient 1
	    F := DOM::normalizeFree(F);
	    // Now gb + F is a Groebner basis of I_<=maxfls
	    // It's in fact reduced w.r.t. I_<=maxfls, since one
	    // is not yet allowed to multiply F by something non
	    // constant, and the leading term of F does not coincide
	    // with any leading term of G*t for G in gb.
	    
	    // Update the set of S-pairs
	    for G in values(gb) do
		addSpair(F, G);
	    end_for;
	    gb[fls] := F;
	    userinfo(NoNL, 1, "\n", map(values(gb), expr@lterm@op,3),"\n");
	    userinfo(NoNL, 1, "[".nops(spairs)."]");
	end_proc;

	addSpair :=
	proc(F, G)
	    local F1, G1, fls, u, v;
	    // Precondition: lt(G) does not divide lt(F)
	begin
	    if DOM=DOM::ringActing and
	       (F[1] = G[1] or
		DOM::hasProp(Cat::CommutativeRing)) and
	       DOM::leadingMonoid::relativelyPrimeStrong(lsupport(F[3]),
							 lsupport(G[3])) then
	        // Standard buchberger criterion between elements that commute
		return();
	    end_if;
	    [u,v] := DOM::spairCoeffs(F[3], G[3]);
	    assert(lsupport(u) <> AOneLs);
	    F1 := F;
	    G1 := G;
	    F1[2] := F1[2]*u;
	    G1[2] := G1[2]*v;
	    case DOM::compare(F1,G1)
	      of 0 do // fls(F*u) = fls(G*v)
		  F1[2] := F1[2] - G1[2];
		  if iszero(F1[2]) then
		      userinfo(NoNL, 1, "!");
		      // The free leading term of the s-pair is strictly smaller
		      // We may ignore this s-pair
		      return();
		  end_if;
		  break;
	      of -1 do // F<G
	          // swap F and G, and u and v accordingly
		  F1 := G1;
	    end_case;
	    // Now, the free leading term of the S-pair is derived
	    // from the free leading term of F1
	    spairs := spairs union
		   {[degree(F1[2])+degrees[F1[1]], // for sorting
		     getfls(F1),      // for sorting
		     F1[2],
		     F, G,    // for computing the spair and possibly
		     u, v     // throwing it away if F or G disapears
		    ]};
	end_proc;
	
	treatSpair :=
	proc(spair)
	    local d, fls, lt, F, G, u, v;
	begin
	    [ d, fls, lt, F, G, u, v] := spair;
	    if isSyzygyFls(fls) then
		userinfo(NoNL, 1, "!");
		return();
	    end_if;
	    if (lsupport(v) = AOneLs) then
		// This spair replaces G in the Groebner basis
		delete(gb[getfls(G)]);
	    end_if;
	    insert([fls[1], lt, F[3]*u-G[3]*v]);
	end_proc;

	nextSpair :=
	proc()
	    local s, s1, i;
	begin
	    if nops(spairs) = 0 then
		return(FAIL);
	    end_if;
	    s := op(spairs,1);
	    for i from 2 to nops(spairs) do
		if lessSpair((s1 := op(spairs,i)), s) then
		    s := s1;
		end_if;
	    end_for;
	    spairs := spairs minus {s};
	    s;
	end_proc;

	for i from 1 to nops(system) do
	    userinfo(NoNL, 1, "+");
	    insert([i, DOM::ringActing::leadingRing::one, system[i]]);
	    while (spair := nextSpair()) <> FAIL do
		treatSpair(spair);
	    end_while;
	end_for;
	userinfo(2, "Syzygies: ", syzygyFlsGenerators);
	return(values(gb));
    end_proc;

    // Buggy attempt 
    groebnerAnyOrder :=
    proc(system: Type::ListOf(dom)) : dom::typeGB
	local gb, gbout, syzygyfls, v, known, spairsToDo, insert, treatSpair, i, DOM, getfls, printfls, sweep, spair;
    begin
	DOM := dom;
	// free leading term of an element p of the module:
	// [i, lt] such that p is of the form system[i] * lt + free-smaller terms;
	// [i, ls] where ls is the support of lt is the free leading support (flt) of p
	//
	// Data structure for gb / gbout
	// table ( fls = [ i, lt, p], ... )
	// Invariants:
	//  - fls and [i,lt] are the free leading support of p
	//  - the elements are mutually reduced (if not fully reduced)
	//  - gb, syzygyfls, and gbout do not have common fls
	gb   := table();
	// Elements that have temporarily be removed from the Grbner basis
	gbout := table();
	// Keys [i,ls] such that we know that there is a relation with free leading term [i,ls]
	syzygyfls := {};
	// Is this fls already treated, or being treated, or in the syzygy initial ideal?
	known :=
	proc(fls) : DOM_BOOL
	begin
	    if contains(gb, fls) or contains(gbout, fls) or contains(syzygyfls,fls) then
		return(TRUE);
	    end_if;
	    for v in syzygyfls do
		if v[1]=fls[1] and dom::leadingMonoid::_divide(fls[2], v[2]) <> FAIL then
		    return(TRUE);
		end_if;
	    end_for;
	    return(FALSE);
	end_proc;

	// s-pairs that remain to be computed
	// { {[i1, ls1], [i2, ls2]}, ... }
	// For each fls=[i,ls] appearing above, there is an element gb[fls]
	spairsToDo := {};

	getfls := F -> [F[1], lsupport(F[2])];
	printfls := F -> [F[1], expr(lterm(F[2]))];
	
	/**
         * Insert F into gb, doing the necessary reductions to ensure that
	 * the elements of gb are free-reduced
	 * Also update the list of s-pairs to be reduced
	 **/
	insert :=
	proc(F: dom::typeGBElement)
	    local f, G, removedKeys, k1, fls, modified, q;
	begin
	    // - F and all the original elements of gb are in todo or gb
	    fls := getfls(F);
	    if known(fls) then
		userinfo(NoNL, 1, "!1", fls);
		return();
	    end_if;
	    removedKeys := {};
	    modified := TRUE;
	    while not iszero(F[3]) and modified do
		// Invariant: the fls of F does not appear in gb
		modified := FALSE;
		for G in values(gb) do
		    if (q := dom::quo(F[3], G[3])) <> FAIL then
			G[2] := G[2]*q;
			case DOM::compare(F,G)
			  of 0 do // F=G
			      F[2] := F[2]-G[2];
			      if iszero(F[2]) then
				  userinfo(NoNL, 1, "!2", fls);
			          // The free leading term of the s-pair is strictly smaller
			          // We may ignore this element
				  return();
			      end_if;
			      break;
			  of -1 do // F<G
			      if known(getfls(G)) then
				  // Hmm, some hypothesis is incorrect here!
				  userinfo(NoNL, 1, "!3", expr(lterm(q)), printfls(F), printfls(G));
				  //return();
			      end_if;
			      F[1] :=  G[1];
			      F[2] := -G[2];
			end_case;
			F[3] := F[3] - G[3]*q;
			modified := TRUE;
			break;
		    end_if;
		end_for;
	    end_while;
	    fls := getfls(F);
	    if iszero(F[3]) then
		userinfo(NoNL, 1, "*");
		// We have detected a relation in the module.
		syzygyfls := syzygyfls union {fls};
		return();
	    end_if;
	    F := DOM::normalizeFree(F);
	    // The elements in gb which are not reduced w.r.t. f are
	    // removed and put in the todo
	    // This ensures that all the elements in gb are mutually
	    // reduced (if not fully reduced)
	    gb := select(gb,
			 proc(eq)
			     local G, fls;
			 begin
			     G := op(eq, 2);
			     if (q := DOM::quo(G[3],F[3])) = FAIL then
				 return(TRUE);
			     end_if;
			     fls := op(eq,1);
			     removedKeys := removedKeys union {fls};
			     gbout[fls] := G;
			     return(FALSE);
			 end_proc);
	    // Update the set of S-pairs
	    spairsToDo := select(spairsToDo,
				 s -> s intersect removedKeys = {})
			union { {k1, fls} $ k1 in keys(gb) };
	    gb[fls] := F;
	    userinfo(3, "add: ", fls, "remove: ", removedKeys);
	end_proc;

	sweep :=
	proc()
	    local eq;
	begin
	    while nops(gbout) > 0 do
		eq := op(gbout,1);
		delete(gbout[op(eq,1)]);
		insert(op(eq,2));
	    end_while;
	    userinfo(NoNL, 1, "\n", map(values(gb), expr@lterm@op,3),"\n");
	    userinfo(NoNL, 1, "[".nops(spairsToDo)."]");
	end_proc;

	treatSpair :=
	proc(flsF: dom::typeGBKey, flsG: dom::typeGBKey)
	    local u, v, lh, tmp, F, G;
	begin
	    // This procedure is the core to the Faugere 5 approach
	    F := gb[flsF];
	    G := gb[flsG];
	    if DOM=DOM::ringActing and DOM::hasProp(Cat::CommutativeRing) and
	       DOM::leadingMonoid::relativelyPrimeStrong(lsupport(F[3]),
							 lsupport(G[3])) then
	        // Buchberger Criterion
		return();
	    end_if;
	    [u,v] := DOM::spairCoeffs(F[3], G[3]);
	    F[2] := F[2]*u;
	    G[2] := G[2]*v;
	    case DOM::compare(F,G)
	      of 0 do // fls(F*u) = fls(G*v)
		  if iszero(F[2]-G[2]) then
		      userinfo(NoNL, 1, "!");
		      // The free leading term of the s-pair is strictly smaller
		      // We may ignore this s-pair
		      return();
		  end_if;
		  break;
	      of -1 do // F<G
	          // swap F and G, and u and v accordingly
		  F := G;
		  u := v;
	    end_case;
	    // We are reduced to the case where the free leading term
            // of the S-pair is derived from the free leading term of F
	    if known(getfls(F)) then
		// There is already an element with that free leading term in treatment
		return();
	    end_if;
	    userinfo(NoNL, 1, "+");
	    insert([F[1], F[2], F[3]*u]);
	end_proc;

	for i from 1 to nops(system) do
	    userinfo(NoNL, 1, "+");
	    insert([i, DOM::ringActing::leadingRing::one, system[i]]);
	    sweep();
	    while spairsToDo <> {} do
		spair := dom::spairSelectionStrategy(gb, spairsToDo);
		spairsToDo := spairsToDo minus {spair};
		treatSpair(op(spair,1), op(spair,2));
		sweep();
	    end_while;
	end_for;
	userinfo(2, "Syzygies: ", syzygyfls);
	return(values(gb));
    end_proc;
end_domain:


// Polynomials in the variables vars=[x,y,z]
domain experimental::Steenrod::Polynomials(Variables, R)
    inherits experimental::Steenrod::PolynomialsWithSkewCoefficientsRight(Variables, dom::thetaVariables, R);
    category Cat::GradedAlgebraWithBasis(R), Cat::CommutativeRing;

    makeThetaVariable := v -> hold(t).v;

    thetaVariables := map(Variables, dom::makeThetaVariable);

    straightenCoefficientRight :=
    proc(m: dom::basisIndices, c: dom::coeffRing) : dom::coeffRing
    begin
	c
    end_proc;
end:

// Polynomials in the variables vars=[x,y,z], over a ring R containing
// the variables [tx,ty,tz], so that tx x = x (tx+1); that is tx is the
// euler operator x d/dx.
domain experimental::Steenrod::PolynomialsWithSkewThetaCoefficients(Variables, R)
    inherits experimental::Steenrod::PolynomialsWithSkewCoefficientsRight(Variables, dom::thetaVariables, R);
    category Cat::GradedAlgebraWithBasis(R);

    makeThetaVariable := v -> hold(t).v;

    thetaVariables := map(Variables, dom::makeThetaVariable);

    euler := dom::makeThetaVariable;

    straightenCoefficientRight :=
    proc(m: dom::basisIndices, c: dom::coeffRing) : dom::coeffRing
    begin
	subs(c, zip(dom::thetaVariables, m, (v,d) -> v = v+d));
    end_proc;
end:

// Polynomials in the variables vars=[x,y,z], over a ring R containing
// the variables [tx,ty,tz], so that tx x = x (tx+q).
// that is tx is the q-euler operator 1 + q x d/dx.
domain experimental::Steenrod::PolynomialsWithSkewQThetaCoefficients(Variables, R, q=hold(q))
    inherits experimental::Steenrod::PolynomialsWithSkewCoefficientsRight(Variables, dom::thetaVariables, R);
    category Cat::GradedAlgebraWithBasis(R);

    makeThetaVariable := v -> hold(t).v;

    thetaVariables := map(Variables, dom::makeThetaVariable);

    euler := v -> (dom::makeThetaVariable(v)-1)/q;

    straightenCoefficientRight :=
    proc(m: dom::basisIndices, c: dom::coeffRing) : dom::coeffRing
    begin
	subs(c, zip(dom::thetaVariables, m, (v,d) -> v = v+q*d));
    end_proc;
end:
