// $Date: 2005/01/26 13:42:23 $ $Author: nthiery $ $Revision: 1.1.1.1 $

/* output::tableForm --- printing a set or list in table form

tableForm(CONT <, seperator> <, OPTIONS >)

CONT      : set or list
seperator : (optional) string
OPTIONS   : (optional) one or a set of options

tableForm(CONT) prints the objects of CONT in table form.
The width of the table depends on the size of TEXTWIDTH.

If seperator is given then it is appended to each object.
Appending spaces to the separator results additionally space
between columns.

OPTIONS is one or a set of

Unquoted            - strings will be printed without quotes
Unique              - all columns are of the same width
Left, Center, Right - the entries will be aligned left, center or right
Width = w           - the table has the max width 'w'
Column = c          - the table has 'c' columns
Append = char       - append 'char' to each entry
Sort = procedure    - the entries will be sorted with the given procedure
                      'procedure' is no procedure => no sorting
Output = file       - output into a file
                      if 'file' is a string => file named 'file' will be
                      opened and overwritten and closed after writing
                      if 'file' is a file descriptor the table will be
                      appended to 'file' without closing 'file'
Format = 

Without "Sort" the objects will be converted to strings an then sorted
alphabetically.

Examples:

>> output::tableForm(anames(DOM_PROC))
_check_global _seqin     alias    domattr funcattr loadproc newfuncarg 
newpurefunc   patchlevel pathname protect read     unalias  unprotect 
version       warning                                                  

>> output::tableForm(map(anames(DOM_PROC), expr2text), ", ", {Right})
"_check_global",     "_seqin",       "alias",    "domattr", "funcattr",
     "loadproc", "newfuncarg", "newpurefunc", "patchlevel", "pathname",
      "protect",       "read",     "unalias",  "unprotect",  "version",
      "warning"                                             

>> output::tableForm(map(anames(DOM_PROC), expr2text), "  ", {Center, Unique, Unquoted})
_check_global     _seqin          alias         domattr       funcattr     
  loadproc      newfuncarg     newpurefunc    oldinterval    patchlevel    
  pathname        protect         read          unalias       unprotect    
   version        warning
*/

alias(allOptions = {hold(Unquoted), hold(Unique), hold(Left), hold(Center), hold(Right),
                    hold(Width), hold(Column), hold(Output), hold(Append), hold(Sort), hold(String)}):
alias(optionTypes= {hold(Width) = Type::PosInt,
                    hold(Column) = Type::PosInt,
                    hold(Sort) = Type::AnyType,
                    hold(Output) = Type::Union(DOM_STRING, Type::NonNegInt),
                    hold(Append) = DOM_STRING}):

output::tableForm:=
  proc(CONT = [], SEP = FAIL, OPT = FAIL)
    local COLW, COLWM, COLN, k, PE, FILE, FC, i, APP, SEPlength;
  begin
    if domtype(OPT) <> DOM_TABLE then
      // Extra patch to use the newest prog::getOptions
      OPT:= combinat::getOptions(allOptions, optionTypes, 1)
    end_if;

    if testargs() then
      if contains([DOM_LIST, DOM_SET, DOM_TABLE, DOM_DOMAIN], type(CONT)) = 0 then
        error("wrong type of first argument")
      end_if;
      if args(0) > 1 then
        if type(SEP) = DOM_STRING or type(SEP) = DOM_LIST then
          if nops(OPT) < args(0) - 2 then // too few arguments
            error("wrong type of additional arguments")
          end_if
        else // no separator given, but additional arguments
          if nops(OPT) < args(0) - 1 then // too few arguments
            error("wrong type of additional arguments")
          end_if          
        end_if;
        eval(map(zip([hold(Width), hold(Column), hold(Sort), hold(Output), hold(Append)],
                     [Type::PosInt, Type::PosInt, Type::Union(DOM_PROC, DOM_FUNC_ENV,
                                                              Type::Boolean),
                                                             Type::Union(DOM_STRING, Type::NonNegInt),
                                                                           DOM_STRING],
                     DOM_LIST),
                 proc(opt)
                 begin
                   if contains(OPT, opt[1]) and not testtype(OPT[opt[1]], opt[2]) then
                     hold(error)("wrong argument of option '".expr2text(opt[1])."'")
                   end_if
                 end_proc))
      end_if
    end_if;

    if type(SEP) <> DOM_STRING and type(SEP) <> DOM_LIST then // OPTIONS without SEPARATOR
      SEP:= " ";
      SEPlength:= 1
    else
      OPT[hold(Separator)]:= SEP; // separator set explicitly
      if type(SEP) = DOM_LIST then
        SEPlength:= max(op(map(SEP, length)))
      else
        SEPlength:= length(SEP)
      end_if
    end_if;

    if contains(OPT, hold(Append)) then
      APP:= OPT[hold(Append)]
    else
      APP:= ""
    end_if;
    
    // sort method failed
    if not contains(OPT, hold(Sort)) then
      OPT[hold(Sort)]:= (X, Y) -> expr2text(X) < expr2text(Y)
    end_if;
               
    // file
    FC:= FALSE;
    FILE:=
      (if domtype(OPT[hold(Output)]) = DOM_STRING then
         if (k:= fopen(Text, OPT[hold(Output)], Write)) <> FAIL then
           FC:= TRUE;
           k
         else
           error("cannot open file '".OPT[hold(Output)]."'")
         end_if
       elif domtype(OPT[hold(Output)]) = DOM_INT then
         if traperror(fprint(Unquoted, OPT[hold(Output)], "")) = 0 then
           OPT[hold(Output)]
         else
           error("cannot write into given file")
         end_if
       else
         0
       end_if);

    // different input objetcs
    if type(CONT) <> DOM_LIST then
      if type(CONT) = DOM_SET then
        CONT:= [op(CONT)]
      elif type(CONT) = DOM_TABLE or type(CONT) = DOM_DOMAIN then
        if not contains(OPT, hold(Column)) then
          OPT[hold(Column)]:= 2
        end_if;
        if not contains(OPT, hold(Separator)) then
          SEP:= " = "
        end_if;
        CONT:= map(sort([op(CONT)], OPT[hold(Sort)]), op);
        OPT[hold(Sort)]:= FALSE; // no further sorting!
        return(output::tableForm(CONT, SEP, OPT))
//       elif type(CONT) = piecewise then
//         if not contains(OPT, hold(Column)) then
//           OPT[hold(Column)]:= 2
//         end_if;
//         if not contains(OPT, hold(Separator)) then
//           SEP:= " if "
//         end_if;
//         output::tableForm(map([op(CONT)], X -> stdlib::branch::expression(X),
//                                                stdlib::branch::condition(X)),
//                           SEP, OPT)
      end_if
    else
      // DOM_LIST
    end_if;

    // sort list
    // assert(domtype(CONT, DOM_LIST));
    if OPT[hold(Sort)] <> FALSE then
      CONT:= sort(CONT, OPT[hold(Sort)])
    end_if;
    
    // container empty
    if nops(CONT) = 0 then
      return()
    end_if;

    CONT:= map(CONT, X -> if contains(OPT, hold(Unquoted)) and type(X) = DOM_STRING then
                            X
                          elif domtype(X) = DOM_PROC then
                            DOM_PROC::print(X)
                          else
                            expr2text(X)
                          end_if);

    // col width
    // determine number of columns
    COLWM:= 0;
    for k in CONT do
      if length(k) > COLWM then
        COLWM:= length(k)
      end_if
    end_for;
    COLWM:= COLWM + length(APP) + SEPlength;

    if COLWM < 1 then
      warning("the table is too wide");
      COLWM:= 1
    end_if;
    
    //
    if contains(OPT, hold(Column)) then
      COLN:= OPT[hold(Column)]
    elif contains(OPT, hold(Width)) then
      COLN:= OPT[hold(Width)] div COLWM
    else
      COLN:= (TEXTWIDTH + length(APP) + SEPlength) div COLWM
    end_if;

    if COLN < 1 then
      warning("the table is too wide");
      COLN:= 1
    end_if;

    // create separator list
    if domtype(SEP) <> DOM_LIST then
      SEP:= [""].[SEP $ COLN]
    else
      SEP:= [""].SEP.(SEP $ (COLN - nops(SEP) + 1) div nops(SEP)); // fill up too short list
    end_if;
    
    // 
    COLW:= [COLWM $ COLN];
    if not contains(OPT, hold(Unique)) then
      (COLW[i]:= max(op(map(zip(CONT, [k $k = 1..nops(CONT)],
                                (X, Y) -> if ((Y - 1) mod COLN) = (i - 1) then X else null() end),
                            length)), 1)
       + length(APP) + length(SEP[i])) $ i=1..COLN
    end_if;
    
    // fill up with empty entries
    if (k:= nops(CONT) - (nops(CONT) div COLN)*COLN) > 0 then
      (CONT:= append(CONT, "")) $ i = 1..COLN - k
    end_if;
    
    PE:=
      proc(X, Y, i)
        local app, sep;
      begin
        sep:= "";
        (if i > 1 and X <> "" then
           sep:= SEP[i]
         else
           ""
         end_if).
        (if X <> "" then
           if Y = "" or Y = FAIL then // end of table
             app:= "" //(_concat(" " $ length(APP)))
           else
             app:= APP
           end_if;
           if i = COLN and contains(OPT, hold(Left)) then
             X.app
           else
             if contains(OPT, hold(Right)) then
               (" " $ COLW[i] - length(app) - length(sep) - length(X)).X.app
             elif contains(OPT, hold(Center)) then
               Y:= COLW[i] - length(app) - length(sep) - length(X);
               (" " $ Y div 2).X.app.(if i = COLN then "" else " " $ Y - (Y div 2) end_if)
             else // left
               if i = COLN then // end of row
                 X.app
               else
                 X.app.(" " $ COLW[i] - length(app) - length(sep) - length(X))
               end_if
             end_if
           end_if
         else
           " " $ COLW[i]
         end_if)
      end_proc;

    // printing
    if contains(OPT, hold(String)) then
      return(_concat((PE(op(CONT, (k - 1)*COLN + i), op(CONT, (k - 1)*COLN + i + 1), i)
                      $ i = 1..COLN, "\n") $ k = 1..nops(CONT) div COLN))
    else
      for k from 1 to nops(CONT) div COLN do
        fprint(Unquoted, FILE,
               PE(op(CONT, (k - 1)*COLN + i), op(CONT, (k - 1)*COLN + i + 1), i) $ i = 1..COLN);
      end_for
    end_if;
    
    if FC then fclose(FILE) end_if;
    
    null()
  end_proc:
