/*****************************************************************************
  $Id: overloaded.mu,v 1.1.1.5 2005/01/26 13:42:23 nthiery Exp $
  File:		lib/OPERATORS/overloaded.mu
  Authors:	Nicolas M. Thiry <nthiery@users.sourceforge.net>
  Created:	28/10/2002
  License:	LGPL
*****************************************************************************/


/**
   Domain for functions that can be overloaded.

   Internal representation:

   1st operand: procedure to be called when no signature match.

   2nd operand: function which returns an association table
     table(signature -> procedure to be called for this signature)
    with this function.   

   3rd operand: function allowing for adding another association
   signature -> procedure
   
   4rd optional operand: a string representing the name of the operator
   
   NOTE: We cannot put directly the table as the 2nd operand
   because we need a reference effect. Therefore, we build two closures
   sharing the same environement where the 2nd operand alloes to get the
   table and the 3rd operand allows to set the table.
**/

alias(GETDEFAULT(F)    = extop(F,1)):
alias(GETSIGS(F)       = extop(F,2)()):
alias(SETSIGS(F,sig,f) = extop(F,3)(sig,f)):
alias(GETNAME(F)       = (if extnops(F) = 4 then extop(F,4) else "anonymous operator ".expr2text(extop(F, 1)) end_if)):

domain operators::overloaded
    inherits Dom::BaseDomain;
    category Cat::BaseCategory;

    interfaceAdd := {hold(declareDomain),
		  hold(declareConversion),
		  hold(declareSignature),
		  hold(convert),
		  hold(convertExplicit),
		  hold(printSignatures),
		  hold(printConversions),
		  hold(conversion)
		 };

    typeSignature := Type::ListOf(DOM_DOMAIN);
    typeProcedure := Type::AnyType;
    typeSignatures := Type::TableOf(dom::typeSignature,
				    dom::typeProcedure);
    typeCost := Type::Union(Dom::Rational, Type::Singleton(infinity));

    // Add the elements of this domain in Type::Function
    // This function is called when domtype(x) = operator::overloaded
    // See email discussion with Ralf on 2004/12/12
    testtype :=
    proc(x, t)
    begin
	if t = Type::Function then
	    TRUE
	else
	    FAIL
	end_if
    end_proc;

    new :=
    // Possible calls:
    //  - operators::overloaded(f, table(...)):
    //  - operators::overloaded(f, table(...), Name = string):
    //  - operators::overloaded(f, Name = string):
    //  - operators::overloaded(string, table(...)):
    proc(default:                   Type::Union(dom::typeProcedure, DOM_STRING), 
	 signatures = table()     : Type::Union(dom::typeSignatures,
						Type::Equation(Type::Singleton(Name), DOM_STRING)),
	 nameOf     = (Name="")   : Type::Equation(Type::Singleton(Name), DOM_STRING))
	option escape, noDebug;
    begin
	if domtype(default) = DOM_STRING then
	    nameOf := default;
	    default := ()->error("Don't know how to compute the ".nameOf." of ".
				    expr2text(map(args(), domtype)));
	else
	    nameOf := op(nameOf, 2);
	end_if;
	if domtype(signatures) = DOM_EXPR then
	    nameOf := op(signatures, 2);
	    signatures := table();
	end_if;
	new(dom, default,
	    ()->signatures, 
	    proc(signature: dom::typeSignature, 
		 f: dom::typeProcedure)
	    begin
		signatures[signature] := f;
	    end_proc,
	    if nameOf <> "" then nameOf else null() end_if);
    end_proc;

    expr := // Should probably also implement a method print which
	    // would use GETNAME(f) as well
    proc(F: dom)
    begin
	dom::key(GETDEFAULT(F), GETSIGS(F));
    end_proc;

    declareSignature :=
    proc(F: dom, 
	 signature: dom::typeSignature, 
	 f: dom::typeProcedure)
    begin
	userinfo(Text, 3, "New signature for ", GETNAME(F), ": ", signature);
	// clear the remember table of dom::bestMatch
	dom::bestMatch := subsop(dom::bestMatch, 5=NIL);

	SETSIGS(F, signature, f);
    end_proc;

    bestMatch :=
    /**
     * @param signatures: all the signatures of an overloaded function
     * @param signature : the list of the domains of the elements in the
     *                    function call being resolved
     *
     * @return: the signature in `signatures` which matches best
     *          `signature`.
    **/
    proc(signatures: dom::typeSignatures, signature: dom::typeSignature)
	option remember; // The result depends on the declared conversions
	                 // and signatures; the remember table should
	                 // be cleared or updated after each change to those
	local bestMatch, bestCost,
	      cost, i, eq, sig;
    begin
	userinfo(Text, NoNL, 3, "Calling operators::overloaded::bestMatch");
	bestMatch := FAIL;
	bestCost := infinity;

	for eq in signatures do
	    sig := lhs(eq);
	    if nops(sig) <> nops(signature) then next; end_if;
	    cost := 0;
	    for i from 1 to nops(sig) do
		cost := cost + dom::conversion(signature[i], sig[i], hold(Cost));
		if cost=infinity then break; end_if;
	    end_for;
	    if cost >= bestCost then next; end_if;
	    bestCost := cost;
	    bestMatch := sig;
	end_for;
	userinfo(Text, 3, " done");
	bestMatch;
    end_proc;

    func_call :=
    proc(f: dom)
	local bestMatch, arguments;
	option noDebug;
    begin
	arguments := context([args(2..args(0))]);
	userinfo(Text, 3, "calling overloaded operator ".GETNAME(f)." with ".expr2text(map(arguments, domtype)));
	bestMatch := dom::bestMatch(GETSIGS(f),
				    map(arguments, domtype));
	userinfo(Text, 3, "best signature match: ".expr2text(bestMatch));
	if bestMatch=FAIL then
	    GETDEFAULT(f)(op(arguments));
	else
	    GETSIGS(f)[bestMatch](op(zip(arguments, bestMatch, dom::convert)));
	end_if;
    end_proc;

    getSignatures :=
    proc(f: dom)
    begin
	GETSIGS(f);
    end_proc;

    getName :=
    proc(f: dom)
    begin
	GETNAME(f);
    end_proc;

    printSignatures :=
    proc(f: dom)
	local sig, eq;
    begin
	fprint(Unquoted, 0, sig) $ sig in sort([lhs(eq) $ eq in GETSIGS(f)]);
    end_proc;

    printInfo :=
    proc(f: dom)
	local sig, eq, signatures, g;
    begin
	print(NoNL, GETNAME(f)." -- an overloaded procedure with signatures:\n");
	signatures := GETSIGS(f):
	for sig in sort([lhs(eq) $ eq in signatures]) do
	    print(NoNL, " - ".expr2text(sig));
	    g := signatures[sig];
	    if domtype(g) = DOM_PROC then
		print(NoNL, "\n   (" . g::dom::expr2textParameters(g) . ")");
		if op(g,8) <> NIL then
		    print(NoNL, " -> ".expr2text(op(g,8)));
		end_if;
	    end_if;
	    print(NoNL, "\n");
	end_for;
    end_proc;

    _index := 
    proc(f : dom) 
	local tableSigs, doms, res;
    begin 
	tableSigs := GETSIGS(f);
	doms := [args(2..args(0))];
	res := tableSigs[doms];
	if res = hold(_index)(tableSigs, doms) then FAIL;
	else res; end_if;
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // Conversions mechanism
    //////////////////////////////////////////////////////////////////////////

    conversions      := table();
    conversionsGraph := Graph([],[], Directed);

    domains        := ()->FAIL;
    makeDomainsEnum := proc()
	local i;
	option escape;
    begin
	i:=0;
	proc(F: DOM_DOMAIN)
	    option remember;
	begin
	    i:=i+1;
	    dom::domains(i) := F;
	    dom::conversionsGraph := 
	    Graph::addVertices(dom::conversionsGraph, [i]);
	    i;
	end_proc:
    end_proc;
    domainsEnum    := dom::makeDomainsEnum();

    // Returns whether the domain F has already been declared
    isDomainDeclared :=
    proc(F: DOM_DOMAIN)
    begin
	contains(op(dom::domainsEnum,5), F);
    end_proc;

    // Raise a warning if the domain F was not declared, and declare it
    checkDomainDeclared :=
    proc(F: DOM_DOMAIN)
    begin
	if not dom::isDomainDeclared(F) then
	    //warning("domain ".expr2text(F)." is not declared. Please declare it with operators::overloaded::declareDomain.");
	    dom::declareDomain(F);
	end_if;
    end_proc;

    // May be called several times on the same domain
    declareDomain := d->(dom::domainsEnum(d); null());

    declareConversion :=
    proc(F: DOM_DOMAIN,
	 G: DOM_DOMAIN,
	 f: dom::typeProcedure,
	 weight = Default: Type::Union(Dom::Rational, DOM_IDENT))
	option remember;
        // It can happen that the exact same conversion is declared
	// twice, typically due to some buggy option remember in
	// initDomain with MuPAD 2.0.0. It's safe to just do the job
	// once
	local Gs;
    begin
	userinfo(Text, 3, "New conversion from ", F, " to  ", G);

	// Clear the outdated information in the remember tables of
	// dom::conversion and dom::bestMatch

	// TO BE OPTIMIZED !!! Idealy, one would use djikstra to
	// locally update the conversions.
	// In the mean time, we content ourselves with clearing from the
	// remember table the conversions F1->G1 where G1 is reachable
	// from G in the conversion graph.
	// Given the shape of the average conversion graph, this should
	// be fairly optimal
	if (weight <> Explicit or contains(dom::conversions, (F,G)))
	   and extop(dom::conversion,5) <> NIL then
	    // the set of the domains which are reachable from G
	    Gs := map(Graph::reachableVertices(dom::conversionsGraph,
						[dom::domainsEnum(G)]),
		      dom::domains);
	    // clears the corresponding entries in the remember table
	    dom::conversion :=
	    subsop(dom::conversion,
		   5=select(extop(dom::conversion, 5),
			    eq -> not contains(Gs, op(eq,[1,2]))));
	end_if;
	dom::bestMatch  := subsop(dom::bestMatch,  5=NIL);

	dom::checkDomainDeclared(F);
	dom::checkDomainDeclared(G);

	if contains(dom::conversions, (F,G)) then
	    // overwrite the former definition
	    //warning("Redefining conversion from ". expr2text(F).
	    //        " to ". expr2text(G));
	    if Graph::isEdge(dom::conversionsGraph,
			      [[dom::domainsEnum(F),
				dom::domainsEnum(G)]]) then
		dom::conversionsGraph :=
		Graph::removeEdge(dom::conversionsGraph,
				  [[dom::domainsEnum(F),
				    dom::domainsEnum(G)]]);
	    end_if;
	end_if;
	dom::conversions[F,G] := f;
	if weight=Explicit then
	    return();
	end_if;
	if domtype(weight)=DOM_IDENT then
	    if testargs() and not contains(dom::conversionsWeights, weight) then
		error("Unknown option: ". weight);
	    end_if;
	    weight:=dom::conversionsWeights[weight];
	end_if;
	dom::conversionsGraph := 
	Graph::addEdges(dom::conversionsGraph,
			[[dom::domainsEnum(F),
			  dom::domainsEnum(G)]],
			EdgeWeights=[weight]);
    end_proc;

    conversionsWeights :=
    table(
	  Facade  = 1/1000,
	  Default = 1
	 );

    /**
     * Computes the optimal implicit conversion from F to G
     *
     * @param opt:
     * @return: the optimal implicit conversion or FAIL if there is none;
     *          if opt is set to Cost, then the cost of the conversion
     *          or infinity if there is none
    **/
    conversion :=
    proc(F: DOM_DOMAIN, G: DOM_DOMAIN, opt=FAIL) :
	Type::Union(DOM_LIST, DOM_FAIL, dom::typeCost)
	option remember; // The result depends on the declared conversions;
	                 // the remember table should be cleared or updated 
			 // after each change to those
	local path, pathCosts, result, i, m, start, cost;
    begin
	userinfo(Text, 3, "requesting for conversion");
	if F=G then
	    if opt = hold(Cost) then
		return(0);
	    else
		return([]);
	    end_if;
	end_if;

	// return FAIL if one of the domains is not declared
	if not (dom::isDomainDeclared(F) and dom::isDomainDeclared(G)) then
	    if opt = hold(Cost) then
		return(infinity);
	    else
		return(FAIL);
	    end_if;
	end_if;

	path:=Graph::shortestPathSingleSource
	      (dom::conversionsGraph, 
	       StartVertex=[dom::domainsEnum(F)], 
	       EndVertex=[dom::domainsEnum(G)], 
	       SearchWith=Bellman);
	pathCosts := op(path, 1);
	path      := op(path, 2);

	if nops(path) = 0 then
	    if opt = hold(Cost) then
		return(infinity);
	    else
		return(FAIL);
	    end_if;
	end_if;
	result := [];
	i := dom::domainsEnum(G);
        start := dom::domainsEnum(F);
	cost := 0;
	while i <> start do
	    cost := cost + pathCosts[i];
	    F := dom::domains(path[i]);
	    m := dom::conversions[F,G];
	    if m<>id then
		result := append(result, m);
	    end_if;
	    i := path[i];
	    G := F;
	end_while;
	userinfo(Text, 4, "conversion done");
	if opt=hold(Cost) then
	    cost;
	else
	    revert(result)
	end_if;
    end_proc;

    // Explicit conversion
    convertExplicit :=
    proc(object, G: DOM_DOMAIN)
	local domains;
	option noDebug;
    begin
	userinfo(Text, 3, "Requesting explicit conversion for ",
		 object, " to ", G);  
	domains := domtype(object), G;
	if contains(dom::conversions, domains) then
	    // If a direct conversion exists, always use it
	    return(dom::conversions[domains](object));
	end_if;
	// Otherwise, try an implicit conversion
	return(dom::convert(args()));
    end_proc;

    // Implicit conversion
    convert :=
    proc(object, G: DOM_DOMAIN)
	local conversion, f;
	option noDebug;
    begin
	userinfo(Text, 3, "Requesting conversion for ", object, " to ", G);  
	if (conversion:=dom::conversion(domtype(object), G)) = FAIL then
	    return(FAIL);
	end_if;
	(object := f(object)) $ f in conversion;
	object;
    end_proc;

    // For debugging purposes
    printConversions :=
    proc()
	local t, l, tabulation;
    begin
	// Don't use the construction ... op(eq,1) ... $ eq in dom::conversions
        // to avoid bug #1158 of MuPAD 2.5
	for t in sort(map([op(dom::conversions)],
			  eq -> map([op(eq,1)], expr2text)),
		      not combinat::words::revLexLess) do
	    l := length(t[1]);
	    tabulation := max(16,ceil(l/8)*8) - l;
	    fprint(Unquoted, 0, t[1], " " $ tabulation, "-> ", t[2]);
	end_for;
    end_proc;

begin
end_domain:

/// The standard binary addition and multiplication operators
operators::plus2 :=
operators::overloaded((x,y)->error("Don't know how to add a ".expr2text(domtype(x))." and a ".expr2text(domtype(y))), Name="plus2"):
operators::mult2 :=
operators::overloaded((x,y)->error("Don't know how to multiply a ".expr2text(domtype(x))." by a ".expr2text(domtype(y))), Name="mult2"):
operators::_power :=
operators::overloaded((x,n)->error("Don't know how to compute a ".expr2text(domtype(x)).
				   " elevated to the power of a ".expr2text(domtype(n))), Name="_power"):

/// The standard Hopf algebra operators
operators::antipode :=
operators::overloaded((x)->error("Don't know how to compute the antipode of a ".expr2text(domtype(x))), Name="antipode"):
operators::coproduct :=
operators::overloaded((x)->error("Don't know how to compute the coproduct of a ".expr2text(domtype(x))), Name="coproduct"):
operators::mu :=
operators::overloaded((x)->error("Don't know how to compute the mu product of a ".expr2text(domtype(x))), Name="mu"):
operators::scalar :=
operators::overloaded((x,y)->error("Don't know how to compute the scalar product of a ".expr2text(domtype(x))." by a ".expr2text(domtype(y))), Name="scalar"):


/// The standard tensor operator
operators::tensor2 :=
operators::overloaded((x,y)->FAIL, Name="tensor2"):
		      //error("Don't know how to compute the tensor product of a ".expr2text(domtype(x))." by a ".expr2text(domtype(y))))
operators::tensor :=
	operators::overloaded
	(proc()
	     option escape;
	     local defaultProc;
	 begin
	     defaultProc := operators::makeAssociative(operators::tensor2);
	     proc()
		 name operators::tensor;
                              local argums, doms, domtens, constr, tmp;
	     begin
		 if (tmp := defaultProc(args())) <> FAIL then
		     tmp;
		 else
		     argums   := [args()];
		     doms    := map(argums, domtype);
		     domtens := Dom::TensorProductOfFreeModules(doms);
		     constr  := domtens::tensorConstructorFromSignature(doms);
		     map(doms, operators::overloaded::declareDomain);
		     operators::overloaded::declareDomain(domtens);
		     operators::overloaded::declareSignature(operators::tensor, doms, constr@DOM_LIST);
		     constr(argums);
		 end_if;
	     end_proc;
	 end_proc(), table(), Name="plus2"):

operators::tensorSymbolic := funcenv(hold(slot)(operators, "tensorSymbolic"), // just for printing. 
			             (x)-> hold(tensor)(op(x))):

operators::setTensorSymbol := 
proc(symb = "#" : DOM_STRING)
    local protLevel;
begin
    protLevel := unprotect(operators);
    operators::tensorSymbolic := funcenv(hold(slot)(operators, "tensorSymbolic"),
				  subsop(extop(_mult, 2), 3=" ".symb." ")); // pretty printing. 
    operator(symb, operators::tensor);
    protect(operators, protLevel);
    operators::tensorSymbolic;
end_proc:

// By default, an error is raised.
// Should it default to _mult ? there is a risk of infinite loop!

/**
 * The standard n-ary addition and multiplication operators
 * a*b*c*d is defined as ((a*b)*c)*d
 * Note that operators::plus2 / mult2 are not supposed to be associative; 
 * It's up to the user to put proper parenthesis whenever he wants the
 * evaluation to be done in some other order.
 **/
operators::_plus := operators::makeAssociative(operators::plus2):
operators::_mult := operators::makeAssociative(operators::mult2):

alias(plus2		= operators::plus2):
alias(mult2		= operators::mult2):
alias(negate		= operators::negate):
alias(declareDomain     = operators::overloaded::declareDomain):
alias(declareConversion = operators::overloaded::declareConversion):
alias(declareSignature  = operators::overloaded::declareSignature):

if version()=[2,0,0] then
    // Force the construction of operators::overloaded::domainsEnum
    // Standard workaround for MuPAD 2.0.0's option remember bug
    operators::overloaded::domainsEnum;
end_if:

/// Basic domains
declareDomain    (DOM_INT):
declareDomain    (DOM_RAT):
declareDomain    (DOM_FLOAT):
declareDomain    (DOM_COMPLEX):
declareDomain    (DOM_LIST):
declareDomain    (DOM_POLY):
declareDomain    (DOM_EXPR):
declareDomain    (DOM_IDENT):

/// Facade domains
declareDomain	 (Dom::Integer):
declareDomain	 (Dom::Rational):
declareDomain	 (Dom::Float):
declareDomain	 (Dom::Real):
declareDomain	 (Dom::Complex):
declareDomain	 (Dom::ExpressionField()):

// Parametrized domains

// Declare implicit conversion from Dom::Integer to Dom::IntegerMod(*)
if slot(extop(Dom::IntegerMod,4), "initDomain")=FAIL then
slot(extop(Dom::IntegerMod,4), "initDomain",
     proc()
     begin
	 declareDomain(dom):
	 declareConversion(Dom::Integer, dom, dom::new):
     end_proc):
end_if:

// Declare implicit conversion from K to Dom::AlgebraicExtension(K)
if slot(extop(Dom::AlgebraicExtension,4), "initDomain")=FAIL then
slot(extop(Dom::AlgebraicExtension,4), "initDomain",
     proc()
     begin
	 declareDomain(dom):
	 declareConversion(dom::groundField, dom, dom::new):
     end_proc):
end_if:

/// overloading of plus and mult
declareSignature (plus2, [Dom::Integer, Dom::Integer], _plus):
declareSignature (mult2, [Dom::Integer, Dom::Integer], _mult):
declareSignature (plus2, [Dom::Rational,Dom::Rational],_plus):
declareSignature (mult2, [Dom::Rational,Dom::Rational],_mult):
declareSignature (plus2, [Dom::Float,   Dom::Float],   _plus):
declareSignature (mult2, [Dom::Float,   Dom::Float],   _mult):
declareSignature (plus2, [Dom::Real,    Dom::Real],    _plus):
declareSignature (mult2, [Dom::Real,    Dom::Real],    _mult):
declareSignature (plus2, [Dom::Complex, Dom::Complex], _plus):
declareSignature (mult2, [Dom::Complex, Dom::Complex], _mult):
declareSignature (plus2, [Dom::ExpressionField(), Dom::ExpressionField()],     _plus):
declareSignature (mult2, [Dom::ExpressionField(), Dom::ExpressionField()],     _mult):

/// Trivial conversions due to inclusion of facade domains
declareConversion(DOM_INT,       Dom::Integer,  id,    Facade):

declareConversion(Dom::Integer,  Dom::Rational, id,    Facade):
declareConversion(DOM_RAT,       Dom::Rational, id,    Facade):

declareConversion(DOM_FLOAT,     Dom::Float,    id,    Facade):

declareConversion(Dom::Float,    Dom::Real,     id,    Facade):
declareConversion(Dom::Rational, Dom::Real,     id,    Facade):

declareConversion(Dom::Float,    Dom::Complex,  id,    Facade):
declareConversion(Dom::Rational, Dom::Complex,  id,    Facade):
declareConversion(DOM_COMPLEX,   Dom::Complex,  id,    Facade):

declareConversion(Dom::Real,	 Dom::ExpressionField(), id, Facade):
declareConversion(Dom::Complex,  Dom::ExpressionField(), id, Facade):
declareConversion(DOM_EXPR,      Dom::ExpressionField(), id, Facade):
declareConversion(DOM_IDENT,     Dom::ExpressionField(), id, Facade):

/// Conversions
declareConversion(Dom::Rational, Dom::Float,    float, 1): // Explicit ?
