/*****************************************************************************
  $Id: repeatedSquaring.mu,v 1.1.1.1 2003/09/15 16:37:46 nthiery Exp $
  File:		lib/OPERATORS/repeatedSquaring
  Authors:	Nicolas M. Thiry <nthiery@users.sourceforge.net>
  Created:	01/05/2003
		(mostly stolen from domains::repeatedSquaring from MuPAD 3.0.0)
  License:	LGPL
*****************************************************************************/

/* $Date: 2003/09/15 16:37:46 $ $Author: nthiery $ $Revision: 1.1.1.1 $
*/
/*++
repeatedSquaring -- create procedure implementing 'repeatedSquaring'

repeatedSquaring(o, <Zero = zero, Opposite = opposite, Source=source>)

D   - domain
o   - binary associative operation from source to itself
zero   - neutral element according to o (if exists)
opposite - inverting operation according to o

repeatedSquaring returns a procedure p(x,i) which calculates 'x o x o
... o x (i times)' for domain elements 'x' and positive integers 'i'
where 'x' 'o' is an operation. If 'o' has a neutral element according
to 'o', it may be given by 'zero'.  In this case the "power" 'i' may
be zero, and p(x,0) returns 'zero'.  If 'o' additionally has an
opposite operation for 'o' the exponents 'i' may be negative; for
negative exponents the inverse of 'x' is squared.

If, for example, o is "_mult" and u is "1" then repeatedSquaring
returns a procedure p to compute x^i for integers i >= 0.  The result
is calculated by repeated squaring.

++*/


operators::repeatedSquaringInner :=
proc(f)
    option escape;
    local zero, source, opposite, options;
begin
    options := combinat::getOptions(dom::repeatedSquaring, 1, Error):
    source  := options[Source];
    zero    := options[Zero];

    if contains(options, Zero) and contains(options, Opposite) then
	opposite := options[Opposite];
	// create 'repeated squaring' using neutral element and inversion
	/* proc(x,n) -- returns 'x oper x .... x (n times)' if n > 0
	  'unit' if n = 0 and 'y oper y .... y (-n times)' if n < 0
	  and y = inv_mthd(x) */
	proc(x: source, n: Type::Integer) : source
	    local p;
	begin
	    if n < 0 then
		x := opposite(x);
		if x = FAIL then return(FAIL) end_if;
		n:= -n;
	    end_if;
	    if n=0 then 
		return(zero);
	    end_if;
	    // Try to avoid using zero if you don't have to ...
	    // This below can probably be simplified
	    if n mod 2 = 1 then
		p:= x;
		n:= n-1
	    else
		p:= f(x, x);
		n:= n-2
	    end_if;
	    while TRUE do
		if n mod 2 = 1 then
		    p:= f(p, x)
		end_if;
		if n < 2 then break end_if;
		n:= n div 2;
		x:= f(x, x)
	    end_while;
	    return(p)
	end_proc;
    elif contains(options, Zero) then
	// create 'repeated squaring' using neutral element
	/* proc(x,n) -- returns 'x oper x .... x (n times)' if n > 0
	  and 'unit' if n = 0 */
	proc(x: source, n: Type::NonNegInt) : source
	    local p;
	begin
	    if n=0 then 
		return(zero);
	    end_if;
	    // Try to avoid using zero if you don't have to ...
	    // This below can probably be simplified
	    if n mod 2 = 1 then
		p:= x;
		n:= n-1
	    else
		p:= f(x, x);
		n:= n-2
	    end_if;
	    // repeated squaring
	    while TRUE do
		if n mod 2 = 1 then
		    p:= f(p, x)
		end_if;
		if n < 2 then break end_if;
		n:= n div 2;
		x:= f(x, x)
	    end_while;
	    return(p)
	end_proc;
    else
	// create 'repeated squaring' without neutral element
	// proc(x,n) -- returns 'x oper x .... x (n times)' for n > 0
	proc(x: source, n: Type::PosInt) : source
	    local p;
	begin
	    if n mod 2 = 1 then
		p:= x;
		n:= n-1
	    else
		p:= f(x, x);
		n:= n-2
	    end_if;
	    // repeated squaring
	    while TRUE do
		if n mod 2 = 1 then
		    p:= f(p, x)
		end_if;
		if n < 2 then break end_if;
		n:= n div 2;
		x:= f(x, x)
	    end_while;
	    return(p)
	end_proc;
    end_if;
end_proc:

operators::repeatedSquaring :=
funcenv(operators::repeatedSquaringInner,
	FAIL,
	table("options"     =      [Zero,
				    Opposite,
				    Source   = Type::AnyType]));
