/*****************************************************************************
  $Id: init.mu,v 1.7 2005/02/07 14:53:51 nthiery Exp $
  File:		lib/init.mu
  Purpose:	Global initialization file for the combinat package,
		including options, domains, axioms, categories, modules, 
		as well as some workarounds to ensure compatibility 
		with all versions of MuPAD >= 2.0.0.
  Authors:	Nicolas M. Thiry <nthiery@users.sourceforge.net>
                Florent Hivert <florent.hivert@univ-mlv.fr>
  Created:	16/02/2001
  License:	LGPL
*****************************************************************************/

if domtype(packages) <> DOM_DOMAIN then
    packages := newDomain("packages"):
    protect(packages):
end_if:

domain packages::Combinat
    version := [1,2,1];

    versionString := "".dom::version[1].".".dom::version[2].".".dom::version[3];

    info_str := "
 +---+
 | T |                    MuPAD-Combinat ". dom::versionString ."
 +---+---+
 | A | K |          an open source MuPAD package for
 +---+---+---+
     | I | N |    research in Algebraic Combinatorics
     +---+---+

This package provides or extends the following libraries:
  combinat, examples, Dom, Cat, output, experimental, IPC

For quick information on a particular library, please type:
  info(library)

For the full html documentation, please browse through:
  http://mupad-combinat.sf.net/ (project web page)
  file:".dom::path."index.html (local mirror)";

end_domain:

// Try to guess the path to the modules directory by looking
// at what was last added to READPATH. Probably not robust!
packages::Combinat::modulesPath := op([READPATH], 1):

// The root directory of the package
packages::Combinat::path :=
proc()
    local s, l;
begin
    if domtype(FILEPATH) = DOM_STRING then
	// Strip away the trailing lib/. Probably not portable!
	substring(FILEPATH,
		  %if "12"[1]="1" then 1 else 0; end_if,
		  length(FILEPATH)-4);
    else
	// In MuPAD <= 2.5.0, FILEPATH is not defined
	// So, we try to guess the root directory from the module path
	s := packages::Combinat::modulesPath;
	// remove the trailing "modules/linux":
	substring(s,0,stringlib::pos(s, "modules")):
    end_if;
end_proc():

/*****************************************************************************
   Compatibility patches for the MuPAD library that need to be applied early
*****************************************************************************/

alias(MuPADVersionAtLeast =
      proc(v) : DOM_BOOL
	  local i;
      begin
	  w := version();
	  for i from 1 to 3 do
	      if v[i] < w[i] then return(TRUE) end_if;
	      if v[i] > w[i] then return(FALSE) end_if;
	  end_for;
	  TRUE
      end_proc):

//////////////////////////////////////////////////////////////////////////////
// Backport of autoload (new in MuPAD 3.1.0) 2004/10/26

if domtype(autoload) = DOM_IDENT then
    autoload := loadproc(autoload, pathname("MUPADBACKPORT", "STDLIB"), "autoload"):
end_if:
autoload(prog::loaddomain): prog::loaddomain:
alias(loaddomain=prog::loaddomain):

//////////////////////////////////////////////////////////////////////////////
// Backport of Cat::BaseCategory: this fixes a buggy testtype in MuPAD
// 2.0.0, and enables the interfaceAdd feature.

if not MuPADVersionAtLeast([3,1,0]) then // Feature test ?
    Cat::BaseCategory :=
    loadproc(Cat::BaseCategory, pathname("MUPADBACKPORT", "DOMAINS", "CATEGORY"), "BaseCat"):
end_if:

//////////////////////////////////////////////////////////////////////////////
// Backport extensions to the DomainConstructor domain for MuPAD < [3,0,0]
// typically the initDomain feature

alias(path=pathname("MUPADBACKPORT", "DOMAINS", "CONSTR")):
if not prog::isDefined(DomainConstructor::getEntryFromSuperDomain) then
    // autoload(..., Lib=..., Provide)
    // autoload(..., Lib=..., Provide, WarnIfDefined)
    DomainConstructor::getEntryFromSuperDomain :=
    loadproc(DomainConstructor::getEntryFromSuperDomain, path, "getEntryFromSuperDomain"):
end_if:
if not prog::isDefined(DomainConstructor::getEntryFromCategory) then
    DomainConstructor::getEntryFromCategory :=
    loadproc(DomainConstructor::getEntryFromCategory, path, "getEntryFromCategory"):
end_if:
if not prog::isDefined(DomainConstructor::getEntryFromAllSuperDomains) then
    DomainConstructor::getEntryFromAllSuperDomains :=
    loadproc(DomainConstructor::getEntryFromAllSuperDomains, path, "getEntryFromAllSuperDomains"):
end_if:
if not prog::isDefined(DomainConstructor::getEntryFromAllCategories) then
    DomainConstructor::getEntryFromAllCategories :=
    loadproc(DomainConstructor::getEntryFromAllCategories, path, "getEntryFromAllCategories"):
end_if:


if not MuPADVersionAtLeast([3,0,0]) then // could we do a feature test instead ?
    DomainConstructor::func_call :=
    loadproc(DomainConstructor::func_call, path, "DCfcall");
    DomainConstructor::getEntry :=
    loadproc(DomainConstructor::getEntry, path, "DCgetEnt");
    // The new getEntry does not play perfectly with the makeEntry of
    // MuPAD < 3.0.0 (methods are named r instead of bla::truc).
    // So, we backport the later as well
    DomainConstructor::makeEntry :=
    loadproc(DomainConstructor::makeEntry, path, "DCmkEnt");
end_if:

if not MuPADVersionAtLeast([3,1,2]) then // feature test ?
    DomainConstructor::printInfo :=
    loadproc(DomainConstructor::printInfo, path, "DCprInfo");
    CategoryConstructor::printInfo :=
    loadproc(CategoryConstructor::printInfo, path, "CCprInfo");
    Category :=
        loaddomain(Category, path, "Category"):
    Dom::BaseDomain :=
        loaddomain(Dom::BaseDomain, pathname("MUPADBACKPORT", "DOMAINS", "DOMAIN"), "BaseDom"):
end_if:
unalias(path):

/*****************************************************************************
   Functionality tests that needs to be done at an early stage

   Using aliases instead of variables keeps them local to this file
*****************************************************************************/

if contains(Cat, "SemiRing") then
    alias(MUPAD_HAS_SEMIRING=TRUE);
else
    alias(MUPAD_HAS_SEMIRING=FALSE);
end_if:

/*****************************************************************************
  Adds "modules/linux/3.0.0/" (or similarwise for other systems & versions of
  mupad) to the front of READPATH. This allows for having dynamic modules
  compiled for multiple versions of MuPAD simultaneously.
*****************************************************************************/

READPATH := (packages::Combinat::modulesPath.
	     pathname("",
		      expr2text(op(version(),1)).".".
		      expr2text(op(version(),2)).".".
		      expr2text(op(version(),3)))),
	     READPATH:

/*****************************************************************************
   Extensions to preexisting MuPAD standard libraries
*****************************************************************************/

//////////////////////////////////////////////////////////////////////////////
// Extensions to the Type library

Type::interface := Type::interface union {
    hold(Polynomial),
    hold(PolynomialDomain),
    hold(Predicate),
    hold(TableOf),
    hold(Intersection),
    hold(Option),
    hold(OptionsTable)
}:

// Backport of extensions already in MuPAD

alias(path=pathname("MUPADBACKPORT", "TYPE")):
if Type::Predicate = FAIL then
    Type::Predicate :=
    loadproc(Type::Predicate, path, "Predicate"):
end_if:
if Type::TableOf = FAIL then
    Type::TableOf :=
    loadproc(Type::TableOf, path, "TableOf"):
end_if:
if Type::Intersection = FAIL then
    Type::Intersection :=
    loadproc(Type::Intersection, path, "Intersection"):
end_if:
Type::Function := loadproc(Type::Function, path, "Function"):
unalias(path):

alias(path = pathname("TYPE")):
Type::Polynomial := Type::AnyType:
Type::PolynomialDomain := DOM_DOMAIN:
Type::Option :=
    Type::Union(Type::Equation(DOM_IDENT, Type::AnyType), DOM_IDENT):
Type::OptionsTable :=
    Type::TableOf(DOM_IDENT):
unalias(path):

//////////////////////////////////////////////////////////////////////////////
// Extensions to the prog library

prog::interface := prog::interface union {
    hold(bless),
    hold(makeTest),
    hold(makeExample),
    hold(testUnit)
}:

alias(path = pathname("PROG")):

autoload(prog::makeTest):
autoload(prog::makeExample):
autoload(prog::bless):
prog::blessAny :=
     loadproc(prog::blessAny, path, "bless"):
prog::domainWrapper :=
    loaddomain(prog::domainWrapper, path, "domainWrapper"):
prog::testUnit :=
    loaddomain(prog::testUnit, path, "testUnit"):
prog::indexedProcedure :=
    loaddomain(prog::indexedProcedure, path, "domainWrapper"):
prog::setMaximalMAXDEPTH :=
    // Set MAXDEPTH to the maximal value accepted by MuPAD
    // Stupid implementation, but sufficient for our needs
    proc()
    begin
	/* Uses 2^30-1 instead of the more natural 2^31-1
	   because of a bug in MuPAD 2.0.0 between
	   MAXDEPTH and loadproc, as in the following ?????
	   >> reset(); MAXDEPTH := 2^31-1; sign; */
	if traperror((MAXDEPTH := 2^30-1)) = 0 then return(); end_if;
	while traperror((MAXDEPTH := MAXDEPTH + 1)) = 0 do;
	end_while
    end_proc:
unalias(path):

//////////////////////////////////////////////////////////////////////////////
// Extensions to the polylib library

polylib::interface := polylib::interface union {
    hold(variables), hold(coeffRing), hold(constructor)
}:

autoload(polylib::support):
autoload(polylib::lsupport):
autoload(polylib::monomials):
autoload(polylib::variables):
autoload(polylib::coeffRing):
autoload(polylib::constructor):

//////////////////////////////////////////////////////////////////////////////
// Extensions to the output library

output::interface := output::interface union {
    hold(asciiArt)
}:
alias(path = pathname("OUTPUT")):
output::asciiArt := // Might go into the kernel some day
    loaddomain(output::asciiArt, path, "asciiArt"):
unalias(path):

//////////////////////////////////////////////////////////////////////////////
// Extensions to the adt library

adt::interface := adt::interface union {
    hold(Stack)
}:
alias(path = pathname("ADT")):
adt::Stack :=
    loaddomain(adt::Stack, path, "Stack"):
unalias(path):

/*****************************************************************************
   Library domains
*****************************************************************************/

alias(path = pathname("LIBFILES")):
operators :=    loaddomain(operators,	       path, "operators"):
protect(operators):

combinat :=     loaddomain(combinat,     path, "combinat"):
protect(combinat):
muEC :=		loaddomain(muEC,         path, "muEC"):
protect(muEC):
IPC :=		loaddomain(IPC,	         path, "IPC"):
protect(IPC):
examples :=	loaddomain(examples,     path, "examples"):
protect(examples):
experimental := loaddomain(experimental, path, "experimental"):
protect(experimental):
unalias(path):

/*****************************************************************************
   Dynamic Modules
*****************************************************************************/
// Autoload for dynamic modules. For the momment they are placed at the 
// toplevel but they can be moved elsewhere. 

alias(path = pathname("MODULES")):
// Should move them to MODULE
Symmetrica := loaddomain(Symmetrica, path, "Symmetrica"):
gordan     := loaddomain(gordan,     path, "gordan"):
protect(Symmetrica):
protect(gordan):
unalias(path):

// This is suggested in ?package in order to print out the
// documentation when loading the package. However Ralph Hillebrand
// does not quite agree with this mechanism (see discussion by email
// with him in march 2001).
stdlib::LIBRARIES := stdlib::LIBRARIES
                    union {"packages::Combinat"}:

/*****************************************************************************
   Domains
*****************************************************************************/

Dom::interface: // force loading 'Dom'

Dom::interface := Dom::interface union {
    hold(FreeModule),
    hold(FreeModuleList),
    hold(FreeModulePoly),
    hold(FreeModuleTable),
    hold(WeylAlgebra_Table),
    hold(SteenrodAlgebraP),
    hold(PermutationGroup),
    hold(SymmetricGroup),
    hold(AlternatingGroup),
    hold(PermutationGroupOnSets),
    hold(PermutationGroupDirectProduct),
    hold(Graph),
    hold(SemiRing),
    hold(BooleanSemiRing),
    hold(MaxMinSemiRing),
    hold(MaxPlusSemiRing),
    hold(MinMaxSemiRing),
    hold(MinPlusSemiRing),
    hold(WeightedAutomaton),
    hold(SparseMatrix_LinBox),
    hold(TensorProductOfFreeModules),
    hold(SubFreeModule),
    hold(QuotientFreeModule)
}:

alias(path = pathname("DOMAINS", "DOMAIN")):
Dom::SubFreeModule :=
    loaddomain(Dom::SubFreeModule, path, "SubFreeModule"):
Dom::QuotientFreeModule :=
    loaddomain(Dom::QuotientFreeModule, path, "QuotientFreeModule"):
Dom::TensorProductOfFreeModules :=
    loaddomain(Dom::TensorProductOfFreeModules, path, "TensorProductOfFreeModules"):

Dom::PermutationGroup_GLIP :=
loaddomain(Dom::PermutationGroup_GLIP, path, "PermutationGroup_GLIP"):
//Dom::BaseDomain :=
//    loadproc(Dom::BaseDomain, path, "BaseDomain"):
Dom::PermutationGroupInvariantRing :=
    loaddomain(Dom::PermutationGroupInvariantRing, path, "PermutationGroupInvariantRing"):
// Note: Dom::PermutationGroup in MuPAD 2.0.0 is not compatible, and
// is stored in a file with a different name.
Dom::PermutationGroup:= loaddomain(Dom::PermutationGroup, path, "PermutationGroup"):
Dom::SymmetricGroup := loaddomain(Dom::SymmetricGroup,   path, "SymmetricGroup"):
Dom::AlternatingGroup :=
    loaddomain(Dom::AlternatingGroup, path, "AlternatingGroup"):
Dom::PermutationGroupOnSets :=
    loaddomain(Dom::PermutationGroupOnSets, path, "PermutationGroupOnSets"):
Dom::PermutationGroupDirectProduct :=
    loaddomain(Dom::PermutationGroupDirectProduct, path, "PermutationGroupDirectProduct"):
Dom::FreeModuleTable := loaddomain(Dom::FreeModuleTable, path, "FreeModuleTable"):
Dom::FreeModulePoly := loaddomain(Dom::FreeModulePoly,  path, "FreeModulePoly"):
Dom::FreeModuleList := loaddomain(Dom::FreeModuleList,  path, "FreeModuleList"):
//Dom::FreeModuleSparseMat :=
//    loaddomain(Dom::FreeModuleSparseMat, path, "FreeModuleSparseMat"):
//Dom::FreeModule := Dom::FreeModuleSparseMat:
//Dom::FreeModule := Dom::FreeModuleMnPoly:
Dom::FreeModule       :=
funcenv(() -> Dom::FreeModulePoly(args()),
	FAIL,
	table("info" = "Dom::FreeModule -- a short hand for Dom::FreeModulePoly, which see")):
//print("Using ".expr2text(Dom::FreeModule)." for default free modules implementations.");
Dom::AlgebraWithBasis_Table :=
    loaddomain(Dom::AlgebraWithBasis_Table, path, "AlgebraWithBasis_Table"):
Dom::WeylAlgebra_Table :=
    loaddomain(Dom::WeylAlgebra_Table, path, "WeylAlgebra_Table"):
Dom::SteenrodAlgebraP :=
    loaddomain(Dom::SteenrodAlgebraP, path, "SteenrodAlgebraP"):
Dom::SparseMatrix_LinBox :=
    loaddomain(Dom::SparseMatrix_LinBox, path, "SparseMatrix_LinBox"):

Dom::SemiRing :=
    loaddomain(Dom::SemiRing, path, "SemiRing"):
Dom::BooleanSemiRing :=
    loaddomain(Dom::BooleanSemiRing, path, "BooleanSemiRing"):
Dom::MaxMinSemiRing :=
    loaddomain(Dom::MaxMinSemiRing, path, "MaxMinSemiRing"):
Dom::MaxPlusSemiRing :=
    loaddomain(Dom::MaxPlusSemiRing, path, "MaxPlusSemiRing"):
Dom::MinMaxSemiRing :=
    loaddomain(Dom::MinMaxSemiRing, path, "MinMaxSemiRing"):
Dom::MinPlusSemiRing :=
    loaddomain(Dom::MinPlusSemiRing, path, "MinPlusSemiRing"):
Dom::WeightedAutomaton :=
    loaddomain(Dom::WeightedAutomaton, path, "WeightedAutomaton"):

unalias(path):

/*****************************************************************************
   Categories
*****************************************************************************/

Cat::interface: // force loading 'Cat'

Cat::interface := Cat::interface union
   {
    hold(AlgebraWithBasis),
    hold(CombinatorialClass),
    hold(DecomposableClass),
    hold(FiniteDimensionalAlgebraWithBasis),
    hold(GradedAlgebraWithBasis),
    hold(GradedHopfAlgebraWithBasis),
    hold(HopfAlgebraWithBasis),
    hold(IntegerListsLexClass),
    hold(ModuleWithBasis),
    hold(PermutationGroup),
    hold(SemiRing),
    hold(SemiRng),
    hold(UseOverloading),
    hold(CombinatorialClassWith2DBoxedRepresentation),
    hold(Operad),
    hold(OperadWithBasis),
    hold(GradedModuleWithBasis),
    hold(ModuleWithSeveralBases),
    hold(GradedCombinatorialClass),
    hold(LeftAModuleWithBasis),
    hold(AlgebraWithSeveralBases),
    hold(HopfAlgebraWithSeveralBases)
   }:

alias(path = pathname("DOMAINS", "CATEGORY")):
Cat::HopfAlgebraWithSeveralBases :=
    loaddomain(Cat::HopfAlgebraWithSeveralBases, path, "HopfAlgebraWithSeveralBases"):
Cat::AlgebraWithSeveralBases :=
    loaddomain(Cat::AlgebraWithSeveralBases, path, "AlgebraWithSeveralBases"):
Cat::LeftAModuleWithBasis :=
    loaddomain(Cat::LeftAModuleWithBasis, path, "LeftAModuleWithBasis"):
Cat::GradedCombinatorialClass :=
    loaddomain(Cat::GradedCombinatorialClass, path, "GradedCombinatorialClass"):
Cat::ModuleWithSeveralBases :=
    loaddomain(Cat::ModuleWithSeveralBases, path, "ModuleWithSeveralBases"):
Cat::GradedModuleWithBasis :=
    loaddomain(Cat::GradedModuleWithBasis, path, "GradedModuleWithBasis"):
Cat::Operad :=
    loaddomain(Cat::Operad, path, "Operad"):
Cat::OperadWithBasis :=
    loaddomain(Cat::OperadWithBasis, path, "OperadWithBasis"):
Cat::CombinatorialClassWith2DBoxedRepresentation :=
    loaddomain(Cat::CombinatorialClassWith2DBoxedRepresentation, path, "CombinatorialClassWith2DBoxedRepresentation"):
//Cat::BaseCategory :=
//    loadproc(Cat::BaseCategory, path, "BaseCategory"):
Cat::FiniteGroupInvariantRing :=
    loaddomain(Cat::FiniteGroupInvariantRing, path, "FiniteGroupInvariantRing"):
Cat::PermutationGroupInvariantRing :=
    loaddomain(Cat::PermutationGroupInvariantRing, path, "PermutationGroupInvariantRing"):
Cat::AlgebraWithBasis :=
    loaddomain(Cat::AlgebraWithBasis,		path, "AlgebraWithBasis"):
Cat::FiniteDimensionalAlgebraWithBasis :=
    loaddomain(Cat::FiniteDimensionalAlgebraWithBasis, path, "FiniteDimensionalAlgebraWithBasis"):
Cat::GradedAlgebraWithBasis :=
    loaddomain(Cat::GradedAlgebraWithBasis, 	path, "GradedAlgebraWithBasis"):
Cat::GradedHopfAlgebraWithBasis :=
    loaddomain(Cat::GradedHopfAlgebraWithBasis, path, "GradedHopfAlgebraWithBasis"):
Cat::ModuleWithBasis :=
    loaddomain(Cat::ModuleWithBasis,		path, "ModuleWithBasis"):
Cat::PermutationGroup :=
    loaddomain(Cat::PermutationGroup,		path, "PermutationGroup"):
Cat::SemiRng :=
    loaddomain(Cat::SemiRng,			path, "SemiRng"):
Cat::SemiRing :=
    loaddomain(Cat::SemiRing,			path, "SemiRing"):
Cat::Ring :=
    loaddomain(Cat::Ring,			path, "Ring"):
Cat::Rng :=
    loaddomain(Cat::Rng,			path, "Rng"):
Cat::SemiGroup :=
    loaddomain(Cat::SemiGroup,			path, "SemiGroup"):
Cat::SemiRing :=
    loaddomain(Cat::SemiRing,			path, "SemiRing"):
Cat::SemiRng :=
    loaddomain(Cat::SemiRng,			path, "SemiRng"):
Cat::UseOverloading :=
    loaddomain(Cat::UseOverloading,		path, "UseOverloading"):
Cat::HopfAlgebraWithBasis :=
    loaddomain(Cat::HopfAlgebraWithBasis, 	path, "HopfAlgebraWithBasis"):
Cat::CombinatorialHopfAlgebra := // Old implementation for backward compatibility. 
    loaddomain(Cat::CombinatorialHopfAlgebra, 	path, "CombinatorialHopfAlgebra"):

/*****************************************************************************
   Categories for combinatorial objects
*****************************************************************************/

Cat::CombinatorialClass := loaddomain(Cat::CombinatorialClass,  path, "CombinatorialClass"):
Cat::IntegerListsLexClass := loaddomain(Cat::IntegerListsLexClass,path, "IntegerListsLexClass"):
Cat::DecomposableClass := loaddomain(Cat::DecomposableClass,   path, "DecomposableClass"):
// FIXME: move into DOMAINS/CATEGORY
Cat::TreesClass := loaddomain(Cat::TreesClass,   pathname("COMBINAT"), "treesGeneric"):

/*
// Temporary alias to allow for using semi-rings as coefficients
// for sparse matrices and such
category Cat::SemiRing
    category Cat::Rng;
    _negate := FAIL;
    // Allows for subtracting an element from itself
    // This is currently required for the method _set_index of Dom::Matrix:
    // to set to a the coefficient of X^i in a polynomial, this
    // method first subtract c*X^i where c is the former coefficient
    // of X^i, and then adds a*X^i
    _subtract :=
    proc(x: dom, y: dom)
    begin
	if x = y then
	    dom::zero
	else
	    error("not implemented");
	end_if;
    end_proc;
end_category:
*/
unalias(path):

/*****************************************************************************
   Axioms
*****************************************************************************/

Ax::interface: // force loading 'Ax'

Ax::interface := Ax::interface union {
    hold(Ax::monoidAlgebra)
}:

alias(path = pathname("DOMAINS", "AXIOM")):
Ax::monoidAlgebra :=
    loaddomain(Ax::monoidAlgebra, path, "monoidAlgebra"):
unalias(path):

/*****************************************************************************
   Options
*****************************************************************************/

proc()
    local opt, packageOptions;
begin
    packageOptions :=
      {hold(MinLength),   // integerListsLexTools & inherited
       hold(MaxLength),   //  ...
       hold(Length),      //  ...
       hold(MinPart),     //  ...
       hold(MaxPart),     //  ...
       hold(Inner),       //  ...
       hold(Outer),       //  ...
       hold(MinSlope),    //  ...
       hold(MaxSlope),    //  ...
       hold(Repeat),      // subwords
       hold(FinalDescent),// permutations & compositions
       // binaryTrees & labelledBinaryTrees
       hold(Complete), hold(Binary), hold(CompleteBinary),
       hold(BinaryTree), hold(CompleteBinaryTree),
       hold(Ordered),
       hold(Label), hold(Leaf), hold(LeafNil), hold(Node),
       hold(Left), hold(Right),
       hold(Prefix), hold(RightPrefix), hold(LeftPrefix),
       hold(Infix), hold(RightInfix), hold(LeftInfix),
       hold(Postfix), hold(RightPostfix), hold(LeftPostfix),
       // output::asciiArt
       hold(Top), hold(Center), hold(Bottom), 
       hold(MinHeight), hold(MinWidth),
       hold(Compact), hold(Packed),
       hold(ColumnSeparator), hold(RowSeparator), hold(Cross), hold(HorizontalPadding),
       hold(VerticalAlignment), hold(Align), hold(MaxRemove),
       // combinat::decomposableObjects
       hold(Boustrophedonic),
       hold(NonTerminal),
       hold(Epsilon), hold(Atom),
       hold(Union), hold(Powerset), hold(Multiset), hold(Sequence), hold(Cycle), hold(Lyndon),
       hold(Predefined), hold(Domain),
       hold(Prod), hold(BoxedProd),
       hold(Int), hold(Theta), hold(Delta), hold(Alias), hold(Lyndon),
       hold(Marked), hold(DivX), hold(MultX),
       hold(Labelled), hold(Products), hold(File), hold(Float), hold(DefaultWeight),
       hold(MainTerminal), hold(MainNonTerminal), hold(Target),
       hold(Lazy), hold(Holonomic), hold(Naive), hold(Auto),
       hold(Count), hold(Random), hold(Unrank), hold(Generator), hold(GeneratorFromRank),
       // Graph.pm from MuPAD 3.0.0
       hold(SearchWith), hold(Bellman),
       hold(StartVertex), hold(EndVertex), hold(EdgeWeights),
       // output::tableForm from MuPAD 3.0.0
       hold(Sort),

       // operators::overloaded
       hold(Explicit), hold(Facade), hold(Cost),
       // operators::make*
       hold(Source), hold(Source1), hold(Source2), hold(ImageSet), hold(Zero), hold(Opposite),
       //skewPartitions
       hold(Connected),
       // prog/combinat::getOptions
       hold(Error), hold(Warning),
       //
       hold(ColumnElimination),
       // examples::SymmetricFunctions
       hold(Lrcalc),
       // combinat::graphs,
       hold(C3Free), hold(C4Free), hold(Bipartite), hold(Biconnected),
       hold(Quiet), hold(Straighten), hold(MinDegree), hold(MaxDegree),
       hold(MinEdges), hold(MaxEdges), hold(GengOptions),
       // combinat::subClass
       hold(InheritsWithParameters), hold(InheritsWithoutParameters),
       hold(Parse), hold(Size), hold(Parameters),

       // Cat::ModuleWithBasis
       hold(ColumnVectors), hold(ColumnIndices), hold(ColumnArgs),
       hold(RowIndices), hold(RowArgs), hold(RowRank)
      }:

    // check if the options do not already have some values
    for opt in packageOptions do
	if opt <> eval(opt) then
	    warning("The option ". expr2text(opt).
		    " has the value ". expr2text(eval(opt)).
		    " before the loading of combinat.\n".
		    " Some functions of the package may behave unexpectedly."
		   );
	end_if;
    end_for;
    stdlib::OPTIONS   :=  stdlib::OPTIONS  union packageOptions:
    map(packageOptions, protect, Warning):
end_proc():

if not contains(stdlib::PROTECTED, hold(testeq)) then
    protect(testeq):
end_if:
if not contains(stdlib::PROTECTED, hold(evalAt)) then
    protect(evalAt):
end_if:

/*****************************************************************************
   Compatibility patches for the MuPAD library
*****************************************************************************/

//////////////////////////////////////////////////////////////////////////////
// Autoload of standard dynamic module with MuPAD <= 3.0.0
alias(path = pathname("MODULE")):
if loadmod() then
    if not contains(stdlib::PROTECTED, hold(shell)) then
	shell := loaddomain(shell, path, "shell");
	protect(shell);
    end_if;
end_if:
unalias(path):

//////////////////////////////////////////////////////////////////////////////
// Backport of matrix stuff for MuPAD < 3.0.0 15/09/2003
// Support for semi rings, sparse matrices, ...


if not MUPAD_HAS_SEMIRING then
    Cat::Matrix :=
    loaddomain(Cat::Matrix, pathname("MUPADBACKPORT", "DOMAINS", "CATEGORY"), "MatCat"):
    Dom::DenseMatrix :=
    loaddomain(Dom::DenseMatrix, pathname("MUPADBACKPORT", "DOMAINS", "DOMAIN"), "DenseMatrix"):
    if not contains(Dom, "SparseMatrixOld") then
	// MuPAD < 2.5
	Dom::SparseMatrixOld := ()->Dom::SparseMatrix(args());
    end_if;
end_if:
if not MuPADVersionAtLeast([3,1,2]) then
    // Feature: matrix::internalPolyVar 10/12/2004
    Dom::Matrix :=
    loaddomain(Dom::Matrix, pathname("MUPADBACKPORT", "DOMAINS", "DOMAIN"), "Matrix"):
    // Bug fix: call to subClass 11/2004
    Dom::SquareMatrix :=
    loaddomain(Dom::SquareMatrix, pathname("MUPADBACKPORT", "DOMAINS", "DOMAIN"), "SqMatrix"):
end_if:

//////////////////////////////////////////////////////////////////////////////
// Backport of Graph for MuPAD <= 3.0.0 05/09/2003
if domtype(Graph)<>DOM_DOMAIN or Graph::addEdges=FAIL then
    Graph := loaddomain(Graph, pathname("MUPADBACKPORT", "LIBFILES"), "Graph"):
    protect(Graph);
end_if:

if generate::sortSums = FAIL then
    generate::sortSums :=
    loadproc(generate::sortSums, pathname("MUPADBACKPORT", "GENERATE"), "sortsum"):
end_if:

if prog::sort = FAIL then
    prog::sort :=
    loadproc(prog::sort, pathname("MUPADBACKPORT", "PROG"), "sort"):
end_if:

/*
// Define mapcoeffs and multcoeffs for matrices
// That forces the loading of Dom::Matrix and Dom::DenseMatrix; not good.
// Better change the files themselves, and backport, as soon as we get
// Walter's agreement.
slot(extop(Dom::Matrix,4), "mapcoeffs", map):
slot(extop(Dom::Matrix,4), "multcoeffs", _mult):
slot(extop(Dom::DenseMatrix,4), "mapcoeffs", map):
slot(extop(Dom::DenseMatrix,4), "multcoeffs", _mult):
*/

Dom::SparseMatrix := () -> Dom::Matrix(args()):
// Should use loadproc's
if domtype(sparsematrix) <> DOM_DOMAIN then
    sparsematrix := Dom::Matrix():
    protect(sparsematrix):
end_if:
if domtype(densematrix) <> DOM_DOMAIN then
    densematrix := Dom::DenseMatrix():
    protect(densematrix):
end_if:
if domtype(matrix) <> DOM_DOMAIN then
    protect(matrix):
end_if:
matrix := Dom::Matrix():
/*
// Make sure that the methods are defined properly for the domains
// that already have been instantiated
for matrixdomain in [sparsematrix, densematrix] do
    for method in ["mapcoeffs", "multcoeffs"] do
	if slot(matrixdomain, method) = FAIL then
	    slot(matrixdomain, method,
		 DomainConstructor::getEntryFromSuperDomain
		 (matrixdomain, matrixdomain::constructor, method));
	end_if;
    end_for;
end_for:
*/
// Dom::DenseMatrix makes reference to Simplify, without actually using
// it unless asked explicitely for it. Simplify does not exists in
// MuPAD < 3.0.0, so the following silents prog::check
if domtype(Simplify) <> DOM_DOMAIN then
    protect(Simplify);
end_if:

//////////////////////////////////////////////////////////////////////////////
// Bug fix backport for Dom::ArithmeticalExpression in MuPAD < 3.0.0
if not testtype(a([1]), Dom::ArithmeticalExpression) then
    Dom::ArithmeticalExpression::convert :=
    proc(e)
    begin
	if e::dom::expr <> FAIL then
	    return(e::dom::expr(e))
	end_if;
	
	if domtype(e) <> DOM_EXPR then
	    if testtype(e, Type::Arithmetical) = TRUE then
		return(e)
	    else
		return(FAIL)
	    end_if
	else
	    // assume everything is ok
	    e
	end_if;
    end_proc;
end_if:

//////////////////////////////////////////////////////////////////////////////
// Bug fixes backports for Dom::ExpressionField() in MuPAD < 3.0.0
if (Dom::ExpressionField())::equal(a[1],a[2])=TRUE then
    slot(Dom::ExpressionField(), "equal",
	 proc(x,y)
	 begin
	     x:= x - y;
	     if testtype(x,Type::Numeric) then
		 return(iszero(x))
	     else
		 return(is(x,Type::Zero))
	     end_if
	 end_proc):
end_if:

if (Dom::ExpressionField()(hold(f)(poly(x))))<>hold(f)(poly(x)) then
    slot(Dom::ExpressionField(), "convert",
	 Dom::ArithmeticalExpression::convert);
end_if:

//////////////////////////////////////////////////////////////////////////////
// In the test files, we use prog::testinit with several arguments to
// do timing checks. This feature of prog::testinit appears in MuPAD
// 2.5.0, while with former versions of MuPAD, calling prog::testinit
// with several arguments raises an error. So we strip the extra
// arguments.
//
// It would be better to do a feature test than a version test, but I
// haven't found a way to do that without having prog::testinit create
// a temporary file as a side effect.
if not MuPADVersionAtLeast([2,5,0]) then
    prog::testinitOrig := prog::testinit;
    prog::testinit := s->prog::testinitOrig(s);
end_if:

//////////////////////////////////////////////////////////////////////////////
// Backport of the improved info mechanism (26/11/2004)
// Also fixes a bug in MuPAD 2.0.0: info(combinat::partitions) raises an error
// (see mail discussion with Ralf on Feb 16 2001)

info :=
loadproc(info, pathname("MUPADBACKPORT", "STDLIB"), "info"):

// The new info mechanism requires output::tableForm with the Append option.
// Backport if if needed
proc()
    local oldtestargs;
begin
    oldtestargs := testargs(TRUE);
    if  traperror(output::tableForm({}, Append=",")) <> 0 then
	output::tableForm :=
	loadproc(output::tableForm, pathname("MUPADBACKPORT", "OUTPUT"), "tabForm"):
    end_if:
    testargs(oldtestargs);
end_proc():

//////////////////////////////////////////////////////////////////////////////
// Backport of Type::PolyExpr which is buggy in MuPAD 2.0.0
proc()
    local oldtestargs;
begin
    oldtestargs := testargs(TRUE);
    if  traperror(solvelib::solve_islinear([a(1)+3*a(3)], [a(1), a(3)])) <> 0 then
	solvelib::solve_islinear := loadproc(solvelib::solve_islinear, pathname("MUPADBACKPORT", "SOLVELIB"), "islin");
    end_if:
    testargs(oldtestargs);
end_proc():

linalg::matlinsolveLUOld := linalg::matlinsolveLU:
linalg::matlinsolveLU :=
proc(L,U,p: Type::ListOf(Type::PosInt), //combinat::permutations
     b)
    local i;
begin
    if args(0) = 3 then
	b := p;
    else
	b := b::dom([b[p[i],1] $i=1..linalg::nrows(b)]);
    end_if;
    linalg::matlinsolveLUOld(L,U,b);
end_proc:

//////////////////////////////////////////////////////////////////////////////
// Backport of poly2list which can't be overloaded in MuPAD < 3.0.0
// (Done at the library level, obviously)

// The feature test below is not too bad; the poly2list method
// has been added to Dom::DistributedPolynomial not so long after
// the change in poly2list (in 3.1.0)

if not contains(extop(Dom::DistributedPolynomial,4), "poly2list") then
    poly2listKernel := poly2list;
    poly2list :=
    proc(p: Type::AnyType)
    begin
	if domtype(p)=DOM_POLY then
	    return(poly2listKernel(p));
	end_if;
	if p::dom::poly2list<>FAIL then
	    return(p::dom::poly2list(p));
	end_if;
	if domtype(extop(p,1))=DOM_POLY then
	    // Fixes previously missing poly2list method in
	    // Dom::UnivariatePolynomial & such
	    return(poly2listKernel(extop(p,1)));
	end_if;
	poly2list(args());
    end_proc:
end_if:

//////////////////////////////////////////////////////////////////////////////
// This let prog::check know that muEC, ... are not global variables.
// With MuPAD > 2.0.0, this is done automatically by protect.
if not contains(stdlib::PROTECTED, hold(muEC)) then
    stdlib::PROTECTED := stdlib::PROTECTED
	union {hold(Graph),
	       hold(muEC),
	       hold(CAT_EXEC), // referenced by accident ? in DOM_EXPR::TeX
	       hold(operators), hold(examples),
	       hold(experimental), hold(IPC),
	       hold(densematrix), hold(sparsematrix),
	       hold(Symmetrica), hold(gordan), hold(shell),
	       // In MuPAD 2.0.0, assert is not defined, and
	       // a call to assert(bla) simply does nothing except
	       // evaluating its argument. We can safely leave it as
	       // is without providing any implementation of assert.
	       // We add it to stdlib::PROTECT to silence prog::check.
	       hold(assert),
	       // For some reason MuPAD >2.0.0 does not complain about
	       // the use of the variable #FloatCheck in Dom::Matrix.
	       hold(`#FloatCheck`), hold(DOM_INTERVAL), hold(frandom), hold(ReturnType)
	      }:
end_if:

//////////////////////////////////////////////////////////////////////////////
// The deprecated warnings are turned on by default when the \mupad-Combinat
// package is explicitly loaded by the user. Otherwise, the default
// setting is initialized from the value of \Mex{Pref::warnChanges()}. }
//
// This ensures some coherency with Pref::warnChanges(), when combinat
// is used as part of the MuPAD library.
combinat::warnDeprecated(TRUE):

//////////////////////////////////////////////////////////////////////////////
// Backward compatibility aliases
//
// Whenever possible, functions should be used so that defining
// the aliases does not trigger the loadproc
//
// TODO: add deprecated warnings

combinat::structure		   := Cat::CombinatorialClass:
combinat::decomposableStructure	   := Cat::DecomposableClass:
combinat::integerListsLexStructure := Cat::IntegerListsLexClass:
Dom::Graph			   := Graph:

Dom::FreeModule_Poly	      := () -> Dom::FreeModulePoly(args()):
Dom::FreeModule_Table	      := () -> Dom::FreeModuleTable(args()):
Dom::FreeModule_List	      := () -> Dom::FreeModuleList(args()):
Dom::FreeModuleMnPoly	      := () -> error("deprecated"):
Cat::FreeModule               := () -> Cat::ModuleWithBasis(args()):
//Cat::CombinatorialHopfAlgebra := () -> Cat::HopfAlgebraWithBasis_old(args()):

//////////////////////////////////////////////////////////////////////////////
// Force loading of operators::overloaded, to allow for hacking
// some domain constructors
operators::overloaded:

// The return value of the package call:
null():
