/*****************************************************************************
  $Id: decomposableStructures.mu,v 1.3 2003/02/09 17:09:37 nthiery Exp $
  File:		lib/COMBINAT/decomposableStructures.mu
  Authors:	Sbastien Cellier <snoteip@yahoo.fr>
		Nicolas M. Thiry <nthiery@users.sourceforge.net>
  Created:	16/05/2002
  Adapted from:	CS package v1.5 for MuPAD 1.4.2
		http://dept-info.labri.u-bordeaux.fr/~dutour/CS
		Sylvie Corteel - LRI
		Alain Denise - LRI
		Isabelle Dutour - LaBRI
		Frdric Sarron - LRI
		Paul Zimmermann - INRIA
  License:	LGPL
*****************************************************************************/

/** specification is a set or list of productions of the form
    A = Epsilon
    A = Atom
    A = Union(B,C,...)
    A = Prod(B,C,...)
    A = Set(B)
    A = Sequence(B)
    A = Predefined(function)

Old doc from CS:
compile(s [, type=labelled] [, target=C] [, mode=graphics] [, library=GMP] 
          [, file="B.c"] [, main=NT]) 
- type is either unlabelled (default) or labelled
- target is either MuPAD (default) or C or MODULE or JAVA
- mode is either text (default) or graphics or applet (in JAVA mode)
- library is either DPE (default) or GMP
- file (string) is the name of the output file (in C, MODULE or JAVA mode)
- main is a non-terminal (in C or JAVA mode)

        >> dom::standardForm({M=Union(Z,Prod(Z,M),Prod(Z,M,M))},unlabelled);

**/

domain combinat::decomposableStructures
    (Specification: Type::Union(DOM_SET,DOM_LIST))

    local Options;
    category combinat::structure;

    info := "Parametrized library 'combinat::decomposableStructures': utilities for decomposable combinatorial structures";

    interface := dom::interface_Structure union 
	{hold(standardForm), hold(standardForm_Count),
	 hold(unrank), hold(random),
	 hold(recurrenceRelation)
	 };

    options := Options;

    specification := Specification;

    count_functions  := table(); // to be renamed ?
    unrank_functions := table(); // to be renamed ?

    //////////////////////////////////////////////////////////////////////////
    // Utilities, ...
    //////////////////////////////////////////////////////////////////////////

    // Types allowed for nonterminals 
    type := Type::AnyType;
    typeStandard := Type::AnyType;
    typeNonTerminal := Type::Union(DOM_IDENT,DOM_FUNC_ENV);
    // Why do we allow DOM_FUNC_ENV ?

    // All the variations on the binary product in a standard specification
    prodlist := {Prod,ProdPrd,ProdSeq,ProdPrdSeq,ProdSet,ProdPrdSet};

    Next:=table(Prod=ProdPrd,
                ProdPrd=ProdPrd,
                ProdSeq=ProdPrdSeq,
                ProdPrdSeq=ProdPrdSeq);

    contextFreeProductions:={Atom,Epsilon,Alias,Union} union dom::prodlist;

    typeofoper :=
    proc(p)
    /**
	Precondition: p is the right hand side of a production 
	of the form A=B or A=Bla(...).
	Returns the type of the production
    **/
    begin
	if testtype(p, dom::typeNonTerminal) then
	    return(NonTerminal);
	end_if;
	if type(p)<>"function" then
	    error("invalid production:", p);
	end_if;
	if contains(dom::prodlist,op(p,0)) then 
	    return(Prod) 
	else
	    return(op(p,0)) 
	end_if
    end_proc;

    // Generates fresh nonterminals named NonTerminal1, NonTerminal2, ...
    genident :=
    proc()
	local i;
	option escape;
    begin
	i:=0;
	() -> (i:=i+1; hold(NonTerminal).i);
    end_proc();

    //////////////////////////////////////////////////////////////////////////
    // Utilities for standard specifications
    //////////////////////////////////////////////////////////////////////////

    /**
        A standard specification is a set of productions of the form
	A = Epsilon(B) 		or A = Epsilon() 
        A = Atom(B)		or A = Atom()
        A = Alias(B)
        A = Union(B,C,...)
        A = Prod(B,C)		or ProdSeq(B,C), or ...
        A = Theta(B)
        A = Int(B)
	A = Delta(B,u)
	A = Predefined(f)
    **/
    typeStandardSpecification :=
        Type::Predicate(dom::testtypeStandardSpecification);
    testtypeStandardSpecification :=
    proc(specification) : DOM_BOOL
	local used, defined, prod, t;
    begin
	if not testtype(specification, 
			Type::TableOfIndex(dom::typeNonTerminal)) then
	    return(FALSE);
	end_if;
	defined := {};  // the non-terminals that are defined in the specif
	used := {};     // the non-terminals that are used in the specif
	for prod in specification do
	    defined := defined union({lhs(prod)});
	    prod := rhs(prod);
	    if type(prod)<>"function" then return(FALSE); end_if;
	    t:=dom::typeofoper(prod);
	    // Check the number of arguments
	    if contains({Epsilon, Atom}, t) then
		if nops(prod) > 1 then return(FALSE); end;
	    elif contains({Alias, Theta, Int, Predefined}, t) then
		if nops(prod) <> 1 then return(FALSE); end;
	    elif contains({Delta, Prod}, t) then
		if nops(prod) <> 2 then return(FALSE); end;
	    elif t<>Union then
		return(FALSE);
	    end_if;
	    // Check that the arguments are all non terminals
	    // As an exception, the second argument of Delta is not a
	    // non terminal, as well as the first argument of Predefined
	    if t=Delta then
		prod := subsop(prod,2=null());
	    elif contains({Predefined, Atom, Epsilon}, t) and nops(prod)=1 then
		prod := subsop(prod,1=null());
	    end_if;
	    if not _and(testtype(A, dom::typeNonTerminal)$A in [op(prod)]) then
		return(FALSE);
	    end_if;
	    used := used union {op(prod)};
	end_for;
	// Finally, all the used non terminals should be defined
	bool(used minus defined = {});
    end_proc;

    optimizeStandardSpecification :=
    proc(specification: dom::typeStandardSpecification)
	: dom::typeStandardSpecification
    /**
	Returns an equivalent standard specification which is hopefully
        smaller. Currently, this is achieved by factoring common
        production rules through aliases.
    **/
	local aliases, newaliases, reverse, modified, eq;
    begin
        // Extract the aliases of the form A=B from the specification
	specification := split(specification,
			       eq -> op(eq,[2,0]) = Alias);
	aliases       := map(specification[1], op, 1);
	// aliases now contain something of the form A=B, B=C
	specification := specification[2];
    
        // Simplify the aliases by transitivity, and apply them on the
        // rest of the specification
	while TRUE do
	    newaliases := subs(aliases, aliases);
	    if newaliases=aliases then break; end_if;
	    aliases := newaliases;
	end_while;
	// aliases now contain something of the form A=C, B=C
	specification := subs(specification, aliases);

        // Try to find a duplicate in the specification, such as
        // A1=Prod(B,C) and A2=Prod(B,C), and replace it by an alias
        // A2=A1. Repeat this process until nothing changes anymore
	repeat
            modified := FALSE;
            reverse := table();
	    for eq in specification do
		if contains(reverse, op(eq,2)) then
		    aliases[op(eq,1)] := reverse[op(eq,2)];
		    delete(specification[op(eq,1)]);
		    specification := subs(specification, 
					  op(eq,1)=reverse[op(eq,2)]);
		    aliases := subs(aliases, 
				    op(eq,1)=reverse[op(eq,2)]);
		    modified := TRUE;
		    break;
		end_if;
		reverse[op(eq,2)] := op(eq,1);
	    end_for;
	until not modified
	end_repeat;
	table(op(specification), op(map(aliases, Alias)));
    end_proc;
    
    /**
        A standard specification is context free if it only uses the
	productions in dom::contextFreeProductions
    **/
    
    typeContextFreeStandardSpecification :=
        Type::Predicate(dom::testtypeContextFreeStandardSpecification);
    
    testtypeContextFreeStandardSpecification :=
    proc(specification) : DOM_BOOL
    begin
	if not testtype(specification, dom::typeStandardSpecification) then
	    return(FALSE);
	end_if;
	_lazy_and(contains(dom::contextFreeProductions, op(prod,[2,0]))
		  $ prod in specification);
    end_proc;

    isContextFree := 
    proc()
//	option remember;
    begin
	testtype(dom::standardForm_Count(),
		 dom::typeContextFreeStandardSpecification)
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // Standard form of the specification
    //////////////////////////////////////////////////////////////////////////
    
    standardForm :=
    proc() : dom::typeStandardSpecification
    /**
       Returns a standard specification which is equivalent.
    **/
	option remember;
	local t,p,spec;
    begin
	dom::standardForm2;    // hacky bug workaround, see standardForm2
	dom::aliases:=table(); // used locally by subprocedures of standardForm1
	spec:=dom::specification;

	// Syntactic sugar: 
        //  - if Z is used but not defined, define it by default as an Atom
	if has(spec,hold(Z)) and not contains(map(spec,op,1),hold(Z)) then
	    spec:=spec union {hold(Z)=Atom} 
	end_if;

        //  - transform A = Atom    into A = Atom(A)
	//              A = Epsilon into A = Epsilon(A) 
	spec := map(spec, p -> case rhs(p)
				 of Atom do
				 of Epsilon do
				     lhs(p)=rhs(p)(lhs(p)); 
				     break;
				 otherwise p;
			       end_case);
	//  - transform the remaining lonely Atom's or Epsilon's
	//    into function calls Epsilon() and Atom()
	spec := misc::maprec(spec,
			     (p->bool(p=Epsilon or p=Atom)) =
			     (p -> p()), 
			     NoOperators);
	t:={};
	for p in spec do
	    if not testtype(p, Type::Equation(dom::typeNonTerminal)) then
		error("invalid production: ".expr2text(p))
	    end_if;
	    t:=t union dom::standardForm2(op(p))
	end_for;
	t:=table(op(t));
	if not testtype(t, dom::typeStandardSpecification) then
	    // Something went wrong; the diagnostic should be improved!
	    error("Invalid standard specification: ". expr2text(t));
	end_if;
	dom::optimizeStandardSpecification(t);
    end_proc;

    // returns a set of productions for the standard form of {A=p} 
    standardForm2 := 
    proc(A,p) 
	// The option remember should not be mandatory;
	// This is just there to work around a bug of MuPAD:
	// The result of dom::standardForm() is not remembered the
	// first time it's called, and moreover the subsequent results
	// are different because of the call to dom::genident()
	option remember;
	local q;
    begin
	case dom::typeofoper(p)
	  of Epsilon do
	  of Atom do
	      return({A=p});
	  of NonTerminal do 
	      return({A=Alias(p)});
	  of Union do return(dom::standardForm3([op(p)],{A=p}))
	  of Prod do
	      return((
		      case nops(p)
			of 0 do {A=Epsilon(Epsilon)}; break
			of 1 do dom::standardForm2(A,op(p)); break
			of 2 do dom::standardForm3([op(p)],{A=p},A,p); break
			otherwise
             // first look for existing aliases 
             /* not used any more because of the decomposition with ProdPrd, ...
               to be changed...
             for i from 2 to nops(p) do
                if type((q:=dom::aliases[op(p,0)(op(p,i-1..i))]))<>"_index" then
                   return(dom::standardForm2(A,subsop(p,i=q,i-1=null())))
                end_if
             end_for;
             */
			   dom::standardForm2(A,op(p,0)(op(p,1),dom::Next[op(p,0)](op(p,2..nops(p)))))
			  end_case))
	  of Theta do 
	  of Int do
	      if nops(p)<>1 then 
		  error("invalid right-hand side of production: ".expr2text(p)) 
	      end_if;
	      return(dom::standardForm3([op(p)],{A=p}))
	  of Delta do
	      if nops(p)<>2 then
		  error("invalid right-hand side of production: ".expr2text(p)) 
	      end_if;
	      return(dom::standardForm3([op(p,1)],{A=p}))
	  otherwise
	      q:=slot(dom,"standardForm_".op(p,0));
	      if domtype(q)<>DOM_PROC then
		  error("invalid right-hand side of production: ".expr2text(p));
	      end_if;
	      q:=q(A,op(p));
	      return((if q=p then {A=p} else dom::standardForm2(A,q) end_if))
	end_case;
    end_proc;

/* for each p in l, if it is not an identifier, creates a new nonterminal
   and substitutes p by it in s */
    standardForm3 := 
    proc(l,s) 
	local p,A,t,oldl;
    begin
    /* first puts everything in l in the form A or f(A,B,...) where
      A,B,... are identifiers */
	l:={op(l)}; t:=null();
	repeat
	    oldl:=l;
	    l:=map(l,proc(x) 
		     begin 
			 if type(x)=DOM_IDENT then
			     x 
			 else
			     x,op(select([op(x)],testtype,"function")) 
			 end_if
		     end_proc);
	until l=oldl end_repeat;
        // then sort by increasing size 
	l:=sort([op(l)], (x,y)->(length(x)<length(y)));
        // then generate aliases by increasing order 
	while l<>[] do
	    p:=l[1]; delete(l[1]);
	    if not testtype(p, dom::typeNonTerminal) then
		if not contains(dom::aliases, p) then
		    A:=dom::genident();
		    dom::aliases[p]:=A
		else
		    A:=dom::aliases[p] // already an alias 
		end_if;
		t:=t,p=A; s:=subs(s,p=A) union dom::standardForm2(A,p);
		l:=subs(l,p=A);
	    end_if
	end_while;
	if args(0)>=4 then 
	    dom::aliases[subs(args(4),t)]:=args(3) 
	end_if;
        // remove trivial expressions 
	select(s, not bool);
    end_proc;

    // standard form for A=Sequence(B,c) 
    standardForm_Sequence :=
    proc(A,B,c) 
	local k;
    begin
	case args(0)
	  of 0 do 
	  of 1 do break;
	  of 2 do 
	      return(Union(Epsilon(),
				 ProdSeq(B,hold(Sequence1)(B))))
	  otherwise // c=Card=k or c=Card>=k or Card<=k 
	      if type(c)="_equal" then // Card=k 
		  k:=op(c,2);
		  if testtype(k,Type::NonNegInt) then
		      c:=Epsilon();
		      while k>1 do c:=ProdPrdSeq(B,c); k:=k-1 end_while;
		      if k>0 then c:=ProdSeq(B,c) end_if;
		      return(c)
		  end_if
	      elif type(c)="_leequal" then
		  if op(c,2)=hold(Card) then // k<=Card 
		      k:=op(c,1);
		      if testtype(k,Type::NonNegInt) then
			  return(ProdSeq(B$k,hold(Sequence1)(B)))
		      end_if
		  elif op(c,1)=hold(Card) then // Card<=k 
		      k:=op(c,2); 
		      if testtype(k,Type::NonNegInt) then
			  c:=Epsilon();
			  while k>1 do c:=Union(Epsilon(),ProdPrdSeq(B,c)); k:=k-1 end_while;
			  if k>0 then c:=Union(Epsilon(),ProdSeq(B,c)) end_if;
			  return(c)
		      end_if
		  end_if
	      end_if
	end_case; error("invalid production: ".expr2text(Sequence(B,c)))
    end_proc;

    standardForm_Sequence1 :=
    proc(A,B)
    begin
	Union(Epsilon(),ProdPrdSeq(B, A))
    end_proc;

    // standard form for A=Set(B,c) 
    standardForm_Set :=
    proc(A,B,c) 
	local k,C,i;
    begin
	if dom::options[Labelled] then
	    C:=Theta(B); 
	else
	    C:=Delta(Theta(B), ()->1) ;
	end_if;
	case args(0)
	  of 0 do 
	  of 1 do break;
	  of 2 do return(Union(Epsilon(),
			       Int(ProdSet(C,hold(Set1)(B)))))
	  otherwise // c=Card>=k or Card<=k 
              if type(c)="_leequal" then
		  if op(c,2)=hold(Card) then // k<=Card 
		      k:=op(c,1);
		      if testtype(k,Type::NonNegInt) then
			  error("not yet implemented")
		      end_if
		  elif op(c,1)=hold(Card) then // Card<=k 
		      k:=op(c,2); 
		      if testtype(k,Type::NonNegInt) then
			  error("not yet implemented")
		      end_if
		  end_if
	      end_if
	end_case; error("invalid production: ".expr2text(Set(B,c)))
    end_proc;

    standardForm_Set1 :=
    proc(A,B) 
	local C;
    begin
	if dom::options[Labelled] then
	    C:=Theta(B); 
	else
	    C:=Delta(Theta(B), ()->1) ;
	end_if;
	Union(Epsilon(),Int(ProdPrdSet(C,A)))
    end_proc;

    // A = Predefined(B) 
    standardForm_Predefined := 
    proc() 
    begin 
	Predefined(args(2..args(0))) 
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // Simplified standard form of the specification for counting
    //////////////////////////////////////////////////////////////////////////

    /**
	standardForm_Count () -> standard specification

	Returns a simplified form of the specification in standard
        form which is equivalent for counting purposes. 

	This is done by erasing the names of the singletons and atoms,
	and forgeting about the different kinds of products, and
        factoring further the result.
    **/
    standardForm_Count :=
    proc() : dom::typeStandardSpecification
	option remember;
	local DOM;
    begin
	DOM := dom;
	dom::optimizeStandardSpecification(
	    map(dom::standardForm(),
		p -> case DOM::typeofoper(p)
		       of Atom    do Atom();           break;
		       of Epsilon do Epsilon();        break;
		       of Prod    do Prod(op(p,1..2)); break;
		       otherwise           p;
		     end_case)
				);
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // count, random, unrank, generator, ...
    //////////////////////////////////////////////////////////////////////////

    count := 
    (
     dom::makeCount();
     proc(n: Type::NonNegInt, terminal=dom::options[MainTerminal])
	 local i,f;
     begin
	 f := dom::count_functions[terminal];
	 // Hack to avoid deep recursion
	 //
	 // If the counting function has an option remember, we try to
	 // compute the missing values in the remember table in order,
	 // so that there is no recursive call.
	 //
	 // For simplicity, we assume that the values stored in the
	 // remember table are contiguous: 0,1,2,3,... If this is not
	 // the case, that won't break the code, but some recursion
	 // may occur.
	 if has([extop(f,3)], hold(remember)) then
	     for i from nops(extop(f,5)) to n-1 do
		 dom::count_functions[terminal](n);
	     end_for;
	 end_if;
	 dom::count_functions[terminal](n);
     end_proc;
    );

    unrank := 
    (
     dom::makeUnrank();
     proc(r: Type::PosInt, n: Type::NonNegInt, terminal=dom::options[MainTerminal])
     begin
	 if r>dom::count_functions[terminal](n) then 
	     return(FAIL);
	 end_if;
	 dom::unstandard(dom::unrank_functions[terminal](n,r))
     end_proc;
    );

    random :=
    proc(n: Type::NonNegInt, terminal=dom::options[MainTerminal])
	local r;
    begin
	r := dom::count(n, terminal);
	if r = 0 then
	    return(FAIL);
	end_if;
	dom::unrank(random(r)()+1, n, terminal)
    end_proc;

    generator :=
    proc(n: Type::NonNegInt, terminal=dom::options[MainTerminal])
	local DOM, r;
	option escape;
    begin
	DOM := dom;
	r := 0;
        () -> (r:=r+1; DOM::unrank(r, n, terminal));
    end_proc;

    // Not necessary anymore; the job is done automatically when calling
    // count, unrank, ...
    compile := 
    proc()
	option remember;
    begin
	dom::standardForm();
	dom::standardForm_Count();
	userinfo(1,"standard form: ", dom::standardForm());
	userinfo(1,"standard form for counting: ", dom::standardForm_Count());
	dom::init();
	dom::makeValuation();
	dom::makeCount();
	dom::makeUnrank();
        //dom::reset();
    end_proc;

    recurrenceRelation :=
    proc(terminal=dom::options[MainTerminal],
	 u = hold(u) : DOM_IDENT,
	 n = hold(n) : DOM_IDENT)
    begin
	dom::speciftorec(dom::standardForm_Count(),terminal,u(n));
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // Tools for the holonomic product
    //////////////////////////////////////////////////////////////////////////

    z  := hold(z);
    Q  := Dom::Fraction(Dom::DistributedPolynomial([dom::z],Dom::Rational));
    MQ := Dom::DenseMatrix();

/* p=Epsilon(A) or Atom(A) or Union(B,C,...)
  or Prod(B,C) or ProdPrd(B,C) 
  or ProdSeq(B,C) or ProdPrdSeq(B,C) 
  or ProdSet(B,C) or ProdPrdSet(B,C) 
  or Theta(B) or Int(B) or Delta(B,u)
*/


// useful functions for getting the list of indices or values of a table t 
listofindices:=
proc(t) begin map([op(t)], op, 1) end_proc;
listofvalues:=
proc(t) begin map([op(t)], op, 2) end_proc;


    //////////////////////////////////////////////////////////////////////////
    // Pieces of cs1....mu
    //////////////////////////////////////////////////////////////////////////

    init:=proc() begin TRUE; end_proc;

    reset:=proc() begin TRUE; end_proc;

    postProcess := 
    if dom::options[Labelled] then
        proc(obj) begin obj end_proc // TODO!
    else
        proc(obj) begin obj end_proc;
    end_if;

    /**
	unstandard(object: dom::typeStandard): dom

	Precondition: object satisfies the specification in standard form
        Returns the corresponding object statisfying the specification
    **/
    unstandard :=
    proc(object: dom::typeStandard): dom
    begin
	case type(object) 
	  of "function" do
	      object := map(object, dom::unstandard);
	      case op(object,0) 
		of ProdPrd do
		of ProdPrdSeq do
		of ProdPrdSet do
		    return(op(object));
		of ProdSeq do
		    if op(object,nops(object))=Epsilon then
		        // ProdSeq(B,C, ..., Epsilon): ignore the Epsilon 
			return(hold(Sequence)(op(object,1..nops(object)-1)));
		    else
			return(hold(Sequence)(op(object)));
		    end_if;
		of ProdSet do
		    if op(object,nops(object))=Epsilon then
		        // ProdSet(B,C, ..., Epsilon): ignore the Epsilon 
			return(hold(Set)(op(object,1..nops(object)-1)));
		    else
			return(hold(Set)(op(object)));
		    end_if;
	      end_case;
	end_case;
	return(object);
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // Constructors for valuation, count, unrank, and random
    //////////////////////////////////////////////////////////////////////////

    /**

        Compute the valuation and degree of each non-terminal and
        store them in dom::valuation and dom::degree. Then, for each
        production rule A = ... we have

	dom::valuation[A] <= size of any element of A <= dom::degree[A]
    */
    makeValuation := 
    proc() 
	option remember;
	local v, d, newv, newd, s, p, f, changed, A, Rhs, t, aliases;
    begin
	s := dom::standardForm_Count();
	v := table();  // v, d: temporary values for dom::valuation and dom::degree
	d := table();
	aliases := table(); // production rules which are aliases
	for p in s do
	    A := lhs(p);
	    case op(rhs(p), 0) 
	      of Epsilon do v[A]:=0; d[A]:=0; delete(s[A]);  break;
	      of Atom    do v[A]:=1; d[A]:=1; delete(s[A]);  break;
	      of Alias   do aliases[A]:=s[A]; delete(s[A]);  break;
	      otherwise 	  v[A]:=infinity; d[A]:=infinity;
	    end_case;
	end_for;
	repeat
	    changed:=FALSE;
	    for p in s do
		A := lhs(p);
		Rhs := rhs(p);
		t:=dom::typeofoper(Rhs);
		case t
		  of Union do
		      newv := min(v[B] $ B in op(Rhs));
		      newd := max(d[B] $ B in op(Rhs));
		      break;
		  of Prod do
		      newv := v[op(Rhs,1)]+v[op(Rhs,2)];
		      newd := d[op(Rhs,1)]+d[op(Rhs,2)];
		      break;
		  of Theta do
		  of Int do  
		      newv := max(1,v[op(Rhs,1)]);
		      newd := d[op(Rhs,1)];
		      break;
		  of Delta do
		      newv := v[op(Rhs,1)];
		      newd := infinity;
		      break;
		  of Predefined do
		      newv := 0;
		      f := op(Rhs,1);
		      while f(newv)=0 do newv := newv+1; end_while;
		      newd := infinity;
		      break;      
		  otherwise   
		      error("unknown constructor: ".expr2text(t))
		end_case;
		if newv<v[A] then v[A]:=newv; changed:=TRUE end_if;
		if newd<d[A] then d[A]:=newd; changed:=TRUE end_if;
	    end_for
	until not changed end_repeat;
	// set valutation and degree for aliases
	for p in aliases do
	    v[lhs(p)] := v[op(rhs(p),1)];
	    d[lhs(p)] := d[op(rhs(p),1)];
	end_for;
	dom::valuation:=v; dom::degree:=d;
    end_proc;

    /** Specifications for the calls to constructors:

	A=Union(B,C,D)

	yields calls to:

	makeCount_Union(B, C, D)
	makeUnrank_Union(A, Union, B, C, D)
    **/

    makeCount := 
    proc() 
	option remember;
	local p, A, Rhs, t, f, i, r;
    begin
	dom::makeValuation();
	for p in dom::standardForm_Count() do
	    A   := lhs(p);
	    Rhs := rhs(p);
	    // Get the constructor for this type of right hand side
	    t:=dom::typeofoper(Rhs);
	    f:=slot(dom, "makeCount_".t);
	    assert(domtype(f) = DOM_PROC);
	    f:=f(op(Rhs));
	    // 
	    for i from 0 to dom::valuation[A]-1 do
		f(i):=0 
	    end_for;
            // simple hack to rename the function
	    dom::count_functions[A] := subsop(f,6=hold(dom::count_functions)[A]);
	end_for;
	
	if dom::options[Products]=Auto then
	    if dom::isContextFree() and not dom::options[Labelled] then
		dom::options[Products]:=Holonomic
	    else
		dom::options[Products]:=Naive // replace this by Lazy soon!
	    end_if;
	end_if;
	
	if dom::options[Products]=Holonomic then
	// ...
	    if testargs() then
		if not dom::isContextFree() then
		    error("Holonomic mode can only be used with context free specifications");
		end_if;
		if dom::options[Labelled] then
		    error("Holonomic and Labelled are currently mutually exclusive");
		end_if;
	    end_if;
	    for p in dom::standardForm_Count() do
		A   := lhs(p);
		if dom::typeofoper(rhs(p)) <> Prod then
		    next;
		end_if;
		r:=dom::speciftorec(dom::standardForm_Count(),A,hold(u)(hold(n)));
		dom::count_functions[A] := 
		    subsop(dom::rectoproc(r,hold(u)(hold(n)),dom::count_functions[A]),
			   6=hold(dom::count_functions)[A])
	    end_for;
	elif dom::options[Products]=Lazy then
	    error("Lazy product temporarily desactivated!");
	end_if;    
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // Utilities for Holonomic counting
    //////////////////////////////////////////////////////////////////////////

    rectoproc :=
    proc(r,uofn,initialValues)
	// r: recurrence relation
	// u: name of the function in the recurrence relation
	// initiaValues: function which returns the values to initiate the
	//               recurence
	local u,n,R,ORDER,lroots,i,result;
	option escape;
    begin
	u:=op(uofn,0);
	n:=op(uofn,1);
	R:=dom::formatrec(r,u,n);
	ORDER:=nops(R)-2;
	R:=subs(R,n=n-ORDER); 
	dom::LIM:=dom::LIM+ORDER;
	lroots:={};// Temporarily desactivated because of a bug in MuPAD3.0.0
//	lroots:=op(solve(R[nops(R)], [n], Domain=Z_),2);
	dom::LIM:=max(dom::LIM, ORDER, op(lroots));
	R := map(R, fp::unapply, n);
	result :=
        proc(n) 
	    local nmax,j;
	    option remember;
	begin
	    // Invariant: the indices in the remember table of result
	    // form an interval [0..nmax]
	    //nmax := max(op(map( op(op(result,5)), op, 1)));
	    nmax:=nops(op(result,5))-1;
	    for j from nmax+1 to n do
		result(j) := - _plus(R[1](j), R[i+1](j)*result(j-ORDER+i-1)$i=1..ORDER) / R[nops(R)](j);
	    end_for;
	end_proc;
	for n from 0 to dom::LIM do
	    result(n) := initialValues(n)
	end_for;
	n->result(n); // keeps result purely lexical without interaction with the big bad exterior
    end_proc;

    formatrec:=
    proc(r,u,n)
	local mi,ma,cr,i;
    begin
	mi:=dom::minindex(r,u,n);
	ma:=dom::maxindex(r,u,n);
	cr:=collect(r,[u(n+i)$i=mi..ma],normal);
	if mi<>0 then
	    cr:=subs(cr,n=n-mi);
	    ma:=ma-mi;
	    dom::LIM:=dom::LIM+mi;
	end_if;
	// I don't know if this test is really useful... 
	if has(map({op(r)},denom),n) then
	    cr:=collect(numer(normal(cr)),[u(n+i)$i=0..ma],normal)
	end_if;
	[subs(cr,[u(n+i)=0$i=0..ma]), coeff(cr,u(n+i),1)$i=0..ma];
    end_proc;
    
    speciftorec := 
    proc(s: dom::typeContextFreeStandardSpecification,
	 N: dom::typeNonTerminal,
	 uofn: DOM_EXPR // an unevaluated function call, like u(n)
	 )
	local algeq,deq;
    begin
	algeq:=dom::speciftoalgeq(s,N);
	deq:=dom::algeqtodiffeq(algeq,N(dom::z));
	dom::diffeqtorec(deq,N(dom::z),uofn)
    end_proc;
    
    speciftoalgeq := 
    proc(s: dom::typeContextFreeStandardSpecification,
	 NT: dom::typeNonTerminal)
	local g,vars,matA,V,nb_lines,pNT,pol,linear_depend,ens,i,T,tc;
    begin
        // computation of groebner basis
	([g,vars]) := dom::groebner_calculus(s);
        // successive normal forms up to linear dependencies 
	matA := [[1]];
	V := [1];  // involved terms 
	nb_lines := 1;
	pNT := poly(NT,vars,dom::Q);
	pol := poly(1,vars,dom::Q);
	linear_depend := FALSE;
	
	while not linear_depend do
	    nb_lines := nb_lines + 1;
	    pol := groebner::normalf(pol*pNT,g,DegInvLexOrder);
	    userinfo(2,"polynomial is: ",expr(pol));
	    ens := {expr(nthterm(pol,i)) $ i=1..nterms(pol)};
	    if nops(ens union {op(V)}) < nb_lines then
		linear_depend := TRUE;
	    end_if;
	    ens := ens minus {op(V)};
	    for T in ens do
		V := append(V, T);
		matA := map (matA, append, 0);
	    end_for;
	    
	    userinfo(2,"involved terms are: ",V);
	    for T in V do
		tc[T] := 0;
	    end_for;
	    for i from 1 to nterms(pol) do
		tc[expr(nthterm(pol,i))] := nthcoeff(pol,i);
	    end_for;
	    
	    matA := append(matA, [ tc[T] $ T in V ]);
	end_while;
	
  // matrix building and gauss elimination 
	for i from 1 to nops(matA) do
	    matA[i] := append(matA[i], NT^(i-1));
	end_for;
	matA:=Dom::DenseMatrix()(matA);
	matA:=linalg::gaussElim(matA);
	
  // take the algebraic equation (numer of the right coefficient) 
	i := 1;
	while not has(matA[i,i], NT) do
	    i := i+1;
	end_while;
	numer(matA[i,i]);
    end_proc;
    
    groebner_calculus := 
    proc(s: dom::typeContextFreeStandardSpecification)
	option remember;
	local vars;
    begin
	// Precondition: s is a standard specification which has been
	// simplified by standardForm_Count(); hence the only productions
	// are of the form:
	//  A=Alias(B)
	//  A=Epsilon()
	//  A=Atom()
	//  A=Union(B, C, ...)
	//  A=Prod(B, C)
	//
	// Returns a reduced system of polynomial equations between
	// the generating series of the non trivial non terminals
	// (i.e. Union and Prod)
	
	// Get rid of the aliases
	s:=select(s, not has, Alias);
	
	// Get rid of the trivial productions A=Atom() and A=Epsilon()
	// by backsubstitution into the others
	s:=split(s, eq -> rhs(eq) in {Atom(),Epsilon()});
	s:=subs(s[2], op(s[1]));

        // Transform the rewriting rules into polynomial equations
	s:=eval(subs(s,Atom()=dom::z, Epsilon()=1, 
		     Union=_plus, Prod=_mult));

	// Transform the polynomial equations into polynomials
	vars := [ lhs(eq) $ eq in s];
	userinfo(1,"vars are: ",vars);
	s := [rhs(eq)-lhs(eq) $ eq in s];
	userinfo(1,"polynomials are: ",s);
	s := map(s, poly, vars, dom::Q);
	
        // Reduction of the system by Groebner basis 
	s := groebner::gbasis(s, DegInvLexOrder);
	userinfo(1,"groebner basis is: ",map(s,expr));
	
	[s,vars];
    end_proc;
    
    algeqtodiffeq := 
    proc(algeq, yofx: DOM_EXPR)
	local y,x,pol,g,u,v,deg,deq,Y,d,matA,V,T,i,inits,tc;
    begin
	y := op(yofx,0); 
	x := op(yofx,1);
	pol := poly(algeq, [y]);
	([g,u,v]) := [gcdex(diff(pol,y),pol)];
	if has(expr(g),y) then 
	    return (dom::algeqtodiffeq(expr(normal(pol/g)),y(x),inits)); 
	end_if;
	deg := degree(pol);
	if deg <= 1 then 
	    deq := subs(expr(pol),y=y(x));
	else
	    Y[1] := map(divide(-u * g * diff(pol,x), pol, Rem), normal);
	    for d from 2 to deg-1 do
		Y[d] := map(divide(diff(Y[d-1],x)+diff(Y[d-1],y)*Y[1], pol, Rem), normal);
	    end_for;
	    matA := [[1, 0 $ i=1..deg-1, 1], [0,1, 0 $ i=1..deg-2, y(x)]];
	    V := [y^i $ i=0..deg-1];  // involved terms 
	    for d from 1 to deg-1 do
		for T in V do tc[T] := 0; end_for;
		for i from 1 to nterms(pol) do
		    tc[expr(nthterm(Y[d],i))] := nthcoeff(Y[d],i);
		end_for;
		matA := append(matA, [ tc[T] $ T in V ]);
	    end_for;
    // matrix building and gauss elimination 
	    for i from 3 to nops(matA) do
		matA[i] := append(matA[i], (D@@(i-2))(y)(x));
	    end_for;
	    matA:=Dom::DenseMatrix()(matA);
	    matA:=linalg::gaussElim(matA);
    // take the differential equation (numer of the right coefficient) 
	    i := 1;
	    while not has(matA[i,i], y) do
		i := i+1;
	    end_while;
	    deq := numer(matA[i,i]);
	end_if;
	deq;
    end_proc;
    
    listofvalues:=
    proc(t) begin map([op(t)], op, 2) end_proc;
    
    diffeqtorec := 
    proc(deq, yofx: DOM_EXPR, uofn: DOM_EXPR)
    begin
	deq:=expand(deq);
   // minimal value of n for which the recurrence is valide 
	dom::LIM:=degree(select(deq,not has,op(yofx,0)),op(yofx,1))+1;
	case type(deq)
	  of DOM_INT do of DOM_IDENT do of "function" do
					of "_mult" do break
					of "_plus" do deq:=op(deq); break
					otherwise error("invalid argument")
	end_case;
	_plus(op(map([deq],dom::diffeqtorec2,op(yofx,0..1),op(uofn,0..1))))
    end_proc;
    
/* t is of the form c*x^j*(D@@k)(y)(x),
	   returns the coefficient of x^n in the g.f. of t in terms of u(n)=[x^n] y(x) */
    diffeqtorec2 := 
    proc(t, y, x, u, n) 
	local c,k,j,i;
    begin
	if type(t)="_mult" then ([t,c,k]):=split(t,has,{x,y}) else c:=1 end_if;
	j:=degree(t,[x]);
	t:=t/x^j;
    // now t = 1 or y(x) or D(y)(x) or (D@@k)(y)(x) 
	if not has(t,y) then 0
	else // t = y(x) or D(y)(x) or D(D(y))(x) or ... 
	    if type(t)<>"function" or [op(t)]<>[x] then
		error("unexpected term: ".expr2text(t))
	    end_if;
	    t:=op(t,0); // now y or D(y) or D(D(y)) or ... 
	    k:=0; while type(t)="D" do k:=k+1; t:=op(t) end_while;
	    if t<>y then error("unexpected term: ".expr2text(t)) end_if;
	    c*_mult(n+i-j $ i=1..k)*u(n+k-j)
	end_if
    end_proc;
    
    minindex:= 
    proc(r,u,n)
    begin
	min(op(map(indets(r,PolyExpr) minus {n},proc() begin op(args(1))-n end_proc )));
    end_proc;
    
// returns the largest i such that u(n+i) appears in a recurrence 
    maxindex := 
    proc(r,u,n)
    begin
	max(op(map(indets(r,PolyExpr) minus {n},proc() begin op(args(1))-n end_proc )));
end_proc;



///////////////////////// End of Holonomic part /////////////////////
    makeUnrank := 
    proc() 
	option remember;
	local p, t, f, A, Rhs;
    begin
	dom::makeValuation();
	dom::makeCount();
	for p in dom::standardForm() do 
	    A   := op(p,1);
	    Rhs := op(p,2);
	    t:=dom::typeofoper(Rhs);
	    f:=slot(dom, "makeUnrank_".t);
	    if domtype(f) <> DOM_PROC then
		error("no unrank rule for constructor ".op(Rhs,0)) 
	    end_if;
	    dom::unrank_functions[A] := f(A, op(Rhs,0), op(Rhs))
	end_for;
    end_proc;

    makeCount_Alias := 
    proc(B) // A=Alias(B)
	option escape;
	local DOM;
    begin
	DOM:=dom;
	n -> DOM::count_functions[B](n);
    end_proc;
    
    makeUnrank_Alias := 
    proc(B) // A=Alias(B)
	option escape;
	local DOM;
    begin
	DOM:=dom;
	(n,r) -> DOM::unrank_functions[B](n,r);
    end_proc;
    
    makeCount_Epsilon := 
    proc()
    begin
	if dom::options[Float] then
	    n -> if n=0 then 1.0 else 0 end_if;
	else
	    n -> if n=0 then 1 else 0 end_if;
	end_if;
    end_proc;

    makeUnrank_Epsilon :=
    proc()
	option escape;
	local label;
    begin 
	if args(0)>=3 then label:=args(3); else label := Epsilon; end_if;
	(n,r) -> if n=0 and r=1 then label else FAIL; end_if;
    end_proc;

    makeCount_Atom := 
    proc()
    begin
	if dom::options[Float] then
	    n -> if n=1 then 1.0 else 0 end_if;
	else
	    n -> if n=1 then 1 else 0 end_if;
	end_if;
    end_proc;

    makeUnrank_Atom:=
    proc()
	local label;
	option escape;
    begin
	if args(0)>=3 then label:=args(3); else label := Atom; end_if;
	(n,r) -> if n=1 and r=1 then label else FAIL; end_if
    end_proc;

    makeCount_Union := 
    proc() // A = Union(B,C,D, ...)
	option escape;
	local operands, DOM;
    begin
	DOM := dom;
	operands := args();
        n -> _plus(DOM::count_functions[o](n) $ o in operands);
    end_proc;

    makeUnrank_Union := 
    proc(A) // A = Union(B,C,D, ...)
	option escape;
	local operands, DOM;
    begin
	DOM := dom;
	operands := args(3..args(0));
        proc(n: Type::NonNegInt, r: Type::PosInt)	
	    local o, r1;
	begin
	    for o in operands do
		r1 := DOM::count_functions[o](n);
		if r <= r1 then 
		    return(DOM::unrank_functions[o](n,r)) 
		end_if;
		r := r - r1;
	    end_for;
	end_proc    
    end_proc;

    makeCount_Theta := 
    proc(B) // A=Theta(B)
	local DOM;
	option escape;
    begin
	DOM := dom;
	n -> n*DOM::count_functions[B](n);
    end_proc;

    makeUnrank_Theta := 
    proc()
	local DOM, B;
	option escape;
    begin
	DOM := dom;
	B:=args(3);
	(n,r) -> DOM::unrank_functions[B](n, r); //(r-1 div n)+1); ???
    end_proc;

    makeCount_Int := 
    proc(B) // A=Int(B)
	local DOM;
	option escape;
    begin
	DOM := dom;
        n -> if n=0 then 0; else (DOM::count_functions[B](n))/n; end_if;
    end_proc;
    
    makeUnrank_Int := 
    proc()
	local DOM, B;
	option escape;
    begin
	DOM := dom;
	B:=args(3);
	(n,r) -> DOM::unrank_functions[B](n, r);
//	(n,r) -> DOM::unrank_functions[B](n, (r-1)*n+1); // ???? CHECK THIS
    end_proc;

    makeCount_Delta := 
    proc(B, u) // A = Delta(B,u) 
	local DOM;
	option escape;
    begin
	DOM := dom;
        proc(n: Type::NonNegInt)
	    option remember; 
	begin
	    _plus(u(k)*DOM::count_functions[B](n/k)
		  $ k in numlib::divisors(n));
	end_proc
    end_proc;

    makeUnrank_Delta := 
    proc()
	local DOM, B, u;
	option escape;
    begin
	DOM := dom;
	B:=args(3);		// CHECK THIS !!!!!
	u:=args(4);
        proc(n: Type::NonNegInt, r: Type::PosInt)
	    local k, r1; 
	begin
	    for k in numlib::divisors(n) do
		r1 := u(k)*DOM::count_functions[B](n/k); 
		if r <= r1 then 
		    return(DOM::unrank_functions[B](n/k, r) $ k)
		end_if;
		r := r - r1;
	    end_for;
	end_proc	
    end_proc;

    makeCount_Predefined := proc(B) begin B end_proc;

    makeUnrank_Predefined  := 
    proc(A) 
	option escape; 
    begin 
	(n,r)->A(n) 
    end_proc;

    /**
	Build a counting procedure for A=Prod(B,C) 

        Note: it seems from the code of isprod1 and isprod2 that the
	holonomic product should not be used for rules of the form
	Prod(Atom, Epsilon) It seems that the lazy product should not
	be used for rules of the form Prod(B, Epsilon) or Prod(Atom, B)
    **/
    makeCount_Prod :=
    //if dom::options[Products]=Naive then
	/**
	    Using the naive product formulas of generating series:

            Unlabelled case:   a_n = \sum_k b_n c_{n-k}
            Labelled case:     a_n = \sum_k b_n c_{n-k} \binom{n,k}
        **/
        proc(B,C)
	    local DOM, vB, vC;
	    option escape;
	begin
	    DOM := dom;
	    vB := DOM::valuation[B];
	    vC := DOM::valuation[C];
	    if DOM::valuation[B] = DOM::degree[B] then
		// Specialized version when the generating series of B
		// has just one term
		if dom::options[Labelled] then
		    proc(n)
			option remember;
		    begin
			binomial(n,vB) * 
			DOM::count_functions[B](vB  ) * 
			DOM::count_functions[C](n-vB);
		    end_proc;
		else
		    proc(n)
			option remember;
		    begin
			DOM::count_functions[B](vB  ) *
			DOM::count_functions[C](n-vB);
		    end_proc;
		end_if;
	    elif DOM::valuation[C]=DOM::degree[C] then
		// Specialized version when the generating series of C
		// has just one term
		if dom::options[Labelled] then
		    proc(n)
			option remember;
		    begin
			binomial(n,vC) * 
			DOM::count_functions[B](n-vC) * 
			DOM::count_functions[C](vC);
		    end_proc;
		else
		    proc(n) 
			option remember;
		    begin
			DOM::count_functions[B](n-vC) * 
			DOM::count_functions[C](vC  ); 
		    end_proc; 
		end_if;
	    else
		if dom::options[Labelled] then
		    proc(n)
			local b,k,s,k2,cB,cC; 
			option remember;
		    begin
			cB := DOM::count_functions[B];
			cC := DOM::count_functions[C];
			k2 := min(DOM::degree[B],n-vC);
			b := binomial(n,vB); // Rewrite as a sum ?
			s := b*cB(vB)*cC(n-vB);
			for k from vB+1 to k2 do
			    b := b*(n-k+1)/k;
			    s := s+b*cB(k)*cC(n-k);
			end_for; 
			s
		    end_proc;
		else
		    proc(n) 
			local k2,cB,cC;  
			option remember;
		    begin
			cB := DOM::count_functions[B];
			cC := DOM::count_functions[C];
			_plus(cB(k) * cC(n-k)
			      $ k=vB..min(DOM::degree[B],n-vC));
		    end_proc;
		end_if
	    end_if;
	end_proc;
/*    elif dom::options[Products]=Lazy then
	if dom::options[Labelled] then 
	    error("Labelled not yet implemented")
	else
	    error("Unlabelled not yet implemented")
	end_if
	
    elif dom::options[Products]=Holonomic then
	
    else error("Unknown product type");
    end_if;*/

    makeUnrank_Prod :=
    proc(A, typeProd, B,C) 
	option escape;
	local DOM, vB,vC;
    begin
	DOM := dom;
	vB:=DOM::valuation[B]; 
	vC:=DOM::valuation[C];
	if vB=DOM::degree[B] then
	    if dom::options[Labelled] then
	        proc(n: Type::NonNegInt, r: Type::PosInt) 
		    local nC;
		begin
		    // For each choice of B and C, there are
		    // binomial(n,vB) choices for relabelling
		    // We unrank them consecutively
		    r := (r-1 mod binomial(n, vB)) + 1;
		    // We unrank the products lexicographically
		    // e.g.: Prod(A,A), Prod(A,B), Prod(B,A), Prod(B,B)
		    nC := DOM::count_functions[C](n-vB);
		    typeProd(DOM::unrank_functions[B](vB  , (r-1 div nC)+1),
			     DOM::unrank_functions[C](n-vB, (r-1 mod nC)+1));
		end_proc
	    else
		proc(n: Type::NonNegInt, r: Type::PosInt) 
		    local nC;
		begin
		    nC := DOM::count_functions[C](n-vB);
		    typeProd(DOM::unrank_functions[B](vB  , (r-1 div nC)+1),
			     DOM::unrank_functions[C](n-vB, (r-1 mod nC)+1));
		end_proc
	    end_if;
	elif vC=DOM::degree[C] then
	    if dom::options[Labelled] then
	        proc(n: Type::NonNegInt, r: Type::PosInt) 
		    local nC;
		begin
		    // For each choice of B and C, there are
		    // binomial(n,vC) choices for relabelling
		    // We unrank them consecutively
		    r := (r-1 mod binomial(n, vC)) + 1;
		    nC := DOM::count_functions[C](vC);
		    typeProd(DOM::unrank_functions[B](n-vC, (r-1 div nC)+1),
			     DOM::unrank_functions[C](vC,   (r-1 mod nC)+1));
		end_proc
	    else
		proc(n: Type::NonNegInt, r: Type::PosInt) 
		    local nC;
		begin
		    nC := DOM::count_functions[C](vC);
		    typeProd(DOM::unrank_functions[B](n-vC, (r-1 div nC)+1),
			     DOM::unrank_functions[C](vC,   (r-1 mod nC)+1));
		end_proc
	    end_if;
	else 
	    if dom::options[Labelled] then
		proc(n: Type::NonNegInt, r: Type::PosInt)
		    local k, cB, cC, nC, r1; 
		begin
		    cB := DOM::count_functions[B];
		    cC := DOM::count_functions[C];
		    for k from vB to n-vC do
			nC := cC(n-k);
			r1 := cB(k)*nC*binomial(n,k);
			if r <= r1 then
			    r := (r-1 mod binomial(n,k))+1;
			    return(typeProd(DOM::unrank_functions[B](k,   (r-1 div nC)+1),
					    DOM::unrank_functions[C](n-k, (r-1 mod nC)+1)));
			end_if;
			r := r - r1;
		    end_for;
		end_proc;
	    else // Unlabelled case
		/*
		    Note: 
  		    The former algorithm was running through vB...n-vC
  		    by alternating low and high values. Was there 
  		    Any point in doing this ???
  		    k  := vB;
   		    repeat
   		        s := s-cB(k)*cC(n-k);
   		        if s<0 then break end_if;
   		        k := n+vB-vC-k;
   		        s := s-cB(k)*cC(n-k);
   		        if s<0 then break end_if;
   		        k := n+vB-vC+1-k;
   		    until FALSE end_repeat;
  		    typeProd(dB(k),dC(n-k))
		*/
		proc(n: Type::NonNegInt, r: Type::PosInt)
		    local k, cB, cC, nC, r1; 
		begin
		    cB := DOM::count_functions[B];
		    cC := DOM::count_functions[C];
		    for k from vB to n-vC do
			nC := cC(n-k);
			r1 := cB(k)*nC;
			if r <= r1 then
			    return(typeProd(DOM::unrank_functions[B](k,   (r-1 div nC)+1),
					    DOM::unrank_functions[C](n-k, (r-1 mod nC)+1)));
			end_if;
			r := r - r1;
		    end_for;
		end_proc;
	    end_if;    
	end_if;
    end_proc;

begin
    // analysis of options, ...
    Options := combinat::getOptions
    (
     table(Labelled = FALSE,	// TRUE
	   Target       = MuPAD,	// (C,GMP), (C,DPE), Module?, 
			         	// (Java,Text), (Java, Applet)?,
				 	// (Java, Graphics)
	   Products     = Auto,		// Auto, Naive, Lazy, Holonomic
	   Float	= FALSE,
	   File         = "cstemp",	//
	   MainTerminal = op(Specification,[1,1])
	  ),
     table(Labelled = DOM_BOOL,
	   Float    = DOM_BOOL,
	   Target   = Type::Union(DOM_IDENT, Type::SequenceOf(DOM_IDENT)),
	   Products = Type::Union(Type::Singleton(Auto),
				  Type::Singleton(Naive),
				  Type::Singleton(Holonomic),
				  Type::Singleton(Lazy)),
           File     = DOM_STRING
	  ),
     1
    );
    // Make sure Specification is a set of production rules
    Specification := { op(Specification) };  
end_domain:
    
stdlib::LoadlibState["combinat::decomposableStructures"] := TRUE:
