/*****************************************************************************
  $Id: integerMatrices.mu,v 1.44 2004/10/02 14:50:18 nthiery Exp $
  File:		lib/COMBINAT/integerMatrices.mu
  Authors:	Teresa Gomez-Diaz <Teresa.Gomez-Diaz@univ-mlv.fr>
                Florent Hivert <florent.hivert@univ-mlv.fr>
		Nicolas M. Thiry <nthiery@users.sourceforge.net>
  Created:	21/11/2002
  License:	LGPL
*****************************************************************************/

// Optimization for MuPAD > 2.0.0: we use RD_INF which is faster than infinity
%if traperror(RD_INF+RD_INF) = 0 then
    alias(infinity  = RD_INF);
end_if:

// Temporary workaround for the min(1, RD_INF) bug for MuPAD <= 2.5
// The bug also appears with RD_NINF. Should we take care of it ? 
%if traperror(min(RD_INF,1)) > 0 or min(1, RD_INF)<>1 or max(RD_INF,1) <> RD_INF then
    alias(min = proc()
		    local l;
		begin
		    l:=select([args()], not _equal, RD_INF);
		    if nops(l)=0 and args(0)>0 then
			// min(RD_INF, RD_INF, RD_INF, ...)
			RD_INF;
		    else
			min(op(l));
		    end_if;
		end_proc);
    alias(max = proc()
		begin
		    if contains({args()}, RD_INF) then
			RD_INF
		    else
			max(args());
		    end_if;
		end_proc);
end_if:

domain combinat::integerMatrices
    inherits Dom::BaseDomain;
    category Cat::CombinatorialClassWith2DBoxedRepresentation;
    axiom    Ax::systemRep;

    info := "Library 'combinat::integerMatrices': utilities for integer matrix";

//////////////////////////////////////////////////////////////////////////
//  An integer matrix m with sum of columns c=[c1,..,ck] and sum of rows l=[l1,..,ln]
//  is a n x k matrix m=m[i,j], where sum(m[i,j], i=1..n)=cj and sum(m[i,j], j=1..k)=li
//  where sum(c[i], i=1..k) is equal to sum(l[i], i=1..n).
//
// - type:		fixed length list of lists of nonnegative integers
//
//  - count    (c, l)	number of integer matrices
//  - list     (c, l)	list of all integer matrices
//  - generator(c, l)	generator for all integer matrices
//
//  - standard (m)              standardised permutation of matrix m
//  - fromStandard (p, cr, cc)  inverse of standard by fixing row and column sums
//  All integer matrices of column sum c=[2,0,5] and line sum l=[3,2,2]:
//
// 	>> combinat::integerMatrices([2,0,5],[3,2,2]);
//////////////////////////////////////////////////////////////////////////


    interfaceAdd :=
        { hold(printPretty), hold(printCompact),
	  hold(nrows), hold(ncols), hold(isSquare),
          hold(rowSums), hold(columnSums),
          hold(standard), hold(fromStandard),
	  hold(fromNonWeakDescentsComposition),
	  hold(transpose)
	};

    domtype := DOM_LIST;

    //////////////////////////////////////////////////////////////////////////
    // nrows, ncols, isSquare
    //////////////////////////////////////////////////////////////////////////

    nrows :=
    proc(m: dom): Type::NonNegInt
    begin
	nops(m);
    end_proc;

    ncols :=
    proc(m: dom): Type::NonNegInt
    begin
	if nops(m)=0 then
	    0 // this is conventional; we can't tell the number of columns of a matrix
	      // with 0 rows
	else
	    nops(m[1]);
	end_if;
    end_proc;
    
    isSquare :=
    proc(m: dom): DOM_BOOL
    begin
	bool(dom::ncols(m) = dom::nrows(m));
    end_proc;
    
    //////////////////////////////////////////////////////////////////////////
    // rowSums
    //////////////////////////////////////////////////////////////////////////

    rowSums :=
    proc(m: dom) : combinat::integerVectors
    begin
	map(m, _plus @ op);
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // columnSums
    //////////////////////////////////////////////////////////////////////////

    columnSums :=
    proc(m: dom) : combinat::integerVectors
        local l, row;
    begin
        if m = [] then return([]);
        end_if;
        l:= [0 $ nops(m[1])];
        (l:= zip(l, row, _plus)) $ row in m;
        return(l);
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // isA
    //////////////////////////////////////////////////////////////////////////

    isA :=
    proc(obj         : Type::AnyType,
         c = FAIL    : Type::Union(combinat::integerVectors, DOM_FAIL), 
	 l = FAIL    : Type::Union(combinat::integerVectors, DOM_FAIL),
         MinPart = 0 : Type::NonNegInt, 
         MaxPart = infinity : combinat::integerListsLexTools::typeIntBound) : DOM_BOOL
    begin
        if not testtype(obj, Type::ListOf(Type::ListOf(Type::NonNegInt))) then
	    return(FALSE);
	end_if;
	if nops({op(map(obj, nops))}) > 1 then return(FALSE); end_if; // same length for the rows
	if args(0) = 1 then return(TRUE); end_if;
        if testargs() and args(0) >= 3 then
            if _plus(op(c)) <> _plus(op(l)) then 
                error("Sum mismatch : ".expr2text(c)." and ".expr2text(l));
            end_if;
        end_if;
	if args(0) >= 2 and dom::columnSums(obj) <> c then return(FALSE); end_if;
	if args(0) >= 3 and dom::rowSums(obj) <> l then return(FALSE); end_if;
	if args(0) < 4 then return(TRUE); end_if;
	if MinPart > min(op(map(obj, op))) then return(FALSE); end_if;
	if MaxPart < max(op(map(obj, op))) then return(FALSE); end_if;
	TRUE;
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // count -> if symmetrica is not loaded, inherited from Cat::CombinatorialClass
    //////////////////////////////////////////////////////////////////////////

    SF    := if domtype(examples) = DOM_DOMAIN and combinat::lrcalc <> FAIL then
		 examples::SymmetricFunctions();
	     end_if;
    count := if dom::SF <> FAIL then
		 userinfo(1, "Using the symmetric functions for counting.");
		 proc(c: combinat::integerVectors, l: combinat::integerVectors,
		      MinPart = 0 : Type::NonNegInt, 
		      MaxPart = infinity : combinat::integerListsLexTools::typeIntBound) : Type::NonNegInt
		     local sc, sl ;
		 begin
		     if args(0) > 2 then dom::countFromGenerator(args());
		     else
			 sc:= _plus(op(c));
			 sl:= _plus(op(l));
			 if testargs() and sc <> sl then
			     error("Sum mismatch : ".expr2text(c)." and ".expr2text(l));
			 end_if;
			 dom::SF::scalar(dom::SF::h(c), dom::SF::h(l));
		     end_if;
		 end_proc;
	     end_if;

    //////////////////////////////////////////////////////////////////////////
    // minus_add_aux
    //////////////////////////////////////////////////////////////////////////

    minus_add_aux :=
    proc(v: Type::ListOf(Type::NonNegInt), lv: dom) : Type::ListOf(Type::NonNegInt)
    // auxiliary function to list
    // minux_add_aux(v, lv) returns v - (add(vi) for vi in lv)
        local aux, vi;
    begin
        aux:= v;
        (aux:= zip(aux, vi, _subtract)) $ vi in lv;
	return(aux);
     end_proc;

    //////////////////////////////////////////////////////////////////////////
    // integerVectors without option parsing
    //////////////////////////////////////////////////////////////////////////

    fastIntegerVectorList :=
	combinat::generators::toList @ dom::fastIntegerVectorGenerator;

    fastIntegerVectorGeneratorMuPAD :=
    proc(n, k, outer, minn, maxx)
	option escape;
    begin
        if minn > min(op(outer)) then 
	    return(combinat::generators::empty); 
	end_if;
	combinat::integerListsLexTools::generator
	    (n, k, k, minn, i->min(outer[i],maxx), -infinity, infinity);
    end_proc;

  fastIntegerVectorGeneratorModule :=
    proc(n, k, outer, minn=0, maxx=2147483647) // 2^(31)-1 for the dynamic module
        option escape;
        local first, newouter, ou;
    begin
        if minn > min(op(outer)) then
	    return(combinat::generators::empty); 
	end_if;
        newouter:= [min(ou, maxx) $ ou in outer];
        first := combinat::MachineIntegerListsLex::WithoutSlopeFirst(n, k, k, 0, [minn$k], newouter):
        combinat::generators::fromNext
	    (first,
	     l -> combinat::MachineIntegerListsLex::WithoutSlopeNext(l, n, k, k, [minn$k], newouter));
    end_proc;

    fastIntegerVectorGenerator :=
	(if (combinat::MachineIntegerListsLex<>FAIL) then
	     userinfo(1, "Using the MachineIntegerListsLex dynamic module for integers lists generation.");
	     dom::fastIntegerVectorGeneratorModule;
         else
	     userinfo(1, "Can't use the MachineIntegerListsLex dynamic module, fall back to MuPAD.");
	     dom::fastIntegerVectorGeneratorMuPAD;
	 end_if);


    //////////////////////////////////////////////////////////////////////////
    // list_aux
    //////////////////////////////////////////////////////////////////////////
 
    list_aux :=
    proc(cc: combinat::integerVectors, kk: Type::NonNegInt, 
         ll: combinat::integerVectors, nnl: Type::NonNegInt, 
         minP: Type::NonNegInt, maxP: combinat::integerListsLexTools::typeIntBound)
        option escape;
        local DOM, aux;
    // does not check sum condition on arguments
    begin
        DOM:= dom;
        if nnl = 1 then
            map(
                DOM::fastIntegerVectorList(ll[1], kk, cc, minP, maxP),
                x -> [x]
            );
        else
            aux:= map(
                DOM::list_aux(cc, kk, [op(ll, 1..nnl-1)], nnl-1, minP, maxP),
                lv -> map(
                    DOM::fastIntegerVectorList(ll[nnl], kk, DOM::minus_add_aux(cc, lv), minP, maxP),
                    (x, lv) -> append(lv, x),
                    lv
                )
            );
            if aux = [] then return(aux); end_if;
            _concat(op(aux));
        end_if;
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // list
    //////////////////////////////////////////////////////////////////////////

    list :=
    proc(c: combinat::integerVectors, l: combinat::integerVectors,
         MinPart = 0 : Type::NonNegInt, 
         MaxPart = infinity : combinat::integerListsLexTools::typeIntBound) : Type::ListOf(dom)
        // recursive version
        local k, nl, sc, sl ;
        option escape;
    begin
        k:= nops(c);
        nl:= nops(l);
        sc:= _plus(op(c));
        sl:= _plus(op(l));
        if testargs() and sc <> sl then
            error("Sum mismatch : ".expr2text(c)." and ".expr2text(l));
        end_if;
        if k = 0 then return([[[] $ nl]]); end_if;
        if MinPart > MaxPart then return([]); end_if;
        dom::list_aux(c, k, l, nl, MinPart, MaxPart);
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // generator_aux
    //////////////////////////////////////////////////////////////////////////

    generator_aux :=
    proc(cc: combinat::integerVectors, kk: Type::NonNegInt, 
         ll: combinat::integerVectors, nnl: Type::NonNegInt,
         minP: Type::NonNegInt, maxP: combinat::integerListsLexTools::typeIntBound)
        option escape;
        local DOM;
    // does not test sum condition on arguments
    begin
        DOM:= dom;
        if nnl = 1 then
            combinat::generators::map(
                DOM::fastIntegerVectorGenerator(ll[1], kk, cc, minP, maxP),
                x -> [x]
            );      
        else    
            combinat::generators::pipe(
                DOM::generator_aux(cc, kk, [op(ll, 1..nnl-1)], nnl-1, minP, maxP),
                lv -> combinat::generators::map(
		      DOM::fastIntegerVectorGenerator(ll[nnl], kk, 
			  DOM::minus_add_aux(cc, lv), minP, maxP),
                    (x, lv) -> append(lv, x),
                    lv      
                )       
            );      
        end_if; 
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // generator
    //////////////////////////////////////////////////////////////////////////

    generator :=
    proc(c: combinat::integerVectors, l: combinat::integerVectors,
         MinPart = 0 : Type::NonNegInt,
         MaxPart = infinity : combinat::integerListsLexTools::typeIntBound)
        local k, nl, sc, sl;
        option escape;
    begin
        k:= nops(c);
	nl:= nops(l);
	sc:= _plus(op(c));
	sl:= _plus(op(l));
	if testargs() and sc <> sl then
	    error("Sum mismatch : ".expr2text(c)." and ".expr2text(l));
        end_if;
	if k = 0 then return(combinat::generators::repeater([[] $ nl],1)); end_if;
        if MinPart > MaxPart then return(combinat::generators::empty); end_if;
	dom::generator_aux(c, k, l, nl, MinPart, MaxPart);
    end_proc;


    //////////////////////////////////////////////////////////////////////////
    // standard
    //////////////////////////////////////////////////////////////////////////

    standard :=
    proc(m: dom) : combinat::permutations::typeStandard
        local sr, sc, dim, nr, nc, lind, perm, indperm, i, j, newi;
    begin
        sr:= dom::rowSums(m);          
        sc:= dom::columnSums(m);      
        dim:= _plus(op(sr));             // dim of permutation matrix
        if testargs() and dim <> _plus(op(sc)) then
            error("Sum mismatch : ".expr2text(m));
        end_if;
        nc:= nops(sc);                   // number of columns
        nr:= nops(sr);                   // number of rows
        lind:= [0 $ nr];                 // index accumulation vector
        for i from 1 to nr-1 do
            lind[i+1]:= lind[i]+sr[i];
        end_for;
        perm:= [0 $ dim];
        indperm:= 1;
        for j from 1 to nc do
            for i from 1 to nr do
                for newi from 1 to m[i][j] do
                    lind[i]:= lind[i] + 1;
                    perm[indperm]:= lind[i];
                    indperm:= indperm + 1;
                end_for;
            end_for;
        end_for;
        return(perm);
    end_proc;


    //////////////////////////////////////////////////////////////////////////
    // fromStandard
    //////////////////////////////////////////////////////////////////////////

    fromStandard :=
    proc(perm: combinat::permutations::typeStandard, 
         compoCol = FAIL : Type::Union(DOM_FAIL, combinat::compositions),
         compoLig = FAIL : Type::Union(DOM_FAIL, combinat::compositions)) : dom
        local nc, nr, mi, m, somme, indperm, i, j, newi;
    begin
        if compoCol = FAIL then 
            compoCol:= combinat::permutations::descentsComposition(perm);
        end_if;
        if compoLig = FAIL then
            compoLig:= combinat::permutations::descentsComposition(
                                  combinat::permutations::inverse(perm));
        end_if;
        nc:= nops(compoCol);
        nr:= nops(compoLig);
        mi:= [0 $ nc];
        m:= [mi $ nr];              // the matrix is given as a list of rows
        somme:= 0;
        for i from 1 to nr do
            indperm:= 1;
            for j from 1 to nc do
                for newi from 1 to compoCol[j] do
                    if somme < perm[indperm] and perm[indperm] <= (compoLig[i] + somme) then 
                        m[i][j]:= m[i][j] + 1;
                    end_if;
                    indperm:= indperm + 1;
                end_for;
            end_for;
            somme:= somme + compoLig[i];
        end_for;
        return(m);
    end_proc;

    transpose :=
    proc(m: dom) : dom
	local i, j;
    begin
	if nops(m)=0 then
	    // There is no way to make a difference between a 0x3 and
	    // 0x4 matrix. So, in this special case of matrices with
	    // 0 rows or 0 lines, transpose is not involutive.
	    return([]);
	end_if;
	[ [ m[i][j] $ i=1..nops(m) ] $ j=1..nops(m[1]) ];
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // Pretty-print
    //////////////////////////////////////////////////////////////////////////
  
    toArray :=
    proc(m: dom) : DOM_ARRAY
        local col, row, width;
    begin
        // Note the max(1, instead of max(0, to make array happy
	// See Cat::CombinatorialClassWith2DBoxedRepresentation
	width := if nops(m)=0 then 1 else max(1,nops(m[1])) end_if;
        array(1..max(1, nops(m)), 1..width,
	      (row, col) = m[row][col]
	      $ col = 1..nops(m[row])
	      $ row = 1..nops(m));
    end_proc;

//    printCompact :=
//    proc(m: dom)
//    begin   
//        %if version()[1] >= 3 then
//            dom::printCompactVersion3(m);
//        else
//            dom::printCompactVersion2_5(m);
//        end_if;
//    end_proc;
//
//    printCompactVersion2_5 :=
//    proc(m: dom) : output::asciiArt
//        local nn, row, col, ll;
//    begin
//        nn:= nops(m[1]);
//        ll:=[];
//        for row from nops(m) downto 1 do
//            ll:= _concat(ll,[_concat("|".m[row][col] $ col=1..nn, "|")]);
//        end_for;
//        output::asciiArt(ll);
//    end_proc;
//
//    printCompactVersion3 :=
//    proc(m: dom) 
//        local nn, mm;
//    begin
//        nn:= nops(m[1]);
//        mm:= map(m, op);
//        stdlib::Exposed(
//           output::tableForm(mm, Sort = FALSE, hold(Column) = nn,
//                             Unique, Right, hold(String))
//        );
//    end_proc;
 
    // For backward compatibility
    prettyPrint := dom::printPretty;

end_domain:

//////////////////////////////////////////////////////////////////////////////
// Descents
//////////////////////////////////////////////////////////////////////////////

/**
 * The combinatorial class of integer matrices M graded by two parameters:
 *  - rowSums: the integer vector of the row sums of M
 *  - c: the non weak descents composition of M
 *
 * j is a (weak) descent of M if there exists a row i0 in M such that
 * M[i,j]=M[i,j+1] for i<i0 and M[i0,j]>M[i0,j+1] (or
 * i0=nops(rowSums)+1).
 *
 * Those are the integer matrices with prescribed row sums which are
 * lexicographically maximal (for the natural left->right top->down
 * reading) in their orbit by the young subgroup S_{c_1} x \dots x
 * S_{c_k} acting by permutation of the columns. In particular, they
 * form a set of orbit representative of integer matrices w.r.t. this
 * young subgroup.
 **/
domain combinat::integerMatrices::fromNonWeakDescentsComposition
    inherits combinat::integerMatrices;
    category Cat::CombinatorialClass;
    axiom Ax::systemRep;

    generator :=
    proc(rowSums : combinat::integerVectors,
	 c: combinat::compositions
	) : combinat::generators
	option escape;
	local DOM;
    begin
	DOM := dom;
	// Base cases
	if nops(rowSums)=0 then
	    return(combinat::generators::repeater([],1));
	end_if;
	if _plus(op(rowSums)) = 0 then
	    return(combinat::generators::repeater([[0$_plus(op(c))]$nops(rowSums)],1));
	end_if;
	if nops(c)=0 then
	    return(combinat::generators::empty);
	end_if;
	// Recursion
	combinat::generators::pipe
	(combinat::integerVectors::fromNonWeakDescentsComposition::generator
	 (rowSums[1], c),
	 proc(v: combinat::integerVectors) : combinat::generators
	     local newc, i, i0, j, s;
		   option escape;
	 begin
	     // Update the non descents composition for lower rows
	     // We could also do it with:
	     // compositions::fromSet(compositions::set(c) union descentsSet(v))
	     newc := [];
	     i0 := 0;
	     j := 1;
	     s := c[1];
	     for i from 1 to nops(v) do
		 assert(s=_plus(op(c, 1..j)));
		 assert(s<=nops(v));
		 assert(i<nops(v) or s=nops(v));
		 if i=s or v[i]>v[i+1] then
		     newc := append(newc, i-i0);
		     i0 := i;
		 end_if;
		 if i=s and i<nops(v) then
		     j:=j+1;
		     s:=s+c[j];
		 end_if;
	     end_for;
//	 print(v, newc, [op(rowSums, 2..nops(rowSums))]);
	     combinat::generators::map
	     (DOM::generator([op(rowSums, 2..nops(rowSums))], newc),
	      m -> [v,op(m)]);
	 end_proc)
    end_proc;

    list := dom::listFromGenerator;
end_domain:
