/*****************************************************************************
  $Id: integerVectors.mu,v 1.24 2004/10/04 16:55:37 nthiery Exp $
  File:		lib/COMBINAT/integerVectors.mu
  Authors:	Nicolas M. Thiry <nthiery@users.sourceforge.net>
                Florent Hivert <florent.hivert@univ-mlv.fr>
  Created:	16/02/2001
  License:	LGPL
*****************************************************************************/

// Optimization for MuPAD > 2.0.0: we use RD_INF which is faster than infinity
%if traperror(RD_INF+RD_INF) = 0 then
    alias(infinity  = RD_INF);
end_if:

domain combinat::integerVectors
    inherits combinat::words;
    category Cat::IntegerListsLexClass;
    axiom    Ax::systemRep;

    info := "Library 'combinat::integerVectors': utilities for integer vectors";

/**
 An integer vector of sum n with k parts is a list of k nonnegative
 integers which sum up to n.

 Integer vectors are naturally ordered lexicographically.
 Several functions of this library are directly inherited from
 Cat::IntegerListsLexClass, and accept generic
 constraints:

  - size of parts: 		MinPart = , MaxPart =
  - slope:			MinSlope = , MaxSlope =
  - containment:		Inner = , Outer =

 See the documentation of Cat::IntegerListsLexClass and the
 examples for details on those options.

 See also Cat::CombinatorialClass.

 - count    (n, k, <Options>)	number of integer vectors
 - list     (n, k, <Options>)	list of all integer vectors
 - generator(n, k, <Options>)	generator for all integer vectors

 - first    (n, k, <Options>)	first integer vector
 - last     (n, k, <Options>)	last integer vector
 - Next     (c, k, <Options>)	next integer vector

 All integer vectors of sum 5 and 3 parts, that are bounded by [4,1,1]:

	>> combinat::integerVectors(5, 3, Outer=[4,1,1]);
		[[4,1,0], [4,0,1], [3,1,1]]

**/

    interface := combinat::words::interface union
        { hold(fromNonWeakDescentsComposition), hold(_plus),
	  hold(first), hold(Next)
	};

// domtype and type ar inherited

    isA := proc(compo : Type::AnyType, summ=FAIL: Type::Union(DOM_FAIL, Type::NonNegInt)) : DOM_BOOL
	   begin
	       _lazy_and(testtype(compo, Type::ListOf(Type::NonNegInt)),
			 _lazy_or(summ=FAIL, _plus(op(compo)) = summ),
			 _lazy_or(args(0) <= 2, 
				  combinat::integerListsLexTools::isCorrect
				  (compo, dom::parseOptions(args(3..args(0)))) 
				 ));
	   end_proc;

    parseOptions :=
    proc(k: Type::NonNegInt)
    begin
        dom::parseOptionsIntegerListsLexClass
        (table(Length	= k,
               MinPart	= 0,
               MaxPart	= infinity,
               Inner	= NIL,
               Outer	= NIL,
	       MinSlope	= -infinity,
	       MaxSlope	=  infinity
              ),
         args(2..args(0))
         ):
    end_proc;

    generator := DomainConstructor::getEntryFromCategory
	       (dom, Cat::IntegerListsLexClass, "generator");

    //////////////////////////////////////////////////////////////////////////
    // count
    //////////////////////////////////////////////////////////////////////////

    count :=
    proc(n: Type::Integer, k: Type::NonNegInt) : Type::NonNegInt
    begin
        if args(0) > 2 then
            // There is no efficient counter yet when there are
            // further constraints.
            return(dom::count_integerListsLexClass(args()));
        end_if;

        if n >= 0 then
            binomial(n+k-1,n)
        else
            0;
        end_if;
    end_proc;

    Next :=
    proc(l: dom) : Type::Union(dom, DOM_FAIL)
    begin
	combinat::integerListsLexTools::Next(l, dom::parseOptions(nops(l), args(2..args(0))));
    end_proc;

    _plus := 
    proc()
	local res, v;
    begin
	res := [];
	for v in [args()] do
	    res := zip(res, v, _plus, 0);
	end_for;
	res;
    end_proc;

begin
end_domain:

//////////////////////////////////////////////////////////////////////////////
// Descents
//////////////////////////////////////////////////////////////////////////////

/**
 * The combinatorial class of integer vectors v graded by two parameters:
 *  - n: the sum of the parts of v
 *  - c: the non descents composition of v
 *
 * In other words: the length of v equals c[1]+...+c[k], and v is
 * descreasing in the consecutive blocs of length c[1], ..., c[k]
 *
 * Those are the integer vectors of sum n which are lexicographically
 * maximal (for the natural left->right reading) in their orbit by the
 * young subgroup S_{c_1} x \dots x S_{c_k}.  In particular, they form
 * a set of orbit representative of integer vectors w.r.t. this young
 * subgroup.
 **/
domain combinat::integerVectors::fromNonWeakDescentsComposition
    inherits combinat::integerVectors;
    category Cat::CombinatorialClass;
    axiom Ax::systemRep;
    
    generator :=
    proc(n,
	 c: combinat::compositions)
	option escape;
    begin
	combinat::generators::pipe
	(// Split up n between the different blocs
	 combinat::integerVectors::generator(n, nops(c)),
	 proc(v)
	     local i;
	     option escape;
	 begin
	     combinat::generators::map
	     (combinat::generators::cartesian
	      (combinat::integerVectors::generator(v[i], c[i], MaxSlope=0) $i=1..nops(c)),
	      map, op)
	 end_proc)
    end_proc;
end_domain:

domain combinat::integerVectors::fromStandard
    inherits combinat::words::fromStandard;
    category Cat::CombinatorialClass;
    axiom    Ax::systemRep;

    info      := "Library 'combinat::integerVectors::fromStandard': utilities for the integerVectors having a given standard permutation";
    typeWords      := combinat::integerVectors; // for correct type checking by inheritance
end_domain:

//////////////////////////////////////////////////////////////////////////////
// shuffle products of two integerVectors
domain combinat::integerVectors::shuffle
    inherits combinat::words::shuffle; 
    category Cat::CombinatorialClass;
    axiom    Ax::systemRep;

    info      := "Library 'combinat::integerVectors::shuffle': utilities for the shuffle product of two integerVectors";
    typeWords      := combinat::integerVectors; // for correct type checking by inheritance
end_domain:
//////////////////////////////////////////////////////////////////////////////
// shuffling products of two words
domain combinat::integerVectors::shufflings
    inherits combinat::words::shufflings; 
    category Cat::CombinatorialClass;
    axiom    Ax::systemRep;

    info      := "Library 'combinat::integerVectors::shuffling': utilities for the shuffling of two integerVectors";
    typeWords      := combinat::integerVectors; // for correct type checking by inheritance
end_domain:
//////////////////////////////////////////////////////////////////////////////
// shuffle products of two words
domain combinat::integerVectors::split
    inherits combinat::words::split;
    category Cat::CombinatorialClass;
    axiom    Ax::systemRep;

    info      := "Library 'combinat::integerVectors::split': utilities for the split of an integerVectors";
    typeWords      := combinat::integerVectors; // for correct type checking by inheritance
end_domain:
//////////////////////////////////////////////////////////////////////////////
// shuffle products of two words
domain combinat::integerVectors::splittings
    inherits combinat::words::splittings; 
    category Cat::CombinatorialClass;
    axiom    Ax::systemRep;

    info      := "Library 'combinat::integerVectors::splitting': utilities for the splitting of an integerVectors";
    typeWords      := combinat::integerVectors; // for correct type checking by inheritance
end_domain:
