/*****************************************************************************
  $Id $
  File:		lib/COMBINAT/necklaces.mu
  Authors:	Xavier Molinero <molinero@lsi.upc.es>
		Nicolas M. Thiry <nthiery@users.sourceforge.net>
  Purpose:	Source code for the domain combinat::necklaces
  Created:	Wed Jan 28 17:13:38 2004
  License:	LGPL
  Project:	MuPAD-Combinat, http://mupad-combinat.sf.net
*****************************************************************************/

domain combinat::necklaces
    inherits Dom::BaseDomain;
    category Cat::CombinatorialClass;
    axiom    Ax::systemRep;

    info := "combinat::necklaces -- the combinatorial class of necklaces";

    interfaceAdd :=
        {hold(fromEvaluation),
	 hold(fromContent)
	};

    domtype := DOM_LIST;

end_domain:

domain combinat::necklaces::fromEvaluation
    inherits Dom::BaseDomain;
    category Cat::CombinatorialClass;
    axiom    Ax::systemRep;

    info := "combinat::necklaces -- the combinatorial class of integer necklaces with given evaluation";

    /**
     * An integer necklace is a necklace whose beads are positive integers
     *
     * The evaluation of an integer necklace is a composition that tells
     * how many 1's, 2's and so it contains (see words::evaluation).
     * (e.g. the evaluation of [1,1,1,2,1,3,3] is [4,1,2]
     *
     * Here, we are restricted to evaluations that are compositions; in other words
     * the necklaces should contains all the integers in a range 1..k
     * (e.g. [2,2,3], or [1,4,3,3] are forbidden)
     **/

    interfaceAdd :=
	       { };

    domtype := DOM_LIST;

    //////////////////////////////////////////////////////////////////////////
    // isA
    //////////////////////////////////////////////////////////////////////////

    isA :=
    proc(object: Type::AnyType,
	 evaluation=FAIL: Type::Union(combinat::compositions, DOM_FAIL))
	: DOM_BOOL
    begin
	if not testtype(object, Type::ListOf(Type::PosInteger)) then
	    FALSE
	elif evaluation <> FAIL and combinat::words::evaluation(object) <> evaluation then
	    FALSE
	else
	    // Check that object is lexicographically smallest!
	    UNKNOWN
	end_if;
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // count
    //////////////////////////////////////////////////////////////////////////

    count :=
    proc(evaluation : combinat::compositions) : Type::NonNegInt
      local n, j, ni, i;    
    begin
      if nops(evaluation)=0 then return(0) end_if;
      n := _plus(i $ i in evaluation);
      _plus(numlib::phi(j) * (n/j)! / (_mult( (ni/j)! $ ni in evaluation ))
	    $ j in numlib::divisors(gcd(op(evaluation)))) / n;
    end_proc;
    
    //////////////////////////////////////////////////////////////////////////
    // generator
    //////////////////////////////////////////////////////////////////////////

    generator :=
    proc(evaluation : combinat::compositions, aperiodic=FALSE)
	option escape;
    local n, k, necklace, generator, i,
	  headValue, nextValue, prevValue, addValue, removeValue, r, s;
    begin
	k := nops(evaluation);      // the alphabet is { 1..k }
	if k=0 then
	    return(combinat::generators::empty)
	end_if;
	n := _plus(op(evaluation)); // the length of the necklace
	necklace := [ 1, k $ n-1 ];
	evaluation[1] := evaluation[1] - 1;
	r := [ 0 $ n]; // r[i] is the number of consecutive k's starting from i
	if k=1 then
	    r[1] := n-1;
	    s := 1;
	else
	    s := 2;
	end_if;

        // We maintain a doubly linked list pointing to the previous
        // and next value j with non-zero evaluation.
	// This data structure ought to be extracted and put somewhere like
	// in adt.
        // Invariant:
        //  - headValue, nextValue(headValue), nextValue(nextValue(...))
        //    yield all the values with non zero evaluation
	headValue := k;
	nextValue := [ i-1 $ i=1..k+1 ];
	prevValue := [ i+1 $ i=1..k+1 ];

	removeValue :=
	proc(j: Type::PosInt)
	    local n, p;
	begin
	    if (j = headValue) then
		headValue := nextValue[j]
	    end_if;
	    n := nextValue[j];
	    p := prevValue[j];
	    if n>0 then
		prevValue[n] := p;
	    end_if;
	    nextValue[p] := n;
	end_proc:

	addValue :=
	proc(j: Type::NonNegInt)
	    local n, p;
	begin
	    n := nextValue[j];
	    p := prevValue[j];
	    if n>0 then
		prevValue[n] := j;
	    end_if;
	    nextValue[p] := j;
	    if (prevValue[j] = k+1) then
		headValue := j
	    end_if
	end;

	if evaluation[1] = 0 then
	    removeValue(1);
	end_if;

	generator :=
	proc(t, p, s)
	    /**
	     * Invariants:
	     *  - necklace[1..t-1] is defined
	     *  - p is the length of the longest aperiodic
	     *    prefix (lyndon word) of necklace[1..t]
	     *  - s: starting point of the current run of s
	     **/
	    option escape;
	    local j, g, s1;
	begin
	    if evaluation[k] = n-t+1 then // their remains only k's to place
		if evaluation[k] > r[t-p] or
		   (evaluation[k] = r[t-p] and
		    if aperiodic then
			n=p
		    else
			n mod p = 0
		    end_if) then
		    combinat::generators::singleton(necklace);
		else
		    combinat::generators::empty;
		end_if;
	    elif evaluation[1] = n-t+1 then // their remains only 1's to place
		combinat::generators::empty;
	    else
		j := headValue;
		g := NIL;
		s1 := s;
		proc()
		    local x, i;
		begin
		    while TRUE do
			if g <> NIL then
			    x := g();
			    if x <> FAIL then
			    return(x);
			    else
			        // That's the place to reset the different
				// information
				if evaluation[j] = 0 then
				    addValue(j);
				end_if;
				evaluation[j] := evaluation[j] + 1;
				necklace[t] := k;
				j := nextValue[j];
			    end_if;
			end_if;
			if j < necklace[t-p] then
			    return(FAIL);
			end_if;
			assert(evaluation[j]>0);
			assert(_and(evaluation[i]=0 $ i = j+1..nextValue[j]-1));
		        // By theorem ..., necklace[1]..necklace[t-1],j is a prenecklace
			r[s] := t-s;
			necklace[t] := j;
			evaluation[j] := evaluation[j] - 1;
			if evaluation[j] = 0 then
			    removeValue(j)
			end_if;
			if j <> k then
			    s1 := t+1;
			end_if;
			g :=
			if j = necklace[t-p] then
			    generator(t+1, p, s1)
			else
			    generator(t+1, t, s1);
			end_if;
		    end_while;
		end_proc;
	    end_if;
	end_proc;
	generator(2, 1, s);
    end_proc;

end_domain:

// Duplicated code with lyndonWords::fromContent
// General framework for both?

domain combinat::necklaces::fromContent
    inherits Dom::BaseDomain;
    category Cat::CombinatorialClass;
    axiom    Ax::systemRep;

    info := "combinat::necklaces -- the combinatorial class of integer lyndon Words with prescribed content";

    interfaceAdd :=
	       { };

    domtype := DOM_LIST;

    //////////////////////////////////////////////////////////////////////////
    // isA
    //////////////////////////////////////////////////////////////////////////

    isA :=
    proc(object: Type::AnyType,
	 content=FAIL: Type::Union(combinat::words, DOM_FAIL))
	: DOM_BOOL
    begin
	if not testtype(object, DOM_LIST) then
	    FALSE
	elif content <> FAIL and sort(object) <> sort(content) then
	    FALSE
	else
	    // Check that object is aperiodic
	    UNKNOWN
	end_if;
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // count
    //////////////////////////////////////////////////////////////////////////

    count :=
    proc(content: combinat::words)
    begin
	combinat::necklaces::fromEvaluation::count
	(combinat::words::evaluation(content, {op(content)}));
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // generator
    //////////////////////////////////////////////////////////////////////////

    generator :=
    proc(content: combinat::words)
	option escape;
	local alphabet, rank, unrank;
    begin
	alphabet := sort([op({op(content)})]);
	unrank := combinat::rankers::fromList(alphabet)[2];
	combinat::generators::map
	(combinat::necklaces::fromEvaluation::generator
	 (combinat::words::evaluation(content, alphabet)),
	 map, unrank);
    end_proc;

end_domain:

/* isA for fromWord
    isA :=
    proc(object: Type::AnyType,
	 evaluation=FAIL: Type::Union(combinat::words, DOM_FAIL))
	: DOM_BOOL
    begin
	if domtype(object) <> DOM_LIST then
	    FALSE
	elif evaluation <> FAIL then
	    bool(sort(object) = sort(evaluation));
	else
	    UNKNOWN
	end_if;
    end_proc;
*/
