/*****************************************************************************
  $Id: ribbons.mu,v 1.19 2004/09/25 22:49:15 nthiery Exp $
  File:		lib/COMBINAT/ribbons.mu
  Authors:	Francois Descouens < francois.descouens@wanadoo.fr >
  Created:	15/11/2003
  License:	LGPL
*****************************************************************************/

//    This file permits manipulation of ribbons which are
//    +---+
//    | 6 |
//    +---+---+---+
//    | 5 | 7 | 8 |
//    +---+---+---+---+
//            | 2 | 4 |
//            +---+---+---+
//                | 1 | 3 |
//                +---+---+
//    its expression representation in MuPAD is a list of filling of each line (from up to down) :
//    [[6],[5,7,8],[2,4],[1,3]]


///////////////////////////////////////////////////////////////////////////////////
//    ribbons domain definition
///////////////////////////////////////////////////////////////////////////////////

domain combinat::ribbons
    inherits Dom::BaseDomain;
    category Cat::CombinatorialClassWith2DBoxedRepresentation;
    axiom    Ax ::canonicalRep;

    info := "Library 'combinat::ribbons': utilities for tableaux with shape of ribbons, particular skew Tableaux with connection conditions";

    interfaceAdd := {hold(isAStandard), hold(isA),  hold(fromShapeAndWord),
                     hold(fromWord), hold(toWord), hold(new)};
  
    //some particular type of ribbons
    //The without-condition filling
    typeFilling        := Type::Predicate("RibbonFilling", dom::isAFilling);
  
    //The semi standard filling of a ribbon 
    typeRibbon         := Type::Predicate("Ribbon", dom::isA);
  
    //The standard fiiling of a ribbon
    typeStandardRibbon := Type::Predicate("StandardRibbon", dom::isAStandard);
  
    ///////////////////////////////////////////////////////////////////////////////////
    /// functions for the new definition of axiom::canonicalRep
    ///////////////////////////////////////////////////////////////////////////////////
  
    new :=
    proc(x : Type::Union(dom::typeFilling, combinat::compositions) )
    begin 
        if args(0) >= 2 then error("too many arguments"); end_if;
        if x=[] then return(new(dom)); end_if;
        if testtype(x, combinat::compositions) then
	    return( dom::list(x));
	elif testtype(x, dom::typeFilling) then
	    return(dom::fromExpr(x));
	else
	    error("illegal arguments");
	end_if;
    end_proc;
  
    fromExpr  := x -> new(dom, op(x));
    expr      := [extop];
  
    _index    := extop;
  
    set_index := 
    proc(t:dom, i:Type::PosInt, value)
    begin
      extsubsop(t, i=value);
    end_proc;
  
    expr2text :=
    proc(t: dom)
    begin
      "new(".expr2text(dom).
      if extnops(t) > 0 then 
	",".expr2text(extop(t));
      else
	""
      end_if.
    ")";
    end_proc;
  
    convert := 
    proc (x: Type::Union(dom, DOM_LIST))
    begin
      case domtype(x)
	of dom do return(x);
	of DOM_LIST do
	  if nops(x) = 0 then return(dom::zero);
	  else return(dom::fromExpr(x));
	  end_if;
      end_case;
    FAIL;
    end_proc;
  
  
    ///////////////////////////////////////////////////////////////////////////////////
    ///    isA functions
    ///////////////////////////////////////////////////////////////////////////////////
  
    isAFilling :=
    proc(x: Type::AnyType) 
    begin
        if testtype(x,dom) then return(TRUE);
	elif testtype(x, Type::ListOf(Type::ListOf(Type::NonNegInt))) then return(TRUE);
	end_if;
        FALSE;
    end_proc;
  
    isA :=
    proc (ribTab: Type::AnyType, shape = FAIL : Type::Union(DOM_FAIL, combinat::compositions ))
	local k,length,i;
    begin
        if not(testtype(ribTab, Type::Union(dom,Type::ListOf(Type::ListOf(Type::NonNegInt)))))
            then return(FALSE);
	end_if;
	// test if ribbon has the good shape
	if args(0)>=2 and
            zip([nops(ribTab[i]) $i=1..nops(ribTab)],shape,_subtract,0)<>[0 $ nops(ribTab)]
                then return(FALSE);
	end_if;
      length:= nops(ribTab);
        for k from 1 to length do 
	    if sort(ribTab[k]) <> ribTab[k] 
		then return(FALSE);
	    end_if;
	end_for;
	for k from 1 to length-1 do
	    if ribTab[k] = [] or  ribTab[k+1] = [] or ribTab[k][nops(ribTab[k])] <= ribTab[k+1][1]
		then return(FALSE);
	    end_if;
	end_for;
	return(TRUE);
    end_proc;

    ///////////////////////////////////////////////////////////////////////////////////
    ///    isAStandard function
    ///////////////////////////////////////////////////////////////////////////////////


    isAStandard:=
    proc (ribTab: Type::AnyType,
              shape = FAIL : Type::Union(DOM_FAIL, combinat::compositions ))
	local length,size,i,boolean;
    begin
	boolean:=dom::isA(ribTab);
	if boolean= FALSE
	    then
	    return(FALSE);
	else
	    length:= nops(ribTab);
	    size:=_plus(nops(ribTab[i]) $i=1..length);
	    if {op(map(ribTab,op))} <> {$1..size}
		then return(FALSE);
	    else
		return(TRUE);
	    end_if;
	end_if;
    end_proc;

    ////////////////////////////////////////////////////////////////////////////////////
    //    fromShapeAndWord function
    ////////////////////////////////////////////////////////////////////////////////////

    fromShapeAndWord :=
    proc(shape: combinat::compositions, word: combinat::words)
	local i,j;
    begin
        if nops(word) <> _plus(shape[i] $i=1..nops(shape)) then return(FAIL); end_if;
	i := 0;		
        combinat::ribbons::new([[op(word,i+1..(i:=i+j))] $ j in shape]) ;
    end_proc;

    ////////////////////////////////////////////////////////////////////////////////////
    //    fromWord function and toWord
    ////////////////////////////////////////////////////////////////////////////////////

    //a tableau of ribbon shape is equivalent to the set of descent of a permutation
    //this function converts a permutation to a tableaux of ribbon shape

    fromWord:=
    proc (p: combinat::compositions)
	local j,i,compo;
    begin
	if p=[] then return(combinat::ribbons([])); end_if;
	compo:= combinat::permutations::descents(p);
	if nops(compo)=0 then return( combinat::ribbons::new([p]) ); end_if;
	combinat::ribbons::new([[p[j]$j=1..compo[1]],
         [p[j]$j=compo[i]+1..compo[i+1]] $i=1..nops(compo)-1,
	 [p[j]$j=compo[nops(compo)]+1..nops(p)]]);
    end_proc;

    //one-one inverse of previous correspondance
    toWord :=
    proc (rtab: dom)
    begin
	//[op(expr(rtab)[i])$i=1..nops(rtab)];
        map(expr(rtab), op);
    end_proc;

    ////////////////////////////////////////////////////////////////////////////////////
    //    list function and count function
    ////////////////////////////////////////////////////////////////////////////////////

    list :=
    proc (shape:combinat::compositions)
    begin
	map(combinat::permutations::fromDescentsComposition(shape),
             dom::fromWord);
    end_proc;

    ////////////////////////////////////////////////////////////////////////////////////
    //    Pretty-print
    ////////////////////////////////////////////////////////////////////////////////////

    toArray :=
    proc(p: Type::AnyType) : DOM_ARRAY
        local M, col, row, width, shift, l;
    begin
	// Note the max(1, instead of max(0, to please array
	// See Cat::CombinatorialClassWith2DBoxedRepresentation
	width := max(1, 1+_plus(nops(l)-1 $ l in p));
	M := array(1..max(1, nops(p)), 1..width);
	shift := 0;
	for row from 1 to nops(p) do
	    for col from 1 to nops(p[row]) do
	        M[row, col+shift] := p[row][col];
	    end_for;
	    shift := shift + nops(p[row])-1;
	end_for;
	M;
    end_proc;

end_domain:

