////////////////////////////////////////////////////////////////////////////////////
//    tableaux of ribbons domain definition
////////////////////////////////////////////////////////////////////////////////////

domain combinat::ribbonsTableaux
    inherits Dom::BaseDomain;
    category Cat::CombinatorialClassWith2DBoxedRepresentation;
    axiom Ax::canonicalRep;
  
    info := "Library 'combinat::ribbonsTableaux': utilities for tableaux of ribbons, a generalization of tableaux";
  
    interfaceAdd := {hold(new), hold(length), hold(evaluation),
                  hold(rCore), hold(rQuotient), hold(fromCoreAndQuotient),
                  hold(stantonWhite), 
                  hold(spinTableau), 
		  hold(spinPolynom), hold(cospinPolynom), hold(spinAndCospinPolynom)
                 };
  
    typeFilling := Type::Predicate("RibbonsTableauFilling", dom::isAFilling);
//////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////

    /////////////////////////////////////////////////////////////////////
    ///   functions for the new definition of axiom Ax::canonicalRep
    /////////////////////////////////////////////////////////////////////
  
    new       :=
    proc (x: Type::Union(combinat::partitions, Type::ListOf(DOM_LIST)) , y = FAIL: Type::Union(Type::ListOf(DOM_INT),FAIL) ,z = FAIL : Type::Union(Type::NonNegInt) )
    begin
      if args(0) >= 2 and testtype(x,combinat::partitions) then return( dom::list(x,y,z) ); end_if;
      if args(0)  = 1 and x = [] then return(dom::zero); end_if;
      if testtype(x, dom::typeFilling) then return(dom::fromExpr(x));
      else 
	  error("illegal arguments");
      end_if; 
    end_proc;
  
    fromExpr  := x->new(dom, op(x)); 
    expr      := [extop];
  
    _index    := extop;
  
    set_index := 
    proc (t:dom, i:Type::PosInt, value)
    begin
      extsubsop(t, i=value);  
    end_proc;
  
    expr2text := 
    proc (t: dom)
    begin
      "new(".expr2text(dom).
          if extnops(t) > 0 then 
	    ",".expr2text(extop(t));
	  else
	    ""
	  end_if.
      ")";
    end_proc;
  
    convert :=
    proc (x)
    begin
      case domtype(x)
	of dom do return(x);
	of Type::ListOf(DOM_LIST) do
	  if nops(x) = 0 then return(dom::zero); 
	  else return(dom::fromExpr(x));
	  end_if;
      end_case;
    FAIL;	  
    end_proc;  
  
///////////////////////////////////////////////////////////////////////////
//        isA functions
///////////////////////////////////////////////////////////////////////////
  
  isAFilling :=
  proc(x: Type::AnyType)
  begin
      if testtype(x,dom) then return(TRUE);
      elif testtype(x,Type::ListOf(Type::ListOf(Type::NonNegInt))) then return(TRUE); 
      end_if;
      TRUE;
  end_proc;
  
//////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////

    /////////////////////////////////////////////////////////////////////
    ///    r-core and r-quotient bijection for partition             ////
    /////////////////////////////////////////////////////////////////////

    //Algorithm in order to find the r-core of a partition
    /////////////////////////////////////////////////////////////////////
 
    rCore := 
    proc (part:combinat::partitions, length: Type::NonNegInt) : combinat::partitions
        local k, i, e;
    begin
        //lenght partition normalization
	if (nops(part) mod length <> 0) then 
	    part := append(part, 0 $ length-irem(nops(part),length));
	end_if;
        //adding canonical vector to the partition 
	part := [part[i] + nops(part)-i $ i=1..nops(part)];
        //Reduced modulo length with succesive representing
	for e from 0 to length-1 do
	    k := e;
	        for i from nops(part) downto 1 do
	            if part[i] mod length = e then
		        part[i] := k; 
		        k       := k+length;
		    end_if;
	        end_for;
	end_for;
	part := revert(sort(part));
        //removing canonical vector after operations
	part := [part[i]-nops(part)+i $ i=1..nops(part)];
        //selecting the r-core
	part := select(part, _not@iszero);
        return(part);
    end_proc;

    //Algorithm in order to find the r-quotient of a partition
    ///////////////////////////////////////////////////////////////////

    rQuotient := 
    proc (part:combinat::partitions, length: Type::NonNegInt) : Type::ListOf(combinat::partitions)
	local k, i, tmp, res, e;
    begin
        //length partition normalization
	if (nops(part) mod length <> 0) then 
	    part := append(part, 0 $ length-irem(nops(part), length));
	end_if;
        //adding canonical vector
	part := [part[i] + nops(part)-i $ i=1..nops(part)];
	res  := [FAIL $ length];
        //reducing vector 
	for e from 0 to length - 1 do
	    k   := e;
	    tmp := [];
	    for i from nops(part) downto 1 do
		if part[i] mod length = e then
		    tmp := append(tmp, (part[i]-k)/length);
		    k   := k + length;
		end_if;
	    end_for;
	    res[e+1] := revert(select(tmp, _not@iszero));
	end_for;
	return(res);
    end_proc;

    //Algorithm in order to find partition from its r-Core and r-Quotient
    /////////////////////////////////////////////////////////////////////

    fromCoreAndQuotient :=  
    proc(core: combinat::partitions, quotient : Type::ListOf(combinat::partitions)) : combinat::partitions
	local len, i, j, k, v, w;
    begin
	len := nops(quotient);
	k   := len*max(ceil(nops(core)/len), 
		       nops(core), op(map(quotient,nops))) + len;
	v   := [core[i]-i+1 $ i=1..nops(core), -i+1 $ i=nops(core)+1..k];
	w   := [select(v, proc(vl) begin bool(vl-i mod len=0) end) $ i=1..len];
	w   := [(op(w,[i,j]) + len*op(quotient,[i,j]) $ j=1..nops(quotient[i]), 
		 op(w,[i,j]) $ j=nops(quotient[i])+1..nops(w[i])
		) $ i=1..len];
	w   := revert(sort(w));
	select([w[i]+i-1 $ i=1..nops(w)], _not@iszero);
    end_proc;

//////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////


    ////////////////////////////////////////////////////////////////
    ///    Stanton-White bijection for ribbons tableaux         ////
    ////////////////////////////////////////////////////////////////


    //  from a ribbon tableau to k-uplets of semi-standard tableaux
    ///////////////////////////////////////////////////////////////

    stantonWhite:=
    proc( tableau: dom) 
        local l, full, k, i, j, res, eval, temp, sw;
    begin
        res  := null();
        l    := dom::length(tableau);
        eval := dom::evaluation(tableau);
        //Using full representation in order to separate each block with the same evaluation
        full :=dom::toFullRepresentation(tableau)[1];
        for k from 1 to nops(eval) do
            //for each evaluation extracting each block
            temp := map(full ,x->select(x, has, [i$i=-1..k]));
            //and computing its r-Quotient
            res := res, dom::rQuotient(revert(select([nops(temp[i]) $ i=1..nops(temp)], _not@has,0)), l);
        end_for;
        res := [res];
        //initializing the solution
        sw  := [    [[nops(eval) $i=1..res[nops(res)][k][j]] $ j=1..nops(res[nops(res)][k])]
		$ k=1..nops(res[nops(res)])    ];
        // filling of sw due to the order of apparition of each case in the previous partition list
        for temp from nops(eval) - 1 downto 1 do
            for i from 1 to l do
                for j from 1 to nops(res[temp][i]) do
                    for k from 1 to res[temp][i][j] do
                         sw[i][j][k] := temp;
                    end_for;
                end_for;
            end_for;
	end_for;
        sw := map(sw, revert);
        sw := map(sw, combinat::tableaux::new);
        return([dom::rCore(revert([nops(tableau[i]) $ i=1..nops(tableau)]), l), sw]);
    end_proc;


  //  from Stanton White to ribbons tableaux
  ///////////////////////////////////////////////////////////////
    

//////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////

    
    ///////////////////////////////////////////////////////////////////////////////////
    ///  function for listing ribbons tableau of a given shape and a given evaluation
    ///////////////////////////////////////////////////////////////////////////////////

    list := 
    proc(part:combinat::partitions, evaluation: Type::ListOf(DOM_INT), length: Type::NonNegInt) : Type::ListOf(dom)
	local core, retire, perms, partp, tmp, selection, work, h, k, i, j, tab;
    begin
        core := dom::rCore(part, length);
        j    := 0;
        j    := _plus(core[i] $ i=1..nops(core));	
        //we must test if arguments are compatible
        if ((_plus(part[i] $ i=1..nops(part))-j)<>
             _plus(evaluation[i] $ i=1..nops(evaluation)) * length) 
            then return(FAIL); 
        end_if;
        tab  := [ [0 $ i=1..part[nops(part)-j+1]] $ j=1..nops(part)];
        //filling with -1 the r-core of the partition
        for j from 1 to nops(core) do 
            for i from 1 to core[j] do
                tab[nops(part)-j+1][i]:=-1;
            end_for;
        end_for;
        //initialization of the vector on which we work
        work := [[tab,part]]; 
        for k from nops(evaluation) downto 1 do
            tmp:=null();
            for i from 1 to nops(work) do
                partp     := combinat::partitions::conjugate(work[i][2]);
                selection := null();
                perms     := combinat::permutations::list(
                             [0 $ (nops(partp)-evaluation[k]), length $ evaluation[k]]);
                //selection of valide permutation vectors
                for j from 1 to nops(perms) do
                    retire := dom::removeRibbons(partp, perms[j]);
                    if (retire <> FAIL)
                        then selection := selection, [retire[1],perms[j]];
      	            end_if;
                end_for;
                selection := [selection];
                //we list all valid tableau at this step (incomplete evaluation)          
                for h from 1 to nops(selection) do
                    tmp := tmp, [dom::insertionTableau(partp,selection[h][2], k, work[i][1]),
                             combinat::partitions::conjugate(selection[h][1])];  
                end_for;
            end_for;
            work := [tmp];
        end_for;
    return([ new(dom,op(work[i][1])) $ i=1..nops(work) ]);
    end_proc;

//////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////

    ///////////////////////////////////////////////////////////////////////////////////
    ///    Function which compute the spin of a given ribbons tableau
    ///////////////////////////////////////////////////////////////////////////////////

    spinTableau := 
    proc(tableau: dom ) : Type::PosRat
	local t, inv, k, i, j, partc, temp, part, l, eval;
    begin
        part  := [nops(tableau[nops(tableau)-k+1]) $ k=1..nops(tableau)];
        inv   := 0;
        eval  := dom::evaluation(tableau);
        l     := dom::length(tableau);
        partc := combinat::partitions::conjugate(part);

        for k from nops(eval) downto 1 do
            temp := null();
            //searching ribbons of eval k on the frontier of the current partition
            for j from 1 to part[nops(part)] do 
            //position into the tableau
                t := nops(tableau)-nops(part)+1;
	        if tableau[t][j] = k then temp := temp, l; 
	        else temp := temp, 0; end_if;
            end_for;
        for i from nops(part)-1 downto 1 do
            t := nops(tableau)-i + 1;
            for j from part[i+1] + 1 to part[i] do
	        if tableau[t][j] = k then temp := temp,l; 
		    else temp := temp, 0; end_if;
            end_for;
        end_for;
        temp := [temp];
	//remove corresponding ribbons on the conjugate current partition
        j    := dom::removeRibbons(partc,temp);
        partc:= j[1];
        part := combinat::partitions::conjugate(partc);
        //computing total number of inversions
        inv  := inv + j[2];
        end_for;
        //convert inversions into spin
        return((_plus(eval[i] $ i=1..nops(eval))*(l-1)-inv)/2);
    end_proc;

    ///////////////////////////////////////////////////////////////////////////////////
    ///  function which both compute spin and cospin polynom 
    ///  for a given shape and evaluation
    ///////////////////////////////////////////////////////////////////////////////////

    spinAndCospinPolynom := 
  proc(part:combinat::partitions, evaluation:combinat::compositions, length: Type::NonNegInt) 
	local m, e, core, retire, perms, partp, tmp, selection, work, h, k, i, j;
    begin
        core := dom::rCore(part, length);
        j    :=_plus(core[i] $ i=1..nops(core));	
        //we must test if arguments are compatible
        if ((_plus(part[i] $ i=1..nops(part)) - j) <>
             _plus(evaluation[i] $ i=1..nops(evaluation)) * length) 
            // or (select(zip(part,evaluation,_subtract), a->is(a,Type::NegInt)=TRUE) <> []) 
            then return(FAIL); 
        end_if;
        //initialization of the vector on which we work
        work :=[[0,part]];
        for k from nops(evaluation) downto 1 do
            tmp:=null();
            for i from 1 to nops(work) do
                partp     := combinat::partitions::conjugate(work[i][2]);
                selection := null();
                perms     := combinat::permutations::list(
							  [ 0 $ (nops(partp)-evaluation[k]), length $ evaluation[k]]);
                //selection of valide permutation vectors
                for j from 1 to nops(perms) do
                    retire:= dom::removeRibbons(partp, perms[j]);
                    if (retire <> FAIL)
                        then selection := selection, retire;
      	            end_if;
                end_for;
                selection := [selection];
                //we list all valid tableau at this step (incomplete evaluation)          
                for h from 1 to nops(selection) do
                    tmp := tmp, [work[i][1] + selection[h][2],
                             combinat::partitions::conjugate(selection[h][1])];  
                end_for;
            end_for;
            work := [tmp];
	end_for;
        if work = [] then return(FAIL); end_if;
        //computing spin and cospin polynom
        e := (_plus(evaluation[j] $ j=1..nops(evaluation)));
        //selecting maximal spin in order to compute the cospin
        m := max((e * (length - 1) - work[i][1])/2 $ i=1..nops(work));
        return( [_plus( hold(q)^((e * (length - 1) - work[i][1])/2) $ i=1..nops(work)),
		_plus( hold(q)^(m - (e * (length - 1) - work[i][1])/2) $ i=1..nops(work))]);
    end_proc;    

    ////////////////////////////////////////////////////////////////////
    // function which select the spin polynom in the previous function 
    ////////////////////////////////////////////////////////////////////

    spinPolynom := 
    proc( part:combinat::partitions, evaluation:combinat::compositions, length:Type::NonNegInt)
        local tmp;  
    begin
        tmp := dom::spinAndCospinPolynom(part, evaluation, length);
        if tmp = FAIL then return(FAIL) else return(tmp[1]); end_if;
    end_proc;

    ////////////////////////////////////////////////////////////////////
    // function which select the cospin polynom in the previous function 
    ////////////////////////////////////////////////////////////////////

    cospinPolynom := 
    proc(part:combinat::partitions, evaluation:combinat::compositions, length:Type::NonNegInt)
        local tmp;  
    begin
        tmp := dom::spinAndCospinPolynom(part, evaluation, length);
        if tmp = FAIL then return(FAIL) else return(tmp[2]); end_if;
    end_proc;

//////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////


    //////////////////////////////////////////////////////////////////////////////////
    //   Auxiliary functions for previous functionnality
    //////////////////////////////////////////////////////////////////////////////////

    //function which compute the length of the ribbons into a ribbons tableaux
    //////////////////////////////////////////////////////////////////////////
		       
    length :=   
    proc( Tableau: dom) : DOM_INT
        local l, k, total;
    begin
        l     := 0; 
        total := 0;
        for k from 1 to nops(Tableau) do	
            total := total + nops(select(Tableau[k],  x->is(x>-1)=TRUE));
            l     := l + nops(select(Tableau[k], x->is(x>0)=TRUE));
	end_for;
        return(total/l);
    end_proc;

    //function which returns the evaluation of a ribbons tableau
    ////////////////////////////////////////////////////////////

    evaluation := 
    proc( Tableau: dom) : combinat::compositions
        local l, k, i, eval, temp;
    begin
        l    := dom::length(Tableau);
        eval := null();
        temp := null();
        k    := 1;
        while(map(expr(Tableau), x->select(x,has,[i $ i=-1..k])) <> expr(Tableau) ) do
	    temp := map(expr(Tableau), x->select(x, has, k));
            eval := eval, _plus(nops(temp[i]) $ i=1..nops(temp));
            k    := k + 1;
        end_while;
        temp := map(expr(Tableau), x->select(x,has,k));
        eval := eval,_plus(nops(temp[i]) $ i=1..nops(temp));
        return([eval]);
    end_proc;

    //function which fill the empty cell of the ribbons tableaux representation
    //and return an other tableaux which gives us separation between cells      
    ///////////////////////////////////////////////////////////////////////////
     
    toFullRepresentation := 
    proc(Tableau: dom)
        local core, le, le2, result, separator, i, j, k, l, m, eval, part, cpart, tmp;
    begin
        //recovering information on the ribbons tableaux
        part      := revert([nops(Tableau[k]) $ k=1..nops(Tableau)]);
        cpart     := combinat::partitions::conjugate(part);
        eval      := dom::evaluation(Tableau);
        le        := dom::length(Tableau);
        core      := dom::rCore( revert([nops(Tableau[i]) $ i=1..nops(Tableau)]),le);
        //Initializing result tableau with empty cells
        result    := [ [0 $ i=1..nops(Tableau[k])] $ k=1..nops(Tableau)];	
        separator := result;
	//we will developp the entire lenght of the ribbons of the tableau
        tmp       := _plus(eval[i]$i=1..nops(eval)) + 1;
        for m from nops(eval) downto 1 do
	    for j from part[1] downto 1 do
                for i from cpart[j] downto 1 do
		    k   := nops(part) - i + 1; 
		    l   := j; 
		    le2 := le;
                    if Tableau[nops(part)-i+1][j]=m then
                    //we are sure the process succesfully stop
                        while le2 > 0 do
			    result[k][l]    := m;
			    separator[k][l] := tmp;
		            le2             := le2 - 1;
                            //we open out ribbons respecting connected condition
		            if   le2 > 0 and nops(Tableau[k]) > l 
			        	 and result[k][l+1] = 0 
		            then l := l+1;
			    elif le2 > 0 and nops(Tableau)>k    
				         and result[k+1][l] = 0 
			    then k := k+1;	 
                            end_if;
			end_while;
		        tmp := tmp - 1;
		    end_if;
                end_for;
            end_for;
        end_for;
        //filling the k-core with -1
        for k from 1 to nops(core) do
            for i from 1 to core[k] do
	        result[nops(result) - k + 1][i]    := -1;
	        separator[nops(result) - k + 1][i] := tmp;
            end_for;
        end_for;
        return([result,separator]);
    end_proc;

    //  function for removing ribbons on a partition which also returns
    //  the number of inversion which permits us computing 
    //  the spin of removed ribbons
    ////////////////////////////////////////////////////////////////////

    removeRibbons := 
    proc (part: Type::ListOf(Type::NonNegInt), plant: Type::ListOf(Type::NonNegInt))
	local res, resinv, delta, i;
    begin
        //compatibility conditions test
        if nops(plant) > nops(part) then return(FAIL); end_if;
	delta  := revert([i $ i=0..nops(part)-1]);
        part   := zip(part, delta, _plus);
        if nops(part) = 0 then return(FAIL); end_if;
        //Unplanting ribbons to the partitions
	res    := zip(part, plant, _subtract);
        if (nops({op(res)}) <> nops(res)) then
	    return(FAIL);
	end_if;
        resinv := res;
	res    := revert(sort(res));
        res    := zip(res,delta,_subtract);
	if (res[nops(res)] < 0 ) 
            then return(FAIL) 
            else 
            return([res,
		    combinat::permutations::inversionsNumber(
                    combinat::words::standard(revert(resinv)))]
                  ); 
        end_if;
    end_proc;   

    // function which permits to fill the tableau from a valide permutation vector 
    ///////////////////////////////////////////////////////////////////////////////

    insertionTableau := 
    proc( partition, perm, eval: Type::NonNegInt, tableau: Type::ListOf(DOM_LIST) )
        local k;
    begin   
        if (nops(perm)>nops(partition)) then return(FAIL); end_if;
        tableau := combinat::tableaux::conjugate(tableau);
        for k from 1 to nops(perm) do
            if perm[nops(perm) - k + 1] <> 0 then
            tableau[nops(tableau)-nops(perm) + k][partition[nops(perm) - k +1 ]] := eval;
            end_if;
        end_for;
        return(combinat::tableaux::conjugate(tableau));
    end_proc;
		       
//////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////


    //////////////////////////////////////////////////////////////////////////////////
    //   PrettyPrint for ribbons Tableaux
    //////////////////////////////////////////////////////////////////////////////////		       
		       
    //as combinat::ribbonsTableaux is into the CombinatorialClassWith2DBoxedRepresentation
    //we must define the variable toArrayVerticalAlignment and the function toArray
    ///////////////////////////////////////////////////////////////////////////////////////       
		       
    toArraySeparator :=  
    proc(tableau: dom): DOM_ARRAY
      local entire, col, row;
    begin
        entire := dom::toFullRepresentation(tableau)[2];
        array(1..max(1, nops(entire)),1..max(1, op(map(entire, nops))),
	      (row,col) = entire[row][col] $ col=1..nops(entire[row]) $ row=1..nops(entire));
    end_proc;  
      
    toArray :=  
    proc(tableau: dom ) : DOM_ARRAY
        local a, entire, col, row;
    begin
        entire := dom::toFullRepresentation(tableau)[1];
        a := array(1..max(1, nops(entire)),1..max(1, op(map(entire, nops))));
        for row from 1 to nops(entire) do
	    for col from 1 to nops(entire[row]) do
                 if entire[row][col] <> -1 then
		     a[row,col] := entire[row][col];
		 else
		     a[row,col] := "";
		 end_if;
	    end_for;  
	 end_for;
      return(a);
    end_proc;
		       
end_domain:












