/*****************************************************************************
  $Id: shuffle.mu,v 1.6 2004/02/18 13:46:55 hivert Exp $
  File:		lib/COMBINAT/shuffle.mu
  Authors:      Florent Hivert <florent.hivert@univ-mlv.fr>
  Created:	11/01/2004
  License:	LGPL
*****************************************************************************/

//////////////////////////////////////////////////////////////////////////////
// shuffle products of two words
// TODO : compute the shuffle product of *several* words.
domain combinat::words::shuffle
    inherits Dom::BaseDomain;
    category Cat::CombinatorialClass;
    axiom    Ax::systemRep;

    info      := "Library 'combinat::words::shuffle': utilities for the shuffle product of two words";
    typeWords := combinat::words;

    isA       := 
    proc(obj            : Type::AnyType,
	 w1    = null() : Type::Union(dom::typeWords, DOM_NULL),
	 w2    = null() : Type::Union(dom::typeWords, DOM_NULL)
	 ) : DOM_BOOL
    begin
	if domtype(obj) <> DOM_LIST then return(FALSE); end_if;
	if args(0) = 1 then return(TRUE); end_if;
	if nops(obj) <> nops(w1) + nops(w2) then return(FALSE) end_if;
	combinat::words::shufflings::exists(w1, w2, obj);
    end_proc;

    count     := 
    proc(word1: dom::typeWords, word2: dom::typeWords)
    begin
	binomial(nops(word1) + nops(word2), nops(word1));
    end_proc;

    generator :=
    proc(word1: dom::typeWords, word2: dom::typeWords)
      option escape;
      local len1, len2;
    begin
      len1 := nops(word1);
      len2 := nops(word2);
      combinat::generators::map
      (combinat::integerVectors::generator(len1, len1+len2, MaxPart=1),
       proc(vect)
           local i1, i2;
       begin
           i1 := 0; i2 := 0;
           map(vect,
               x -> if x=1 then i1:=i1+1; word1[i1];
                           else i2:=i2+1; word2[i2];
                    end_if
              )
       end_proc)
    end_proc;
begin
end_domain:


domain combinat::words::split
    inherits Dom::BaseDomain;
    category Cat::CombinatorialClass;
    axiom    Ax::systemRep;

    info := "Library 'combinat::splitNK': utilities for spliting of words";
    typeWords := combinat::words;

    isA  := 
    proc(obj : Type::AnyType, 
	 w = null()  : Type::Union(DOM_NULL, DOM_LIST), 
	 k = null()  : Type::Union(DOM_NULL, Type::NonNegInt))
    begin
	if not testtype(obj, Type::ListProduct(DOM_LIST $ 2))
	   then return(FALSE)
	end_if;
	if args(0) > 2 and nops(obj[1]) <> k then return(FALSE) end_if;
	if args(0) > 1 then 
	    return(combinat::words::splittings::exists(op(obj), w));
	end_if;
	return(TRUE);
    end_proc;

    count  := 
    proc(w : dom::typeWords, k: Type::NonNegInt)
    begin
	combinat::binomialRecursive(nops(w), k);
    end_proc;

    random := 
    proc(w : dom::typeWords, k : Type::NonNegInt)
	local s1,s2,i;
    begin
	[s1,s2] := combinat::splitNK::random(nops(w), k);
	[[w[i] $ i in s1], [w[i] $ i in s2]];
    end_proc;
		       
    unrank := 
    proc(r : Type::PosInt, w : dom::typeWords, k: Type::NonNegInt)
	local s1,s2,tmp,i;
    begin
	if (tmp := combinat::splitNK::unrank(r, nops(w), k)) = FAIL then return(FAIL) end_if;
	[s1,s2] := tmp;
	[[w[i] $ i in s1], [w[i] $ i in s2]];
    end;

    // TODO: add generator once there is one implemented in splitNK
    
end_domain;


// Shuffling the word w1 with the word w2 to obtain the word w.
//*********************************************************************************    
// A shuffling is a pair of stricly increasing functions 
//    fi : [1..nops(wi)] -> [1..nops(w)]
// such that w[fi[x]] = wi[x]
// and {image(fi), i=1..2} is a partition of {1..nops(w)}.
// Here shufflings are represented by lists.
// For example [[3,5,9,10],   [1,2,4,6,7,8]] is a shuffling of
//           w1=[1,2,1,2], w2=[1,1,2,3,2,1],  w=[1,1,1,2,2,3,2,1,1,2]
domain combinat::words::shufflings
    inherits Dom::BaseDomain;
    category Cat::CombinatorialClass;
    axiom    Ax::systemRep;

    info      := "Library 'combinat::words::shufflings': utilities for the shufflings of words";
    typeWords := combinat::words;

    shufflingsGeneric := 
    proc(empty, trivial, addKToSol, addSols)
	option escape;
    begin
    // Goes along the set of shufflings and construct the set of solutions
    // - empty        is the empty set of solution.
    // - trivial(l,k) is the singleton of the trivial solution [[$1..l], [$l+1..k]]
    // - addKToSol(sol,k,i) add k to the i-th functions (i=1,2) of the solutions of the sets sol
    // - addSols      add two solutions sets. 
	proc(w1 : dom::typeWords, w2 : dom::typeWords, w : dom::typeWords)
        local alpha, len1, len2, len, shufflings, newSh, k, l;
	begin
	    len  := nops(w); len1 := nops(w1); len2 := nops(w2);
	    alpha  := listlib::removeDuplicates(w.w1.w2);
	    if zip(combinat::words::evaluation(w1, alpha), 
		   combinat::words::evaluation(w2, alpha), _plus) 
	       <>  combinat::words::evaluation(w, alpha) then return(empty); end_if;
    // Loop invariant : 
    // At step k, Shufflings[i] contains the list of the shufflings of
    // the factor w1[1..i] with w2[1..k-i] which give w[1..k]
    // At the begining Shufflings[0] contains the trivial shuffling [[], []].
	    shufflings := array(0..nops(w));
	    shufflings[0] := trivial(0,0);
	    for k from 1 to len do
		if k <= len1 and shufflings[k-1] <> empty and w1[k] = w[k] then // Case : w1[1..k] = w[1..k]
		    shufflings[k] := trivial(k,k);
		else 
		    shufflings[k] := empty;
		end_if;
		for l from min(k-1, len1) downto max(1, k-len2) do 
		    if w1[l] = w[k] then 
			newSh := addKToSol(shufflings[l-1], k, 1);
		    else
			newSh := empty;
		    end_if;
		    if w2[k-l] = w[k] then 
			newSh := addSols(newSh, 
					 addKToSol(shufflings[l], k, 2));
		    end_if;
		    shufflings[l] := newSh;
		end_for;
		if k<=len2 and shufflings[0] <> empty and w2[k] = w[k] then // Case : w2[1..k] = w[1..k]
		    shufflings[0] := trivial(0,k);
		else 
		    shufflings[0] := empty;
		end_if;	
	    end_for;
	    shufflings[len1];
	end_proc;
    end_proc;

    list := dom::shufflingsGeneric([], (l,k) -> [ [[$1..l], [$l+1..k]] ],
        (suf,k,i) -> map(suf, proc(pair) begin subsop(pair, i=append(pair[i], k)) end_proc),
        _concat);
    count  := dom::shufflingsGeneric(0,     (l,k) -> 1,    (suf,k,i) -> suf, _plus);
    exists := dom::shufflingsGeneric(FALSE, (l,k) -> TRUE, (suf,k,i) -> suf, _or);
end_domain;


// splittings the word w to obtain the pair of words w1,w2.
//*********************************************************************************    
domain combinat::words::splittings
    inherits combinat::words::shufflings;
    category Cat::CombinatorialClass;
    axiom    Ax::systemRep;

    info      := "Library 'combinat::words::splittings': utilities for the splittings of words";

    list := dom::shufflingsGeneric([], (l,k) -> [ [1$l, 2$(k-l)] ],
        (suf,k,i) -> map(suf, append, i),
        _concat);
end_domain;


