/*****************************************************************************
  $Id: skewPartitions.mu,v 1.29 2004/07/12 20:23:45 nthiery Exp $
  File:		lib/COMBINAT/skewPartitions.mu
  Authors:	Teresa Gomez-Diaz <Teresa.Gomez-Diaz@univ-mlv.fr>
                Florent Hivert <florent.hivert@univ-mlv.fr>
		Nicolas M. Thiry <nthiery@users.sourceforge.net>
  Created:	05/05/2003
  License:	LGPL
*****************************************************************************/

domain combinat::skewPartitions
    inherits Dom::BaseDomain;
    category Cat::CombinatorialClassWith2DBoxedRepresentation;
    axiom    Ax::systemRep;

    info := "Library 'combinat::skewPartitions': utilities for skew partitions";

    interfaceAdd :=
        { hold(typeConnected), 
          hold(size), hold(overlap), hold(rowLenghts), 
          hold(printPretty), hold(printCompact),
          hold(conjugate), hold(jacobiTrudy),
          hold(outerCorners), hold(innerCorners),
          hold(printPrettyCorners), hold(printCompactCorners)
//          // "fast" standard filling
//          hold(fillStandard)
        };

    // needed because Ax::systemRep
    domtype := DOM_LIST;
 
    typeConnected := Type::Predicate("SkewPartition", dom::isA, Connected);

// Definitions:
// A skew partition skp is a list of two partitions [p, q], the second one 
// (inner partition) included in the first one (outer partition). 
// A skew partition is represented graphically by the boxes in the outer
// partition which are not in the inner partition.
// The overlap of two consecutive rows in a skew partition is the number
// of pairs of boxes (one in each row) which have a common edge.
// This number can be positive, zero  or negative :
//       ####           ####         ##
//         ###              ##           ###
//    overlap = 2    overlap = 0    overlap = -2
//    connected      unconnected    disconnect
// The overlap of a skp is the minimum of the overlaps of the consecutive rows,
// or infinity in the case of at most one row.
// If the overlap is positive, the skp is called "Connected" (reserved word).
// [[][]] has overlap infinity, size 0, rowLenghts [0] and is Connected.
 
    //////////////////////////////////////////////////////////////////////////
    // type functions
    //////////////////////////////////////////////////////////////////////////

    isA :=
    proc(skp : Type::AnyType,
        sizeN     = Connected : Type::Union(Type::NonNegInt, Type::Singleton(Connected)),
        overlapN  = Connected : Type::Union(Type::Integer, 
                                           stdlib::Infinity, Type::Singleton(Connected)))
    begin
       if args(0) > 3 then error("wrong number of arguments!"); end_if;
       if args(0) = 1 then return(dom::isSkp(skp)); end_if;
       if args(0) = 2 then 
           if args(2) = Connected then return(dom::isSkp(skp) and dom::isOverlapN(skp, 1));
           else return(dom::isSkp(skp) and dom::isSizeN(skp, sizeN));
           end_if;
       end_if;
       // arguments are 3
       if args(2) = Connected then error("second argument should be of type NonNegInt");
       end_if;
       if overlapN = Connected then overlapN:= 1; end_if;
       return(dom::isSkp(skp) and dom::isSizeN(skp, sizeN) and dom::isOverlapN(skp, overlapN));
    end_proc;

    isSkp :=
    // The skew partitions are by default disconnected ie. overlap = -infinity
    // The default case for count, list, generator is unconnected, ie. overlap = 0
    proc(skp)
        local p, q;
    begin
	if domtype(skp) <> DOM_LIST then return(FALSE); end_if;
        if nops(skp) <> 2 then return(FALSE); end_if;
        [p, q]:= skp;
        if not(testtype(p, combinat::partitions) and testtype(q, combinat::partitions))
        then return(FALSE);
        end_if;
        if p = [] then return(bool(q = [])); end_if;
        if q = [] then return(TRUE); end_if;
        return(bool(select(dom::rowLenghts_aux(skp), _less, 0) = []));
    end_proc;

    rowLenghts_aux := skp -> if skp[1] = [] then [0];
                             else zip(skp[1], skp[2], _subtract, 0);
                             end_if;

    size_aux := skp -> _plus(op(dom::rowLenghts_aux(skp)));

    isSizeN := (skp, N) -> bool(N = dom::size_aux(skp));

    overlap_aux := 
    proc(skp)
        local p, q;
    begin
        [p, q]:= skp;
        if nops(p) <= 1 then return(infinity); end_if;
        if q = [] then return(min(op(p))); end_if;
        q:= [op(q,1)].q;
        return(min(op(dom::rowLenghts_aux([p,q]))));
    end_proc;

    isOverlapN := (skp, N) -> bool(N <= dom::overlap_aux(skp));

    //////////////////////////////////////////////////////////////////////////
    // size, overlap, rowLenghts
    //////////////////////////////////////////////////////////////////////////

    size :=
    proc(skp : dom) : Type::NonNegInt
    begin  
        dom::size_aux(skp);
    end_proc;

    overlap :=
    proc(skp : dom) : Type::Union(Type::Integer, stdlib::Infinity)
    begin
        dom::overlap_aux(skp);
    end_proc;

    rowLenghts :=
    proc(skp : dom) : Type::ListOf(Type::NonNegInt)
    begin
        dom::rowLenghts_aux(skp);
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // Pretty-print
    //////////////////////////////////////////////////////////////////////////
    toArray :=
    proc(skp : dom) : DOM_ARRAY
        local p, q, col, row;
    begin
        [p, q]:= skp;
	// revert the two partitions once for all, and fill q with zeroes
	p := revert(p);
        q := revert(q.[0 $ nops(p)-nops(q)]);
        // Note the max(1, instead of max(0, to make array happy
	// See Cat::CombinatorialClassWith2DBoxedRepresentation
	array(1..max(1, nops(p)), 1..max(1, op(p)),
	      (row, col) = ""
	      $ col = q[row]+1..p[row]
	      $ row = 1..nops(p));
    end_proc;

    toArrayVerticalAlignment := -1; // alignement on the bottom row of the array

    printCompact :=
    proc(skp : dom) : output::asciiArt
        local p, q, row, col;
    begin
        [p, q]:= skp;
        q:= q.[0 $ nops(p)-nops(q)];
        output::asciiArt([_concat(" " $ col=1..q[row], 
                                  "#" $ col=q[row]+1..p[row]) $ row=1..nops(p)]);
    end_proc;
 
    //////////////////////////////////////////////////////////////////////////
    // count
    //////////////////////////////////////////////////////////////////////////

    count_slide:=
    // auxiliary function for count
    // count_slide(compo, overlap) counts all the skew partitions related to 
    // the composition compo by "sliding". (compo has is the list of rowLenghts).
    // See fromRowLengths_aux function and note that if connected, skew partitions
    // are nn = min(ck_1, ck)-1 but the for loop begins in step 0.
    // The skew partitions are unconnected if overlap = 0 (default case).
    proc(compo: combinat::compositions,
         overlap = 0 : Type::Integer) : Type::NonNegInt
        local nn, resu, i, nomb;
    begin
        nn:= nops(compo);
        resu:= 1;
        if nn > 1 then
            for i from 1 to nn-1 do
                nomb:= min(op(compo,i), op(compo, i+1));
                nomb:= nomb + 1 - overlap;
                resu:= resu*nomb;
            end_for;
        end_if;
        return(resu);
    end_proc;

    count:=
    proc(n: Type::NonNegInt,
        overlap = 0 :
        Type::Union(Type::Integer, Type::Singleton(Connected))) : Type::NonNegInt
       local gg, suma, co;
    begin
        if n = 0 then return(1); end_if;
        if overlap = Connected then overlap:= 1; end_if;
        gg := combinat::compositions::generator(n):
        if overlap > 0 then
            gg := combinat::compositions::generator(n, MinPart = overlap):
        end_if;
        suma:= 0;
        while (co := gg()) <> FAIL do suma:= suma + dom::count_slide(co, overlap); end;
        return(suma);
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // fromRowLengths-list 
    //////////////////////////////////////////////////////////////////////////

    fromRowLengths_aux:=
    // auxiliary function for fromRowLengths
    // fromRowLengths_aux(skp, ck_1, ck, overlap) is a step in the computation of 
    // the skew partitions related to the composition [c1,c2,..ck-1,ck].
    // skp corresponds to a skew partition related to [c1,c2,..ck-1],
    // and fromRowLengths_aux computes all the possibilities when sliding part 
    // "ck" added to the top of skp. (old "slide function")
    // The skew partitions are unconnected if overlap = 0 (default case).
    proc(sskp, ck_1, ck, overlap = 0)
        local i, lskp, nn, mm, skp1, skp2;
    begin
        lskp:= [];
        nn:= min(ck_1, ck);
        mm:= max(0, ck-ck_1);
        // nn should be >= 0. In the case of positive overlap,
        // the minPart conditon insures ck >= overlap for all k
        nn:= nn - overlap;
        for i from 0 to nn do
            [skp1, skp2]:= sskp;
            skp2:= skp2.[0 $ nops(skp1)-nops(skp2)];
            skp1:= map(skp1, _plus, mm, i);
            skp1:= skp1.[ck];
            skp2:= map( skp2, _plus, mm, i );
            skp2:= select(skp2, _unequal,0);
            lskp:= lskp.[ [skp1,skp2] ];
        end_for;
        return(lskp);
    end_proc;

    fromRowLengths:=
    // fromRowLengths(compo, connect) recursively computes all the skew partitions 
    // related to composition compo. The composition gives the rowLenghts of all
    // the generated skew partitions.
    // The skew partitions are unconnected if overlap = 0 (default case).
    proc(compo: combinat::compositions,
         overlap = 0 : Type::Integer) : Type::ListOf(dom)
        local DOM, nn;
        option escape;
    begin
        if compo = [] then return([[[],[]]]);  
        end_if;
        DOM:= dom;      
        nn:= nops(compo);
        if nn = 1 then return([[[op(compo,1)],[]]]);
        end_if;
        return(map(
            DOM::fromRowLengths([op(compo, 1..nn-1)], overlap),
            op@DOM::fromRowLengths_aux, op(compo, nn-1), op(compo, nn), overlap));
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // list
    //////////////////////////////////////////////////////////////////////////

    list :=
    // second argument stablish connected or unconnected type for elements
    // in the list of results, unconnected is assumed as default
    proc(n: Type::NonNegInt,
         overlap = 0 :
         Type::Union(Type::Integer, Type::Singleton(Connected))) : Type::ListOf(dom)
        local DOM;
    begin
        DOM:= dom;
        if overlap = Connected then overlap:= 1; end_if;
        if overlap > 0 then
            map(combinat::compositions::list(n, MinPart = overlap), 
                                                 op@DOM::fromRowLengths, overlap);
        else
            map(combinat::compositions::list(n), op@DOM::fromRowLengths, overlap);
        end_if;
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // fromRowLenghts-generator
    //////////////////////////////////////////////////////////////////////////

    slide_gen_aux:=
    // auxiliary function for fromRowLenghts_generator
    // slide_gen_aux(skp, ck_1, ck, overlap) is a step in the computation of 
    // the skew partitions related to the composition [c1,c2,..ck-1,ck].
    // skp corresponds to a skew partition related to [c1,c2,..ck-1],
    // and slide_gen_aux computes all the possibilities when sliding part 
    // "ck" added to the top of skp
    // The skew partitions are unconnected if overlap = 0 (default case).
    proc(sskp, ck_1, ck, overlap = 0)
        local i, mm, nn;
        option escape;
    begin
        nn:= min(ck_1, ck);
        mm:= max(0, ck - ck_1);
        // nn should be >= 0. In the case of positive overlap,
        // the minPart conditon insures ck >= overlap for all k
        nn:= nn - overlap;
        i:= -1;
        proc()
            local skp1, skp2;
        begin
            i:= i+1;
            [skp1, skp2]:= sskp;
            skp2:= skp2.[0 $ nops(skp1)-nops(skp2)];
            skp1:= map(skp1, _plus, mm, i);
            skp1:= skp1.[ck];
            skp2:= map( skp2, _plus, mm, i );
            skp2:= select(skp2, _unequal,0);
            if i <= nn then return([skp1,skp2]);
            else FAIL;
            end_if;
        end_proc;
    end_proc;

    fromRowLenghts_generator :=
    // fromRowLenghts_generator(compo, overlap) recursively computes all the skew partitions 
    // related to composition compo, which gives the rowLenghts of the skew partition. 
    // The skew partitions are unconnected if overlap = 0 (default case).
    proc(compo: combinat::compositions,
         overlap = 0 : Type::Integer) : combinat::generators
        local DOM, nn;
        option escape;
    begin
        if compo = [] then return(combinat::generators::fromList([[[],[]]]));  
        end_if;
        DOM:= dom;      
        nn:= nops(compo);
        if nn = 1 then return(combinat::generators::fromList([[[op(compo,1)],[]]]));
        end_if;
        combinat::generators::pipe(
            DOM::fromRowLenghts_generator([op(compo, 1..nn-1)], overlap),
            DOM::slide_gen_aux, op(compo, nn-1), op(compo, nn), overlap);
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // generator
    //////////////////////////////////////////////////////////////////////////

    generator :=
    // second argument stablish connected or unconnected type for elements
    // in the generator of results, unconnected is assumed as default
    proc(n: Type::NonNegInt,
         overlap = 0 :
         Type::Union(Type::Integer, Type::Singleton(Connected))) : combinat::generators
        local DOM;
        option escape;
    begin
        DOM:= dom;
        if overlap = Connected then overlap:= 1; end_if;        
        if overlap > 0 then
            combinat::generators::pipe(combinat::compositions::generator(n, MinPart = overlap), 
                                                          DOM::fromRowLenghts_generator, overlap);
        else
            combinat::generators::pipe(combinat::compositions::generator(n), 
                                                          DOM::fromRowLenghts_generator, overlap);
        end_if;
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // conjugate
    //////////////////////////////////////////////////////////////////////////

    conjugate :=
    proc(skp: dom) : dom
    begin
        map(skp, combinat::partitions::conjugate);
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // jacobiTrudy
    //////////////////////////////////////////////////////////////////////////

    jacobiTrudy :=
    proc(skp: dom, funct = (i -> h[i])) : Dom::Matrix()
    // matrix indices are [[(1,1),(2,1)...],[(1,2),(2,2)...]]
        local nn, p, q, row, col, M;
    begin
        [p, q]:= skp;
        nn:= nops(p);
        q:= q.[0 $ nn-nops(q)];
//      M:= matrix(nn,nn);       (bug !)
//      (M[row,col] := (op(p,col)-op(q,row)-col+row);) $ col= 1..nn $ row= 1..nn;
        M:= array(1..nn,1..nn,
		   [[op(p,col)-op(q,row)-col+row $ col= 1..nn] $ row= 1..nn]);
        M:= map(M, i-> if i < 0 then 0 else funct(i); end_if);
        return(matrix(M));
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // corners
    //////////////////////////////////////////////////////////////////////////

    outerCorners :=
    proc(skp: dom) : Type::ListOf(Type::ListOf(Type::NonNegInt))
    begin   
        combinat::partitions::corners(op(skp,1));
    end_proc;

    innerCorners :=
    proc(skp: dom) : Type::ListOf(Type::ListOf(Type::NonNegInt))
        local skp2, lcors, nn, i, pi;
    begin
        skp2:= op(skp,2);
        if skp2 = [] then
            if op(skp,1) = [] then return([]);
            else return([[1,1]]);
            end_if;
        end_if;
        lcors:= [[1, op(skp2,1)+1]];
        nn:= nops(skp2);
        if nn <> 1 then
            for i from 2 to nn do
                pi:= op(skp2,i);
                if pi <> op(skp2,i-1) then
                    lcors:= lcors.[[i,pi+1]];
                end_if;
            end_for;
        end_if;
        lcors:= lcors.[[nn+1,1]];
        return(lcors);
    end_proc;


    //////////////////////////////////////////////////////////////////////////
    // Pretty-print for corners
    //////////////////////////////////////////////////////////////////////////
 
    // "*" for outer corners, "x" for inner, and "%" for 
    // corners which are both inner and outer at the same time

    printPrettyCorners :=
    proc(skp : dom) : output::asciiArt
        local p, q, row, i,
	      fil, innerCornerRows, outerCornerRows;
    begin
        [p, q]:= skp;
	// Should use dom::rowLengths!
	fil := [ [ "" $ i ] $ i in revert(zip(p,q, _subtract, 0))];
	// We should check the convention for the coordinates
	// returned by inner/outerCorners
	innerCornerRows := map(dom::innerCorners(skp), c -> nops(p)+1-op(c,1));
	outerCornerRows := map(dom::outerCorners(skp), c -> nops(p)+1-op(c,1));
	for i in innerCornerRows do
	    fil[i][1] := "x";
	end_for;
	for row in outerCornerRows do
	    if fil[row][nops(fil[row])] = "x" then
		fil[row][nops(fil[row])] := "%"
	    else
		fil[row][nops(fil[row])] := "*"
	    end_if;
	end_for;
	combinat::skewTableaux::printPretty([q, fil]);
    end_proc;
 
    printCompactCorners :=
    proc(skp : dom) : output::asciiArt
        local i, p, q, lasci, oc, ic, row, col;
    begin   
        [p, q]:= skp;
        oc:= combinat::partitions::corners(p);
        ic:= dom::innerCorners(skp);
        q:= q.[0 $ nops(p)-nops(q)];
        lasci:= [_concat(" " $ col=1..q[row],
                         "#" $ col=q[row]+1..p[row]) $ row=1..nops(p)];
        %if "12"[1]="1" then // work around a change in strings indexed 
                             // access from MuPAD 3.0.0 on
            (lasci[oc[i][1]][oc[i][2]]:= "*";) $ i = 1..nops(oc);
            (if lasci[ic[i][1]][ic[i][2]] = "*" then lasci[ic[i][1]][ic[i][2]]:= "%";
             else lasci[ic[i][1]][ic[i][2]]:= "x"; end_if;) $ i = 1..nops(ic);
        else
            (lasci[oc[i][1]][oc[i][2]-1]:= "*";) $ i = 1..nops(oc);
            (if lasci[ic[i][1]][ic[i][2]-1] = "*" then lasci[ic[i][1]][ic[i][2]-1]:= "%";
             else lasci[ic[i][1]][ic[i][2]-1]:= "x"; end_if;) $ i = 1..nops(ic);
        end_if;
        output::asciiArt(lasci);
    end_proc;

//    //////////////////////////////////////////////////////////////////////////
//    // "fast" standard filling
//    //////////////////////////////////////////////////////////////////////////
//
//    cornerGenerator_aux :=
//    // pp, qq have same lenght as lists, qq is a inner partition of partition pp, 
//    // qq can have 0 as entries
//    // rr is the number of corners in each row in every iteration (ie, pp - qq)
//    proc(pp, rr)
//        local lindex, i, ind, lcors, nn, newind, newcor;
//    begin
//        lindex:= [select(i $ i = 1..nops(rr), x-> op(rr,x) <> 0)];
//        if lindex = [] then
//            return(combinat::generators::empty);
//        else
//            ind:= op(lindex, 1);
//            lcors:= [[ind, op(pp, ind) - op(rr, ind) + 1]];
//            nn:= nops(lindex);
//            if nn <> 1 then
//                for i from 2 to nn do
//                    ind:= op(lindex, i - 1);
//                    newind:= op(lindex, i);
//                    newcor:= op(pp, newind) - op(rr, newind) + 1;
//                    if newcor <> op(pp, ind) - op(rr, ind) + 1 then
//                        lcors:= lcors.[[newind, newcor]];
//                    end_if;
//                end_for;
//            end_if;
//        end_if;
//        return(lcors);
//    end_proc;
//
//    cornerGenerator :=
//    // given a skewPartition skp, this function returns a generator
//    // to travel over the corners of skp, starting by the inner corner
//    // in the first row, second the innerCorner in the second row, ...
//    // when all the inner corners are exhausted, the process is iterated even
//    // if there is not a skewPartition any more.
//// example g:=combinat::skewPartitions::cornerGenerator([[5,4,3,3,1],[]])
//// example g:=combinat::skewPartitions::cornerGenerator([[5,4,3,3,1],[4,1,1]])
//    proc(skp : dom) : combinat::generators
//        local p, q, r, lcorners;
//    begin   
//        [p, q]:= skp;
//        r:= zip(p, q, _subtract, 0);
//        if q = [] then
//            r[1]:= r[1] - 1;
//            //needs to iterate over r (differences)
//            lcorners:= [[1,1]].dom::cornerGenerator_aux(p, r)
//        else
//            //needs to iterate over r (differences)
//            lcorners:= dom::cornerGenerator_aux(p, r);
//        end_if;
//        return(lcorners);
//    end_proc;
//
//    fillStandard :=
//    // a skew standard tableau is [[2,1], [[0,0, 1,2,3], [0, 4,5], [6]]]
//    // where [2,1] is the inner partition of skp
//    // and [2+nops([1,2,3]), 1+ nops([2,3]), nops([4])] is the outer partition
//    //                     +---+
//    //                     | 6 |
//    //                     +---+---+---+
//    //                         | 4 | 5 |
//    //                         +---+---+---+---+
//    //                             | 1 | 2 | 3 |
//    //                             +---+---+---+
//    proc(skp : dom) : Type::ListOf(Type::ListOf(Type::NonNegInt))
//        local p, q, corgen, res, i, cor, row;
//    begin   
//        [p, q]:= skp;
//        //r:= zip(p, q, _subtract, 0);
//        //nn:= dom::size(skp);
//        corgen:= dom::cornerGenerator(skp);
//        res:= [[0 $ p[row]] $ row=1..nops(p)];
//        i:= 1;
//        while (cor := corgen()) <> FAIL do 
//            res[op(cor,1)][op(cor,2)]:= i;
//            i:= i+1;
//        end_while;
//        //////////// type of result !!
//        return(res);
//    end_proc;
//
//    //////////////////////////////////////////////////////////////////////////
//    // Pretty-print standard filling
//    //////////////////////////////////////////////////////////////////////////
// 
//    printPrettyFilling :=
//    //proc(skp: dom, fil: Type::ListOf(Type::ListOf(Type::NonNegInt)) ) : output::asciiArt
//    proc(skp: dom) : output::asciiArt
//        local M, col, row, fil, np, p, q;
//    begin
//        [p, q]:= skp;
//        fil:= dom::fillStandard(skp);
//        np:= nops(p);
//        q:= q.[0 $ np-nops(q)];
//        M := array(1..np, 1..max(1, op(p)));
//        (M[row,col] := "") $ col=q[row]+1..p[row] $ row=1..np;
//        (M[row,col] := fil[row][col]) $ col=q[row]+1..p[row] $ row=1..np;
//        output::asciiArt::fromArray(M);
//    end_proc;

end_domain:
