/*****************************************************************************
  $Id: structure.mu,v 1.24 2003/03/06 18:26:46 nthiery Exp $
  File:		lib/COMBINAT/structure.mu
  Authors:	Nicolas M. Thiry <nthiery@users.sourceforge.net>
                Florent Hivert <florent.hivert@univ-mlv.fr>
  Created:	16/02/2001
  License:	LGPL
*****************************************************************************/

category combinat::structure
    category Cat::BaseCategory;

    info := "Category 'combinat::structure': generic utilities for combinatorial structures";

/*****************************************************************************
Many operations, like counting or generating, are common to all
combinatorial structures. This category defines a standardized
interface, and generic tools.

 Standardized entries (not all of them need to be implemented):
 - type				MuPAD type for the structure
 - testtype(p)			test if p has the correct type
 - check ???			thorough test ?
 - size(s)			size of the structure s
 - count(<size>)		number of structures of size <size>
 - generatingSeries(<ident>)	generating series by size
 - generator(<size>)		generator for all the structures of size <size>
 - list(<size>)			list of all the structures of size <size>
 - first(<size>)		first structure of size <size>, or FAIL
 - last(<size>)			last structure of size <size>, or FAIL
 - Next(s)			next structure of size <size>, or FAIL
 - previous(s)			previous structure of size <size>, or FAIL
 - rank(s)			rank of the structure s
 - unrank(n, <size>)		n-th structure of size <size>, or FAIL
 - random(<size>)		random structure of size <size>
 - _less(s1,s2)			comparison of the structures s1 and s2

 Moreover, calling directly combinat::mystructure(<size>) is a
 shortcut for combinat::mystructure::list(<size>). See
 combinat::partitions for an exception to this rule.

 Required entries:
 - type
 - generator or list

   Warning: if none of them is provided, the default implementations
   that are provided will call each other, and cause an infinite loop!

 Default implementations are provided for:
 - count
 - list
 - new
 - generator
*****************************************************************************/

    interface_Structure :=
    { hold(type), hold(testtype), // hold(size),
      hold(count), hold(list), hold(generator)
    };

    interface := dom::interface_Structure; // to allow for reuse

    //////////////////////////////////////////////////////////////////////////
    // Type checking
    //////////////////////////////////////////////////////////////////////////

    /// the domtype of the elements of this domain (for facade domains)
    domtype  :=
    if dom::hasProp(Ax::systemRep) then
	toBeDefined;
    end;

    /// the type of the elements of this domain (for facade domains)
    /// defaults to domtype
    type :=
    if dom::hasProp(Ax::systemRep) then
	dom::domtype;
    end;

    /// default implementation of testtype for facade domains using dom::type
    testtype :=
    if dom::hasProp(Ax::systemRep) then
	proc(p,T)
	    option noDebug;
	begin
	    if T = dom then
		testtype(p, dom::type);
	    else
		FAIL
	    end_if;
	end_proc:
    end_if;

    /**
     * Syntactic sugar: when there is no conflict with the construction
     * of domain elements (typically for facade domain), calling the
     * domain constructor is a shortcut for list. This way one can write
     * combinat::compositions(4) to get the list of the compositions of 4
     *
     * For non facade domains, this is the same as Cat::BaseCategory::new
     * except that dom::list is called if the conversion returns FAIL.
    **/
    new :=
    if dom::hasProp(Ax::systemRep) then
	dom::list;
    else
	proc()
	begin
	    dom::convert(args());
	    if % = FAIL then
		dom::list(args());
	    else
		%
	    end_if
	end_proc;

    end_if;

    convert :=
    if dom::hasProp(Ax::systemRep) then
	x -> if domtype(x)=dom::domtype then x else FAIL; end_if;
    end_if;

    expr :=
    if dom::hasProp(Ax::systemRep) then
	expr;
    end_if;

// Tentative implementation with isA.
//    testtype := (p,T) -> if T = dom then dom::isA(x); else FAIL end_if;
//    type      := Type::Predicate(dom::isA);
//    isA       := dom::testtype;

    //////////////////////////////////////////////////////////////////////////
    //

    count     := dom::count_FromGenerator;
    list      := dom::list_FromGenerator;

    generator := (if dom::first <> FAIL and dom::Next <> FAIL then
                      dom::generator_FromNext;
                  elif dom::last <> FAIL and dom::previous <> FAIL then
                      dom::generator_FromPrevious;
                  else
		      if dom::generatorFromListLoopGard = TRUE then
			  ()->error(expr2text(dom).
				    "::generator called but not implemented");
		      else
			  dom::generatorFromListLoopGard := TRUE;
			  dom::generator_FromList;
		      end_if;
                  end_if);

    //////////////////////////////////////////////////////////////////////////
    // Default implementations
    //////////////////////////////////////////////////////////////////////////
    // With Pref::typeCkeck(Interactive) the use of ... @ ... ensures that
    // the type checking is done 
    // (cf mailing list, Christopher Creutzig, 2002 june 27) 
    ///////
    count_FromGenerator := combinat::generators::count  @ dom::generator;
    list_FromGenerator  := combinat::generators::toList @ dom::generator;
    
    generator_FromList  := combinat::generators::fromList @ dom::list;
    //////////////////////////////////////////////////////////////////////////
    // One has to use auxiliary function:
    // generator_FromNext := (z->[...]::fromNext(z, dom::Next)) @ dom::first;
    // does not work.
    // (cf mailing list, Christopher Creutzig, 2002 june 27 
    ///////
    generator_FromNextAux  := 
        z -> combinat::generators::fromNext(z, dom::Next); 
    generator_FromPreviousAux := 
        z -> combinat::generators::fromNext(z, dom::previous);    
    generator_FromNext     := dom::generator_FromNextAux @ dom::first;
    generator_FromPrevious := dom::generator_FromPreviousAux @ dom::last;

    generator_FromUnrank   := 
    proc(n: Type::NonNegInt)
	local DOM, r, options;
	option escape;
    begin
	DOM := dom;
	r := 0;
	options := args();
        () -> (r:=r+1; DOM::unrank(r, options));
    end_proc;

begin
end_category:

stdlib::LoadlibState["combinat::structure"] := TRUE:
