/*****************************************************************************
  $Id: words.mu,v 1.48 2004/09/29 22:29:26 nthiery Exp $
  File:		lib/COMBINAT/words.mu
  Authors:	Nicolas M. Thiry <nthiery@users.sourceforge.net>
  		Florent Hivert <florent.hivert@univ-mlv.fr>
  Created:	16/02/2001
  License:	LGPL
*****************************************************************************/

domain combinat::words
    inherits Dom::BaseDomain;
    category Cat::CombinatorialClass;
    axiom    Ax::systemRep;

    info := "Library 'combinat::words': utilities for words";

    interfaceAdd :=
       { hold(evaluationTable), hold(evaluationPartition),
	 hold(evaluation), hold(evaluationSparse),
	 hold(alphabetToOrder), hold(standard),
	 hold(fromStandardAndEvaluation),
	 hold(shuffleAugmented),
	 hold(swap),
	 // Descents
	 hold(descents), hold(descentsComposition), hold(descentsNumber),
	 hold(lexLess),    hold(degLexLess),
	 hold(invLexLess), hold(degInvLexLess),
	 hold(revLexLess), hold(degRevLexLess),
	 hold(unpairedLetters)
         };

    interfaceAddIfExists :=
        {// those are domains and might not exists in subdomains
         hold(fromStandard),
	 hold(shuffle), hold(shufflings),
	 hold(split),   hold(splittings)
	};

    domtype := DOM_LIST;
    typeAlphabet := Type::Union(DOM_LIST, DOM_SET, Type::NonNegInt);
    typeOrder    := Type::Function;
    typeStandard := combinat::permutations::typeStandard;


    //////////////////////////////////////////////////////////////////////////
    // isA
    //////////////////////////////////////////////////////////////////////////

    isA :=
    proc(obj         : Type::AnyType,
	 alph = FAIL : Type::Union(dom::typeAlphabet, DOM_FAIL),
	 k    = FAIL : Type::Union(dom::typeAlphabet, DOM_FAIL)
	 ) : DOM_BOOL
    begin
	if domtype(obj) <> DOM_LIST then return(FALSE); end_if;
	if args(0) = 1 then return(TRUE); end_if;
	if testtype(alph, Type::NonNegInt) then 
	    alph := {$1..alph};
	else
	    alph := {op(alph)};
	end_if;
	if {op(obj)} minus alph <> {} then return(FALSE); end_if;
	if args(0) = 2 or nops(obj) = k then 
	    TRUE
	else
	    FALSE
	end_if;
    end_proc;
	
    //////////////////////////////////////////////////////////////////////////
    // count
    //////////////////////////////////////////////////////////////////////////

    count :=
    proc(n=FAIL: dom::typeAlphabet, k=FAIL: Type::NonNegInt)
	: Type::Union(Type::NonNegInt, Type::Singleton(infinity))
    begin
	// Extension that allows to use it as Domain argument of
	// combinat::decomposableObject:
	if args(0) = 0 then return(infinity) end_if;
        if testargs() and args(0) <> 2 then
            error("Wrong number of arguments") ;
        end_if ;

        if not testtype(n, Dom::Integer) then
            n:=nops(n);
        end_if;
        return(n^k);
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // generator (special case of the generator of combinat::cartesian)
    //////////////////////////////////////////////////////////////////////////

    generator :=
    proc(A: dom::typeAlphabet, k: Type::NonNegInt)
        : combinat::generators
    /* 
       generator(A, k) 		generator of all words of length k of A
       generator(n, k)	 	generator of all words of length k of [1..n]

       Each call of the generator yields the following word
       lexicographically or FAIL if there is no more. There is
       absolutely no waranty if the generator is called again after
       FAIL was returned.

       Is the result reasonable when a A contains twice an element ? */
        local word, word_int, n, first_time;
        option escape;
    begin
        if testargs() and args(0) <> 2 then
            error("Wrong number of arguments") ;
        end_if ;

        if testtype(A, Dom::Integer) then
            A:=[$1..A];
        end_if;
        n:=nops(A);
        if k=0 then
            return(combinat::generators::repeater([],1));
        end_if;
        if n=0 then
            return(combinat::generators::empty);
        end_if;
        word_int:=[1 $ k];
        word:=[op(A,1) $ k];
        first_time:=TRUE;
        // Invariants:
        //  - word_int is a word of k integers in [1..n]
        //  - word is the corresponding word of k elements in A
        //  - first_time is true if and only if it's the first run
        return(proc(): Type::Union(dom, DOM_FAIL)
                   local index;
               begin
                   if first_time then
                       first_time:=FALSE;
                       return(word);
                   end_if;
                   for index from k downto 1 do
                       if word_int[index]<n then
                           word_int[index]:=word_int[index]+1;
                           word[index]:=op(A, word_int[index]);
                           return(word);
                       else 
                           word_int[index]:=1;
                           word[index]:=op(A,1);
                       end_if;
                   end_for;
                   return(FAIL);
               end_proc)
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // Methods
    //////////////////////////////////////////////////////////////////////////

    /**
	evaluationTable(word) - evaluation of word, as a table

        This is a table such that for each letter l appearing in word,
	t[l] is the number of occurences of l in word.

	Example:
		combinat::words::evaluationTable([b,a,d,b,c,d,b]);
			table(a=1, b=3, c=1, d=2)

    **/
    evaluationTable :=
    proc(word: dom) : DOM_TABLE
        local l,t;
    begin
	t:=table();
        for l in word do
            if contains(t,l) then
                t[l] := t[l]+1;
            else
                t[l] := 1;
            end_if;
        end_for;
	t;
    end_proc;

    /**
	evaluationPartition(word) - evaluation of word, as a partition

        This is the list of number of occurences of each letters in
	word, sorted decreasingly.

	Example:
		combinat::words::evaluationPartition([b,a,d,b,c,d,b]);
			[3,2,1,1]
    **/
    evaluationPartition :=
    proc(word: dom) : combinat::partitions
    begin
        revert(sort(map([op(dom::evaluationTable(word))], op, 2)));
    end_proc;


    /**
	evaluation(word) - evaluation of word, as an integer vector

        This is the list of number of occurences of each letters in
	word. Either the word is a word on integers or you must
	provide the alphabet as a list.

	Example:
		combinat::words::evaluationPartition([b,a,d,b,c,d,b]);
			[3,2,1,1]
    **/
    evaluation :=
    proc(word  : dom,
	 alpha = FAIL : Type::Union(dom::typeAlphabet, DOM_FAIL)
	 ) : combinat::integerVectors
	local tmp;
    begin
	if alpha=FAIL then
	    if word = [] then return([]);
	    elif testtype(word, Type::ListOf(Type::NonNegInt)) then
		alpha := [$1..max(op(word))];
	    else
		error("For non numeric word you must provide an alphabet: ".
		      expr2text(word));
	    end_if;
	elif domtype(alpha) = DOM_INT then
	    alpha := [$1..args(2)];
	elif domtype(alpha) = DOM_SET then
	    alpha := [op(alpha)];
	end_if;
	tmp := dom::evaluationTable(word);
	map(alpha, (i -> if contains(tmp,i) then tmp[i] else 0 end_if));
    end_proc;

    /**
	evaluationSparse(word) - evaluation of word, as a "sparse" composition

        This is the list of number of occurences of each letters in
	word. Either the word is a word on integers or you must
	provide the alphabet as a list.

	Example:
		combinat::words::evaluationPartition([b,a,d,b,c,d,b]);
			[3,2,1,1]
    **/
    evaluationSparse :=
    proc(word: dom) : Type::ListOf(Type::ListOf(Type::PosInt))
        local eq;
    begin
	prog::sort([[op(eq)] $ eq in dom::evaluationTable(word)],
		   op, 1);
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // Standardization
    //////////////////////////////////////////////////////////////////////////

    alphabetToOrder := 
    proc(A: dom::typeAlphabet) : dom::typeOrder
        option escape;
        local rank, i;
    begin
        if testtype(A, Dom::Integer) then
            _less;
        else
            (rank(op(A,i)):=i)$i=1..nops(A);
            (x,y) -> (rank(x) < rank(y));
        end_if;
    end_proc;

    standard :=
    proc(word         : dom, 
         order = FAIL : Type::Union(dom::typeOrder, dom::typeAlphabet, DOM_FAIL)
        ) : dom::typeStandard
	local t, k, keys, tmp, offset;
    begin
        if order = FAIL
            then order := _less;
        elif testtype(order, dom::typeAlphabet)
            then order := dom::alphabetToOrder(order);
        end_if;

        t:=dom::evaluationTable(word);
        
        keys := sort(map([op(t)], op, 1), order);
        offset:=0;
        tmp:=0;
        for k in keys do
            tmp   := t[k];
            t[k]  := offset;
            offset := offset+tmp;
        end_for;
        
        map(word, l->(t[l]:=t[l]+1));
    end_proc;

    // In some sense, the standard describes the shape of the
    // word, while the evaluation describes its content. So, the
    // name fromStandardAndEvaluation and argument order is
    // coherent with the construction of matrices in MuPAD where you
    // first describe the size of the matrix, and then its content,
    // and as well as with the various fromShapeAndWord
    
    fromStandardAndEvaluation :=
    proc(std : dom::typeStandard,
	 ev  : Type::ListOf(Type::NonNegInt), 
	 alpha = FAIL : Type::Union(dom::typeAlphabet, DOM_FAIL) 
	 ) : dom
	local partSum, nextPartSum, i, k, sub;
    begin
	if _plus(op(ev)) <> nops(std) then
	    error("Incompatible evaluation and standard: ".expr2text(ev)
		  .", ".expr2text(std));
	end_if;
	if alpha = FAIL then 
	    alpha := [$1..nops(ev)];
	elif testtype(alpha, Type::NonNegInt) then
	    alpha := [$1..alpha];
	end_if;
	partSum := 0;
	sub := [];
	for i from 1 to nops(ev) do
	    nextPartSum := partSum + ev[i];
	    sub := append(sub, k=alpha[i] $ k=partSum+1..nextPartSum);
	    partSum := nextPartSum;
	end_for;
	subs(std, sub);
    end_proc;
  
  charge := 
    proc(w : dom)
	// the evaluation of w should be a partition
	local charge, i, l, index;
    begin
	charge := 0; 
	while nops(w) <> 0 do
	    i := 1; l := 1; index := 0;
	    while nops(w) <> 0 and l <= max(op(w)) do
		while w[i] <> l do 
		    i := i + 1;
		    if i > nops(w) then i := 1; index := index+1; end_if;
		end_while;
		charge := charge + index;
		l := l + 1;
		w := subsop(w, i=null());
		if i > nops(w) then i := 1; index := index+1; end_if;
	    end_while;
	end_while;
	charge;
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // Descents
    //////////////////////////////////////////////////////////////////////////

/**
    descents(p)
    Returns the descent List of the permutation p
 
    That is the list of the i for which p[i]>p[i+1]

    the option FinalDescent adds a descent at the end
**/
    descents :=
    proc(p: dom):  Type::ListOf(Type::NonNegInt)
        local result, i;
    begin
        result:=[];
        for i from 1 to nops(p)-1 do
            if p[i]>p[i+1] then
                result:=append(result,i);
            end_if
	end_for;
	if args(0) > 1 and args(2) = FinalDescent and p <> [] then
	    // the empty permutation has no final descent!!!???
	    result := append(result, nops(p));
	end_if;
	result;
    end_proc;
    descentsComposition := combinat::compositions::fromDescents @
		       dom::descents @ ((p) -> (p, FinalDescent));

    descentsNumber :=
    proc(p: dom):  Type::ListOf(Type::NonNegInt)
    begin
	nops(dom::descents(args())):
    end_proc;

    /**
     * Swap entries at positions i and i+1 in w
     **/
    swap :=
    proc(w: dom, i: Type::PosInt)
	local tmp;
    begin
	if testargs() and i>nops(w) then
	    error("index out of range");
	end_if;
	tmp := w[i];
	w[i] := w[i+1];
	w[i+1] := tmp;
	w;
    end_proc;

    /**
     * exchange entries at positions i and i+1 in w if w[i] > w[i+1]
     **/
    swapIncrease :=
    proc(w: dom, i: Type::PosInt)
	local tmp;
    begin
	if testargs() and i>nops(w) then
	    error("index out of range");
	end_if;
	if w[i] > w[i+1] then
	    tmp := w[i];
	    w[i] := w[i+1];
	    w[i+1] := tmp;
	end_if;
	w;
    end_proc;

    /**
     * exchange entries at positions i and i+1 in w if w[i] < w[i+1]
     **/
    swapDecrease :=
    proc(w: dom, i: Type::PosInt)
	local tmp;
    begin
	if testargs() and i>nops(w) then
	    error("index out of range");
	end_if;
	if w[i] < w[i+1] then
	    tmp := w[i];
	    w[i] := w[i+1];
	    w[i+1] := tmp;
	end_if;
	w;
    end_proc;

    /**
     * Lexicographic comparison of words
     * (currently restricted to words whose letters can be compared by <)
     **/
    lexLess :=
    proc(x1: dom, x2: dom) : DOM_BOOL
        local i;
    begin
        i := 1;
        while (i<=nops(x1)) and (i<=nops(x2)) and (x1[i]=x2[i]) do
            i := i+1;
        end_while;
        if (i>nops(x2)) then return(FALSE) end_if;
        if (i>nops(x1)) then return(TRUE) end_if;
        bool(x1[i]<x2[i])
    end_proc;

    /**
     * Degree lexicographic comparison of words
     * (currently restricted to words whose letters can be compared by <)
     **/
    degLexLess :=
    proc(x1: dom, x2: dom) : DOM_BOOL
        local d1, d2;
    begin
        d1 := _plus(op(x1)); d2 := _plus(op(x2));
	if d1<>d2 then
	    return(bool(d1<d2));
	end_if;
	dom::lexLess(x1,x2);
    end_proc;

    /**
     * Inverse lexicographic comparison of words
     * (currently restricted to words whose letters can be compared by <)
     **/
    invLexLess :=
    proc(x1: dom, x2: dom) : DOM_BOOL
        local i;
    begin
	if nops(x1)<>nops(x2) then
	    return(bool(nops(x1)<nops(x2)));
	end_if;
	for i from nops(x1) downto 1 do
	    if x1[i]<>x2[i] then
		return(bool(x1[i]<x2[i])); 
	    end_if;
	end_for;
	return(FALSE);
    end_proc;

    /**
     * Standard degree inverse lexicographic comparison of words
     * (currently restricted to words whose letters can be compared by <)
     **/
    degInvLexLess :=
    proc(x1: dom, x2: dom) : DOM_BOOL
        local d1, d2;
    begin
        d1 := _plus(op(x1));
	d2 := _plus(op(x2));
	if d1<>d2 then
	    return(bool(d1<d2));
	end_if;
	dom::invLexLess(x1,x2);
    end_proc;

    /**
     * Reverse lexicographic comparison of words
     * (currently restricted to words whose letters can be compared by <)
     **/
    revLexLess :=
    proc(x1: dom, x2: dom) : DOM_BOOL
        local i;
    begin
	if nops(x1)<>nops(x2) then
	    return(bool(nops(x1)>nops(x2)));
	end_if;
	for i from nops(x1) downto 1 do
	    if x1[i]<>x2[i] then
		return(bool(x1[i]>x2[i]));
	    end_if;
	end_for;
	return(FALSE);
    end_proc;

    /**
     * Standard degree reverse lexicographic comparison of words
     * (currently restricted to words whose letters can be compared by <)
     **/
    degRevLexLess :=
    proc(x1: dom, x2: dom) : DOM_BOOL
        local d1, d2;
    begin
        d1 := _plus(op(x1));
	d2 := _plus(op(x2));
	if d1<>d2 then
	    return(bool(d1<d2));
	end_if;
	dom::revLexLess(x1,x2);
    end_proc;

    shuffleAugmented :=
    proc(word1: dom, word2: dom)
    begin
	if nops(word1)=0 then
	    [word2];
	elif nops(word2)=0 then
	    [word1];
	else
	map(dom::shuffleAugmented([op(word1, 1..nops(word1)-1)], word2),
	    append, word1[nops(word1)]) .
	    map(dom::shuffleAugmented(word1, [op(word2, 1..nops(word2)-1)]),
		append, word2[nops(word2)]) .
	    map(dom::shuffleAugmented([op(word1, 1..nops(word1)-1)],
				      [op(word2, 1..nops(word2)-1)]),
		append, word1[nops(word1)]+word2[nops(word2)]);
	end_if;
    end_proc;

  unpairedLetters := 
    proc(w : words, i : Type::PosInt)
	local height, unpairedI, unpairedI1, k;
    begin
	height := 0;
	unpairedI := [];
	for k from 1 to nops(w) do
	    if w[k] = i then
	    if height > 0 then
		height := height - 1;
	    else
		unpairedI := append(unpairedI, k);
	    end_if;
	    elif w[k] = i+1 then
	    height := height + 1;
	    end_if;
	end_for;
	height := 0;
	unpairedI1 := [];
	for k from nops(w) downto 1 do
	if w[k] = i+1 then
	    if height > 0 then
		height := height - 1;
	    else
		unpairedI1 := append(unpairedI1, k);
	    end_if;
	elif w[k] = i then
	    height := height + 1;
	end_if;
	end_for;
	[unpairedI, sort(unpairedI1)];
    end_proc;

    /// Pretty print a word of integers as a function plot
    printFunctionPlot :=
    proc(w: Type::ListOf(Type::NonNegInt)) : output::asciiArt
	local result, x, l;
    begin
	result := output::asciiArt();
	x := 1;
	for l in w do
	    result[x,l+1] := expr2text(l);
	    x := x + length(expr2text(l));
	end_for;
	result;
    end_proc;

end_domain:

// Rename to combinat::words::fromStandard ?


domain combinat::words::fromStandard
    inherits Dom::BaseDomain;
    category Cat::CombinatorialClass;
    axiom    Ax::systemRep;
    info      := "Library 'combinat::words::fromStandard': utilities for the words having a given standard permutation";
    typeWords := combinat::words;
    generator :=
    proc(std   : combinat::words::typeStandard,
       alpha : combinat::words::typeAlphabet
      )
      option escape;
      local nalpha, recoil;
    begin
      if testtype(alpha, Type::NonNegInt) then
          nalpha := alpha;
      else
          nalpha := nops(alpha);
      end_if;
      recoil := combinat::compositions::fromDescents(
                    (combinat::permutations::descents(
                           combinat::permutations::inverse(std), FinalDescent)));
      combinat::generators::map(
           combinat::compositions::finerVectors::generator(recoil, nalpha),
             proc(ev)
           begin
               combinat::words::fromStandardAndEvaluation(std, ev, alpha);
           end_proc);
    end_proc;
begin
end_domain:

// shuffling and spliting of words. 
//*********************************
combinat::words::shuffle   := 
    loadproc(combinat::words::shuffle,   pathname("COMBINAT"), "shuffle"):
combinat::words::shufflings := 
    loadproc(combinat::words::shufflings, pathname("COMBINAT"), "shuffle"):
combinat::words::split := 
    loadproc(combinat::words::split,     pathname("COMBINAT"), "shuffle"):
combinat::words::splittings := 
    loadproc(combinat::words::splittings, pathname("COMBINAT"), "shuffle"):
