/*++
Method of characteristics for nonlinear equations of first order

Three methods: detools::nonlinearCharODESystem for finding the system of odes which
                                   will determine the characteristics 
	       detools::nonlinearCharacteristics for computing characteristics
	       detools::nonlinearCharSolve for solving PDEs	       

detools::nonlinearCharODESystem(nldf,var,<initcond,pars>)
    where nldf is a nonlinear differential function
    (nonlinearCharSystem will consider nldf to be equivalent to the
    equation nldf=0), var is a variable to be used as the independent
    variable in the characteristics, initcond is a set of optional
    initial conditions (an initial curve for the characteristics), for
    example, if x,t are the independent variables, u the dependent
    variable, an initial condition could be {x=T^2,t=2*T,u=T-1}.

detools::nonlinearCharacteristics(nldf,var,<initcond,pars>)
    where nldf, var and initcond are as above.

detools::nonlinearCharSolve(nldf,init,par)
    where nldf is as above, init is a set parameterizing the initial
    "curve" (an n-1 dimensional surface if there are n independent
    variables) and par is the set of parameters used in init (so there
    should be n-1 of them).  This method will try to return the
    solution of nldf = 0 passing through the initial curve.  It will
    call nonlinearCharacteristics(nldf,init).  If
    nonlinearCharacteristics does not return a solution, this method
    will return the original information.  If nonlinearCharacteristics
    does return a solution, this method will attempt to eliminate the
    parameters and return the function.  If this is not possible, this
    method will return the characteristics.


Author: Jay Belanger (belanger@truman.edu)
	(adapted to detools library by Werner M. Seiler)
Created: 	Thu Apr 29 19:50:48 1999
Last update: 	Fri Apr 30 11:16:15 1999
++*/

alias(CANNOT_SOLVE_ODESYS=return(hold(detools::nonlinearCharODESystem)(args())),
      CANNOT_SOLVE_CHAR=return(hold(detools::nonlinearCharacteristics)(args())),
      CANNOT_SOLVE_CHARSOLVE=return(hold(detools::nonlinearCharSolve)(args())),
      DAT(x)=detools::data[x],
      CDAT(x)=contains(detools::data,x)):

detools::nonlinearCharODESystem :=
    proc()
    local sys:DOM_SET, fn:DOM_SET, u:DOM_EXPR, i:DOM_INT, j:DOM_INT,
	  DVI:DOM_LIST, dv, vars_to_functions:DOM_PROC, 
	  nldf, var:DOM_IDENT, p, nn:DOM_INT, DV:DOM_DOMAIN, DF:DOM_DOMAIN,
          ndv, k:DOM_INT, rh, du, ueqn, peqn, xeqn, peqns, psoln, prhs,
          pars;
    begin
	// Analysis of input
        if args(0) > 0 then
            detools::getdata(testargs(),args());
        end_if;
//	if DAT("Linear") then error("linear equation") end_if;
        DV := (DAT("DVin"));
        DF := (DAT("DFin"));
	nldf := DAT("Eqs")[1];
	if testargs() then
            if DAT("System") then 
                userinfo(3, "nonlinearCharODESystem cannot handle systems");
                CANNOT_SOLVE_ODESYS //error("cannot handle systems") 
            end_if;
	    if args(0)<2 or args(0)>4 then 
                userinfo(3, "nonlinearCharODESystem given the wrong number of arguments");
		CANNOT_SOLVE_ODESYS //error("wrong number of args") 
	    end_if;
	    if DF::order(nldf)<>1 then
                userinfo(3, "nonlinearCharODESystem can only handle first order expressions");
		CANNOT_SOLVE_ODESYS 
                    //error("Differential equation must be first order");
	    end_if;
	end_if;
	if DAT("Parms")=[] then error("no variable name given") end_if;
	var := DAT("Parms")[1];
        userinfo(2, "nonlinearCharODESystem called with expression " .
            expr2text(nldf) . " and variables " .expr2text(var));
	// Now, we need to make sure that there is only one dependent 
	// variable.  "dv" will be a the set of all dependent
	// variables in the equation.
	dv := DV::depVariables();
	// If "dv" has more than one element, we need to signal an error,
	// otherwise make "u" the dependent variable
	if nops(dv) <> 1 then
            userinfo(3, "nonlinearCharODESystem must be given exactly one dependent variable");
	    CANNOT_SOLVE_ODESYS 
                //error("There must be exactly one dependent variable");
	end_if;	
	dv := op(dv);
        // The independent variables
	DVI := DV::indVariables;
        DVI := map(DVI, expr);
        nn := DV::numIndVar;
        // The dependent variable
	u := DV::expr(dv);
        // The variables for the derivatives
        for i from 1 to nn do
            p[i] := genident("p");
        end_for;
	// "vars_to_functions" will turn independent variables (e.g. "x")
	// into functions (e.g. "x(S)").
	vars_to_functions:= 
	    proc (ex)
	    local k:DV;
	    begin
                // replace any u([var]) by p[i]
                for i from 1 to nn do
                 ex := subs(ex, 
                  DV(diff(expr(op(DV::depVariables()))(DVI[i]),DVI[i])) = p[i](var));
                end_for;
		ex := expr(ex);
                ex := subs(ex,u=u(var));
		for k in DVI do
		    ex := subs(ex,k=k(var));
		end_for;
	    end_proc;
	// "sys" will be the set of odes
        sys := {};
        // First, the x'(s) = ... equations and the u'(s) = ...
        du := 0;
        for k from 1 to nn do
            ndv := DV(expr(DV::depVariables()[1])([DVI[k]]));
            rh := diff(nldf,ndv);
            rh := vars_to_functions(rh);
            sys := sys union {diff(DVI[k](var),var) = rh};
            du := du + p[k](var)*rh;
        end_for;
        sys := sys union {diff(u(var),var) = du};
        // Finally, the p'(s) = ... equations
        for k from 1 to nn do
            sys := sys union
             {diff(p[k](var),var) = 
                -1*vars_to_functions(diff(nldf,DV(u)))*p[k](var) -
                vars_to_functions(diff(nldf,DV(DVI[k])))};
        end_for;
        // Finish with the ODES
	// Next, check for optional initial conditions.  If given,
	// they are contained in detools::data["Parms"] (we assume
	// that no other parameters are given).
	if nops(DAT("Parms"))>1 then
            pars := [op(DAT("Parms")[3])];
	    for j in DAT("Parms")[2] do
		// Add the intial conditions to sys, replacing
		// an equation like x=T by x(0)=T
                if op(j,1) = u then 
                    ueqn:=op(j,2); 
                end_if;
                for k from 1 to nn do
                    if op(j,1) = DVI[k] then 
                        xeqn[k]:=op(j,2);
                    end_if;
                end_for;
		sys := sys union {op(j,1)(0)=op(j,2)};
	    end_for;
            // Next, find the p(0) = ... initial conditions.
            for k from 1 to (nn - 1) do
                prhs := 0;
                for i from 1 to nn do
                    prhs:=prhs + p[i]*diff(xeqn[i],pars[k]);
                end_for;
                peqn[k]:= diff(ueqn,pars[k]) = prhs;
            end_for;
            for i from 1 to nn do
                nldf := subs(nldf,
                DV(diff(expr(op(DV::depVariables()))(DVI[i]),DVI[i])) = p[i]);
            end_for;
            nldf := expr(nldf);
            // Solve the given equations for the ps
            peqns := {nldf = 0, peqn[i]$i=1..(nn-1)};
            psoln := solve(peqns,{p[i]$i=1..nn});
            if psoln = FAIL then
                error("Cannot solve for initial conditions for p");
            else
                if psoln = {} then
                    error("Cannot solve for initial conditions for p");
                else
                    // In the case of more than one solution, take the first one.
                    // This should be fixed.
                    psoln:=psoln[1];
                    for k in psoln do
                        prhs := op(k,2);
                        prhs := subs(prhs, u=ueqn);
                        for i from 1 to nn do
                            prhs := subs(prhs, DVI[k]=xeqn[k]);
                        end_for;
                        sys := sys union {op(k,1)(0)=prhs};
                    end_for;
                end_if;
            end_if;
	end_if;
	// Now, either return the solved system of odes or the unsolved
	// system of odes.
        fn := {u(var)};
        for k from 1 to nn do
            fn := fn union {DVI[k](var), p[k](var)}
        end_for;
	return(ode(sys,fn));
    end_proc:



detools::nonlinearCharacteristics :=
    proc()
    local sys:DOM_SET, soln:DOM_SET, nldf, dv, DVI, k, var, DV:DOM_DOMAIN,
          i, nn;
    begin
        // Analysis of input
        if args(0) > 0 then
            detools::getdata(testargs(),args());
        end_if;
	nldf := DAT("Eqs")[1];
        DV := (DAT("DVin"));
 	if testargs() then
 	    if args(0)<2 or args(0)>4 then 
                userinfo(3, "nonlinearCharacteristics given the wrong number of arguments");
                CANNOT_SOLVE_CHAR 
            end_if; //error("wrong no of args") end_if;
 	    if (DAT("DFin"))::order(nldf)<>1 then
                userinfo(3, "nonlinearCharacteristics must be given a first order expression");
 		CANNOT_SOLVE_CHAR 
                // error("Differential equation must be first order");
 	    end_if;
 	end_if;
 	if DAT("Parms")=[] then 
            userinfo(3, "nonlinearCharacteristics must be given a variable name");
            CANNOT_SOLVE_CHAR // error("no variable name given") 
        end_if;
	var := DAT("Parms")[1];
        userinfo(2, "nonlinearCharacteristics called with expression " .
            expr2text(nldf) . " and variables " .expr2text(var));
	dv := DV::depVariables();
	// If "dv" has more than one element, we need to signal an error,
	// otherwise make "u" the dependent variable
	if nops(dv) <> 1 then
	    error("There must be exactly one dependent variable");
	end_if;	
	dv := op(dv);
	DVI := DV::indVariables;
        DVI := map(DVI, expr);
        nn := DV::numIndVar;
       	sys := detools::nonlinearCharODESystem();
        sys := solve(sys);
        sys := sys[1];
        // Now extract the relevant equations.
        soln := {};
        dv := expr(dv);
        for k in sys do
            for i from 1 to nn do
                if (op(k,1) = DVI[i](var) or op(k,1) = dv(var))then
                    soln := soln union {k};
                end_if;
            end_for;
        end_for;
    return(soln);
    end_proc:


detools::nonlinearCharSolve :=
    proc()
    local soln, fn, var:DOM_IDENT, char, k, dv, functions_to_vars:DOM_PROC,
	  init:DOM_SET, par:DOM_SET, nldf, DVI:DOM_LIST,
          getsol:DOM_PROC, depars, notDV:DOM_PROC, hasPars:DOM_PROC,
          psoln, DV:DOM_DOMAIN,
          DF:DOM_DOMAIN;
    begin
	// Analysis of input
        if args(0)>0 then
            detools::getdata(testargs(),args());
        end_if;
	if DAT("System") then 
            userinfo(3, "nonlinearCharSolve cannot handle systems");
            CANNOT_SOLVE_CHARSOLVE
            // error("cannot handle systems") 
        end_if;
        DV := (DAT("DVin"));
        DF := (DAT("DFin"));
	nldf := DAT("Eqs")[1];
	if testargs() then
	    if not (nops(DAT("Parms"))<>3) then 
                userinfo(3, "nonlinearCharSolve was given the wrong number of arguments");
		CANNOT_SOLVE_CHARSOLVE //error("wrong number of arguments");
	    end_if;
	    if DF::order(nldf)<>1 then
                userinfo(3, "nonlinearCharSolve must be given a first order expression");
		CANNOT_SOLVE_CHARSOLVE 
                // error("Differential equation must be first order");
	    end_if;
            if nops(DAT("Parms")) <> 2 then
                userinfo(3, "nonlinearCharSolve must be given initial conditions");
                CANNOT_SOLVE_CHARSOLVE //error("Initial conditions must be given");
            end_if;
	end_if;
	dv := DV::depVariables();
	// If "dv" has more than one element, we need to signal an error,
	// otherwise make "u" the dependent variable
	if nops(dv) <> 1 then
	    error("There must be exactly one dependent variable");
	end_if;	
	dv := op(dv);
        dv := expr(dv);
	DVI := DV::indVariables;
        DVI := map(DVI, expr);
        init := DAT("Parms")[1];
        par := {op(DAT("Parms")[2])};
        userinfo(2, "nonlinearCharSolve called with expression " .
            expr2text(nldf) . " initial conditions " . expr2text(init)
            . " and parameters " .expr2text(par));
	// This function should replace expressions of the form "x(S) = ..."
	// by "x = ..."
	functions_to_vars :=
	    proc(l)
	        begin op(l,[1,0]) = op(l,2) 
	    end_proc;
	// First of all, we need a new variable for "characteristics"
	var := genident();
	// Now we need to try to get the characteristics.
	// If the system of odes cannot be solved by MuPAD, we'll return
	// the original information
//        sysassign(detools::data["Parms"],[var,init]);
        char := detools::nonlinearCharacteristics(nldf,var,init,par);
//	char := detools::nonlinearCharacteristics();
	if type(char)="solve" then return(nldf,init,par) end_if;
	// For the time being, we'll assume that all "solve"s return only
	// one solution set (or rather, we'll use the first solution set).
	// If necessary, this should be easy enough to change.
//	soln := op(char,1);
	// After "characteristics" goes to the trouble of changing the
	// independent variables into functions, we need to change them
	// back into variables.
	soln := map(char,functions_to_vars);
	// Next, seperate the actual function that we are trying to solve
	// for from the other variables.
	// We need to find out which dependent variable appears in the
	// equation.  ("characteristics" already does this, but I'm
	// not sure that we want another argument passed.)
	soln := split(soln,has,dv);
	fn := op(soln[1],1);
        fn := rhs(fn);
	soln := soln[2];
	// Next, we'll solve for var and parameters in terms of the 
	// independent variables (if this cannot be done, we'll return the 
	// characteristics)
	soln := solve(soln, par union {var});
	// If there are no solutions, then the characteristics lie in the
	// initial "curve" (or somesuch)  I'm not sure how this should be
	// handled, for the time being we'll just return the latest result
	if domtype(soln)=DOM_FAIL then return(char) end_if;
	if type(soln)="solve" then return(char) end_if;
        getsol :=
            proc(sn, f)
            local k;
            begin
                for k in sn do
                    f := subs(f, k);
                end_for;
                return(f);
            end_proc; /* getsol */
        if type(soln)=piecewise then 
            depars := DF::indets(nldf) union indets(init);
            notDV :=
                proc(v)
                begin
                    if domtype(v)=DV or has({op(DVI)} union {dv}, v) 
                        then
                        return(FALSE);
                    else
                        return(TRUE);
                    end_if;
                end_proc; /* notDV */
            depars := select(depars, notDV);
            hasPars :=
                proc(v)
                begin
                    for k in depars do
                    if has(v,k) then
                        return(TRUE);
                    end_if;
                    end_for;
                    return(FALSE);
                end_proc; /* hasPars */
            psoln := piecewise::selectConditions(soln,hasPars);
            if psoln=undefined then
                soln := piecewise::disregardPoints(soln);
            else
                soln := psoln;
            end_if;
        end_if;
        if type(soln)=piecewise then
            fn:=piecewise::extmap(soln, getsol, fn);
        else
            fn := getsol(soln, fn);
        end_if;
	// Finally, in the expression for the function we are solving for,
	// replace var and the parameters with the independent variables.
	// tadaa
	return({fn});
    end_proc:


unalias(DAT,CDAT):



