/*++
A method for solving second order constant coefficient linear partial 
differential equations in two variables by using Riemann's function.

detools::riemannSolve(pde,fn)
     where pde is the partial differential equation (required to 
     be linear, second order, constant coefficent, and with two
     variables), fn is the function u(x,y)

Author: Jay Belanger (belanger@truman.edu)
++*/

alias(CANNOT_SOLVE=return(hold(detools::riemannSolve)(args())),
      DAT(x)=detools::data[x]);
      
detools::riemannSolve :=
    proc()
    local // The local variables from the getdata part
          DF:DOM_DOMAIN, DV:DOM_DOMAIN, rde, DVI:DOM_LIST, 
          DVD:DOM_LIST, x, y, u, a, b, c, d, e, f, g,
          // The local variables from the convertindvars part
          disc, root1, root2, x1, y1, a0, b1, c0, c1, d1, e0, e1,
          // The local variable sfrom the convertdepvar part
          lambda, ff,
          // The local variables from the main part
          X, Y, fff, A, AX, AY, AXY, C, F, G, v,
          // The local variables from the convertback part
          U;
    begin
        // The getdata part
        // This part should take the input, a linear second order 
        // pde with two variables,
        //  a u_{xx} + b u_{xy} + c u_{yy} + d u_x + e u_y + f u = g
        // the function in question
        // u(x,y)
        // It will result in the data
        // x, y, u
        // a, b, c, d, e, f, g
        if args(0) > 0 then
            detools::getdata(testargs(),args());
        end_if;
        if testargs() then
            if DAT("System") then 
                userinfo(3, "riemannSolve cannot handle systems");
                CANNOT_SOLVE; //error("Cannot handle systems") 
            end_if;
            if not DAT("Linear") then 
                userinfo(3, "riemannSolve cannot handle non-linear equations");
                CANNOT_SOLVE; //error("Cannot handle non-linear equations") 
            end_if;
            if not DAT("ConstCoeff") then 
                userinfo(3, "riemannSolve cannot handle equations with non-constant coefficients");
                CANNOT_SOLVE;
                  //error ("Cannot handle equations without constant coefficients")
            end_if;
        end_if;
        DF := DAT("DFin");
        rde := DAT("Eqs")[1];
        userinfo(2, "riemannSolve called with expression " . expr2text(rde));
        DV := DAT("DVin");
        DVI := DV::indVariables;
        DVD := DV::depVariables;
        if testargs() then
            if DF::order(rde)<>2 then
                userinfo(3, "riemannSolve can only solve second order expressions");
                CANNOT_SOLVE;
                    // error("Differential equation must be second order");
            end_if;
            if nops(DVI) <> 2 then
                userinfo(3,"riemannSolve can only solve expressions with two variables");
                CANNOT_SOLVE;
                    //error("Differential equation must have two variables");
            end_if;
        end_if;
        // Get the variables
        x  := expr(DVI[1]);
        y  := expr(DVI[2]);
        u  := expr(DVD[1]);
        // Now, get coefficients
        a := expr(DF::coeff(rde,DV::newRepeated("Deriv",1,[1,1])));
        b := expr(DF::coeff(rde,DV::newRepeated("Deriv",1,[1,2])));
        c := expr(DF::coeff(rde,DV::newRepeated("Deriv",1,[2,2])));
        d := expr(DF::coeff(rde,DV::newRepeated("Deriv",1,[1])));
        e := expr(DF::coeff(rde,DV::newRepeated("Deriv",1,[2])));
        f := expr(DF::coeff(rde,DV::newRepeated("Deriv",1)));
        g := -1*expr(op(rde,2));
        // The convertindvars part
        // At this point, we have 
        //  x, y, u,
        //  a, b, c, d, e, f, g
        // This part will convert the differential equation to the form
        // u_{x1,y1} + b1 u_x1 + c1 u_y1 + d1 u = e1
        // This will result in
        //    x1, y1
        //    b1, c1, d1, e1
        disc := b^2 - 4*a*c;
        userinfo(2, "The discriminant of the expression is " . expr2text(disc));
        if is(disc, Type::Real) = TRUE then
           if disc <= 0 then
              userinfo(3, "riemannSolve was given a non-hyperbolic equations");
              CANNOT_SOLVE;
                //error("This is not a hyperbolic equation");
           end_if;
        end_if;
        x1 := genident("x");
        y1 := genident("y");
        if a <> 0 then
            disc := sqrt(disc);
            root1 := (-1*b + disc)/(2*a);
            root2 := (-1*b - disc)/(2*a);
            // x1 := y + root1*x;
            // y1 := y + root2*x;
            a0 := 2*c - b^2/a;
            b1 := (d*root1 + e)/a0;
            c1 := (d*root2 + e)/a0;
            d1 := f/a0;
            e0 := subs(g, x = (x1-y1)/(root1-root2), 
                       y = (root2*x1-root1*y1)/(root2-root1));
            e1 := e0/a0;
        elif c <> 0 then
            // x1 := x;
            // y1 := x -b*y/c;
            a0 := -1* b^2/c;
            b1 := d/a0;
            c1 := (d-e*b/c)/c0;
            d1 := f/c0;
            e0 := subs(g, x=x1, 
                       y = c*(x1-y1)/b);
            e1 := e0/a0;
        else
            b1 := d/b;
            c1 := e/b;
            d1 := f/b;
            e1 := g/b;
        end_if;    
        // The convertdepvar part
        // This part will convert the differential equation from
        //   u_{x1,y1} + b1 u_x1 + c1 u_y1 + d1 u = e1
        // to
        //  v_{x1,y1} + lambda v = ff
        // It starts with
        //    x1, y1, u, b1, c1, d1, e1
        // and end up with
        //    x1, y1, lambda, f
        lambda := (d1-b1*c1);
        ff := e1*exp(b1*y1 + c1*x1);
        // The main part
        // At this point, we should have a differential equation
        // v_x1y1 + lambda*v = f(x1,y1)
        X := genident("X");
        Y := genident("Y");
        fff := subs(ff,x1=X,y1=Y);
        // Next, create the Riemann function and its derivatives
        A   := besselJ(0,2*sqrt(lambda*y1*x1));
        AX  := besselJ(0,2*sqrt(lambda*y1*(x1-X)));
        AY  := besselJ(0,2*sqrt(lambda*x1*(y1-Y)));
        AXY := besselJ(0,2*sqrt(lambda*(x1-X)*(y1-Y)));
        // The arbitrary functions and constant
        C := genident("C");
        F := genident("F");
        G := genident("G");
        // Finally, the solution
        if lambda=0 then
            if fff = 0 then
                v := F(y1) + G(x1);
            else
                v := F(y1) + G(x1) +
                   int(int(fff*AXY,Y=0..y1),X=0..x1);
           end_if;
        elif fff = 0 then
            v := A*C +
               int(F(Y)*AY,Y=0..y1) +
               int(G(X)*AX,X=0..x1);
        else
            v := A*C +
               int(F(Y)*AY,Y=0..y1) +
               int(G(X)*AX,X=0..x1) +
               int(int(fff*AXY,Y=0..y1),X=0..x1);
        end_if;
        // The convertback part
        // At this point, we have the function 
        // v=v(x1,y1)
        // where v=u*exp(b1*y1 + c1*x1)
        // and x1 = y + root1*x
        //     y1 = y + root2*x
        // We want the function u=u(x,y)
        if a <> 0 then
            U := subs(v,x1=y+root1*x,y1=y+root2*x);
            U := U*exp(-1*b1*(y+root2*x) - c1*(y+root1*x));
        elif c <> 0 then
            U := subs(v,x1=x,y1=x-b*y/c);
            U := U*exp(-1*b1*(x-b*y/c) - c1*x);
        else
            U := subs(v,x1=x,y1=y);
            U := U*exp(-1*b1*y - c1*x);
        end_if;
        return({U});
    end_proc; /* riemannSolve */
