/*****************************************************************************
  $Id: IntegerListsLexClass.mu,v 1.10 2004/07/12 19:05:10 nthiery Exp $
  File:		lib/COMBINAT/IntegerListsLexClass.mu
  Authors:	Florent Hivert <Florent.Hivert@univ-mlv.fr
		Nicolas M. Thiry <nthiery@users.sourceforge.net>
  Created:	16/06/2001
  License:	LGPL
*****************************************************************************/

// Optimization for MuPAD > 2.0.0: we use RD_INF which is faster than infinity
%if traperror(RD_INF+RD_INF) = 0 then
    alias(infinity  = RD_INF);
end_if:

alias(Tools = combinat::integerListsLexTools):

category Cat::IntegerListsLexClass
    category Cat::CombinatorialClass, Cat::OrderedSet;
    axiom    Ax::systemRep;

    info := "Category 'Cat::IntegerListsLexClass': wrapper around combinat::integerListsLexTools";

    interfaceAdd := { hold(first), hold(Next), hold(_less) };

    /// Returns constraints for integerListsLexTools from user-friendly options
    parseOptions := toBeDefined;

    /**	Helper for constructing the option parser dom::parseOptions

        This generic parser makes it easy for a domain of the category
        combinat::IntegerListsLexClass to build his own option parser.

	Example:

        domain ivectors
            inherits Dom::BaseDomain;
	    category Cat::IntegerListsLexClass;

	    parseOptions :=
	    proc(k)
	    begin
		dom::parseOptionsIntegerListsLexClass
		   (table(
                          Length	= k,
                          MinPart	= 0,
               		  MaxPart	= infinity,
               		  Inner	= NIL,
               		  Outer	= NIL,
	       		  MinSlope= -infinity,
	       		  MaxSlope= infinity
			 ),
		    args());
	    end_proc;
	end_domain;
	ivectors::list(5,3, MinLength=3, Outer=[1,3,4], MaxSlope=4);

        It's up to the user (or the domain!) to ensure that Inner and
        Outer are compatible with MinPart, MaxPart, MinSlope and MaxSlope.
    **/

    // for backward compatibility
    parseOptions_IntegerListLexBuilding :=
    dom::parseOptionsIntegerListsLexClass;

    parseOptionsIntegerListsLexClass :=
    proc(defaultOptions: Type::OptionsTable)
        : Type::SequenceOf(Type::AnyType,6,6)
         // sequence of 6 constraints
        local options, inner, outer, minpart;
        option escape;
    begin
        options := combinat::getOptions(defaultOptions,
					dom::parseOptions_OptionTypes,
					1, Error);
        if options[Length] <> NIL then
            options[MinLength] := options[Length];
            options[MaxLength] := options[Length];
        end_if;
        minpart := options[MinPart];
        if options[Inner] <> NIL then
            if domtype(options[Inner]) = DOM_LIST then
                inner := options[Inner];	// slight optimization ?
                options[Inner] :=
                    i->if i>nops(inner) then minpart else inner[i]; end_if;
                options[MinLength] := max(options[MinLength],
                                                nops(inner));
            end_if;
        else
	    // Note: a constant can be considered as a function.
            options[Inner] := minpart;
        end_if;
        if options[Outer] <> NIL then
            if domtype(options[Outer]) = DOM_LIST then
                outer := options[Outer];	// slight optimization ?
                options[Outer] := i-> outer[i];
                options[MaxLength] := min(options[MaxLength],
                                                nops(outer));
            end_if;
        else
	    // Note: a constant can be considered as a function.
            options[Outer] := options[MaxPart];
        end_if;
        
        options[MinLength], options[MaxLength],
        options[Inner], options[Outer],
        options[MinSlope], options[MaxSlope];
    end_proc;

    parseOptions_OptionTypes :=
    table(
          MinLength = Type::NonNegInt,
          MaxLength = Tools::typeIntBound,
	  Length    = Type::Union(Type::NonNegInt, DOM_NIL),
          MinPart   = Type::NonNegInt,
          MaxPart   = Tools::typeIntBound,
	  Inner     = Type::Union(Tools, DOM_NIL),
          Outer     = Type::Union(Tools::typeCompositionBound, DOM_NIL),
	  MinSlope  = Tools::typeSlopeBound,
	  MaxSlope  = Tools::typeSlopeBound
         );

    //////////////////////////////////////////////////////////////////////////

    domtype := DOM_LIST;

    isA := proc(compo : Type::AnyType, summ) 
	       local startOptions;
	   begin
	       if not testtype(compo, Type::ListOf(Type::NonNegInt)) then return(FALSE) end_if;
	       if args(0) > 1 and domtype(summ) = DOM_INT then
		   startOptions := 3;
		   if _plus(op(compo)) <> summ then return(FALSE); end_if;
	       else 
		   startOptions := 2;
	       end_if;
	       Tools::isCorrect(compo, dom::parseOptions(args(startOptions..args(0))));
	   end_proc;

    size := l -> _plus(op(l));

    // Not very useful yet, as domains of this category could as well
    // call dom::countFromGenerator right away. But if some time in
    // the future we can improve integerListsLexTools::count, we will
    // just have to make a modification here to let all domains of
    // this category get the benefit.

    count_integerListsLexClass := dom::countFromGenerator;
    count := count_integerListsLexClass;

    //////////////////////////////////////////////////////////////////////////
    // wrapper methods around combinat::integerListsLexTools
    // first, Next, ...
    //////////////////////////////////////////////////////////////////////////

    first :=
    proc(n: Type::Integer)
        : Type::Union(dom, DOM_FAIL)
    begin
	Tools::first(n, dom::parseOptions(args(2..args(0))));
    end_proc;

    last :=
    proc(n: Type::Integer) 
        : Type::Union(dom, DOM_FAIL)
    begin
	Tools::last(n, dom::parseOptions(args(2..args(0))));
    end_proc;

    Next :=
    proc(l: dom) : Type::Union(dom, DOM_FAIL)
    begin
	Tools::Next(l, dom::parseOptions(args(2..args(0))));
    end_proc;

    previous :=
    proc(l: dom) 
        : Type::Union(dom, DOM_FAIL)
    begin
        Tools::previous(l, dom::parseOptions(args(2..args(0))));
    end_proc;

    generator :=
    proc(n: Type::Integer)
        : combinat::generators
        local constraints;
    begin
        constraints := dom::parseOptions(args(2..args(0)));
        combinat::generators::fromNext
        (Tools::first(n, constraints),
         Tools::Next, constraints);
    end_proc;

    _less :=
    proc(l1: dom, l2: dom)
	local i;
    begin
	for i from 1 to min(nops(l1), nops(l2)) do
	    if l1[i] <> l2[i] then
		return(l1[i] < l2[i]);
	    end_if;
	end_for;
	nops(l1) < nops(l2);
    end_proc;

begin
end_category:

