/* 
Cat::InvolutiveDivision -- the category of involutive divisions

Author: Marcus Hausdorf, Uni Mannheim
        (hausdorf@math.uni-mannheim.de)
Created: 	Mon Dec 10 16:18:31 2001
Last update: 	Mon Mar 31 13:48:25 2003
*/


/* 
Arguments:

nV: length of the multiindices
	
Entries:

========== obsolete ==========

separationTableMult(mu_l: list of multiindices of length nV):
    sep_t: table(mu -> set of multiplicative variables for mu_l[i] $ mu in mu_l)
    
separationTableNonMult(mu_l: list of multiindices of length nV):
    sep_t: table(mu -> set of nonmultiplicative variables for mu_l[i] $ mu in mu_l)
    
separationTableMultAdd(sep_t, mu):
    add multiindex mu to separation table and recompute multiplicative variables
    
separationTableMultRemove(sep_t, mu):
    remove multiindex mu from separation table and recompute multiplicative variables

separationTableNonMultAdd, separationTableNonMultRemove:
    same for nonmultiplicative variables
    
multDirs(mu, <mu_l>):
    compute the multiplicative directions of mu w.r.t mu_l (can be omitted for a global
    division
    
==============================

sepListMult(mu_l):
    returns a list of the form [mu, mv_s] giving for each multi index mu in mu_l its 
    multiplicative indices mv_s with respect to the involutive division and the multi indices 
    in mu_l
    
sepListMultAdd(sep_l, mu):
    given a separation list (exactly as returned by sepListMult; this is important for some
    involutive divisions, e.g. the Janet division, give their result as an ordered list w.r.t.
    some ordering), the multi index mu is inserted into sep_l and the multiplicative indices 
    are recomputed
    
sepListMultRem(sep_l, mu):
    same as sepListMultRem, except that mu is removed form sep_l
    
sepListNonMult, sepListNonMultAdd, SepListNonMultRem: 
    same as the respective methods above, only for nonmultiplicative indices
    
isInvDivisor(mu, nu, mv_s):
    returns mu - nu if nu lies in the cone of mu restricted by the directions in the set mv_s
    and 0 otherwise
    
autoreduce(mu_l):
    returns an autoreduced list of multi indices with the same involutive span as the multi 
    indices in mu_l
    
invAutoreduce(mu_l):
    returns an involutively autoreduced list of multi indices with the same involutive span
    as the multi indices in the original list mu_l
    
complete(mu_l):
    returns an involutively autoreduced list of multi indices with the same span as the multi
    indices in the original list mu_l, i.e. a (strong) involutive basis

isInvDivisor(mu, nu: multiindices of length n):
    0, if mu does not involutively divide nu;
    (nu - mu) otherwise 
*/


category Cat::InvolutiveDivision(nV:Type::PosInt)
local i, avars;

    category Cat::BaseCategory;
    
    //==================
    // undefined methods
    //==================
    
    // obsolete; will be removed some time
    
    separationTableMult;

    // separationTableMultAdd;
    
    // separationTableMultRemove;
    
    //================== 
    
    sepListMult;
    
    sepListMultAdd;
    
    sepListMultRem;
    
    // multDirs;
    
    //==================
    
    numVars := nV;
    
    // obsolete; will be removed some time
    
    separationTableNonMult :=
	(x -> map(dom::separationTableMult(x), l -> ({i $ i = 1..nV} minus l)));
    
    separationTableNonMultAdd := 
	((t, mi) -> map(dom::separationTableMultAdd(t, mi), l -> ({i $ i = 1..nV} minus l)));
    
    separationTableNonMultRemove := 
	((t, mi) -> map(dom::separationTableMultRemove(t, mi), l -> ({i $ i = 1..nV} minus l)));

    //==================
    
    sepListNonMult := 
	(x -> map(dom::sepListMult(x), l -> [l[1], avars minus l[2]]));
	
    sepListNonMultAdd :=
	((t, mi) -> map(dom::sepListMultAdd(map(t, l -> [l[1], avars minus l[2]]), mi),
	    l -> [l[1], avars minus l[2]]));
	
    sepListNonMultRem :=
	((t, mi) -> map(dom::sepListMultRem(map(t, l -> [l[1], avars minus l[2]]), mi),
	    l -> [l[1], avars minus l[2]]));
	    
    isInvDivisor := proc(mu:Type::ListOf(Type::NonNegInt), nu:Type::ListOf(Type::NonNegInt),
	mV_s:Type::SetOf(Type::PosInt))
	local tmp, k;
	begin
	    tmp := zip(nu, mu, _subtract);
	    for k from 1 to nV do
		if tmp[k] < 0 or (tmp[k] > 0 and not contains(mV_s, k)) then
		    return(0);
		end_if;
	    end_for;
	    tmp;
	end_proc;
	
    autoreduce := proc(mu_l:Type::ListOf(Type::ListOf(Type::NonNegInt))):
	    Type::ListOf(Type::ListOf(Type::NonNegInt))
	local res_l, mu, nu, reduced;
	begin
	    if nops(mu_l) = 1 then
		return(mu_l);
	    end_if;
	    mu_l := sort([op({op(mu_l)})], 
		(l1, l2) -> bool(0 > (select(zip(l1, l2, _subtract), not iszero))[1]));
	    res_l := [mu_l[1]];
	    delete mu_l[1];
	    repeat
		mu := mu_l[1];
		delete mu_l[1];
		reduced := FALSE;
		for nu in res_l do
		    if select(zip(mu, nu, _subtract), l -> bool(l < 0)) = [] then
			reduced := TRUE;
			break;
		    end_if;
		end_for;    
		if not reduced then
		    res_l := append(res_l, mu);
		end_if;
	    until mu_l = [] end_repeat;
	    res_l;
	end_proc;
	
    invAutoreduce := proc(mu_l:Type::ListOf(Type::ListOf(Type::NonNegInt))):
	    Type::ListOf(Type::ListOf(Type::NonNegInt))
	local res_l, mu, nu, reduced, sepm_l, sepm_t;
	begin
	    if nops(mu_l) = 1 then
		return(mu_l);
	    end_if;
	    if args(0) = 2 then
		sepm_l := args(2);
	    else
		sepm_l := dom::sepListMult(mu_l);
	    end_if;
	    sepm_t := table(op(map(sepm_l, l -> (l[1] = l[2]))));
	    mu_l := sort([op({op(mu_l)})], 
		(l1, l2) -> bool(0 > (select(zip(l1, l2, _subtract), not iszero))[1]));
	    res_l := [mu_l[1]];
	    delete mu_l[1];
	    repeat
		mu := mu_l[1];
		delete mu_l[1];
		reduced := FALSE;
		for nu in res_l do 
		    if not iszero(dom::isInvDivisor(nu, mu, sepm_t[nu])) then
			reduced := TRUE;
			sepm_l := dom::sepListMultRem(sepm_l, mu);
			sepm_t := table(op(map(sepm_l, l -> (l[1] = l[2])))); 
			break;
		    end_if;
		end_for;
		if not reduced then
		    res_l := append(res_l, mu);
		end_if;
	    until mu_l = [] end_repeat;
	    res_l;
	end_proc;
	    	
    complete := proc(mu_l:Type::ListOf(Type::ListOf(Type::NonNegInt)))
	local donenmp_t, sepm_l, sepm_t, sepnm_t, involutive, reduced, tmp, tmp2, mu, k;
	begin
	    if nops(mu_l) = 1 then
		return(mu_l);
	    end_if;
	    donenmp_t := table(op(map(mu_l, _equal, {})));
	    sepm_l := dom::sepListMult(mu_l);
	    repeat
		sepm_t := table(op(map(sepm_l, l -> (l[1] = l[2]))));
		sepnm_t := map(sepm_t, l -> avars minus l);
		mu_l := sort(mu_l, (l1, l2) -> (_plus(op(l1)) <= _plus(op(l2))));
		involutive := TRUE;
		for k from 1 to nops(mu_l) do
		    for i in (sepnm_t[mu_l[k]] minus donenmp_t[mu_l[k]]) do
			tmp := subsop(mu_l[k], i = mu_l[k][i] + 1);
			tmp2 := _plus(op(tmp));
			donenmp_t[mu_l[k]] := donenmp_t[mu_l[k]] union {i};
			reduced := FALSE;
			for mu in mu_l do
			    if mu = mu_l[k] then next end_if;
			    if (_plus(op(mu)) > tmp2) then break end_if;
			    if not iszero(dom::isInvDivisor(mu, tmp, sepm_t[mu])) then
				reduced := TRUE;
				break;
			    end_if;
			end_for;
			if not reduced then
			    involutive := FALSE;
			    mu_l := listlib::singleMerge(mu_l, [tmp], 
				(l1, l2) -> (_plus(op(l1)) <= _plus(op(l2))));
			    sepm_l := dom::sepListMultAdd(sepm_l, tmp);
			    donenmp_t[tmp] := {};
			    break;
			end_if;
		    end_for;
		    if not involutive then break end_if;
		end_for;
	    until involutive end_repeat;			
	    dom::invAutoreduce(mu_l, sepm_l);
	end_proc;
	
    begin    
	if args(0) <> 1 then error("wrong no of args") end_if;
	avars := {i $ i = 1..nV};

    end_category:
