/*****************************************************************************
  $Id: ModuleWithBasis.mu,v 1.16 2004/10/05 16:20:25 nthiery Exp $
  File:		lib/DOMAINS/CATEGORY/FreeModule.mu
  Authors:	Nicolas M. Thiry <nthiery@users.sourceforge.net>
  Created:	16/02/2001
  License:	LGPL
*****************************************************************************/

/**
    Cat::ModuleWithBasis(R) -- R-Free module over some basis dom::basisIndices

    Assumptions:
     - dom::basisIndices should be a combinatorial class with a working
       testtype function.  If it's a facade domain, then it should
       represent a subset of a true domain, and provide a domtype
       entry containing this true domain.
     - By default, multiplication by scalars on the left (respectively
       on the right) is implemented by multiplying accordingly the
       coefficients. Skew scalar multiplications can be achieved by
       redefining multcoeffs and multcoeffsLeft.

    Required entries:
     - basis
     - monomial
     - toList
     - coeff
     - _plus, _mult

    Possible optimizations:
      - define and use Rmult2
	(by order of preference: R::mult2, R::mult, R::_mult, *)
      - define and use Rnegate ?
**/

alias(plus2		= operators::plus2):
alias(mult2		= operators::mult2):
alias(declareDomain     = operators::overloaded::declareDomain):
alias(declareConversion = operators::overloaded::declareConversion):
alias(declareSignature  = operators::overloaded::declareSignature):
alias(rankers		= combinat::rankers):

category Cat::ModuleWithBasis(R: DOM_DOMAIN)

    category
	if R::hasProp(Cat::Field) then
	    Cat::VectorSpace(R)
	else
	    Cat::Module(R)
	end_if,
	Cat::UseOverloading;
    axiom
	if R::hasProp(ax) then
	    ax;
	end_if $ ax in [Ax::canonicalRep, Ax::normalRep];

    info := "Cat::ModuleWithBasis: category of modules with a distinguished basis";

    interfaceAdd :=
        {// hold(options), 
	 hold(basisIndices), hold(basis),
	 hold(coeffRing),
         hold(zero), hold(iszero),
         // hold(generate), 
	 hold(monomial),
         hold(expr), hold(TeX), hold(convert),
         hold(lsupport), hold(support),
         hold(nterms), hold(lterm), hold(terms),
         hold(lcoeff),
         hold(degree), hold(degreevec),
         hold(mapcoeffs), hold(mapsupport), hold(selectSupport),
         hold(_plus), hold(_negate), hold(_mult),
	 hold(multcoeffs), hold(multcoeffsLeft),
	 hold(genericCoefficients), hold(genericElement),
	 hold(sparseVector), hold(fromSparseVector),
	 hold(endomorphismMatrix)
         };

    // characteristic := R::characteristic; ???


    /////////////////////////////////////////////////////////////////////////
    // Basic information on the free module
    /////////////////////////////////////////////////////////////////////////

    basisIndices   := toBeDefined;
    basisName      := hold(B);
    basis          := combinat::imageClass(dom::basisIndices, dom::term);
    coeffRing      := R;

    //////////////////////////////////////////////////////////////////////////
    // Conversions and overloading
    //////////////////////////////////////////////////////////////////////////

    // Default initialization using the overloaded and conversion
    // mechanism from operators::overloaded.
    //
    // A domain in this category can set its own initialization by
    // overriding initDomainFreeModule

    initDomainAdd := dom::initDomainFreeModule;
    initDomainFreeModule :=
    proc()
	local basisDomtype;
    begin
	// make sure that the coeff ring is declared:
	declareDomain(R);
	// make sure that the basis is declared
	if dom::basisIndices::hasProp(Ax::systemRep) then
	    basisDomtype := dom::basisIndices::domtype;
	    if basisDomtype = FAIL then
		warning("The domain ".expr2text(dom::basisIndices).
			" has axiom Ax::systemRep but do not provide the domtype entry; ".
			expr2text(dom)."::new may not work properly.");
	    end_if;
	else
	    basisDomtype := dom::basisIndices;
	end_if;
	if basisDomtype <> FAIL then
	    declareDomain(basisDomtype);
	    // Declare the conversion from the basis
	    // Warning: dom::term is not called at the interactive level,
	    // and does no type checking!
	    declareConversion(basisDomtype,
			      dom,
			      if dom::straightenBasis<>FAIL then
				  dom::straightenBasis
			      else
				  dom::term;
			      end_if, Explicit);
	end_if;
	//if dom::_plus = operators::_plus then
	    declareSignature(plus2, [dom, dom], dom::plus2);
	//end_if;
	//if dom::_mult = operators::_mult then
	    declareSignature(mult2, [dom, R  ], dom::multcoeffs);
	    declareSignature(mult2, [R,   dom], dom::multcoeffsLeft);
	//end_if;
	if dom::dual <> FAIL then
	    declareSignature(operators::scalar, [dom, dom::dual], dom::dualityScalarProduct);
	end_if;
    end_proc;

    /////////////////////////////////////////////////////////////////////////
    // Constructors and constants
    /////////////////////////////////////////////////////////////////////////

    zero     := dom::fromList([]);
    monomial := toBeDefined;

    term :=
    proc(support: dom::basisIndices) : dom
    begin
	dom::monomial(R::one, support)
    end_proc;

    iszero :=
    proc(x: dom) : DOM_BOOL
    begin
	bool(dom::nterms(x)=0);
    end_proc;

    isMonomial :=
    proc(x:dom) : DOM_BOOL
    begin
	bool(dom::nterms(x)=1)
    end_proc;

    // for backward compatibility
    indexed := dom::domainWrapper;

    domainWrapper := prog::domainWrapper(dom);
    _indexDomain :=
    if dom::basisIndices::hasProp(Ax::systemRep) and dom::basisIndices::domtype=DOM_LIST then
	dom@DOM_LIST;
    else
	dom;
    end_if;

    /////////////////////////////////////////////////////////////////////////
    // Conversions
    /////////////////////////////////////////////////////////////////////////

    // Note: in expr/print/TeX, it's not critical if we don't
    // recognize a coefficient as being one. But it's very bad to
    // write one for a coefficient that is *not* one. So we use
    // c=R::one instead of R::isone(c), which seems safer.

    // In particular, this works around MuPAD's bug in
    // Dom::ExpressionField()::equal in MuPAD <= 2.5.2.
    // (isn't this bug already fixed in lib/init.mu ?)

    expr :=
    proc(x: dom)
    begin
	case dom::nterms(x)
	  of 0 do return(0);
	  of 1 do return(dom::exprMonomial(dom::lmonomial(x)));
	  otherwise
	      _plus(dom::exprMonomial(x) $ x in dom::monomials(x));
	end_case;
    end_proc;

    exprMonomial :=
    proc(x: dom)
    begin
	if dom::lcoeff(x) = R::one then
	    dom::exprTerm(dom::lsupport(x))
	else
	    _mult(expr(dom::lcoeff(x)),
		  dom::exprTerm(dom::lsupport(x)));
	end_if;
    end_proc;

    exprTerm :=
	x -> dom::basisName(expr(x));
    exprTermIndex :=
	x -> dom::basisName[op(expr(x))];

    print :=
    proc(x: dom)
    begin
	case dom::nterms(x)
	  of 0 do return(0);
	  of 1 do return(dom::printMonomial(x));
	  otherwise
	      hold(_plus)(dom::printMonomial(x) $ x in dom::monomials(x));
	end_case;
    end_proc;

    printMonomial :=
    proc(x: dom)
    begin
	if dom::lcoeff(x) = R::one then
	    dom::printTerm(dom::lsupport(x))
	else
	    hold(_mult)(dom::lcoeff(x),
			dom::printTerm(dom::lsupport(x)));
	end_if;
    end_proc;

    // The trick below makes printTerm a reference to exprTerm, instead
    // of a copy. This allows for redefining on the fly exprTerm
    printTerm := ()->dom::exprTerm(args());

    // This implementation of expr2text tries to make sure that the
    // result can be parsed back into the same element of the domain
    // Note: a domain that inherits from this one and adds stuff to the
    // internal representation will have to redefine this method!
    expr2text :=
    proc(x: dom) : DOM_STRING
    begin
	"(".expr2text(dom).")::fromList(".expr2text(dom::toList(x)).")";
    end_proc;


    TeXTerm := (z, p) -> generate::tex(dom::exprTermIndex(z), p);
		     
    // Reuse the TeX fonctions from MuPAD's Free Modules
    TeXMonomial :=
    proc(x: dom, p:DOM_INT):DOM_STRING
	local s: dom::basisIndices, c: dom::coeffRing;
    begin
	c := dom::lcoeff(x);
	s := dom::lsupport(x);
	if s = dom::oneBasis then
	    generate::tex(c,p)
	elif c = dom::coeffRing::one then
//	    generate::tex(s,p)
	    dom::TeXTerm(s, p);
	else
	    generate::TeXoperator(" ",p,output::Priority::Mult,c,dom::term(s))
	end_if;
    end_proc;

    TeX :=
        proc(x:dom,p:DOM_INT) : DOM_STRING
	local res: DOM_STRING, s: DOM_STRING;
        begin 
	    res := "";
	    for s in map(dom::monomials(x),dom::TeXMonomial,output::Priority::Plus) do
		if res <> "" then
		    // Look if the next string starts with a "-" sign.
		    // Note: the %if works around a change in strings indexed
		    // access from MuPAD 3.0.0 on
		    if s[%if "12"[1]="1" then 1 else 0; end_if] = "-" then
			res := res." "
		    else
			res := res." + "
		    end_if;
		end_if;
		res := res.s;
	    end_for;
	    generate::TeXquote(res,p,output::Priority::Plus);
	end_proc;

    TeX2 :=
    proc(x: dom)
	local c,s,t;
    begin
	s:="";
	for t in poly2list(x) do
	    if s<>"" then s:=s."+"; end_if;
	    c:=op(t,1);
	    t:=op(t,2);
	    s:=s. (if c = R::one then "" else generate::TeX(c) end_if)
		. generate::TeX(t);
	end_for;
    end_proc;


    /////////////////////////////////////////////////////////////////////////
    // Access to terms, leading terms, monomials, coefficients, ...
    /////////////////////////////////////////////////////////////////////////

    coeff     := toBeDefined;
    toList    := toBeDefined;

    poly2list := dom::toList;

    fromList :=
    if dom::plus <> FAIL then
	proc(l: Type::ListOf(Type::ListProduct(dom::coeffRing, dom::basisIndices))) : dom
	    local m;
	begin
	    dom::plus(dom::monomial(op(m)) $ m in l);
	end_proc;
    else
	proc(l: Type::ListOf(Type::ListProduct(dom::coeffRing, dom::basisIndices))) : dom
	    local m;
	begin
	    if nops(l) = 0 then
		dom::zero
	    else
		_plus(dom::monomial(op(m)) $ m in l);
	    end_if;
	end_proc;
    end_if;
    nterms :=
    proc(x: dom) : Type::NonNegInt
    begin
	nops(dom::toList(x));
    end_proc;

    support :=
    proc(x: dom): Type::ListOf(dom::basisIndices)
    begin
	map(dom::toList, op, 2);
    end_proc;

    coeffs :=
    proc(x: dom) : Type::ListOf(dom::coeffRing)
    begin
	map(dom::toList(x), op, 1);
    end_proc;

    terms :=
    proc(x: dom) : Type::ListOf(dom)
    begin
	map(dom::toList(x), dom::term@op, 2);
    end_proc;

    monomials :=
    proc(x: dom) : Type::ListOf(dom)
    begin
	map(dom::toList(x), dom::monomial@op);
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // lsupport, lterm, lcoeff, lmonomial
    //////////////////////////////////////////////////////////////////////////
    lsupport :=
    proc(x: dom) : Type::Union(dom::basisIndices, DOM_FAIL)
    begin
	if dom::iszero(x) then return(FAIL) end_if;
	op(dom::toList(x),[1,2]);
    end_proc;
    // for backward compatibility
    degreevec         := dom::lsupport;
    monomial2exponent := dom::lsupport;
    lexponent	      := dom::lsupport;

    lterm :=
    proc(x: dom) : dom
	local s;
    begin
	if (s := dom::lsupport(x)) = FAIL then return(dom::zero); end_if;
	dom::term(s);
    end_proc;

    lcoeff :=
    proc(x: dom) : dom::coeffRing
	local s;
    begin
	if (s := dom::lsupport(x)) = FAIL then return(dom::coeffRing::zero); end_if;
	dom::coeff(x, s);
    end_proc;

    lmonomial :=
    proc(x: dom) : dom
	local s;
    begin
	if (s := dom::lsupport(x)) = FAIL then return(dom::zero); end_if;
	dom::monomial(dom::coeff(x,s), s);
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // tsupport, tterm, tcoeff, tmonomial
    //////////////////////////////////////////////////////////////////////////

    tsupport :=
    proc(x: dom) : Type::Union(dom::basisIndices, DOM_FAIL)
	local n;
    begin
	if (n:=dom::nterms(x)) = 0 then
	    FAIL;
	else
	    dom::nthsupport(x, n);
	end_if;
    end_proc;

    tterm :=
    proc(x: dom) : dom
	local n;
    begin
	if (n:=dom::nterms(x)) = 0 then
	    dom::zero
	else
	    dom::nthterm(x, n);
	end_if;
    end_proc;

    tcoeff :=
    proc(x: dom) : dom::coeffRing
	local n;
    begin
	if (n:=dom::nterms(x)) = 0 then
	    dom::coeffRing::zero
	else
	    dom::nthcoeff(x, n);
	end_if;
    end_proc;

    tmonomial :=
    proc(x: dom) : dom
	local n;
    begin
	if (n:=dom::nterms(x)) = 0 then
	    dom::zero
	else
	    dom::nthmonomial(x, n);
	end_if;
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // nthsupport, nthterm, nthcoeff, nthmonomial
    //////////////////////////////////////////////////////////////////////////

    nthsupport :=
    proc(x: dom, i: Type::PosInt) : Type::Union(dom::basisIndices, DOM_FAIL)
    begin
	if dom::nterms(x) < i then return(FAIL); end_if;
	dom::toList(x)[i][2];
    end_proc;

    nthterm :=
    proc(x: dom, i: Type::PosInt) : Type::Union(dom, DOM_FAIL)
        local s;
    begin
	if (s := dom::nthsupport(x,i)) = FAIL then return(FAIL); end_if;
	dom::term(dom::nthsupport(x,i));
    end_proc;

    nthcoeff :=
    proc(x: dom, i: Type::PosInt) : Type::Union(dom::coeffRing, DOM_FAIL)
    begin
	if dom::nterms(x) < i then return(FAIL); end_if;
	dom::toList(x)[i][1];
    end_proc;

    nthmonomial :=
    proc(x: dom, i: Type::PosInt) : Type::Union(dom, DOM_FAIL)
    begin
	if dom::nterms(x) < i then return(FAIL); end_if;
	dom::monomial(op(dom::toList(x)[i]));
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // Operations on all coefficients / support elements: mapcoeffs, mapsupport
    //////////////////////////////////////////////////////////////////////////

    // ZERO CHECKS

    mapcoeffsNoChecks :=
    proc(x: dom, f)
	local options;
    begin
	options := args(3..args(0));
	// Would it be faster by using subsop ?
        dom::fromList(map(dom::toList(x), t -> [f(t[1], options), t[2]]));
    end_proc;

    // same as mapcoeffs, but ensures that zero coefficients are removed
    // should this be the default?
    mapcoeffs :=
    proc(x: dom, f)
	local options;
    begin
	options := args(3..args(0));
	// Would it be faster by using subsop ?
        dom::fromList(select(map(dom::toList(x), t -> [f(t[1], options), t[2]]),
			     not iszero @ op, 1));
    end_proc;

    mapsupport :=
    proc(x: dom, f) : dom
	local options;
    begin
	options := args(3..args(0));
        dom::fromList(map(dom::toList(x), t -> [t[1], f(t[2], options)]));
    end_proc;

    mapterms := dom::mapsupport; // backward compatibility

    selectSupport :=
    proc(x: dom, f: Type::Function) : dom
	local options;
    begin
	options := args(3..args(0));
        dom::fromList(select(dom::toList(x), t -> f(t[2], options)));
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // Grading-related operations
    // Deprecated by the entries in Cat::GradedModuleWithBasis!
    // To be removed!
    //////////////////////////////////////////////////////////////////////////

    degreeBasis :=
    if dom::basisIndices::degree <> FAIL then
	// the id@ prevents changing dom in dom::basisIndices::degree
	id@dom::basisIndices::degree;
    else
	degree; // or nothing ???
    end_if;

    degree :=
    if dom::degreeBasis<>FAIL then
        proc(x: dom)
	    local s;
	begin
	    max(0,dom::degreeBasis(s, args(2..args(0))) $ s in dom::support(x));
	end_proc:
    end_if;

    /////////////////////////////////////////////////////////////////////////
    // Arithmetic operations
    /////////////////////////////////////////////////////////////////////////

    _plus := toBeDefined;
    _mult := toBeDefined;

    plus :=
    if dom::plus2 <> FAIL then
	operators::makeAssociative
	(dom::plus2, Zero = dom::zero, Source=dom, ImageSet=dom);
    end_if;

    _negate :=
    proc(x: dom)
    begin
	dom::mapcoeffs(x, R::_negate);
    end_proc;

    _subtract := (x,y) -> x+(-y);

    intmult := dom::multcoeffs;

    multcoeffs :=
    proc(x: dom, c: dom::coeffRing) : dom
    begin
        dom::mapcoeffs(x, _mult, c);
    end_proc;

    multcoeffsLeft :=
    if dom::coeffRing::hasProp(Cat::CommutativeRing) then
	// Little optimization when the coefficient ring is commutative
	proc(c: dom::coeffRing, x: dom) : dom
	begin
	    dom::mapcoeffs(x, _mult, c);
	end_proc;
    else
	proc(c: dom::coeffRing, x: dom) : dom
	begin
	    dom::mapcoeffs(x, c1 -> c*c1);
	end_proc;
    end_if;

    /////////////////////////////////////////////////////////////////////////
    // Simplifications / ...
    /////////////////////////////////////////////////////////////////////////

    simplify :=
    proc(x: dom)
    begin
	mapcoeffs(x, simplify, args(2..args(0)))
    end_proc;

    normal :=
    proc(x: dom)
    begin
	mapcoeffs(x, normal, args(2..args(0)))
    end_proc;

    expand :=
    proc(x: dom)
    begin
	mapcoeffs(x, expand, args(2..args(0)))
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    /// Basis related things

    genericCoefficients :=
    proc(f: Type::AnyType)
    begin
	map(dom::basisIndices::list(args(2..args(0))), f);
    end_proc;

    genericElement :=
    proc(f: Type::AnyType)
	local basis, t;
    begin
	basis := dom::basisIndices::list(args(2..args(0)));
	dom::fromList([[f(t), t] $ t in basis]);
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // Interface with linalg / Dom::/*Sparse*/Matrix
    //////////////////////////////////////////////////////////////////////////

    rankBasis   := combinat::rankers::fromType(dom::basisIndices)[1];
    unrankBasis := combinat::rankers::fromType(dom::basisIndices)[2];

    sparseVectorsDomain := Dom::SparseMatrix(dom::coeffRing);
    sparseMatricesDomain := Dom::SparseMatrix(dom::coeffRing);
    sparseVectorsType   := dom::sparseVectorsDomain;

    sparseVector :=
    // Builds a sparse column vector from an element of this domain
    // The ranking used to associate a row number to a basis element is
    // done as follow:
    //  - The second optional argument can provide a ranker or a list
    //    from which the ranking will be done
    //  - By default, the ranker from dom::basisIndices is used

    // If dom::dimension is defined, then the vectors will all have
    // dom::dimension rows. Otherwise, they will have variable number of rows
    proc(p: dom,
	 rank = dom::rankBasis :
	   Type::Union(rankers::typeRanker, Type::ListOf(dom::basisIndices)))
	: dom::sparseVectorsType
	local s, t, eq;
    begin
	if domtype(rank) = DOM_LIST then
	    rank := rankers::rankFromList(rank);
	end_if;
	t := table(rank(op(s,2))=op(s,1) $ s in dom::toList(p));
	dom::sparseVectorsDomain(if dom::dimension <> FAIL then
				     dom::dimension
				 else
				     max(0,op(eq,1) $ eq in t)
				 end_if, 1,
				 t);
    end_proc;

    fromSparseVector :=
    proc(vector: dom::sparseVectorsDomain,
	 unrank = dom::unrankBasis :
	   Type::Union(rankers::typeUnranker, Type::ListOf(dom::basisIndices)))
	: dom
    begin
	if domtype(unrank) = DOM_LIST then
	    unrank := rankers::unrankFromList(unrank);
	end_if;
	// Warning: this relies on the internal structure of sparse matrices!
	dom::fromList(map(poly2list(op(extop(vector,3),1)),
			  monomial -> [op(monomial,1),unrank(op(monomial,2))]));
    end_proc;

    endomorphismMatrix :=
    proc(endo)
	local b;
    begin
	_concat(dom::sparseVector(endo(b))
		$ b in dom::basis::list(args(2..args(0))));
    end_proc;
			    
    // Note: find a name for all the operations concerning a bunch of vectors

    // Should be optimized by using lower level methods
    // such as listOfVectorsToMatrix / listOfVectorsFromMatrix
    // that could be specialized
    basisVectors :=
    proc(l = Type::ListOf(dom))
	local mat, elim, rank, i;
    begin
	mat   := _concat(op(map(l, dom::sparseVector)));
	elim  := dom::sparseMatrix::gaussElim(mat, hold(ColumnElimination));
        mat   := elim[1];
	rank  := elim[2];
	[dom::fromSparseVector(mat::dom::col(mat,i)) $ i = 1..rank];
    end_proc;

    /////////////////////////////////////////////////////////////////////////
    // Duality and scalar product
    /////////////////////////////////////////////////////////////////////////
    
    scalar := if dom::dual <> FAIL then operators::scalar; end_if;

    dualityScalarProduct := 
    if dom::dual <> FAIL then
	proc(v1 : dom, v2 : dom::dual) : dom::coeffRing
	    name dualityScalarProduct;
	    local sv1, sv2, supint, g;
	begin
	    sv1 := dom::support(v1);
	    sv2 := dom::dual::support(v2);
	    supint := select(sv2, (z) -> contains(sv1, z)<>0);
	    if supint = [] then 
		dom::coeffRing::zero; // to ensure that the scalarProduct is in the good ring. 
	    else
		_plus(dom::coeff(v1, g) * dom::dual::coeff(v2, g) $ g in supint);
	    end_if;
	end_proc;
    end_if;

end_category:
