/* 
Cat::SolvableAlgebra -- the category of polynomial algebras of solvable type

Author: Marcus Hausdorf, Uni Heidelberg
        (marcus.hausdorf@iwr.uni-heidelberg.de)
Created: 	Tue Oct  1 13:43:28 2001
Last update: 	Thu Sep 18 11:50:03 2003
*/

/*  
Cat::SolvableAlgebra(R) is the category of solvable algebras of polynomial type, i.e. its 
elements are polynomial rings with a (not necessarily commutative) multiplication '.' 
satisfying a.f = a*f and le(f.g)=le(f)+le(g) for all a in R and the chosen term order. In such 
algebras the computation of Groebner and involutive bases is possible.

The following methods have to be provided by domains in this category:

coeffRing: R

variables: names of the variables the exponent vectors refer to.

numVars: number of variables resp. length of the exponent vectors.

leader(f:dom): extract the 'leading information' from a domain element. The result is 
    list with two elements: the first is the leading coefficient (i.e. an element from
    R), the second the leading 'exponent'. Its definition may vary for different 
    domains, but exponents must be comparable by the methods exponentOrder/exponentLess.
    
exponentOrder(mu1, mu2): compare the exponents mu1 and mu2; return
    -1 if mu1 < mu2,
     0 if mu1 = mu2,
     1 if mu1 > mu2.

allMonomials(f:dom): convert f into a list of 'terms' of the form [coefficient, exponent]
    (similar to poly2list).
    
exponent2monomial(mu): convert an exponent into a domain element.
*/

alias(A_ERR = error("wrong type of arguments"),
      N_ERR = error("wrong number of arguments"));
  
category Cat::SolvableAlgebra(R)
    
    local _list2table;
    category Cat::Ring, Cat::LeftModule(R);
    
    //==================
    // undefined methods
    //==================
    
    coeffRing;
    variables;
    numVars;
    exponent2monomial;
    // monomial2exponent;
    exponentOrder;
    // reduce;
    allMonomials;
    leader;
    
    //================
    // generic methods
    //================
    
    exponentLess := ((l1, l2) -> (bool(dom::exponentOrder(l1, l2) = -1)));
    
    autoreduce := proc(F:Type::ListOf(dom))
	local n:DOM_INT, G:DOM_LIST, g:dom, f:dom, reduced:DOM_BOOL, tmp, hr_f:DOM_BOOL,
	    DOM:DOM_DOMAIN;
	begin
	    if args(0) < 1 or args(0) > 2 then
		N_ERR;
	    end_if;
	    // F := sort(F, dom::elemLess);
	    DOM := dom;
	    tmp := map(F, l -> DOM::leader(l));
	    if args(0) = 2 and args(2) = hold(Head) then
		hr_f := TRUE;
		F := zip(F, tmp, (l1, l2) -> table("elem" = l1, "mons" = [l2], "lmon" = l2));
	    else
		hr_f := FALSE;
		F := zip(F, tmp, (l1, l2) -> table("elem" = l1, "mons" = DOM::allMonomials(l1),
		    "lmon" = l2));
	    end_if;
	    F := sort(F, (l1, l2) -> DOM::exponentLess(l1["lmon"][2], l2["lmon"][2]));
	    repeat
		reduced := FALSE;
		n := nops(F) + 1;
		repeat
		    n := n - 1;
		    G := F;
		    f := F[n];
		    delete G[n];
		    g := dom::internalNForm(f, G, hr_f);
		    reduced := bool(f <> g);
		until (n = 1 or reduced) end_repeat;
		if reduced then
		    delete F[n];
		    if not iszero(g["elem"]) then
			F := listlib::insert(F, g, 
			    (l1, l2) -> DOM::exponentLess(l1["lmon"][2], l2["lmon"][2]));
		    end_if;
		end_if;
	    until not reduced end_repeat;
	    map(F, l -> l["elem"]);
	end_proc;
	
    nForm := proc(f:dom, F:Type::ListOf(dom))
	local a1, a2, DOM:DOM_DOMAIN, res, o_t;
	begin
	    o_t := prog::getOptions([hold(Head) = FALSE, hold(Coeffs) = FALSE], 
		[hold(Head) = DOM_BOOL, hold(Coeffs) = DOM_BOOL], 2);
	    if o_t[hold(Head)] then
		a1 := table("elem" = f, "mons" = [dom::leader(f)]);
	    else
		a1 := table("elem" = f, "mons" = dom::allMonomials(f));
	    end_if;
	    DOM := dom;
	    a2 := map(F, l -> table("elem" = l, "lmon" = DOM::leader(l)));
	    res := dom::internalNForm(a1, a2, o_t[hold(Head)], o_t[hold(Coeffs)]);
	    if o_t[hold(Coeffs)] then
		return([res[1]["elem"], res[2]]);
	    else
		return([res["elem"]]);
	    end_if;
	end_proc;
	
    internalNForm := proc(f, G, hr_f:DOM_BOOL, co_f = FALSE)
	local reduced:DOM_BOOL, t, d, isDivisor:DOM_PROC, c, tmp1, tmp2, c_l, k; 
	begin
	    isDivisor := proc(mu1, mu2)
	    local tmp, k;
	    begin
		tmp := zip(mu2, mu1, _subtract);
		for k from 1 to nops(mu1) do
		    if tmp[k] < 0 then
			return(0);
		    end_if;
		end_for;
		tmp;
	    end_proc;	    
	
	    if co_f then 
		c_l := [dom::zero $ nops(G)];
	    end_if;
	    repeat
		reduced := FALSE;
		for k from 1 to nops(G) do
		    for t in f["mons"] do
			if not iszero((d := isDivisor(G[k]["lmon"][2], t[2]))) then
			    reduced := TRUE;
			    // f := dom::reduce(f["elem"], g["elem"], g["lmon"], t, d);
			    tmp1 := dom::exponent2monomial(d);
			    tmp2 := tmp1 * G[k]["elem"];
			    c := (t[1]) / (dom::leader(tmp2)[1]);
			    if co_f then
				c_l[k] := c_l[k] + dom::multcoeffs(tmp1, c);
			    end_if;
			    f := f["elem"] - (dom::multcoeffs(tmp2, c)); 
			    f := table("elem" = f, "lmon" = dom::leader(f));
			    if hr_f then
				f["mons"] := [f["lmon"]];
			    else
				f["mons"] := dom::allMonomials(f["elem"]);
			    end_if;
			end_if;
			if reduced then break end_if;
		    end_for;
		    if reduced then break end_if;
		end_for;
	    until (dom::iszero(f["elem"])) or not reduced end_repeat;
	    if co_f then
		return([f, c_l]);
	    else
		return(f);
	    end_if;
	end_proc;			

    
    //=======================
    // involutive normal form
    //=======================
    
    // involutiveNForm(f, F, L) computes the involutive normal form of f modulo the set F 
    // w.r.t. the involutive division L. The switch HeadReduce (default: FALSE) determines 
    // whether full or only head reductions are used. 
     
    involutiveNForm := proc(f:dom, F:Type::ListOf(dom), L:DOM_DOMAIN)
	local a1, a2, DOM:DOM_DOMAIN, i, res, o_t;
	begin
	    o_t := prog::getOptions([hold(Head) = FALSE, hold(Coeffs) = FALSE], 
		[hold(Head) = DOM_BOOL, hold(Coeffs) = DOM_BOOL], 3);
	    if o_t[hold(Head)] then
		a1 := table("elem" = f, "mons" = [dom::leader(f)]);
	    else
		a1 := table("elem" = f, "mons" = dom::allMonomials(f));
	    end_if;
	    DOM := dom;
	    a2 := map(F, l -> table("elem" = l, "lmon" = DOM::leader(l)));
	    res := dom::internalInvolutiveNForm(a1, a2, 
		_list2table(L::sepListMult([a2[i]["lmon"][2] $ i = 1..nops(a2)])), L, 
		o_t[hold(Head)], o_t[hold(Coeffs)]);
	    if o_t[hold(Coeffs)] then
		return([res[1]["elem"], res[2]]);
	    else
		return([res["elem"]]);
	    end_if;
       end_proc; 
	    	
    internalInvolutiveNForm := proc(f, G, sep_t, L, hr_f, co_f = FALSE)
	local reduced:DOM_BOOL, t, d, c, tmp1, tmp2, c_l, k;
	begin
	    if co_f then
		c_l := [dom::zero $ nops(G)];
	    end_if;
	    repeat
		reduced := FALSE;
		for k from 1 to nops(G) do
		    for t in f["mons"] do
			if not iszero((d := L::isInvDivisor(G[k]["lmon"][2], t[2], 
				sep_t[G[k]["lmon"][2]]))) 
			then
			    reduced := TRUE;
			    // f := dom::reduce(f["elem"], g["elem"], g["lmon"], t, d);
			    tmp1 := dom::exponent2monomial(d);
			    tmp2 := tmp1 * G[k]["elem"];
			    c := (t[1]) / (dom::leader(tmp2)[1]);
			    if co_f then
				c_l[k] := c_l[k] + dom::multcoeffs(tmp1, c);
			    end_if;			    
			    f := f["elem"] - (dom::multcoeffs(tmp2, c)); 
			    f := table("elem" = f, 
				"lmon" = dom::leader(f));
			    if hr_f then
				f["mons"] := [f["lmon"]];
			    else
				f["mons"] := dom::allMonomials(f["elem"]);
			    end_if;
			end_if;
			if reduced then break end_if;
		    end_for;
		    if reduced then break end_if;
		end_for;
	    until (dom::iszero(f["elems"])) or not reduced end_repeat;
	    if co_f then
		return([f, c_l]);
	    else
		return(f);
	    end_if;
	end_proc;			

    involutiveAutoreduce := proc(F:Type::ListOf(dom), ID:DOM_DOMAIN)
	local n:DOM_INT, G:DOM_LIST, g:dom, f:dom, reduced:DOM_BOOL, tmp, hr_f:DOM_BOOL,
	    DOM:DOM_DOMAIN;
	begin
	    if args(0) < 1 or args(0) > 3 then
		N_ERR;
	    end_if;
	    // F := sort(F, dom::elemLess);
	    DOM := dom;
	    tmp := map(F, l -> DOM::leader(l));
	    if args(0) = 3 and args(3) = hold(Head) then
		hr_f := TRUE;
		F := zip(F, tmp, (l1, l2) -> table("elem" = l1, "mons" = [l2], "lmon" = l2));
	    else
		hr_f := FALSE;
		F := zip(F, tmp, (l1, l2) -> table("elem" = l1, "mons" = DOM::allMonomials(l1),
		    "lmon" = l2));
	    end_if;
	    F := sort(F, (l1, l2) -> DOM::exponentLess(l1["lmon"][2], l2["lmon"][2]));
	    repeat
		reduced := FALSE;
		n := nops(F) + 1;
		repeat
		    n := n - 1;
		    G := F;
		    f := F[n];
		    delete G[n];
		    g := dom::internalInvolutiveNForm(f, G, 
			_list2table(ID::sepListMult(map(G, l -> l["lmon"][2]))), ID, hr_f);
		    reduced := bool(f <> g);
		until (n = 1 or reduced) end_repeat;
		if reduced then
		    delete F[n];
		    if not iszero(g["elem"]) then
			F := listlib::insert(F, g, 
			    (l1, l2) -> DOM::exponentLess(l1["lmon"][2], l2["lmon"][2]));
		    end_if;
		end_if;
	    until not reduced end_repeat;
	    map(F, l -> l["elem"]);
	end_proc;    
	    
    involutiveComplete := proc(F:Type::ListOf(dom), L)
	local DOM:DOM_DOMAIN, T:DOM_LIST, Q:DOM_LIST, sepm_t:DOM_TABLE, sepnm_t:DOM_TABLE,  
	    criterion:DOM_PROC, g, h, hr_f:DOM_BOOL, tmp, j, k:DOM_INT, noMore:DOM_BOOL, i, p_f;
	begin
	    // argument checks 
	    if args(0) > 4 or args(0) < 2 then
		N_ERR;
	    end_if;
	    hr_f := FALSE;
	    p_f := FALSE;
	    if args(0) > 2 then
		if contains({args(3..args(0))}, hold(Head)) then
		    hr_f := TRUE;
		end_if;
		if contains({args(3..args(0))}, hold(Output)) then
		    p_f := TRUE;
		end_if;
	    end_if;
	    if (domtype(L) <> DOM_DOMAIN
		or L::hasProp = FAIL
		or not L::hasProp(Cat::InvolutiveDivision))
	    then
		A_ERR;
	    end_if;
	    DOM := dom;
	    // Buchberger's chain (resp. second) criterion
	    criterion := proc(g)
		local f;
		begin
		    for f in T do
			if (not iszero(L::isInvDivisor(f["lmon"][2], g["lmon"][2], 
			    sepm_t[f["lmon"][2]]))
			    and DOM::exponentLess(zip(g["crit"], f["crit"], 
				(l1, l2) -> max(l1, l2)), g["lmon"][2]))
			then
			    if p_f then
				print(Unquoted, "The C R I T E R I O N has been invoked");
			    end_if;
			    return(TRUE);
			end_if;
		    end_for;
		    FALSE;
		end_proc;
	    // autoreduce F. The result is an ordered list.
	    F := dom::autoreduce(F);
	    // set up T and Q; doneNMP_t lists all nonmultiplicative prolongations
	    // carried out so far.
	    // Data structures: row - "elem": element represented by the row
	    //			      "mons": all monomials with respect to which the row
	    //			              may be reduced
	    // 			      "lmon": leading monomial
	    //			      "crit": multi index used for Buchberger criterion
	    //			      
	    //		        (monomial = [coefficient, multi index])
	    //			
	    // T - [row], Q - [row], doneNMP - (row["elem"] - {int})						
	    // sepm_t contains the multiplicative variables for the elements in T.
	    F := map(F, l -> table("elem" = l, "lmon" = DOM::leader(l), "nmp" = {}));
	    for k from 1 to nops(F) do
		F[k]["crit"] := F[k]["lmon"][2];
	    end_for;  
	    T := [F[1]];
	    Q := [F[i] $ i = 2..nops(F)];
	    sepm_t := _list2table(L::sepListMult([T[1]["lmon"][2]]));
	    repeat
		if p_f then
		    print(Unquoted, "=======================");
		    print(Unquoted, "recomputing T and Q ...");
		    print(Unquoted, "=======================");
		end_if;
		h := table("elem" = dom::zero);
		while Q <> [] and dom::iszero(h["elem"]) do
		    g := Q[1];
		    delete Q[1];
		    if not criterion(g) then
			if not contains(g, "mons") then
			    if hr_f then
				g["mons"] := [g["lmon"]];
			    else
				g["mons"] := dom::allMonomials(g["elem"]);
			    end_if;
			end_if;
			h := dom::internalInvolutiveNForm(g, T, sepm_t, L, hr_f);
		    end_if    
		end_while;
		if not dom::iszero(h["elem"]) then
		    if g["elem"] <> h["elem"] then 
		        h["crit"] := h["lmon"][2];
			h["nmp"] := {};
			tmp := [];
			for k from nops(T) downto 1 do
			    if dom::exponentLess(h["lmon"][2], T[k]["lmon"][2]) then
				tmp := append(tmp, T[k]);
				delete T[k];
			    else
				break;
			    end_if;
			end_for;
			if tmp <> [] then
			    Q := listlib::singleMerge(tmp, Q, 
				(l1, l2) -> (DOM::exponentLess(l1["lmon"][2], l2["lmon"][2])));
			end_if;
			T := append(T, h);
		    else
			T := append(T, g);
		    end_if;	
		end_if;
		if p_f then
		    print(Unquoted, "T: " , map(T, l -> l["elem"]));
		    print(Unquoted, "Q: " , map(Q, l -> l["elem"]));			
		    print(Unquoted, "=======================");
		end_if;
		noMore := FALSE;
		repeat 
		    if p_f then
			print(Unquoted, "T: " , map(T, l -> l["elem"]));
			print(Unquoted, "-----------------------");
		    end_if;			
		    sepm_t := _list2table(L::sepListMult([T[i]["lmon"][2] $ i = 1..nops(T)]));
		    sepnm_t := map(sepm_t, l -> ({i $ i = 1..DOM::numVars} minus l));
		    if p_f then
			print(Unquoted, ".......................");
			print(Unquoted, "non-multiplicative variables: " , sepnm_t);
			print(Unquoted, ".......................");
		    end_if;
		    tmp := sort([[T[i], subsop(T[i]["lmon"][2], j = T[i]["lmon"][2][j] + 1), j, 
			i] $ j in (sepnm_t[T[i]["lmon"][2]] minus T[i]["nmp"])
		    	$ i = 1..nops(T)], (l1, l2) -> (DOM::exponentLess(l1[2], l2[2])));
		    if (tmp = [] or 
			(Q <> [] and not dom::exponentLess(tmp[1][2], Q[1]["lmon"][2]))) 
		    then
			noMore := TRUE;
		    else
			if p_f then
			    print(Unquoted, "prolonging " , tmp[1][1]["elem"] , " w.r.t " 
			    , tmp[1][3]);
			end_if;         
			T[tmp[1][4]]["nmp"] := T[tmp[1][4]]["nmp"] union {tmp[1][3]};
			g := dom::exponent2monomial(subsop([0 $ dom::numVars], tmp[1][3] = 1)) 
			    * tmp[1][1]["elem"];
			if p_f then
			    print(Unquoted, "yielding " , g); 
			end_if;
			g := table("elem" = g, "lmon" = dom::leader(g), 
			    "crit" = tmp[1][1]["lmon"][2]);
			if criterion(g) then
			    h := table("elem" = dom::zero);
			else
			    if hr_f then
				g["mons"] := [g["lmon"]];
			    else
				g["mons"] := dom::allMonomials(g["elem"]);
			    end_if;
			    h := dom::internalInvolutiveNForm(g, T, sepm_t, L, hr_f);
			end_if;
			if p_f then
			    print(Unquoted, "reducing to " , h["elem"]);
			    print(Unquoted, "-----------------------"); 
			end_if;
			if not dom::iszero(h["elem"]) then
			    h["nmp"] := {};
			    if h["elem"] <> g["elem"] then 
				h["crit"] := h["lmon"][2];				
				tmp := [];
				for k from nops(T) downto 1 do
				    if dom::exponentLess(h["lmon"][2], T[k]["lmon"][2]) then
					tmp := append(tmp, T[k]);
					delete T[k];
				    else
					break;
				    end_if;
				end_for;
				T := append(T, h);
				if tmp <> [] then
				    Q := listlib::singleMerge(tmp, Q, 
				    (l1, l2) -> (DOM::exponentLess(l1["lmon"][2], l2["lmon"][2])));
				end_if;
			    else  // no reduction
				h["crit"] := g["crit"];
				T := append(T, h);
			    end_if;
			end_if; 
		    end_if;
		until noMore end_repeat;
	    until Q = [] end_repeat;
	    map(T, l -> DOM::multcoeffs(l["elem"], 1 / l["lmon"][1]));
	end_proc;
	
    obstructions2Involution := proc(F:Type::ListOf(dom), ID:DOM_DOMAIN)
	local nmp_t:DOM_TABLE, tmp:DOM_LIST, f, v;
	begin
	    nmp_t := _list2table(ID::sepListNonMult(map(map(F, dom::leader), op, 2)));
	    tmp := [];
	    for f in F do
		for v in nmp_t[dom::leader(f)[2]] do
		    if not dom::iszero(dom::involutiveNForm(
			dom::convert(dom::variables[v]) * f, F, ID)) 
		    then
			tmp := append(tmp, [f, dom::variables[v]]);
		    end_if;
		end_for;
	    end_for;
	    tmp;
	end_proc;
		
    begin    
    if args(0) <> 1 then error("wrong no of args") end_if;
    if not R::hasProp(Cat::Field) then
	error("not a field")
    end_if;
    
    _list2table := (l -> table(op(map(l, l1 -> l1[1] = l1[2]))));
    
end_category:
    
unalias(A_ERR, N_ERR):
