/*****************************************************************************
  $Id: FreeModulePoly.mu,v 1.20 2004/09/08 12:37:04 hivert Exp $
  File:		lib/DOMAINS/DOMAIN/FreeModulePoly.mu
  Authors:	Florent Hivert <florent.hivert@univ-mlv.fr>
  Created:	10/01/2003
  License:	LGPL
*****************************************************************************/

/**
    Dom::FreeModulePoly(R,Basis) -- R-Free module over the basis Basis

    R: ring
    Basis: domain

    For now, R needs to be commutative.

    This is an alternative implementation of Dom::FreeModule using Polynomials.
    TODO: Mimic SparseMatrix for handling the coefficient ring R.
**/

alias(rankers     = combinat::rankers):
alias(REP(x)      = extop(x,1)):
alias(REPCOEFF(x) =
      %if coeff(poly(0,[hold(x)])) = null() then
	  [coeff(REP(x))];
      else
	  if x=dom::zero then
	      null();
	  else
	      [coeff(REP(x))];
	  end_if;
      end_if);


domain Dom::FreeModulePoly
    (R: DOM_DOMAIN,
     BasisIndices: DOM_DOMAIN)

    local Options,
	  N, Rank, Unrank,
        RpolyOne,
	R2polyCoerce,    // Coerce : R external -> R internal
	Rcoerce,         // Coerce : R internal -> R external
	Rzero,           // internal rep of the 0 of R (<> R::zero)
	Rone,            // internal rep of the 1 of R (<> R::one)
	Rcharacteristic, // characteristic of R
	PolyVariables,   // Variables for internal polynomials
	PolyR;           // Ring for internal polynomials

    inherits Dom::BaseDomain;
  
    category Cat::ModuleWithBasis(R);
  
    info := "Dom::FreeModuleTable: domain for free modules, implemented with polynomials";

    interfaceAdd := {
		     hold(plus), hold(plus2)
		    };

    //////////////////////////////////////////////////////////////////////////
    // These are needed to buid optimized tensor products
    //////////////////////////////////////////////////////////////////////////
    getRankerParameters := 
    proc() begin [N, Rank, Unrank, PolyVariables]; end_proc;

    // The following is only needed by the sub domain TensorProductOfFreeModule. 
    // But I don't know to access to it in a simpler method. 
    getPolyRingParameters := 
    proc() begin [PolyR, R2polyCoerce, Rcoerce]; end_proc;

    ranker      := if N = 1 then [Rank, Unrank]; end_if;
    rankBasis   := if N = 1 then Rank; end_if;
    unrankBasis := if N = 1 then Unrank; end_if;

    //////////////////////////////////////////////////////////////////////////
    // Conversions and overloading
    //////////////////////////////////////////////////////////////////////////

    _plus := operators::_plus;
    _mult := operators::_mult;

    /////////////////////////////////////////////////////////////////////////
    // Basic information on the free module
    /////////////////////////////////////////////////////////////////////////

    basisIndices       := BasisIndices;

    /////////////////////////////////////////////////////////////////////////
    // Constructors and constants
    /////////////////////////////////////////////////////////////////////////

    // Unused ... For compatibility with standard mupad domains ?????
//     generate := 
//     proc(x: DOM_) : dom
//     begin 
// 	new(dom, x); 
//     end_proc;

    term := // called many time, trying to optimize 
    proc(support: dom::basisIndices) : dom
    begin
	new(dom, poly([[RpolyOne, Rank(support)]], PolyVariables, PolyR));
    end_proc;

    monomial :=
    proc(c: R, support: dom::basisIndices) : dom
    begin
	new(dom, poly([[R2polyCoerce(c), Rank(support)]], PolyVariables, PolyR));
    end_proc;

    zero := new(dom, poly(R2polyCoerce(0), PolyVariables, PolyR));

    iszero :=
    proc(x: dom) : DOM_BOOL
    begin
	bool(iszero(REP(x))); 
    end_proc;

    /////////////////////////////////////////////////////////////////////////
    // Access to terms, leading terms, monomials, coefficients, ...
    /////////////////////////////////////////////////////////////////////////

    nterms :=
    proc(x: dom)
    begin
	nterms(REP(x));
    end_proc;

    toList :=
    if Unrank = id then
	proc(x: dom)
	    : Type::ListOf(Type::ListProduct(dom::coeffRing, dom::basisIndices))
	begin
	    map(poly2list(REP(x)), t -> [Rcoerce(t[1]), t[2]]);
	end_proc;
    else
	proc(x: dom)
	    : Type::ListOf(Type::ListProduct(dom::coeffRing, dom::basisIndices))
	begin
	    map(poly2list(REP(x)), t -> [Rcoerce(t[1]), Unrank(t[2])]);
	end_proc;
    end_if;

    fromList :=
    if Rank = id then
	proc(l: Type::ListOf(Type::ListProduct(dom::coeffRing, dom::basisIndices)))
	    : dom
	begin
	    new(dom, poly(map(l, t -> [R2polyCoerce(t[1]), t[2]]), 
			  PolyVariables, PolyR));
	end_proc;
    else
	proc(l: Type::ListOf(Type::ListProduct(dom::coeffRing, dom::basisIndices)))
	    : dom
	begin
	    new(dom, poly(map(l, t -> [R2polyCoerce(t[1]), Rank(t[2])]),
			  PolyVariables, PolyR));
	end_proc;
    end_if;

    support :=
    proc(x:dom) : Type::ListOf(dom::basisIndices)
    begin
	map(poly2list(REP(x)), Unrank@op, 2);
    end_proc;

    // Note: multivariate coeff(poly(x*y,[x,y]), [1,1])) is only implemented
    // in the kernel since may 2003, in 3.0.0 devel; with previous versions
    // we use the select hack proposed by Christopher
    coeff :=
    if N=1 or traperror(coeff(poly(hold(x)*hold(y),[hold(x),hold(y)]), [1,1])) = 0 then
	proc(x: dom, s=FAIL: Type::Union(DOM_FAIL, dom::basisIndices))
	    local i;
	begin
	    if args(0)=2 then
		Rcoerce(coeff(REP(x), Rank(s)));
	    else
		Rcoerce(i) $ i in REPCOEFF(x);
	    end_if;
	end_proc;
    else
	proc(x: dom, s=FAIL: Type::Union(DOM_FAIL, dom::basisIndices))
	    local i;
	begin
	    if args(0)=2 then
		x := select(poly2list(REP(x)), has, [Rank(s)]);
		if nops(x) = 0 then
		    R::zero;
		else
		    Rcoerce(op(x,[1,1]));
		end_if;
	    else
		Rcoerce(i) $ i in REPCOEFF(x);
	    end_if;
	end_proc;
    end_if;

    // nthcoeff, nthsupport, nthterm, nthmonomial

    nthcoeff :=
    proc(x: dom, n: Type::PosInt) : Type::Union(DOM_FAIL,dom::coeffRing)
    begin
	Rcoerce(nthcoeff(REP(x), n));
    end_proc;

    nthsupport :=
    proc(x: dom, n: Type::PosInt) : Type::Union(DOM_FAIL,dom::basisIndices)
    begin
	x := nthterm(REP(x), n);
	if x=FAIL then
	    FAIL
	else
	    Unrank(op(poly2list(x), [1,2]));
	end_if;
    end_proc;

    nthterm :=
    proc(x: dom, n: Type::PosInt) : Type::Union(DOM_FAIL, dom)
    begin
	x := nthterm(REP(x), n);
	if x=FAIL then
	    FAIL
	else
	    new(dom, x);
	end_if;
    end_proc;

    nthmonomial :=
    proc(x: dom, n: Type::PosInt) : Type::Union(DOM_FAIL, dom)
    begin
	x := nthmonomial(REP(x), n);
	if x=FAIL then
	    FAIL
	else
	    new(dom, x);
	end_if;
    end_proc;

/*--
    mapcoeffs ( x, f, [args] )

    map the function f on all the coefficients of x. f must return a
    valid coefficient! As with the usual map function, [args] are
    passed as further arguments for f.
--*/

  // Beware that mapcoeffs on DOM_POLY uses syntactical equality to
  // test for zero coefficients. See MuPAD Bug #1402.

  mapcoeffs :=
    proc(x: dom, f: Type::AnyType)
	local options;
    begin
	options := args(3..args(0));
	new(dom, mapcoeffs(REP(x), z-> R2polyCoerce(f(Rcoerce(z),  options)) ));
    end_proc;

/*--
    multcoeffs (x, r)

    Multiply the coefficients of the terms of x by r.
--*/

    multcoeffs :=
    proc(x: dom, r: R)
    begin
	new(dom, multcoeffs(REP(x), R2polyCoerce(r)));
    end_proc;

    // Note: when left and right scalar multiplication coincide,
    // multcoeffsLeft could be implemented in a more efficient
    // way. Should we define an axiom for this ?

/**
 * mapterms ( x, f, [args] )

 * Map the function f on all the terms (elements of the support) of
 * x. The function f must always return a valid term! no check is
 * done. As with the usual map function, [args] are passed as further
 * arguments for f.
 **/

    mapterms :=
    proc(x: dom, f)
	local options;
    begin
	// TODO : Optimize this by using reverse poly2list
	options := args(3..args(0));
        new(dom, poly(map(poly2list(REP(x)),
			  t -> [t[1], Rank(f(Unrank(t[2]), options))]),
		      PolyVariables, PolyR));
    end_proc;

    /////////////////////////////////////////////////////////////////////////
    // Printing :
    // The following function is by default inherited from BaseDomain
    // we inherit from the category
    /////////////////////////////////////////////////////////////////////////

    TeX := DomainConstructor::getEntryFromCategory
	 (dom, Cat::ModuleWithBasis(R), "TeX");

    /////////////////////////////////////////////////////////////////////////
    // Arithmetic operations
    /////////////////////////////////////////////////////////////////////////

    plus2 := // called many time, trying to optimize
    proc(x : dom, y: dom)
    begin
	new(dom, _plus(REP(x), REP(y)));
    end_proc;

    plus :=
    proc()
	local i;
    begin
	if args(0)=0 then return(dom::zero); end_if;
	if testargs() then
	    for i from 1 to args(0) do
		if not testtype(args(i), dom) then
		    error("Wrong type of ".i.". argument (type '".expr2text(dom)."' expected,\ngot argument '".expr2text(args(i)).");");
		end_if;
	    end_for;
	end_if;
	new(dom, _plus(REP(args(i)) $ i = 1..args(0)));
    end_proc;

    // Added _subtract and _negate to work with the double inheritance
    // problem with Cat:AbelianGroup see mailing list (23 Feb 2004)
    _subtract := proc(x,y) begin dom::_plus(x, _negate(y)) end_proc;
			       
    _negate := // much better that de default version of Cat::ModuleWithBasis. 
    proc(x: dom)
    begin
	new(dom, _negate(REP(x)));
    end_proc;

    //////////////////////////////////////////////////////////////////////////
    // Interface with linalg / Dom::/*Sparse*/Matrix
    //////////////////////////////////////////////////////////////////////////

    // for backward compatibility
    sparseMatrix := dom::sparseVectorsDomain;

    // Optimized implementations of sparseVector / fromSparseVector
    // when N=1 and no ranker is provided
    sparseVectorCatFreeModule :=
        DomainConstructor::getEntryFromCategory
        (dom, Cat::ModuleWithBasis(R), "sparseVector");

    /// See specifications in Cat::ModuleWithBasis
    sparseVector :=
    if N=1 and PolyVariables[1] = hold(_X) then
    proc(p: dom,
	 rank=FAIL: Type::Union(DOM_FAIL,
				rankers::typeRanker, Type::ListOf(dom::basisIndices)))
	: dom::sparseVectorsDomain
    begin
	if args(0)=1 then
	    p := REP(p);
	   // Warning: this relies on the internal structure of sparse matrices!
	    new(dom::sparseVectorsDomain, if dom::dimension <> FAIL then
					      dom::dimension()
					  else
					      degree(p)
					  end, 1, [p], "FLAG");
	else
	    dom::sparseVectorCatFreeModule(args());
	end_if;
    end_proc;
    end_if;

// Cant be reused due to the bug of FreeModules with IntegerMod Coefficients. 
//     fromSparseVectorCatFreeModule :=
//         DomainConstructor::getEntryFromCategory
// 	(dom, Cat::ModuleWithBasis(R), "fromSparseVector");

    fromSparseVector :=
    if N=1 and PolyVariables[1] = hold(_X) then
    proc(vector: dom::sparseVectorsDomain,
	 unrank = dom::unrankBasis : 
	   Type::Union(rankers::typeUnranker, Type::ListOf(dom::basisIndices)))
	: dom
    begin
	if args(0)=1 then
	    return(new(dom, op(extop(vector,3),1)));
	elif domtype(unrank) = DOM_LIST then
	    unrank := rankers::unrankFromList(unrank);
	end_if;
	dom::fromList(map(poly2list(op(extop(vector,3),1)),
			  monomial -> [Rcoerce(op(monomial,1)),unrank(op(monomial,2))]));
    end_proc;
    end_if;

    //////////////////////////////////////////////////////////////////////////
    // Tools to build operators
    //////////////////////////////////////////////////////////////////////////

  makeLinearFunction :=
    proc(f)
    /**
        f is a function from dom::basisIndices to some module over
	dom::coeffRing, where the multiplication by scalars can
	be done via multcoeffs

	Returns the linear extension of f to dom;

	Extra arguments of the constructed operator are passed down to f.
    **/
        option escape;
    begin
        proc(x: dom)
	    local s;
	begin
	    _plus(multcoeffs(f(Unrank(op(s,2)), args(2..args(0))),
			     op(s,1))
		  $ s in poly2list(REP(x)));
	end_proc;
    end_proc;

  makeBilinearFunction :=
    proc(f)
    /**
        f is a function from dom::basisIndices x dom::basisIndices to some module over
	dom::coeffRing, where the multiplication by scalars can
	be done via multcoeffs

	Returns the linear extension of f to dom;

	Extra arguments of the constructed operator are passed down to f.

	Has it ever been tested ?
    **/
	option escape;
    begin
        proc(x1: dom, x2: dom)
	    local s1, s2;
	begin
	    _plus(multcoeffs(f(Unrank(op(s1,1)),
			       Unrank(op(s2,1)), args(3..args(0))),
			       op(s1,2) *op(s2,2))
		  $ s1 in poly2list(REP(x1))
		  $ s2 in poly2list(REP(x2)));
	end_proc;
    end_proc;

    initDomainAdd :=
    proc()
    begin
    end_proc;

begin
    Options := combinat::getOptions
	     ([hold(Variables), hold(RankerBasis), hold(MultirankerBasis)],
	      [hold(Variables)        = Type::Union(Type::PosInt, DOM_LIST),
	       hold(RankerBasis)      = Type::ListProduct(Type::AnyType $2),
	       hold(MultirankerBasis) = Type::ListProduct(Type::AnyType $2)
	      ],
	      2, Error);

    if contains(Options, hold(Variables)) then
	PolyVariables := Options[hold(Variables)];
	if domtype(PolyVariables) = DOM_INT then
	    PolyVariables := [hold(_X).i $ i=1..PolyVariables];
	end_if;
    else
	PolyVariables := [hold(_X)];
    end_if;
    N := nops(PolyVariables);

    // It's not safe to do this in initDomainAdd: that will indeed
    // happen after the initDomainAdd of the categories, and those
    // may want to create domains elements (like Cat::AlgebraWithBasis
    // which will create dom::one to define the conversion from the
    // coefficient ring)

    if N=1 then
	if contains(Options, hold(MultirankerBasis)) then
	    [Rank, Unrank] := Options[hold(MultirankerBasis)];
	    Rank   := op@Rank;
	    Unrank := Unrank@DOM_LIST;
	elif contains(Options, hold(RankerBasis)) then
	    [Rank, Unrank] := Options[hold(RankerBasis)];
//	elif dom::multirankBasis <> FAIL then
//	    Rank   := op@dom::multirankBasis;
//	    Unrank := dom::multiunrankBasis@DOM_LIST;
	else
	    // Probably safer if this coincides with the default
	    // definition of rankBasis from Cat::ModuleWithBasis
	    [Rank, Unrank] := combinat::rankers::fromType(BasisIndices);
	end_if;
    else
	if contains(Options, hold(MultirankerBasis)) then
	    [Rank, Unrank] := Options[hold(MultirankerBasis)];
	else
	    error("No basis ranker provided; please use option MultirankerBasis");
	end_if;
    end_if;

    //*******************************************************************//
    //         Taken from SparseMatrix to ensure compatibility with it   //
    //*******************************************************************//
    if R <> Dom::ExpressionField() and R::hasProp( Cat::Rng ) <> TRUE then
        error("coefficients must be from a Cat::Rng")
    end_if;

    // Default case: the internal polynomials use the
    // same ring as the sparse matrix. 
    PolyR := R; 
    Rzero := R::zero;
    Rone  := R::one;
    // Exceptions (for speed):
    //  matrix ring = Dom::ExpressionField() --> poly ring = Expr
    //  matrix ring = Dom::IntegerMod(n)     --> poly ring = IntMod(m)
    // These exceptional cases are handled below

    // coefficient domain R is passed to polynomials for internal
    // representation. Replace Dom::ExpressionField()/Dom::IntegerMod
    // by their analogues Expr/IntMod for faster polynomial arithmetic:    
    Rcharacteristic:= R::characteristic;
    if Rcharacteristic <> FAIL and 
       Rcharacteristic > 0 and
       R = Dom::IntegerMod(Rcharacteristic) then
         PolyR:= IntMod(Rcharacteristic);
         Rzero:= 0;
         Rone := 1;
    end_if;
    if R::hasProp(Ax::systemRep) then
         PolyR:= Expr;
         Rzero:= R::zero:
         Rone := R::one;
    end_if;

    //----------------------------------------------------------
    // Rcoerce is used in 2 ways:
    // 1) coercing input (but no further coercion to poly, this
    //                    is then done by R2polyCoerce)
    // 2) coercing internal poly representation to output
    // Treat the most important case R = Dom::ExpressionField
    // as a special case for efficiency:
    //----------------------------------------------------------
    Rcoerce:=
      if R = Dom::ExpressionField()
      then id  // no coercion necessary
      else R::coerce 
      end_if;
    //----------------------------------------------------------
    // coercing input (already converted by Rcoerce)
    // further to internal poly representation.
    //----------------------------------------------------------
    R2polyCoerce:=
      if PolyR = IntMod(Rcharacteristic)
      then expr; // poly(.., IntMod(n)) does not
                 // accept Dom::IntegerMod(n) coeffs ;-(
      else id; // all other cases were converted by Rcoerce.
               // Now it is up to poly to accept and
               // handle these coefficients
      end_if;
    //----------------------------------------------------------
    RpolyOne := R2polyCoerce(R::one);

end_domain:

unalias(REP):
