/* Dom::SolvableModule(A, mdim) - Domain of free modules over a solvable
    algebra of polynomial type

Author: Marcus Hausdorf, Uni Mannheim
        (hausdorf@math.uni-mannheim.de)
Created: 	Mon Apr  7 15:06:19 2003
Last update: 	Thu Sep 18 12:55:31 2003
*/

/*
Dom::SolvableModule(A, G) represents finitely generated free modules over 
polynomial algebras of solvable type, i.e. elements of A^|G|. The names 
of the generators are provided by G, either as a list of identifiers or
as an indexed identifier. A must be contained in Cat::SolvableAlgebra.
The representation is based on Dom::FreeModuleOld; especially all the rules 
for creating and converting elements apply here.

The main motivation of this domain is to provide a very general setting 
for involutive bases methods. The core of the domain consists of the method
involutiveComplete which computes an involutive basis for the ideal a given 
set of generators spans with resepct to an involutive division from 
Cat::InvolutiveDivision. The algorithm used is a slightly modified version
of the one used in Cat::SolvableAlgebra in order to cope with the different
data structures. 

For SolvableModules, the 'leading information' looks like this: 
 
    [leading coefficient, [leading multi index, leading generator]]
    
					    =
					    
				     leading exponent    
*/

				    

alias(A_ERR = error("wrong type of arguments"),
      N_ERR = error("wrong number of arguments"),
      REP(x) = extop(x, 1));

domain Dom::SolvableModule(A, mdim:Type::PosInt)
local mOrder, i, spos_t, _list2table, G;
inherits Dom::FreeModuleOld(A, G);
category Cat::LeftModule(A), 
	 (if A::hasProp(Cat::Field) then Cat::VectorSpace(A) end_if);
axiom (if A::hasProp(Ax::canonicalRep) then Ax::canonicalRep end_if),
      (if A::hasProp(Ax::normalRep) then Ax::normalRep end_if);

//===============
// "past" methods
//===============

exponentOrder := if dom::moduleOrder = "top" then
    proc(mu1, mu2):DOM_INT
    local tmp;
    begin
	tmp := A::exponentOrder(mu1[1], mu2[1]);
	if iszero(tmp) then
	    if mu1[2] > mu2[2] then
		return(1);
	    elif mu1[2] < mu2[2] then
		return(-1);
	    else
		return(0);
	    end_if;
	else
	    return(tmp);
	end_if;
    end_proc;
else
    proc(mu1, mu2):DOM_INT
    begin
	if mu1[2] > mu2[2] then
	    return(1);
	elif mu1[2] < mu2[2] then
	    return(-1);
	else
	    return(A::exponentOrder(mu1[1], mu2[1]));
	end_if;
    end_proc
end_if;

exponentLess := ((l1, l2) -> (bool(dom::exponentOrder(l1, l2) = -1)));

leader := if dom::moduleOrder = "top" then
    proc(x:dom)
    local le_l, tmp, k;
    begin
	if dom::iszero(x) then
	    return([A::zero, [[0 $ A::numVars], 1]]);
	end_if;
	le_l := map(map(REP(x), op, 1), A::leader);
	le_l := zip(le_l, REP(x), (l1, l2) -> [l1[1], [l1[2], spos_t[l2[2]]]]);
	tmp := le_l[1];
	for k from 2 to nops(le_l) do
	    if dom::exponentLess(tmp[2], le_l[k][2]) then
		tmp := le_l[k];
	    end_if;
	end_for;
	tmp;
    end_proc;
else
    proc(x:dom)
    local tmp, k;
    begin
	if dom::iszero(x) then
	    return([A::zero, [[0 $ A::numVars], 1]]);
	end_if;
	tmp := dom::elem2list(x);
	for k from 1 to nops(tmp) do
	    if not A::iszero(tmp[k]) then
		tmp := A::leader(tmp[k]);
		return([tmp[1], [tmp[2], k]]);
	    end_if;
	end_for;
    end_proc;
end_if;
    
exponent2monomial := (x -> new(dom, [[A::exponent2monomial(x[1]), G[x[2]]]]));

// monomial2exponent := (x -> dom::allMonomials(x)[1][2]);  

allMonomials := proc(x:dom)
    local tmp, i;
    begin
	if dom::iszero(x) then
	    return([]);
	end_if;
	tmp := map(REP(x), l -> ([A::allMonomials(l[1]), spos_t[l[2]]]));
	map(tmp, l -> ([l[1][i][1], [l[1][i][2], l[2]]] $ i = 1..nops(l[1])));
    end_proc;	
			
autoreduce := proc(F:Type::ListOf(dom))
    local n:DOM_INT, H:DOM_LIST, g:dom, f:dom, reduced:DOM_BOOL, tmp, hr_f:DOM_BOOL,
	DOM:DOM_DOMAIN;
    begin
	if args(0) < 1 or args(0) > 2 then
	    N_ERR;
	end_if;
	if nops(F) = 1 then return(F) end_if;
	DOM := dom;
	tmp := map(F, dom::leader);
	if args(0) = 2 and args(2) = hold(Head) then
	    hr_f := TRUE;
	    F := zip(F, tmp, (l1, l2) -> table("elem" = l1, "mons" = [l2], "lmon" = l2));
	else
	    hr_f := FALSE;
	    F := zip(F, tmp, (l1, l2) -> table("elem" = l1, "mons" = DOM::allMonomials(l1),
	        "lmon" = l2));
	end_if;
	F := sort(F, (l1, l2) -> DOM::exponentLess(l1["lmon"][2], l2["lmon"][2]));
	repeat
	    reduced := FALSE;
	    n := nops(F) + 1;
	    repeat
		n := n - 1;
		H := F;
		f := F[n];
		delete H[n];
		g := dom::internalNForm(f, H, hr_f);
		reduced := bool(f <> g);
	    until (n = 1 or reduced) end_repeat;
	    if reduced then
		delete F[n];
		if not iszero(g["elem"]) then
		    F := listlib::insert(F, g, 
			(l1, l2) -> DOM::exponentLess(l1["lmon"][2], l2["lmon"][2]));
		end_if;
	    end_if;
	until not reduced end_repeat;
	map(F, l -> l["elem"]);
    end_proc;

nForm := proc(f:dom, F:Type::ListOf(dom))
    local a1, a2, DOM:DOM_DOMAIN, res, o_t;
    begin
	o_t := prog::getOptions([hold(Head) = FALSE, hold(Coeffs) = FALSE], 
	    [hold(Head) = DOM_BOOL, hold(Coeffs) = DOM_BOOL], 2);
	if o_t[hold(Head)] then
	    a1 := table("elem" = f, "mons" = [dom::leader(f)]);
	else
	    a1 := table("elem" = f, "mons" = dom::allMonomials(f));
	end_if;
	DOM := dom;
	a2 := map(F, l -> table("elem" = l, "lmon" = DOM::leader(l)));
	res := dom::internalNForm(a1, a2, o_t[hold(Head)], o_t[hold(Coeffs)]);
	if o_t[hold(Coeffs)] then
	    return([res[1]["elem"], res[2]]);
	else
	    return([res["elem"]]);
	end_if;
    end_proc;
    
// nForm := proc(f:dom, F:Type::ListOf(dom))
//     local hr_f:DOM_BOOL, a1, a2, DOM:DOM_DOMAIN;
//     begin
// 	if args(0) < 2 or args(0) > 3 then
// 	    N_ERR;
// 	end_if;
// 	if args(0) = 3 and args(3) = hold(Head) then
// 	    hr_f := TRUE;
// 	    a1 := table("elem" = f, "mons" = [dom::leader(f)]);
// 	else
// 	    hr_f := FALSE;
// 	    a1 := table("elem" = f, "mons" = dom::allMonomials(f));
// 	end_if;
// 	DOM := dom;
// 	a2 := map(F, l -> table("elem" = l, "lmon" = DOM::leader(l)));
// 	(dom::internalNForm(a1, a2, hr_f))["elem"];
//     end_proc;

internalNForm := proc(f, G, hr_f, co_f = FALSE)
    local reduced:DOM_BOOL, t, d, isDivisor:DOM_PROC, c, tmp1, tmp2, c_l, k; 
    begin
	isDivisor := proc(mu1, mu2)
	local tmp, k;
	begin
	    if mu1[2] <> mu2[2] then
		return(0);
	    end_if;
	    tmp := zip(mu2[1], mu1[1], _subtract);
	    for k from 1 to nops(mu1[1]) do
		if tmp[k] < 0 then
		    return(0);
		end_if;
	    end_for;
	    tmp;
	end_proc;	    

	if co_f then 
	    c_l := [A::zero $ nops(G)];
	end_if;
	repeat
	    reduced := FALSE;
	    for k from 1 to nops(G) do
		for t in f["mons"] do
		    if not iszero((d := isDivisor(G[k]["lmon"][2], t[2]))) then
			reduced := TRUE;
			// f := dom::reduce(f["elem"], g["elem"], g["lmon"], t, d);
			tmp1 := A::exponent2monomial(d);
			tmp2 := dom::_mult(tmp1, G[k]["elem"]);
			c := (t[1]) / (dom::leader(tmp2)[1]);
			if co_f then
			    c_l[k] := c_l[k] + A::multcoeffs(tmp1, c);
			end_if;
			f := f["elem"] - dom::_mult(c, tmp2);
			f := table("elem" = f, "lmon" = dom::leader(f));
			if hr_f then
			    f["mons"] := [f["lmon"]];
			else
			    f["mons"] := dom::allMonomials(f["elem"]);
			end_if;
		    end_if;
		    if reduced then break end_if;
		end_for;
		if reduced then break end_if;
	    end_for;
	until (dom::iszero(f["elem"])) or not reduced end_repeat;
	if co_f then
	    return([f, c_l]);
	else
	    return(f);
	end_if;
    end_proc;			
    
//=======================
// involutive normal form
//=======================

// involutiveNForm(f, F, L) computes the involutive normal form of f modulo the set F 
// w.r.t. the involutive division L. The switch HeadReduce (default: FALSE) determines 
// whether full or only head reductions are used. 

// involutiveNForm := proc(f:dom, F:Type::ListOf(dom), L:DOM_DOMAIN)
//     local hr_f:DOM_BOOL, a1, a2, DOM:DOM_DOMAIN;
//     begin
// 	if args(0) < 3 or args(0) > 4 then
// 	    N_ERR;
// 	end_if;
// 	if args(0) = 4 and args(4) = hold(Head) then
// 	    hr_f := TRUE;
// 	    a1 := table("elem" = f, "mons" = [dom::leader(f)]);
// 	else
// 	    hr_f := FALSE;
// 	    a1 := table("elem" = f, "mons" = dom::allMonomials(f));
// 	end_if;
// 	DOM := dom;
// 	a2 := map(F, l -> table("elem" = l. "lmon" = DOM::leader(l)));
// 	(dom::internalInvolutiveNForm(a1, a2, [0 $ nops(G)], L, hr_f))[1]["elem"];
//    end_proc; 

involutiveNForm := proc(f:dom, F:Type::ListOf(dom), L:DOM_DOMAIN)
    local a1, a2, DOM:DOM_DOMAIN, res, o_t;
    begin
	o_t := prog::getOptions([hold(Head) = FALSE, hold(Coeffs) = FALSE], 
	    [hold(Head) = DOM_BOOL, hold(Coeffs) = DOM_BOOL], 3);
	if o_t[hold(Head)] then
	    a1 := table("elem" = f, "mons" = [dom::leader(f)]);
	else
	    a1 := table("elem" = f, "mons" = dom::allMonomials(f));
	end_if;
	DOM := dom;
	a2 := map(F, l -> table("elem" = l, "lmon" = DOM::leader(l)));
	res := dom::internalInvolutiveNForm(a1, a2, [0 $ nops(G)], L, 
	    o_t[hold(Head)], o_t[hold(Coeffs)]);
	if o_t[hold(Coeffs)] then
	    return([res[1]["elem"], res[3]]);
	else
	    return([res[1]["elem"]]);
	end_if;
    end_proc; 

internalInvolutiveNForm := proc(f, G, mvt_l, L, hr_f, co_f = FALSE)
    local reduced:DOM_BOOL, t, d, c, lpos, tmp1, tmp2, c_l, k;
    begin
	if co_f then
	    c_l := [A::zero $ nops(G)];
	end_if;
	repeat
	    for t in f["mons"] do
		lpos := t[2][2];
		if iszero(mvt_l[lpos]) then
		    mvt_l[lpos] := _list2table(L::sepListMult(
			map(select(G, l -> bool(l["lmon"][2][2] = lpos)), 
			    l2 -> l2["lmon"][2][1]))); 
		end_if;	
		reduced := FALSE;
		for k from 1 to nops(G) do
		    if G[k]["lmon"][2][2] <> lpos then next end_if;
		    if not iszero((d := L::isInvDivisor(G[k]["lmon"][2][1], t[2][1], 
			    mvt_l[lpos][G[k]["lmon"][2][1]]))) 
		    then
			reduced := TRUE;
			// f := dom::reduce(f["elem"], g["elem"], g["lmon"], t, d);
			tmp1 := A::exponent2monomial(d);
			tmp2 := dom::_mult(tmp1, G[k]["elem"]);
			c := (t[1]) / (dom::leader(tmp2)[1]);
			if co_f then
			    c_l[k] := c_l[k] + A::multcoeffs(tmp1, c);
			end_if;
			f := f["elem"] - dom::_mult(c, tmp2);
			f := table("elem" = f, 
			    "lmon" = dom::leader(f));
			if hr_f then
			    f["mons"] := [f["lmon"]];
			else
			    f["mons"] := dom::allMonomials(f["elem"]);
			end_if;
		    end_if;
		    if reduced then break end_if;
		end_for;
		if reduced then break end_if;
	    end_for;
	until (dom::iszero(f["elem"])) or not reduced end_repeat;
	if co_f then
	    return([f, mvt_l, c_l]);
	else
	    return([f, mvt_l]);
	end_if;
    end_proc;
    
// reduce := proc(f:dom, g:dom, mu, t)
//     local d, tmp, tmp2;
//     begin
// 	if args(0) = 5 then
// 	    d := args(5);
// 	else
// 	    if (mu[2][2] <> t[2][2]
// 		or select((d := zip(t[2][1], mu[2][1], _subtract)), 
// 		    l -> bool(l < 0)) <> 0)
// 	    then
// 		error("cannot reduce");
// 	    end_if;
// 	end_if;
// 	tmp := dom::_mult(A::exponent2monomial(d), g);
// 	tmp2 := dom::leader(tmp);
// 	dom::_mult(tmp2[1], f) - dom::_mult(t[1], tmp);
//     end_proc;
						
moduleOrder := mOrder;

involutiveAutoreduce := proc(F:Type::ListOf(dom), ID:DOM_DOMAIN)
    local n:DOM_INT, H:DOM_LIST, g:dom, f:dom, reduced:DOM_BOOL, tmp, hr_f:DOM_BOOL,
	DOM:DOM_DOMAIN;
    begin
	if args(0) < 1 or args(0) > 3 then
	    N_ERR;
	end_if;
	if nops(F) = 1 then return(F) end_if;
	DOM := dom;
	tmp := map(F, dom::leader);
	if args(0) = 3 and args(3) = hold(Head) then
	    hr_f := TRUE;
	    F := zip(F, tmp, (l1, l2) -> table("elem" = l1, "mons" = [l2], "lmon" = l2));
	else
	    hr_f := FALSE;
	    F := zip(F, tmp, (l1, l2) -> table("elem" = l1, "mons" = DOM::allMonomials(l1),
	        "lmon" = l2));
	end_if;
	F := sort(F, (l1, l2) -> DOM::exponentLess(l1["lmon"][2], l2["lmon"][2]));
	repeat
	    reduced := FALSE;
	    n := nops(F) + 1;
	    repeat
		n := n - 1;
		H := F;
		f := F[n];
		delete H[n];
		g := (dom::internalInvolutiveNForm(f, H, [0 $ nops(H)], ID, hr_f))[1];
		reduced := bool(f <> g);
	    until (n = 1 or reduced) end_repeat;
	    if reduced then
		delete F[n];
		if not iszero(g["elem"]) then
		    F := listlib::insert(F, g, 
			(l1, l2) -> DOM::exponentLess(l1["lmon"][2], l2["lmon"][2]));
		end_if;
	    end_if;
	until not reduced end_repeat;
	map(F, l -> l["elem"]);
    end_proc;
    
involutiveComplete := proc(F:Type::ListOf(dom), L)
    local DOM:DOM_DOMAIN, T:DOM_LIST, Q:DOM_LIST, criterion:DOM_PROC, g, h, 
	hr_f:DOM_BOOL, tmp, j, k:DOM_INT, noMore:DOM_BOOL, i, p_f, Tmv_l, Tnmv_l;
    begin
	// argument checks 
	if args(0) > 4 or args(0) < 2 then
	    N_ERR;
	end_if;
	hr_f := FALSE;
	p_f := FALSE;
	if args(0) > 2 then
	    if contains({args(3..args(0))}, hold(Head)) then
		hr_f := TRUE;
	    end_if;
	    if contains({args(3..args(0))}, hold(Output)) then
		p_f := TRUE;
	    end_if;
	end_if;
	if (domtype(L) <> DOM_DOMAIN
	    or L::hasProp = FAIL
	    or not L::hasProp(Cat::InvolutiveDivision))
	then
	    A_ERR;
	end_if;
	DOM := dom;
	// Buchberger's chain (resp. second) criterion
	criterion := proc(g)
	    local f, tmp;
	    begin
		for f in T do
		    tmp := f["lmon"][2][2];
		    if iszero(Tmv_l[tmp]) then
			Tmv_l[tmp] := _list2table(L::sepListMult(
			    map(select(T, l -> bool(l["lmon"][2][2] = tmp)), 
				l2 -> l2["lmon"][2][1])));
		    end_if;
		    if (not iszero(L::isInvDivisor(f["lmon"][2][1], g["lmon"][2][1], 
			Tmv_l[tmp][f["lmon"][2][1]]))
			and A::exponentLess(zip(g["crit"], f["crit"], 
			    (l1, l2) -> max(l1, l2)), g["lmon"][2][1]))
		    then
			if p_f then
			    print(Unquoted, "The C R I T E R I O N has been invoked");
			end_if;
			return(TRUE);
		    end_if;
		end_for;
		FALSE;
	    end_proc;
	// autoreduce F. The result is an ordered list.
	F := dom::autoreduce(F);
	// set up T and Q; doneNMP_t lists all nonmultiplicative prolongations
	// carried out so far.
	// Data structures: row - "elem": element represented by the row
	//			      "mons": all monomials with respect to which the row
	//			              may be reduced
	// 			      "lmon": leading monomial
	//			      "crit": multi index used for Buchberger criterion
	//			      
	//		        (monomial = [coefficient, multi index])
	//			
	// T - [row], Q - [row], doneNMP - (row["elem"] - {int})	    
	F := map(F, l -> table("elem" = l, "lmon" = DOM::leader(l), "nmp" = {}));
	for k from 1 to nops(F) do
	    F[k]["crit"] := F[k]["lmon"][2][1];
	end_for;
	T := [F[1]];
	Tmv_l := [0 $ nops(G)];
	Q := [F[i] $ i = 2..nops(F)];
	// sepm_t := _list2table(L::sepListMult([T[1]["lmon"][2]]));
	repeat
	    if p_f then
		print(Unquoted, "=======================");
		print(Unquoted, "recomputing T and Q ...");
		print(Unquoted, "=======================");
	    end_if;
	    h := table("elem" = dom::zero);
	    while Q <> [] and dom::iszero(h["elem"]) do
		g := Q[1];
		delete Q[1];
		// if not criterion(g) then
		    if not contains(g, "mons") then
			if hr_f then
			    g["mons"] := [g["lmon"]];
			else
			    g["mons"] := dom::allMonomials(g["elem"]);
			end_if;
		    end_if;
		    h := dom::internalInvolutiveNForm(g, T, Tmv_l, L, hr_f);
		    Tmv_l := h[2];
		    h := h[1];
		// end_if    
	    end_while;
	    if not dom::iszero(h["elem"]) then
		if g["elem"] <> h["elem"] then 
		    h["crit"] := h["lmon"][2][1];
		    h["nmp"] := {};
		    tmp := [];
		    for k from nops(T) downto 1 do
			if dom::exponentLess(h["lmon"][2], T[k]["lmon"][2]) then
			    tmp := append(tmp, T[k]);
			    Tmv_l[T[k]["lmon"][2][2]] := 0;
			    delete T[k];
			else
			    break;
			end_if;
		    end_for;
		    if tmp <> [] then
			Q := listlib::singleMerge(tmp, Q, 
			    (l1, l2) -> (DOM::exponentLess(l1["lmon"][2], l2["lmon"][2])));
		    end_if;
		    T := append(T, h);
		    Tmv_l[h["lmon"][2][2]] := 0;
		else
		    T := append(T, h);
		    Tmv_l[h["lmon"][2][2]] := 0;
		end_if;	
	    end_if;
	    if p_f then
		print(Unquoted, "T: " , map(T, l -> l["elem"]));
		print(Unquoted, "Q: " , map(Q, l -> l["elem"]));			
		print(Unquoted, "=======================");
	    end_if;
	    noMore := FALSE;
	    repeat 
		if p_f then
		    print(Unquoted, "T: " , map(T, l -> l["elem"]));
		    print(Unquoted, "-----------------------");
		end_if;			
		for k from 1 to nops(G) do
		    if iszero(Tmv_l[k]) then
			Tmv_l[k] := _list2table(L::sepListMult(
			    map(select(T, l -> bool(l["lmon"][2][2] = k)), 
				l2 -> l2["lmon"][2][1])));
		    end_if; 
		end_for;
		Tnmv_l := map(Tmv_l, l -> map(l, l2 -> {i $ i = 1..A::numVars} minus l2));
		if p_f then
		    print(Unquoted, ".......................");
		    print(Unquoted, "non-multiplicative variables: " , Tnmv_l);
		    print(Unquoted, ".......................");
		end_if;
		tmp := sort([[T[i], 
			      [subsop(T[i]["lmon"][2][1], j = T[i]["lmon"][2][1][j] + 1), T[i]["lmon"][2][2]], 
			      j, 
			      i] 
			  $ j in (Tnmv_l[T[i]["lmon"][2][2]][T[i]["lmon"][2][1]] minus T[i]["nmp"])
			  $ i = 1..nops(T)], 
		      (l1, l2) -> (DOM::exponentLess(l1[2], l2[2])));
		if (tmp = [] or 
		    (Q <> [] and not dom::exponentLess(tmp[1][2], Q[1]["lmon"][2]))) 
		then
		    noMore := TRUE;
		else
		    if p_f then
			print(Unquoted, "prolonging " , tmp[1][1]["elem"] , " w.r.t " 
			, tmp[1][3]);
		    end_if;         
		    T[tmp[1][4]]["nmp"] := T[tmp[1][4]]["nmp"] union {tmp[1][3]};
		    g := dom::_mult(A::exponent2monomial(subsop([0 $ A::numVars], tmp[1][3] = 1)), 
			tmp[1][1]["elem"]);
		    if p_f then
			print(Unquoted, "yielding " , g); 
		    end_if;
		    g := table("elem" = g, "lmon" = dom::leader(g), 
			"crit" = tmp[1][1]["lmon"][2][1]);
		    // if criterion(g) then
		    //	h := table("elem" = dom::zero);
		    // else
			if hr_f then
			    g["mons"] := [g["lmon"]];
			else
			    g["mons"] := dom::allMonomials(g["elem"]);
			end_if;
			h := dom::internalInvolutiveNForm(g, T, Tmv_l, L, hr_f);
			Tmv_l := h[2];
			h := h[1];
		    // end_if;
		    if p_f then
			print(Unquoted, "reducing to " , h["elem"]);
			print(Unquoted, "-----------------------"); 
		    end_if;
		    if not dom::iszero(h["elem"]) then
			h["nmp"] := {};
			if h["lmon"] <> g["lmon"] then 
			    h["crit"] := h["lmon"][2][1];				
			    tmp := [];
			    for k from nops(T) downto 1 do
				if dom::exponentLess(h["lmon"][2], T[k]["lmon"][2]) then
				    tmp := append(tmp, T[k]);
				    Tmv_l[T[k]["lmon"][2][2]] := 0;
				    delete T[k];
				else
				    break;
				end_if;
			    end_for;
			    T := append(T, h);
			    Tmv_l[h["lmon"][2][2]] := 0;
			    if tmp <> [] then
				Q := listlib::singleMerge(tmp, Q, 
				(l1, l2) -> (DOM::exponentLess(l1["lmon"][2], l2["lmon"][2])));
			    end_if;
			else  // no reduction
			    h["crit"] := g["crit"];
			    T := append(T, h);
			    Tmv_l[h["lmon"][2][2]] := 0;
			end_if;
		    end_if; 
		end_if;
	    until noMore end_repeat;
	until Q = [] end_repeat;
	map(T, l -> l["elem"]);
    end_proc;
    
//===============
// Input / Output
//===============

// conversion between representation of free modules and list representation

list2elem := proc(s_l:Type::ListOf(A, mdim, mdim)):dom
    local i;
    begin
	new(dom, select([[s_l[i], G[i]] $ i = 1..mdim], l -> not A::iszero(l[1])));
    end_proc;
    
elem2list := proc(x:dom):DOM_LIST
    begin
	subsop([A::zero $ mdim], op(map(REP(x), l -> _equal(spos_t[l[2]], l[1])))); 
    end_proc;
    
// new convert and expr methods

convert := proc(x)
    local tmp;
    begin
	if domtype(x) = DOM_LIST then
	    tmp := map(x, A::coerce);
	    if iszero(contains(tmp, FAIL)) 
		and nops(tmp) = mdim 
	    then
		return(dom::list2elem(tmp));
	    end_if;
	end_if;
	dom::moduleConvert(x);
    end_proc;

expr := (x -> map(dom::elem2list(x), A::expr));

dimension := mdim;
    
//===============
// body of domain
//===============

begin
    if args(0) < 2 or args(0) > 3 then
	N_ERR;
    end_if;
    // first argument must be in Cat::SolvableAlgebra
    if domtype(A) <> DOM_DOMAIN
	or A::hasProp = FAIL
	or not A::hasProp(Cat::SolvableAlgebra) 
    then
	A_ERR;
    end_if;
    // second must be positive integer
    G := [genident("_sm_") $ i = 1..mdim];
    // optional third argument denotes the module ordering
    // default: "top"
    if args(0) = 3 then
	if not contains ({"top", "pot"}, args(3)) then
	    A_ERR;
	else
	    mOrder := args(3);
	end_if;
    else
	mOrder := "top";
    end_if;
    spos_t := table((G[i] = i) $ i = 1..nops(G));
    _list2table := (l -> table(op(map(l, l1 -> l1[1] = l1[2]))));
end_domain:
