/**
  generate::isNeg  :  

  Domains with own representation which may contain signed values
  have to provide a isNeg method in order to produce good looking 
  outputs.  It order to check some operands of an element these
  methods may call this function.


  WARNING:  This is a library equivalent of an internal kernel
            function.  Changes have to be done in MIO_is_neg as
            well !!!
*/
generate::isNeg :=
proc(x) : DOM_BOOL
  local isneg, re, negate;
  option noDebug;
begin
  negate := x::dom::_negate;
  if (isneg := x::dom::isNeg) <> FAIL and negate <> FAIL then
    return(isneg(x));
  end_if;
  
  case domtype(x)
    of DOM_INT do
    of DOM_RAT do
    of DOM_FLOAT do
       return(bool(x<0));
       break;

    of DOM_COMPLEX do
       re := op(x,1);
       if not iszero(re) then
         return(bool(re<0));
       else
         return(bool(op(x,2)<0));
       end_if;
       break;

    of DOM_EXPR do
       case type(x) 
           of "_negate" do
               return(TRUE);
	       break;

           of "_plus" do
               return(_and(map(op(x), generate::isNeg)));
	       break;

	   of "_mult" do
	       if contains({DOM_COMPLEX, DOM_FLOAT, DOM_INT, DOM_RAT}, 
			   domtype(op(x, nops(x)))) then
		  return(generate::isNeg(op(x, nops(x))));
               end_if;
	       if contains({DOM_COMPLEX, DOM_FLOAT, DOM_INT, DOM_RAT}, 
			   domtype(op(x, 1))) then
		  return(generate::isNeg(op(x, 1)));
               end_if;
               return(FALSE);
               break;
       end_case;
   end_case;
   FALSE
end_proc:


/**
  generate::negate  :  


  If generate::isNeg returns TRUE for an expression, generate::negate
  can be used to get rid of the minus-sign.  In contrast to simply
  calling _negate, genrate::negate does not do any simplifications,
  which is necessary for the output of unsimplified expressions.

  WARNING:  This is a library equivalent of an internal kernel
            function.  Changes have to be done in MIO_negate as
            well !!!
*/
generate::negate :=
proc(ex)
  local len, value, res;
begin
  if domtype(ex) = DOM_COMPLEX then
    return(-ex)
  end_if;
  
  if type(ex) = "_mult" then
    len := nops(ex);
    
    if testtype(op(ex, len), Type::Numeric) then
      // negate last number
      value := (-1)*op(ex, len);
      if value = 1 then
        res := hold(_mult)(op(ex, 1..len-1));
      else
        res := hold(_mult)(op(ex, 1..len-1), value);
      end_if;
      if nops(res) = 1 then
        // product contains only one factor
        res := op(res);
      end_if;
      return(res);
    elif testtype(op(ex, 1), Type::Numeric) then
      // negate first number
      value := (-1)*op(ex, 1);
      if value = 1 then
        res := hold(_mult)(op(ex, 2..len));
      else
        res := hold(_mult)(value, op(ex, 2..len));
      end_if;
      if nops(res) = 1 then
        // product contains only one factor
        res := op(res);
      end_if;
      return(res);
    else
      return(hold(_mult)(op(ex), -1))
    end_if;
  end_if;
  
  if type(ex) = "_negate" then
    return(op(ex, 1))
  end_if;

  _negate(ex)
end_proc:
/*--
generate::sortSum  --  rearranges the succession of the summands in
                       a sum. The first summand is positive if
                       possible. 

generate::sortSum( SUM )

       SUM  --   a "_plus"- expression

Puts the first nonnegative summand in the first position of
the sum.

>> -a*b*3-c*d*(-2);

                              - 3 a b + 2 c d

>> generate::sortSum(%);

                               2 c d - 3 a b

This function is used by generate::TeX and by the
typesetting system for determining the output order of terms of a
_plus-expression.

--*/

generate::sortSum :=
proc(SuM)
  local i, list, lastOp, keepOrder, bases, p, prMonomial;
//  option noDebug;
begin
  if not testtype(SuM, "_plus") then
    return(SuM);
  end_if;
  if nops(SuM) = 1 then
    return(SuM);
  end_if;

  keepOrder := Pref::keepOrder();
  if keepOrder = hold(Always) or
     keepOrder = hold(DomainsOnly) and _domainOutputMode = TRUE then
    return(SuM)
  end;
    
  list := [op(SuM)];
  lastOp := op(list, nops(list));
  if domtype(lastOp) = DOM_COMPLEX and op(lastOp, 1) <> 0 then
    list := [op(SuM, 1..nops(list)-1), op(lastOp,1), I*op(lastOp,2)]
  end_if;
  if hastype(float(list), "_power") then
    // get all bases of powers
    bases := DOM_SET::sort(indets(map(select({op(map(list, op)), op(list)}, testtype, "_power"), op, 1), RatExpr) minus Type::ConstantIdents);
//    bases := DOM_SET::sort(indets(list) minus Type::ConstantIdents);
    if bases = [] or nops(SuM) = 2 and
      map({op(list)}, generate::isNeg) = {TRUE, FALSE} then
      p := FAIL
    else
      if expr(poly(SuM, bases)) = SuM then
        p := poly2list(hold(_plus)(op(list)), bases);
      else
        p := FAIL;
      end_if;
    end_if;
    if p <> FAIL and indets(map(p, op, 1)) intersect {op(bases)} = {} then
      prMonomial :=
      proc(m)
        local c, pow;
      begin
        if m[1] = 1 then
          c := null();
        else
          c := m[1]
        end_if;
        pow := null();
        for i from 1 to nops(m[2]) do
          case op(m[2], i)
            of 0 do
              // c*x^0
              break;
              
            of 1 do
              pow := pow, bases[i];
              break;
              
            otherwise
              pow := pow, hold(_power)(bases[i], op(m[2], i));
          end_case;
        end_for;
        if c = null() and pow = null() then
          c := 1;
        end_if;
        if nops([c, pow]) = 1 then
          c := c, pow;
          if type(c) = "_plus" then
            op(c)
          else
            c
          end_if
        else
          if testtype(c, Type::Numeric) then
            hold(_mult)(pow, c)
          elif type(c) = "_mult" then
            if testtype(op(c, nops(c)), Type::Numeric) then
               hold(_mult)(op(c, 1..nops(c)-1), pow, op(c, nops(c)))
           else
              hold(_mult)(op(c), pow)
            end_if
          else
            hold(_mult)(c, pow)
          end_if
        end_if;
      end:
      return(hold(_plus)(prMonomial(p[i]) $ i = 1..nops(p)))
    end_if
  end_if;
  
  if generate::isNeg( op(list,1)) then
    for i from 2 to nops(list) do
      if not generate::isNeg(op(list, i)) then
        return(hold(_plus) (op(subsop(list, 1=op(list, i), i=op(list, 1), Unsimplified))));
      end_if;
    end_for;
  end_if;
  hold(_plus)(op(list));
end_proc:

generate::sortSums :=
proc(ex)
  local ex2;
begin
  if ex::dom::sortSums <> FAIL then
    return(ex::dom::sortSums(ex))
  end_if;
  if Pref::keepOrder() = Always then ex
  else
    ex2 := misc::maprec(ex,
                        (x->testtype(x, "_plus"))
                        = generate::sortSum, Unsimplified);
    // Do not write '-a-b' as '-(a+b)' at top level.
    if op(ex2, 0) = hold(_mult)
      and nops(ex2) = 2
      and op(ex2, [1, 0]) = (_plus)
      and op(ex2, 2) = -1 then
      ex
    else
      ex2
    end_if;
  end_if
end_proc:

