/*--
generate::splitProduct  --  decomposes a product into a list
			    containing the sign, the numerator
			    and the denumerator

  generate::splitProduct( multExpr )

  multExpr --  an expression with type "_mult"

  splitProduct returns a list with the sign of the product
  (-1 or 1) and two lists with the numerator and the denum-
  erator. If the numerator is empty, a 1 is inserted.
  For the decision, if a factor has to be in the numerator 
  or not, exponents of powers are analysed. 
  Numerical factors (which are the last factors in the normal
  form) are splitted (rationals) and put in the first position.
  The result maybe used for output of "_mult" expressions.

  Examples:

>> generate::splitProduct( a / b);

                               [1, [a], [b]]

>> generate::splitProduct( - a / b);

                              [-1, [a], [b]]

>> generate::splitProduct( - a / b^(-x));

                                       x
                             [-1, [a, b ], []]

>> generate::splitProduct( - 2/3 * a / b^(x));

                                             x
                           [-1, [2, a], [3, b ]]

--*/

 generate::splitProduct :=
   //  analyse a _mult-expr and return a list of a 1 number and 2 
   //  lists; the number is the sign of the product 1 or -1, the  
   //  1. list contains the numerator and the 2. the denominator. 
 
   proc(multExpr)
     local numList, denumList, sigN, num, i, expo, normalForm,
           toNumList   /* what should be inserted in numList? */,
           toDenumList /* what should be inserted in denumList? */,
           insertFront : DOM_BOOL /* insert at the front or the end ? */;
   begin 
     if testargs() then
       if type( multExpr ) <> "_mult" then
         error("_mult-expr expected");
       end_if;
     end_if;
 
     normalForm :=
       (
        num := select({op(multExpr)}, testtype, Type::Numeric);
        if nops(num) = 0 then TRUE 
        elif nops(num) = 1 and  // the last one is the number ? 
	  testtype(op(multExpr, nops(multExpr)), Type::Numeric) then 
	  TRUE
        else FALSE end_if;
        ):   // if normalForm = FALSE don't change the order of terms 
     if nops(multExpr) = 1 then  // _mult(a+b) 
       return([1, [op(multExpr,1)], []]);
     end_if;
 
     sigN := 1;
     numList := [];
     denumList := [];
     
     for i in multExpr do
       toNumList := null();  toDenumList := null();
       if normalForm and testtype(i, Type::Constant) then
         insertFront := TRUE;
       else
         insertFront := FALSE;
       end;
       
       if testtype( i, Type::Real ) or i::dom::isNeg <> FAIL then
         if generate::isNeg(i) then
           i := -i;
           sigN := sigN * (-1);
           if i = 1 then next end_if;
         end_if;
         if testtype( i, DOM_RAT) then  // fraction
           num := op(i,1);
           if num <> 1 then  // ignore 1 ( a/2 -> [[a],[2]] not [[1,a],[2]] )
             toNumList := num;
           end_if;
           toDenumList := op(i,2);
         else
           // real number, but no fraction
           toNumList := i;
         end_if;
       elif testtype( i, DOM_COMPLEX ) then  // insert at the beginning
         toNumList := i;
       elif testtype( i, "_power" ) then
	 		 // no number, so insert at the end 
         // take a look at the exponent 
         expo := op(i ,2);
         if testtype( expo, "_mult") then
           if testtype( op(expo, nops(expo)), Type::Real ) then
             if  op(expo, nops(expo)) < 0 then
               toDenumList := 1/i;
             else
               toNumList := i;
             end_if; // product contains a real number 
           else
             // exponent is a product with no real number 
             toNumList := i;
           end_if;
          // exponent is no product 
          // if exponent is less zero, so insert 1/i in denumerator 
          // but not if base an exponent are numbers                
          elif testtype(expo, Type::Real) and
               not testtype(op(i, 1), Type::Real) then
            if expo = -1 then
              // special case for -1 because of internal normal form
              toDenumList := op(i, 1);
            elif expo < 0 then
              toDenumList := 1/i;
            else
              toNumList := i;
            end_if
         else   // just insert
           toNumList := i;
         end_if;
       else   // factor is no _power, just insert in numerator 
         toNumList := i;
       end_if;
       if insertFront then
         numList := [toNumList].numList;
         denumList :=[toDenumList].denumList;
       else
         numList := numList.[toNumList];
         denumList := denumList.[toDenumList];
       end_if;
     end_for;
 
     if numList = [] then   // dont't return empty numerator 
       numList := [ 1 ];
     end_if;
 
     return( [sigN, numList, denumList] );
   end_proc:
