

/**********************************
numlib::pi(x)
x - integer, rational, or float

determines the number of primes below x 


For a description of the algorithm (Meissel-Lehmer method),
see Lagarias, Miller, Odlyzko, "Computing \pi(x): The Meissel-Lehmer method",
Math. Comp., Vol. 44, No. 170, April 1985, pages 537-560
(in particular see p. 540 for reducing the problem to a call to
numlib::partialSievePhi)

**********************************/





numlib::pi:=
proc(x)
  local a, i, p, Li, P2: DOM_PROC;
begin
  
  
  // local methods

  Li:= x -> -Re(Ei(-ln(x))); 


  // P2(a) is the number of products of exactly two primes below x,
  // with none of the two primes being among the a smallest primes
  // We do not put up a sieve here (which might save time but cost memory)
  
  P2:=
  proc(x, a)
    local p, pxhalf, s;
  begin
    p:= ithprime(a+1);
    assert(p^3 > x);
    pxhalf:= a;
    s:= 0;
    while p <= floor(x^(1/2)) do
      pxhalf:= pxhalf + 1;
      s:= s + numlib::pi(x/p);
      p:= nextprime(p+1)
    end_while;
    // now pxhalf = pi(x^(1/2)
    a*(a-1)/2 - pxhalf*(pxhalf-1)/2 + s
  end_proc;
    
  
  case domtype(x)
    of DOM_RAT do
    of DOM_FLOAT do
      x:= floor(x);
      break
    of DOM_INT do
      break;
    of DOM_COMPLEX do
      error("Illegal argument")
    otherwise
      if testtype(x, Type::Arithmetical) then
        return(procname(x))
      else
        error("Illegal argument")
      end_if
  end_case;

  if x < 2 then
    return(0)
  end_if;
  
  // handle small values using ithprime
  if x <= max(stdlib::ifactor(PrimeLimit), numlib::thePrimes[-1]) then
    x:= numlib::prevprime(x);
    // guess a position
    i:= floor(Li(float(x)) - Li(float(x)^(1/2)));
    p:= ithprime(i);
    if p <= x then
      while p < x do
        p:= nextprime(p+1);
        i:= i+1
      end_while
    else
      while p > x do
        p:= numlib::prevprime(p-1);
        i:= i-1
      end_while
    end_if;
    return(i)
  end_if;
  
  // we apply the Meissel - Lehmer method.
  // If a > pi(x^(1/3)), then
  // we have phi(x, a) = 1 + pi(x) - a + P2(x, a)
 
  a:= numlib::pi(ceil(x^(1/3)));
  return(numlib::partialSievePhi(x, a) - 1 + a - P2(x, a))
  
end_proc:
      
