
/* ode::abel(eq,y,x) tries to solve the first-order ODE eq=0 wrt y(x)
   when eq=0 is of Abel type. 
   Reference: Murphy pp23-26 and Kamke pp24-28.

   Applicable to equations of the form :
    --> y' = f3(x) y^3 + f2(x) y^2 + f1(x) y + f0(x)
        (Abel equations of the first kind)

    --> (g0(x) + g1(x) y) y' = f2(x) y^2 + f1(x) y + f0(x)
        (Abel equations of the second kind)
        It can be always be transformed to an Abel ODE of first kind.

   Algorithm used is not complete for solving Abel ODEs of first kind
   (for a complete one, one may have a look at :
     E.S. Cheb-Terrab and A.D. Roche, Abel ODEs: Equivalence and Integrable
     Classes, Comp. Phys. Comm., 130 (2000) 197, p. 204-241.
   The only "easy" cases are implemented...

   Consider :
      s3:=f0f3^2+1/3(2f2^3/9-f1f2f3+f3f2'-f2f3') and
      s5=f3s3'-3s3(f3'+f1f3-f2^2/3)
   What is implemented for the moment are the cases s3=0 and s5^3/s3^5 is
   constant. They both reduces to separable ODEs.
   

   Examples:

1) Equation of second kind :

>> ode::solve(x*(y(x)+4)*diff(y(x),x)-y(x)^2-2*y(x)-2*x, y(x));


{    / / 4 C3 + C3 x - 4 \1/2        3/2
{ x, | | --------------- |    (x + 4)    -
{    \ \      x + 4      /
{

     / 4 C3 + C3 x - 4 \1/2        1/2      1/2 \
   4 | --------------- |    (x + 4)    + 4 x    | /
     \      x + 4      /                        /

   / / 4 C3 + C3 x - 4 \1/2        1/2    1/2 \
   | | --------------- |    (x + 4)    - x    |, -
   \ \      x + 4      /                      /

   /   / 4 C3 + C3 x - 4 \1/2        1/2
   | 4 | --------------- |    (x + 4)    -
   \   \      x + 4      /

   / 4 C3 + C3 x - 4 \1/2        3/2      1/2 \
   | --------------- |    (x + 4)    + 4 x    | /
   \      x + 4      /                        /

   / / 4 C3 + C3 x - 4 \1/2        1/2    1/2 \ }
   | | --------------- |    (x + 4)    + x    | }
   \ \      x + 4      /                      / }
                                                }


>> ode::solve(x*(2*y(x)+x-1)*diff(y(x),x)-y(x)*(y(x)+2*x+1), y(x));

                              3          2            2          2
{0, x - 1} union RootOf(4 C6 y  + 12 C6 y  - 12 C6 x y  + 12 C6 x  y +

                                                   2         3
   12 C6 y - 27 x y - 24 C6 x y - 12 C6 x + 12 C6 x  - 4 C6 x  + 4 C6, y)


1) Equation of first kind :

>> ode::solve(diff(y(x),x)-x^a*y(x)^3+3*y(x)^2-x^(-a)*y(x)-x^(-2*a)+a*x^(-a-1),y(x));

{                         /      x     \2
{                      exp| ---------- |
{                         |  a         |
{  1   1                  \ x  (a - 1) /
{ --, -- + ------------------------------------------,
{  a   a   /          /                          \1/2
{ x   x    |         |   a    /      x     \4    |
{          | C9 - 2  |  x  exp| ---------- |  dx |
{          |        /         |  a         |     |
{          \                  \ x  (a - 1) /     /

                       /      x     \2             }
                    exp| ---------- |              }
                       |  a         |              }
    1                  \ x  (a - 1) /              }
   -- - ------------------------------------------ }
    a   /          /                          \1/2 }
   x    |         |   a    /      x     \4    |    }
        | C9 - 2  |  x  exp| ---------- |  dx |    }
        |        /         |  a         |     |    }
        \                  \ x  (a - 1) /     /    }


*/

ode::abel :=
proc(eq, y, x)
  local yp,P,g,d,g0,g1,f0,f1,f2,f3,F1,F2,F3,aux,Sol,F;

begin
  yp := genident();
  eq := eval(subs(eq,diff(y(x),x)=yp,y(x)=y));
  userinfo(2,"trying to recognize an Abel equation");
  if testtype(eq,Type::PolyExpr(yp)) then
    if degree((eq:=poly(eq,[yp])))=1 then
      if testtype((P:=coeff(eq,0)), Type::PolyExpr(y)) then
        P:=poly(-P, [y]);
        if testtype((g:=coeff(eq,1)),Type::PolyExpr(y)) then
          d:=degree((g:=poly(g,[y])));
          if d=1 and degree(P)<=2 then
            userinfo(1,"Abel equation of second kind");
            g1:=coeff(g,1);
            g0:=coeff(g,0)/g1;
            f2:=coeff(P,2)/g1; f1:=coeff(P,1)/g1; f0:=coeff(P,0)/g1;
            F3:=-combine(normal(f2*g0^2-f1*g0+f0));
            F2:=-combine(normal(f1-2*f2*g0+diff(g0,x)));
            F1:=-f2;
            userinfo(2,"transformed Abel equation of first kind ".
                     "is:", diff(y(x),x)=F3*y(x)^3+F2*y(x)^2+F1*y(x));
            if (aux:=ode::Abel1st(F3,F2,F1,0,x))<>FAIL then
//              The solution of the initial ODE is 1/u-g0 where u is in aux
              userinfo(1,"Abel method worked");
              if type(aux)=DOM_SET then
                return(map(aux, f->
                       if f<>0 then
                         normal(expand(1/f-g0))
                       else
                         null()
                       end_if))
              else // aux is of type union 
                return(map(aux, f->
                  if type(f)="solve" then
                    Sol:=subs(op(f,1), op(f,2)=1/(op(f,2)+g0));
//                    eval(subs(f, op(f,1)=Sol));
                    solvelib::discreteSolve(Sol, op(f,2));
                  else // args(1) is a set

                           if type(f)=RootOf then
                      F:=subs(op(f,1),op(f,2)=1/(op(f,2)+g0));
                      solvelib::discreteSolve(F, op(f,2))
                    else
                        map(f, ()-> if args(1)<>0 then
                            normal(expand(simplify(1/args(1)-g0,sqrt)))
                                    else
                            null()
                                    end_if)
                    end_if
                  end_if))
              end_if  
            end_if
          elif d=0 and degree(P)<=3 then
            userinfo(1,"Abel equation of first kind");
            g0:=coeff(g,0);
            f3:=combine(normal(coeff(P,3)/g0));
            f2:=combine(normal(coeff(P,2)/g0));
            f1:=combine(normal(coeff(P,1)/g0));
            f0:=combine(normal(coeff(P,0)/g0));
            if (aux:=ode::Abel1st(f3,f2,f1,f0, x)) <> FAIL then
              userinfo(1,"Abel method worked");
              return(aux)
            end_if
          end_if
        end_if
      end_if
    end_if
  end_if;
  FAIL
end_proc:



//  Main procedure for solving Abel equations of first kind.
//  Only the simple cases are implemented, i.e. when s3=0 ans s5^3/s3^5=Ct

ode::Abel1st :=
proc(f3,f2,f1,f0, x)
  local y,z,C,s3,F,v,s5,inv,X,crs3,i,c0,c1,cubic,zero,integral,u,aux,sol;
  
begin
  y:= genident();
  C:= genident("C");
  u:= genident();
  if f3<>0 then
    s3:=combine(normal(f0*f3^2+1/3*(2*f2^3/9-f1*f2*f3+f3*diff(f2,x)
                       -f2*diff(f3,x))));
    if s3=0 then
      F:=normal(-f2/3/f3);
      v:=expand(exp(int(f1+f2*F, x)));
      z:=int(f3*v^2, x);
//  The solution is y(x)=u(z)*v(x)+F(x) where u'(z)=u(z)^3 ,
//      i.e. u(z)=0 or (+/-)1/sqrt(C-2*z)
      return({F, v/sqrt(-2*z+C)+F, -v/sqrt(-2*z+C)+F});
    else
      s5:=normal(expand(combine(f3*diff(s3,x)
                                -3*s3*(diff(f3,x)+f1*f3-f2^2/3))));
      if not has((inv:=normal(s5^3/s3^5)), x) then
/*   Here are some explanations of the following....
   The theory says (see Murphy p.24) that in this case the general solution
   is y(x)=u(x)v(x)+F(x), where F=-f2/3/f3, v=1/f3*s3^(1/3) and u is solution
   of the ODE :
          u'(x)=X(x)*(1-A*u(x)+u(x)^3)
   with (3A)^3=inv and X=1/f3*s3^(2/3).
   Therefore, u is solution of the following algebraic equation :
          solve(-int(X,x)+int(1/(1-A*u+u^3), u)+C=0, u)
       or u=RootOf(_Z^3-A*_Z+1, _Z).
   And it's easy to see that u(x)*v(x) is solution of :
          solve(-int(X,x)+v^2*int(1/(c0-c1*y+y^3), y)+C=0, y)
       or a solution of _Z^3-c1*_Z+c0=0
   with c0=s3/f3^3 and c1=1/3/f3^2/s3*s5.
   Here I computed directly the integral :
          integral:=int(1/(c0-c1*y+y^3), y)
                   =sum(ln(u-R)/(3*R^2-c1)) for R in solve(c0-c1*y+y^3, y)
   and therefore, y=uv+F is solution of
          solve(-int(X,x)+v^2*subs(integral, u=u-F)+C=0, u)
       or a solution of (_Z-F)^3-c1*(_Z-F)+c0=0.
   Since X=v^2/f3, we can ignore the content in the factorization of s3
   and simplify the cubic root of s3 (crs3 in the next) : only divide
   the exponents by 3.*/
   
        s3:=factor(s3);
        crs3:=_mult(s3[2*i]^(s3[2*i+1]/3) $ i=1..nops(s3) div 2);
        v:=normal(1/f3*crs3);
        F:=normal(-f2/3/f3);
        X:=normal(1/f3*crs3^2);
//   The solution is y(x)=u(x)*v(x)+F(x) where u'(x)=X(x)(1-A*u(x)+u(x)^3)
        c1:=normal(1/3*s5/f3^2/s3);
        c0:=normal(s3/f3^3);
        cubic:=c0-c1*y+y^3;
        if type((zero:=solvelib::discreteSolve(cubic, y)))=DOM_SET then
          if nops(zero)=3 then
            integral:=_plus(ln(u-F-i)/(3*i^2-c1) $ i in zero)
          else
            integral:=int(1/subs(cubic, y=u), u)
          end_if:
        else //zero is {RootOf(cubic,y)}
          integral:=_plus(ln(u-F-i)/(3*i^2-c1) $ i in op(zero))
        end_if;
        aux:=factor(v^2*integral-int(X, x));
        if not has(aux[1], {u,x}) then
          aux:=aux/aux[1]
        end_if;
        sol:=ode::mapsol(solvelib::discreteSolve(aux - ln(C), u), expand);
        return(solvelib::discreteSolve((y-F)^3 - c1*(y-F) + c0, y) union sol);
      else
        userinfo(1,"Abel equation failed");
      end_if
    end_if
  end_if;
  FAIL
end_proc:
  