/*  ode::autonomousEq(eq,y,x,n) solves the autonomous equation 
   eq=0 of order n in y(x),
   using the algorithm described in Zwillinger, section 36 page 153.

   Output: FAIL or a list of solutions for y(x).

   Example: 
   >> ode::autonomousEq(diff(y(x),x,x)-diff(y(x),x)-2*y(x)*diff(y(x),x),y,x,2);

         {        /           1/2          \                     }
         {        | (4 C3 - 1)    (C5 + x) |           1/2       }
         {     tan| ---------------------- | (4 C3 - 1)          }
         {        \            2           /                     }
         { C6, ------------------------------------------- - 1/2 }
         {                          2                            }

*/


ode::autonomous:=
  proc()
    local components;
  begin
    components:=ode::getComponents(args());
    if nops(components[2])>1 then
      error("autonomous method works only for scalar equations.")
    end_if;
    components[4]:=ode::order(components[1],{components[2]});
    ode::autonomousEq(components);
  end_proc:
    
ode::autonomousEq :=
 proc(eq,y,x,n) 
   local eq0,i,u,solu,soly,s,s1,sol,ind,elim;
 begin
   userinfo(2,"trying to recognize an autonomous equation");
   u:=genident("u");
   eq0:=eq;
   eq:=eval(subs(eq,diff(y(x),x$i)=ode::auto_transform(y,u,i) $i=1..n,y(x)=y));
   if has(eq,x) then FAIL 
   else 
      userinfo(1,"autonomous equation");
      sysassign(ode::depth,ode::depth+1);
      solu:=ode::solve_eq(eq,u,y);
      sysassign(ode::depth,ode::depth-1);
      if type(solu)="solve" or has(solu, solve) or has(solu,FAIL) then
         userinfo(1,"autonomous method failed");
         return(FAIL)
      elif domtype(solu)=RootOf then
        solu:={solu}
      end_if;
      userinfo(2,"solutions for ".expr2text(u)."(y) are",solu);
      // now each solution corresponds to u that is y'(x) 
      s:=eval(subs(eq,u(y)=0));
      /* if s=0, then u(y) is a factor of eq, therefore u=0 (y=C) is 
        already in solu */
      if s=0 then soly:={} 
      else soly:=solvelib::discreteSolve(s,y); //,MaxDegree=1);
        if calledFromWithinDSolve = TRUE and not ode::isResultOfSolveOK(soly)
          then userinfo(4,"autonomous method worked, but only get the ".
                        "implicit solution ".expr2text(soly)." and skip");
        end_if;
      end_if;
      if type(solu)="_union" then 
        solu:=map(solu,()->(if type(args(1))=DOM_SET
	                       then args(1) else {args(1)} end_if)) 
      end_if;
      for sol in solu do
         if sol=0 then s:={genident("C")} // y'=0 
         elif sol=FAIL then s:={}
         else
            userinfo(2,"integrating ".expr2text(1/sol)." in ".expr2text(y));
            // level needed for u=(y^2 + 1)^(3/2) for instance 
//            s:=int(1/sol,y); //s:=level(s,2);
//            s:=solvelib::discreteSolve(s=x+genident("C"),y);//,PrincipalValue);                                                      //MaxDegree=1);
/*       Maybe the following will not be needed later but it simplifyes things
       when we make solve(s=x+genident("C"),y). Look at the example :
       diff(y(x),x$2)-y(x)*diff(y(x),x)=0 for instance : due to integration, an
       arctan is not recognized and the above solve is very ugly (solution
       should be returned with a tan).*/
            ind:=indets(solu) minus {y,x} minus {PI,CATALAN,EULER};
            map(ind, assume, Type::Positive);
            s:=simplify(expand(eval(int(1/sol, y))));
            map(ind, unassume);
            s:=intlib::complexLnConversion(s, y);
            s1:=solvelib::discreteSolve(s-x-genident("C"),y);
            if s1 = FAIL then
              userinfo(1,"autonomous method failed");
              return(FAIL)
            else
              s:=s1
            end_if;
            userinfo(2,"solution is",s);
         end_if;
//if s<>FAIL then soly:=soly union s
         if (calledFromWithinDSolve = TRUE and ode::isResultOfSolveOK(s,"AllowImplicit"))
           or (calledFromWithinDSolve <> TRUE and
               ode::isResultOfSolveOK(s,"AllowImplicit"))
           then soly:=soly union s
         else
           userinfo(10,
                    "problems occurred in autonomous, solution(s) will be lost")
         end_if;
      end_for;
      if soly<>{} then
        userinfo(1,"autonomous method worked");
/*        It may happen that we get redundant solutions : same solutions but
        with a different constant. The following procedure tries to eliminate
        such solutions. */
        elim:=proc(l, eq)
                local ind, ll, f, indC, C, lC, indCC;
          begin
            if type(l)<>DOM_SET or nops(l)<2 then
              return(l)
            else
              ind:=indets(l) minus indets(eq) minus {PI,CATALAN,EULER};
              if nops(ind)>=2 then
                ll:={}; indC:=[];
                for f in l do
                  if has(f, ind) then
                    C:=[op(indets(f) intersect ind)];
                    indCC:=select(indC, ()-> nops(args(1))=nops(C));
                    if indC={} or indCC={} or (map(ll,expand) intersect map({subs(f, {C[i]=lC[i] $ i=1..nops(C)}) $ lC in indCC}, expand) = {}) then
                      ll:=ll union {f};
                      indC:=append(indC, C);
                    end_if;
                  else
                    ll:=ll union {f};          
                  end_if
                end_for;
                return(ll)
              else
                return(l)  
              end_if  
            end_if
          end_proc:
        if type(soly)=DOM_SET then
          return(elim(soly, eq0))
        else
          return(map(soly, elim, eq0))
        end_if;
//        soly;
      else FAIL
      end_if
   end_if
end_proc:

// gives the value of diff(y(x),x$n) in terms of u(y)=diff(y(x),x) 
ode::auto_transform:=proc(y,u,n) local e;
begin
   e:=u(y);
   while n>1 do e:=expand(diff(e,y)*u(y)); n:=n-1 end_while;
   e
end_proc:
