

/* ode::chini(eq,y,x) tries to solve the first-order ODE eq=0 wrt y(x)
   when eq=0 is of type Chini. 
   Reference: Kamke I.55 p303.

   Applicable to equations of the form :
     (eq) y' = f(x) y^n + g(x) y + h(x)

   Only a simple case is implemented : denote z=(h/f)^(1/n), if (z'-gz)/h
   is a constant alpha not depending on x then a solution of (eq) is
   z*u(x) where int(1/(u^n-alpha*u+1),u)+C=int((f/h)^(1/n)*h,x)

   Examples:

>> ode::solve(x*diff(y(x),x)*ln(x)-y(x)^2*ln(x)-(2*ln(x)^2+1)*y(x)-ln(x)^3,y(x))


                  {                 /      1          \ }
                  { -ln(x), - ln(x) | ----------- + 1 | }
                  {                 |           2     | }
                  {                 | C6   ln(x)      | }
                  {                 | -- + ------     | }
                  {                 \  2      2       / }


>> ode::solve(2*f(x)*diff(y(x),x)+2*f(x)*y(x)^2-diff(f(x),x)*y(x)-2*f(x)^2,y(x))


   {                                  /              /              \ }
   {       1/2      1/2        1/2    |       PI    |        1/2    | }
   { - f(x)   , f(x)   , I f(x)    tan| C10 + -- -  |  I f(x)    dx | }
   {                                  \        2   /                / }

*/


ode::chini :=
proc(eq, y, x)
  local yp,P,n,i,f,g,h,aux,alpha,hf,sing;

begin
  yp := genident();
  eq := eval(subs(eq,diff(y(x),x)=yp,y(x)=y));
  userinfo(2,"trying to recognize an integrable Chini equation");
  if testtype(eq,Type::PolyExpr(yp)) then
    if degree((eq:=poly(eq,[yp])))=1 then
      if testtype((P:=coeff(eq,0)/coeff(eq,1)), Type::PolyExpr(y)) then
        P:=poly(-P, [y]); n:=degree(P);
        if contains({{0},{}},{coeff(P,y,i) $ i=2..n-1}) then
          f:=combine(coeff(P,y,n)); g:=combine(coeff(P,y,1));
          h:=combine(coeff(P,y,0));
          aux:=normal(expand(1/n/f/h^2*(f*diff(h,x)-diff(f,x)*h-n*f*h*g)));
          alpha:=normal(expand(aux^n*h/f));
          if not has(alpha,x) then
            userinfo(2,"Integrable Chini equation");
            hf:=factor(combine(h/f));
            hf:=_mult(hf[1]^(1/n), hf[2*i]^(hf[2*i+1]/n) $ i=1..nops(hf) div 2);
            if alpha=0 then
              eq:=int(1/(y^n+1),y)=genident("C")+int(1/hf*h,x);
              sing:=solvelib::discreteSolve(y^n+1,y);
            else
              alpha:=normal(expand(hf*aux));
              eq:=int(1/(y^n-alpha*y+1),y)+1/alpha*genident("C")-int(h/hf,x);
              sing:=solvelib::discreteSolve(y^n-alpha*y+1,y);
            end_if;
            eq:=solvelib::discreteSolve(eq,y) union sing;
            return(ode::mapsol(eq, _mult, hf));
          end_if;
        end_if;
      end_if;
    end_if;
  end_if;
  FAIL
end_proc;

