/* ode::combination(sys,unk,t) tries to solve the system of
   ODEs sys by the method of integrable combinations.
   The implementation below works only for first-order systems with
   polynomial coefficients.

   It returns either FAIL or a sequence of solutions.

   Reference: Zwillinger pages 238-240.

   Example: 
>> ode::solve({diff(x(t),t)=-3*y(t)*z(t),diff(y(t),t)=3*x(t)*z(t),
                   diff(z(t),t)=-x(t)*y(t)},{x(t),y(t),z(t)});

                       2 1/2                            2 1/2
{[x(t) = - (C7 + 3 z(t) )   , y(t) = (- C1 - C7 - 3 z(t) )   ],

                       2 1/2                              2 1/2
   [x(t) = (C7 + 3 z(t) )   , y(t) = - (- C1 - C7 - 3 z(t) )   ],

                         2 1/2                              2 1/2
   [x(t) = - (C7 + 3 z(t) )   , y(t) = - (- C1 - C7 - 3 z(t) )   ],

                       2 1/2                            2 1/2
   [x(t) = (C7 + 3 z(t) )   , y(t) = (- C1 - C7 - 3 z(t) )   ]}

*/

ode::combination := proc(sys,unk,t) local l,i,n,v,eq;
begin
   userinfo(1,"trying integrable combinations method");
   n:=nops(unk);
   l:=[diff(op(unk,i),t)=genident()$i=1..n, op(unk,i)=op(unk,[i,0])$i=1..n];
   sys:=subs(sys,l);
   sys:=map(sys,()->(if type(args(1))="_equal" 
                       then op(args(1),1)-op(args(1),2) else args(1) end_if));
   v:=map(l,op,2);
   for eq in sys do
      if not testtype(eq,Type::PolyExpr(v)) then return(FAIL) end_if
   end_for;
   userinfo(1,"trying to find integrable combinations");
   sys:=[op(map(sys,poly,v))];
   /* traperror is needed because gbasis can fail on non rational polynomials
     (Error: Illegal argument [icontent]) */
   if traperror((sys:=groebner::gbasis(sys)))<>0 then return(FAIL) end_if;
   sys:=map(sys,expr);
   sys:=subs(sys,op(l,[2*n-i,2])=op(l,[2*n-i,1])$ i=0..2*n-1);
   sys:=map(sys,int,t);
   sys:=select(sys,not has,int);
   if nops(sys)=0 then return(FAIL) end_if;
   sys:={(numer(sys[i])+genident("C")) $ i=1..nops(sys)};
   sys:=solvelib::discreteSolve(sys,unk);
   userinfo(1,"integrable combinations method worked");
   sys
end_proc:
