/* ode::exact_second_order(eq,y,x) tries to solve the second order ODE eq=0
   with respect to y(x) using the method described in Zwillinger, chap. 51 p. 198.
   Namely, if eq is of the form f(x,y,y') y'' + g(x,y,y') =0, with the two conditions

   diff(f,x,x)+2*p*diff(f,x,y)+p^2*diff(f,y,y)=diff(g,x,p)+p*diff(g,y,p)-diff(g,y)
   diff(f,x,p)+p*diff(f,y,p)+2*diff(f,y)=diff(g,p,p)

   then eq is the total differential of some function.

   Input: eq is a second-order equation in y(x).
   Output: FAIL or a list of solutions.

   Example:

ode::solve(x*y(x)*diff(y(x),x,x)+x*diff(y(x),x)^2+y(x)*diff(y(x),x),y(x));

             1/2                1/2     1/2                1/2
           {2    (C2 ln(x) - C4)   , - 2    (C2 ln(x) - C4)   }

*/

ode::exact_second_order :=
proc(eq,y,x)
  local eq0,f,g,p,ypp,cf,cond1,cond2,firint,res;
begin
  p:= genident();
  ypp:= genident();
  eq0:=eq;
  eq:=subs(eq,diff(y(x),x)=p,diff(y(x),x,x)=ypp,y(x)=y);
  if testtype(eq,Type::PolyExpr(ypp)) then
    if degree(eq,[ypp])=1 then
      f:=normal(coeff(eq,[ypp],1));
      cf:=factor(f)[1]; f:=f/cf;
      g:=normal(coeff(eq,[ypp],0)/cf); 
      // check second condition first
      cond2:=normal(diff(f,x,p)+p*diff(f,y,p)+2*diff(f,y)-diff(g,p,p));
      if has(cond2, {cos,sin}) then
        cond2:=simplify(rewrite(cond2, sincos))
      end_if;
      if cond2=0 then
        cond1:=normal(diff(f,x,x)+2*p*diff(f,x,y)+p^2*diff(f,y,y)
                      -(diff(g,x,p)+p*diff(g,y,p)-diff(g,y)));
        if has(cond1, {cos,sin}) then
          cond1:=simplify(rewrite(cond1, sincos))
        end_if;
        if cond1 = 0 then
          userinfo(1,"exact second order equation");
          if not has((firint:=ode::firint_second_order(eq0,y,x)),FAIL) then
            if has((res:=ode::solve_eq(firint,y,x)),FAIL) then
              userinfo(2,"exact second order method failed but found a first".
                          " order integral", firint);
              return(FAIL)
            else
              return(res)
            end_if
          end_if
        end_if
      end_if
    end_if
  end_if;
  FAIL
end_proc:




// This function tries to compute a first integral for the 2nd order ODE eq
// which is supposed to be exact
ode::firint_second_order :=
proc(eq,y,x)
  local p,ypp,f,cf,g,h,f0,f1,intf0,u,firint;
begin
  p:= genident();
  ypp:= genident();
  eq:=subs(eq,diff(y(x),x)=p,diff(y(x),x,x)=ypp,y(x)=y);
  if testtype(eq,Type::PolyExpr(ypp)) and degree(eq,[ypp])=1 then
    f:=normal(coeff(eq,[ypp],1));
    cf:=factor(f)[1]; f:=f/cf;
    g:=normal(coeff(eq,[ypp],0)/cf); 
    f:=int(f,p);
               // solution is phi=constant where phi=int(f,p)+h(x,y) 
               // we should also have g = diff(phi,x)+diff(phi,y)*y' 
    g:=normal(expand(g-diff(f,x)-diff(f,y)*p));
    if has(g, {cos,sin}) then
      g:=simplify(rewrite(g, sincos))
    end_if;
               // thus now we should have g = diff(h,x)+diff(h,y)*p 
               /* we deal only with the three following easy cases:
                  (1) no y in g ==> h=int(g,x)
                  (2) no x in g ==> h=int(g/p,y)
                  (3) g=f0(x)+f1(y)*p ==> h=int(f0,x)+int(f1,y) */
    if not has(g,y) and not has(g,p) then
      h:=int(g,x)
    elif not has(g,x) and not has(g,p) then
      h:=int(g/p,y)
    elif testtype(g,Type::PolyExpr(p)) and degree(g,[p])<=1 then
      f1:=normal(coeff(g,[p],1)); f0:=normal(coeff(g,[p],0));
      if not has(f1,x) and not has(f0,y) then
        h:=int(f1,y)+int(f0,x)
      else
        if type(f0)="_plus" then
          intf0:=_plus(map(op(f0), int, x))
        else
          intf0:=int(f0,x)
        end_if;
        h:=intf0+int(normal(f1-diff(intf0,y)),y);
      end_if
    else
      userinfo(2,"cannot find a first order integral");
      return(FAIL)
    end_if;

    if type(f)="_plus" then
//      remove unnecessary constants
      f:=expand(f-select(f, testtype, Type::Constant))
    end_if:

    if type(h)="int" and has(h,y) then
      u:=genident("u");
      h:=int(subs(op(h,1), y=u), u=0..y(x));
           userinfo(2,"find a first order integral", subs(f, y=y(x))+h,
                      "but cannot solve it");
      return(FAIL);
    end_if:
    
               // now we obtain a first integral
    if type(f)="ln" or type(solvelib::combineln(f))="ln" then
//  to avoid to solve eq:=ln(diff(y(x),x))+....
      firint:=expand(exp(f+h))-genident("C")
    elif type(f)="arctan" or type(combine::arctan(f))="arctan" then
//  to avoid to solve eq:=arctan(diff(y(x),x))+....
      firint:=expand(tan(combine(f+h,arctan)))-genident("C")
    elif type(f)="int" then
      u:=genident("u");
      f:=int(subs(op(f,1), p=u), u=0..diff(y(x),x));
            userinfo(2,"find a first order integral", f+subs(h, y=y(x)),
                     "but cannot solve it");
      return(FAIL);
    else
      firint:=f+h-genident("C");
    end_if;
    
    firint:=subs(firint,y=y(x),p=diff(y(x),x));
    userinfo(2,"first order integral is",firint);
    return(firint)
  end_if:
  FAIL
end_proc:
