
/* ode::homogeneousC(eq,y,x) tries to solve the ODE eq with respect to y(x)
   using the method of homogeneous C equations (I take the same name as in
   Maple; the letter C comes from the classification of homogeneous ODE by
   Kamke on p.19-20). It returns either a set of solutions or FAIL.

   Applicable to: first order ODEs of the form :
     (1)    y'=f((a1x+b1y+c1)/(a2x+b2y+c2))
   where f is any fonction and D=a1b2-a2b1<>0

   Reference: Zwillinger p.278, Kamke p. 19.

   Method : when doing the change of variables :
     X = x - (b1c2-c1b2)/D
     Y = y - (-a1c2+c1a2)/D
   it reduces to solve the homogeneous equation :
     (2)    Y'=f((a1X+b1Y)/(a2X+b2Y))

   If an ODE y'=F(x,y) is of the form (1) then we have
     Fx/Fy=(Dy+a1c2-c1a2)/(-Dx+b1c2-b2c1)
   where Fx (resp. Fy) is the partial derivative of F in x (resp. y).

   Example:

>> ode::homogeneousC((y(x)+1)*diff(y(x),x)-y(x)-x, y,x);

     /                                                     1/2
     | /                                  1/2        1/2 \5
solve| |    3         x        2 v1      5        x 5    |     /
     | | ------- - ------- + ------- - ------- + ------- |
     | \ 2 x - 2   2 x - 2   2 x - 2   2 x - 2   2 x - 2 /
     |
     \

   /
   |              1/2      1/2     5              1/2      1/2     5
   | (2 v1 - x + 5    - x 5    + 3)  (2 v1 - x - 5    + x 5    + 3)
   |
   \

                                                       1/2 \
   /                                  1/2        1/2 \5    |
   |    3         x        2 v1      5        x 5    |     | =
   | ------- - ------- + ------- + ------- - ------- |     |
   \ 2 x - 2   2 x - 2   2 x - 2   2 x - 2   2 x - 2 /     /

                                                          \
   exp(10 ln(x - 1) - 10 C1)                              |
   -------------------------, v1, IgnoreProperties = TRUE | union
                     10                                   |
         1024 (x - 1)                                     |
                                                          |
                                                          /

   {         /        1/2 \              /  1/2       \     }
   {         |       5    |              | 5          |     }
   { (x - 1) | 1/2 - ---- | - 1, (x - 1) | ---- + 1/2 | - 1 }
   {         \         2  /              \   2        /     }


*/


ode::homogeneousC :=
proc(eq,y,x) 
  local yp,eqx,eqy,R,num,den,delta,A,B,aux,Sol,F;
begin
  yp:= genident();
  userinfo(2, "trying to recognize an homogeneous equation of type C");
  eq:= subs(eq, diff(y(x), x) = yp, y(x)=y);
  if testtype(eq, Type::PolyExpr(yp)) then
    if degree(eq, [yp], yp) = 1 then
      eq :=normal(-coeff(eq, [yp], yp, 0) / coeff(eq, [yp], yp, 1));
      eqx:=diff(eq,x); eqy:=diff(eq,y);
      if eqx <> 0 and eqy <> 0 then
        R:=normal(eqx/eqy);
        if testtype((num:=numer(R)), Type::PolyExpr(y)) and
           testtype((den:=denom(R)), Type::PolyExpr(x)) and
           not has(num,x) and not has(den, y) then
           if degree(num, [y], y) = 1 and degree(den, [x], x) = 1 then
             if (delta:=coeff(num, [y], y, 1))+coeff(den, [x], x, 1) = 0 then
               userinfo(1,"type C homogeneous equation");
               A:=coeff(den, [x], x, 0)/delta;
               B:=-coeff(num, [y], y, 0)/delta;
               eq:=subs(eq, {x=x+A, y=y+B});
               aux:=ode::homogeneous(diff(y(x),x)-subs(eq, y=y(x)), y, x);
               if aux <> FAIL then
                 userinfo(1,"homogeneous C worked");
                 aux:=subs(aux, x=x-A);
                 if type(aux)=DOM_SET then
                   return(map(aux, _plus, B))
                 else // aux is of type union 
                   return(map(aux, f->
                     if type(f)="solve" then
                  /*     sol:=op(f,1);
                       y:=op(f,2);*/
                       Sol:=subs(op(f,1), op(f,2)=op(f,2)-B);
                       eval(subs(f, op(f,1)=Sol));
                     else // args(1) is a set
                       if domtype(f)=RootOf then
                         F:=subs(f, op(f,1)=subs(op(f,1),op(f,2)=op(f,2)-B));
                         eval(F)
                       else
                         map(f, _plus, B)
                       end_if
                     end_if))
                 end_if
               else
                 userinfo(1,"homogeneous C failed");
               end_if;
             end_if;   
           end_if;        
        end_if;
      end_if;
    end_if;
  end_if;
  FAIL
end_proc:

      