
/* ode::homogeneousD(eq,y,x) tries to solve the ODE eq with respect to y(x)
   using the method of homogeneous D equations (I take the same name as in
   Maple; the letter D comes from the classification of homogeneous ODE by
   Kamke on p.19-20). It returns either a set of solutions or FAIL.

   Applicable to: first order ODEs of the form :
     (1)    y'=y/x+f(x)*g(y/x)

   Reference: Kamke p. 20.

   Method : by the change of variable y=x*u(x) in the ODE (1) we get the
   separable equation : x*u'(x)=f(x)g(u).

   Example:

>> ode::solve(x*(x^2+y(x)^2)*diff(y(x),x)-(x^2+x^4+y(x)^2)*y(x),y(x));

          {         /  2      \                                }
          {         | x       |                  2         1/2 }
          {    x exp| -- - C3 | lambertW(0, exp(x  - 2 C3))    }
          {         \  2      /                                }
          { 0, ----------------------------------------------- }
          {                        2        1/2                }
          {                   exp(x  - 2 C3)                   }


*/

ode::homogeneousD :=
proc(eq,y,x) 
  local yp,u,aux;
begin
  yp:= genident();
  userinfo(2, "trying to recognize an homogeneous equation of type D");
  eq:= subs(eq, diff(y(x), x) = yp);
  if testtype(eq, Type::PolyExpr(yp)) then
    if degree(eq, [yp], yp) = 1 then
      eq :=normal(-coeff(eq, [yp], yp, 0) / coeff(eq, [yp], yp, 1));
      eq :=normal(eq-y(x)/x);
      u := genident("u");
      eq := normal(subs(eq, y(x) = u(x)*x));
      aux:=ode::separate(x*diff(u(x),x)-eq, u,x,1);
      if aux <> FAIL then
        userinfo(1,"homogeneous D worked");
        return(ode::mapsol(aux, _mult, x));
      end_if;
    end_if;
  end_if;
  FAIL
end_proc:
      