
/* created by Cormier, 4/04/02.

   ode::homogeneousG(eq,y,x) tries to solve the ODE eq with respect to y(x)
   using the method of homogeneous G (gleichgradig in German) equations. It
   returns either a set of solutions or FAIL.

   Applicable to: first order ODEs of the form y'=x^(n-1)f(y/x^n), also called
   Isobaric equations, where f is an analytic function and n a rational number.

   Reference: Murphy, p36.

   Examples :

>> ode::homogeneousG(diff(y(x),x)-a*y(x)^3-b/x^(3/2), y,x);


{                           3       }            /
{ RootOf(2 b + X15 + 2 a X15 , X15) }            |
{ --------------------------------- } union solve|
{                 1/2               }            \
{                x                  }

      /       /                        2         2              2    2
      |       | X29 + 9 a b X13 - 2 X29  + 54 a b  X29 - 108 a b  X29  + 1
   sum| X29 ln| ----------------------------------------------------------
      \       \                            9 a b

   \                                                        \
   |                    3                  2    3           |         ln(x)
   |, X29 = RootOf(4 X33  - 3 X33 + 216 a b  X33  - 1, X33) | = C16 + -----
   /                                                        /           2

         \
         |    1/2
   , X13 | / x
         /


>> ode::solve((4*x*y(x)+3*y(x)^4)+(2*x^2+5*x*y(x)^3)*diff(y(x),x), y(x));

                               3  5    4  2
                       RootOf(x  y  + x  y  - C3, y)

*/


ode::homogeneousG :=
proc(eq,y,x) 
  local yp, zz, fx, fy, alpha, sol, u;
begin
  yp:= genident();
  u:= genident();
  userinfo(2, "trying to recognize an homogeneous equation of type G");
  eq:= subs(eq, {diff(y(x), x) = yp, y(x)=y});
  if testtype(eq, Type::PolyExpr(yp)) then
    if degree(eq, [yp], yp) = 1 then
      eq :=-coeff(eq, [yp], yp, 0) / coeff(eq, [yp], yp, 1);
      zz := normal(x*eq/y);
      if has(zz, y) then
        fx := normal(x*diff(zz, x));
        fy := normal(y*diff(zz, y));
        if fy <> 0 then 
          alpha := normal(expand(normal(fx/fy))) // alpha=-n in Murphy p36
        else
          return(FAIL);
        end_if
      elif not has(zz, x) then
        alpha := 0
      else return(FAIL);
      end_if;
      if not has(alpha, [x, y]) then
        userinfo(1,"type G homogeneous equation");
        sol:=ode::separate(x*diff(u(x),x)-
                 combine(expand(u(x)*(subs(zz, y=u(x)/x^alpha)+alpha))),u,x,1);
        if sol <> FAIL then
          userinfo(1,"homogeneous G worked");
          return(ode::mapsol(sol, _mult, x^(-alpha)));
        else
          userinfo(1,"homogeneous G failed");
        end_if;
      end_if;
    end_if;
  end_if;
  FAIL
end_proc:
