/* ode::interchange(eq,y,x,n) tries to solve the n-th order ODE eq=0 wrt y(x)
  by interchanging y and x (see Zwillinger page 263).
  There is a typo in (68.5): the solution is x(y)=sqrt(A*exp(2*y^3/3)-y^3-3/2).

Examples:
>> ode::interchange(diff(y(x),x)-x/(x^2*y(x)^2+y(x)^5),y,x,1);

                   /         /    3 \                      \
                   |         | 2 y  |      3      2        |
              solve| 4 C2 exp| ---- | - 2 y  = 2 x  + 3, y |
                   \         \   3  /                      /

>> ode::interchange(diff(y(x),x)^3-4*x*y(x)*diff(y(x),x)+8*y(x)^2,y,x,1);

                         {    3       2       2 }
                         { 4 x   (4 C4  x - 1)  }
                         { ----, -------------- }
                         {  27            6     }
                         {           64 C4      }

>> setuserinfo(Any,1):
>> ode::interchange(diff(y(x),x,x)+x*y(x)*diff(y(x),x)^3,y,x,2);

              solve(C2 airyBi(y, 0) + C3 airyAi(y, 0) = x, y)

*/

ode::interchange :=
proc(eq,_0_y,_0_x,n) 
  local i,l,sols, X, r, intterm;
  //save X;
begin
  // delete X;
  X:=genident();
   /* the following lines are needed because one cannot deal with other
     functions of x, for example p(x)*diff(y(x),x)+q(x) */
  l:=select(map(select(indets(eq,PolyExpr),testtype,"diff"),op,1),has,_0_x);
  if l minus {_0_y(_0_x)} <> {} then return(FAIL) end_if;
//   Do not interchange if there are expressions like int(f(x),x) :
  intterm:=misc::subExpressions(eq, "int");
  if has(map(intterm, ()->bool(op(args(1),2)=_0_x)), TRUE) then
    return(FAIL)
  end_if;

  userinfo(1,"trying to interchange dependent and independent variable");
  l:=[1/diff(X(_0_y),_0_y)];
  for i from 2 to n do
    l:=append(l,expand(diff(l[i-1],_0_y)*l[1]))
  end_for;
  eq:=numer(subs(eq,diff(_0_y(_0_x),_0_x$i)=l[i]$ i=1..n,
                 _0_y(_0_x)=_0_y,_0_x=_0_x(_0_y),X=_0_x));
   /* the following line avoids that we try to interchange back,
     but the call of "numer" above is crucial to have a normal form */
  sysassign(ode::interchange(eq,_0_x,_0_y,n), FAIL);
   // check the depth here in order not to enter an infinite loop 
  sysassign(ode::depth,ode::depth+1);
  eq:=ode::solve_eq(eq,_0_x,_0_y,{});
  sysassign(ode::depth,ode::depth-1);
  if has(eq,FAIL) then
    userinfo(1,"interchanging failed");
    FAIL
  else
//    sols:=[solvelib::discreteSolve(op(eq,i)=_0_x,_0_y)$ i=1..nops(eq)];
    sols:={};
    if type(eq)="solve" then eq:={eq} end_if:
    for i from 1 to nops(eq) do
      if type(op(eq, i))=DOM_SET then
        sols:=sols union op(map(op(eq, i),
            proc() begin
               if (r:=solvelib::discreteSolve(args(1)-_0_x, _0_y))=FAIL
               then {} else r;
               end_if;
            end_proc));
      else
        if traperror(
          (r:=solvelib::discreteSolve(op(eq,i)-_0_x,_0_y)))<>0 then
          return(FAIL)
        elif r<>FAIL then
          sols:=sols union r;
        else
          userinfo(2,"a solution of ",op(eq,i)-_0_x=0, "in", _0_y, " is a ".
           "solution but unable to solve it");
          userinfo(1,"interchanging failed");
          return(FAIL)
        end_if;
      end_if;
    end_for;
    userinfo(1,"interchanging successful");  
    if type(sols) = "_union" then
      _union(op(select(sols,ode::isResultOfSolveOK,"AllowImplicit")))
////      _union(op(select(sols,ode::isResultOfSolveOK,"AllowImplicit"))) ???
//    _union(op(select(sols,ode::isResultOfSolveOK)))
    else
      if ode::isResultOfSolveOK(sols,"AllowImplicit") = TRUE then
        sols
      else
        {}
      end_if;
    end_if;
  end_if;
end_proc:
