// $Date: 2003/06/17 15:28:48 $ $Author: cormier $ $Revision: 1.8 $ 

/* ode::int_factors(eq,y,x) tries to solve the ODE eq with respect to y(x)
   using the method of integrating factors (see Zwillinger's book page 259).
   Currently this works only for first order equations.

   Output: either FAIL or a set of solutions.

   Examples:
   ode::int_factors(diff(y(x),x)+y(x)/x-x^2,y,x);

                              {           4 }
                              {   4 C4 - x  }
                              { - --------- }
                              {      4 x    }

   ode::int_factors(y(x)*(y(x)^2-x)+x^2*diff(y(x),x),y,x);

                   {           1/2            1/2      }
                   {          2    x         2    x    }
                   { 0, - -------------, ------------- }
                   {                1/2            1/2 }
                   {      2 (C5 + x)     2 (C5 + x)    }

*/

ode::int_factors :=
proc(eq,y,x)
  local N,M,_f,g,u,eq0,yp,uu,sing,solu,s;
// save yp;
begin
   //delete yp;
  yp:=genident();
  eq0:=eq;
  userinfo(2,"trying to find an integrating factor");
  eq:=subs(eq,diff(y(x),x)=yp,y(x)=y);
  if testtype(eq,Type::PolyExpr(yp)) then
    eq:=poly(eq,[yp]);
    if degree(eq)=1 then
      M:=coeff(eq,0); N:=coeff(eq,1); // M*dx+N*dy=0
      eq:=diff(M,y)-diff(N,x);
      if not has((_f:=normal(eq/N)),y) then // case (A)
        u:=int(_f,x)
      elif not has((g:=normal(-eq/M)),x) then // case (B)
            // there is a typo in Zwillinger for the sign of g 
        u:=subs(int(g,y),y=y(x))
      else
        u:=FAIL
      end_if;
/* We have to take care of the following fact : when multiplying by u
 in the next ode::exact.... then we may loose some solution when the
 denominator of u vanishes. This is the case for the example :
 y^2+(xy-1)y'=0 */        
      if u<>FAIL and type(u)<>"int" then
        u:=expand(exp(u));
        sing:={};
        if not has((uu:=subs(u, y(x)=y)), x) then
          if (solu:=solvelib::discreteSolve(1/uu, y))<>FAIL then
            for s in solu do
              if iszero(expand(eval(subs(eq0, y(x)=s)))) then
                sing:=sing union {s}
              end_if;
            end_for;
          end_if;
        end_if;
        userinfo(1,"found integrating factor",u);
        u:=ode::exact_first_order(u*eq0,y,x);
        if not has(u,FAIL) then
          userinfo(1,"integrating factor method worked");
          return(sing union u);
        end_if;
      end_if
    end_if;
  end_if;
  FAIL
end_proc:
