
/* ode::int_factors2(eq,y,x) tries to solve the second order ODE eq with
   respect to y(x) using the method of integrating factors (see Murphy p165).

   Output: either FAIL or a set of solutions.

   The method used to find an integration factor is based on the paper by
   E.S. Cheb-Terrab and A.D. Roche :
   "Integrating factors for second order ODEs",
   J. Symb. Comp. 27(5), 505-519, 1999.
   Once one has found an integration factor I, then I*eq is exact.
   The notations are, as far as possible, those used in the article and
   so we refer to the paper for the explanations of the algorithm.

   Examples:

1) Equation of Liouville's type :

>> ode::solve(2*y(x)*(1-y(x))*diff(y(x),x,x)-(1-2*y(x))*diff(y(x),x)^2+y(x)*(1-y(x))*diff(y(x),x)*f(x), y(x));

        {     /    /             /                       \     \2 }
        {     |    |            |           1            |     |  }
        {     | exp| - C3 - C2  |  ------------------ dx | + 1 |  }
        {     |    |           /      /   /         \    |     |  }
        {     |    |                  |  |          |    |     |  }
        {     |    |                  |  |  f(x) dx |    |     |  }
        {     |    |                  | /           |    |     |  }
        {     |    |               exp| ----------- |    |     |  }
        {     \    \                  \      2      /    /     /  }
        { C4, --------------------------------------------------- }
        {              /             /                       \    }
        {              |            |           1            |    }
        {         4 exp| - C3 - C2  |  ------------------ dx |    }
        {              |           /      /   /         \    |    }
        {              |                  |  |          |    |    }
        {              |                  |  |  f(x) dx |    |    }
        {              |                  | /           |    |    }
        {              |               exp| ----------- |    |    }
        {              \                  \      2      /    /    }


2) Integrating factor :

>> ode::solve((x+y(x)^2)*diff(y(x),x,x)-2*(x-y(x)^2)*diff(y(x),x)^3+diff(y(x),x)*(1+4*y(x)*diff(y(x),x)), y(x));

         1/2        1/2                                      2
    {(-x)   , - (-x)   } union solve(exp(C10 y - C11 - 2) - y  = x, y)

*/

ode::int_factors2 :=
proc(eq,y,x)
  local yp,p,q,eq0,parameq,Phi,Phi0,found,a,b,c,X,Y,phi,tau,A,B,nu,paramnu,
        nu1,nu2,denomnu,sing,eq1,eq2,firint1,firint2,dery,sys,res,gama,intay,
        F,mu,Fmu,muc,H,Hy,Hp,y2,mutilde,foncF,elim,err;
begin
  yp:=genident(); y2:=genident();
  p:=genident("p");
  eq0:=eq;
  parameq:=indets(eq) minus {x,y,PI,CATALAN,EULER};
  userinfo(2,"trying to find an integrating factor");
  eq:=subs(eq,diff(y(x),x$2)=yp,diff(y(x),x)=p,y(x)=y);
  if testtype((eq:=poly(eq,[yp])),Type::PolyExpr(yp)) and degree(eq)=1 then
    Phi:=normal(-coeff(eq,yp,0)/coeff(eq,yp,1));
    Phi0:=Phi; found:=FALSE;
    if testtype(Phi,Type::PolyExpr(p)) and degree((Phi:=poly(Phi,[p])))<=2 then
      
      a:=normal(coeff(Phi,2)); b:=normal(coeff(Phi,1));
      c:=normal(coeff(Phi,0));
        
      if c=0 and not has(a,x) and not has(b,y) then
 //   The ODE is a Liouville's equation (c.f. Murphy p.385) and so we can
//    solve it directly.
        found:=TRUE;
        userinfo(2,"Liouville's equation : integrating factor is :".
                 " 1/diff(y(x),x)");
        X:=simplify(int(-b,x)); Y:=simplify(int(-a,y));
        res:=solvelib::discreteSolve(int(expand(exp(Y)),y)
                  +genident("C")*int(expand(exp(-X)),x)+genident("C"),y):
        if not has(res, FAIL) then
          return(res union {genident("C")})
        else
          userinfo(2,"Liouville's equation failed : problem with solve")
        end_if:
         
          
      elif normal(2*diff(a,x)-diff(b,y))=0 then
//   In this case we have two integrating factors of the form mu(x,y)
        phi:=diff(c,y)-a*c;
        if normal(diff(a,x$2)-diff(a,x)*b-diff(phi,y))=0 then
          found:=TRUE;
          userinfo(2,"two integrating factors of the form mu(x,y)");
          tau:=diff(int(a,y),x);
          A:=2*tau-b;
          B:=phi+tau*(b-tau)-diff(b-tau,x);
          nu:=genident();
          eq:=diff(nu(x),x$2)-A*diff(nu(x),x)-B*nu(x);
          userinfo(2,"solving linear 2nd ODE for the integrating factors :",
                   eq=0);
          sysassign(ode::depth,ode::depth+1);
          nu:=ode::solve_eq(eq,nu,x);
          sysassign(ode::depth,ode::depth-1);
          if nu<>{FAIL} then
            nu:=map(nu, _mult, exp(-int(a,y)));
            paramnu:=indets(nu) minus {x,y,PI,CATALAN,EULER} minus parameq;
            nu1:=simplify(op(subs(nu, paramnu[1]=0, paramnu[2]=1)));
            nu2:=simplify(op(subs(nu, paramnu[1]=1, paramnu[2]=0)));
            userinfo(2,"the two integrating factors found are :", nu1, nu2);
//            when multiplying by the integrating factor nu1 one may lose some
//            solution of the initial DE eq0: so one has to check this !
            denomnu:=denom(normal(nu1));
            sing:=solvelib::discreteSolve(denomnu, y);
            sing:=select(sing,
                         proc(s)
                         begin
                           if traperror((res:=eval(subs(eq0, y(x)=s))))=0 then
                             return(iszero(expand(res)))
                           end_if:
                           FALSE
                         end_proc):

//            We have two different ways now to proceed the resolution :
//            - compute the two first integrals with respect to nu1 and nu2 and
//              eliminate the derivative y' in these 2 equations to get the
//              solution of the original equation
//            - solve one the above first integral
            
            eq1:=subs(nu1, y=y(x))*(diff(y(x),x$2)-subs(op(Phi)[1],y=y(x),
                                                       p=diff(y(x),x)));
            eq2:=subs(nu2, y=y(x))*(diff(y(x),x$2)-subs(op(Phi)[1],y=y(x),
                                                       p=diff(y(x),x)));
 //           eq1 and eq2 are both exact differential equations
            firint1:=ode::firint_second_order(eq1,y,x);
            firint2:=ode::firint_second_order(eq2,y,x);
            firint1:=subs(firint1, diff(y(x),x)=p, y(x)=y);
            firint2:=subs(firint2, diff(y(x),x)=p, y(x)=y);
            dery:=solvelib::discreteSolve(firint1,p);
            if type(dery)=DOM_SET then
              sys:={subs(firint2, p=q) $ q in dery};
              err:=traperror((res:=_union(
                 solvelib::discreteSolve(numer(simplify(q)),y) $ q in sys)));
              if err=0 and not has(res, FAIL) then
                return(map(res,simplify) union sing)
              end_if:
            else
              userinfo(2,"solving new equation :", firint1=0);
              return(sing union ode::solve_eq(firint1,y,x))
            end_if:
          else
            userinfo(2,"integrating factors failed");
            return(FAIL)
          end_if
        end_if
        
      else
//   In this case we have one integrating factor of the form mu(x,y)        
        phi:=diff(c,y)-a*c-diff(b,x);
        gama:=diff(a,x$2)+diff(a,x)*b+diff(phi,y);
        if normal(diff(gama,y)-diff(a,x))=0 and
           normal(diff(gama,x)+phi+b*gama-gama^2)=0 then
          found:=TRUE;
          intay:=int(a,y);
          mu:=expand(exp(int(-gama+diff(intay,x),x)-intay));
          if mu<>1 then
            userinfo(2,"one integrating factor of the form mu(x,y)");
            userinfo(2,"integrating factor found is :", mu);
            eq:=subs(mu, y=y(x))*(diff(y(x),x$2)-subs(op(Phi)[1], y=y(x),
                     p=diff(y(x),x)));
            userinfo(2,"solving new equation :", eq=0);
            
//            when multiplying by the integrating factor mu one may lose some
//            solution of the initial DE eq0: so one has to check this !
            denomnu:=denom(normal(mu));
            sing:=solvelib::discreteSolve(denomnu, y);
            sing:=select(sing,
                         proc(s)
                         begin
                           if traperror((res:=eval(subs(eq0, y(x)=s))))=0 then
                             return(iszero(expand(res)))
                           end_if:
                           FALSE
                         end_proc):


            if has((res:=ode::exact_second_order(eq,y,x)),FAIL) then
              userinfo(2,"integrating factor method failed but found a first".
                        " order integral")
            else
              return(res union sing)
            end_if;
          end_if;
        end_if;
      end_if;
    end_if;

//////////////// SUBROUTINES ///////////////////////////////

/* The following procedure mutilde computes the function \tilde{mu}(x) of the
   Lemma 2 of Cheb-Terrab & A.D. Roche. */

mutilde :=
proc(phi,F,x,y,p)
  local phiy,phi1,phi2,phi3,phi4,res;
begin
  phiy:=diff(phi,y);
  phi2:=normal(diff(phiy*F,p));
  phi1:=normal(phiy*F-p*phi2);
  phi3:=normal(-diff(phi*F,p));
  phi4:=normal(diff(F,p));
  if phi2<>0 then
    if not has((res:=normal(1/phi2*(diff(phi1,y)-diff(phi2,x)))),{p,y}) then
      return(simplify(expand(exp(int(res,x)))))
    end_if;
  else
    if not has((res:=normal(1/phi4*(diff(phi3,p)-diff(phi4,x)))),{p,y}) then
      return(simplify(expand(exp(int(res,x)))))
    end_if;
  end_if;
  FAIL
end_proc:


/* The following procedure foncF computes the function {\cal F}(x,y') of
   the lemma 3 of Cheb-Terrab & A.D. Roche.
   Notations are, as far as possible, the same as the article. */

foncF :=
proc(phi,x,y,p)
  local gama,fgama,gamay,fgamap,i,w,wx,wy,pp,C1,lambda,Psi,lambdax,lambdap,
  Psix,Psip,ps,eq;
begin
  gama:=normal(diff(phi,y));
  fgama:=factor(gama);
  pp:=genident("p"); ps:=genident();
  if gama<>0 then
    
    gamay:=normal(diff(gama,y));
    if normal(diff(gamay/gama,p))<>0 then // Case A
      fgamap:=select([i $ i=1..nops(fgama) div 2], ()->
                   has(fgama[2*args(1)], p) and not has(fgama[2*args(1)], y));
      return(expand(_mult(fgama[1]*fgama[2*i]^(-fgama[2*i+1])
                          $ i in fgamap)))
    else
      fgamap:=select([i $ i=1..nops(fgama) div 2], ()->
                 has(fgama[2*args(1)], y) and not has(fgama[2*args(1)], p));
      w:=expand(_mult(fgama[1]*fgama[2*i]^(fgama[2*i+1]) $ i in fgamap));
      
      if gamay<>0 then
        
        if normal(diff(diff(ln(w),y),x))<>0 and
              normal(diff(diff(ln(w),y),y))<>0 then // Case C
          wx:=diff(w,x); wy:=diff(w,y);
          pp:=normal((diff(wx,y)*w-wx*wy)/(diff(wy,y)*w-wy^2));
          return(normal(w*(pp+p)/gama));
          
        elif not has((C1:=diff(ln(w),y)), {x,y,p}) then // Case E
          lambda:=normal(C1*exp(C1*y)/gama);
          Psi:=normal(phi*lambda-C1*exp(C1*y));
          lambdax:=diff(lambda,x); lambdap:=diff(lambda,p);
          Psix:=diff(Psi,x); Psip:=diff(Psi,p);
          eq:=normal((ps+pp^2*C1)*lambdap+pp*(p*lambdap*C1+lambda*C1
                     +diff(lambdax,p)+diff(Psip,p))+2*Psip+lambdax
                     +p*diff(lambdax,p)+p*diff(Psip,p));
          if normal(lambdap)=0 then
            pp:=op(solvelib::discreteSolve(factor(eq), pp))
          else
            eq:=normal(eq/lambdap);
            if has(eq, p) then
              pp:=op(solvelib::discreteSolve(factor(normal(diff(eq,p))), pp))
            else  // this is case F : not yet implemented...
              pp:=FAIL
            end_if
          end_if;
          if pp<>FAIL and not has(pp, {y,p}) then
            return(normal(w*(pp+p)/gama))
          else
            return(FAIL)
          end_if
        end_if;

      else // Case D
        lambda:=normal(1/gama);
        Psi:=normal(1/gama*phi-y);
        lambdax:=diff(lambda,x); lambdap:=diff(lambda,p);
        Psix:=diff(Psi,x); Psip:=diff(Psi,p);
        if normal(lambdap)=0 then
          pp:=normal(-p-(lambdax+2*Psip)/(diff(lambdax,p)+diff(Psip,p)))
        else
          ps:=normal(-((diff(lambdax,p)+diff(Psip,p))*(p+pp(x))
                       +lambdax+2*Psip)/lambdap);
          eq:=normal(lambda*diff(ps,x)+(diff(lambdax,x)+diff(Psip,x))*(p+pp(x))
                     +(lambdax+Psip)*ps+Psix-pp(x));
          eq:=subs(eq, diff(pp(x),x)=ps);
          pp:=op(solvelib::discreteSolve(eq, pp(x)));
        end_if;
        if not has(pp, {y,p}) then
          return(normal(w*(pp+p)/gama))
        end_if;
      end_if;
    end_if;  
      
  end_if;
  FAIL
end_proc:

////////////////// END OF SUBROUTINES ////////////////////////////////

//   If we have not found an integrating factor of the form mu(x,y) then
//   we look for one of the form mu(x,y')
    if not found and ((F:=foncF(Phi0,x,y,p);))<>FAIL then
        if (mu:=mutilde(Phi0,F,x,y,p))<>FAIL then
//   We have now to check for exactness condition by applying Euler's operator
//   to the total derivative of H=mu*(y''-phi) :
          H:=normal(F*mu*(y2-Phi0));
          Hy:=subs(diff(H,y), y=y(x), p=diff(y(x),x), y2=diff(y(x),x$2));
          Hp:=subs(rewrite(diff(H,p), D), y=y(x), p=diff(y(x),x),
                   y2=diff(y(x),x$2));
          Fmu:=normal(F*mu);
          muc:=subs(Fmu, y=y(x), p=diff(y(x),x));
          if muc<>1 and normal(Hy-diff(Hp,x)+diff(muc,x$2))=0 then
            userinfo(2,"integrating factor found is :", muc);
            eq:=muc*(diff(y(x),x$2)-subs(Phi0, y=y(x), p=diff(y(x),x)));
            userinfo(2,"solving new equation :", eq=0);

            denomnu:=denom(Fmu);
            if has(denomnu, p) then
              sing:=map(solvelib::discreteSolve(denomnu, p), int, x);
              sing:=map(sing, _plus, genident("C"));
            elif has(denomnu, y) then
              sing:=solvelib::discreteSolve(denomnu, y);
            else
              sing:={}
            end_if:
            sing:=select(sing,
                         proc(s)
                         begin
                           if traperror((res:=eval(subs(eq0, y(x)=s))))=0 then
                             return(iszero(expand(res)))
                           end_if:
                           FALSE
                         end_proc):

            
            if has((res:=ode::exact_second_order(eq,y,x)),FAIL) then
              userinfo(2,"integrating factor method failed but found a first".
                          " order integral")
            else
/*     ACHTUNG : we have to check that the solutions does not vanish for the
         integrating factor mu.
     Example : for exp(y(x))+1/x*diff(y(x),x)+diff(y(x),x$2)=0
     one obtain two solutions in res, one of which is not a solution! */

              elim:=proc(l)
              begin
                if type(l)=DOM_SET then
                  return(select(l, ()->
                         not(iszero(eval(subs(muc, y(x)=args(1))))))):
                else
                  return(l)
                end_if:
              end_proc:
            
              if type(res)=DOM_SET then
                return(sing union elim(res))
              else
                return(sing union map(res, elim))
              end_if:
            end_if:
          end_if;
        end_if;
    end_if
  end_if;
  FAIL
end_proc:


