/* ode::lagrange(eq,y,x) tries to solve the first-order ODE eq=0 wrt y(x)
   trying to recognize a Lagrange equation, i.e. an equation of the form

                          y = x*F(y')+G(y')

   If F=1, then we have a Clairaut's equation.

Examples: 
>> ode::solve(y(x)-2*x*diff(y(x),x)+a*diff(y(x),x)^3,y(x));

{                  /  3      2          3 1/2            \1/2
{         1/2  1/2 | x  - ((x  - 3 C3 a) )    + 9 C3 a x |
{      2 2    3    | ----------------------------------- |
{                  \                  a                  /
{ 0, - ------------------------------------------------------,
{                                 9

               /  3      2          3 1/2            \1/2
      1/2  1/2 | x  - ((x  - 3 C3 a) )    + 9 C3 a x |
   2 2    3    | ----------------------------------- |
               \                  a                  /
   ------------------------------------------------------,
                              9

                 /    2          3 1/2    3            \1/2
        1/2  1/2 | ((x  - 3 C3 a) )    + x  + 9 C3 a x |
     2 2    3    | ----------------------------------- |
                 \                  a                  /
   - ------------------------------------------------------,
                                9

               /    2          3 1/2    3            \1/2 }
      1/2  1/2 | ((x  - 3 C3 a) )    + x  + 9 C3 a x |    }
   2 2    3    | ----------------------------------- |    }
               \                  a                  /    }
   ------------------------------------------------------ }
                              9                           }


>> ode::solve(y(x)-2*x*diff(y(x),x)+diff(y(x),x)^2,y(x));

                        3      2  2                      3       2
    {0} union RootOf(4 y  - 3 x  y  - 18 C7 x y + 12 C7 x  + 9 C7 , y)


>> ode::solve(diff(y(x),x)^4-3*(1-x)*diff(y(x),x)^2+3*(1-2*y(x))*diff(y(x),x)+3*x, y(x));

{                                    2 3 1/2              3
{                   C11   ((4 x + C11 ) )      C11 x   C11
{ 5/6 - x, x + 1/6, --- - ------------------ - ----- - ---- + 1/2,
{                    2            12             2      12

                    2 3 1/2              3       }
   C11   ((4 x + C11 ) )      C11 x   C11        }
   --- + ------------------ - ----- - ---- + 1/2 }
    2            12             2      12        }

*/

ode::lagrange:=proc(eq,y,x) local p,F,G,eqinit,eq1,eq2,r,rr,t,sing,C,eqC,solC;
begin
   userinfo(2,"trying to recognize a Lagrange equation");
   p:=genident("p");// to avoid name space pb
   eqinit:=eq;
   eq:=subs(eq,diff(y(x),x)=p,y(x)=y);
   if testtype(eq,Type::PolyExpr(y)) then
      if degree((eq:=poly(eq,[y])))=1 then
         eq:=-coeff(eq,0)/coeff(eq,1); // should match x*F(y')+G(y') 
         if testtype(eq,Type::PolyExpr(x)) then
            if degree((eq:=poly(eq,[x])))=1 then
               userinfo(1,"Lagrange equation");
               F:=normal(coeff(eq,1)); G:=normal(coeff(eq,0));
// If F=p then we can't divide by (p-F) and the result is given
// by ode::clairaut(eq,y,x)
               if F=p then
                 return(ode::clairaut(eqinit,y,x))
               else
                eq:=diff(x(p),p)-x(p)*diff(F,p)/(p-F)-diff(G,p)/(p-F);
                userinfo(2,"new equation is",eq);
                sysassign(ode::depth,ode::depth+1);
                eq:=ode::solve_eq(eq,x,p,{});
                sysassign(ode::depth,ode::depth-1);
                if has(eq,FAIL) then
                   userinfo(1,"Lagrange method failed");
                   return(FAIL) 
                end_if;
                if testtype((sing:=solvelib::discreteSolve(p-F,p,MaxDegree=3)),
                           DOM_SET) then
                  sing:=map(sing, int, x)
                else
                  sing:={}
                end_if;
                sing:=map(sing,
                      proc()
                      begin
                        C:=genident("C");
                        eqC:=subs(eqinit,y(x)=args(1)+C);
                        if traperror((eqC:=eval(eqC)))=0 then 
                          eqC:=normal(expand(eqC));
                          if (solC:=solvelib::discreteSolve(eqC,C))<>{} then
                            return(args(1)+op(solC))
                          end_if
                        end_if;
                        null() 
                      end_proc);
                eq1:=map(eq,proc() begin args(1)-x end_proc );
                eq2:=map(map(eq1,numer),ode::eliminate,numer(y-x*F-G),p);
               // p=y' should not appear any more now 
                if has(eq2,FAIL) or has(eq2,p) or not has(eq2,{x,y}) then
                  if has(eq1, FAIL) then
                    userinfo(1,"Lagrange method failed");
                    return(FAIL)
                  else
                    eq1:=select(eq1, has, p);
                    r:=solvelib::discreteSolve(op(eq1), p);
                    if testtype(r, DOM_SET) and not hastype(r, RootOf) then
                      userinfo(1,"Lagrange method worked");
                      {expand(x*subs(F, p=rr)+subs(G, p=rr)) $ rr in r};
                      return(simplify(%) union sing)
                    else
                      userinfo(1,"Lagrange method worked, parametric solution");
                      t:=genident("_T");
                      return(
                        simplify(subs({[x(t)=rr,y(t)=expand(rr*F+G)] $ rr in eq}, p=t)));
                    end_if:
                  end_if;
                else
                  eq2:=map(eq2,()->solvelib::discreteSolve(expand(
                    args(1)/content(args(1),[y])),y,MaxDegree=2));
                  userinfo(1,"Lagrange method worked");
                  return(op(eq2) union sing)
                end_if
               end_if  
            end_if
         end_if
      end_if
   end_if;
   FAIL
end_proc:




// eliminates x between f and g 
ode::eliminate :=
proc(f,g,x)
begin
  f:=ode::topoly(f,x);
  if f<>FAIL then
    g:=ode::topoly(g,x);
    if g<>FAIL then
      return(polylib::resultant(f,g,x))
    end_if
  end_if;
  FAIL
end_proc:

/* tries to convert expression f into a polynomial in x
  (this can add some other solution like in resultant) */
ode::topoly :=
proc(f,x)
  local k,good,g;
begin
  if testtype(f,Type::PolyExpr(x)) then
    return(f)
  elif type(f)="_plus" then
    good:=select(f,testtype,Type::PolyExpr(x));
    f:=f-good;
    if type(f)="_mult" then
      g:=select(f,testtype,Type::PolyExpr(x));
      f:=f/g;
      if type(f)="_power" then
        if type((k:=op(f,2)))=DOM_RAT then // good+g*bad^k=0
          return((-good)^op(k,2)+g^op(k,2)*f^op(k,2))
        end_if
      end_if
    end_if
  end_if;
  FAIL
end_proc:
