// $Date: 2003/06/17 15:30:02 $ $Author: cormier $ $Revision: 1.27 $ 

/* ode::riccati(eq,y,x) tries to solve the first-order ODE eq=0 wrt y(x)
   by rewriting it as a linear second order ODE. 
   Reference: Zwillinger pages 288-291.

   Applicable to: equations of the form y' = a(x) y^2 + b(x) y + c(x).
   With the change y=-w'/w/a, they rewrite to w''-(a'/a+b)w'+a*c*w=0.

   Examples:
>> ode::riccati(diff(y(x),x)-exp(x)*y(x)^2+y(x)-exp(-x),y,x);

                  {          cos(x) + C2 sin(x)        }
                  { - -------------------------------- }
                  {   exp(x) sin(x) - C2 cos(x) exp(x) }


>> ode::riccati(diff(y(x),x)-y(x)^2+x*y(x)-1,y,x);

              {                          /  2 \             }
              {                          | x  |             }
              {                     2 exp| -- |             }
              {                          \  2 /             }
              { x, x + ------------------------------------ }
              {                            1/2          1/2 }
              {        2 C4 + I erf(1/2 I 2    x) (2 PI)    }

*/

ode::riccati :=
proc(eq,y,x) 
  local a,b,c,d,e,eq0,parameq,s,zz,C, yp, w;
// save yp,w;
begin
  yp := genident();
  w  := genident();
   // delete yp,w;
  eq0:= eq;
  parameq:=indets(eq) minus {x,y,PI,EULER,CATALAN};
  userinfo(2,"trying to recognize a Riccati equation");
  eq := eval(subs(eq,diff(y(x),x)=yp,y(x)=y));
  if testtype(eq,Type::PolyExpr(yp)) then
      if degree((eq:=poly(eq,[yp])))=1 then
        eq:=-coeff(eq,0)/coeff(eq,1); // should match a*y^2+b*y+c
        if testtype(eq,Type::PolyExpr(y)) then
          if degree((eq:=poly(eq,[y])))=2 then
            userinfo(1,"Riccati equation");
            a:=coeff(eq,2); b:=coeff(eq,1); c:=coeff(eq,0);
               // first try to find a particular solution
//            The method to find a particular solution is quite exotic!!
            userinfo(2,"try to find a particular solution");
            eq:=solvelib::discreteSolve(eval(subs(eq0,y(x)=w*x)),w);
               //eq:=solvelib::discreteSolve(eval(subs(eq0,y(x)=w*x)),w,
               //                            MaxDegree=ode::maxdegree);
            if type(eq)=DOM_SET then
              for s in eq do
                if has(s,x) or domtype(s)=RootOf then
                  next
                end_if;
                s:=s*x; // particular solution
                userinfo(1,"Riccati method worked");
                     // Eq. (76.2)
                return({s} union
                       map(ode::bernoulliEq(-2*a*s-b,a,2,x),_plus,s))
              end_for
            end_if;

            eq:=solvelib::discreteSolve(eval(subs(eq0,y(x)=w/x)),w);
            if type(eq)=DOM_SET then
              for s in eq do
                if has(s,x) or domtype(s)=RootOf then
                  next
                end_if;
                s:=s/x; // particular solution
                userinfo(1,"Riccati method worked");
                     // Eq. (76.2) 
                return({s} union
                       map(ode::bernoulliEq(-2*a*s-b,a,2,x),_plus,s))
              end_for
            end_if;

 //           Another way to find a particular solution in a particular case :
            if b=0 and not has(a,x) then
              zz:={};
              misc::maprec(c, {"function"} =
                           (t-> (if nops(t)=1 and op(t,1)=x then
                                   zz:=zz union {t}
                                 end_if;
                                 t)));
              if nops(zz)=1 then
                C:=genident();
                s:=C*zz[1];
                zz:=solvelib::discreteSolve(diff(s,x)-a*s^2-c, C);
                zz:=select(zz, not has, x);
                if zz<>{} then
                  s:=subs(s, C=op(zz,1));
                  userinfo(1,"Riccati method worked");
                  return({s} union map(ode::bernoulliEq(-2*a*s,a,2,x),_plus,s))
                end_if;
              end_if;
            end_if;
                                                                      
               /* otherwise try to solve the associated second order 
                 linear equation */
            d:=expand(-(diff(a,x)/a+b)); e:=expand(a*c);
            if has(d, {sin,cos}) or has(e, {sin,cos}) then
              d:=(simplify@expand@simplify)(d);
              e:=(simplify@expand@simplify)(e);
            else
              d:=normal(d); e:=normal(e);
            end_if;
	       // check, if constant coefficients 
            if not has([d,e],x) then
              eq:=ode::solve_eq(diff(w(x),x,x)+d*diff(w(x),x)+e*w(x),
                                w,x,{});
              if not has(eq,FAIL) then
                userinfo(1,"Riccati transformation worked");
                d:=indets(eq) minus {x,PI,EULER,CATALAN} minus parameq;
                if nops(d)=2 then eq:=subs(eq,op(d,2)=1) end_if;
                return(map(eq,()->normal(expand(-diff(args(1),x)/args(1)/a ))))
              end_if;
            end_if;
	       /* check, if coefficients are rational functions over the
		 rational numbers */
//            if map({d,e}, testtype, Type::RatExpr(x,Type::Rational))={TRUE}
            if map({d,e}, testtype, Type::RatExpr(x))={TRUE}
              then
              userinfo(2,"particular solution failed");
              userinfo(2,"do Riccati transformation");
              eq:=ode::secondOrder(diff(w(x),x,x)+d*diff(w(x),x)+e*w(x),
                                   w,x,{});
              if eq={} then
                eq:=ode::specfunc(diff(w(x),x,x)+d*diff(w(x),x)+e*w(x),w,x);
              end_if;
              if eq<>{} then
                userinfo(1,"Riccati transformation worked");
                if hastype(eval(eq), RootOf) then
                  if ode::printWarningsFlag then
                    warning("Cannot do back transformation,\n".
                            "solution of transformed equation \n".
                            expr2text(diff(w(x),x,x)+d*diff(w(x),x)+e*w(x)).
                            "\n is \n".expr2text(eq)."\n solution of ".
                            "given equation is diff(Y,".expr2text(x).
                            ")/Y/".expr2text(a));
                  end_if;
                  return(FAIL)
                else
                  eq:={op(eq,1)*genident("C")+op(eq,2)}
                end_if;
                return(map(eq,()->normal(expand(-diff(args(1),x)/args(1)/a ))))
              end_if;
            end_if;
            userinfo(1,"Riccati method failed");
          end_if
        end_if
      end_if
  end_if;
  FAIL
end_proc:
