/* ode::separate(eq,y,z,n) tries to solve the n-th order equation eq=0 for y(z)
  either recognizing a direct n-th order equation (D@@n)(y)(x)=f(x)
  or separate equations

  Examples: 
  >> ode::separate(diff(y(x),x,x)-exp(x),y,x,2);

                           {C3 + exp(x) + C2 x}

  >> ode::separate(diff(y(x),x)+y(x)^2,y,x,1);

                               {       1   }
                               { 0, ------ }
                               {    C5 + x }


  >> ode::separate(x+y(x)+1+(2*x+2*y(x)-1)*diff(y(x),x),y,x,1);

         {2 - x} union solve(2 y + 3 ln(x + y - 2) = - C2 - x, y)

*/

ode::separate :=
proc(eq,y,z,n)
  local s,parameq,i,l,inty,u,fu,yp,s1,soll,ind,norms,nums,dens,sing,
        newconst,C,newC,valnewC;
// save yp;
begin

   // first try to recognize (D@@n)(y)(x)=f(x) 
   // cf Zwillinger page 147 
   // delete yp;
  yp:= genident();
  s:=eval(subs(eq,diff(y(z),z$n)=yp));
  parameq:=indets(eq) minus {y,z,PI,CATALAN,EULER};
  newC:=genident("C");
    // using option 'PrincipalValue' leads to less efficiency and
    // in many cases to more bad results
  s:=solvelib::discreteSolve(s,yp,MaxDegree=1);
  if ode::isResultOfSolveOK(s,"NoRootOf") and nops(s)=1 then
    s:=op(s);
    if not has(s,{y,FAIL}) then // try to integrate directly
      userinfo(1,"trivial n-th order equation");
      s:=simplify(s);
      for i from 1 to n do
        userinfo(2,"integrand is",s);
        s:=int(s,z)+genident("C");
      end_for;
      return({s})
    elif n=1 then // try to separate
      s:=Factored::convert_to(factor(s),DOM_EXPR);
      if type(s)<>"_mult" then s:=hold(_mult)(s) end_if;
      ([l,s,inty]):=split(s,has,y);
      l:=subs(l,y(z)=y);
      if not has(l,z) then // int(dy/l)=int(s,z)
        userinfo(2,"separate equation:",diff(y(z),z)/l=s);
// We have to be careful when we divide by l, we may loose some
//   singular solutions.
// I've made some stuff here to simplify things : have a look at the following
// example : ode::solve(x*diff(y(x),x$2)+(y(x)-1)*diff(y(x),x), y(x))
// before and after the changes.
// If soll has an element, say l, containing only one parameter, say C, one
// introduce a new parameter C' such that l=C'. This may simplify the resolution
// when working with the new parameter C' instead of C, as in the above example.
        soll:={};
        if contains(indets(l), y) and
          (soll:=solvelib::discreteSolve(l,y))<>FAIL then
            if not has(soll, solve) then
              if (newconst:=select(soll,
                  ()->has(args(),parameq) and not(args() in parameq)
                                   and nops(indets(args()))=1))<>{} then
                C:=op(indets(newconst[1]));
                valnewC:=solvelib::discreteSolve(newconst[1]=newC, C);
                if type(valnewC)=DOM_SET then
                  valnewC:=valnewC[1];
                  assume(newC, Type::Positive);
                  l:=simplify(subs(l, C=valnewC));
                  s:=simplify(subs(s, C=valnewC));
                  soll:=simplify(subs(soll, C=valnewC));
                end_if:
              end_if;
              soll:=map(soll, normal);
            end_if;
        end_if;

        inty:=int(1/l,y);
        if type(-inty)="ln" then
          inty:=-inty; s:=-s;
        end_if;
        
        if type(inty)="ln" then // put the integration constant in the log
          u:=factor(combine(expand(exp(int(s,z)))));
 //         do not take into account factors of u that are constant :
          fu:=select([i $ i=1..nops(u) div 2], ()-> has(u[2*args(1)],{y,z}));
          s:=op(inty)-genident("C")*normal(_mult(u[1]*u[2*i]^u[2*i+1]
             $ i in fu));
          if (s1:=solvelib::discreteSolve(op(inty),y))<>FAIL then
            soll:=soll minus s1
          end_if;
        elif type((s1:=solvelib::combineln(numer(inty))))="ln" and
          nops(parameq)<=1 then
//          the same as above when inty can be combined into ln and when
//          the equation has no more than one parameter (otherwise the result
//          may be too complicated to handle for solve). 
          u:=factor(combine(expand(exp(int(denom(inty)*s,z)))));
          fu:=select([i $ i=1..nops(u) div 2], ()-> has(u[2*args(1)],{y,z}));
          s:=op(s1)-genident("C")*normal(_mult(u[1]*u[2*i]^u[2*i+1]
             $ i in fu));
          if (s1:=solvelib::discreteSolve(op(s1),y))<>FAIL then
            soll:=soll minus s1
          end_if;
        else
          s:=inty-int(s,z)+genident("C");
        end_if;
/*   Maybe the following will not be needed later but it simplifyes things
   when we make solve(s=x+genident("C"),y). Look at the example :
   diff(y(x),x)-c-y(x)^2=0 for instance : due to integration, an
   arctan is not recognized and the above solve is very ugly (solution
   should be returned with a tan, or an exp).*/
        ind:=indets(s) minus {y,z} minus {PI,CATALAN,EULER} minus parameq;
        map(ind, assume, Type::Positive);
        s:=simplify(expand(s));
        map(ind, unassume);
//        s:=intlib::complexLnConversion(op(s, 1), y)=op(s, 2);
        s:=intlib::complexLnConversion(s, y);
 
        norms:=normal(s); nums:=numer(norms); dens:=denom(norms);
        if not has(dens, z) then
          s:=nums
        end_if;
        s1:=solvelib::discreteSolve(s,y);
        if s1<>FAIL then
          if not(has(s1, solve)) and (has(s1,ln) or has(s1, exp)) then
            s:=map(s1, simplify)
          elif type(s1)=DOM_SET and type(op(s1))=RootOf then
            s:=op(s1)
          else
            s:=s1
          end_if;
        else
          return(FAIL)
        end_if:
        if calledFromWithinDSolve <> TRUE or
              ode::isResultOfSolveOK(s,"AllowImplicit") or 
              map({op(s)}, ode::isResultOfSolveOK,"AllowImplicit")={TRUE}
              
//              ode::isResultOfSolveOK(s,"AllowImplicit")
//          ode::isResultOfSolveOK(s)
          then userinfo(1,"separate method worked");
          if soll <> FAIL then
            if has(s, newC) then
              return(simplify(subs(soll, newC=newconst[1])) union
                     simplify(subs(s, newC=newconst[1])))
            else
              return(soll union s)
            end_if:
          else
            if has(s, newC) then
              return(simplify(subs(s, newC=newconst[1])))
            else
              return(s)
            end_if:
          end_if;
        else
          userinfo(4,"separate method worked, but only get the ".
                   "implicit solution ".expr2text(s)." and skip");
        end_if;
      else // try with a change of function
        l:=Factored::convert_to(factor(l),DOM_EXPR);
        if type(l)="_mult" then u:=numer(normal(select(l,has,y))) else u:=l end_if;
 //       if type(l)="_mult" then u:=op(select(l,has,y),1) else u:=l end_if;
        l:=s*l;
        u:=solvelib::discreteSolve(u,y);
        if type(u)=DOM_SET and nops(u)<>0 then
          u:=op(u,1);
          l:=normal(subs(l,y=y+u));
          // we have now a DE for yy=y-u
          if not has(l,z) and not has((i:=diff(u,z)),z) then
		  // y' = yy' + i = l
            sing:=solvelib::discreteSolve(subs(l-i,y=y-u),y);
            s:=factor(int(expand(1/(l-i)),y));
            if type(s[2])="ln" then
              s:=expand(exp(subs(s[2],y=y-u)))-genident("C")*exp(z/s[1]);
              if has(solvelib::discreteSolve(l,y), 0) then
                sing:=sing minus {u-i}
              end_if;
            else
              s:=subs(s,y=y-u)-z;
              if type(s)="_plus" then
                s:=select(s,not testtype,Type::Constant);
                s:=s+genident("C")
              end_if;
            end_if;
            s:=solvelib::discreteSolve(s,y,MaxDegree=1);
            
            if calledFromWithinDSolve <> TRUE or
              ode::isResultOfSolveOK(s,"AllowImplicit") then
//              ode::isResultOfSolveOK(s) then
              userinfo(1,"separate method worked");
              return(sing union s);
            else
              userinfo(4,"separate method worked, but only get the ".
                       "implicit solution ".expr2text(s)." and skip");
            end_if
          end_if
        end_if;
      end_if
    end_if
  end_if;
  FAIL
end_proc:

