// $Date: 2003/07/16 17:59:17 $ $Author: cormier $ $Revision: 1.21 $ 

/* ode::solvable_for_y(eq,y,x) tries to solve the first-order ODE eq=0 wrt y(x)
   using the method of "solvable in y" equations (cf Zwillinger pages 305-306).
   There is a typo in the example given by Zwillinger: equation 
   (82.6) should be
   p'=p/x/(1-p^2) instead of p'=px/(1-p^2), but the solution (82.7) is correct.

   Output: either FAIL or a set of solutions.

   Example:
>> ode::solvable_for_y(y(x)-x-x*diff(y(x),x),y,x);

                           {x (C2 - ln(x) + 1)}

>> ode::solvable_for_y(x-y(x)*diff(y(x),x)+x*diff(y(x),x)^2,y,x);

{     2           1/2                     2                2             }
{ (- x  exp(2 C4))    (exp(2 C4) - exp(C4)  lambertW(0, - x  exp(2 C4))) }
{ ---------------------------------------------------------------------- }
{                                             2           1/2            }
{            exp(C4) exp(2 C4) lambertW(0, - x  exp(2 C4))               }

*/

ode::solvable_for_y := proc(eq,y,x) local p,yy,sol,sol1;
begin
   userinfo(2,"trying to recognize an equation solvable for y");
   p:=genident("p");
   eq:=subs(eq,diff(y(x),x)=p,y(x)=y);
   // Note, I use MaxDegree only, since otherwise solve needs too much time
   sol:=solvelib::discreteSolve(eq,y,MaxDegree=1); //,MaxDegree=1);
   if not ode::isResultOfSolveOK(sol,"NoRootOf") or sol={0} or nops(sol)<>1
     then return(FAIL)
   end_if;
   //if type(sol)="solve" or sol=FAIL then return(FAIL) end_if;
   //sol:=select(sol,proc() begin bool(domtype(args(1))<>RootOf) end_proc );
   //if nops(sol)=0 then return(FAIL) end_if;
   userinfo(1,"equation solvable for y");
   yy:=op(sol,1); // yy=f(x,p) 
   sol:=diff(subs(yy,p=p(x)),x)-p(x);
   sysassign(ode::depth,ode::depth+1);
   sol:=ode::solve_eq(sol,p,x,{});
   sysassign(ode::depth,ode::depth-1);
   if type(sol)<>DOM_SET or has(sol,FAIL) or (type(sol)=DOM_SET and hastype(sol,DOM_LIST)) then
      userinfo(1,"solvable for y method failed");return(FAIL) 
   end_if;
   sol:=map(sol,proc() begin subs(eq,p=args(1)) end_proc );
   sol1:=map(sol,solvelib::discreteSolve,y);
   if sol1={FAIL} then
       sol:=map(sol, RootOf, y)
   else
       sol:=sol1
   end_if;
   userinfo(1,"solvable for y method worked");
   map(sol,()->(if type(args(1))="solve" or domtype(args(1))=RootOf then
                  args(1) else op(args(1)) end_if))
end_proc:
