// $Date: 2004/08/05 10:42:04 $ $Author: stefanw $ $Revision: 1.21 $ 

/******************************************************************
ode::solve_sys(sys, l, z, inits)

sys - a set of expressions
l   - a set of identifiers {x1, x2, ...}
z   - variable
inits - set of initial conditions

returns a set of lists {[x1(z) = ..., x2(z) = ...]}

*******************************************************************/


ode::solve_sys:=
proc(sys: DOM_SET,_l: DOM_SET, z,inits) // unks is a set
  local is_linear,eq,maxord,s,i,j,A,ll,_b,m,k,sol,n,indet,
  initialConditions,unks;
begin
  // handle initial conditions
  // because of the inconsistent return value formats of the
  // particular solvers called, sol may be a set of lists
  // {[x(t) =...., y(t)=...,. ...],...}
  // or a set of sets
  // {{x(t) = ..., y(t) = ..., ...}, ...}
  initialConditions:=
  proc(sol: DOM_SET, inits: DOM_SET, indet, _l, z)
  begin
    // convert to lists
    sol:= map(sol, x -> if type(x) = DOM_LIST then x else [op(x)] end_if);
    _l:= {op(_l)};
    if inits <> {} then
      sol:= map(sol, ode::solvesysinit, indet, inits, _l, z)
    end_if;
    /*
    if domtype(sol)=DOM_SET and type(op(sol))=DOM_SET then
      sol:= {[op(op(sol))]}
    end_if;
    */
    sol
  end_proc;

  /********
   m a i n
   ********/
 
  // Get a list of the indeterminates from the original system
  // this will be needed to handle initial conditions
  indet := {};
  for eq in sys union inits do
    indet := indet union indets(eq);
  end_for;

  n:=nops(_l); // number of unknowns functions
  unks:=map(_l,e->e(z)); // set of unknowns functions

  // check the equations and compute the maximal order
  maxord:=ode::order(sys,_l);
  userinfo(1,"differential system of order ".maxord." with ".n." functions");

  // recognize linear systems
  _l:=[op(_l)]; // to set the order
  s:=[(diff(_l[i](z),z$j) $ i=1..n) $ j=0..maxord];
  m:=n*(maxord+1);
  is_linear:=TRUE;
  _b:=[];
  for k from 0 to maxord do
    A[k]:=[[0$n]$n]
  end_for;
  for i from 1 to n do
    if (ll:=Type::Linear(expand(op(sys,i)),s))<>FALSE then
      _b:=append(_b,ll[m+1]);
      for k from 0 to maxord do
        for j from 1 to n do
          A[k][i][j]:=ll[k*n+j]
        end_for;
      end_for;
    else
      is_linear:=FALSE;
      break
    end_if
  end_for;
  sol:=FAIL;
  if is_linear then
    sol:= ode::linsys(A,_b,_l,z,n,maxord,n)
  else
    userinfo(1,"nonlinear system");
    if maxord = 1 then
      sol:=ode::combination(sys,unks,z);
      if sol = FAIL then
        sol:=ode::matrixRiccati(sys,unks,z)
      end_if;
    end_if;
  end_if;
  if sol <> FAIL then
    return(initialConditions(sol,inits,indet,_l,z))
  end_if;

  // try to recognize a triangular system
  for eq in sys do
    s:= misc::subExpressions(eq, "function") intersect unks;
    if nops(s) = 1 then
      userinfo(1,"found one independent equation");
      s:= op(s); // x(t)
      sol:= ode::solve_eq(eq,op(s,0),op(s));
      if has(sol, FAIL) then
        next
      end_if;
      assert(domtype(sol) = DOM_SET);
      map(sol,
        proc(solution, sys, unks, s)
          begin
            solution:= {s = solution};
            sys:= eval(subs(sys minus {eq},solution));
            unks:= unks minus {s};
            if nops(unks) = 1 then
              s:=op(unks);
              sys:=ode::solve_eq(op(sys),op(s,0),op(s));
              if has(sys, FAIL) then
                return(FAIL)
              end_if;
              
          sys:= ode::mapsol(sys,
                            proc(sol)
                            begin
                              {s = sol}
                            end_proc
                            );
          sys:= map(sys,
                    proc()
                    begin
                      if type(args(1))=DOM_SET then
                        args(1)
                      else
                        {args(1)}
                      end_if
                    end_proc
                    );
              
          else
          sys:=ode::solve_sys(ode::getComponents(sys,unks));
          assert(testtype(sys, Type::SetOf(Type::ListOf("_equal"))))
          end_if; // nops(unks) = 1
          if type(sys)=DOM_SET then
            return(initialConditions(map(sys,
                                         proc(e)
                                         begin
                                           if domtype(e)=DOM_LIST then
                                             e.[op(solution)]
                                           else
                                             [op(e), op(solution)]
                                           end_if
                                         end_proc),
                                     inits,indet,_l,z))
          else
          return(FAIL)
          end_if
          end_proc,
          sys, unks, s);
      if contains(%, FAIL) then
        return(FAIL)
      else
        return(_union(op(%)))
      end_if
    end_if
  end_for;

  /*
  if has(sol, FAIL) then
    FAIL
  else
    initialConditions({sol},inits,indet,_l,z)
  end_if;
  */
  FAIL
end_proc:


/*******************************************
ode::solvesysinit - initial condition solver

arguments:
sol - solution of a system of diff. eq. (list Of equations)
indet - set of free parameters
inits - set of initial conditions
fns   - set of identifiers representing the functions to solve for
z     - the variable on which the functions in fns depend

********************************************/
ode::solvesysinit:=
proc( sol: Type::ListOf("_equal"), indet: DOM_SET,
     inits: Type::SetOf("_equal"), fns: DOM_SET, z: DOM_IDENT)
  local
  csts: DOM_SET, // set of constants
  eq, s, Lhs, a, i, csteq, aux, e, t, spsol, soln;
begin
  // Find the constants
  csts := indets(sol) minus (indet union fns union {z});
  userinfo(3,"general solution is",sol);
  // here solve_eq.mu has a test to see if the solution has type
  // RootOf
  userinfo(3,"initial conditions are",inits);
  // Now put the initial conditions into a more useful form
  s := {};
  for eq in inits do
    if has(op(eq,2),fns) then
      error("dependent variable not allowed in rhs of init conds");
    end_if;
    Lhs := op(eq,1);
    i := 0;
    a := op(Lhs);
    Lhs := op(Lhs,0); // This should be f or D(...(D(f))...)
    case type(Lhs)
      of DOM_IDENT do
        break
      of "D"       do
        repeat
          Lhs := op(Lhs);
          i := i+1;
        until type(Lhs) <> "D"
        end_repeat;
        break
      otherwise
        error("invalid initial conditions")
    end_case;
    if not(contains(fns,Lhs)) then
      error("invalid initial conditions")
    end_if;
    s := s union {[Lhs,i,a,op(eq,2)]};
	// [f,i,a,b] means (D@@i)(f)(a)=b
	
	// solve_eq has a check here to make sure that the denominator
	// of eq doesn't vanish

  end_for;

    // solve_eq.mu has the next part in "traperror"
    // it will take the initial conditions and create a system of 
    // equations which can be solved for the constants
  csteq := {};
 
  for t in s do
    if has(sol, t[1]) then
      // get the right hand side of t[1](z) = ....
      eq := op(select(sol,has,t[1]),[1,2]) 
    end_if;
    // I'm not sure of why there is the substitution "int=fun(0)"
    // in the next line, but it is from solve_eq.mu
    csteq := csteq union {eval(subs(diff(eq,z$t[2]),
                                    hold(int)=proc()
                                              begin
                                                0
                                              end_proc ,
                                    z=t[3])) = t[4]};
  end_for;
 
 
  // Now solve the equations for the constants
  if map(csteq,
         proc()
           begin
           nops(_intersect(indets(args()) minus Type::ConstantIdents, csts))
         end_proc
         )
    ={1}
    // solve the equations independently because solvelib::discreteSolve may introduce
    // some output of type "_in" for instance.
    then
    aux:=[];
    for e in csteq do
      aux:=append(aux, op(op(solvelib::discreteSolve(e, csts))))
    end_for:
    csteq:={aux};
  else
    csteq := solvelib::discreteSolve(csteq,csts)
  end_if:
  // I would think that csteq would be a set of sets or lists,
  // but I again I get the following from solve_eq.mu
  csteq := select(csteq,proc()
                        begin
                          contains({DOM_LIST,DOM_SET},
                                   type(args(1)))
                        end_proc );
  
  // What to do if the system of equations for the constants cannot
  // be solved?  Return the general solution with the initial
  // conditions, I suppose
  
  if type(csteq) <> DOM_SET then
    if ode::printWarningsFlag then
      warning("Initial conditions cannot be solved")
    end_if;
    return(sol.[op(inits)])
  elif nops(csteq) = 0 then
    // no equations need to be satisfied
    return(sol)
  else
    spsol := {};
    for soln in sol do
      soln := (eval(subs(soln,op(csteq,i))) $ i = 1..nops(csteq));
      if has(soln,{sin,cos}) then
        soln := combine(expand(soln),sincos)
      end_if;
      spsol := spsol union {soln}
    end_for;
    //return(spsol);
    return([op(spsol)]);
  end_if;
end_proc: /* ode::solvesysinit */


