// $Date: 2004/05/12 09:13:55 $  $Author: tonner $   $Revision: 1.8 $

/* -----------------------------------------------------------
Author: SO
Status: final (Walter 23.02.03: added viewing box)

    Cone -- The graphical primitive for 3D cones

    Syntax:

    Cone(baseRadius, baseCenter, <, topRadius = 0>, topCenter <, op1, op2, ...>)

    baseRadius, topRadius : expressions
    baseCenter, topCenter : list of 3 expressions
    op1, op2, ...         : options of the form 'option = value'

    Example:
    >> plot::Cone(1, [0, 0, 0], [0, 0, 3])
    >> plot::Cone(1, [0, 0, 0], 0.2, [0, 0, 3])
----------------------------------------------------------------*/ 

plot::createPlotDomain("Cone",
                       "graphical primitive for 3D cones"):

//-----------------------------------------------------------------------------
plot::Cone::styleSheet:= table(
      LinesVisible = TRUE,
      LineColor    = RGB::Black.[0.25],
      FillColor = RGB::LightBlue,
      BaseRadius   = 1,
      BaseX        = 0,
      BaseY        = 0,
      BaseZ        = 0,
      TopRadius    = 0,
      TopX         = 0,
      TopY         = 0,
      TopZ         = 1,
      LineColorDirectionY=0
):

//------------------------------------------------------------------------
// Hints for parents:
plot::Cone::hints := {
      Scaling = Constrained
}:

// methods yet to be implemented:  convert, convert_to, expr

//-----------------------------------------------------------------------------
plot::Cone::new:=
  proc()
    local object, other;
  begin
    // check all known options from the argument list
    object := dom::checkArgs([], args());
    
    // get arguments which are not yet processed
    other := object::other;
    
    if testargs() and nops(other) > 0 then
      if nops(other) < 3 then
        error("at least 3 arguments expected")
      end_if;
      
      if nops(other) >=1 and not testtype(other[1], Type::Arithmetical) then
        error("1st argument: expecting an expression (the base radius)")
      end_if;
      
      if nops(other) >= 2 and
        not testtype(other[2],
                     Type::ListOf(Type::Arithmetical, 3, 3)) and
        not (other[2])::dom::hasProp(Cat::Matrix)=TRUE then
        error("2nd argument: expecting a list of 3 expressions (the base center)")
      end_if;

      if nops(other) >= 3 and
        (not testtype(other[3], Type::Arithmetical)) and
        (not testtype(other[3], Type::ListOf(Type::Arithmetical, 3, 3))) and
        not (other[3])::dom::hasProp(Cat::Matrix)=TRUE then
        error("3rd argument: expecting an expression (the top radius) ".
              "or a list of 3 expressions (the top center)")
      end_if;
      if nops(other) >= 4 and
        (not testtype(other[4], Type::ListOf(Type::Arithmetical, 3, 3))) and
        not (other[4])::dom::hasProp(Cat::Matrix)=TRUE then
        error("4th argument: expecting a list of 3 expressions (the top center)")
      end_if;
      if nops(other) > 4 then
        error("unexpected argument: ".expr2text(other[5]))
      end_if;
    end_if;
    
    if nops(other) > 0 then
      object::BaseRadius := other[1];
      object::Base       := other[2];
      if nops(other) = 3 then
        object::Top        := other[3];
      else
        object::TopRadius  := other[3];
        object::Top        := other[4];
      end_if;
    end_if;

    // semantically check for validity
    dom::checkObject(object);
  end_proc:

plot::Cone::print :=
  obj -> hold(plot::Cone)(
                          obj::BaseRadius, 
                          obj::Base, 
                          obj::TopRadius,
                          obj::Top
                         ):

//--------------------------------------------------------------
plot::Cone::MuPlotML:= proc(obj, attrib)
local b, br, t, tr, n, nn, tmp,
      xmin, xmax, ymin, ymax, zmin, zmax;
begin
    b:=  float([attrib[BaseX], attrib[BaseY], attrib[BaseZ]]);
    br:= float(attrib[BaseRadius]);
    t:=  float([attrib[TopX], attrib[TopY], attrib[TopZ]]);
    tr:= float(attrib[TopRadius]);
    plot::MuPlotML::beginElem("Cone",
       "Base" = (b[1], b[2], b[3]), "Top" = (t[1], t[2], t[3]),
       "BaseRadius" = br, "TopRadius" = tr,
    Empty);
    //--------------------------------------------------------------
    // Compute the viewing box. The disc of radius r centered at the
    // vector c with normal vector n has the viewing box components
    // vbox[k] = c[k] - r*sqrt(1-n[k]^2/|n|^2) .. 
    //           c[k] + r*sqrt(1-n[k]^2/|n|^2)
    //--------------------------------------------------------------
    n:= [t[1]- b[1], t[2]- b[2], t[3]- b[3]]; // the direction of the cone
    nn:= n[1]^2 + n[2]^2 + n[3]^2;
    // If Base = Top, the normal of the disc is not defined.
    // Choose a default direction:
    if iszero(nn) then 
       n := float([0, 0, 1]);
       nn:= float(1); 
    end_if;
    tmp:= sqrt((n[2]^2+n[3]^2)/nn); 
    xmin:= min(b[1]-br*tmp, t[1]-tr*tmp); 
    xmax:= max(b[1]+br*tmp, t[1]+tr*tmp);
    tmp:= sqrt((n[1]^2+n[3]^2)/nn);
    ymin:= min(b[2]-br*tmp, t[2]-tr*tmp); 
    ymax:= max(b[2]+br*tmp, t[2]+tr*tmp);
    tmp:= sqrt((n[1]^2+n[2]^2)/nn);
    zmin:= min(b[3]-br*tmp, t[3]-tr*tmp); 
    zmax:= max(b[3]+br*tmp, t[3]+tr*tmp);
    return([xmin .. xmax, ymin .. ymax, zmin .. zmax]);
end_proc:
