/* -------------------------------------------------------------
Author: W. Oevel

Status 22.2.03: Everything should work. The strategy for 
                logarithmic plots still needs to be tested. 

-----------------------------------------------------------------
plot::Curve2d -- The graphical primitive for 2D curves

Call(s): Curve2d([x(t), y(t)], 
                 t = TMin .. TMax
                <a = AMin .. AMax>,
                <Color = c>,
                <Mesh = n>,
                <Submesh = m>,
                <AdaptiveMesh = l>
                <DiscontinuitySearch = TRUE/FALSE>
                <opt1, opt2, ...>)
Parameters:
    x, y      : arithmetical expression (depending on t)
                and, possibly, on an animation parameter
    t         : the curve parameter: DOM_IDENT or indexed identifier
    TMin, TMax: numerical real values or univariate expressions
                of the animation  parameter
    a         : the animation parameter: DOM_IDENT or indexed identifier
    AMin, AMax: numerical real values
    c         : an RGB value, an RGBa value or a procedure
    n         : the number of mesh points: an integer > 0
    m         : the number of 'inner mesh points': an integer >= 0
    l         : the number of recursive refinements in the 
                adaptive mode: an integer >= 0

Options:
    opt1, opt2, ... : general 2d plot options

Examples:
 >> plot::Curve2d([t*sin(t), t*cos(t)], t = 0..2*PI )
 >> plot::Curve2d([sin(t)/t, cos(t)/t], t = 0..2*PI, Color = RGB::Blue )
----------------------------------------------------------------*/
plot::createPlotDomain("Curve2d",
                       "graphical primitive for 2D curves"):

// methods yet to be implemented:  convert, convert_to, expr

plot::Curve2d::styleSheet := table(LegendEntry = TRUE,
                                   UMesh=121):

plot::Curve2d::new:= proc()
local object, other, X, Y;
begin
    // check all known options from the argument list
    object := dom::checkArgs(["U"], args());
    // get arguments which are not yet processed
    other := object::other;
  
  if nops(other) > 0 and (other[1])::dom::hasProp(Cat::Matrix)=TRUE then
    other[1] := [op(other[1])];
  end_if;
  
    if testargs() and nops(other) > 0 then
      if domtype(other[1]) <> DOM_LIST or
        nops(other[1]) <> 2 then
        error("expecting a list of two function expressions");
      end_if;
      if nops(other) > 1 then
        error("unexpected argument: ".expr2text(other[2]));
      end_if;
    end_if;
    
    if nops(other) > 0 then
      [X, Y]    := other[1];
      object::XFunction := X;
      object::YFunction := Y;
    end_if;
    
    // semantically check for validity
    dom::checkObject(object);
end_proc:

plot::Curve2d::print :=
  obj -> hold(plot::Curve2d)([obj::XFunction, obj::YFunction],
                             obj::UName=obj::URange):

plot::Curve2d::MuPlotML := proc(object, attributes, inheritedAttributes,
                                flag = FALSE)
local mesh, submesh, adaptivemesh, discontinuitysearch,
      X, Y, u, umin, umax, x, xmin, xmax, y, ymin, ymax,
      colorfunction, viewingbox, singularities, 
      i, branch, branches, clipFactor, clipBranch,
      _min, _max, _range;
begin
    mesh    := attributes[UMesh];
    submesh := attributes[USubmesh];
    u       := attributes[UName];
    umin    := float(attributes[UMin]);
    umax    := float(attributes[UMax]);
    X       := attributes[XFunction];
    Y       := attributes[YFunction];
    adaptivemesh:= attributes[AdaptiveMesh];
    discontinuitysearch:=
               attributes[DiscontinuitySearch];
    [xmin,xmax,ymin,ymax]:= [Automatic $ 4]:
    if contains(attributes, ViewingBoxXMin) then
         xmin:= attributes[ViewingBoxXMin]:
    end_if;
    if contains(attributes, ViewingBoxYMin) then
         ymin:= attributes[ViewingBoxYMin]:
    end_if;
    if contains(attributes, ViewingBoxXMax) then
         xmax:= attributes[ViewingBoxXMax]:
    end_if;
    if contains(attributes, ViewingBoxYMax) then
         ymax:= attributes[ViewingBoxYMax]:
    end_if;
    if contains(attributes, LineColorFunction) then 
      colorfunction := attributes[LineColorFunction];
    else colorfunction := () -> null();
    end_if;
    //-------------------------------------------------------------
    // Set the parameters:
    //-------------------------------------------------------------
    [umin, umax, xmin, xmax, ymin, ymax]:=
          float([umin, umax, xmin, xmax, ymin, ymax]):
    mesh:= 1 + (mesh - 1)*(submesh + 1);
    if mesh <= 1 then
       warning("expecting a number of mesh points >= 2, got: ".
               expr2text(mesh).
               ". Setting the number of mesh points to 2.");
       mesh:= 2:
    end_if:
    //-------------------------------------------------------------
    // Modifications for logarithmic plots
    //  !!! This still needs to be tested !!!
    //-------------------------------------------------------------
    if attributes[CoordinateType] in {LogLin, LogLog} then
       if (domtype(xmin) = DOM_FLOAT and xmin <= 0) then
          error("expecting a positive X range for a logarithmic X axis. ".
                "Got Xmin = ".expr2text(xmin));
       end_if:
       if (domtype(xmax) = DOM_FLOAT and xmax <= 0) then
          error("expecting a positive X range for a logarithmic X axis. ".
                "Got Xmax = ".expr2text(xmax));
       end_if:
       if xmin <> Automatic then
          xmin:= ln(xmin);
       end_if;
       if xmax <> Automatic then
          xmax:= ln(xmax);
       end_if;
       X:= ln@X:
    end_if;
    if attributes[CoordinateType] in {LinLog, LogLog} then
       if (domtype(ymin) = DOM_FLOAT and ymin <= 0) then
          error("expecting a positive Y range for a logarithmic Y axis. ".
                "Got Ymin = ".expr2text(ymin));
       end_if:
       if (domtype(ymax) = DOM_FLOAT and ymax <= 0) then
          error("expecting a positive Y range for a logarithmic Y axis. ".
                "Got Ymax = ".expr2text(ymax));
       end_if:
       if ymin <> Automatic then
          ymin:= ln(ymin);
       end_if;
       if ymax <> Automatic then
          ymax:= ln(ymax);
       end_if;
       Y:= ln@Y;
    end_if;
    //-------------------------------------------------------------
    // call the plot::curveEval utility
    //-------------------------------------------------------------
    [viewingbox, branches, singularities]:=
       plot::curveEval([X, Y], u = umin .. umax, 
                       [xmin .. xmax, ymin .. ymax], mesh,
                       AdaptiveMesh = adaptivemesh,
                       DiscontinuitySearch = discontinuitysearch):
    //-------------------------------------------------------------
    // We need to undo the logarithmic modifications
    // before writing the data to the XML file.
    //  !!! This still needs to be tested !!!
    //-------------------------------------------------------------
    if attributes[CoordinateType] in {LogLin, LogLog} then
       // x -> exp(x)
       if viewingbox <> NIL then
          viewingbox:= [map(viewingbox[1], exp), viewingbox[2]];
       end_if;
       branches:= map(branches, map, pt -> [pt[1], exp(pt[2]), pt[3]]);
    end_if;
    if attributes[CoordinateType] in {LinLog, LogLog} then
       // y -> exp(y)
       if viewingbox <> NIL then
          viewingbox:= [viewingbox[1], map(viewingbox[2], exp)];
       end_if;
       branches:= map(branches, map, pt -> [pt[1], pt[2], exp(pt[3])]);
    end_if;
    // ------------------------------------------------------------
    // Special: clip out points that are really far away.
    // This seems necessary for Windoof, where plotting
    // raises an error if there are points way out of the
    // viewing area (even if they are clipped).
    // ------------------------------------------------------------
    if viewingbox <> NIL then 
      clipFactor:= 1000:
      // clipBranch:= (t, i) -> (_min < t[i] and t[i] < _max);
      clipBranch:= proc(t, i) begin
                     if t[i] < _min then t[i]:= _min; end_if;
                     if t[i] > _max then t[i]:= _max; end_if;
                     t;
                   end_proc:
      for i from 1 to 2 do // i = 1 = x, i = 2 = y
        _min:= op(viewingbox, [i, 1]);
        _max:= op(viewingbox, [i, 2]);
        _range:= _max - _min;
        if iszero(_range) then  
          _range:= _plus(op(viewingbox, [i, 2]) - op(viewingbox, [i, 1]) $ i=1..2);
           if iszero(_range) then
              _range:= 1;
           end_if;
        end_if;
        _min:= _min - clipFactor*_range;
        _max:= _max + clipFactor*_range;
        branches:= map(branches, /*select*/ map , clipBranch, i + 1);
      end_for:
    end_if:
    //-----------------------------------------
    // When Hatch calls:
    //-----------------------------------------
    if flag = TRUE then // The MuPlotML method was called by the Hatch object.
       // The singularities returned by plot::curveEval is a set
       // {u1, u2, ...} of u-values of the singularities of the curve.
       return([viewingbox, branches, singularities])
    end_if;
    //-----------------------------------------
    // write the graphical data to the XML file
    //-----------------------------------------
    for branch in branches do
      plot::MuPlotML::beginElem("Poly2d", "Filled"=FALSE, "Closed"=FALSE);
      for i from 1 to nops(branch) do
        [u, x, y]:= branch[i];
        plot::MuPlotML::prP2(x, y, colorfunction(u, x, y));
      end_for;
      plot::MuPlotML::endElem("Poly2d"): // end of Poly2d definition
    end_for: // for branch in branches
    //-------------------------------------------------------------
    // Finally, return the viewing box
    //-------------------------------------------------------------
    if viewingbox = NIL then
         return();
    end_if;
    return(viewingbox);
end_proc:
