/* ----------------------------------------------------------------------------
plot::Octahedron -- graphical primitive of a octahedron

Author: A. Sorgatz
Status: 2003-08-02

Call(s):
    plot::Octahedron( <args, ...> )

Parameters:
    -

Options:
    attr, ...   : plot options, i.e., equations Attribute = value, where
                  Attribute is one of the attributes listed below or a hint.

Related Domains:
    plot::Tetrahedron, plot::Hexahedron, plot::Icosahedron, plot::Dodecahedron,
    plot::Box, plot::Cone, plot::Cylinder, plot::Ellipsoid, plot::Sphere, plot::Rotate3d,
    plot::Scale3d, plot::Translate3d

Example:
    >> plot( plot::Octahedron() );

Further Reading:
    Refer to Tetrahedron.mu

---------------------------------------------------------------------------- */


plot::createPlotDomain("Octahedron",
                       "graphical primitive for octahedrons",
                       3,
   [
    CenterX, CenterY, CenterZ, Center, Radius,
    LinesVisible, LineStyle, LineWidth,
    LineColor, LineColorType, LineColor2, LineColorFunction, Color,
    Filled, FillColor, FillColorType, FillColor2, FillColorFunction,
    Shading,
    PointsVisible, PointStyle, PointSize,
    LineColorDirectionX, LineColorDirectionY, LineColorDirectionZ,
    LineColorDirection, FillColorDirection,
    FillColorDirectionX, FillColorDirectionY, FillColorDirectionZ
   ]
):

//-----------------------------------------------------------------------------
plot::Octahedron::styleSheet:= table(
      LinesVisible  = TRUE,
      LineColor     = RGB::Black.[0.25],
      Radius        = 1,
      CenterX       = 0,
      CenterY       = 0,
      CenterZ       = 0,
      LineColorDirectionY=0
):

//-----------------------------------------------------------------------------
plot::Octahedron::hints:= {
      Scaling = Constrained
}:

//-----------------------------------------------------------------------------
plot::Octahedron::new:= proc()
local  object, other;
option escape;
begin
    // filter out all generic arguments
    object:= dom::checkArgs([],args());

    // process the special attributes in object::other
    other:= object::other;
    if nops(other) > 0 then
      error("unexpected argument: ".expr2text(other[1]));
    end_if;

    // check all attributes
    dom::checkObject(object);
end_proc:

//-----------------------------------------------------------------------------
plot::Octahedron::print:= obj -> "plot::Octahedron(...)":

//-----------------------------------------------------------------------------
plot::Octahedron::MuPlotML:= proc(object, attributes, inheritedAttributes)
    // object:              object that was created by the new method above.
    // attributes:          table of all attributes. hints are not included.
    // inheritedAttributes: just ignore it.
local
    cx, cy, cz, r,
    triangles,
    fillcolorfunction,
    linecolorfunction, haslinecolorfunction,
    tn, tr,
    entries;

begin
    // check for fill-color function
    if contains(attributes, FillColorFunction) then
      fillcolorfunction:= float@(attributes[FillColorFunction]); // procedure
    else fillcolorfunction:= () -> null();
    end_if;

    // check for line-color function
    if contains(attributes, LineColorFunction) then
      linecolorfunction:= float@(attributes[LineColorFunction]); // procedure
      haslinecolorfunction:= TRUE;
    else linecolorfunction:= () -> null();
         haslinecolorfunction:= FALSE;
    end_if;

    cx:= float(attributes[CenterX]);
    cy:= float(attributes[CenterY]);
    cz:= float(attributes[CenterZ]);
    r := float(attributes[Radius]);

    triangles:= [
      [cx+r,cy+0,cz], [cx+0,cy+r,cz], [cx+0,cy+0,cz+r],
      [cx+r,cy+0,cz], [cx+0,cy-r,cz], [cx+0,cy+0,cz+r],
      [cx-r,cy-0,cz], [cx+0,cy-r,cz], [cx+0,cy+0,cz+r],
      [cx-r,cy-0,cz], [cx-0,cy+r,cz], [cx+0,cy+0,cz+r],
      [cx+r,cy+0,cz], [cx+0,cy+r,cz], [cx+0,cy+0,cz-r],
      [cx+r,cy+0,cz], [cx+0,cy-r,cz], [cx+0,cy+0,cz-r],
      [cx-r,cy-0,cz], [cx+0,cy-r,cz], [cx+0,cy+0,cz-r],
      [cx-r,cy-0,cz], [cx-0,cy+r,cz], [cx+0,cy+0,cz-r]

//    [cx+r,cy+r,cz], [cx-r,cy+r,cz], [cx+0,cy+0,cz+r],
//    [cx+r,cy+r,cz], [cx+r,cy-r,cz], [cx+0,cy+0,cz+r],
//    [cx-r,cy-r,cz], [cx+r,cy-r,cz], [cx+0,cy+0,cz+r],
//    [cx-r,cy-r,cz], [cx-r,cy+r,cz], [cx+0,cy+0,cz+r],
//    [cx+r,cy+r,cz], [cx-r,cy+r,cz], [cx+0,cy+0,cz-r],
//    [cx+r,cy+r,cz], [cx+r,cy-r,cz], [cx+0,cy+0,cz-r],
//    [cx-r,cy-r,cz], [cx+r,cy-r,cz], [cx+0,cy+0,cz-r],
//    [cx-r,cy-r,cz], [cx-r,cy+r,cz], [cx+0,cy+0,cz-r]
    ];

    // start to write triangle mesh to XML stream
    plot::MuPlotML::beginElem("Surf3d",
                              "Type"          = "Triangles",
                              "HasLineColors" = FALSE);

    tn:= 3;
    for tr in triangles do
      plot::MuPlotML::prP3(tr[1], tr[2], tr[3],
                           fillcolorfunction(tn div 3, tr[1], tr[2], tr[3]));
      tn:= tn + 1;
    end_for;

    plot::MuPlotML::endElem("Surf3d"):

    // start to write triangle lines to XML stream
    plot::MuPlotML::beginElem("Poly3d", "Filled" = FALSE, "Closed" = TRUE);

    tn:= 3;
    for tr in triangles do
      plot::MuPlotML::prP3(tr[1], tr[2], tr[3],
                           linecolorfunction(tn div 3, tr[1], tr[2], tr[3]));
      tn:= tn + 1;
    end_for;

    plot::MuPlotML::endElem("Poly3d"):

    // finally, return the viewing box
    entries:= (i) -> map(triangles,op,i);
    return( [min(entries(1))..max(entries(1)),
             min(entries(2))..max(entries(2)),
             min(entries(3))..max(entries(3))] );
  end_proc:
//------------------------------------------------------------------------