// $Date: 2004/01/05 13:53:22 $  $Author: tonner $   $Revision: 1.12 $

/* -----------------------------------------------------------
    Piechart3d -- The graphical primitive for 3D Pie charts

    Syntax:
    Piechart3d(data<, Titles=t><, Moves=m><, Colors=c><, op1, op2, ...>)

    data                  : list of expressions; the data set
    Titles                : list of equations pos=string; titles for the pies
    Moves                 : list of equations pos=offset; pie positions 
    Colors                : list of colors; pie colors
    op1, op2, ...         : options of the form 'option = value'

Example:
>> plot::Piechart3d([5,12,38,14,25],
                    Moves=[4=6],
                    Titles=[4="Neue Option: Moves"])

>> plot::Piechart3d([5,12,38],
                    Colors=[RGB::RoyalBlue,RGB::VioletRed,RGB::GreenPale])

>> plot::Piechart3d([5,6,12,4],
                    Header="CICHLIDENZUCHT:",
                    Titles=[1="Aulonocara", 2="Cynotilapia",
                            3="Pseudotropheus",4="Steatocranus"])

>> plot::Piechart3d(matrix([1,5,3,6,5]))

>> plot::Piechart3d(array(1..5, (1)=9, (2)=8, (3)=5, (4)=8, (5)=5 ))

>> plot::Piechart3d([5, 6, 12, 4], Header = "CICHLIDENZUCHT:",
                    Titles = [1 = "Aulonocara", 2 = "Cynotilapia",
                              3 = "Pseudotropheus", 4 = "Steatocranus"], 
                    Moves=[i=exp(-(a-10*i)^2)*10$i=1..4], a=5..45)

// extreme animation
>> cond := (i-1)*PI/2<a and i*PI/2>a:
   piece := piecewise([cond, sin(2*a+(i-1)*PI)/2], [not cond, 0]):
   p3 := plot::Piechart3d([5+a/4, 2, 1+a/2, 4], Header = "CICHLIDENZUCHT:",
                          Center = [sin(a), cos(a), sin(2*a)],
                          Radius = 10+sin(2*a),
                          Heights = [1.5+sin(a), 1.5+cos(2*a),
                                     1.5+sin(a), 1.5+cos(4*a)],
                          Titles = [1 = "Aulonocara", 2 = "Cynotilapia",
                                    3 = "Pseudotropheus", 4 = "Steatocranus"], 
                          Moves=[i=piece$i=1..4], a=0..2*PI):
  plot(p3):
----------------------------------------------------------------*/ 

plot::createPlotDomain("Piechart3d",
                       "graphical primitive for 3D pie charts",
                       3,
       [[Data, ["Mandatory", NIL],
			 ["Definition", "ExprSeq", FAIL,
			  "Data to plot.", TRUE]],
        CenterX, CenterY, CenterZ, Center, Radius,
        [Moves, ["Optional", NIL],
         ["Definition", "ExprSeq", FAIL, "Pie offsets", FALSE]],
        [Heights, ["Optional", NIL],
         ["Definition", "ExprSeq", FAIL, "Pie heights", FALSE]],
        [Titles,   ["Optional", NIL],
         ["Annotation", "ExprSeq", FAIL, "Pie titles", FALSE]],
        Colors, Color,
        LineColor, LineWidth, LineStyle, LinesVisible,
        Filled, TextFont, Billboarding]
):

plot::Piechart3d::styleSheet :=
   table(Colors=RGB::ColorList[1..10],
         LineColor=RGB::Black.[0.25],
         CenterX = 0,
         CenterY = 0,
         CenterZ = 0,
         Radius = 1,
         Heights = [0.3],
         Moves= [0],
         Titles= [""]
         ):

// Hints for parents:
plot::Piechart3d::hints := {Axes = None,
                            Scaling = Constrained,
                            CameraDirection = [0, -10, 14]
                     //     CameraDirectionX =  0,  // do not set the single
                     //     CameraDirectionY = -100,// components (it will
                     //     CameraDirectionZ =  140 // turn up on the help page!)
                           }:

// methods yet to be implemented:  convert, convert_to, expr

// here we have to do something
plot::Piechart3d::new:=
  proc()
    local object, other;
  begin
    // check all known options from the argument list
    object := dom::checkArgs([], args());
    
    // get arguments which are not yet processed
    other := object::other;

    if nops(other) > 1 then
      error("unexpected argument: ".expr2text(other[2]))
    end_if;
    
    if nops(other) <> 0 then
      dom::changeNotifier(object, "Data" = other[1]);
    end_if;

    // semantically check for validity
    dom::checkObject(object);
  end_proc:

plot::Piechart3d::print :=
  obj -> hold(plot::Piechart3d)(obj::Data):

plot::Piechart3d::changeNotifier :=
  proc(obj, eq)
    local l;
  begin
    if op(eq, 1) = "Data" then
      l := op(eq, 2);
      if not domtype(l)=DOM_LIST
        and not domtype(l) = DOM_ARRAY
        and not l::dom::hasProp(Cat::Matrix) = TRUE then
        error("1st argument: list or array or matrix expected")
      end_if;
      if domtype(l)=DOM_LIST then
        if not testtype(float(l), Type::ListOf(Type::Arithmetical)) then
          error("1st argument: entries of list must be arithmetical expressions")
        end_if:
      elif domtype(l) = DOM_ARRAY then
        if op(l,[0,1]) <> 1
          then error("1st argument: array must be one-dimensional")
        end_if;
      elif l::dom::hasProp( Cat::Matrix ) = TRUE then
        if linalg::nrows(l) <> 1 and linalg::ncols(l) <> 1 then
          error("1st argument: matrix must have one row or one column")
        end_if;
      end_if;
      //--------------------------------------------
      // Matrix, Array --> List :
      if domtype(l) = DOM_ARRAY or l::dom::hasProp( Cat::Matrix ) = TRUE then 
        l := [op(l)];
      end_if;
      // now l is a list
      dom::extslot(obj, "Data", l);
      //-----------------------------------------------
      return(FALSE);
    end_if;
    TRUE
  end_proc:

plot::Piechart3d::MuPlotML:=
  proc(obj, attrib)
    local l, i, m, c, t, ss, pi, db, a, b, xPos, yPos, hAlign, 
          xMin, xMax, yMin, yMax, printPie, totalMesh, heights, arc1, arc2,
          r, center, num, maxheight;
  begin
    // mesh for a full circle
    totalMesh := 120;


    //------------  begin:  read values from attrib -------------------------
    if contains(attrib, hold(Data)) then
      l := map(attrib[hold(Data)], float);
      num := nops(l);
    else
      error("data set no longer available; please assign the plot::Piechart3d object to an identifier");
    end:

    if nops(l) = 0 then
      // no data given, do nothing
      return([0.0..0.0, 0.0..0.0, 0.0..0.0])
    end_if;
    // radius of the pie
    r := float(attrib[Radius]);
    // center of the pie
    center := [attrib[CenterX], attrib[CenterY], attrib[CenterZ]];
    c := attrib[Colors];
    if nops(c) < num then
      // cycle trough givn colors
      c := _concat(c $ (num div nops(c)) + 1);
    end_if;
    heights := plot::listAttribute2table(Heights, num, Type::Real,
                                         attrib[Heights], 0.3);
    maxheight := max(op(map([op(heights)], op, 2)));
    m := plot::listAttribute2table(Moves, num, Type::NonNegative,
                                   attrib[Moves], 0.0);
    t := plot::listAttribute2table(Titles, num, DOM_STRING,
                                   attrib[Titles], "");
    if  min(op(l)) < 0 then
      error("Data: list entries must be non-negative")
    end_if;
    //------------  end:  read values from attrib -------------------------
   
    printPie :=
    proc(centerX, centerY, centerZ, angleBegin, angleEnd, color, height)
      local circle, i, angleDiff, mesh;
    begin
      // mesh for this pie
      mesh := ceil((angleEnd-angleBegin)*totalMesh/2/float(PI));
      if mesh < 2 then
        // otherwise division by 0 below
        mesh := 2
      end_if;
      angleDiff := angleEnd-angleBegin;

      circle := [NIL $ mesh];

      // compute circle
      for i from 1 to mesh do
        circle[i] := [centerX + r*cos(angleBegin+(i-1)*angleDiff/(mesh-1)),
                      centerY + r*sin(angleBegin+(i-1)*angleDiff/(mesh-1))]
      end_for;
      
      // Bottom plane
      plot::MuPlotML::beginElem("Surf3d", "Type"="TriangleFan",
                                "MeshVisible"=FALSE,
                                "FillColor"=color,
                                "FillColorType"="Flat");
      plot::MuPlotML::prP3(centerX, centerY, centerZ);
      for i from 1 to mesh do
        plot::MuPlotML::prP3(circle[i][1], circle[i][2], centerZ);
      end_for;
      plot::MuPlotML::endElem("Surf3d");
      
       // Border plane
      plot::MuPlotML::beginElem("Mesh3d",
                                "UMesh" = 2,    "VMesh" = 2,
                                "USubmesh" = 0, "VSubmesh" = 0,
                                "UseXYLinesVisible"=FALSE,
                                "PointsVisible"=FALSE,
                                "ULinesVisible"=FALSE,
                                "VLinesVisible"=FALSE,
                                "FillColor"=color,
                                "FillColorType"="Flat");
      plot::MuPlotML::prP3(centerX, centerY, centerZ);
      plot::MuPlotML::prP3(centerX, centerY, centerZ+height);
      plot::MuPlotML::prP3(circle[1][1], circle[1][2], centerZ);
      plot::MuPlotML::prP3(circle[1][1], circle[1][2], centerZ+height);
      plot::MuPlotML::endElem("Mesh3d");
      
      plot::MuPlotML::beginElem("Mesh3d",
                                "UMesh" = 2,    "VMesh" = mesh,
                                "USubmesh" = 0, "VSubmesh" = 0,
                                "UseXYLinesVisible"=FALSE,
                                "PointsVisible"=FALSE,
                                "ULinesVisible"=FALSE,
                                "VLinesVisible"=FALSE,
                                "FillColor"=color,
                                "FillColorType"="Flat");
      for i from 1 to mesh do
        plot::MuPlotML::prP3(circle[i][1], circle[i][2], centerZ);
        plot::MuPlotML::prP3(circle[i][1], circle[i][2], centerZ+height);
      end_for;
      plot::MuPlotML::endElem("Mesh3d");
      
      plot::MuPlotML::beginElem("Mesh3d",
                                "UMesh" = 2,    "VMesh" = 2,
                                "USubmesh" = 0, "VSubmesh" = 0,
                                "UseXYLinesVisible"=FALSE,
                                "PointsVisible"=FALSE,
                                "ULinesVisible"=FALSE,
                                "VLinesVisible"=FALSE,
                                "FillColor"=color,
                                "FillColorType"="Flat");
      plot::MuPlotML::prP3(circle[mesh][1], circle[mesh][2], centerZ);
      plot::MuPlotML::prP3(circle[mesh][1], circle[mesh][2], centerZ+height);
      plot::MuPlotML::prP3(centerX, centerY, centerZ);
      plot::MuPlotML::prP3(centerX, centerY, centerZ+height);
      plot::MuPlotML::endElem("Mesh3d");
      
      // Top plane
      plot::MuPlotML::beginElem("Surf3d", "Type"="TriangleFan",
                                "MeshVisible"=FALSE,
                                "FillColor"=color,
                                "FillColorType"="Flat");
      plot::MuPlotML::prP3(centerX, centerY, centerZ+height);
      for i from 1 to mesh do
        plot::MuPlotML::prP3(circle[i][1], circle[i][2], centerZ+height);
      end_for;
      plot::MuPlotML::endElem("Surf3d");
      
      plot::MuPlotML::beginElem("Lines3d", "Type"="LineLoop");
      plot::MuPlotML::prP3(centerX, centerY, centerZ);
      for i from 1 to mesh do
        plot::MuPlotML::prP3(circle[i][1], circle[i][2], centerZ);
      end_for;
      plot::MuPlotML::endElem("Lines3d");
      plot::MuPlotML::beginElem("Lines3d", "Type"="LineLoop");
      plot::MuPlotML::prP3(centerX, centerY, centerZ+height);
      for i from 1 to mesh do
        plot::MuPlotML::prP3(circle[i][1], circle[i][2], centerZ+height);
      end_for;
      plot::MuPlotML::endElem("Lines3d");
      
      plot::MuPlotML::beginElem("Lines3d", "Type"="Lines");
      plot::MuPlotML::prP3(centerX, centerY, centerZ);
      plot::MuPlotML::prP3(centerX, centerY, centerZ+height);

      plot::MuPlotML::prP3(circle[1][1], circle[1][2], centerZ);
      plot::MuPlotML::prP3(circle[1][1], circle[1][2], centerZ+height);

      plot::MuPlotML::prP3(circle[mesh][1], circle[mesh][2], centerZ);
      plot::MuPlotML::prP3(circle[mesh][1], circle[mesh][2], centerZ+height);
      plot::MuPlotML::endElem("Lines3d");
    end_proc:
    
    // initial ViewingBox
    xMin := center[1];  xMax := center[1];
    yMin := center[2];  yMax := center[2];

    // sum of data values
    ss := _plus(op(l)); 
    pi := float(PI); b := 0; 
    
    for i from 1 to num do
      db:= l[i]/ss*2*pi;
      a := b; b := b+db;
      xPos := float(center[1]+m[i]*r*cos((a+b)/2));
      yPos := float(center[2]+m[i]*r*sin((a+b)/2));
      arc1 := xPos+r*cos(a...b);
      arc2 := yPos+r*sin(a...b);
      xMin := min(xMin, op(arc1));
      xMax := max(xMax, op(arc1));
      yMin := min(yMin, op(arc2));
      yMax := max(yMax, op(arc2));
      printPie(xPos, yPos, center[3], a, b, c[i], heights[i])
    end_for;//i
    // 2. loop:  print titles over (after) pies
    for i from 1 to num do
      db:= l[i]/ss*2*pi;
      a := b; b := b+db;
      if t[i] <> "" then
        xPos := float(center[1]+1.1*(1+m[i])*r*cos((a+b)/2));
        yPos := float(center[2]+1.2*(1+m[i])*r*sin((a+b)/2));
        xMin := min(xMin, xPos);
        xMax := max(xMax, xPos);
        yMin := min(yMin, yPos);
        yMax := max(yMax, yPos);
        if xPos > center[1] then
          hAlign := Left;
        else
          hAlign := Right;
        end_if;
        plot::MuPlotML::beginElem("Text3d",
                                  "BaseDir"=(1.0, 0.0, 0.0),
                                  "FaceDir"=(0.0, 0.0, 1.0),
                                  "HorizontalAlignment"=hAlign,
                                  "VerticalAlignment"=Top,
                                  "Position"=(xPos, yPos,
                                              center[3]+maxheight));
        plot::MuPlotML::prCdata((t[i]));
        plot::MuPlotML::endElem("Text3d");
      end_if;
    end_for;//i

    [xMin..xMax, yMin..yMax, center[3]..center[3]+maxheight];
  end_proc:
