// Adaptive Surface tesselation, based on :
//    Luiz Velho, Luiz Henrique de Figueriredo, and Jonas Gomes
//    A Unified Approach for Hierarchical
//    Adaptive Tesselation of Surfaces
//    ACM Transaction on Graphics, Vol. 18, No. 4, Oct. 1999, pp.329-360
// ToDo: 
//       protection agains 'cannot convert to C double'
//       ignore complex values
//       special case empty plot (purely complex data)
//
// Structure of daptiveSurfaceEval:
// --------+
// Storage |
// --------+
// The data are stored in 3 tables and two lists of sets:
//   POINTS =    table(i = [u, v, x, y, z])
//   EDGES  =    table(i = [[pt1, pt2], [], subdivlevel])
//               with pt1, pt2 = integer indices of POINTS
//               and  tr1, tr2 = integer indices of TRIANGLES
//               The [] is to store intermediate information
//               after splitting this edge
//   TRIANGLES = table(i = [e1, e2, [e3]]
//               with e1, e2, e3 = integer indices of EDGES
//               [e3] indicates that this edge is to be taken
//               in the opposite direction as stored in EDGES[e3]
//   ULINES =    [{i1, i2, i3, ... }, ...]
//   VLINES =    [{i1, i2, i3, ... }, ...]
//               with i1, i2, ... being indices into EDGES, such that
//               EDGES[i1], EDGES[i2], ... form the U/V-Lines
//-----------+
// Utilities |
//-----------+
// findBorder(a, b) // not implemented, yet
//    a = [ua, va, xa, ya, za] is a point, where x,y,z evaluate
//    to real floats. 
//    b = [ub, vb] is a point, where x,y,z, did not evaluate 
//    to real floats.
//    Returns a point c = [uc,vc,xc,yc,zc] between a and b (as close
//    as possible to b) where x, y, z evaluate to a real float
// subdivide(edge) // only partially implemented
//    Tries the midpoint c = (a+b)/2 of [a,b] = edge[1].
//    1) If c can be evaluated and satisfies the max bend criterion 
//       --> return([[a,b], edge[2], MAXSUBDIVLEVEL)
//    2) If c can be evaluated and does not satisfy the max bend criterion 
//       --> return([[a,c,b], edge[2], edge[3] + 1)
//    3) If c cannot be evaluated then
//          c1 = findBorder(a, c), add_new_point(c1)
//          c2 = findBorder(b, c), add_new_point(c2)
//       --> return([[a,c1,c2,b], edge[2], edge[3] + 1)
// splitTriangle(triangle) // only partially implemented
//    inspects the 3 edges [e1,e2,e3] = triangle, which were marked
//    for subdivision. There is a total of 27 cases dependening on
//    the form of the edges (an edge can have 2 points (-> nothing to do),
//    or 3 points (-> split at the midpoint) or 4 points (-> there is
//    a hole in the middle).
//    For each case do:
//        add_new_point()  // only for some of the 27 cases
//        add_new_edges()  // for the edges that have to be split
//        add_new_triangles() // corresponding to the split edges
//        delete TRIANGLES[triangle]
//--------------+
// Main program |
//--------------+
//    initialize()  // try evaluation on a coarse regular mesh
//    for subdivlevel from 1 to MAXSUBDIVLEVEL do
//        Edges:= map(Edges, subdivide) // mark the edges to be subdivided
//        for triangle in TRIANGLES do
//            splitTriangle(tr1);
//        end_for:
//    end_for:
//    conversion to Triangles[i] = [pt1, pt2, pt3]
//    addnormals();
//    return([ViewingBox, Points]])
//--------------------------------------------------------------------

// There's a *lot* of assertions in this file.
// If not debugging, it saves time to disable them completely:
// alias(assert(x)=null()):

plot::adaptiveSurfaceEval:= proc(fX, fY, fZ,
                                 umesh, usubmesh, umin_umax,
                                 vmesh, vsubmesh, vmin_vmax,
                                 xrange, yrange, zrange 
                              // AdaptiveMesh = adaptivemesh
                              // "NoSingularities"
                                 )
local adaptivemesh,
      nosingularities,
      MAXBEND, MAXSUBDIVLEVEL, 
      POINTS, EDGES, TRIANGLES,
      ULINES, VLINES,
      NEXTEDGE, NEXTTRIANGLE,
      MINFLOAT, MAXFLOAT, ST,
      WORKTODO,
      umin, vmin, xmin, ymin, zmin,
      umax, vmax, xmax, ymax, zmax,
      vbox, ubbLO, ubbHI, uwidth, vwidth,
      getPoint, trackRange,
      F, tmp, 
      fmin, fmax, Fmin, Fmax,
      _umin, _umax, _vmin, _vmax,
      i, j, u, du, v, dv, x, triangle, subdivlevel, p2, p3, 
      e1, e2, e3, e4, e5, p1, p4, um1, vm1, pt,
      revertEdge, subdivide, splitTriangle,
      XYZratio, XYZratio2, oldXYZratio2, minlength2, 
      negativeCurvature,
      getAutomaticViewingBox, refine, recompute,
      add_normals1, add_normals2,
      checktriangle,    // for debugging
      checkorientation  // for debugging
    ;
begin
  adaptivemesh:= op(select([args()], has, AdaptiveMesh), [1,2]):
  nosingularities:= has([args()], "NoSingularities"):
  //-------------------------------------------------------
  // set the 'tuning parameters' MAXSUBDIVLEVEL and 'MAXBEND'
  //-------------------------------------------------------
  MAXSUBDIVLEVEL:= adaptivemesh:  // recursive depth of subdivisions
  //------------------------------------------------------------------------
  // The following 'max bend angle' of 10 degrees is mentioned explicitly
  // in the documentation of PLOT/ATTRIBUTES/AdaptiveMesh.tex. Do not change
  // this value without keeping the documentation up to date!
  //------------------------------------------------------------------------
  MAXBEND:= tan::float(10/180*PI/2)^2/2;

  MINFLOAT:= 1e-300: // smallest float that can be converted to a C double
  MAXFLOAT:= 1e300:  // largest float that can be converted to a C double

  //-------------------------------------------------------
  // float input parameters
  //-------------------------------------------------------
  [umin, umax]:= float([op(umin_umax)]);
  [vmin, vmax]:= float([op(vmin_vmax)]);
  uwidth:= umax - umin;
  vwidth:= vmax - vmin;
  if iszero(uwidth) then
     umesh:= 2; // for speed
  end_if;
  if iszero(vwidth) then
     vmesh:= 2; // for speed
  end_if;
  
  ULINES := [{} $ umesh];
  VLINES := [{} $ vmesh];
  
  umesh:= (umesh - 1)*(1 + usubmesh) + 1:
  du:= 1.435107e-6*uwidth: // 'random' shift used to avoid singularities

  vmesh:= (vmesh - 1)*(1 + vsubmesh) + 1:
  dv:= 1.435107e-6*vwidth: // 'random' shift used to avoid singularities

  //-------------------------------
  // initialize the data containers
  //-------------------------------
  POINTS:= table():    // container
  EDGES:= table():     // container
  TRIANGLES:= table(): // container

  NEXTEDGE:= 1:     // initialize edge count
  NEXTTRIANGLE:= 1: // initialize triangle count

  //--------------------------
  // initialize various tables
  //--------------------------
  [fmin, fmax, Fmin, Fmax,
   _umin, _umax, _vmin, _vmax,
   ubbLO, ubbHI, XYZratio
  ] := [table() $ 11]:

  //-------------------
  // utility revertEdge
  //-------------------
  revertEdge:= edge -> [revert(edge[1]), revert(edge[2]), edge[3]]:

  // --------------------------
  // initialize the viewing box
  // --------------------------
  vbox:= [xrange, yrange, zrange];
  for i from 1 to 3 do
    if vbox[i] = Automatic then
       vbox[i]:= Automatic .. Automatic;
    end_if;
  end_for:
  for i from 1 to 3 do
    // initialization for later refinement
    if op(vbox[i], 1) = Automatic then
         ubbLO[i]:= FALSE;    // userboundingbox?
         fmin[i]:=  MAXFLOAT; // The actual viewing box found during the evaluation
         Fmin[i]:= -MAXFLOAT; // The viewing box to be passed to the renderer
    else ubbLO[i]:= TRUE:     // userboundingbox?
         fmin[i]:= op(vbox[i], 1):
         Fmin[i]:= fmin[i];   // The bounding box requested by the user
    end_if;
    if op(vbox[i], 2) = Automatic then
         ubbHI[i]:= FALSE;    // userboundingbox?
         fmax[i]:= -MAXFLOAT; // The actual viewing box found during the evaluation
         Fmax[i]:=  MAXFLOAT; // The viewing box to be passed to the renderer
    else
         ubbHI[i]:= TRUE;     // userboundingbox?
         fmax[i]:= op(vbox[i], 2):
         Fmax[i]:= fmax[i];   // The bounding box requested by the user
    end_if;
  end_for:

  //---------------------------------------------
  // The parametrization functions of the surface
  //---------------------------------------------
  F:= [float@fX, float@fY, float@fZ];

   //-----------------------------------------------------
   // utility getPoint(u,v): try to evaluate X(u,v), Y(u,v), Z(u,v)
   // If the result consist of real numbers, assign them to the
   // variables x = [x[1],x[2],x[3]](of adaptiveSurfaceEval) 
   // (SIDE EFFECT!!) and return TRUE.
   // If an error occurs, FALSE is retuned.
   // If a non-real value X(u,v) or Y(u,v) or Z(u,v) is found, it
   // is ignored and FALSE is returned.
   // The call getPoint(u,v) tracks fmin/max[i] ==  xmin, .. , zmax
   // as a SIDE EFFECT.
   // The call getPoint(u,v, any_third_argument) does not track
   // xmin, xmax, ymin, ymax, zmin, zmax
   //-----------------------------------------------------
   getPoint:= proc(u,v) 
   // assigns x:= [X(u,v),Y(u,v),Z(u,v)] as a side effect
   local i;
   begin
     x:= [0 $ 3]; // this is a variable of plot::adaptiveSurfaceEval!
     if traperror(((x[i]:= F[i](u,v)) $ i=1..3)) <> 0 then
//        singularities:= singularities union {[u,v]};
          return(FALSE)
     elif _lazy_and(
             _lazy_or(domtype(x[i]) = DOM_FLOAT,
                      domtype((x[i]:= numeric::complexRound(x[i]))) = DOM_FLOAT)
             $ i = 1..3,
             x[i] >= -MAXFLOAT $ i = 1..3,
             x[i] <=  MAXFLOAT $ i = 1..3,
             _lazy_or((iszero(x[i]),
                       x[i] >= MINFLOAT,
                       x[i] <=-MINFLOAT) $ i = 1..3)
            ) then
//        if args(0) = 2 then
          trackRange(u,v,x);
//        end_if;
          return(TRUE)
     else return(FALSE);
     end_if;
   end_proc:
   //------------------------------------------------------
   // utility trackRange(u,v,x): check the global extremal
   // values fmin/max[i] (== xmin, ... , zmax) and redefine
   // them if necessary.
   //------------------------------------------------------
   trackRange:= proc(u,v,x)
   local i;
   begin
     for i from 1 to 3 do
       if x[i] < fmin[i] then 
          fmin[i]:= x[i]; // the value of the minimum
          _umin[i]:= u;   // the location of the minimum
          _vmin[i]:= v;   // the location of the minimum
       end_if;
       if x[i] > fmax[i] then 
          fmax[i]:= x[i]; // the value of the maximum
          _umax[i]:= u;   // the location of the maximum
          _vmax[i]:= v;   // the location of the maximum
       end_if;
     end_for:
   end_proc:

  //---------------------------------------------
  // Evaluation on the coarse mesh
  //---------------------------------------------

//userinfo(3, "starting evaluation on the initial regular mesh. ");
  ST:= time():
  proc()
    local line_index;
  name plot::Surface::evalOnRegularMesh;
  begin
    if umesh = 1 then
      umesh:= 2;
    end_if;
    um1:= umesh - 1:
    if vmesh = 1 then
      vmesh:= 2;
    end_if;
    vm1:= vmesh - 1:
    for i from 1 to vmesh do
      // Beware: do not change this computation of v. 
      // It prevents roundoff effects!
      v:= ((vmesh - i)/vm1)*vmin + ((i - 1)/vm1)*vmax;
      for j from 1 to umesh do
        // Beware: do not change this computation of u. 
        // It prevents roundoff effects!
        u:= ((umesh - j)/um1)*umin + ((j - 1)/um1)*umax;
        // x = [X, Y, Z] is assigned as a side effect
        if   getPoint(u   ,v   ) then POINTS[j+(i-1)*umesh]:= [u   , v   ,op(x)];
        elif getPoint(u+du,v   ) then POINTS[j+(i-1)*umesh]:= [u+du, v   ,op(x)];
        elif getPoint(u-du,v   ) then POINTS[j+(i-1)*umesh]:= [u-du, v   ,op(x)];
        elif getPoint(u   ,v+dv) then POINTS[j+(i-1)*umesh]:= [u   , v+dv,op(x)];
        elif getPoint(u   ,v-dv) then POINTS[j+(i-1)*umesh]:= [u   , v-dv,op(x)];
        elif getPoint(u+du,v+dv) then POINTS[j+(i-1)*umesh]:= [u+du, v+dv,op(x)];
        elif getPoint(u+du,v-dv) then POINTS[j+(i-1)*umesh]:= [u+du, v-dv,op(x)];
        elif getPoint(u-du,v+dv) then POINTS[j+(i-1)*umesh]:= [u-du, v+dv,op(x)];
        elif getPoint(u-du,v-dv) then POINTS[j+(i-1)*umesh]:= [u-du, v-dv,op(x)];
        else
           // ToDo
           error("cannot evaluate to a real numerical value near the point ".expr2text(u, v));
        end_if;
      end_for;
    end_for;
    // The points on the regular mesh are computed. 
    // Collect them to edges and triangles
    for i from 1 to vmesh-1 do
      for j from 1 to umesh-1 do
                                           //  p3---e5---p4
        p1:= j + (i-1)*umesh;              //   |\_      |
        p2:= p1 + 1;                       //  e3  \e2_  e4
        p3:= p1 + umesh; // = j + i*umesh; //   |      \ |
        p4:= p3 + 1;                       //  p1---e1---p2
        //---------------------------------------------------------------------
        // insert lower left triangle with corners p1, p2, p3
        //---------------------------------------------------------------------
        e1:= nops(EDGES)+1; EDGES[e1] := [[p1,p2], [], 0];
        e2:= nops(EDGES)+1; EDGES[e2] := [[p2,p3], [], 0];
	e3:= nops(EDGES)+1; EDGES[e3] := [[p3,p1], [], 0];
	if modp(j-1, 1+usubmesh)=0 then
	  line_index := ((j-1) div (1+usubmesh))+1;
	  ULINES[line_index] := ULINES[line_index] union {e3};
	end_if;
	if modp(i-1, 1+vsubmesh)=0 then
	  line_index := ((i-1) div (1+vsubmesh))+1;
	  VLINES[line_index] := VLINES[line_index] union {e1};
	end_if;
        NEXTEDGE:= NEXTEDGE + 3:
        TRIANGLES[NEXTTRIANGLE]:= [e1, e2, e3]:
        NEXTTRIANGLE:= NEXTTRIANGLE + 1:
        //---------------------------------------------------------------------
        // insert upper right triangle with corners p2, p4, p3
        //---------------------------------------------------------------------
        if j < umesh - 1 then
             e4:= e3 + 3;
        else e4:= e3 + 1;
	  EDGES[e4] := [[p4,p2], [], 0];
	  NEXTEDGE:= NEXTEDGE + 1:
	  line_index := nops(ULINES);
	  ULINES[line_index] := ULINES[line_index] union {e4};
        end_if;
        if i < vmesh - 1 then
             e5:= e1 + 3*umesh - 2;
        else e5:= e1 + 3*(umesh - 1 - j) + 3 + j;
        end_if;
        TRIANGLES[NEXTTRIANGLE]:= [[e4], [e5], [e2]]:
        NEXTTRIANGLE:= NEXTTRIANGLE + 1:
      end_for;
    end_for:
    i:= vmesh;
    assert(modp(i-1, 1+vsubmesh)=0);
    line_index := nops(VLINES);
    for j from 1 to umesh -1 do
        p1:= j + (i-1)*umesh;
        p2:= p1 + 1;
        e1:= nops(EDGES)+1; EDGES[e1] := [[p1,p2], [], 0];
        NEXTEDGE:= NEXTEDGE + 1:
        VLINES[line_index] := VLINES[line_index] union {e1};
    end_for:
  end_proc();

  userinfo(3, "evaluation on the initial regular mesh finished. ".
              "Took: ".expr2text(time() - ST)." msec");
//------------------------------------------
// For triangle = [edge1, edge2, edge3] with 
// edge.i = [pt.i.1, pt.i.2], check that the 
// points do indeed form a closed polygon.
//------------------------------------------
checktriangle:= proc(triangle)
local ne1, ne2, ne3, e1, e2, e3, p1, p2, p3, p4, p5, p6;
begin
     [ne1, ne2, ne3]:= TRIANGLES[triangle]:
     if domtype(ne1) = DOM_INT then
           e1:= EDGES[ne1]; 
     else ne1:= op(ne1);
           e1:= revertEdge(EDGES[ne1]);
     end_if;
     if domtype(ne2) = DOM_INT then
           e2:= EDGES[ne2]; 
     else ne2:= op(ne2);
           e2:= revertEdge(EDGES[ne2]);
     end_if;
     if domtype(ne3) = DOM_INT then
           e3:= EDGES[ne3]; 
     else ne3:= op(ne3);
           e3:= revertEdge(EDGES[ne3]);
     end_if;
     assert(ne1 <> ne2 and ne2 <> ne3 and ne1 <> ne3);
     [p1, p2]:= e1[1]:
     [p3, p4]:= e2[1]:
     [p5, p6]:= e3[1]:
     assert(p2 = p3 and p4 = p5 and p6 = p1);
end_proc:
//-------------------------------------------------------
// Do check the triangles created by evalOnRegularMesh;
//-------------------------------------------------------
userinfo(10, "time for checking initial triangles = ".
             expr2text( time((
                               for triangle in TRIANGLES do
                                  checktriangle(op(triangle, 1));
                               end_for:
                               ))). " msec");
   userinfo(3, expr2text(nops(POINTS)). " points, ".
               expr2text(nops(EDGES))." edges, ".
               expr2text(nops(TRIANGLES))." triangles");
   //-------------------------------------------------------

// begin 'just for debugging'
  //---------------------------------------------
  // Evaluation on the coarse mesh is finished.
  // First estimation of the viewing box:
  //---------------------------------------------
  tmp:= {op(pt,2)[3] $ pt in POINTS};
  [xmin, xmax]:= [min(op(tmp)), max(op(tmp))];
  tmp:= {op(pt,2)[4] $ pt in POINTS};
  [ymin, ymax]:= [min(op(tmp)), max(op(tmp))];
  tmp:= {op(pt,2)[5] $ pt in POINTS};
  [zmin, zmax]:= [min(op(tmp)), max(op(tmp))];

  if not ubbLO[1] then
     assert(xmin = fmin[1]):
  end_if;
  if not ubbHI[1] then
     assert(xmax = fmax[1]):
  end_if;
  if not ubbLO[2] then
     assert(ymin = fmin[2]):
  end_if;
  if not ubbHI[2] then
     assert(ymax = fmax[2]):
  end_if;
  if not ubbLO[3] then
     assert(zmin = fmin[3]):
  end_if;
  if not ubbHI[3] then
     assert(zmax = fmax[3]):
  end_if;
// end 'just for debugging'

  //-----------------------------------------------------------
  // edges with length^2 <= minlength2 will not be subdivided. 
  // Note that relative measures are used (think of the objects
  // as being scaled to a cubic viewing box of length 1 in
  // each direction.
  //-----------------------------------------------------------
  minlength2:= 1e-4;

  //---------------------------------------------
  // utility subdivide(edge)
  //---------------------------------------------
  subdivide:= proc(edge)
  local p1, p2, p3, u1, u2, u3, v1, v2, v3,
        x1, x2, x3, y1, y2, y3, z1, z2, z3,
        du, dv, distlr2, pt3, bend;
  name plot::Surface::subdivide;
  begin
      // edge = [[p1, p2], [], subdivlevel] 
      if edge[3] >= MAXSUBDIVLEVEL then
         return(edge);
      end_if;
      if nops(edge[1]) <> 2 then 
         return(subsop(edge, 3 = MAXSUBDIVLEVEL));
      end_if;
      [p1, p2]:= edge[1]:
      [u1, v1, x1, y1, z1]:= POINTS[p1][1..5]:
      [u2, v2, x2, y2, z2]:= POINTS[p2][1..5]:
      distlr2 :=   (x1 - x2)^2/XYZratio2[1]
                 + (y1 - y2)^2/XYZratio2[2]
                 + (z1 - z2)^2/XYZratio2[3];
      if distlr2 <= minlength2 then
         return(subsop(edge, 3 = MAXSUBDIVLEVEL));
      end_if;
      // the interpolated middle:
      [u3, v3, x3, y3, z3]:= [(u1 + u2)/2, (v1 + v2)/2,
                              (x1 + x2)/2, (y1 + y2)/2, (z1 + z2)/2];
      // the exact middle:
      // getPoint assigns x = [X,Y,Z] as a side effect
      du:= 0.849187*uwidth/10^6; // choose a du different to the one in 
      dv:= 0.849187*vwidth/10^6; // evalOnRegularMesh to avoid hitting the singularity 
      if   getPoint(u3     , v3     ) then pt3:= [u3     , v3,      op(x)];
      elif getPoint(u3 + du, v3     ) then pt3:= [u3 + du, v3,      op(x)];
      elif getPoint(u3 - du, v3     ) then pt3:= [u3 - du, v3,      op(x)];
      elif getPoint(u3     , v3 + dv) then pt3:= [u3     , v3 + dv, op(x)];
      elif getPoint(u3     , v3 - dv) then pt3:= [u3     , v3 - dv, op(x)];
      elif getPoint(u3 + du, v3 + dv) then pt3:= [u3 + du, v3 + dv, op(x)];
      elif getPoint(u3 + du, v3 - dv) then pt3:= [u3 + du, v3 + dv, op(x)];
      elif getPoint(u3 - du, v3 + dv) then pt3:= [u3 - du, v3 + dv, op(x)];
      elif getPoint(u3 - du, v3 - dv) then pt3:= [u3 - du, v3 - dv, op(x)];
      else
         // ToDo
         error("cannot evaluate to a finite real value near the point ".expr2text(u3, v3));
      end_if;
      bend:= ((x3 - pt3[3])^2/XYZratio2[1]
            + (y3 - pt3[4])^2/XYZratio2[2]
            + (z3 - pt3[5])^2/XYZratio2[3])/distlr2:
      if bend <= MAXBEND then // do not introduce a mid point
         return(subsop(edge, 3 = MAXSUBDIVLEVEL));
      end_if;
      WORKTODO:= TRUE: // the flag WORKTODO informs splitTriangle
                       // that at least one edge was subdivided
      p3:= nops(POINTS) + 1: // mark for splitting at the mid point
      POINTS[p3]:= pt3;
      [[p1, p3, p2], [], edge[3] + 1];
  end_proc:

  //---------------------------------------------
  // utility splitTriangle(triangle)
  //---------------------------------------------
  splitTriangle:= proc(triangle)
  local e1, e2, e3, ne1, ne2, ne3, rev1, rev2, rev3,
        p1, p2, p3, p4, p5, p6, q1, q2, q3,
        ne14, ne24, ne25, ne35, ne36, ne42, ne43, ne45, ne46, ne51, ne54, ne53, ne61, ne65,
        b, subdivlevel, splitUVedge
       ;
  begin
     [ne1, ne2, ne3]:= TRIANGLES[triangle]:
     if domtype(ne1) = DOM_INT then
          e1:= EDGES[ne1]; 
          rev1:= FALSE;
     else ne1:= op(ne1);
          e1:= revertEdge(EDGES[ne1]);
          rev1:= TRUE;
     end_if;
     if domtype(ne2) = DOM_INT then
          e2:= EDGES[ne2]; 
          rev2:= FALSE;
     else ne2:= op(ne2);
          e2:= revertEdge(EDGES[ne2]);
          rev2:= TRUE;
     end_if;
     if domtype(ne3) = DOM_INT then
          e3:= EDGES[ne3]; 
          rev3:= FALSE;
     else ne3:= op(ne3);
          e3:= revertEdge(EDGES[ne3]);
          rev3:= TRUE;
     end_if;
     assert(ne1 <> ne2 and ne2 <> ne3 and ne1 <> ne3);
    
    splitUVedge := proc(old, new1, new2)
		   begin
		     ULINES := subs(ULINES, old=(new1, new2));
		     VLINES := subs(VLINES, old=(new1, new2));
		   end_proc:
    
     //--------------------------------------------------------------
     // combine the cases that are equal up to permutation
     //--------------------------------------------------------------
     case [nops(e1[1]), nops(e2[1]), nops(e3[1])]
     of [2, 2, 2] do 
        return();
     //-----------------------------------
     of [2, 3, 2] do  // reduce to [3, 2, 2]
        [e1,   e2,  e3]:= [ e2,  e3,  e1];
        [ne1, ne2, ne3]:= [ne2, ne3, ne1];
        [rev1, rev2, rev3]:= [rev2, rev3, rev1];
        break;
     of [2, 2, 3] do // reduce to [3, 2, 2]
        [e1,   e2,  e3]:= [ e3,  e1,  e2];
        [ne1, ne2, ne3]:= [ne3, ne1, ne2];
        [rev1, rev2, rev3]:= [rev3, rev1, rev2];
        break;
     of [3, 2, 2] do
        break;
     //-----------------------------------
     of [3, 3, 2] do // reduce to [2, 3, 3]
        [ e1,  e2,  e3]:= [ e3,  e1,  e2];
        [ne1, ne2, ne3]:= [ne3, ne1, ne2];
        [rev1, rev2, rev3]:= [rev3, rev1, rev2];
        break;
     of [3, 2, 3] do // reduce to [2, 3, 3]
        [e1, e2, e3]:=    [ e2,  e3,  e1];
        [ne1, ne2, ne3]:= [ne2, ne3, ne1];
        [rev1, rev2, rev3]:= [rev2, rev3, rev1];
        break;
     of [2, 3, 3] do
        break;
     //-----------------------------------
     of [3, 3, 3] do
        break;
     end_case;

     // the recursive division level of the edges
     subdivlevel[1]:= e1[3]:
     subdivlevel[2]:= e2[3]:
     subdivlevel[3]:= e3[3]:

     //--------------------------------------------------------------
     // If an edge has not been touched before in splitTriangle, its
     // 2nd entry is []. Split such an edge into 2 subedges and write 
     // the indices of the subedges into the edge.
     // If the edge was touched before, its 2nd entry is a list of 
     // 2 subedge indices. Pick these subedges and delete the edge.
     //--------------------------------------------------------------

     case [nops(e1[1]), nops(e2[1]), nops(e3[1])]
     of [2, 2, 2] do
        //------------
        //     p3 
        //    / \
        //  p1---p2  
        //------------
        // nothing to be done
        return();
     //---------------------------------------------
     //---------------------------------------------
//   of [2, 3, 2] do // reduced to [3, 2, 2] above
//   of [2, 2, 3] do // reduced to [3, 2, 2] above
     of [3, 2, 2] do
        //----------------
        //    /\  /\
        //   /  p3  \
        //  e3 /| \ e2
        // /  / e4 \  \
        // \ /  |   \ / 
        //  p1--p4--p2  
        //  |___e1___|
        //----------------
        [p1, p4, p2]:= e1[1]:
        [q2, p3]:= e2[1]:
        [q3, q1]:= e3[1]:
        assert(p1 = q1 and p2 = q2 and p3 = q3);
        //--------------------------------------------------------------
        // add the outer edges
        //--------------------------------------------------------------
        if nops(e1[2]) = 0 then
             ne14:= NEXTEDGE: EDGES[ne14]:= [[p1, p4], [], subdivlevel[1]]: 
             ne42:= ne14 + 1: EDGES[ne42]:= [[p4, p2], [], subdivlevel[1]]:
             NEXTEDGE:= NEXTEDGE+2;
             if rev1 then
                  EDGES[ne1]:= subsop(EDGES[ne1], 2 = [ne42, ne14]);
             else EDGES[ne1]:= subsop(EDGES[ne1], 2 = [ne14, ne42]);
             end_if;
        else [ne14, ne42]:= e1[2]; 
             if EDGES[ne14][1] <> p1 then ne14:= [ne14]: end_if;
             if EDGES[ne42][2] <> p2 then ne42:= [ne42]: end_if;
             delete EDGES[ne1]; 
	end_if;
	// adjust ULINES and VLINES, if they contain e1:
        splitUVedge(ne1, ne14, ne42);
        if nops(e2[2]) <> 0 then
             error("should not arrive here"):
        elif rev2 then ne2:= [ne2];
        end_if;
        if nops(e3[2]) <> 0 then
             error("should not arrive here"):
        elif rev3 then ne3:= [ne3];
        end_if;
        //--------------------------------------------------------------
        // add the inner edge
        //--------------------------------------------------------------
        ne43:= NEXTEDGE: EDGES[ne43]:= [[p4, p3], [], subdivlevel[1]]:
        NEXTEDGE:= NEXTEDGE+1;
        //--------------------------------------------------------------
        // delete the old triangle, add the new triangles
        //--------------------------------------------------------------
        delete TRIANGLES[triangle]:
        TRIANGLES[NEXTTRIANGLE  ]:= [ ne14 ,  ne43 ,  ne3  ]:   
        TRIANGLES[NEXTTRIANGLE+1]:= [ ne42 ,  ne2  , [ne43]]: 
        checktriangle(NEXTTRIANGLE);
        checktriangle(NEXTTRIANGLE + 1);
        NEXTTRIANGLE:= NEXTTRIANGLE + 2;
        break;
     //---------------------------------------------
     //---------------------------------------------
//   of [3, 3, 2] do // reduced to [2, 3, 3] above
//   of [3, 2, 3] do // reduced to [2, 3, 3] above
     of [2, 3, 3] do
        //----------------------------------
        //     /\  /\             /\  /\
        //    /  p3  \           /  p3  \
        //   e3 /  \ e2         e3 /  \ e2
        //  /  p5--p4  \  bzw. /  p5--p4  \
        //  \ / __/  \ /       \ /  \__ \ /
        //   p1/_____p2         p1_____\p2
        //   |___e1___|         |___e1___|
        //----------------------------------
        [p1, p2]:= e1[1]:
        [q2, p4, p3]:= e2[1]:
        [q3, p5, q1]:= e3[1]:
        assert(p1 = q1 and p2 = q2 and p3 = q3);
        //--------------------------------------------------------------
        // add the outer edges
        //--------------------------------------------------------------
        if nops(e1[2]) <> 0 then
             error("should not arrive here");
        elif rev1 then ne1:= [ne1]; 
        end_if;
        if nops(e2[2]) = 0 then
             ne24:= NEXTEDGE: EDGES[ne24]:= [[p2, p4], [], subdivlevel[2]]:
             ne43:= ne24 + 1: EDGES[ne43]:= [[p4, p3], [], subdivlevel[2]]:
             NEXTEDGE:= NEXTEDGE+2;
             if rev2 then
                  EDGES[ne2]:= subsop(EDGES[ne2], 2 = [ne43, ne24]);
             else EDGES[ne2]:= subsop(EDGES[ne2], 2 = [ne24, ne43]);
             end_if;
        else [ne24, ne43]:= e2[2]; 
             if EDGES[ne24][1] <> p2 then ne24:= [ne24]: end_if;
             if EDGES[ne43][2] <> p3 then ne43:= [ne43]: end_if;
             delete EDGES[ne2]; 
        end_if;
	// adjust ULINES and VLINES, if they contain e2:
        splitUVedge(ne2, ne24, ne43);
        if nops(e3[2]) = 0 then
             ne35:= NEXTEDGE: EDGES[ne35]:= [[p3, p5], [], subdivlevel[3]]:
             ne51:= ne35 + 1: EDGES[ne51]:= [[p5, p1], [], subdivlevel[3]]:
             NEXTEDGE:= NEXTEDGE+2;
             if rev3 then
                  EDGES[ne3]:= subsop(EDGES[ne3], 2 = [ne51, ne35]);
             else EDGES[ne3]:= subsop(EDGES[ne3], 2 = [ne35, ne51]);
             end_if;
        else [ne35, ne51]:= e3[2]; 
             if EDGES[ne35][1] <> p3 then ne35:= [ne35]: end_if;
             if EDGES[ne51][2] <> p1 then ne51:= [ne51]: end_if;
             delete EDGES[ne3]; 
        end_if;
	// adjust ULINES and VLINES, if they contain e3:
        splitUVedge(ne3, ne35, ne51);
        //--------------------------------------------------------------
        // add the 2 inner edges
        //--------------------------------------------------------------
        // Introduce the shorter of the edges p1-p4 and p2-p5:
        if   (POINTS[p5][3] - POINTS[p2][3])^2/XYZratio2[1] +
             (POINTS[p5][4] - POINTS[p2][4])^2/XYZratio2[2] +
             (POINTS[p5][5] - POINTS[p2][5])^2/XYZratio2[3] 
          >= (POINTS[p4][3] - POINTS[p1][3])^2/XYZratio2[1] +
             (POINTS[p4][4] - POINTS[p1][4])^2/XYZratio2[2] +
             (POINTS[p4][5] - POINTS[p1][5])^2/XYZratio2[3] then
           ne14:= NEXTEDGE: EDGES[ne14]:= [[p1, p4], [], subdivlevel[2]]:
           ne45:= ne14 + 1: EDGES[ne45]:= [[p4, p5], [], min(subdivlevel[2],subdivlevel[3])]:
           b:= FALSE;
        else
           ne25:= NEXTEDGE: EDGES[ne25]:= [[p2, p5], [], subdivlevel[3]]:
           ne45:= ne25 + 1: EDGES[ne45]:= [[p4, p5], [], min(subdivlevel[2],subdivlevel[3])]:
           b:= TRUE;
        end_if;
        NEXTEDGE:= NEXTEDGE+2;
        //--------------------------------------------------------------
        // delete the old triangle, add the new triangles
        //--------------------------------------------------------------
        delete TRIANGLES[triangle]:
        if b then
             TRIANGLES[NEXTTRIANGLE+1]:= [  ne1 ,  ne25,  ne51 ]: 
             TRIANGLES[NEXTTRIANGLE  ]:= [ ne24 ,  ne45, [ne25]]:   
             TRIANGLES[NEXTTRIANGLE+2]:= [[ne45],  ne43,  ne35 ]: 
        else TRIANGLES[NEXTTRIANGLE  ]:= [  ne1 ,  ne24, [ne14]]: 
             TRIANGLES[NEXTTRIANGLE+1]:= [ ne14 ,  ne45,  ne51 ]:   
             TRIANGLES[NEXTTRIANGLE+2]:= [[ne45],  ne43,  ne35 ]: 
        end_if;
        checktriangle(NEXTTRIANGLE);
        checktriangle(NEXTTRIANGLE + 1);
        checktriangle(NEXTTRIANGLE + 2);
        NEXTTRIANGLE:= NEXTTRIANGLE + 3;
        break;
     //---------------------------------------------
     //---------------------------------------------
     of [3, 3, 3] do
        //----------------
        //    /\    /\
        //   /   p3  \
        //  e3  /  \  e2
        // /  p6 -- p5  \
        // \ /  \  /  \ / 
        //  p1---p4---p2  
        //  |____e1____|
        //----------------
        [p1, p4, p2]:= e1[1]:
        [q2, p5, p3]:= e2[1]:
        [q3, p6, q1]:= e3[1]:
        assert(p1 = q1 and p2 = q2 and p3 = q3);
        //--------------------------------------------------------------
        // add the outer edges
        //--------------------------------------------------------------
        if nops(e1[2]) = 0 then
             ne14:= NEXTEDGE: EDGES[ne14]:= [[p1, p4], [], subdivlevel[1]]: 
             ne42:= ne14 + 1: EDGES[ne42]:= [[p4, p2], [], subdivlevel[1]]:
             NEXTEDGE:= NEXTEDGE+2;
             if rev1 then
                  EDGES[ne1]:= subsop(EDGES[ne1], 2 = [ne42, ne14]);
             else EDGES[ne1]:= subsop(EDGES[ne1], 2 = [ne14, ne42]);
             end_if;
        else [ne14, ne42]:= e1[2]; 
             if EDGES[ne14][1] <> p1 then ne14:= [ne14]: end_if;
             if EDGES[ne42][2] <> p2 then ne42:= [ne42]: end_if;
             delete EDGES[ne1]; 
        end_if;
	// adjust ULINES and VLINES, if they contain e1:
        splitUVedge(ne1, ne14, ne42);
        if nops(e2[2]) = 0 then
             ne25:= NEXTEDGE: EDGES[ne25]:= [[p2, p5], [], subdivlevel[2]]:
             ne53:= ne25 + 1: EDGES[ne53]:= [[p5, p3], [], subdivlevel[2]]:
             NEXTEDGE:= NEXTEDGE+2;
             if rev2 then
                  EDGES[ne2]:= subsop(EDGES[ne2], 2 = [ne53, ne25]);
             else EDGES[ne2]:= subsop(EDGES[ne2], 2 = [ne25, ne53]);
             end_if;
        else [ne25, ne53]:= e2[2]; 
             if EDGES[ne25][1] <> p2 then ne25:= [ne25]: end_if;
             if EDGES[ne53][2] <> p3 then ne53:= [ne53]: end_if;
             delete EDGES[ne2]; 
        end_if;
	// adjust ULINES and VLINES, if they contain e2:
        splitUVedge(ne2, ne25, ne53);
        if nops(e3[2]) = 0 then
             ne36:= NEXTEDGE: EDGES[ne36]:= [[p3, p6], [], subdivlevel[3]]:
             ne61:= ne36 + 1: EDGES[ne61]:= [[p6, p1], [], subdivlevel[3]]:
             NEXTEDGE:= NEXTEDGE+2;
             if rev3 then
                  EDGES[ne3]:= subsop(EDGES[ne3], 2 = [ne61, ne36]);
             else EDGES[ne3]:= subsop(EDGES[ne3], 2 = [ne36, ne61]);
             end_if;
        else [ne36, ne61]:= e3[2]; 
             if EDGES[ne36][1] <> p3 then ne36:= [ne36]: end_if;
             if EDGES[ne61][2] <> p1 then ne61:= [ne61]: end_if;
             delete EDGES[ne3]; 
        end_if;
	// adjust ULINES and VLINES, if they contain e3:
        splitUVedge(ne3, ne36, ne61);
        //--------------------------------------------------------------
        // add the 3 inner edges
        //--------------------------------------------------------------
        ne46:= NEXTEDGE: EDGES[ne46]:= [[p4, p6], [], min(subdivlevel[1],subdivlevel[3])]:
        ne54:= ne46 + 1: EDGES[ne54]:= [[p5, p4], [], min(subdivlevel[1],subdivlevel[2])]:
        ne65:= ne54 + 1: EDGES[ne65]:= [[p6, p5], [], min(subdivlevel[2],subdivlevel[3])]:
        NEXTEDGE:= NEXTEDGE+3;
        //--------------------------------------------------------------
        // delete the old triangle, add the new triangles
        //--------------------------------------------------------------
        delete TRIANGLES[triangle]:
        TRIANGLES[NEXTTRIANGLE  ]:= [ ne14 ,  ne46 ,  ne61 ]:   
        TRIANGLES[NEXTTRIANGLE+1]:= [ ne42 ,  ne25 ,  ne54 ]: 
        TRIANGLES[NEXTTRIANGLE+2]:= [ ne53 ,  ne36 ,  ne65 ]: 
        TRIANGLES[NEXTTRIANGLE+3]:= [[ne54], [ne65], [ne46]]: 
        checktriangle(NEXTTRIANGLE);
        checktriangle(NEXTTRIANGLE + 1);
        checktriangle(NEXTTRIANGLE + 2);
        checktriangle(NEXTTRIANGLE + 3);
        NEXTTRIANGLE:= NEXTTRIANGLE + 4;
        break;
     otherwise
        error("should not arrive here");
     end_case;
  end_proc:

  //-----------------------------------------------------
  // This routine does the adaptive work: 
  //-----------------------------------------------------
  refine:= proc() begin
    for subdivlevel from 1 to MAXSUBDIVLEVEL do
        WORKTODO:= FALSE; // may be reset by subdivide
        EDGES:= map(EDGES, subdivide); // mark the edges to be subdivided
        context(hold(userinfo)(3, output::ordinal(subdivlevel). " subdivision of the edges. ".
                 "Took: ".expr2text(time() - ST)." msec"));
        // subdivide set the flag WORKTODO to FALSE if no edge was subdivided
        if not WORKTODO then
           break;
        end_if;
        ST:= time():
        for triangle in TRIANGLES do
            splitTriangle(op(triangle, 1));
        end_for:
        context(hold(userinfo)(3, output::ordinal(subdivlevel). 
                               " subdivision of the triangles. ".
                               "Took: ".expr2text((time() - ST))." msec"));
        ST:= time():
        context(hold(userinfo)(3, expr2text(nops(POINTS)). " points, ".
                                  expr2text(nops(EDGES))." edges, ".
                                  expr2text(nops(TRIANGLES))." triangles"));
    end_for:
  end_proc:

  //---------------------------------------------------------
  // utility to decide whether the extrema are singularities.
  // The idea is that the second derivative near a singularity
  // must have the same sign as the function value.
  // We just need the sign of the second derivative along the
  // steepest descent direction (the gradient of fproc).
  // fproc : the function (procedure) to be investigated
  // f1    : fproc(u1, v1)
  // u1, v1: the coordinates of the point to be investigated
  //---------------------------------------------------------
  negativeCurvature:= proc(fproc, f1, u1, v1)
  local eps, u0, v0, fu, fv, tmp, u2, v2, f0, f2;
  save DIGITS;
  begin
     eps:=  0.975*10^(2- DIGITS);
     u0 := u1 - eps*uwidth;
     v0 := v1 - eps*vwidth;
     if traperror((
        fu:= (fproc(u0, v1) - f1)/uwidth; // u-derivative
        fv:= (fproc(u1, v0) - f1)/vwidth; // v-derivative
        )) = 0 then
        tmp:= max(specfunc::abs(fu), specfunc::abs(fv));
        if iszero(tmp) then
           // zero gradient. The second derivative would be
           // dominated by round-off
           return(UNKNOWN);
        end_if;
     else return(UNKNOWN);
     end_if;
     fu:= fu/tmp; // u-component of gradient(f)
     fv:= fv/tmp; // v-component of gradient(f)
     u0 := u1 - eps*uwidth*fu; // perturb
     v0 := v1 - eps*vwidth*fv; // along the
     u2 := u1 + eps*uwidth*fu; // gradient
     v2 := v1 + eps*vwidth*fv; // direction
     DIGITS:= DIGITS + 10;
     if traperror((f0:= fproc(u0,v0); f2:= fproc(u2,v2))) = 0 or // derivative along gradient
        traperror((f0:= fproc(u0,v1); f2:= fproc(u2,v1))) = 0 or // u-derivative at v1
        traperror((f0:= fproc(u1,v0); f2:= fproc(u1,v2))) = 0 or // v-derivative at u1
        traperror((f0:= fproc(u0,v0); f2:= fproc(u2,v0))) = 0 or // u-derivative at v0
        traperror((f0:= fproc(u0,v0); f2:= fproc(u0,v2))) = 0 or // v-derivative at u0
        traperror((f0:= fproc(u0,v2); f2:= fproc(u2,v2))) = 0 or // u-derivative at v2
        traperror((f0:= fproc(u2,v0); f2:= fproc(u2,v2))) = 0    // v-derivative at u2
        then
        tmp:= 2*f1 - (f0 + f2);
        if iszero(tmp) or {domtype(f0), domtype(f2)} <> {Dom::Float} then
             return(UNKNOWN)
        else return(bool(tmp > 0));
        end_if;
     end_if;
     return(UNKNOWN);
  end_proc;
  
  //----------------------------------------------------------------
  // getAutomaticViewingBox tries to clip the ranges in a smart way.
  // It checks heuristically whether very large/small data should be
  // interpreted as singularities.
  // If not, it falls back to the values xmin, .. , zmax found during
  // the numerical evaluation. These are stored as fmin[i], fmax[i] with
  // i = 1,2,3 Finally, it returns the viewing box Fmin .. Fmax in the
  // i-th direction.
  //----------------------------------------------------------------
  getAutomaticViewingBox:= proc(i) // i = 1 = x-direction
                                   // i = 2 = y-direction
                                   // i = 3 = z-direction
  local eq, fdata, tmp, eq1, eq5, eq9;
  name plot::adaptiveSurfaceEval::getAutomaticViewingBox;
  begin
    ST:= time():
    context(hold(userinfo)(3, "trying to compute a decent viewing box in the ".
                              output::ordinal(i)." direction")):
    if fmin[i] = MAXFLOAT or fmax[i] = -MAXFLOAT then
       // no real & finite plot point was found: this plot is empty
       [Fmin[i], Fmax[i]]:= [float(0) $ 2]:
       return():
    end_if;
    // The following context construct serves for activating the
    // userinfo commands inside this sub procedure of adaptiveSurfaceEval
    // by pretending that the userinfo was called in the context
    // of the calling function, i.e., plot::adaptiveSurfaceEval
    fdata:= map([op(POINTS)], op, [2, 2 + i]);
    if nops(fdata) = 0 then
       return();
    end_if;

    eq:= stats::empiricalQuantile(fdata):
    if not ubbHI[i] then
         eq9:= eq(0.9);
    else eq9:= Fmax[i]; // the user defined viewing box border
    end_if:
    if not ubbLO[i] then
         eq1:= eq(0.1);
    else eq1:= Fmin[i]; // the user defined viewing box border
    end_if:

    if not ubbHI[i] then
      //----------------------------------------------
      // determine Fmax[i]
      //----------------------------------------------
      // Heuristics to decide whether the maximum is a proper maximum or a singularity
      Fmax[i]:= fmax[i];
      assert(fmax[i] = F[i](_umax[i], _vmax[i]));
      if fmax[i] > 0 and negativeCurvature(F[i], fmax[i], _umax[i], _vmax[i]) <> TRUE then
//print("positiveSingularity?");
         // this may be a positive singularity
         if fmax[i] >= eq9 + 100 * (eq9 - eq1) then // this does look like a singularity
//print("positiveSingularity!");
           Fmax[i]:= eq9;
         end_if;
      end_if;
    end_if;
    if not ubbLO[i] then
      //----------------------------------------------
      // determine Fmin[i]
      //----------------------------------------------
      // Heuristics to decide whether the minimum is a proper minimum or a singularity
      Fmin[i]:= fmin[i];
      assert(fmin[i] = F[i](_umin[i], _vmin[i]));
      if fmin[i] < 0 and negativeCurvature(F[i], fmin[i], _umin[i], _vmin[i]) <> FALSE then
//print("negativeSingularity?");
         // this may be a negative singularity
         if fmin[i] <= eq1 - 100 * (eq9 - eq1) then // this does look like a singularity
//print("negativeSingularity!");
           Fmin[i]:= eq1;
         end_if;
      end_if;
    end_if;
    if not (ubbHI[i] and ubbLO[i]) then
      //----------------------------------------------
      // symmetrize Fmin[i], Fmax[i]
      //----------------------------------------------
      // If there is a positive and a negative singularity, symmetrize the
      // viewing box around the 50% quantile eq5 by taking the harmonic
      // mean of its distance to the upper and lower viewingbox bounds
      if Fmax[i] < fmax[i] and Fmin[i] > fmin[i] then
        eq5:= eq(0.5):
        tmp:= sqrt(specfunc::abs(eq5 - Fmin[i])*specfunc::abs(Fmax[i] - eq5));
        if not iszero(tmp)     and
           fmin[i] < eq5 - tmp and
             eq1   > eq5 - tmp and
           fmax[i] > eq5 + tmp and
             eq9   < eq5 + tmp then
             if not ubbLO[i] then
                Fmin[i]:= eq5 - tmp;
             end_if;
             if not ubbHI[i] then
                Fmax[i]:= eq5 + tmp;
             end_if;
        end_if;
      end_if;
    end_if;

    //-----------------------------------------------------
    // automatic viewing box Fmin[i] .. Fmax[i] is computed
    //-----------------------------------------------------

    context(hold(userinfo)(3, "automatic viewing box in the ".
                              output::ordinal(i)." direction: ".
                              expr2text(Fmin[i] .. Fmax[i]).
                              " (original: ".expr2text(fmin[i] .. fmax[i]).
                              "). Took ".expr2text(time()- ST)." msec"
                           )):
  end_proc:

  //--------------------------------------------------------------------
  //-----------------------------------------
  // Estimate the aspect ratios x:y:z needed
  // for the adaptive refinement.
  //-----------------------------------------
  for i from 1 to 3 do
     if ubbLO[i] and ubbHI[i] then
       XYZratio[i]:= Fmax[i] - Fmin[i]; // requested by the user
     elif ubbLO[i] and not ubbHI[i] then
       XYZratio[i]:= fmax[i] - Fmin[i]; 
       if XYZratio[i] < 0 then
          // can only happen if a viewing box was specified and
          // no point in this range was found:
          XYZratio[i]:= 1.0:
       end_if;
     elif (not ubbLO[i]) and ubbHI[i] then
       XYZratio[i]:= Fmax[i] - fmin[i]; 
       if XYZratio[i] < 0 then
          // can only happen if a viewing box was specified and
          // no point in this range was found:
          XYZratio[i]:= 1.0:
       end_if;
     elif (not ubbLO[i]) and (not ubbHI[i]) then
       XYZratio[i]:= fmax[i] - fmin[i]; // from the numerical evaluation
     else
       error("should not arrive here");
     end_if;
     assert(XYZratio[i] >= 0);
     if iszero(XYZratio[i]) then
        XYZratio[i]:= 1.0;
     end_if;
  end_for;
  tmp:= max(XYZratio[1], XYZratio[2], XYZratio[3]):
  XYZratio2:= [(XYZratio[1]/tmp)^2, (XYZratio[2]/tmp)^2, (XYZratio[3]/tmp)^2];

  //--------------------------------------------------------------------
  // Do the adaptive refinement
  //--------------------------------------------------------------------
  refine();
  //--------------------------------------------------------------------
  // call getAutomaticViewingBox
  //--------------------------------------------------------------------
  for i from 1 to 3 do
    if not (ubbLO[i] and ubbHI[i]) then 
       if nosingularities then
            Fmin[i]:= fmin[i];
            Fmax[i]:= fmax[i];
       else getAutomaticViewingBox(i);
       end_if;
    end_if;
    if Fmin[i] > Fmax[i] then 
       // this can only happen if at least one of the
       // viewing box borders is given and no point
       // inside this range was found.
       if ubbLO[i] then // Fmin is binding
          Fmax[i]:= Fmin[i];
       elif ubbHI[i] then // Fmax is binding
          Fmin[i]:= Fmax[i];
       else
          assert(FALSE); // should not arrive here
          // Just to be on the safe side for the release version:
          [Fmin[i], Fmax[i]] := [(Fmin[i] + Fmax[i])/2 $ 2];
       end_if;
    end_if;
  end_for;
  //---------------------------------------------------------------
  // If the aspect ratio changed drastically, we need to recompute.
  // Estimate the new aspect ratios x:y:z after the decision on the
  // viewing box by getAutomaticViewingBox.
  //---------------------------------------------------------------
  recompute:= FALSE;
  oldXYZratio2:= XYZratio2;
  for i from 1 to 3 do
     if not (ubbLO[i] and ubbHI[i]) and
        ( XYZratio[i] < (Fmax[i] - Fmin[i])/10 or
          XYZratio[i] > (Fmax[i] - Fmin[i])*10) then
        XYZratio[i]:= Fmax[i] - Fmin[i]; 
        if iszero(XYZratio[i]) then
           XYZratio[i]:= 1.0;
        end_if;
        recompute:= TRUE:
     end_if;
     assert(XYZratio[i] >= 0);
  end_for:

  //-----------------------------------------
  // do recompute
  //-----------------------------------------
  if recompute then 
     // clear the subdivlevel entry in EDGES
     EDGES:= map(EDGES, subsop, 3 = 0);
     // We need to set the new XYZratio2 need by subdivide (called in refine())
     tmp:= max(XYZratio[1], XYZratio[2], XYZratio[3]):
     XYZratio2:= [(XYZratio[1]/tmp)^2, (XYZratio[2]/tmp)^2, (XYZratio[3]/tmp)^2];

     //--------------------------------------------------------------------
     // Do a second adaptive refinement
     //--------------------------------------------------------------------
     userinfo(3, "the aspect ratios have changed from ".
                  expr2text([oldXYZratio2[i]^(1/2) $ i = 1..3]).
                 " to ".
                  expr2text([XYZratio2[i]^(1/2) $ i = 1..3]));
     userinfo(3, "starting a second adaptive refinement");
     ST:= time();
     refine();
/*
     //--------------------------------------------------------------------
     // 2nd call of getAutomaticViewingBox Problem: the refinement 
     // may have given too much emphasis on the singularities implying
     // that the heuristics in getAutomaticViewingBox breaks down.
     //--------------------------------------------------------------------
     for i from 1 to 3 do
       if not (ubbLO[i] and ubbHI[i]) then 
          if nosingularities then
               Fmin[i]:= fmin[i];
               Fmax[i]:= fmax[i];
          else getAutomaticViewingBox(i);
          end_if;
       end_if;
     end_for;
*/
  end_if;

  //--------------------------------------------------------------------
  //--------------------------------------------------------------------
  //--------------------------------------------------------------------
  // The evaluation of the surface is done
  //--------------------------------------------------------------------
  //--------------------------------------------------------------------
  //--------------------------------------------------------------------

  //----------------------------------------------------------
  // convert TRIANGLE = table() to TRIANGLE = list
  //----------------------------------------------------------
  TRIANGLES:= map([op(TRIANGLES)], op, 2);

  //----------------------------------------------------------
  // convert triangle = [edge1, edge2, edge3] do
  //         triangle = [point1, point2, point3]
  //----------------------------------------------------------
  TRIANGLES:= map(TRIANGLES, proc(edges)
                             local e1, e2, e3;
                             name edges2points;
                             begin
                               [e1, e2, e3]:= edges;
                               if domtype(e1) = DOM_INT then
                                    e1:= EDGES[e1][1][1];
                               else e1:= EDGES[op(e1)][1][2];
                               end_if;
                               if domtype(e2) = DOM_INT then
                                    e2:= EDGES[e2][1][1];
                               else e2:= EDGES[op(e2)][1][2];
                               end_if;
                               if domtype(e3) = DOM_INT then
                                    e3:= EDGES[e3][1][1];
                               else e3:= EDGES[op(e3)][1][2];
                               end_if;
                               assert(e1 <> e2 and e2 <> e3 and e3 <> e1);
                               [e1, e2, e3];
                             end_proc);
  //----------------------------------------------------------
  // Add normals to the point data:
  //     pt = [u,v,x,y,z]   -->   pt = [u,v,x,y,z,nx,ny,nz]
  //----------------------------------------------------------
  // Version 1: compute the normal at a point as the mean
  // of the normals of all triangles sharing this point.
  //----------------------------------------------------------
  add_normals1:= proc()
  local normals, i, j, pt1, pt2, pt3, t21, t31, pt;
  begin
    normals:= [0 $ nops(TRIANGLES)]:
    POINTS := map(POINTS, append, {});
    for i from 1 to nops(TRIANGLES) do
      // add the triangle index as extra info
      // to the points: pt[6] becomes a set of
      // all triangles that share the point pt.
      pt1:= POINTS[TRIANGLES[i][1]]:
      pt2:= POINTS[TRIANGLES[i][2]]:
      pt3:= POINTS[TRIANGLES[i][3]]:
      t21:= [(pt2[j] - pt1[j]) $ j = 3..5]:
      t31:= [(pt3[j] - pt1[j]) $ j = 3..5]:
      // compute the normal of the i-th triangle and
      // store it as the i-th entry of the normals list.
      normals[i]:=[t21[2]*t31[3] - t21[3]*t31[2],
                   t21[3]*t31[1] - t21[1]*t31[3],
                   t21[1]*t31[2] - t21[2]*t31[1]];
      POINTS[TRIANGLES[i][1]][6]:= pt1[6] union {i};
      POINTS[TRIANGLES[i][2]][6]:= pt2[6] union {i};
      POINTS[TRIANGLES[i][3]][6]:= pt3[6] union {i};
    end_for:
    for i from 1 to nops(POINTS) do
      // pt[6] is a set of indices of all triangles that
      // share the point pt. Average the normals of these
      // triangles and add it to the point such that it
      // becomes pt = [u, v, x, z, z, nx, ny, nz].
      // This works properly only if the triangles are
      // oriented in the same fashion!
      pt:= POINTS[i]:
      pt[6]:= _plus(normals[j][i] $ j in pt[6])/nops(pt[6]) $ i = 1..3;
      POINTS[i]:= pt;
    end_for:
  end_proc:
  //-----------------------------------------------------------
  // Version 2: compute some normals via numeric differentiation
  //-----------------------------------------------------------
  add_normals2:= proc()
  local du, dv, umid, vmid, computenormal;
  begin
    du:= 2.516e-6*uwidth: // make sure these perturbations do not
    dv:= 2.516e-6*vwidth: // coincide with the perturbations when
                          // evaluating the function near s singularity
                          // (otherwise we may end up directly on the
                          // singularity)
    if not (iszero(du) or iszero(dv)) then
      umid:= (umin + umax)/2:
      vmid:= (vmin + vmax)/2:
      computenormal:= proc(pt)
      local ndu, ndv, u, v, uu, vv, X, Y, Z, tu, tv;
      name plot::Surface::computenormal;
      begin
        [u, v, X, Y, Z]:= pt[1..5];
        //-----------------------------------------------------
        // Use numeric differentiation only along the boundary
        // of the u-v rectangle. This avoids visible contours
        // in case of closed surfaces.
        // We assume that the points were pre-processed by
        // add_normals1();
        // Remove the following lines to use numeric differentiation
        // for **all** points.
        if u > umin + du and u < umax - du  and
           v > vmin + dv and v < vmax - dv  then
          return(pt);
        end_if:
        //-----------------------------------------------------
        ndu:= du: if u > umid then ndu:= -du; end_if;
        ndv:= dv; if v > vmid then ndv:= -dv; end_if;
        uu:= u + ndu; vv:= v + 0.1*ndv;
        // getPoint(uu, vv) assigns x = [X,Y,Z] as a side effect
        if   getPoint(uu, vv)                             or                          
             getPoint(uu + uwidth/10^9, vv              ) or
             getPoint(uu - uwidth/10^9, vv              ) or
             getPoint(uu,               vv + vwidth/10^9) or
             getPoint(uu,               vv - vwidth/10^9) or
             getPoint(uu + uwidth/10^9, vv + vwidth/10^9) or
             getPoint(uu + uwidth/10^9, vv - vwidth/10^9) or
             getPoint(uu - uwidth/10^9, vv + vwidth/10^9) or
             getPoint(uu - uwidth/10^9, vv - vwidth/10^9) then 
             tu:= [(x[1] - X)/ndu, (x[2] - Y)/ndu, (x[3] - Z)/ndu];
        else // ToDo
             error("cannot evaluate to a finite real value near the point ".expr2text(uu, vv));
        end_if;
        uu:= u + 0.1*ndu; vv:= v + ndv;
        // getPoint(uu, vv) assigns x = [X,Y,Z] as a side effect
        if   getPoint(uu, vv)                             or  
             getPoint(uu + uwidth/10^9, vv              ) or
             getPoint(uu - uwidth/10^9, vv              ) or
             getPoint(uu,               vv + vwidth/10^9) or
             getPoint(uu,               vv - vwidth/10^9) or
             getPoint(uu + uwidth/10^9, vv + vwidth/10^9) or
             getPoint(uu + uwidth/10^9, vv - vwidth/10^9) or
             getPoint(uu - uwidth/10^9, vv + vwidth/10^9) or
             getPoint(uu - uwidth/10^9, vv - vwidth/10^9) then
             tv:= [(x[1] - X)/ndv, (x[2] - Y)/ndv, (x[3] - Z)/ndv];
        else // ToDo
             error("cannot evaluate to a finite real value near the point ".expr2text(uu, vv));
        end_if;
        [op(pt, 1..5),-tu[3]*tv[2] + tu[2]*tv[3],
                      -tu[1]*tv[3] + tu[3]*tv[1],
                      -tu[2]*tv[1] + tu[1]*tv[2]];
      end_proc:
      POINTS:= map(POINTS, computenormal);
      assert(nops(POINTS[1]) = 8):
    end_if:
  end_proc:

  //------------------------------------------------------------
  // Add the normals to the point data
  //------------------------------------------------------------
  add_normals1(); // average triangle normals
  add_normals2(): // numerical differentiation along the u-v boundaries

  //----------------------------------------------
  // debugging utility
  //----------------------------------------------
  checkorientation:= proc(tr)
                     local pt1, pt2, pt3, t21, t31;
                     begin
                        pt1:= POINTS[tr[1]];
                        pt2:= POINTS[tr[2]];
                        pt3:= POINTS[tr[3]];
                        t21:= [pt2[i] - pt1[i] $ i= 1..2];
                        t31:= [pt3[i] - pt1[i] $ i= 1..2];
                        bool(t21[2]*t31[1] - t21[1]*t31[2] <= 0);
                     end_proc;

  assert(_and(checkorientation(triangle) $ triangle in TRIANGLES));
  //---------------------------------------------
  // The adaptive refinement is finished. Return
  // the data.
  //---------------------------------------------
  
  // ULINES, VLINES: from edges to points
  ULINES := map(ULINES, map, i -> (POINTS[EDGES[i][1][1]],
				   POINTS[EDGES[i][1][2]]));
  VLINES := map(VLINES, map, i -> (POINTS[EDGES[i][1][1]],
				   POINTS[EDGES[i][1][2]]));
   userinfo(1, "viewing box = ".expr2text([Fmin[i] .. Fmax[i] $ i = 1..3]));
  [
   [Fmin[i] .. Fmax[i] $ i = 1..3],  // viewing box
   [(POINTS[TRIANGLES[i][j]] $ j=1..3) $ i=1..nops(TRIANGLES)],
   ULINES, VLINES
  ];
end_proc:
//----------------------------------------------------------------
