// $Date: 2004/08/05 14:52:36 $ $Author: stefanw $ $Revision: 1.13 $


/*++
faclib::factor_expr -- factor expression with given factorizer

faclib::factor_expr(p, fac)

p   - polynomial expression over the rationals
fac - factorizer

Returns a list of the form [c, f1, e1,...,fn, en] such that p=c*f1^e1*...
*fn^en and c is the content of p.

fac must be a procedure which takes a rational polynomial q and returns
a list of the form [c, f1, e1,...,fn, en] such that the fi are polynomials
and q = c*f1^e1*...*fn^en.
++*/

faclib::factor_expr:=
proc(p, fac, options = table(MaxDegree = 1): DOM_TABLE)
  local l, oldl, i, j, f, e, c, newp;
begin
  // special cases which may save some time...
  case type(p)
    of DOM_INT do
    of DOM_RAT do
      return([p]);

    of "_power" do
      j:= op(p,2);
      if domtype(j) <> DOM_INT then
        return([1, p, 1])
      end_if;
      l:= faclib::factor_expr(op(p,1), fac, options);
      return([ l[1]^j, (l[2*i], l[2*i+1]*j) $ i=1..(nops(l) div 2) ]);

    of "_mult" do
      c:= FAIL;
      f:= table();
      // find common extension containing all coefficients
      // of all factors
      newp:= polylib::makerat([op(p)], options[MaxDegree]);
      if type(op(newp,[1,1]))<>DOM_POLY then
        // all indets are algebraic over Q, each el. of op(newp,1) is a unit
        return([p])
      end_if;

    // factors of the numerator
      
      for e in op(newp,1) do
        e:= fac(e);
        c:= if c=FAIL then e[1] else c*e[1] end_if;
        if iszero(c) then return( [c] ) end_if;

        // collect factors in f:
        for j from 2 to nops(e) step 2 do
          if contains(f, e[j]) then
            f[e[j]]:= f[e[j]] + e[j+1]
          else
            f[e[j]]:= e[j+1]
          end_if
        end_for;
      end_for;

      // factors of the denominator
      // denoted by negative multiplicities
      
      for e in op(newp,2) do
        e:= fac(e);
        c:= c/ e[1];
        if iszero(c) then
          error("division by zero")
        end_if;

        for j from 2 to nops(e) step 2 do
          if contains(f, e[j]) then
            // this may well happen:
            // the product had not been normalized
            f[e[j]]:= f[e[j]] - e[j+1]
          else
            f[e[j]]:= -e[j+1]
          end_if
        end_for;
      end_for;
        
      // deleting zero entries
      f:=select(f, x-> not iszero(op(x,2)));
      l:= map([ c, op(op(f,i)) $ i=1..nops(f) ], expr);
      repeat
        oldl:= l;
        l:= subs(l, op(newp,3))
      until l=oldl end_repeat; 
      break

  otherwise

      newp:=normal(p);
      if contains({DOM_INT, DOM_RAT, "_power", "_mult"}, type(newp)) then
    // simple case after normalizing, start from top
        return(faclib::factor_expr(newp, fac, options))
      end_if;
      newp:=polylib::makerat(newp, options[MaxDegree]);

      if type(op(newp,1))<>DOM_POLY then
    // all indets are algebraic over Q, p is a unit
        return([p])
      end_if;
    
       // eventually factor using fac
      f:= fac(op(newp,1));
      e:= fac(op(newp,2));
      l:= f . [(op(e,2*i), -op(e,2*i+1)) $ i=1..nops(e) div 2];
      // adjust unit
      l[1]:=l[1]/op(e,1);
        // back substitution:
      
      l:= subs(map(l, expr));
    
      repeat
        oldl:= l;
        l:= subs(l, op(newp,3))
      until l=oldl end_repeat;

          
  end_case;

 
  // adjust signs. After back substitution, previously positive
  // factors may now have
  // a negative sign, therefore we have to exchange signs
  for j from 2 to nops(l) step 2 do
    // if stdlib::hasmsign(l[j]) then
    f:=poly(l[j]);
    if f = FAIL then
      e:= stdlib::hasmsign(l[j])
    else
      e:= sign(lcoeff(f)) = -1
    end_if;
    if e then
      l[j]:= - l[j];
      // if n is even, x^n may be replaced by (-x)^n,
      // otherwise we have to adjust the unit
      if l[j+1] mod 2 = 1 then
        l[1]:= -l[1]
      end_if;
    end_if
  end_for;

  return(l)

end_proc:

