/* --------------------------------------------------------------- 

   rootbound -- returns a bound on the absolute value of a
                polynomials roots.

Call:    polylib::rootbound(p <, option> );

Parameter:  p -- a univariate polynomial (DOM_POLY).  
                 Its coefficients must be integers or rationals 
                 or convertible to floats via MuPADs float command.
            option -- description forthcoming

Return value: a float value.

Synopsis:    
              The return value is an upper bound on the absolute value
              for the roots of the polynomial p, i.e.:
              
              p(x) = 0 ==> |x| <= polylib::rootbound(p)
              
              Three different bounds are calculated, from which
              the smallest one is chosen.
              
              For details see W. Oevel: "Einfuehrung in die numerische
              Mathematik", Springer 1996, pp 69--70.
              
Remark:       The computation uses floating-point arithmetic.
              
Personal remark:  should be improved.

Examples:

  >> polylib::rootbound(poly(x^5+2*x^4+3*x^3+4*x^2+5*x+6,[x]));

                                    4.0
  >> polylib::rootbound(poly(x^5+x^3+x-1,[x]));

                                    2.0
  >> polylib::rootbound(poly((x-3)*(x-1)*(x+5),[x]));

                                8.246211251
  >> polylib::rootbound(poly(x^2+10*x+300,[x]));

                                   30.0

Author: Peter Horn 8-8-2001 hornp@upb.de
--------------------------------------------------------------- */

polylib::rootbound := proc(p : DOM_POLY, flag=FAIL)
local b1, b2, b3, p2, b4, L, n, i, xdiv, A, tmp, tmp2, u, l, t, x;
save DIGITS;
begin
	userinfo(9,"polylib::rootbound called with polynomial of degree ".degree(p));
    
    if testargs() then
        if nops(op(p, 2)) <> 1 then
            error("The polynomial must be univariate.");
        end_if;
        p2 := map(poly2list(p), x -> float(op(x,1)));
        if {op(map (p2, X -> testtype(X,Type::Numeric)))} <> {TRUE} then
            error("The polynomial must have proper numbers as coefficients.");
        end_if;
    end_if;
	
    x := op(op(p,2),1);
    
    case flag
        of FAIL do
        of hold(Traditional) do
            userinfo(9,"calculating the tree 'tradional' rootbounds");
        	// get the coefficient-vector of p 
        	n := degree(p);
        	L := array(0..n, [(abs(coeff(p,i))) $ i=0..n]);
            //DIGITS := max(DIGITS, op(map(L, x -> _if(x=0,1,ceil(log(10,x))))))+3;
        
        	// (a) first bound
        	// max{|p0/pn|, 1+|p1/pn|,...,1+|pn-1/pn|}
        	b1 := max(L[0]/L[n], 1+L[i]/L[n] $ i=1..n-1);
        	userinfo(1,"polylib::rootbound: First bound: ", float(b1));
        
        	// (b) second bound
        	// max{|p0/p1|, 2*|p1/p2|,...,2*|pn-1/pn|}
        	xdiv := proc(a,b) begin if iszero(b) then return(infinity) else
        		return(a/b) end_if; end_proc;
        	b2 := max(xdiv(L[0],L[1]), 2*xdiv(L[i],L[i+1]) $ i=1..n-1);
        	userinfo(1,"polylib::rootbound: Second bound: ", float(b2));
        	
        	// (c) third bound
        	// 2*max{|pn-1/pn|, |p0/pn|^(1/n),...,|pn-2/pn|^(1/2)}
        	b3 := 2*max(float(L[n-1]/L[n]), float(L[i]/L[n])^(1/(n-i)) $ i=0..n-2);
        	userinfo(1,"polylib::rootbound: Third bound: ", float(b3));
        	
        	// (d) fourth bound
        	// the positive real root of ...
        	p2 := poly(x^n - _plus(L[i]/L[n]*x^i $ i=0..n-1));
        	DIGITS := 5;
        	u := float(min(b1,b2,b3)); l := 0.0;
        	repeat
        	    t := (u+l)/2;
        	    if p2(t) > 0 then
        	        u := t;
        	    else
        	        l := t;
        	    end_if;
        	until (u-l) < 0.5 end_repeat;
        	b4 := u;
        	userinfo(1,"polylib::rootbound: Fourth bound: ", b4);
        	        	
        	// return the smallest one of the three.
        	return(float(min(b1,b2,b3,b4)));
        	
        of hold(Exact) do
            userinfo(9,"calculating floating point approximation of all roots");
            DIGITS := 8;
            A := numeric::polyroots(p, FixedPrecision);
            A := map(A, abs);
            b1 := ceil(max(op(A)) * 1.1);
            return(b1);
            
        of hold(Iterate) do
            b3 := polylib::rootbound(poly(p(x)), hold(Traditional));
            b1 := polylib::rootbound(poly(p(x+1)), hold(Traditional))+1;
            b2 := polylib::rootbound(poly(p(x-1)), hold(Traditional))+1;
            if b1 >= b3 and b2 >= b3 then 
                return(b3);
            end_if;
            tmp := min(b1, b2);
            u := _if(b1 < b2, 1, -1);
            l := u;
            // look for an interval that may hold a 'small' rootbound
            repeat
                u := l;
                tmp2 := tmp;
                l := 2*l;
                tmp := polylib::rootbound(poly(p(x+l)), hold(Traditional))+l;
            until tmp >= tmp2 end_repeat;
            
            repeat
                if polylib::rootbound(poly(p(x+u)))+u 
                  > polylib::rootbound(poly(p(x+l)))+l then
                    u := round((u+l)/2);
                else
                    l := round((u+l)/2);
                end_if
            until abs(u-l)<=1 end_repeat;
            return (polylib::rootbound(poly(p(x+u)))+u);
    end_case;
end_proc;
