// $Date: 2004/09/23 07:53:10 $ $Author: bij $ $Revision: 1.96 $

/*++
prog::test -- test function

prog::test()
prog::test(a)
prog::test(a, b)
prog::test(a, b, <c>, <Status = status_name>, // general status
                      <Status(failed_result) = status_name ...> // one errors status 
                      <Message(failed_result) = string ...>)    // one errors message

a, b - expressions
c    - string

prog::test prints - depending upon the parameters - the current function
and test number onto the test results file 'test.res'. This signals
an error in the current test.

prog::test() prints the current function and test number unconditional.

prog::test(a) prints the current function and test number if a is not
equal to TRUE or if an error occures during the evaluation of a.

prog::test(a, b) prints the current function and test number if a is not
equal to b (in the sense of _equal) of if an error occures during the
evaluation of a or b.

prog::test(a,b,c) works like prog::test(a, b) but append string c to 
the output.
++*/


prog::test:=
  proc(stmt, result = TRUE)
    option hold;
    local
          deleteName,        // to delete unwanted procedure names
          r,                 // evaluated stmt
          res,               // evaluated result
          ERR,               // error while evaluating stmt
          ERRes,             // error while evaluating result
          printMessage, message, TE, PREF, addArgs,
          TIME,              // time for evaluation of stmt
          OPTS,              // table collecting options
          FILE,              // log file: keep last test call if kernel crashes
          StatusStrings,     // all valid status strings
          allOptions,        // valid option names
          compare;           // compare of stmt and res
  begin
    // all valid status strings
    StatusStrings := {"Analyzed", "Registered", "ToDo"}:
    // valid option names
    allOptions := [hold(Status), hold(Message), hold(TrapError), hold(Timeout), hold(Method)]:

    // create output message
    message:=
      proc(call, got)
        local le, ind;
      begin
        if ERRes + ERR <> 0 then
          le := getlasterror()[2];
          // remove 
          if (ind := strmatch(le, "Error: ", Index)) <> FALSE then
            le[_range(op(ind))] := "" // remove 'Error: '
          end_if;
          if (ind := strmatch(le, ";\?while reading \*tst'", Index)) <> FALSE then
            le[_range(op(ind))] := "" // remove substring
          end_if;
        end_if;
        call := call,
                " [",
                if ERRes <> 0 then // unexpected error while evaluation of result
                     "TrapError = ", ERRes, " result: ", le
                elif ERR <> 0 then // unexpected error while evaluation of stmt
                  "TrapError = ", ERR, ": ", le
                  (if ERR = 1320 and testtype(OPTS[hold(Timeout)], Type::PosInt) then
                     " (", OPTS[hold(Timeout)], " s)"
                   else
                     null()
                   end_if)
// 		  getlasterror()[2],
// 		  (if testtype(OPTS[hold(Timeout)], Type::PosInt) then
// 		     " (", OPTS[hold(Timeout)], " s)"
// 		   else
// 		     null()
// 		   end_if)
                else              // unexpected result
                  got
                end_if,
                "]",
                if contains(OPTS, hold(Message)(r)) then // concat message
                  " (", OPTS[hold(Message)(r)], ")"
                elif contains(OPTS, hold(Message)) then
                  " (", OPTS[hold(Message)], ")"
                elif contains(OPTS, hold(Message)(hold(TrapError) = ERR)) then
                  " (", OPTS[hold(Message)(hold(TrapError) = ERR)], ")"
                else
                  ""
                end_if;

        // remove all newlines in message
        map(call, x -> if domtype(x) = DOM_STRING then
                         stringlib::collapseWhitespace(x)
                       else x end)
      end_proc;
    
    // procedure to print out error messages
    printMessage:=
      proc(COND, call, got)
        name prog::test;
        local MESS, STAT, k;
        // uses r, ERR, and ERRes from outer procedure
      begin
        // process state
        if ERR <> 0 and contains(OPTS, hold(Status)(hold(TrapError) = ERR)) then
          STAT:= "*".OPTS[hold(Status)(hold(TrapError) = ERR)]."*"
        elif contains(OPTS, hold(Status)(r)) then // error specific status
          STAT:= "*".OPTS[hold(Status)(r)]."*"
        elif contains(OPTS, hold(Status)) then // general status
          STAT:= "*".OPTS[hold(Status)]."*"
        else
          STAT:= "*Old*";
          
          // consider global test status set by prog::testerrors
          if prog::TestErrorTable <> table() and prog::TestErrorTable <> FAIL then
            for k in prog::TestErrorTable do
              if contains(op(k, 2), prog::TestNo) > 0 then
                STAT:= "*".op(k, 1)."*";
                break
              end_if
            end_for
          end_if
        end_if;
          
        // keep all calls!
        prog::testcall(prog::TestNo, call);
          
        if COND then // error!
          // build up message
          MESS := message(call, got);

          // logging in file XXX.res
          if domtype(prog::TestFile) = DOM_INT then
            fprint(Unquoted, prog::TestFile, prog::TestFunc, " ",
                   prog::TestNo, ": ", MESS)
          end_if;

          // for logging
          r:= stringlib::collapseWhitespace(expr2text(r));

          // change status to '*New*' if error is new but not registered
          if prog::TestOldErrors <> FAIL then
            if not contains(prog::TestOldErrors, [prog::TestFunc, call, r])
              and STAT = "*Old*" then // only unregistered errors
              STAT:= "*New*";
            end_if;
          elif STAT = "*Old*" then
              STAT := null();
          end_if;

          fprint(Unquoted, 0, "Error", STAT, " in test '", prog::TestFunc, "' ",
                 prog::TestNo, ": ", MESS);

          if length(STAT) > 2 then
            // use the new table, default is used to mark running tests
            prog::testerrors(STAT[2..-2], prog::TestNo, New); // remove * *
          else
            prog::testerrors("New", prog::TestNo, New); // new error
          end_if;

          prog::lasterror("Error".STAT." in test '".prog::TestFunc. "' ".
                          expr2text(prog::TestNo).": ".MESS);
          
          if prog::TestErrors <> FAIL then // log errors in ".err"
            sysassign(prog::TestErrors[[prog::TestFunc, call, r]], TRUE)
          end_if;
          sysassign(prog::TestErrorCount, prog::TestErrorCount + 1) // count errors
          
        else // no error
          if prog::TestOldErrors <> FAIL and // log results
            contains(prog::TestOldErrors, [prog::TestFunc, call]) then
              fprint(Unquoted, 0, "Info: Test '", prog::TestNo, "' corrected [prog::test]")
          end_if;
          if STAT <> "*Old*" then // status defined
            fprint(Unquoted, 0, "Error".STAT,
                   ": test '", prog::TestFunc, " ",
                   prog::TestNo, (if contains(OPTS, hold(Message)) then
                                    "' (" , OPTS[hold(Message)] , ")"
                                  else
                                    "'"
                                  end_if),
                   " error fixed; remove status flag [prog::test]");
            prog::testerrors("Fixed", prog::TestNo);
          end_if;
          prog::lasterror("");
        end_if
      end_proc;

    // no result file opened
    if domtype(prog::TestFile) <> DOM_INT then
      userinfo(10, "result file does not exist (no logging)")
    end_if;

    // no name for test
    if prog::TestFunc = FAIL then
      userinfo(10, "test function not defined (interactive mode)");
      sysassign(prog::TestFunc, "interactive");
      sysassign(prog::TestNo, 1)
    end_if;
    
    // global test counter
    userinfo(Text, 10, "count the test '", stmt = result, "'");
    sysassign(prog::TestCount, prog::TestCount + 1);

    // keep 'Pref::typeCheck' and 'testargs' state
    if prog::TestNoTestargs <> TRUE then
      userinfo(10, "switch ON testargs and typecheck");
      PREF:= Pref::typeCheck(Always), testargs(TRUE)
    else
      userinfo(10, "reset testargs and typecheck to default (OFF)");
      PREF:= Pref::typeCheck(NIL), testargs(FALSE)
    end_if;

    //////////////////////////////////////////////////////////////////////
    // collect and check additional arguments
    // evaluate additional arguments
    if args(0) > 2 then
      addArgs := [context(args(3..args(0)))]
    else
      addArgs := []
    end_if;

    // special calls - obsolete?
    r:= op(split(addArgs, X -> domtype(X) = DOM_STRING), 1);

    OPTS:= prog::getOptions(allOptions, 2);
    if nops(OPTS) + nops(r) < args(0) - 2 then // too few arguments
      //fprint(Unquoted, 0, addArgs, r);
      error("wrong type of additional arguments")
    end_if;
    // append single comment (without 'Message = comment')
    if r <> [] then
      OPTS[hold(Message)]:= op(r)
    end_if;
    // obsolete, but always provided
    if contains(OPTS, hold(Quiet)) then
      OPTS[hold(Status)]:= "Registered";
      warning("old option 'Quiet' used in 'prog::test' - please replace it by 'Status = Registered'")
    end_if;
    // process status flags
    OPTS:= table(op(map([op(OPTS)], proc(X)
                                      name prog::test::checkStatus;
                                      local reg;
                                    begin
                                      if has(X, hold(Status)) then
                                        reg:= prog::getname(op(X, 2));
                                        if not contains(StatusStrings, reg) then
                                          error("wrong status flag used")
                                        else
                                          op(X, 1) = reg
                                        end_if;
                                      else
                                        X
                                      end_if
                        end_proc)));

    if type(result) = "_equal" and eval(op(result, 1)) = hold(TrapError) then
      TE:= op(result, 2);
    else
      TE:= 0
    end_if;

    // log current example
    if prog::TestLogExample <> FAIL then
      if (FILE:= fopen(Text, prog::TestLogExample, Write)) <> FAIL then
        fprint(NoNL, FILE, prog::TestFunc, " ", prog::TestNo,
                           (if contains(OPTS, hold(Message)) then
                              " (" , OPTS[hold(Message)] , ")"
                            else
                              " (no message)"
                            end_if));
        fclose(FILE)
      else
        warning("cannot open '".prog::getname(prog::TestLogExample)."' for writing")
      end_if
    end_if;
    
    // init time and results
    TIME:= 0; r:= FAIL; res:= FAIL;
    userinfo(Text, 1, "testing '", prog::TestFunc, "' ", prog::TestNo);
    if args(0) = 0 then
      // support obsolete behaviour
      print(prog::TestNo);
      sysassign(prog::TestNo, prog::TestNo - 1)
    else // standard
      userinfo(Text, 2,"  for ",stmt," = ",_if(TE > 0,  "Error No. ".expr2text(TE), result));

      // evaluate stmt and keep result, time and error code if it fails
      if testtype(OPTS[hold(Timeout)], Type::PosInt) then
        ERR:= traperror((TIME:= time((r:= context(stmt)))), OPTS[hold(Timeout)])
      else
        ERR:= traperror((TIME:= time((r:= context(stmt)))))
      end_if;
      
      // the same for the result (without time)
      ERRes:= traperror((res:= context(result)));

      // local procedure deleteName
      // this is to handle the following problem:
      // if stmt is a procedure without name, assigning context(stmt)
      // to some local var r will change the name, such that
      // prog::test(x -> x, x -> x)
      // might not work
      deleteName :=
      proc(flist, originallist)
        local f, original;
      begin
        f:= op(flist);
        original:= op(originallist);
        if domtype(f) = DOM_PROC and domtype(original) = DOM_PROC then
          subsop(f, 6 = op(original, 6))
        else
          f
        end_if
      end_proc;
      
      //
      if testtype(OPTS[hold(Method)], Type::Function)
         or prog::testmethod() <> FAIL then
        if not testtype(OPTS[hold(Method)], Type::Function) then
          OPTS[hold(Method)] := prog::testmethod()
        end_if;
        if traperror((compare := OPTS[hold(Method)](deleteName([r], [stmt]), // remove name from procedures
                                      deleteName([res], [result])))) <> 0 then
          printMessage(// compare fails
                       TRUE,
                     
                       // remove all newlines in input
                       stringlib::collapseWhitespace(prog::getname(OPTS[hold(Method)]).
                                                     "(".expr2text(stmt, result).")"),
                       "compare with ".prog::getname(OPTS[hold(Method)]));
        else
          printMessage(// whether test fails
                       ERR <> TE or ERRes <> 0 or
                       (ERR + ERRes = 0 and // both evals are without errors
                        
                        compare <> TRUE),
                       
                       // remove all newlines in input
                       stringlib::collapseWhitespace(prog::getname(OPTS[hold(Method)]).
                                                     "(".expr2text(stmt, result).")"),
                       "got: ".stringlib::collapseWhitespace(expr2text(compare)))
        end_if
      else
        printMessage(// whether test fails
                     ERR <> TE or ERRes <> 0 or
                     (ERR + ERRes = 0 and // both evals are without errors
                      
                      not _equal(deleteName([r], [stmt]), // remove name from procedures
                                 deleteName([res], [result]))),
                     
                     // remove all newlines in input
                     stringlib::collapseWhitespace(expr2text(_equal(stmt, result))),
                     "got: ".stringlib::collapseWhitespace(expr2text(r)))
      end_if;
    end_if;
    
    // log each test with its time for prog::testtime
//     sysassign(prog::TestTimeTable, append(prog::TestTimeTable,
//                                           prog::TestNo = stringlib::formatf(float(TIME/1000), 1)/*,
//                                           message(hold(_equal)(stmt, result))*/));
    
    sysassign(prog::TestTimeTable[prog::TestNo], stringlib::formatf(float(TIME/1000), 1));
    
    sysassign(prog::TestNo, prog::TestNo + 1);
    sysassign(prog::TestTime, prog::TestTime + TIME);

    if prog::TestNoTestargs <> TRUE then
      Pref::typeCheck(op(PREF, 1));
      testargs(op(PREF, 2))
    end_if;

    null()
  end_proc:

// end of file 
