// $Date: 2003/09/05 13:58:26 $ $Author: tonner $ $Revision: 1.70 $

/*
after prog::trace(f) where f is a procedure or a function environment, 
all the calls of f will print their input and output.

Examples:

>> prog::trace(has): has(x,y);
enter 'has'                            with args   : x, y
leave 'has'                            with result : FALSE

                                   FALSE

>> prog::untrace(has): has(x,y);

                                   FALSE

>> fib := proc(n) begin if n<2 then n else fib(n-1)+fib(n-2) end_if end_proc:
>> prog::trace(fib): fib(2);
enter 'fib'                            with args   : 2
enter 'fib'                            with args   : 1
leave 'fib'                            with result : 1
enter 'fib'                            with args   : 0
leave 'fib'                            with result : 0
leave 'fib'                            with result : 1

                                     1

>> g:= x->x + 1: prog::trace(g): g(3);
enter 'g'                              with args   : 3
leave 'g'                              with result : 4

                                     4

>> prog::trace(domtype(infinity)):
>> infinity+3;
enter 'stdlib::Infinity::_plus'        with args   : infinity, 3
enter 'stdlib::Infinity::thisDomain'   with args   : infinity
leave 'stdlib::Infinity::thisDomain'   with result : TRUE
enter 'stdlib::Infinity::thisDomain'   with args   : 3
leave 'stdlib::Infinity::thisDomain'   with result : FALSE
enter 'stdlib::Infinity::thisDomain'   with args   : infinity
leave 'stdlib::Infinity::thisDomain'   with result : TRUE
enter 'stdlib::Infinity::thisDomain'   with args   : 3
leave 'stdlib::Infinity::thisDomain'   with result : FALSE
enter 'stdlib::Infinity::removeRe'     with args   : 3
leave 'stdlib::Infinity::removeRe'     with result : 0
leave 'stdlib::Infinity::_plus'        with result : infinity

                                 infinity
*/

/* This table will contain the original code of all traced functions and is
   used for prog::untrace */
prog::traceTab := table():

// Option Tree
prog::trace_depth:= -1:
// Option Mem
prog::BYTES:= table():

alias(allOptions = {hold(Backup), hold(Force), hold(None), hold(Mem),
                    hold(Plain), hold(NoArgs), hold(Depth), //hold(Eval),
                    hold(Proc), hold(Name), hold(All), hold(Parent),
                    hold(Width), hold(All)}):
// param options    hold(Depth) = d, hold(Width) = w, hold(Name) = n

alias(untracedDomainMethods = {"print", "expr2text"}):

prog::trace:=
proc(_f_ = (()->()), METHODS = null())
  option hold, escape;
  local ff, p, Print, trace_internal, trace_proc, trace_func_env,
        bytes_used_in, bytes_used_out, OX, OPT, SFORMAT, FROM, k;
begin
  if args(0) = 0 then
    warning("no object to trace");
    // reset display depth
    sysassign(prog::trace_depth, -1);
    return()
  end_if;

  SFORMAT:=
    proc(STR)
    begin
      substring(STR."                                                                ",
                1..(if contains(OPT, hold(Parent)) then 64 else 32 end_if))
    end_proc;

  FROM:=
    proc(n)
    begin
      substring(substring("'".n."'", 1..min(length(n)+2, 32)).
                " from '".
                (if strmatch((n := prog::getname(context(hold(context(hold(context(procname))))))),
                             "context(hold(context\*") = FALSE then
                   n
                 else
                   "interactive"
                 end_if).
                "'                                                                ",
                1..64)
    end_proc;

  bytes_used_in:=
    proc(n)
      name prog::trace::bytes_used_in;
    begin
      if not(contains(prog::BYTES, n)) then sysassign(prog::BYTES[n], []) end_if;
      assert(testtype(prog::BYTES[n], Type::ListOf(Type::PosInt)));
      sysassign(prog::BYTES[n], append(prog::BYTES[n], op(bytes(), 1)));
      null()
    end_proc;

  bytes_used_out:=
    proc(n)
      name prog::trace::bytes_used_out;
      local used;
    begin
      used:= prog::BYTES[n][-1];
      assert(testtype(used, Type::PosInt));
      sysassign(prog::BYTES[n], subsop(prog::BYTES[n], nops(prog::BYTES[n]) = null()));
      if prog::BYTES[n] = [] then sysdelete(prog::BYTES[n]) end_if;
      used:= trunc((op(bytes(), 1) - used)/1000.0);
      expr2text(used)
    end_proc;

  // ---- internal routines ---- //
  Print := 
  /* print all procedures in the argument list in short Form */
  proc(aaa)
    name prog::trace::Print;
    option noDebug;
  begin
    map(aaa, proc(x)
               local str;
             begin
               if domtype(x) = DOM_PROC then
                 str:= DOM_PROC::print(x)
               else
                 str:= expr2text(x)
               end_if;
               if length(str) > 1024 then
                 substring(str, 1, 1024)." ..."
               else
                 str
               end_if
               end_proc);
               zip(%, [", "$nops(%)-1], _exprseq, "");
               _concat(op(%))
             end_proc;

  trace_internal :=    /* trace for kernel functions */
  proc(p,n)
    option escape, noDebug;
  begin
    subsop(proc()
             local  RES, ARGS;
             option noDebug;
           begin
             sysassign(prog::trace_depth, prog::trace_depth + 1); // Depth ++
             ARGS := args();
             if prog::trace_depth < "DEPTH" then
               fprint(Unquoted, 0, "  " $ prog::trace_depth,
                      "enter ", "NAME", " with args   : ", Print([ARGS]))
             end_if;
             RES:= p(ARGS);
             if prog::trace_depth < "DEPTH" then
               fprint(Unquoted, 0, "  " $ prog::trace_depth,
                      "leave ", "NAME", " with result : ", Print([RES]))
             end_if;
             sysassign(prog::trace_depth, prog::trace_depth - 1); // Depth --
             RES
           end_proc, 6 = _f_, // the name of the kernel function //text2expr(OPT[hold(Name)]),
           // indentation
           _if(contains(OPT, hold(Plain)),
               ([4,1] = NIL, [4,6] = NIL, [4,3,3] = "", [4,5,3] = ""), null()),
           _if(OPT[hold(Depth)] = infinity,
               ([4,3,1] = TRUE, [4,5,1] = TRUE),
               ([4,3,1,2] = OPT[hold(Depth)], [4,5,1,2] = OPT[hold(Depth)])),
           _if(contains(OPT, hold(Parent)),
               ([4,3,2,5] = hold(id)(n), [4,3,2,5,0] = FROM),
               [4,3,2,5] = SFORMAT("'".n."'")),
           [4,5,2,5] = SFORMAT("'".n."'"),
           _if(contains(OPT, hold(NoArgs)),
               ([4,3,6] = "", [4,3,7] = "", [4,5,6] = "", [4,5,7] = ""), null())
           )
  end_proc;
	       
  trace_proc :=      /* trace for procedures */
  proc(p, n)
    option escape;
    name prog::trace::trace_proc;
    local trace_exit;
  begin
//     /* check if function is already traced */
//     if contains(prog::traceTab, p) then
//       return(p)
//     end_if;
    
    trace_exit :=    /* print result on exit ; also used for return */
      subsop(proc(n)
               option escape;
               name prog::trace::trace_exit;
             begin
               proc()
               begin
                 if prog::trace_depth < NIL then
                   fprint(Unquoted, 0,
                          "  " $ prog::trace_depth,
                          "leave ".SFORMAT("'".n."'"),
                          " with result ", "[".bytes_used_out(n)." kB]", ": ".Print([args()]))
                 end_if;
                 sysassign(prog::trace_depth, prog::trace_depth - 1);
                 args()
               end_proc:
             end_proc,
             _if(OPT[hold(Depth)] = infinity,
                 [4,4,1,1] = TRUE,
                 [4,4,1,1,2] = OPT[hold(Depth)]),
             _if(contains(OPT, hold(Plain)),
                 ([4,4,1,2,3] = "", [4,4,2] = NIL), null()),
             _if(contains(OPT, hold(Mem)), null(), [4,4,1,2,6] = ""),
             _if(contains(OPT, hold(NoArgs)), ([4,4,1,2,5] = "", [4,4,1,2,7] = ""), null())
             );

    /* create new Proc with trace-token, enter-info, old code, exit-info */
    subsop(p, 4 =
	   hold(_stmtseq)(
                          hold(sysassign(prog::trace_depth, prog::trace_depth + 1)),
                          subsop(hold(id)(n), 0 = bytes_used_in, Unsimplified),
                          hold(_if)(hold(_less)(hold(prog::trace_depth), OPT[hold(Depth)]),
                                    hold(fprint)(Unquoted, 0,
                                                 hold("  " $ prog::trace_depth),
                                                 "enter ",
                                                 SFORMAT("'".n."'"),
                                                 " with args   : ",
                                                 hold(NIL([args()]))),
                                    NIL),
                          "original BODY",
                          trace_exit(n),
                          hold(%(%2))),
           _if(contains(OPT, hold(Plain)), ([4,1] = NIL, [4,3,2,3] = ""), null()),
           _if(contains(OPT, hold(Mem)), null(), [4,2] = NIL),
           [4,3,2,7,0] = Print,
           _if(OPT[hold(Depth)] = infinity, [4,3,1] = TRUE, null()),
           // BODY replacement
           [4, 4] = subsop(op(p, 4), k = return@trace_exit(n)
                                       $ k in map({prog::find(p, hold(return), hold(NoProc))},
                                                  X -> subsop(X, 1 = null()))),
           _if(contains(OPT, hold(NoArgs)), ([4,3,2,6] = "", [4,3,2,7] = ""), null()),
           _if(contains(OPT, hold(Parent)),
               ([4, 3, 2, 5] = hold(id)(n), [4, 3, 2, 5, 0] = FROM),
               null()),
           //6 = text2expr(n), // Name
	   Unsimplified)
  end_proc;

  // trace function environment 'ENV'
  trace_func_env:=
    proc(ENV, NAME)
      local Pr1, funcTab;
      name prog::trace::trace_func_env;
    begin
      userinfo(1, "trace function environment '".prog::getname(ENV)."'");
      Pr1:= op(ENV, 1);
      case domtype(Pr1)
	of DOM_PROC do
	  Pr1:= trace_proc(op(ENV, 1), NAME);
	  break;
	  
	of DOM_EXEC do
	  Pr1:= trace_internal(op(ENV, 1), NAME);
	  break;
      end_case;
      //ENV:=context(ENVID);  /* get changes from last step in ff */
      if domtype(op(ENV,3)) = DOM_TABLE then
	/* trace all slot's of the DOM_FUNC_ENV */
	funcTab := op(ENV,3);
	map(funcTab, context);
	//ENV:=context(ENVID);  /* get changes from last step in ff */
	//funcTab := op(ENV,3);
        // trace procedures
        userinfo(2, "trace all slots of '".prog::getname(ENV)."' that are procedures");
        map([op(select(funcTab, x->domtype(op(x,2))=DOM_PROC))], 
            proc(x)
              local Pr;
              name prog::trace::trace_func_env;
            begin
              Pr:= op(x,2);
              userinfo(5, "trace slot '".prog::getname(ENV)."::".op(x, 1)."'");
              funcTab[op(x,1)] := trace_proc(Pr, prog::getname(Pr))
            end_proc);

        // trace function environments, that does not have
        // procedures or func-envs as slot
        userinfo(2, "trace all slots of '".prog::getname(ff)."' that are function environments");
        map([op(select(funcTab, x->bool(domtype(op(x,2))=DOM_FUNC_ENV
                                        and select(op(x, [2, 3]),
                                                   x->bool(domtype(op(x,2))=DOM_PROC
                                                           or domtype(op(x,2))=DOM_FUNC_ENV)) = table())))],
            proc(x)
              local Pr;
              name prog::trace::trace_func_env;
            begin
              Pr:= genident("TRACE");
              evalassign(Pr, op(x, 2), 2); // DOM_FUNC_ENV
              userinfo(5, "trace slot '".prog::getname(ENV)."::".op(x, 1)."'");
              funcTab[op(x, 1)]:= trace_func_env(op(x, 2), NAME."::".op(x, 1))
            end_proc);

        
	subsop(ENV, 1 = Pr1, 3 = funcTab)
      else // internal function
        trace_internal(ENV, NAME)
      end_if
    end_proc;

// --------------------------- 

  // functions that cannot be traced
  if contains({hold(fprint), hold(hold), hold(slot), hold(DOM_SET::sort),
               hold(return)}, _f_) then
    error("this function cannot be traced")
  elif contains({hold(_concat), hold(_procdef), 
                 hold(length), hold(Pref), hold(strmatch),
                 hold(anames), hold(type), hold(substring), hold(domtype),
                 hold(contains), hold(sysassign)}, _f_) then
    warning("this function should not be traced, unexpected sideeffects breaks out")
  elif Pref::typeCheck() <> hold(None) and contains({hold(testtype)}, _f_) then
    error("this function cannot be traced, when 'Pref::typeCheck' is not 'None'")
  end_if;

  ff:=context(_f_);
  prog::init(ff);

  OPT := prog::getOptions(allOptions, (if domtype(ff) = DOM_DOMAIN
                                           and (domtype(METHODS) = DOM_SET
                                                or domtype(METHODS) = DOM_LIST) then
                                          2
                                        else
                                          1
                                       end_if), Warning);
  if not contains(OPT, hold(Depth)) then
    OPT[hold(Depth)] := infinity
  end_if;

  /* check if function is already traced */
  if contains(prog::traceTab, _f_) then
    if contains(OPT, hold(Force)) then
      warning("backup of object '".expr2text(_f_)."' is replaced");
      //
    elif contains(OPT, hold(Backup)) then
      warning("backup of object '".expr2text(_f_)."' is traced");
      //context(hold(prog::untrace)(_f_, Quiet))
      //sysevalassign(_f_, prog::traceTab[_f_], 2)
      ff := prog::traceTab[_f_]
    else
      warning("object '".expr2text(_f_)."' is already traced");
      return()
    end_if
  end_if;

  // explicit name set
  if not contains(OPT, hold(Name)) then
    OPT[hold(Name)] := expr2text(_f_)
  elif domtype(OPT[hold(Name)]) <> DOM_STRING then
    OPT[hold(Name)] := expr2text(OPT[hold(Name)])
  end_if;
  
  if contains(OPT, hold(Width)) and domtype(OPT[hold(Width)]) = DOM_INT then
    // manipulate Print
    Print := subsop(Print,
                    [4, 1, 2, 4, 2, 1, 1] = OPT[hold(Width)],
                    [4, 1, 2, 4, 2, 2, 1, 3] = OPT[hold(Width)])
  end_if;

  // force evaluation
  //if contains(OPT, hold(Eval)) then
  //  _f_:= context(hold(level)(_f_, 1))
  //end_if;
  
  case domtype(ff)
    /*  ff is a internal function or procedure with slot's */
    of DOM_FUNC_ENV do
      userinfo(1, "trace function environment '".prog::getname(ff)."'");
      if contains(OPT, hold(Proc)) then
        return(trace_func_env(ff, OPT[hold(Name)]))
      else
        sysassign(prog::traceTab[_f_], ff);
        stdlib::syseval(evalassign(_f_, trace_func_env(ff, OPT[hold(Name)]), 1));
      end_if;
      break
       
    of DOM_PROC do
      userinfo(1, "trace procedure '".prog::getname(ff)."'");
      if contains(OPT, hold(Proc)) then
        return(trace_proc(ff, OPT[hold(Name)]))
      else
        if type(_f_) = "slot"
           and domtype(context(op(_f_, 1))) = DOM_DOMAIN then
          sysassign(prog::traceTab[_f_], ff);
          stdlib::syseval((evalassign(_f_,trace_proc(ff, OPT[hold(Name)]), 1)))
        else
          sysassign(prog::traceTab[_f_], ff);
          stdlib::syseval((evalassign(_f_,trace_proc(ff, OPT[hold(Name)]), 1)))
        end_if
      end_if;
      break

    of DOM_DOMAIN do
      userinfo(1, "trace domain '".prog::getname(ff)."'");
      if contains(OPT, hold(Proc)) then
        error("option 'Proc' cannot be used with domains")
      end_if;
      if domtype(METHODS) = DOM_SET or domtype(METHODS) = DOM_LIST then
        for p in METHODS do
          if domtype(p) <> DOM_STRING then
            p:= expr2text(p)
          end_if;
          if not contains(untracedDomainMethods, p)
            and domtype(slot(ff,p)) = DOM_PROC then
            if contains(prog::traceTab, hold(slot)(_f_, p))
              and not contains(OPT, hold(Force)) then
              warning("method '".expr2text(hold(slot)(ff, p))."' is already traced")
            else
              userinfo(Text, 5, "trace domain method '", expr2text(hold(slot)(_f_, p)),
                         "' with name '", prog::getname(slot(ff, p)), "'");
              sysassign(prog::traceTab[hold(slot)(_f_,p)], slot(ff, p));
              sysassign(slot(ff,p),
                        trace_proc(slot(ff,p), prog::getname(slot(ff, p))
                                   /*expr2text(hold(slot)(ff,p))*/ ))
            end_if
          else
            warning("no object to trace [".expr2text(ff)."::".p."]")
          end_if
        end_for
      else // trace all Methods
        sysassign(prog::traceTab[_f_], "traced domain");

        // not traced methods/only traced methods
        if contains(ff, "trace") and not contains(OPT, hold(All)) then
          userinfo(Text, 5, "use 'trace' slot of the domain");
          OX := ff::trace
        else
          userinfo(Text, 5, "trace all methods of the domain");
          OX := map({op(ff)}, op, 1); // only method names as strings
        end_if;
        if contains(ff, "noTrace") then
          userinfo(Text, 5, "use 'noTrace' slot of the domain");
          OX:= OX minus map(slot(ff, "noTrace"), X -> if domtype(X) = DOM_STRING then
                                                        X
                                                      else
                                                        expr2text(X)
                                                      end_if)
        end_if;
        OX := OX minus untracedDomainMethods; // except untrace-able methods
        
        for p in op(ff) do
          if contains(OX, op(p,1)) then // no trace of excluded methods
            if domtype(op(p,2)) = DOM_PROC then
              // traced?
              if contains(prog::traceTab, hold(slot)(_f_,op(p,1)))
                 and not contains(OPT, hold(Force)) then
                warning("method '".expr2text(hold(slot)(ff,op(p,1)))."' is already traced")
              else
                userinfo(Text, 5, "trace domain method '", expr2text(hold(slot)(_f_, op(p, 1))),
                                  "' with name '", prog::getname(slot(ff, op(p, 1))), "'");
                sysassign(prog::traceTab[hold(slot)(_f_,op(p,1))], op(p,2));
                sysassign(slot(ff,op(p,1)),
                          trace_proc(op(p,2), prog::getname(slot(ff, op(p, 1)))
                                     /*expr2text(hold(slot)(ff,op(p,1)))*/))
              end_if
            elif domtype(op(p,2)) = DOM_FUNC_ENV then
              // traced?
              if contains(prog::traceTab, hold(slot)(_f_,op(p,1)))
                 and not contains(OPT, hold(Force)) then
                warning("method '".expr2text(hold(slot)(ff,op(p,1)))."' is already traced")
              else
                userinfo(Text, 5, "trace domain method '", expr2text(hold(slot)(_f_, op(p, 1))),
                                  "' with name '", prog::getname(slot(ff, op(p, 1))), "'");
                sysassign(prog::traceTab[hold(slot)(_f_,op(p,1))], op(p,2));
                sysassign(slot(ff,op(p,1)),
                          trace_func_env(op(p,2), OPT[hold(Name)]."::".op(p, 1)))
              end_if
            end_if
	  end_if
        end_for
      end_if;
      break
    of DOM_SET do
      map(_f_, prog::trace, op(OPT)); break
    otherwise
      warning("no object to trace (".OPT[hold(Name)].")")
  end_case;

  null()
end_proc:

prog::traceall:=
  proc()
    local GF, DF;
  begin
    // DOM_PROC /////////////////////////////////////////////////////////////
    userinfo(Text, 1, "pre-init all DOM_PROC");
    map(sort([op(select(stdlib::anames(3), (X)->(type(context(X)) = DOM_PROC)))],
	     (X, Y)->(expr2text(X)<expr2text(Y))),
	prog::init);
    // DOM_FUNC_ENV /////////////////////////////////////////////////////////
    userinfo(Text, 1, "pre-init all DOM_FUNC_ENV");
    map(sort([op(select(stdlib::anames(3), (X)->(type(context(X)) = DOM_FUNC_ENV)))],
	     (X, Y)->(expr2text(X)<expr2text(Y))),
	prog::init);
    // DOM_DOMAIN ///////////////////////////////////////////////////////////
    userinfo(Text, 1, "pre-init most DOM_DOMAIN");
    GF:= sort([op(select(stdlib::anames(3), (X)->(type(context(X)) = DOM_DOMAIN)))],
	      (X, Y)->(expr2text(X)<expr2text(Y)));
    for DF in GF do
      DF:= context(DF);
      if domtype(DF) = DOM_DOMAIN then
	//if contains(context(DF), "constructor") then
	  prog::init(DF);
	  sysdelete(prog::init(DF))
	//end_if
      end_if
    end_for;
    null()
  end_proc:

// 
prog::untrace:=
  proc(_f_, OPT = hold(Warning))
    option hold;
    local p, tabIndex;
  begin
//     if OPT <> hold(Force) then
//       case prog::traced(_f_, hold(Quiet))
//         of UNKNOWN do
//           error("object '".prog::getname(_f_).
//                 "' seems to be traced, but no backup exists. Cannot restore original function.");
//           break
//         of FALSE do
//           error("object '".prog::getname(_f_).
//                 "' seems not to be traced, but a backup exist. Use option 'Force' to redefine with the backup.");
//           break
//       end_case
//     end_if;
    
    if args(0) = 0 then
      //* untrace all traced functions */
      for p in prog::traceTab do
        if prog::traceTab[op(p, 1)] = "traced domain" then
          // 
        else
          sysevalassign(op(p, 1), op(p, 2), 2);
        end_if
      end_for;
      sysassign(prog::traceTab, table());
      return();
    end_if;
  
    if type(context(_f_)) = DOM_DOMAIN then
      //* untrace all procedures in this domain */
      if contains(prog::traceTab, _f_) then
        sysdelete(prog::traceTab[_f_])
      end_if;
      for p in prog::traceTab do
        if type(op(p,1)) = "slot" and op(p,[1,1]) = _f_ then
	  sysevalassign(op(p,1),op(p,2), 2);
	  sysdelete(prog::traceTab[op(p,1)]);
        end_if;
      end_for;
    elif type(_f_) = "slot" then
      if type(_f_) = "slot"
         and domtype(context(op(_f_, 1))) = DOM_DOMAIN then
        sysdelete(prog::traceTab[op(_f_, 1)])
      end_if;
      tabIndex := hold(slot)(op(_f_,1),op(_f_,2));
      if not contains(prog::traceTab, tabIndex) then
        error("function was not traced")
      end_if;
   
      stdlib::syseval(evalassign(tabIndex, prog::traceTab[tabIndex], 2));
      sysdelete(prog::traceTab[tabIndex]);
    else
      p:= eval(_f_);
      if not contains(prog::traceTab, _f_) then
        if context(hold(prog::traced)(_f_, hold(Quiet))) = UNKNOWN then
          error("the function seems to be traced, however, no backup exists")
        else
          error("function was not traced")
        end_if
      end_if;
      if context(hold(prog::traced)(_f_, hold(Quiet))) = UNKNOWN and OPT <> hold(Force) then
        warning("the function seems not to be traced! Use option 'Force' to restore");
        return()
      end_if;
      stdlib::syseval(evalassign(_f_, prog::traceTab[_f_], 2));
      sysdelete(prog::traceTab[_f_]);
    end_if;
		   
    null()
  end_proc:

// X     - object
// OPT   - Quiet
//         return: TRUE    
//                 FALSE   
prog::traced:=
  proc(X, OPT = FAIL)
    option hold;
    local k, _X, RET, PRINC, CHECK;
  begin
    PRINC := proc(MSG, L)
             begin
               if OPT = hold(Quiet) then
                 RET:= L + 1
               else
                 fprint(Unquoted, 0, MSG)
               end_if        
             end_proc:

    CHECK := proc(obj)
             begin
               bool(_lazy_and(nops(prog::find(obj, "enter ")) > 0,
                              nops(prog::find(obj, "leave ")) > 0))
             end_proc;
    
    RET:= 1;
    OPT := context(OPT);
    if args(0) = 0 then
      if nops(prog::traceTab) > 0 then
        output::tableForm(sort(map([op(prog::traceTab)], op, 1)), ", ")
      else
        PRINC("No objects traced.", 0)
      end_if
    else
      if contains(prog::traceTab, X) then
        PRINC("A backup of object '" . prog::getname(X) . "' exists.", RET)
      end_if;
      _X:= context(X);
      case domtype(_X)
        of DOM_FUNC_ENV do
          if //type(op(_X, [1, 4])) = "_stmtseq"
             //and op(_X, [1, 4, 1, 0]) = "traced" then
             //and op(_X, [1, 4, 1]) = "traced" then
            CHECK(op(_X, 1)) then
            PRINC("Object '" . expr2text(X) . "' seems to be traced.", RET)
          end_if;
          break
        of DOM_PROC do
          if //type(op(_X, 4)) = "_stmtseq"
             //and op(_X, [4, 1, 0]) = "traced" then
             //and op(_X, [4, 1]) = "traced" then
            CHECK(_X) then
            PRINC("Object '" . expr2text(X) . "' seems to be traced.", RET)
          end_if;
          break
        of DOM_DOMAIN do
          if contains(prog::traceTab, X) or contains(prog::traceTab, _X) then
            fprint(Unquoted, 0, "Domain '", expr2text(X), "' seems to be traced.")
          else
            for k in map(op(_X), op, 1) do
              if contains(prog::traceTab, hold(slot)(X, k)) then
                PRINC("A backup of object '" . expr2text(hold(slot)(X, k)) . "' exists.", RET)
              end_if;
              if domtype(slot(_X, k)) = DOM_PROC then
                if //type(op(slot(_X, k), 4)) = "_stmtseq"
                   //and op(slot(_X, k), [4, 1]) = "traced" then
                   CHECK(slot(_X, k)) then
                  PRINC("Object '" . expr2text(hold(slot)(X, k)) . "' seems to be traced.", RET)
                end_if
              elif domtype(slot(_X, k)) = DOM_FUNC_ENV then
                if //type(op(slot(_X, k), [1, 4])) = "_stmtseq"
                   //and op(slot(_X, k), [1, 4, 1]) = "traced" then
                   CHECK(op(slot(_X, k), 1)) then
                  PRINC("Object '" . expr2text(hold(slot)(X, k)) . "' seems to be traced.", RET)
                end_if
              //elif domtype(slot(_X, k)) = DOM_DOMAIN then
              //  prog::traced(slot(_X, k), OPT)
              end_if
            end_for
          end_if;
          break
        otherwise
          PRINC("Nothing known about object '" . expr2text(X) . "'.", 0)
      end_case
    end_if;
    if OPT = hold(Quiet) then
      return(op([FALSE, UNKNOWN, TRUE], RET))
    else
      return()
    end_if
  end_proc:

//prog::traced := prog::remember(prog::traced, hold(PreventRecursion)):