/*
 * series slot of lambertW - computes series(lambertW(f), x, n, dir)
 *
 * computes undirected expansions around x0 for -exp(-1) <= x0 < infinity
 * returns FAIL for +-infinity
 * returns a symbolic Taylor expansion for all other values of x0
 *
 * Author: Juergen Gerhard, 18.09.01
 *
 * Changed 26.4.02 Christopher Creutzig: Use Jrgen's code for the principal
 * branch only; implement asymptotic formula
 */

lambertW::series :=
proc(branch, f, x, n, dir)
  local s, t, k, mu, alpha, j, l, nextbranch;
begin
  // recursively expand the argument
  t := Series::series(f, x, n, dir);

  if domtype(t) = Series::Puiseux then
    k := Series::Puiseux::ldegree(t);
    if k = FAIL then // t = O(..)
      k = Series::Puiseux::order(t);
      if k > 0 then
        return(t)
      end_if;
      Series::error("order too small")
    end_if;
    if k > 0 then // expansion around 0
      if branch = 0 then
        s := Series::Puiseux::create(1, 1, n+1,
                                     Series::gen["lambertW/0/0"](n), x, 0, dir);
        if f = x then
          return(s)
        else
          return(s @ t)
        end_if
      else
        s := Series::gseries::create(Series::gen["lambertW/k/asympt"](branch, n, x),
                                     (1/(ln(x)+branch*I*2*PI))^(n+1), x=0);
        if f = x then
          return(s);
        else
          return(s @ t);
        end_if
      end_if
    elif k < 0 then // expansion around infinity
      if dir <> Undirected then
        l := limit(lmonomial(t), x, dir);
        if l = infinity then
          if dir=Left then
            f := -f;
            dir = Right;
          end_if;
          if dir=Right then
            // Maple 6 simply returns lambertW(x)
            // Asymptotic expansion of de Bruijn/Comtet,
            // formulation of Corless, Jeffrey, Knuth 1997
            s := Series::gseries::create(Series::gen["lambertW/k/asympt"](branch, n, 1/x),
                                         (1/(ln(x)+branch*I*2*PI))^(n+1), x=infinity);
            f := 1/f;
            if f=x then
              return(s);
            else
              return(s @ Series::series(f, x, n, dir));
            end_if
          end_if
        end_if
      end_if;
      // we can't do it
      return(FAIL)
    else // k = 0, expansion around a finite point x0 <> 0
      t := lmonomial(t, Rem);
      if branch=0 then
        if combine(t[1], exp) = -exp(-1) then
          // This is from the ISSAC'97 article of Corless, Jeffrey & Knuth,
          // section 3
          mu[0] := -1;
          mu[1] := 1;
          alpha[0] := 2;
          alpha[1] := -1;
          for k from 2 to n - 1 do
            alpha[k] := _plus(mu[j]*mu[k + 1 - j] $ j = 2..(k - 1));
            mu[k] := (k - 1)/(k + 1)*(mu[k - 2]/2 + alpha[k - 2]/4)
                     - alpha[k]/2 - mu[k - 1]/(k + 1);
          end_for;
          s := Series::Puiseux::create(2, 0, n,
                 [mu[k]*2^(k/2)*exp(1)^(k/2) $ k = 0..(n-1)], x, 0, dir);
          return(s @ t[2]);
        end_if;
      end_if;
      if branch=-1 then
        if combine(t[1], exp) = -exp(-1) then
          // This is from the ISSAC'97 article of Corless, Jeffrey & Knuth,
          // section 3.
          mu[0] := -1;
          mu[1] := 1;
          alpha[0] := 2;
          alpha[1] := -1;
          for k from 2 to n - 1 do
            alpha[k] := _plus(mu[j]*mu[k + 1 - j] $ j = 2..(k - 1));
            mu[k] := (k - 1)/(k + 1)*(mu[k - 2]/2 + alpha[k - 2]/4)
                     - alpha[k]/2 - mu[k - 1]/(k + 1);
          end_for;
          s := Series::Puiseux::create(2, 0, n,
                                       [(-1)^k*mu[k]*2^(k/2)*exp(1)^(k/2)
                                        $ k = 0..(n-1)], x, 0, dir);
          return(s @ t[2]);
        end_if
      end_if
    end_if;
    if (branch = 0 and is(t[1] < -exp(-1))=TRUE)
      or (branch <> 0 and is(t[1] < 0)=TRUE) then
      nextbranch := branch-1;
      if branch = 1 and is(t[1] > -exp(-1))=TRUE then
        nextbranch := -1;
      end_if;
      return((signIm(f)+1)/2 * Series::unknown(lambertW(branch, f), x, n, dir)
             - (signIm(f)-1)/2 * Series::unknown(lambertW(nextbranch, f), x, n, dir));
    end_if
  end_if;

  Series::unknown(lambertW(branch, f), x, n, dir)
end_proc:

// ensure that domain Series is loaded
eval(Series):

// expansion of lambertW(0, x) around x = 0:
Series::gen["lambertW/0/0"] := proc(n)
begin
  [(-i)^(i - 1)/fact(i) $ i = 1..n]
end_proc:

// asymptotic expansion of lambertW(k, z) for z -> infinity
Series::gen["lambertW/k/asympt"] :=
proc(branch, order, z)
  local ln_k, ln_ln_k, n, m;
begin
  ln_k := ln(z) + I*branch*2*PI;
  ln_ln_k := ln(ln_k);
  [[ln_k-ln_ln_k, 1],
   [(-1)^(n+1)
    *_plus((combinat::stirling1(n, n-m+1)
            *ln_ln_k^m/fact(m))$m=1..n), ln_k^(-n)] $ n=1..order];
end_proc:
