// $Date: 2002/11/21 13:35:06 $ $Author: ccr $ $Revision: 1.27 $ 

// kg, 13/01/94 

/*++
arccos -- the inverse cosine

arccos(x)

x - expression

NB:  The branch cuts are the real intervals (-infinity, -1) and
     (1, infinity). 
     Note: for x >=1 :  arccos(x) =  limit(arccos(x+eps*I),eps=0, eps>0) 
                                  = -limit(arccos(x-eps*I),eps=0, eps>0) 
           for x <=-1:  arccos(x) =  limit(arccos(x-eps*I),eps=0, eps>0) 

     I.e., on the left real line the values coincide with the limits from
     below, on the right real line with the limit from above. OK.

     Consistent with arccos(x)+arccos(-x) = PI. OK.
++*/

arccos:=
proc(x)
  local f;
  name arccos;
  option noDebug;
begin
  if args(0) = 0 then error("no arguments given")
  elif x::dom::arccos <> FAIL then return(x::dom::arccos(args()))
  elif args(0) <> 1 then error("wrong no of args") 
  end_if;

  case domtype(x)
    of DOM_SET do
      return(map(x,arccos))

    of DOM_INT do
    of DOM_RAT do
	if x < 0 then return(PI - arccos(-x)) end_if;
	break;
    
    of DOM_FLOAT do
	return(arccos::float(x));

    of DOM_COMPLEX do
	if domtype(op(x,1)) = DOM_FLOAT or
	   domtype(op(x,2)) = DOM_FLOAT then
	    return(arccos::float(x))
	end_if;
	if op(x,1) = 0 then return(PI/2 - arcsin(x)) end_if;
	break;
  end_case;

  if not testtype(x,Type::Arithmetical) then
      error("argument must be of 'Type::Arithmetical'")
  end_if;

  case type(x)
    of "_mult" do
	f:= op(x, nops(x));
	if testtype(f, Type::Real) then
	    if f < 0 then return(PI - arccos(-x)) end_if;
	end_if;
	break;

    of "cos" do // arccos : [-1..1] --> [PI..0] 
       if (is(op(x)>=0) and is(op(x)<=PI))=TRUE then return(op(x))
       elif type((f:=floor(Re(op(x))/PI)))=DOM_INT then
          case f mod 2
          of 0 do return(op(x)-f*PI)
          of 1 do return((f+1)*PI-op(x))
          end_case;
       end_if; 
    end_case;

    procname(x)
end_proc:

arccos( 0):=  PI/2:  
arccos( 1):= 0:
arccos(-1):= PI:
arccos( 1/2):=   PI/3:
arccos(-1/2):= 2*PI/3:
arccos( (3^(1/2))/2):=   PI/6:
arccos(-(3^(1/2))/2):= 5*PI/6:
arccos( (2^(1/2))/2):=   PI/4:
arccos(-(2^(1/2))/2):= 3*PI/4:
arccos( 1/4*2^(1/2)*(5+5^(1/2))^(1/2)):=   PI/10:
arccos(-1/4*2^(1/2)*(5+5^(1/2))^(1/2)):= 9*PI/10:
arccos( 1/4*2^(1/2)*(5-5^(1/2))^(1/2)):= 3*PI/10:
arccos(-1/4*2^(1/2)*(5-5^(1/2))^(1/2)):= 7*PI/10:
arccos( ((2-2^(1/2))^(1/2))/2):= 3*PI/8:
arccos(-((2-2^(1/2))^(1/2))/2):= 5*PI/8:
arccos( ((2+2^(1/2))^(1/2))/2):=   PI/8:
arccos(-((2+2^(1/2))^(1/2))/2):= 7*PI/8:
arccos( 1/4*(5^(1/2)+1)):=   PI/5:
arccos(-1/4*(5^(1/2)+1)):= 4*PI/5:
arccos( 1/4*(5^(1/2)-1)):= 2*PI/5:
arccos(-1/4*(5^(1/2)-1)):= 3*PI/5:
arccos( (6^(1/2)-2^(1/2))/4):= 5*PI/12:
arccos(-(6^(1/2)-2^(1/2))/4):= 7*PI/12:
arccos( (6^(1/2)+2^(1/2))/4):=   PI/12:
arccos(-(6^(1/2)+2^(1/2))/4):=11*PI/12:

arccos( infinity):=   -I*infinity:
arccos(-infinity):= PI+I*infinity:
arccos( I*infinity):= PI/2 - I*infinity:
arccos(-I*infinity):= PI/2 + I*infinity:

arccos:= funcenv(arccos, op(specfunc::arccos, 2)):
arccos::info := "arccos -- the inverse cosine":
arccos::print := "arccos":
arccos::type := "arccos":
arccos::float := specfunc::arccos:
arccos::hull := DOM_INTERVAL::arccos@hull:

arccos::inverse := "cos":

arccos::diff :=
  proc(f)
    local op1;
  begin 
    op1 := op(f, 1);
    -diff(op1, args(2..args(0))) / sqrt(1-op1^2)
  end_proc:

arccos::complexDiscont :=
   loadproc(arccos::complexDiscont, pathname("STDLIB", "DISCONT"), "arccos"):
  
arccos::realDiscont := {}:
arccos::undefined := {}:

arccos::series := loadproc(arccos::series, pathname("SERIES"), "arccos"):

arccos::rectform :=
    loadproc(arccos::rectform, pathname("STDLIB","RECTFORM"), "acos"):

arccos::Content := (Out, data) -> Out::Capply(Out::Carccos, Out(op(data))):

// end of file 
