


// O. Cormier : July 2002

/*
kummerU(a,b,z) is, with kummerM(a,b,z)=1F1(a,b,z), a solution of the
Kummer ODE : z*y''(z)+(b-z)*y'(z)-ay(z)=0

Some particular cases are implemented for particular values of a and b.

Reference : Abramowitz and Stegun (A&S), Chap. 13
*/


kummerU:=
proc(a, b, z)
local fa, fb, fz, Z, ra, f;
//option remember;
begin
  if args(0) <> 3 then
    error("wrong number of arguments")
  elif z::dom::kummerU <> FAIL then
    return( z::dom::kummerU(args()) )
  end_if:

  if testargs() then
    if not (testtype(a, Type::Arithmetical) and
            testtype(b, Type::Arithmetical) and
            testtype(z, Type::Arithmetical)) then
      error("arguments must be of 'Type::Arithmetical'")
    end_if;
  end_if:

  fa:= domtype(float(a));
  fb:= domtype(float(b));
  fz:= domtype(float(z));
  if (fa=DOM_FLOAT or fa=DOM_COMPLEX) and (fb=DOM_FLOAT or fb=DOM_COMPLEX)
    and (fz=DOM_FLOAT or fz=DOM_COMPLEX)
    and has([map((op(a),op(b),op(z)),type)],DOM_FLOAT) then
      return(kummerU::float(a,b,z))
  end_if:

  Z:=genident();
  
//  if a=1 then kummerU(1,b,z)=z^(1-b)*kummerU(2-b,2-b,z) by A&S 13.1.29,
//    and then use A&S 13.6.28 : kummerU(1-a,1-a,z)=exp(z)*igamma(a,z) :
  if iszero(a-1) then
    if iszero(z) then
      if fb=DOM_COMPLEX or fb=DOM_FLOAT then
        return(limit(expand(Z^(1-b)*exp(Z)*igamma(b-1,Z)), Z=0))
      end_if:
    else
      return(expand(z^(1-b)*exp(z)*igamma(b-1,z)))
    end_if:
  end_if:

//  if a is a positive integer greater than 1 then we use recurrence
//    relation A&S 13.4.17 :
  ra:=round(a);
  if domtype(ra)=DOM_INT and iszero(a-ra) and ra>1 then
    return(1/(a-1)*(kummerU(a-1,b,z)-kummerU(a-1,b-1,z)))
  end_if:

//  if a is zero or a negative integer, then kummerU can be expressed
//    in terms of the Laguerre polynomial (c.f. A&S 13.6.27) :
  ra:=round(-a);
  if iszero(a) or (domtype(ra)=DOM_INT and ra > 0 and iszero(-a-ra)) then
      f:=expand((-1)^(ra)*Z^(-b+1)*exp(Z)*diff(exp(-Z)*Z^(ra+b-1),Z$(ra)));
      return(normal(expand(subs(f, Z=z))))
  end_if:

//  same thing if a=b-1 using A&S 13.1.29 :
  ra:=round(-a+b-1);
  if iszero(-a+b-1) or (domtype(ra)=DOM_INT and ra>0 and iszero(-a+b-1-ra))
    then
      if iszero(z) then
        if fb=DOM_COMPLEX or fb=DOM_FLOAT then
          return(limit(Z^(1-b)*kummerU(-ra,2-b,Z), Z=0))
        end_if:
      else
        return(z^(1-b)*kummerU(-ra,2-b,z))
      end_if:
  end_if:

//  if b=2a then kummerU can be expressed in terms of modified besselK function
//    using A&S 13.6.21 :
  if iszero(b-2*a) then
    if iszero(z) then
      if fa=DOM_COMPLEX or fb=DOM_FLOAT then
        return(limit(expand(1/sqrt(PI)*exp(Z/2)*Z^(-a+1/2)
                            *besselK(a-1/2,Z/2)), Z=0))
      end_if:
    else
      return(simplify(expand(1/sqrt(PI)*exp(z/2)*z^(-a+1/2)
                             *besselK(a-1/2,z/2))))
    end_if:
  end_if:

//  if a=b then one uses A&S 13.6.28 :
  if iszero(a-b) then
    if iszero(z) then
      if fa=DOM_COMPLEX or fb=DOM_FLOAT then
        return(limit(expand(exp(Z)*igamma(1-a,Z)),Z=0))
      end_if:
    else
      return(expand(exp(z)*igamma(1-a,z)))
    end_if:
  end_if:

  procname(a, b, z)        
end_proc:


kummerU := funcenv(kummerU):
kummerU::type := "kummerU":
kummerU::print := "kummerU":
kummerU::info := "kummerU(a,b,z) -- confluent hypergeometric function,
solution of the Kummer differential equation.":


kummerU::float :=
proc()
local a, b, z, fa, fb, fz, Z, ra, f, rb, s1, n, pochhammer, k, s2, l, C, S,
  s2k, extra;
  save DIGITS;
begin
  a:=args(1); b:=args(2); z:=args(3);
  fa:=float(a); fb:=float(b); fz:=float(z);
  if not((domtype(fa)=DOM_FLOAT or domtype(fa)=DOM_COMPLEX) and
         (domtype(fb)=DOM_FLOAT or domtype(fb)=DOM_COMPLEX) and
         (domtype(fz)=DOM_FLOAT or domtype(fz)=DOM_COMPLEX)) then
    return(kummerU(fa,fb,fz))
  end_if:

//  if z is large (let say |z|>10) then one uses A&S 13.5.2.
//  Apparently, this formula seems to give wrong results with MuPAD when
//  z is negative (I've made some comparisons with Maple and Mathematica),
//  so I added the following testtype :
  if abs(fz) >= float(10) and not(testtype(fz, Type::Negative)) or
    abs(fz) >= float(50) then
      return(float(z^(-a)*hypergeom([a,1+a-b],[],-1/z)))
  end_if:

  Z:=genident();
  if iszero(a-1) then
    if iszero(fz) then
      return(float(limit(expand(Z^(1-b)*exp(Z)*igamma(b-1,Z)), Z=0)))
    else
      return(float(combine(z^(1-b)*exp(z)*igamma(b-1,z),exp)))
    end_if:
  end_if:

  ra:=round(a);
  if domtype(ra)=DOM_INT and iszero(a-ra) and ra>1 then
    return(float(1/(a-1)*(kummerU(a-1,b,z)-kummerU(a-1,b-1,z))))
  end_if:

  ra:=round(-a);
  if iszero(a) or (domtype(ra)=DOM_INT and ra > 0 and iszero(-a-ra)) then
    f:=expand((-1)^(ra)*Z^(-b+1)*exp(Z)*diff(exp(-Z)*Z^(ra+b-1),Z$(ra)));
    return(float(normal(expand(subs(f, Z=z)))))
  end_if:

  ra:=round(-a+b-1);
  if iszero(-a+b-1) or (domtype(ra)=DOM_INT and ra > 0 and iszero(-a+b-1-ra))
    then
      if iszero(z) then
        return(float(limit(Z^(1-b)*kummerU(-ra,2-b,Z), Z=0)))
      else
        return(float(z^(1-b)*kummerU(-ra,2-b,z)))
      end_if:
  end_if:
    
  if iszero(b-2*a) then
    if iszero(z) then
      return(float(limit(expand(1/sqrt(PI)*exp(Z/2)*Z^(-a+1/2)
                                *besselK(a-1/2,Z/2)), Z=0)))
    else
      return(float(simplify(expand(1/sqrt(PI)*exp(z/2)*z^(-a+1/2)
                                   *besselK(a-1/2,z/2)))))
    end_if:
  end_if:
  
  if iszero(a-b) then
    if iszero(z) then
      return(float(limit(expand(exp(Z)*igamma(1-a,Z)),Z=0)));
    else
      return(float(expand(exp(z)*igamma(1-a,z))))
    end_if:
  end_if:


//  if b is not an integer and 1+a-b is not a negative integer, then one uses
//    formula A&S 13.1.3 :
  rb:=round(b); ra:=round(1+a-b);
  if not(domtype(rb)=DOM_INT and iszero(rb-b)) and
    not(domtype(ra)=DOM_INT and ra < 0 and iszero(ra-1-a+b)) then
      if iszero(z) and float(Re(b)) > float(1) then
        return(undefined)
      else
//   We have to add some extra digits since otherwise we may loose some
//        precision of the result. I hope extra is sufficient .... This comes
//        from the fact that, when z>0 and z->+infinity, then 1F1(a,b,z) has
//        exp(z)z^(a-b) has a factor : c.f. A&S 13.5.1
        if not iszero(z) then // expression below is not always defined for z=0
          extra:=max(0, abs(Re(round(log(10, exp(z)*z^(a-b))))))+1;
          DIGITS:=DIGITS+extra;
        end_if:
        return(float(PI/sin(PI*b)*(hypergeom([a],[b],z)/gamma(1+a-b)/gamma(b)
                     -z^(1-b)*hypergeom([1+a-b],[2-b],z)/gamma(a)/gamma(2-b))))
      end_if:
  end_if:

//  if b is zero or a negative integer then 2-b is a positive integer
//    and we use A&S 13.1.28 :
  if iszero(b) or (domtype(rb)=DOM_INT and rb < 0) then
    if iszero(z) then
      return(float(limit(Z^(1-b)*kummerU(a+1-b,2-b,Z), Z=0)))
    else
      return(float(z^(1-b)*kummerU(a+1-b,2-b,z)))
    end_if:
  else
//    if b is a positive integer then we use A&S 13.1.6 :
    if iszero(z) then
      Z:=genident();
      if traperror((l:=limit(kummerU(a,b,Z), Z)))=0 and not has(l,limit) then
          return(l)
        end_if:
    end_if:
    
    s1:=0; n:=rb-1;

    pochhammer :=
    proc(x,m)
    local t, k;
      begin
      t:=1;
      for k from 1 to m do t:=t*(x+k-1) end_for:
      return(t)
    end_proc:

//    We have to add extra digits not to loose precision, and for that, we
//      have that, when z>0 and z->infinity, the hypergeometric term in S
//      beyond is a factor of exp(z)*z^(a-n+1) :
    if Re(z) > 0 then
      extra:=max(0, abs(Re(round(log(10, exp(z)*z^(a-n+1))))))+1
    else
      extra:=max(0, abs(Re(round(log(10, exp(z)*(-z)^(-a))))))+1
    end_if:
    DIGITS:=DIGITS+extra;
    
    if not(iszero(n)) then
      s1:=_plus(pochhammer(a-n,k)/pochhammer(1-n,k)/fact(k)*z^k $ k=0..n-1);
      s1:=float(s1*z^(-n)*fact(n-1)/gamma(a));
    end_if;

    C:=float((-1)^(n+1)/fact(n)/gamma(a-n));
    S:=float(C*(hypergeom([a],[n+1],z)*ln(z))+s1);
    s2:=0; k:=0;
    repeat
      s2k:=float(gamma(a+k)/gamma(a)*gamma(n+1)/gamma(n+k+1)/fact(k)*z^k*(
                 psi(a+k)-psi(1+k)-psi(1+n+k)));
      s2:=float(s2+C*s2k);
      S:=float(S+C*s2k);
      k:=k+1;
    until abs(float(s2k)) < 10^(-DIGITS) end_repeat:
    return(S)
    
  end_if:

end_proc:


// derivation of the kummerU fonction is given by several formulas
// A&S 13.4.21 ... 13.4.28. We use all these relations to simplify
// as much as possible the derivative of kummerU, in such a way that
// it is a solution of the differential equation.
kummerU::diff :=
proc(e, x)
local a, b, f, Select, Select2;
begin
  if has({op(e,1),op(e,2)}, [args(2..args(0))]) then hold(diff)(args())
  elif args(0)>2 then diff(diff(e,x),args(3..args(0)))
  else
    a:=op(e,1); b:=op(e,2); f:=op(e,3);

    Select:= X -> _lazy_or(_lazy_and(type(X) = "_plus",
           _lazy_and(select([op(X)], testtype, Type::NegRat) = [],
                   not(Select(select(_plus(op(X)), testtype, DOM_COMPLEX))))),
         testtype(Re(X), Type::PosRat) ):

    Select2:= X -> _lazy_or(_lazy_and(type(X) = "_plus",
           _lazy_and(select([op(X)], testtype, Type::PosRat) = [],
                   not(Select(select(_plus(op(X)), testtype, DOM_COMPLEX))))),
         testtype(Re(X), Type::NegRat) ):
    
    if Select(a) then 
      if Select(b) then // 13.4.27
        (1/f*(1-b+f)*kummerU(a,b,f)-1/f*kummerU(a-1,b-1,f))*diff(f,x)
      elif Select2(b) and not(iszero(a-b)) then
        (f*(a-b+f)/(b-a)*kummerU(a,b+1,f)+f/(a-b)*kummerU(a-1,b,f))*diff(f,x)
      else // 13.4.26
        (1/f*(a-b+f)*kummerU(a,b,f)-1/f*kummerU(a-1,b,f))*diff(f,x)
      end_if:
    elif Select2(a) then
      if Select(b) then
        (a*(1-b)/f*kummerU(a+1,b,f)-a/f*kummerU(a,b-1,f))*diff(f,x)
      elif Select2(b) then
        (-a*kummerU(a+1,b+1,f))*diff(f,x)
      else
        (-a/f*kummerU(a,b,f)-a/f*(b-a-1)*kummerU(a+1,b,f))*diff(f,x)
      end_if:
    elif Select(b) then // 13.4.24
        (-1/f*(1+a-b)*kummerU(a,b-1,f)+1/f*(1-b)*kummerU(a,b,f))*diff(f,x)
    elif Select2(b) then // 13.4.25
        (kummerU(a,b,f)-kummerU(a,b+1,f))*diff(f,x)
    else // 13.4.23
      (-a/f*kummerU(a,b,f)-a/f*(b-a-1)*kummerU(a+1,b,f))*diff(f,x)
    end_if:
  end_if:
end_proc:


kummerU::series :=
proc(a, b, f, z, o, dir)
local fb, t, l, s, P, rb, ra, pochhammer, n, s1, s2, k;
begin
  fb:=float(b);

  t:=Series::series(f, z, o, dir);
  if domtype(t) = Series::Puiseux then
    k:=Series::Puiseux::ldegree(t);
    if k = FAIL then
      k:=Series::Puiseux::order(t);
      if k > 0 then
        return(t)
      end_if;
      Series::error("order too small")
    end_if;

    if k < 0 then // expansion at infinity : A&S 13.5.2
      l:=limit(f,z,dir);
      s:=Series::series(-1/f, z, o, dir);
      P:=series(hypergeom([a,1+a-b],[],z), z, o, dir) @ s;
      return(P * Series::series(f^(-a), z, o, dir));

    elif iszero(coeff(t,z,0)) and (type(fb)=DOM_FLOAT or type(fb)=DOM_COMPLEX)
    then  // expansion at 0 when b is not symbolic
      rb:=round(b); ra:=round(1+a-b);
  
      if not(domtype(rb)=DOM_INT and iszero(rb-b)) and
         not(domtype(ra)=DOM_INT and ra < 0 and iszero(ra-1-a+b)) then
// A&S 13.1.3
           return(series(PI/sin(PI*b)*(hypergeom([a],[b],f)/gamma(1+a-b)/gamma(b)-f^(1-b)*hypergeom([1+a-b],[2-b],f)/gamma(a)/gamma(2-b)),z,o))
      end_if:
    
      if iszero(b) or (domtype(rb)=DOM_INT and rb < 0) then //apply A&S 13.1.29
        return(series(f^(1-b)*kummerU(a+1-b,2-b,f),z,o))
      else // formula A&S 13.1.6
        b:=rb; s1:=0; n:=b-1;

        pochhammer :=
        proc(x,m)
        local t, k;
        begin
          t:=1;
          for k from 1 to m do t:=t*(x+k-1) end_for:
          return(t)
        end_proc:


        if not(iszero(n)) then
          s1:=_plus(pochhammer(a-n,k)/pochhammer(1-n,k)/fact(k)*f^k $k=0..n-1);
          s1:=s1*f^(-n)*fact(n-1)/gamma(a);
        end_if;
        s2:=_plus(pochhammer(a,k)/pochhammer(n+1,k)/fact(k)*f^k*(psi(a+k)
                  -psi(1+k)-psi(1+n+k)) $ k=0..o);
        return(series((-1)^(n+1)/fact(n)/gamma(a-n)*
                      (hypergeom([a],[n+1],f)*ln(f)+s2)+s1,z,o))
      end_if
    end_if:
  end_if:

  Series::unknown(kummerU(a,b,f), z, o, dir)
end_proc:
