/*===========================================================================

sum -- (in)definite summation 

Calling sequence:
sum(f, x)

Parameters:
f -- an expression (rational function)
x -- identifier

Summary:
sum(f,x) computes the indefinite sum of f(x) with respect to x.

Copyright for the original version:
Paul Zimmermann and Inria, November 1994

--------------------------------------------------------------------------

Examples for indefinite sums:

1) Polynomial case:

>> sum(1,k);

                                     k

>> sum(k^2,k);

                                     2    3
                                k   k    k
                                - - -- + --
                                6   2    3
 
>> sum(a*x^2+b*x+c,x);
 
                    3
                 a x      / a   b     \    2 /   a   b  \
                 ---- + x | - - - + c | + x  | - - + -  |
                  3       \ 6   2     /      \   2   2  /
 
 
2) Fractional case: 

a) Decomposition into the polynomial part and the rational part:
 
>> sum(k/(k+1),k);
 
                              k - psi(k + 1)
 
>> sum((x^2+1)/(x-1),x);
 
                              2
                         x   x     
                         - + -- + 2 psi(x - 1) 
                         2   2 
 
 
b) Reduced rational expressions:
   (degree of numerator is less than degree of denominator)

>> sum(1/k/(k-2),k);

                                1           1    
                          - --------- - ---------
                            2 (k - 1)   2 (k - 2)


>> sum(1/k/(2*k-1),k);

                          - psi(k) + psi(k - 1/2)


==========================================================================*/


sum :=
proc(s, x)
  local k, l, r, a, y, t, branches, conditions, options: DOM_TABLE;
  
begin
  if args(0) = 0 then
    error("Wrong number of arguments")
  end_if;
  
  // overloading
  if s::dom::sum <> FAIL then
    s::dom::sum(args())
  end_if;
  
  if args(0) < 2 then
    error("at least two arguments expected")
  end_if;

  if testargs() then
    if type(x) = "_equal" then
      if domtype(op(x, 1)) <> DOM_IDENT and type(op(x,1)) <> "_index" then
        error("summation variable must be an identifier or indexed identifier")
      elif type(op(x, 2)) = "_range" then
        if map(op(x, 2), testtype, Type::Arithmetical) <> (TRUE..TRUE) then
          error("bounds must be of Type::Arithmetical")
        end_if;
        if is(op(x, [2, 1]) <= op(x, [2, 2])) = FALSE then
          error("Left border must not exceed right border")
        end_if;
      elif domtype(op(x, 2)) <> RootOf then
        error("range or RootOf expected as rhs")
      elif has(op(x, 2), op(x, 1)) then
        error("summation variable occurs in rhs")
      elif stdlib::hasfloat(op(x, 2)) then
        error("floating-point numbers not allowed in rhs")
      end_if;
    elif type(x) = "_in" then
      if domtype(op(x, 1)) <> DOM_IDENT and type(op(x,1)) <> "_index" then
        error("summation variable must be an identifier or indexed identifier")
      elif testtype(op(x, 2), Type::Set) = FALSE then
        error("summation variable must range over a set")
      elif has(op(x, 2), op(x, 1)) then
        error("summation variable occurs in rhs")
      end_if
    elif domtype(x) <> DOM_IDENT and type(x) <> "_index" then
      error("identifier, indexed identifier, or equation expected ".
            "as 2nd argument")
    end_if
  end_if;


  
  // simple form
   if type(x) = "_equal" and type(op(x, 2)) = "_range" then
    // definite summation; make assumptions about the summation
    // variable if possible
     k := op(x, 1);
     [l, r]:= [op(op(x, 2))];
     if protected(k) = hold(None) then
       save k;
       assume(k, property::IV([l, r]));
       assume(k, Type::Integer, _and);
     end_if;

     a:= r - l;
     for y in select(indets(a) minus Type::ConstantIdents,
                     ind -> protected(ind) = hold(None)) do
       // we solve the condition a >= 0 for y, provided that
       // a is linear in y
       if (t:= Type::Linear(a, [y])) <> FALSE then
         // a = t[1] * y + t[2]
         // thus a >= 0 iff t[1] is positive and y >= -t[2]/t[1]
         // or t[1] is negative and y <= -t[2]/t[1] or t[1] = 0
         if is(t[1] > 0) = TRUE then
           save y;
           assume(y >= -t[2]/t[1], _and)
         elif is(t[1] < 0) = TRUE then
           save y;
           assume(y <= -t[2]/t[1], _and)
         end_if
       end_if;
     end_for;


     s:= eval(s)
   end_if;


  if testargs() then
    if domtype(s) <> DOM_POLY and not testtype(s, Type::Arithmetical) then
      error("1st argument must be of Type::Arithmetical")
    end_if
  end_if;

  if type(s) = "solve" then
    // cannot sum over solve(ode(..)) and solve(rec(..))
    return(procname(args()))
  end_if;

  options:= prog::getOptions(sum, 2, hold(Error));
  
  if domtype(s) = DOM_POLY then
    s:= expr(s)
  end_if;
  
  a:= Sum::sum(s, x, options);

  if domtype(a) <> piecewise then
    return(a)
  end_if;
  
  // fill "uncovered case" if the result is a piecewise
  // and the condition allows to replace an identifier
  t:= not _and(op(map([op(a)], op, 1)));
  if type(t) = "_or" then
    branches:=[op(t)]
  else
    branches:= [t]
  end_if;

  conditions:= map({extop(a)}, op, 1);
  for t in branches do
    if testtype(t, DOM_IDENT = Type::Constant) then
      t:= piecewise::flattenBranch([t, sum(subs(s, t), subs(x, t), options)]);
      for y in [t] do
        if not contains(conditions, op(y, 1))
          then
          conditions:= conditions union {op(y, 1)};
          a:= piecewise::insert(a, y, 1)
        end_if
      end_for;
    end_if;
  end_for;

  a
  
end_proc:

sum:= funcenv(sum,
proc(a)
  local eq, sumSign;
begin
  if PRETTYPRINT = TRUE then
    sumSign := "---",
               "\\  ",
               "/  ",
               "---":
    if type(op(a, 2)) = "_equal" then
      // definite sum or RootOf
      eq := op(a, [2,2]);
      if domtype(eq) = RootOf then
        _outputSequencePlus(stdlib::Exposed(stdlib::align(sumSign,
                                                      expr2text(op(a, [2,1])=eq))),
                        stdlib::Exposed(" "), op(a, 1));
      else
        _outputSequencePlus(stdlib::Exposed(stdlib::align(expr2text(op(eq, 2)),
                                                      sumSign,
                                                      expr2text(op(a, [2,1])=op(eq, 1)))),
                        stdlib::Exposed(" "), op(a, 1));
       end_if;
     else
      // indefinite sum
      _outputSequencePlus(stdlib::Exposed(stdlib::align(sumSign,
                                                    expr2text(op(a, 2)))),
                      stdlib::Exposed(" "), op(a, 1));
              
     end_if;
  else
    FAIL
  end:
end_proc):
sum:= slot(sum, "type", "sum"):
sum:= slot(sum, "print", "sum"):

sum::options:= table(MaxOrder = 2,
                     ExpandFinite = 1000):
sum::optionTypes:= table(MaxOrder = Type::PosInt,
                         ExpandFinite = Type::NonNegInt):


sum::int:=
proc(s,x)
begin
  sum(int(op(s,1),x),op(s,2))
end_proc:

sum::Simplify:=
proc(S)
  local newS;
begin
  // simplify only the summand
  newS:= Simplify(op(S, 1));
  if newS <> op(S, 1) then
    eval(hold(sum)(newS, op(S, 2..nops(S))))
  else
    S
  end_if
end_proc:


sum::Content      :=
  proc(Out, data)
    local bvar;
  begin
    if type(op(data, 2)) = "_equal" then
      if type(op(data, [2,2])) = "_range" then
        Out::Capply(Out::Csum,
                    Out::Cbvar(Out(op(data, [2,1]))),
                    Out::Clowlimit(Out(op(data, [2,2,1]))),
                    Out::Cuplimit( Out(op(data, [2,2,2]))),
                    Out(op(data,1)));
      else
        bvar := Out::Cbvar(Out(op(data, [2,1])));
        Out::Capply(Out::Csum, bvar,
                    Out::Ccondition(
                          Out::Capply(Out::Cin,
                                      bvar,
                                      Out(op(data, [2,2])))),
                    Out(op(data,1)));
      end_if:
    else
      Out::Capply(Out::Csum,
                  Out::Cbvar(Out(op(data, 2))),
                  Out(op(data,1)));
    end_if;
  end_proc:

sum::float := loadproc(sum::float, pathname("NUMERIC"), "floatsum"):

sum::diff :=
proc(e) // e = sum(f, x) 
  local l, _x;
begin
  l := [args(2..args(0))]; // variables to differentiate
  if map({op(l)}, testtype, Type::Indeterminate) = {TRUE}
    and not has((_x := op(e, 2)), l) then
     // summation independent of differentiation 
    sum(diff(op(e, 1), op(l)), _x)
  else
    hold(diff)(e, op(l))
  end_if
end_proc:

sum::expand:=
proc()
  local s, d, l, k, f, u;
begin
  if type(args(2)) = "_equal" and
    type((k:= op(args(2), 2))) = "_range" and
    domtype((d:= op(args(2),[2,2]) - op(args(2), [2, 1]))) = DOM_INT then
    // left border must not exceed right border, this should
    // have been checked before
    assert(d>=0);
    l:= op(args(2), [2, 1]);
    k:= op(args(2), 1);
    f:= subs(args(1), k = k + l);
    f:= _plus(eval(subs(f, k=u)) $ u=0..d);
    return(eval(%))
  end_if;
  s:=expand(args(1));
  if type(s)="_plus" then
    _plus(map(s, expand@sum,args(2)))
  elif type(s)="_mult" then
    split(s,has,(if type(args(2))=DOM_IDENT
                   then
                   args(2)
                 else
                   op(args(2),1)
                 end_if));
    %[2]*hold(sum)(%[1],args(2))
  else
    hold(sum)(args())
  end_if
end_proc:

//--------------------------------------------------------------------------
Sum:= newDomain("Sum"):
Sum::create_dom:=hold(Sum):
Sum::info:= "Sum: Utilities for definite and indefinite summation":
Sum::interface:= {}:

Sum::sum:=
proc(s: Type::Arithmetical, x, options: DOM_TABLE)
  local f,f_orig,newn,fnew, g,gg,_k_,Rec,RRec,u,_d_,l,lc, r, _p,n,q,z,clean,
  i, j,
  printWarnings: DOM_BOOL;
begin
  if stdlib::hasfloat(s) then
    s:= numeric::rationalize(s);
    return(float(sum(s, x)))
  end_if;
   
    
  
  n:= if type(x) = DOM_IDENT or type(x) = "_index" then
        x
      else
        op(x, 1)
      end_if;
  
  f:= misc::maprec(s, {"cos", "sin", "cosh", "sinh"} =
                   proc(X)
                     begin
                       if has(X, n) then
                         rewrite(X, exp)
                       else
                         X
                       end_if
                   end_proc
                          );

  if has(f, exp) then
    f:=eval(subs(f,hold(exp) = (x->`#E`^x)))
  end_if;

  // include, because in some branches `#E` survived. Walter, 24.10.99
  //  sum( exp(n*x)/n, n=0..infinity);
  clean:=
  proc(result)
  begin
    if domtype(result) = piecewise then
      return(piecewise::extmap(piecewise::mapConditions(result, clean), clean))
    end_if;
    if has(result, `#E`)
      then
      subs(misc::maprec(combine(result),
                        {"_power"}= proc(x)
                                    begin
                                      if op(x,1)=`#E` then
                                        exp(op(x,2))
                                      else
                                        x
                                      end_if
                        end_proc
                        ), `#E`=exp(1));
    else
      result;
    end_if;
  end_proc;

  
  if domtype(x) = DOM_IDENT or type(x) = "_index" then
// to do: if f is a sum, try to sum the individual operands first
    f := Sum::indefinite(f, x)
  elif type(x) = "_in" then
    // can only sum over finite sets and RootOf's at the moment
    case type(op(x, 2))
      of RootOf do
        f:= Sum::rootOf(op(x, 2), op(x, 1), s);
        if f = FAIL then
          return(hold(sum)(s, x))
        else
          return(f)
        end_if
      of DOM_SET do
        return(_plus(subs(s, op(x,1) = op(x, [2, j])) $j=1..nops(op(x, 2)) ))
      otherwise
        return(hold(sum)(s, x))
    end_case;
  elif type(x)="_equal" then
    _k_:= op(x, 1);
    if type(op(x, 2))="_range" then
      // definite summation

      if map({op(op(x, 2))}, domtype) intersect
        {DOM_FLOAT, DOM_RAT, DOM_COMPLEX} <> {} then
        error("Illegal range")
      end_if;

      l:= op(x, [2, 1]);
      r:= op(x, [2, 2]);

      if l=r then // one-point range
        return(Sum::myeval(s, _k_, l))
      end_if;

     
      if domtype((g:= r - l)) = DOM_INT and g < options[hold(ExpandFinite)]
        then
        // finite range
        assert(g>=0);
        fnew:= eval(_plus(eval(subs(s,_k_=u+l)) $ u=0..g));

        //------------------------------------------------------
        // Walter 11.1.2005: sum(matrix([[1]])^k/k!, k = 1..2)
        // produced matrix([[1/1! + 1/2!]]) due to insufficient
        // evaluation. The following map(.., eval) cures this.
        if fnew::dom::hasProp( Cat::Matrix ) = TRUE then
           fnew:= map(fnew, eval):
        end_if;
        //------------------------------------------------------

        // if fnew is so simple that no improvement seems possible, return
        if testtype(fnew, Type::Numeric) then
          return(fnew)
        end_if;
        // otherwise, compare with result obtained by using indefinite summatio
        if traperror((g := Sum::indefinite(f, _k_))) <> 0  
        or type(g) = "sum" then
          return(fnew)
        end_if;
        printWarnings := stdlib::limit::printWarnings(FALSE):
        g:= clean(g);
        if traperror(
                     (g:= Sum::myeval(g,_k_,r+1)
                      - Sum::myeval(g,_k_, l))
                     ) <>0 then
          return(fnew)
        end_if;
        f:= normal(g);
        if length(f) > length(g) then
          f:= g
        end_if;
        if length(fnew) < length(f) then
          return(fnew)
        else
          return(f)
        end_if         
        
      end_if;
      
      
      // apply some heuristics for infinite ranges

      if l = -infinity and r = infinity then
        // as experience shows, splitting saves time here although
        // two sums have to be computed
        return(Sum::sum(f, _k_ = l .. -1, options) +
               Sum::sum(f, _k_ = 0 .. r, options))
      
      elif l = -infinity and not has(r, infinity) then
        // sums of the form \sum_{k=-\infty}^{n} f(k)
        // are normalized to \sum_{k=-n}^{\infty} f(-k)
        f:= subs(f, _k_= -_k_);
        l:= -r;
        r:= infinity;
        x:= _k_ = l .. r
      end_if;

      if r = infinity and testtype(f, Type::RatExpr(_k_)) = TRUE then
        f:= normal(f);
        n:= degree((u:= numer(f)), [_k_]) - degree((q:= denom(f)), [_k_]);
        assert(type(n) = DOM_INT);
        u:= lcoeff(u, [_k_]);
        q:= lcoeff(q, [_k_]);
        if is(u<>0 and q<>0) = TRUE and n>= -1 then
          // the sum cannot converge
          lc:= u/q;
          if n mod 2 = 1 and l = -infinity then
            return(undefined)
          else
            return(sign(lc) * infinity)
          end_if
        end_if;
      end_if;
      
        
      g := Sum::indefinite(f, _k_);
      if type(g) = "sum" or
        piecewise::disregardPoints(g) = undefined then
        if type(f) = "_mult" then
          // do not use expand, since we only
          // want to expand products, but we
          // do not apply addition theorems for sin, cos, etc.
          l:= [op(f)];
          if (u:= contains(map(l, type), "_plus")) > 0 then
            q:= l[u];
            delete l[u];
            _mult(op(l));
            fnew:= _plus(op(q, i)*%  $i=1..nops(q))
          else
            fnew:= f
          end_if
        else
          fnew:= f;
        end_if;
        if type(fnew) = "_plus" then // try to map
        l := [];
          for q in fnew do
            u := Sum::sum(q,x, options);
            if type(u) = "sum" or type(u) = "solve" then
            // no success: either an unevaluated sum
            // or an unsolved rec was returned  
              break
            else
              l := append(l, u)
            end_if
          end_for;
          if nops(l) = nops(fnew) then
            return(clean(_plus(op(l))));
          end_if;
        end_if;
        
        userinfo(1,"indefinite summation failed");
        if Sum::ishypergeom(f,_k_) then
          userinfo(1,"summand is hypergeometric");
          // try Zeilberger's algorithm 
          // choose in priority indets that are in the bounds 
          l:= indets(op(x,2)) minus Type::ConstantIdents minus {`#E`};
          l:=[op(l), op((indets(f) minus Type::ConstantIdents minus {`#E`})
                        minus (l union {_k_}))];
          newn:= genident();
          for n in l do
            // replace n by newn where it cannot be helpful because
            // Sum::ratio surely would not produce a rational expression in n
            // we may then compute the sum and substitute back, since n
            // is independent of the summation variable  
            f_orig:= f;
            f:= misc::maprec
            (f,
             (t -> not contains({"_power", "_plus", "_mult", DOM_IDENT,
                                "binomial", "fact", "gamma"}, type(t))) =
             (a -> subs(a, n = newn)),
             PostMap);
            f:= misc::maprec(f, {"_power"} =
                             (a -> subsop(a, 2 = subs(op(a, 2), n = newn))));
            if Sum::ishypergeom(f,n) then
              userinfo(1,"trying Zeilberger's algorithm with respect to",n);
              u:= genident("u");
              Rec:= Sum::zeilberger(f, _k_, n, u, op(op(x,2)), options);
              if Rec = FAIL then
                userinfo(1,"Zeilberger's algorithm fails");
                f:= f_orig;
                next
              else
                userinfo(1,"Zeilberger's algorithm succeeds")
              end_if;
              gg:=Rec[2]; RRec:=Rec[1];
              _d_:=op(RRec,[1,1])-n;
              _p:= proc(z)
                     name Sum::sum;
                   begin
                     eval(subs( hold(sum)(f,_k_=op(x,2)), n=z ))
                   end_proc;

              // it seems that more than one shift is necessary in
              // some cases
              while traperror((g:={(u(z)=_p(z))$ z=gg-_d_..gg}))<>0 do
                gg:= gg+1
              end_while;
              
              Rec:=solve(rec(RRec,u(n),g));
              if type(Rec) = "solve" then
                // We specified one more initial condition than necessary in
                // the previous call.
                // ( This is necessary since otherwise, due to the changed
                //   output format of rec::solve, the
                //   solution returned by rec::solve may still contain free
                //   parameters ... CRUDE HACK!
                //   Example: sum(k*binomial(n, k), k = 0..n) )
                // Since it was not successful, it
                // sometimes helps to specify one less initial condition.
                // Example: sum(binomial(d,j)*binomial(d-r-j-1,r-j),j=0..r)
                // ( Reason: rec::solve uses linsolve which does not recognize
                //   the functional identity of the gamma function)
                if traperror((g:={(u(z)=_p(z))$ z=gg-_d_..gg-1}))<>0 then
                  g:={(u(z)=_p(z))$ z=gg-_d_+1..gg}
                end_if;
                Rec:=solve(rec(RRec,u(n),g));
              end_if:
              userinfo(5, "Solving recurrence equation gives ".
                       expr2text(Rec));
              if not has(Rec,undefined) then
                Rec:= piecewise::extmap
                (Rec,
                 proc(Rc)
                 begin
                   if type(Rc)=DOM_SET and nops(Rc)>=1 then
                     op(Rc,1)
                   else
                     Rc
                   end_if
                 end_proc);
                delete _p;
                return(clean(subs(Rec, newn = n)))
              end_if;
              delete _p;
            end_if; // ishypergeom(f, n)
            f:= f_orig;
          end_for; // for n in l
          // f is hypergeometric but Zeilberger's algorithm fails 
          userinfo(1,"trying definite summation of hypergeometrics");
          g:=Sum::hypergeomtype(f,_k_);
          g:= clean(op(g, 1)), clean(op(g, 2)), clean(op(g, 3));
          userinfo(2,"hypergeometric type is",g);
          if g<>FAIL then
            g:=Sum::hypergeomeval(g,op(op(x,2)),clean(f),_k_);
            if g<>FAIL then
              return(g)
            end_if
          end_if;
          if g=FAIL then g:=Sum::recognize_gf(f,_k_,op(op(x,2)));
            if g<>FAIL then
              return(clean(g))
            end_if
          end_if;
          userinfo(1,"definite summation of hypergeometrics failed");
        end_if;

            // not hypergeometric in k or in another variable 
// changed (Walter, 13.6.01): why should unsucessful sum return 
// the *changed* argument rather than the original argument?
// So: do *not* replace s (= the original argument) by f (= the changed argument)
     // s:= f; // return changed argument
        f:= FAIL // hold(sum)(f,x)



      else // indefinite summation succeeded 
        userinfo(2,"indefinite summation gave",g);
        printWarnings := stdlib::limit::printWarnings(FALSE):
        g:= clean(g);
        if traperror(
                     (g:= Sum::myeval(g,_k_,op(x,[2,2])+1)
                      - Sum::myeval(g,_k_,op(x,[2,1])))
                     )=0 then
          f:= normal(g);
          if length(f) > length(g) then
            f:= g
          end_if;
        else
          // Sum::myeval raises an error iff evaluating 
          error("Singularity");
        end_if:
        stdlib::limit::printWarnings(printWarnings):
      end_if
    elif type((_p:=op(x,2))) = RootOf then // sum(..., k=RootOf(q,z)) 
      userinfo(1,"sum over RootOf");
      f:= Sum::rootOf(_p, _k_, f)
    end_if // type(op(x, 2))
  end_if; // type(x)

  if f = FAIL
    then
          // changed (Walter 13.6.01): s is now the *original* argument as provided
          // by the user. No need to clean it:
          // return( hold(sum)(eval(clean(s)),x) )
    return(hold(sum)(s, x));
  else
    return(clean(f));
  end_if;
end_proc:

Sum::rootOf:=
proc(p: RootOf, k, f)
  local z, q;
begin
  z:=op(p,2);
  q:=eval(subs(op(p,1),z = k));
  // sum(..., x=RootOf(1, y)) simplifies to zero
  if contains({DOM_INT,DOM_RAT},type(q)) then
    return(0)
  end_if;
  if testtype(f, Type::RatExpr(k)) then
// first simplify to a sum over a polynomial
    userinfo(2,"summand is rational");
    f:= stdlib::RootOf_evala(f,q,k)
  end_if;
  if testtype(f,Type::PolyExpr(k)) then // evaluate
    f:= stdlib::RootOf_polysum(f,q,k)
  else
    f:= FAIL // hold(sum)(args())
  end_if;
  f
end_proc:

/* the following extension was suggested by Daniel Duparc
  <dduparc@club-internet.fr> */
Sum::recognize_gf:=proc(f,n,a,b) local s,P,x,Q,lc,d;
begin
   P:=genident("P"); x:=genident("x");
   // recognizes sum(P(n)*x^n/n!, n=0..infinity) 
   if a=0 and b=infinity then
      s:=intlib::match(n!*f, _pattern(P,Type::PolyExpr(n))*_pattern(x)^n);
      if s<>FAIL then
         P:=eval(subs(P,s)); x:=eval(subs(x,s)); Q:=0;
         // x^i*exp(x) -> n*(n-1)*...*(n-i+1) = n!/(n-i)! 
         while P<>0 do
            lc:=lcoeff(P,[n]); d:=degree(P,[n]);
            Q:=Q+lc*x^d; P:=expand(P-lc*n!/(n-d)!);
         end_while;
         return(Q*exp(x));
      end_if
   end_if;
   FAIL
end_proc:

// no longer used and therefore commented out. Use Sum::myeval below instead.
// (jngerhar, 10.10.01)
//Sum::eval := proc(f,x,v) // evaluate value of f at x=v 
//begin
//   if v=infinity then
//      limit(f,x=infinity)
//   else /* do no eval here ==> problems if f contains the name "x",
//          for example sum(x^k*y^(n-k),k=0..n) or sum(x^k,k=1..2) */
//      eval(subs(f,x=v))
//   end_if
//end_proc:

// returns h such that h(x+1)-h(x)=f(x) 
Sum::indefinite:=
proc(f,x)
  local prod;
begin
  userinfo(1,"entering indefinite summation");
  // handle ln separately
  if type(f) = "ln" then
    // use sum(ln(g(n)), n) = ln(product(g(n), n))
    prod:= product(op(f, 1), x);
    if type(prod) <> "product" then
      return(ln(prod))
    end_if;
  end_if;
  // first: expand all factorials
  f:= misc::maprec(f, {"fact"} = expand);
  if testtype(f,Type::PolyExpr(x,Type::AnyType)) then
    return(Sum::poly(f,x))
  elif testtype(f,Type::RatExpr(x,Type::AnyType)) then
    // do not use properties in this step
    save x;
    unassume(x);
    return(Sum::rat(f,x))
  elif type(f) = "_plus" and nops(f) = 2 then
    if subs(op(f, 1), x = x - 1) = -op(f, 2) then
       // f(x) is syntactically  h(x + 1) - h(x)
      return(-op(f, 2))
    elif subs(op(f, 1), x = x + 1) = -op(f, 2) then
       // f(x) is syntactically  h(x) - h(x + 1)
      return(-op(f, 1))
    end_if
  end_if;
  return(Sum::gosper(f,x))
end_proc:

// p is a polynom in x, returns h such that p = eval(subs(h,x=x+1))-h 
Sum::poly := proc(p,x)
local d,i,b,a;
begin
  if not has(p,x) then p*x
  else // method of indeterminate coefficients 
   p:=poly(p,[x]); d:=degree(p);
   // first tabulate a[i]=(x+1)^i 
   a[1]:=poly(x+1,[x]); for i from 2 to d+1 do a[i]:=a[1]*a[i-1] end_for;
   for i from d+1 downto 2 do
      b[i]:=coeff(p,i-1)/i;
      p:=p+multcoeffs(lmonomial(a[i],Rem)[2],-b[i]);
   end_for;
   b[1]:=coeff(p,0);
   _plus(b[i]*x^i $ i=1..d+1)
  end_if
end_proc:

Sum::rat := proc(f,x)
local num,den,d,dnum,dden;
begin
   num:=numer(f); den:=denom(f);
   dnum:=degree(num,x); dden:=degree(den,x);
   if dnum>=dden then
      d:=divide(poly(num,[x]),poly(den,[x]));
      Sum::poly(expr(d[1]),x)+Sum::rat(expr(d[2])/den,x);
   else // now deg(num)<deg(den) 
      d:=Sum::rat_abramov(f,x);
      if d=FAIL then hold(sum)(f,x) else d end_if
   end_if
end_proc:

/* f is minimal : sum(f,x) can not be written p+sum(r,x)
  where p is a polynomial and denom(r) has lower degree than denom(f)
*/
Sum::to_psi :=
proc(f,x)
  local ff,t,s,a,d,lin;
begin
  ff:=partfrac(f,x,hold(full));
  userinfo(3,"full partial fraction decomposition=",ff);
  if type(ff)="_plus" then
    ff:=[op(ff)]
  else
    ff:=[ff]
  end_if;
  s:=0;
  for t in ff do
    d:= denom(t);
    d:=d/maprat(d,proc() begin content(args(1),[x]) end_proc );
// primpart(d,[x]) expands too !!!
    if type(d)="_power" and testtype(op(d,2),Type::NonNegInt) then
      lin:=op(d,1)
    else
      lin:=d
    end_if;
    if testtype(lin,Type::PolyExpr(x,Type::AnyType)) then
      case degree(lin,[x])
        of 1 do
          a:=coeff(lin,[x],0)/coeff(lin,[x],1);
          s:=s+subs(Sum::to_psi2(normal(eval(subs(t,x=x-a))),x),x=x+a);
          break;
        otherwise
          return(FAIL)
      end_case
    else // non polynomial
      return(FAIL)
    end_if
  end_for;
  s
end_proc:

Sum::is_linear:=
proc(p,x) // checks for a*x+b 
begin
  if testtype(p,Type::PolyExpr(x,Type::AnyType)) then
    if degree(p,x)=1 then
      return(TRUE)
    end_if
  end_if;
  FALSE
end_proc:

Sum::to_psi2 := proc(f,x) // the denominator looks like a*x^j 
// f must be in normal form 
local n,j,res,d,c,dd,t;
begin
   res:=0;
   d:=denom(f);
   c:=maprat(d,proc() begin content(args(1),[x]) end_proc );
   d:=normal(d/c); // should be given by primpart 
   if d=-x then d:=x; c:=-c end_if;
   if not ((d=x) or (type(d)="_power" and op(d,1)=x and type(op(d,2))=DOM_INT))
   then return(FAIL) // should not happen 
   end_if;
   dd:=degree(d,x)-1;
   n:=numer(f);
   if type(n)="_plus" then n:=[op(n)] else n:=[n] end_if;
   for t in n do
      j:=dd-degree(t,x);
      res:=res+maprat(t,proc() begin lcoeff(args(1),[x]) end_proc )/c*(-1)^j/fact(j)*psi(x,j);
   end_for;
   res
end_proc:

Sum::rat_abramov :=
proc(f,x)
begin
   userinfo(1,"entering Abramov's algorithm");
   Sum::abr(f, Sum::dispersion(denom(normal(f)),x),x)
end_proc:

Sum::abr :=
proc(f,dis,x)
  local p,q,cp,vwp,vwm,ba,u,newf;
begin
  userinfo(2,"dispersion=",dis);
  if f=0 then
    return(0)
  end_if;
  if dis=0 then
    return(Sum::to_psi(f,x))
  end_if;
  p:=numer(f); q:=denom(f);
  cp:=gcd(q,subs(q,x=x+dis));
  if cp=1 then return(Sum::abr(f,dis-1,x)) end:
  vwp:=Sum::abr_part(q,cp, x);
  vwm:=Sum::abr_part(q,subs(cp,x=x-dis), x);
  if degree(vwm[1],x)>degree(vwp[1],x) then
    ba:=solvelib::pdioe(vwp,p,x);
    u:=subs(ba[2]/vwp[1],x=x-1);
    newf:=normal(ba[1]/vwp[2]+u);
  else
    ba:=solvelib::pdioe(vwm,p,x);
    u:=-ba[2]/vwm[1];
    newf:=normal(ba[1]/vwm[2]+subs(ba[2]/vwm[1],x=x+1))
  end_if;
  u+Sum::abr(newf,dis-1,x)
end_proc:

Sum::abr_part :=
proc(p, h, x: DOM_IDENT)
  local g,v,w;
begin
  g:=gcd(p,h); v:=1; w:=p;
  while degree(g, x) > 0 do
    w:=divide(w,g,Quo);
    v:=v*g;
    g:=gcd(w,v)
  end_while;
  v,w
end_proc:

Sum::dispersion :=
proc(q,x) // q is a polynomial 
  local X,r,i,s;
begin
  X:= genident("X"):
  q:=normal(q/content(q,[x]));
  r:=polylib::resultant(q,subs(q,x=x+X),x);
  r:=factor(r);
  s:={};
  for i from 2 to nops(r) step 2 do
    // this is because iroots does not like other variables
    if indets(op(r,i))={X} then
      s:=s union solvelib::iroots(op(r,i))
    end_if;
  end_for;
   // no multiplicity for iroots 
  max(op(s))
end_proc:

Sum::ishypergeom :=
proc(f,n)
begin
  testtype(Sum::ratio(f,n),Type::RatExpr(n))
end_proc:

/* f being hypergeometric in k, puts f(k+1)/f(k) in the form
  (k+a1)...(k+ap)/(k+b1)/.../(k+bq) x/(k+1)
  and returns [a1,...,ap], [b1,...,bq], x,
  cf Wilf page 5 */
Sum::hypergeomtype :=
proc(f,k)
  local g,P,Q,a,b,x,t,d;
begin
  g:=normal(subs(Sum::ratio(f,k),k=k+1));
  P:=Factored::convert_to(factor(numer(g)),DOM_EXPR);
  if type(P)<>"_mult" then
    P:=[P]
  end_if;
  a:=[]; b:=[]; x:=1;
  for t in P do
    if not has(t,k) then
      x:=x*t
    else
      if type(t)="_power" then
        d:=op(t,2);
        t:=op(t,1)
      else
        d:=1
      end_if;
      if degree(t,k)<>1 then
        return(FAIL)
      end_if;
      x:=x*coeff(t,k,1);
      a:=append(a,coeff(t,k,0)/coeff(t,k,1) $ d);
    end_if
  end_for;
  Q:=Factored::convert_to(factor(denom(g)),DOM_EXPR);
  if type(Q)<>"_mult" then
    Q:=[Q]
  end_if;
  for t in Q do
    if not has(t,k) then
      x:=x/t
    else
      if type(t)="_power" then
        d:=op(t,2);
        t:=op(t,1)
      else
        d:=1
      end_if;
      if degree(t,k)<>1 then
        return(FAIL)
      end_if;
      x:=x/coeff(t,k,1);
      b:=append(b,coeff(t,k,0)/coeff(t,k,1) $ d);
    end_if
  end_for;
  t:=contains(b,1);
  if t=0 then // no factor k+1 in the denominator, add one
    a:=append(a,1)
  else // delete it
    delete b[t]
  end_if;
  a,b,x
end_proc:


/****
Sum::hypergeomeval(P, Q, z, a, b, f, k)

computes
sum(f,k=a..b) when 
sum(f,k={0,1}..infinity) = pFq(P,Q,z) * f(0)

P = [a1, ..., ai], Q= [b1, ..., bj]
means f(k)/f(k+1) = (k+a1)(k+a2)..(k+ai) / ((k+b1)..(k+bj) * z/(k+1)

where z does not depend on k

****/

Sum::hypergeomeval :=
proc(P,Q,z,a,b,f,k)
  local init, kk, pos, n, l;
begin
  assert(not has(z, `#E`));
  // reduce hypergeomtype by shift if possible:
  // the numerator must contain k+1, and the denominator must contain k+b
  // for some integer b
  if (pos:= contains(P, 1)) > 0 and
    nops((l:= select(Q, testtype, DOM_INT))) >= 1 then
    userinfo(1, "reducing hypergeomtype");
    n:= - min(l) + 1;
    // substituting k by k+n will reduce the hypergeomtype
    // at first, we cancel the k+1 in the numerator and the k+1 in z/(k+1)
    delete P[pos];
    // then we have to produce a new k+1 in the denominator by shifting
    P:= map(P, _plus, n);
    Q:= map(Q, _plus, n);
    a:= a-n;
    b:= b-n;
    f:= subs(f, k = k+n);
    pos:= contains(Q, 1);
    assert(pos > 0);
    delete Q[pos];
  end_if;
  userinfo(1,"hypergeometric of type ".nops(P)."F".nops(Q));
  case [nops(P), nops(Q)]
    of [1, 0] do // 1F0
      if testtype(a,Type::NonNegInt) and b=infinity then
        userinfo(2,"1F0 with right bound at infinity");
        if P=[0] then
          // f(1) * sum(z^k/k, k=1..infinity)
          if a = 0 then
            error("Zero denominator in first summand")
          end_if;

          init:=_plus(eval(subs(f,k=kk)) $ kk=1..a-1);
          return(piecewise([abs(z) <= 1 and not z=1,
                            -ln(1-z)*eval(subs(f,k=1))/z-init  ],
                           [z >= 1, infinity * subs(f, k=1)],
                           [abs(z) > 1 and not z>=1, undefined]))
        else
          if testtype(P[1], Type::NegInt) then
            // e.g., sum((-1)^k/k/(k-1), k = 2..infinity)
            // not really satisfying, but otherwise the call
            // above provokes an error
            return(FAIL)
          end_if;
          init:=_plus(eval(subs(f,k=kk)) $ kk=0..a-1);
          return(((1-z)^(-P[1])-init)*eval(subs(f,k=0)))
        end_if
      end_if;
      break
    of [0, 0] do // 0F0
      if testtype(a, Type::NegInt) then
        error("One summand has zero denominator")
      end_if;
      if testtype(a,Type::NonNegInt) and b = infinity then
        init:=eval(_plus(eval(subs(f,k=kk)) $ kk = 0..a-1));
        return(exp(z)*eval(subs(f,k=0))-init)
      end_if
    of [2, 1] do // 2F1  
      return((if P=[1/2,1] and Q=[3/2] and a=0 and b=infinity then
                arctan(I*sqrt(z))/I/sqrt(z) * eval(subs(f, k=0))
              elif P=[0,0] and Q=[1] and testtype(a,Type::PosInt) and
                b=infinity then
                init:=_plus(eval(subs(f,k=kk)) $ kk=1..a-1);
                polylog(2,z)*eval(subs(f/z,k=1))-init
              else
                FAIL
              end_if));
      // NOT REACHED
      assert(FALSE)
     of [3, 2] do // 3F2 
      if P=[1,1,1] and Q=[2,3/2] and z=1/4 and b=infinity and
        testtype(a,Type::NonNegInt) then // sum(n!^2/(2*n+2)!,n=a..infinity)
        _plus(f $ k=0..a-1);
        return((PI^2/9-2*eval(%))*eval(subs(f,k=0)))
      end_if;
      if P = [1/2, 1/2, 1] and Q = [3/2, 3/2] and z=-1 and b=infinity and
        testtype(a,Type::NonNegInt) then
        init:= _plus(eval(subs(f,k=kk)) $ kk=0..a-1);
        return(CATALAN - init)
      end_if;
  end_case;
  FAIL
end_proc:

// computes f/subs(f,n=n-1) and returns a normalized result
// see also Koepf, Hypergeometric summation, page 15

Sum::ratio :=
proc(f, n) 
  local p,a,b,i;
begin
  // not really correct, but best in practice:
  if iszero(f) then
    return(f)
  end_if;
  case type(f)
    of "_mult" do
      return(normal(map(f,Sum::ratio,n)))
    of "_power" do
      // special case
      if op(f, 1) = -1 then
        f:= Simplify::minus1Power(f)
      end_if;
      if not has(op(f,2),n) then
        return(Sum::ratio(op(f,1),n)^op(f,2))
      elif not has(op(f,1),n) then
        return(op(f,1)^(op(f,2)-subs(op(f,2),n=n-1)))
      else
        break
      end_if
   of "fact" do
      if not has((p:=Sum::ratio(rewrite(f,gamma),n)),gamma) then
        return(p)
      end_if;
      break
   of "gamma" do
      if not has(f,n) then
        return(1)
      elif Sum::is_linear((p:=op(f)),n) then
        if type((a:=coeff(p,n,1)))=DOM_INT then
          b:=coeff(p,n,0);
          if a>0 then
            return(_mult(a*n+b-i$i=1..a))
          elif a<0 then
            return(1/eval(_mult(a*n+b+i$i=0..-a-1))) // PR-6
            // a cannot be 0 otherwise p would be independent of n 
          end_if
         end_if
      end_if;
   end_case;
   // other cases 
 
  if type(f) = "binomial" then
    return(Sum::ratio(expand(f),n))
  end_if;
  normal(f/expand(subs(f,n=n-1)))
end_proc:

// from GKP pages 224-226 
Sum::gosper :=
proc(f,_n_) // returns g such that f = subs(g,n=n+1)-g 
  local rat,qn,rn,res;
begin
  userinfo(1,"enter Gosper");
  if not has(f,_n_) then
    f*_n_
  else
      /* write f(k+1)/f(k) = p(k+1)/p(k) * q(k)/r(k+1) (5.117)
        where roots of q and r do not differ by an integer */
    rat:=Sum::ratio(f,_n_);
      // check for geometric summations 
    if not has(rat,_n_) then
      return(eval(subs(f,_n_=0))*
             piecewise([rat <> 1, rat^_n_/(rat-1)],
                       [rat = 1, _n_])
             )
    end_if;
    if has(rat,{binomial,fact}) then
      rat:=normal(expand(rat))
    end_if;
    userinfo(2,"f(".expr2text(_n_)." / f( ".expr2text(_n_-1)." = ".
             expr2text(rat));
    qn:=numer(rat);
    rn:=denom(rat);
    if not (testtype(qn,Type::PolyExpr(_n_,Type::AnyType))
            and testtype(rn,Type::PolyExpr(_n_,Type::AnyType))) then
      return(hold(sum)(f,_n_))
    end_if;
    userinfo(1,"input is hypergeometric");
    res:=Sum::gosper2(1,qn,rn,_n_,{},f);
    if res=FAIL then
      userinfo(1,"Gosper's algorithm fails");
      hold(sum)(f,_n_)
    else
      normal(res*f)
    end_if
  end_if
end_proc:

/* returns R=g/f such that f = subs(g,n=n+1)-g for f(n)/f(n-1)=p(n)/p(n-1)*qn/rn
  with pn,qn,rn polynomials,
  inds is a set of auxiliary unknowns (used only in Zeilberger).
  As f[n] = g[n+1]-g[n]. we must have f[n]=R[n+1]*f[n+1]-R[n]*f[n],
  i.e. 1 = R[n]*p(n)/p(n-1)*qn/rn-R[n-1]
*/
Sum::gosper2 :=
proc(pn,qn,rn,n,inds,f)
  local X,rnj,res,i,p_zeros,z,j,gn,gnj,dp,R,dR,Q,dQ,d,k0,fn,eq,F,ii,m,unk,ff,
  cond;
begin
  userinfo(3,"pn=",pn,"qn=",qn,"rn=",rn);

  X:= genident("X");
  rnj:=subs(rn/content(rn,[n]),n=n+X); // do not expand !
  res:=polylib::resultant(qn,rnj,n);
  res:=factor(res);
  p_zeros:=[];
  for i from 1 to nops(res) div 2 do
    ff:=op(res,2*i);
    if degree(ff,X)=1 then
      z:=-coeff(ff,X,0)/coeff(ff,X,1);
      if testtype(z,Type::PosInt) then p_zeros:=append(p_zeros,z) end_if
    end_if
  end_for;
  for j from nops(p_zeros) downto 1 do
    gn:=gcd(qn,expand(eval(subs(rn,n=n+p_zeros[j]))));
    qn:=divide(qn,gn,Exact);
    gnj:=gn;
    for i from 1 to p_zeros[j] do
      pn:=expand(pn*gnj);
      gnj:=expand(subs(gn,n=n-i));
    end_for;
    rn:=normal(rn/gnj); // divide(rn,gnj,Exact) ?
  end_for;
      // make the polynoms monic (cf Prop. 3.2 of [LiPaSt93]) 
  qn:=qn/lcoeff(rn,[n]);
  rn:=rn/lcoeff(rn,[n]);
  qn:=expand(subs(qn,n=n+1));
  dp:=degree(pn,[n]);
  R:=expand(qn+rn); dR:=degree(R,[n]);
  Q:=expand(qn-rn); dQ:=degree(Q,[n]);
  if dR<=dQ then // case 1
    userinfo(2,"case 1 of Gosper's algorithm");
    if dQ=0 then // polynomial case
      d:=dp+1
    else
      d:=dp-dQ
    end_if
  else // case 2 : deg(qn)=deg(rn) and lcoeff(qn)=lcoeff(rn)
    m:=dR; // to have the same notation as [LiPaSt93]
    k0:=coeff(rn-qn,n,m-1); // (5.2) of [LiPaSt93]
    if type(k0)<>DOM_INT or m=1 // cf [LiPaSt93] page 255
      or testtype(f,Type::RatExpr(n,Type::AnyType)) then // case 2a
      userinfo(2,"case 2a of Gosper's algorithm");
      d:=dp-m+1
    else // case 2b
      userinfo(2,"case 2b of Gosper's algorithm");
      d:=max(k0,dp-m+1)
    end_if
  end_if;
  userinfo(2,"d=",d);
  if d>=0 then
    F:= genident("F");
    fn:=_plus(F[ii]*n^ii$ii=0..d);
         // eq. (GE) of [LiPaSt93] p. 248 
         // where qn is q(n+1) 
         // put minus sign before fn because of PR302-303-310 
    eq:=expand(pn+qn*subs(-fn,n=n+1)+rn*fn);
    userinfo(2,"number of equations is",degree(eq,[n])+1);
    eq:=poly(eq, [n]);
         // necessary to avoid problems with e.g. eq=param^(-1)*n 
    eq:={coeff(eq,i)$ i=0..degree(eq)};
    unk:=inds union {F[ii]$ii=0..d};
    if nops(eq)<nops(unk) then
      eq:=eq union {F[0]=0}
    end_if;
    userinfo(3,"system to solve is",eq);
    eq:=linsolve(eq,unk, hold(ShowAssumptions));
    cond:= op(op(eq, 2)) and op(op(eq, 3));
    if hastype(cond, "_equal") then
      return(FAIL)
    end_if;
    eq:= op(eq, 1);
    userinfo(1,"solution is",eq);
    if eq = FAIL then
             // system has no solution, or solution is only valid if the
             // parameters satisfy certain constraints 
      return( FAIL )
    else
             /* do the substitution into pn too because it can
               contain some indeterminates C[l] in Zeilberger's alg. */
      fn:=Sum::normal(eval(subs(fn/pn,eq))*rn);
      if has(fn,F) then
        return( FAIL )
      elif inds={} then
        return( piecewise([cond, fn]))   // for Sum::gosper
      else
        return( piecewise([cond, fn]), eq)             // for Sum::zeilberger
      end_if
    end_if
  else
    return( FAIL )
  end_if
end_proc:

// required because normal expands 
Sum::normal :=
proc(f)
begin
  Sum::factor(numer(f))/Sum::factor(denom(f))
end_proc:

Sum::factor :=
proc(f)
  local i,l;
begin
  l:=factor(f);
  _mult(l[1],l[2*i]^l[2*i+1]$i=1..nops(l) div 2)
end_proc:

// returns a..b such that f=0 for j<a or j>b 
Sum::support :=
proc(f,j)
  local a,b,x,k,c,d;
begin
  a:=-infinity; b:=infinity;
  for x in misc::subExpressions(f,"binomial") do
    for k in [op(x,2),op(x,1)-op(x,2)] do
         // k must be >= 0 
      if testtype(k,Type::PolyExpr(j)) then
        if degree(k,[j])=1 then
          d:=coeff(k,[j],1); c:=-coeff(k,[j],0)/d;
          if is(coeff(k,[j],1)>0) then
            if a=-infinity then
              a:=c
            else
              a:=max(a,c)
            end_if
          elif is(coeff(k,[j],1)<0) then
            if b=infinity then
              b:=c
            else
              b:=min(b,c)
            end_if
          end_if
        end_if
      end_if
    end_for
  end_for;
  a..b
end_proc:

// CHANGED (jngerhar, 28.9.00)
// Try to compute limit(f, x = a) without using limit if possible, since
// limit is often much much too slow when f contains parameters.
Sum::myeval :=
proc(f, x, a)
  local ff, fa, ord, t, j;
begin
  if type(f) = piecewise then
    f:= piecewise::mapConditions(f, subs, x=a);
    if type(f) = piecewise then
      return(piecewise::extmap(f, Sum::myeval, x, a))
    end_if
  end_if;
  ff := expr(factor(f));
  // first try subs
  if domtype(a) <> stdlib::Infinity
    and traperror((fa := eval(subs(ff, x = a)))) = 0 then
    // rewrite all binomials
      fa:= misc::maprec(fa, {"binomial"} =
                        proc(b)
                          local n, k;
                        begin
                          n:= op(b, 1);
                          k:= op(b, 2);
                          if testtype(k, Type::NonNegInt) then
                            _mult((n-j+1)/j $j=1..k)
                          elif testtype(n-k, Type::NegInt) then
                            0
                          else
                            b
                          end_if
                        end_proc
                        );
      return(eval(expr(factor(fa))))
  else
    // next try series with orders 1,2,3,4
    for ord from 1 to 4 do
      if traperror((fa := series(ff, x = a, ord, Real, NoWarning))) = 0 then
        if domtype(fa) = Series::Puiseux or domtype(fa) = Series::gseries then
          // expand is necessary to simplify products of gammas. This
          // was formerly done by series.
          // combine is necessary to cancel, e.g., 2^x/2^x
          if (t := lmonomial(fa)) <> FAIL then
            t := eval(expr(factor(expand(combine(rewrite(t, gamma))))));
            if not has(t, x) then
              return(t);
            else
              break;
            end_if
          end_if
        elif type(fa) = "series" then
          break;
        end_if;
      end_if;
    end_for;
  end_if;
  // Finally, use limit as a last resort
  save x;
  unassume(x);
  t:= limit(ff, x = a);
  if type(a) <> stdlib::Infinity and has(t, infinity) then
    error("Singularity")
  else
    t
  end_if
end_proc:

// u is the name of the output recurrence 
Sum::zeilberger:=
proc(f,_k,n,u,a,b, options) // sum(f,k=a..b) 
local ratn,d,l,m,C,p,ratk,res,rnk,cert,va,vb,A,/*fcert,*/corr,i,sup,aa,bb,
  printWarnings;
begin
   if has(a,n) then return(FAIL) end_if; // not yet treated 

   ratn:=Sum::ratio(f,n); // f(n)/f(n-1) 
   if not has(ratn,{n,_k}) then 
	return(FAIL) 
   elif not testtype(ratn,Type::RatExpr(n)) then 
	return(FAIL)
   elif not testtype(ratn,Type::RatExpr(_k)) then 
	return(FAIL)
   end_if;

   d:=1;
   C:= genident("C");
   repeat
      rnk:=1; m:=1;
      for l from 1 to d do
         m:=m*subs(ratn,n=n+l); // now m=f(n+l)/f(n) 
         rnk:=rnk+C[l]*m;
      end_for;
      // now apply Gosper to rnk*f with respect to k 
      /* i.e. search if F=f(n)+C[1]*f(n+1)+...+C[d]*f(n+d)
        satisfies F(k)=G(k+1)-G(k) */
      userinfo(3,"rnk=",rnk);
      l:=C[l]$ l=1..d;
      p:=numer(rnk);
      ratk:=Sum::ratio(f/denom(rnk),_k);
      res:=Sum::gosper2(p,numer(ratk),denom(ratk),_k,{l},f);
      if res<>FAIL then 
         cert:=eval(subs(res[1]*rnk,res[2])); // g = cert*f 
         userinfo(1,"certificate=",cert);
         /* one should have subs(G,k=k+1)-G-F=0 where
           F=f(n)+C[1]*f(n+1)+...=rnk*f and G=(cert/rnk)*F=cert*f,
           i.e. subs(cert*f,k=k+1)/f-cert-rnk=0 */
         /* thus sum(f,k=a..b)+...+sum(subs(f,n=n+d),k=a..b) 
		= subs(f*cert,k=b+1) - subs(f*cert,k=a) */
         /* only deal with the cases where a is independent from n
           and b is independent from n or b=n */
//         fcert:=Sum::normal(f*cert);
//         /* The following is a VERY CRUDE hack to make the limit
//            computations below work correctly. Remove this as
//            soon as the corresponding limit bug has been fixed!
//         */
//         fcert := numer(fcert)/expand(denom(fcert));
// CHANGED (jngerhar, 28.9.00): the above lines are no longer necessary,
// see the changes below. fcert is no longer used anywhere.
         sup:=Sum::support(f,_k);
         // if a<=op(sup,1) we can extend the lower bound to -infinity 
         aa:=simplify(min(a,op(sup,1)));
//         if traperror((va:=eval(subs(expand(subs(f*cert,_k=a+_k)),_k=0))))<>0 
//         then va:=hold(limit)(fcert,_k=a)
//         end_if;
// CHANGED (jngerhar, 28.9.00): replaced the above three lines by Sum::myeval
         va := Sum::myeval(f*cert, _k, a);
         // if op(sup,2)<=b we can extend the upper bound to infinity 
         bb:=simplify(max(b,subs(op(sup,2),n=n+d)));
//         if traperror((vb:=eval(subs(expand(subs(f*cert,_k=b+1+_k)),_k=0))))<>0
//         then vb:=hold(limit)(fcert,_k=b+1)
//         end_if;
// CHANGED (jngerhar, 28.9.00): replaced the above three lines by Sum::myeval
         vb := Sum::myeval(f*cert, _k, b + 1);
         A:=_plus(u(n),C[l]*u(n+l)$ l=1..d);
         printWarnings := stdlib::limit::printWarnings(FALSE):
         A:=A+eval(va-vb);
         stdlib::limit::printWarnings(printWarnings):
         if A=undefined then return(FAIL)
         // correcting terms if b depends on n 
         elif bb<>b and not has(b-n,n) then
           if traperror
             ((
              corr:=_plus((C[l]*eval(subs(f,n=n+l,_k=b+i))$ i=1..l) $ l=1..d)
              )) <> 0 then
             return(FAIL)
           end_if;
	       A:=A-corr
         else
           return(FAIL)
         end_if;
         A:=expand(eval(subs(A,res[2])));
         /* to check: subs(rnk,res[2])-subs(cert*Sum::ratio(f,k),k=k+1)+cert
	   should be zero */
         userinfo(1,"recurrence=",A);
         res:=numer(A);
         res:=-coeff(res,u(n+d),0)/coeff(res,u(n+d),1);
         res:=u(n+d)=collect(res,[u(n+l)$ l=0..d-1],x->Factored::convert_to(factor(x),DOM_EXPR));
         if iszero(eval(subs(denom(cert),n=0))) then d:=d+1 end_if;
         return([res,d])
      else
         userinfo(1,"Gosper's algorithm fails for order ",d);
      end_if;
      d:= d+1; // try a higher order 
   until d > options[hold(MaxOrder)] end_repeat;
   FAIL
end_proc:



