alias( laplace  = transform::laplace ):

// A couple of routines computing laplace transforms
// written by Stephan Huckemann Sept - Nov 2002


laplace::compute_x_n_exp_c_over_x := proc(_a,_n,_c,_fu, var1, var2)
		// computes laplace transform of _fu = (_a*var1)^(_n)*exp(_c/var1)
		// transform exists 
		// for all _n if Re(_c) < 0 or
		// for _n with Re(_n) > -1 if only Re(_c) <= 0 and _c <> 0
		// note that in Oberhettinger's table of laplace transforms
		// formula 5.29 holds only for Re(a) > 0 and
		// formula 5.34 holds also for Re(a) >= 0, a <> 0 and Re(nu) > 0
local bol1, bol2;
begin
		   bol1:=FALSE;
		   bol2:=FALSE;		   
		   if testtype(_c, Type::Numeric) then
		       if Re(_c) < 0 then 
				bol1:=TRUE; 
				bol2:=TRUE; 
		       elif Re(_c) = 0 then    // _c = 0 should not happen
			  if testtype(_n, Type::Numeric) then
			     if Re(_n) > -1 then
				bol1:=TRUE;
				bol2:=TRUE;
			     end_if;
			  else // n is not numeric
				bol1:=TRUE;
				bol2:=is(Re(_n)>-1);
			  end_if;
		       end_if;
                   else // _c is not numeric
		       if testtype(_n, Type::Numeric) then
			     if Re(_n) > -1 then
				bol1:=is(Re(_c)<=0);
				bol2:=TRUE;
			     else // Re(_n) =< -1 
				bol1:=is(Re(_c)<0);
				bol2:=TRUE;
			     end_if;
			else   //_n is not numeric
			     if is(Re(_n) > -1) = TRUE then
				bol1:=is(Re(_c)<=0);
				bol2:=TRUE;
			     else // _n is anything
				bol1:=is(Re(_c)<0);
				bol2:=TRUE;
			     end_if;
			end_if;
                   end_if;
		   if bol1 = TRUE and bol2 = TRUE then
	            userinfo(10, "formula from frankp");
        	    return( _a^(_n)* 2*besselK(_n+1,2*sqrt(-_c*var2))*(-_c/var2)^(_n/2+1/2)  );
		   elif bol1 = TRUE and bol2=UNKNOWN then
		    warning("cannot determine the sign of ".expr2text(1 + Re(_n)));
		   elif bol2 = TRUE and bol1=UNKNOWN then
		    warning("cannot determine the sign  of ".expr2text(Re(_c)));
		   elif bol1=UNKNOWN and bol2=UNKNOWN then
		    warning("cannot determine the signs of "
				.expr2text(1 + Re(_n))." and ".expr2text(Re(_c)));
		   end_if;
		   return(new(laplace, _fu, var1, var2));  // cannot be computed
end_proc:


laplace::compute_x_n_sin_x := proc(_g,_d,_k,_c,_fu, var1, var2)
		// computes laplace transform of _fu = (_d*var1)^(_k)*sin(_c*var1)
		// transform exists 
		// for all _k with Re(_k) > -2 
		// there are, however, roots involved leading to various
		// branches in the _c times _var2 plane
local branch, dummy, trafo;
begin
	      case _k
	      of -1 do
		return(arctan(_c/var2)/_d);
	      of -1/2 do 	// pick the correct branch of the root
		branch := laplace::pick_branch(_c, var1);
			// branch is only zero when evaluation of 
			// assumptions on symbolics might lead further
			// then do a simplify later
		if not iszero(branch) and testtype(branch, Type::Numeric) then
			return( branch * (sqrt(var2^2+_c^2)-var2)^(1/2)
				  /sqrt(var2^2+_c^2)*_d^_k*sqrt(PI/2));
		end_if;
		break;
	      of -3/2 do	// pick the correct branch of the root
		branch := laplace::pick_branch(_c, var1);
			// branch is only zero when evaluation of 
			// assumptions on symbolics might lead further
			// then do a simplify later
		if not iszero(branch) and testtype(branch, Type::Numeric) then
			return(branch* (sqrt(var2^2+_c^2)-var2)^(1/2)*_d^_k*sqrt(2*PI));
		end_if;
		break;
	      otherwise
		if (testtype(_k, Type::Numeric) and Re(_k) > -2)
			// case _k = integer > 0 is already covered somewhere else
		    or // symbolic and assumption
		    (not testtype(_k, Type::Numeric) and (is(Re(_k) > -2) = TRUE)) then	
			      return( gamma(_k+1)*(var2^2+_c^2)^(-(_k+1)/2)*
					sin((_k+1)*arctan(_c/var2))*_d^_k)
		end_if;
	      end_case;
		dummy := genident(); // if var2 is numeric 
//            trafo:=(transform::laplace)(_g, var1, var2); 
            trafo:=(transform::laplace)(_g, var1, dummy); 
            userinfo(10, "sin-modulation");
//             return( 1/(2*I)*(subsex(trafo, var2=var2-I*_c)-
//                             subsex(trafo, var2=var2+I*_c)));
             return( 1/(2*I)*(subsex(trafo, dummy=var2-I*_c)-
                             subsex(trafo, dummy=var2+I*_c)));

end_proc:


laplace::pick_branch := proc(_c, var2)
			// in general the following holds with branch = + or - 1:
			//
			// sqrt(I*var2 +_c) - sqrt(I*var2 - _c)
			//   = 2 * branch * sqrt(sqrt(var2^2 + _c^2) - var2)
			//
			// we compute branch for numerical _c in some cases we also
			// need to know about the location of var2, we shall only do
			// it for numericals then. 
			// Otherwise (symbolic case + assumptions), 
			// for reasons of speed
			// simplify shall do it with a similar routine
			// 

local branch, erg1, erg2;
begin
	branch := 0;
	if testtype(_c, Type::Numeric) then
	// only in the following cases we may pick a universal branch
		// -PI/4 < arg(_c) <= PI/4
		if  Re(_c) >= 0  and Im(_c) <= Re(_c) and -Im(_c) < Re(_c) then
			branch :=1;
		elif	// 3 PI/4 < arg(_c) <= 5 PI/4
		    Re(_c) <= 0  and Im(_c) < -Re(_c) and -Im(_c) <= -Re(_c) then
			branch :=-1;
		end_if;
	else // _c is symbolic, testing for assumptions on _c
	// again only in the following cases there is a univeral branch
		// -PI/4 < arg(_c) <= PI/4
		if  is(Re(_c) >= 0) = TRUE and
			( (is(Im(_c) >= 0) = TRUE and is(Im(_c) <= Re(_c)) = TRUE) or
			  (is(Im(_c) <= 0) = TRUE and is(-Im(_c) < Re(_c)) = TRUE) 
			)    then
			branch :=1;
		elif	// 3 PI/4 < arg(_c) <= 5 PI/4
		      is(Re(_c) <= 0) = TRUE and 
			( (is(Im(_c) >= 0) = TRUE and is(Im(_c) < -Re(_c)) = TRUE) or
			  (is(Im(_c) <= 0) = TRUE and is(-Im(_c) <= -Re(_c))= TRUE) 
			) then
			branch :=-1;
		end_if;
	end_if;
	if iszero(branch) then
	// _c is numeric or symbolic and we don't know whether
	// it lies within the two above specified cones in the plane
	// just compute the branch:
		erg1 := sqrt(var2 + I*_c); // var2 must be restricted such that
		erg2 := sqrt(var2 - I*_c); // both ly in the right half-plane
		branch := sign(Im(erg1 - erg2)) - (1 - abs(sign(Im(erg1 - erg2))))
							* sign(Re(erg1 - erg2));
		if branch <> 1 and branch <> -1 then
			// it's an expression, we don't want to return that
			branch := 0;	
		end_if;
	end_if;
	return(branch);

end_proc:

unalias( laplace ):
