/*-
 Jrn Maas - 02.05.2002
 Graph
 Library package for working with all kind of Graphs.

 Style Guide:
 All parameters start with uppercase (i._edge. vertices = Vertices)
 All local variables that are used in several methods with the same meaning have the same name (i._edge. leftOperator)
 All variables having the same meaning as the functions are used with an underline in front (i._edge. _vertexWeights)
-*/

// Joern: Please check all my comments (search for 'ccr')

Dom::Graph := newDomain("Dom::Graph"):
Dom::Graph::create_dom:=hold(Dom::Graph):
Dom::Graph::info := "Library 'Graph': package for handling graphs.":
Dom::Graph::interface := {
hold(addEdge), hold(addVertex), hold(bipartite), hold(breadthFirstSearch), hold(checkForVertices), hold(chromaticPolynomial), 
hold(createCircleGraph), hold(createCompleteGraph), hold(createRandomEdgeWeights), hold(createRandomEdgeCosts), 
hold(createRandomVertexWeights), hold(createRandomGraph), hold(depthFirstSearch), hold(getAdjacentEdgesEntering), 
hold(getAdjacentEdgesLeaving), hold(getBestAdjacentEdge), hold(getEdgeCost), hold(getEdgeDescriptions), 
hold(getEdges), hold(getEdgesEntering), hold(getEdgesLeaving), hold(getEdgeWeights), hold(getEdgeNumber), hold(getSubGraph), 
hold(getVertexNumber), hold(getVertexWeights), hold(getVertices), hold(graphCheck), hold(inDegree), hold(isConnected), 
hold(isDirected), hold(isEdge), hold(isVertex), hold(minimumSpanningTree), hold(minSort), hold(outDegree), 
hold(plotBipartiteGraph), hold(plotCircleGraph), hold(plotGridGraph), hold(printGraphInformation), 
hold(printVertexInformation), hold(printEdgeInformation), hold(printEdgeDescInformation), hold(printEdgeCostInformation), 
hold(printEdgeWeightInformation), hold(removeEdge), hold(removeVertex), hold(residualGraph), hold(revertGraph), hold(selectEdge),
hold(selectVertex), hold(setEdgeCosts), hold(setEdgeDescriptions), hold(setEdgeWeights), hold(setVertexWeights),
hold(shortestPathSingleSource), hold(shortestPathAllPairs), hold(stronglyConnectedComponents), hold(topSort),
hold(contract),
hold(convertSSQ), hold(admissibleFlow), hold(minCut), hold(minCost), hold(maxFlow), hold(residualGraph),
hold(levelDecomposition), hold(isAcyclic)
}:
  

Dom::Graph::slot :=
proc(G: Dom::Graph, slot)
begin
  if slot = "vertexWeights" then
    dom::getVertexWeights(G)
  elif slot = "edges" then
    dom::getEdges(G)
  end_if;
end_proc:

/*-
 new - generates an undirected or directed Graph.

 new(Vertices, Edges <, Directed> <, VertexWeights=[v1,..,vn]>) <, EdgeDescriptions=[d1,..,dm]> <, EdgeWeights=[t1,..,tm]> <, EdgeCosts=[c1,..,cm]>)

 Parameters:
 -----------
   mandatory:
   ----------
   Vertices         - list of one or more vertices 
   Edges            - list of lists of one or more edges
   
   optional:
   ---------
   Directed         - if set, the graph will be directed. Default is undirected
   VertexWeights    - list of maximal weights used for one or more vertices.
   EdgeDescriptions - list of descriptions for one ore more edges.
   EdgeWeights      - list of weights for one or more edges.
   EdgeCosts        - list of costs for one or more edges.   

 ===========================================================================================================
 As long as no vertex is set with an explicit weight, the call VertexWeights will return FAIL (to make some algorithms faster)
 As long as no edge is set with an explicit weight, the call edgeWeight will return FAIL (to make some algorithms faster)
 As long as no edge is set with an explicit cost, the call EdgeCosts will return FAIL (to make some algorithms faster)
 As long as no edge is set with an explicit description, the call EdgeCosts will return FAIL (to make some algorithms faster)
-*/
Dom::Graph::new := proc(Vertices : DOM_LIST, Edges  : Type::ListOf(DOM_LIST))
local _edges, _edgesEntering, _edgesLeaving, leftOperator, rightOperator, _vertexWeights,
      _edgeWeights, _edgeCosts, _edgeDescriptions, _directed, _argument, _arguments, i, j, tmpList;

  begin
    if testargs() then
      if args(0) < 2 then
        error("Wrong number of arguments! See help for more details.");
      end_if;
    end_if;

    // initialize the Lists
    _vertexWeights := FAIL; 

    _edgeWeights := FAIL;
    _edgeCosts := FAIL;
    _edgeDescriptions := FAIL;

    // Default is not directed
    _directed := FALSE;

    _arguments := args();    
    
    for i from 3 to nops(_arguments) do
      if args(i) = hold(Directed) then
        _directed := TRUE;
        delete _arguments[i];     
      end_if;
      if args(i) = hold(Undirected) then
        delete _arguments[i];     
      end_if;
    end_for;

    // Duplicate edges
    tmpList:= [op({op(Edges)})];
    if nops(Edges) <> nops(tmpList) and nops(tmpList) > 0 then
      error("Duplicate edges were given.");     
    end_if;

    // Duplicate edges if graph is undirected. 
    // Sort edges with sysorder and then check again for duplicates. (Normal check was already done !)
    if not _directed then
      tmpList := Edges;
      for i from 1 to nops(tmpList) do
        tmpList[i] := sort(tmpList[i]);
      end_for;
      tmpList:= [op({op(tmpList)})];
      if nops(Edges) <> nops(tmpList) and nops(tmpList) > 0 then
        error("Duplicate edges were given.");
      end_if;
    end_if;
   
    // Duplicate vertices
    tmpList:= [op({op(Vertices)})];
    if nops(Vertices) <> nops(tmpList) then
      error("Duplicate vertices were defined.");      
    end_if;

    tmpList := Dom::Graph::checkForVertices(Edges, Vertices);
    if not tmpList = [] then
      error("One or more edges contain vertices that are not in list: " . expr2text(tmpList) . "\n" );
    end_if;

    // first two parameters (Vertices and Edges) are mandatory, whereas all the rest is optional
    for i from 3 to nops(_arguments) do
      _argument := _arguments[i];
      if not testtype(_argument, "_equal") then
        error("Wrong type of ".expr2text(i).". argument: equation expected!");
      end_if;
      leftOperator := op(_argument,1); 
      rightOperator := op(_argument,2);

      if domtype(rightOperator) <> DOM_LIST then
        error("Wrong argument type for " . expr2text(leftOperator) . ": list expected!");
      end_if;

      case leftOperator
        of hold(VertexWeights) do
          if testargs() then
            if nops(rightOperator) <> nops(Vertices) then
              error("Number of vertex weights differs from number of vertices");
              //rightOperator := rightOperator.[0 $ j=1..(nops(Vertices)-nops(rightOperator))];
            end_if;
            if {op(map(rightOperator, x->bool(x=hold(None) or x=infinity or x=-infinity or domtype(float(x))=DOM_FLOAT)))} <> {TRUE} then
              error("Vertex weights must be real or +/-infinity");                
            end_if;
          end_if;
          _vertexWeights := table(Vertices[j] = rightOperator[j] $ j=1..nops(Vertices));
          //_vertexWeights := _vC;
          for j in _vertexWeights do
             if op(j,2) = hold(None) then
               delete _vertexWeights[op(j,1)];
             end_if;
          end_for;
          break;
        of hold(EdgeWeights) do
          if testargs() then
            if nops(rightOperator) <> nops(Edges) then
              error("Number of edge weights differs from number of edges");
              //rightOperator := rightOperator.[0 $ (nops(Edges)-nops(rightOperator))];
            end_if;
          end_if;
          _edgeWeights := table(Edges[j] = rightOperator[j] $ j=1..nops(Edges));
          // remove redundant elements
          _edgeWeights := select(_edgeWeights, x->not op(x,2)=hold(None));

          // If graph is undirected also add the reversed edges to it.
          if _directed = FALSE then        
            for j in _edgeWeights do
              _edgeWeights[revert(op(j,1))] := op(j,2);
            end_for;
          end_if;
          break;
        of hold(EdgeCosts) do
          if testargs() then
            if nops(rightOperator) <> nops(Edges) then
              error("Number of edge costs differs from number of edges");
              //rightOperator := rightOperator.[0 $ (nops(Edges)-nops(rightOperator))];
            end_if;
          end_if;
          _edgeCosts := table(Edges[j] = rightOperator[j] $ j=1..nops(Edges));
          _edgeCosts := select(_edgeCosts, x-> not op(x,2)=hold(None));
          //_edgeCosts := select(_edgeCosts, () -> not bool( rhs(args() ) = None) );
          // if graph is undirected also add reverted edges to it.
          if _directed = FALSE then        
            for j in _edgeCosts do
              _edgeCosts[revert(op(j,1))] := op(j,2);
            end_for;
          end_if;
          if nops(_edgeCosts) = 0 then
            _edgeCosts := FAIL;
          end_if;
          break;
        of hold(EdgeDescriptions) do
          if testargs() then
            if nops(rightOperator) <> nops(Edges) then
              error("Number of edge descriptions differs from number of edges");
            end_if;
          end_if;
          _edgeDescriptions := table(Edges[j] = rightOperator[j] $ j=1..nops(Edges));
          _edgeDescriptions := select(_edgeDescriptions, x-> not op(x,2)=hold(None));
          // if graph is undirected also add reverted edges to it.
          if _directed = FALSE then        
            for j in _edgeDescriptions do
              _edgeDescriptions[revert(op(j,1))] := op(j,2);
            end_for;
          end_if;
          if nops(_edgeDescriptions) = 0 then
            _edgeDescriptions := FAIL;
          end_if;
          break;        
        otherwise
          error("Illegal Argument: " . expr2text(leftOperator) . ". Please consult help for more information.");
      end_case;
    end_for;

   if _directed then
     _edges := Edges;
   else
     _edges := Edges . map(Edges, revert);
// map(...) is functionally equivalent to:
//     for _edge in Edges do
//      _edges := append(_edges, revert(_edge));
//     end_for;
   end_if;

   _edges := listlib::removeDuplicates(_edges, KeepOrder);
    
   /*-
    Generate the adjacency lists
   -*/
   _edgesEntering := table((Vertices[i]=sort( map(select(_edges, _edge->op(_edge,2)=Vertices[i]), op, 1) ) $ i=1..nops(Vertices))):
   _edgesLeaving  := table((Vertices[i]=sort( map(select(_edges, _edge->op(_edge,1)=Vertices[i]), op, 2) ) $ i=1..nops(Vertices))):
   
   // edges are of form [[a, b], [c, d], ... ] (for example _edge = ([1,2], [1,3])
   // _edges... are of form [a=[], b=[], ... ] (for example _edgesEntering = (1=[] 2=[1] 3=[])

  Vertices := sort(Vertices);
  _edges := sort(_edges, (x, y) -> (if op(x,1) = op(y,1) then	sysorder(op(x,2), op(y,2)); else sysorder(op(x,1), op(y,1)); end_if;));

   new(Dom::Graph, Vertices, _edges, _vertexWeights, _edgeDescriptions, _edgeWeights,
                  _edgeCosts, _edgesEntering, _edgesLeaving, _directed);
end_proc:

/*-
 Output of dom
-*/
Dom::Graph::print := proc()
begin 
   "Dom::Graph(...)"
end_proc:

/*
isEmpty - checks if the graph has no vertices and edges
isEmpty(G)
Parameters:
  G - Dom::Graph
*/
Dom::Graph::isEmpty := proc(G : Dom::Graph)
begin
  if Dom::Graph::getEdgeNumber(G) = 0 and Dom::Graph::getVertexNumber(G) = 0 then
  	 TRUE;
  else
    FALSE;
  end_if;
end_proc:

/*-
 getVertices - returns a list with all vertices of a Dom::Graph
 getVertices(G)
 Parameters:
   G - Dom::Graph
-*/
Dom::Graph::getVertices := proc(G : Dom::Graph)
begin
   extop(G,1);
end_proc:

/*-
 getEdges - returns a list with all edges of a Dom::Graph
 getEdges(G)
 Parameters:
   G - Dom::Graph
-*/
Dom::Graph::getEdges := proc(G : Dom::Graph)
begin
   extop(G,2);
end_proc:

/*-
 getVertexWeights - returns a list with all vertice weights of a Dom::Graph
 getVertexWeights(G)
 Parameters:
   G - Dom::Graph
-*/
Dom::Graph::getVertexWeights := proc(G : Dom::Graph) 
begin
   extop(G,3);
end_proc:

/*-
 setVertexWeights - changes the weight of one ore more edges of a Dom::Graph
 setVertexWeights(G, Vertex, VertexWeights)
 setVertexWeights(G, [v1,..,vn], [c1,..,cn])
 Parameters:
   G                - Dom::Graph
   Vertex           - a list of one or more vertices
   VertexWeights    - a list of one or more vertice weights
-*/
Dom::Graph::setVertexWeights := proc(G : Dom::Graph, Vertex : DOM_LIST, VertexWeights : Type::Union(DOM_LIST, DOM_TABLE))
local _vertices, _existingVertices, position, x, _vertexWeights;
begin
  if testargs() then
    if args(0) <> 3 then
      error("Wrong number of arguments! See help for more details.");
    end_if;
  end_if;

  _vertices := Dom::Graph::getVertices(G);
  _vertexWeights := Dom::Graph::getVertexWeights(G);
  _existingVertices := Dom::Graph::selectVertex(G, Vertex, TRUE, Show=FALSE);

  if _existingVertices = [] then
    warning("No vertex weight to be changed.");
  else
    if domtype(VertexWeights) = DOM_TABLE then
      for x in _existingVertices do 
        if VertexWeights[x] = hold(None) then
          delete _vertexWeights[x];
        else
          _vertexWeights[x] := VertexWeights[x];
        end_if;
      end_for;
    else
      for x in _existingVertices do
        position := contains(Vertex, x);
        // just overwrite the predefined weight or create a new one if not defined before.
        if VertexWeights[position] = hold(None) then
           // remove vertex from list, because it is not neccessary.
           delete _vertexWeights[op(x,1)];
        else
          _vertexWeights[x] := VertexWeights[position]; 
        end_if;
      end_for;
    end_if;
  end_if;

  subsop(G, 3=_vertexWeights);
end_proc:

/*-
 getEdgeDescriptions - returns a list with all the descriptions of the edges of a Dom::Graph
 getEdgeDescriptions(G)
 Parameters:
   G - Dom::Graph
-*/
Dom::Graph::getEdgeDescriptions := proc(G : Dom::Graph)
begin
   extop(G,4)
end_proc:

/*-
 setEdgeDescriptions - changes the Description of one or more edges of a Dom::Graph
 setEdgeDescriptions(G, Edge, EdgeDescriptions <,OnlySpecifiedEdges=TRUE|FALSE>)
 setEdgeDescriptions(G, [e1,..,en], [d1,..,dn] <,OnlySpecifiedEdges=TRUE|FALSE>)
 Parameters:
   G            - Dom::Graph
   Edge         - a list of lists of one or more edges
   EdgeWeights - a list of one or more edge descriptions
-*/
Dom::Graph::setEdgeDescriptions := proc(G : Dom::Graph, Edge : Type::ListOf(DOM_LIST), EdgeDescriptions : Type::Union(DOM_LIST, DOM_TABLE))
local _edges, _existingEdges, _edgeDescriptions, position, x, _onlySE, reverted;
begin
  if testargs() then
    if args(0) < 3 then
      error("Wrong number of arguments! See help for more details.");
    end_if;
  end_if;

  _onlySE := FALSE;
  if args(0) = 4 then
     if args(4) = hold(OnlySpecifiedEdges) then
       _onlySE := TRUE;
     end_if;
  end_if;  

  _edges := Dom::Graph::getEdges(G);
  _edgeDescriptions := Dom::Graph::getEdgeDescriptions(G);
  _existingEdges := Dom::Graph::selectEdge(G, Edge, TRUE, Show=FALSE, OnlySpecifiedEdges=_onlySE);

  if _existingEdges = [] then
    warning("No edge cost to be changed.");
  else
    if domtype(EdgeDescriptions) = DOM_TABLE then
      for x in _existingEdges do
        reverted := FALSE;
        if contains(EdgeDescriptions, x) = FALSE then // check the reverted edge if original edge does not exist!
          x := [op(x,2),op(x,1)];
          reverted := TRUE;
        end_if;
        if EdgeDescriptions[x] = hold(None) then
          // remove edge from list, because it is not necessary.
          if reverted then
            x := [op(x,2),op(x,1)];
            delete _edgeDescriptions[x];
          else  
            delete _edgeDescriptions[x];
          end_if;
        else
          if reverted then
            if contains(EdgeDescriptions, x) then
              _edgeDescriptions[[op(x,2),op(x,1)]] := EdgeDescriptions[x];            	
            end_if;
          else
            if contains(EdgeDescriptions, x) then
              _edgeDescriptions[x] := EdgeDescriptions[x];
            end_if;
          end_if;
        end_if;
      end_for;
    else
      for x in _existingEdges do
        reverted := FALSE;
        position := contains(Edge, x);

        if position = 0 then
          x := [op(x,2),op(x,1)];
 	        position := contains(Edge, x);
          reverted := TRUE;
        end_if;

        // just overwrite the predefined weight or create a new one if not defined before.
        if EdgeDescriptions[position] = hold(None) then
          // remove edge out of list, because it is not necessary.
          if reverted then
            delete _edgeDescriptions[[op(x,2),op(x,1)]];
          else  
            delete _edgeDescriptions[x];
          end_if;
        else
          if not type(EdgeDescriptions[position]) = DOM_STRING then
            if reverted then
              if position > 0 then
                _edgeDescriptions[[op(x,2),op(x,1)]] := expr2text(EdgeDescriptions[position]);
              end_if;
            else  
              if position > 0 then
                _edgeDescriptions[x] := expr2text(EdgeDescriptions[position]);
              end_if;
            end_if;
          end_if;
          if reverted then
            if position > 0 then
              _edgeDescriptions[[op(x,2),op(x,1)]] := EdgeDescriptions[position];            	
            end_if;
          else  
            if position > 0 then
              _edgeDescriptions[x] := EdgeDescriptions[position];            	
            end_if;
          end_if;
        end_if;
      end_for;
    end_if;
  end_if;

  subsop(G, 4=_edgeDescriptions);
end_proc:

/*-
 getEdgeWeights - returns a list with all edge weights of a Dom::Graph
 getEdgeWeights(G)
 Parameters:
   G - Dom::Graph
-*/
Dom::Graph::getEdgeWeights := proc(G : Dom::Graph)
begin
   extop(G,5);
end_proc:

/*-
 setEdgeWeights - changes the weight of one ore more edges of a Dom::Graph
 setEdgeWeights(G, Edge, EdgeWeights <,OnlySpecifiedEdges=TRUE|FALSE>)
 setEdgeWeights(G, [e1,..,en], [c1,..,cn] <,OnlySpecifiedEdges=TRUE|FALSE>)
 Parameters:
   G            - Dom::Graph
   Edge         - a list of lists of one or more edges
   EdgeWeights  - a list of one or more edge weights
-*/
Dom::Graph::setEdgeWeights := proc(G : Dom::Graph, Edge : Type::ListOf(DOM_LIST), EdgeWeights : Type::Union(DOM_LIST, DOM_TABLE))
local _edges, _existingEdges, _edgeWeights, position, x, _onlySE, reverted;
begin
  if testargs() then
    if args(0) < 3 then
      error("Wrong number of arguments! See help for more details.");
    end_if;
  end_if;

  _onlySE := FALSE;
  if args(0) = 4 then
     if args(4) = hold(OnlySpecifiedEdges) then
       _onlySE := TRUE;
     end_if;
  end_if;  

  _edges := Dom::Graph::getEdges(G);
  _edgeWeights := Dom::Graph::getEdgeWeights(G);
  _existingEdges := Dom::Graph::selectEdge(G, Edge, TRUE, Show=FALSE, OnlySpecifiedEdges=_onlySE);

  if _existingEdges = [] then
    warning("No edge cost to be changed.");
  else
    if domtype(EdgeWeights) = DOM_TABLE then
      for x in _existingEdges do
        reverted := FALSE;
        if contains(EdgeWeights, x) = FALSE then // check the reverted edge if original edge does not exist!
          x := [op(x,2),op(x,1)];
          reverted := TRUE;
        end_if;
        if EdgeWeights[x] = hold(None) then
          // remove edge from list, because it is not necessary.
          if reverted then
            x := [op(x,2),op(x,1)];
            delete _edgeWeights[x];
          else  
            delete _edgeWeights[x];
          end_if;
        else
          if reverted then
            if contains(EdgeWeights, x) then
              _edgeWeights[[op(x,2),op(x,1)]] := EdgeWeights[x];
            end_if;
          else  
            if contains(EdgeWeights, x) then
              _edgeWeights[x] := EdgeWeights[x];            	
            end_if;
          end_if;
        end_if;
      end_for;
    else
      for x in _existingEdges do
        reverted := FALSE;
        position := contains(Edge, x);

        if position = 0 then
          x := [op(x,2),op(x,1)];
 	        position := contains(Edge, x);
          reverted := TRUE;
        end_if;
        // just overwrite the predefined weight or create a new one if not defined before.
        if EdgeWeights[position] = hold(None) then
          // remove edge out of list, because it is not necessary.
          if reverted then
            delete _edgeWeights[[op(x,2),op(x,1)]];
          else  
            delete _edgeWeights[x];
          end_if;
        else
          if reverted then
            if position > 0 then
              _edgeWeights[[op(x,2),op(x,1)]] := EdgeWeights[position];            	
            end_if;
          else  
            if position > 0 then
              _edgeWeights[x] := EdgeWeights[position];            	
            end_if;
          end_if;
        end_if;
      end_for;
    end_if;
  end_if;

  subsop(G, 5=_edgeWeights);
end_proc:

/*-
 getEdgeCosts - returns a list with the edge costs of a Dom::Graph
 getEdgeCosts(G)
 Parameters:
   G - Dom::Graph
-*/
Dom::Graph::getEdgeCosts := proc(G : Dom::Graph)
begin
   extop(G,6)
end_proc:

/*-
 setEdgeCosts - changes the costs of one ore more edges of a Dom::Graph
 setEdgeCosts(G, Edge, EdgeCosts <,OnlySpecifiedEdges=TRUE|FALSE>)
 setEdgeCosts(G, [e1,..,en], [c1,..,cn] <,OnlySpecifiedEdges=TRUE|FALSE>)
 Parameters:
   G        - Dom::Graph
   Edge     - a list of lists of one or more edges
   EdgeCosts - a list or table of one or more edge costs
-*/
Dom::Graph::setEdgeCosts := proc(G : Dom::Graph, Edge : Type::ListOf(DOM_LIST), EdgeCosts : Type::Union(DOM_LIST, DOM_TABLE))
local _edges, _existingEdges, _edgeCosts, position, x, _onlySE, reverted;
begin
  if testargs() then
    if args(0) < 3 then
      error("Wrong number of arguments! See help for more details.");
    end_if;
  end_if;

  _onlySE := FALSE;
  if args(0) = 4 then
     if args(4) = hold(OnlySpecifiedEdges) then
       _onlySE := TRUE;
     end_if;
  end_if;  

  _edges := Dom::Graph::getEdges(G);
  _edgeCosts := Dom::Graph::getEdgeCosts(G);
  _existingEdges := Dom::Graph::selectEdge(G, Edge, TRUE, Show=FALSE, OnlySpecifiedEdges=_onlySE);

  if _existingEdges = [] then
    warning("No edge cost to be changed.");
  else
    if domtype(EdgeCosts) = DOM_TABLE then
      for x in _existingEdges do
        reverted := FALSE;
        if contains(EdgeCosts, x) = FALSE then // check the reverted edge if original edge does not exist!
          x := [op(x,2),op(x,1)];
          reverted := TRUE;
        end_if;
        if EdgeCosts[x] = hold(None) then
          // remove edge from list, because it is not necessary.
          if reverted then
            x := [op(x,2),op(x,1)];
            delete _edgeCosts[x];
          else  
            delete _edgeCosts[x];
          end_if;
        else
          if reverted then
            if contains(EdgeCosts, x) then
              _edgeCosts[[op(x,2),op(x,1)]] := EdgeCosts[x];            	
            end_if;
          else  
            if contains(EdgeCosts, x) then
              _edgeCosts[x] := EdgeCosts[x];
            end_if;
          end_if;
        end_if;
      end_for;
    else
      for x in _existingEdges do
        reverted := FALSE;
        position := contains(Edge, x);

        if position = 0 then
          x := [op(x,2),op(x,1)];
 	        position := contains(Edge, x);
          reverted := TRUE;
        end_if;
        // just overwrite the predefined weight or create a new one if not defined before.
        if EdgeCosts[position] = hold(None) then
          // remove edge out of list, because it is not necessary.
          if reverted then
            delete _edgeCosts[[op(x,2),op(x,1)]];
          else  
            delete _edgeCosts[x];
          end_if;
        else
          if reverted then
            if position > 0  then
              _edgeCosts[[op(x,2),op(x,1)]] := EdgeCosts[position];            	
            end_if;
          else  
            if position > 0 then
              _edgeCosts[x] := EdgeCosts[position];            	
            end_if;  
          end_if;
        end_if;
      end_for;
    end_if;
  end_if;

  subsop(G, 6=_edgeCosts);
end_proc:

/*-
 getEdgesEntering - returns the adjacency list for incoming edges of a Dom::Graph
 getEdgesEntering(G)
 Parameters:
   G - Dom::Graph
-*/
Dom::Graph::getEdgesEntering := proc(G : Dom::Graph)
begin
   extop(G,7)
end_proc:

/*-
 getEdgesLeaving - returns the adjacency list for outgoing edges of a Dom::Graph
 getEdgesLeaving(G)
 Parameters:
   G - Dom::Graph
-*/
Dom::Graph::getEdgesLeaving := proc(G : Dom::Graph)
begin
   extop(G,8)
end_proc:

/*-
 isDirected - returns TRUE if the Graph is directed, FALSE otherwise
 isDirected(G)
 Parameters:
   G - Dom::Graph
-*/
Dom::Graph::isDirected := proc(G : Dom::Graph)
begin
   extop(G,9)
end_proc:

/*-
 isVertex - checks whether a vertex is contained in a Dom::Graph

 isVertex(G, Vertex)
 Parameters:
   G      - Dom::Graph
   Vertex - a list of one or more vertices
-*/
Dom::Graph::isVertex := proc(G : Dom::Graph, Vertex : DOM_LIST)
local _vertices, i;
begin
  if testargs() then
    if args(0) <> 2 then
    error("Wrong number of arguments! See help for more details."):
    end_if;
  end_if;

  _vertices := Dom::Graph::getVertices(G);

  for i in Vertex do
    if not bool(contains(_vertices, i) <> 0) then
    	 return (FALSE);
    end_if;  	
  end_for;
  TRUE;
end_proc:

/*-
 isEdge - checks whether an edge is contained in a Dom::Graph
 Because of the new data structure the check is done in O(1).
 If the graph is directed just check if 
 Table[x,y] is not NIL (if there is only one edge allowed between two vertices)
 If the graph is undirected check:
 if y < x switch x and y and check if Table[x,y] is not NIL.

 isEdge(G,_edge)
 Parameters:
   G    - Dom::Graph
   Edge - a list of lists of one or more edges
   
-*/
Dom::Graph::isEdge := proc(G : Dom::Graph, Edge : Type::ListOf(DOM_LIST))
local i, _edges;
begin
   /*- Because of the data-structure (undirected graphs are handled like directed graphs) if the graph is undirected the
       edge (a, b) has to be checked and the edge (b, a), too.
   -*/
   if testargs() then
      if args(0) <> 2 then
      error("Wrong number of arguments! See help for more details."):
      end_if;
   end_if;

   _edges := Dom::Graph::getEdges(G);

   if Dom::Graph::isDirected(G) then
      // directed graph, only check the edge given.
      for i in Edge do
        if not bool(contains(_edges, i) <> 0) then
        	 return (FALSE);
        end_if;      	
      end_for;
   else
      // undirected graph, check the edge (a, b) and the edge (b, a)
     for i in Edge do
       if not bool(contains(_edges, op(Edge)) <> 0 and contains(_edges, op(revert(Edge))) <> 0) then
     	   return (FALSE);
       end_if;
     end_for;
   end_if;

   TRUE;
end_proc:

/*-
 graphCheck - checks if G has the type Dom::Graph
 Parameters:
   G - Dom::Graph
-*/
Dom::Graph::graphCheck := proc(G)
begin
  if not testtype(G, Dom::Graph) then
// ccr: Try
// context(hold(error)("Wrong argument type: Dom::Graph expected!"))
// and find out the difference.
    error("Wrong argument type: Dom::Graph expected!");
  end_if;
end_proc:

/*-
 printGraphInformation - prints some information of a Dom::Graph
 Parameters:
   G - Dom::Graph
-*/
Dom::Graph::printGraphInformation := proc(G : Dom::Graph)
save PRETTYPRINT;
begin

  PRETTYPRINT := FALSE;
  if Dom::Graph::getVertices(G) = FAIL then
    print(Unquoted, "Vertices: no vertices.");
  else
    print(Unquoted, "Vertices: ".expr2text(Dom::Graph::getVertices(G)));
  end_if;
 
  if Dom::Graph::getEdges(G) = FAIL then
     print(Unquoted, "Edges: no edges.");
  else
     print(Unquoted, "Edges: ".expr2text(Dom::Graph::getEdges(G)));
  end_if;

  if Dom::Graph::getVertexWeights(G) = FAIL then
    if nops(Dom::Graph::getVertices(G)) > 0 then
      print(Unquoted, "Vertex weights: no vertex weights.");     
    end_if; 
  else
     print(Unquoted, "Vertex weights: ".expr2text(op(Dom::Graph::getVertexWeights(G)))." (other existing vertices have no weight)");  
  end_if;

  if Dom::Graph::getEdgeDescriptions(G) = FAIL then
    if nops(Dom::Graph::getEdges(G)) > 0 then
      print(Unquoted, "Edge descriptions: no edge descriptions.");
    end_if;
  else
     print(Unquoted, "Edge descriptions: ".expr2text(op(Dom::Graph::getEdgeDescriptions(G))));  
  end_if;

  if Dom::Graph::getEdgeWeights(G) = FAIL then
    if nops(Dom::Graph::getEdges(G)) > 0 then
       print(Unquoted, "Edge weights: no edge weights.");
    end_if;
  else
     print(Unquoted, "Edge weights: ".expr2text(op(Dom::Graph::getEdgeWeights(G)))." (other existing edges have no weight)"); 
  end_if;

  if Dom::Graph::getEdgeCosts(G) = FAIL then
    if nops(Dom::Graph::getEdges(G)) > 0 then
      print(Unquoted, "Edge costs: no edge costs.");
    end_if;
  else
    print(Unquoted, "Edge costs: " . expr2text(op(Dom::Graph::getEdgeCosts(G))) . " (other existing edges have costs zero)"); 
  end_if;

  if Dom::Graph::getEdgesLeaving(G) = FAIL then
    if nops(Dom::Graph::getEdges(G)) > 0 then
      print(Unquoted, "Adjacency list (out): no edges.");
    end_if;
  else
     print(Unquoted, "Adjacency list (out): ".expr2text(op(Dom::Graph::getEdgesLeaving(G)))); 
  end_if;

  if Dom::Graph::getEdgesEntering(G) = FAIL then
    if nops(Dom::Graph::getEdges(G)) > 0 then
      print(Unquoted, "Adjacency list (in): no edges.");
    end_if;
  else
     print(Unquoted, "Adjacency list (in): ".expr2text(op(Dom::Graph::getEdgesEntering(G))));
  end_if;

  if Dom::Graph::isDirected(G) then
    print(Unquoted, "Graph is directed.");
  else
    print(Unquoted, "Graph is undirected.");
  end_if;

  null();

end_proc:

/*-
 printVertexInformation - prints detailed information about the vertices of a Dom::Graph
 Parameters:
   G - Dom::Graph
-*/
Dom::Graph::printVertexInformation := proc(G : Dom::Graph)
save PRETTYPRINT;
local v, _vertices, _vertexWeights;
begin

  PRETTYPRINT := FALSE;
  if Dom::Graph::getVertices(G) = FAIL then
    print(Unquoted, "No vertices.");
  else
    print(Unquoted, "Vertices existing in the graph:");
    print(Unquoted, "-------------------------------");
    _vertices := Dom::Graph::getVertices(G);
    _vertexWeights := Dom::Graph::getVertexWeights(G);
    for v in _vertices do
      print(NoNL, "Vertex " . expr2text(v) . " has weight ");
      if (_vertexWeights = FAIL) or (contains(_vertexWeights,v) = FALSE)then
        print(Unquoted, "None");
      else
        print(Unquoted, expr2text(_vertexWeights[v]));
      end_if;
    end_for;
  end_if;
  null();
end_proc:

/*-
 printEdgeInformation - prints detailed information about the edges of a Dom::Graph
 Parameters:
   G - Dom::Graph
-*/
Dom::Graph::printEdgeInformation := proc(G)
save PRETTYPRINT;
begin
  if testargs() then
    Dom::Graph::graphCheck(G);
  end_if;

  PRETTYPRINT := FALSE;
  if Dom::Graph::getEdges(G) = FAIL then
    print(Unquoted, "No edges.");
  else
    print(Unquoted, "Edges existing in the graph:");
    print(Unquoted, "----------------------------");
    print(Unquoted, expr2text(op(Dom::Graph::getEdges(G))));
  end_if;
  null();
end_proc:

/*-
 printEdgeDescriptionInformation - prints detailed information about the edge descriptions of a Dom::Graph
 Parameters:
   G - Dom::Graph
-*/
Dom::Graph::printEdgeDescInformation := proc(G)
save PRETTYPRINT;
local _edge, _edges, _edgeDescriptions;
begin
  if testargs() then
    Dom::Graph::graphCheck(G);
  end_if;

  PRETTYPRINT := FALSE;
  _edges := Dom::Graph::getEdgeDescriptions(G);
  if _edges = FAIL then
    print(Unquoted, "No edge descriptions defined.");
  else
    print(Unquoted, "Edge descriptions existing in the graph:");
    print(Unquoted, "----------------------------------------");
    _edgeDescriptions := Dom::Graph::getEdgeDescriptions(G);
    for _edge in _edges do
      print(Unquoted, "Edge " . expr2text(_edge) . " = " . expr2text(_edgeDescriptions[_edge]));  
    end_for;
  end_if;
 null();
end_proc:

/*-
 printEdgeCostInformation - prints detailed information about the edge costs of a Dom::Graph
 Parameters:
   G - Dom::Graph
-*/
Dom::Graph::printEdgeCostInformation := proc(G)
save PRETTYPRINT;
local _edge, _edges, _edgeCosts;
begin
  if testargs() then
    Dom::Graph::graphCheck(G);
  end_if;

  PRETTYPRINT := FALSE;
  _edges := Dom::Graph::getEdgeCosts(G);
  if _edges = FAIL then
    print(Unquoted, "No edge costs defined.");
  else
    print(Unquoted, "Edge costs existing in the graph:");
    print(Unquoted, "----------------------------------------");
    _edgeCosts := Dom::Graph::getEdgeCosts(G);
    for _edge in _edges do
      print(NoNL, "Edge " . expr2text(_edge) . " has cost ");
      if (_edgeCosts = FAIL) or (contains(_edgeCosts, _edge) = FALSE)then
        print(Unquoted, "0");
      else
        print(Unquoted, expr2text(_edgeCosts[_edge]));
      end_if;
      // print(Unquoted, "Edge " . expr2text(_edge) . " has cost " . expr2text(_edgeCosts[_edge])); 
    end_for;
  end_if;
  null();
end_proc:

/*-
 printEdgeWeightInformation - prints detailed information about the edge weights of a Dom::Graph
 Parameters:
   G - Dom::Graph
-*/
Dom::Graph::printEdgeWeightInformation := proc(G)
save PRETTYPRINT;
local _edge, _edges, _edgeWeights;
begin
  if testargs() then
    Dom::Graph::graphCheck(G);
  end_if;

  PRETTYPRINT := FALSE;
  if Dom::Graph::getEdgeWeights(G) = FAIL then
    print(Unquoted, "No edge weights defined.");
  else
    print(Unquoted, "Edge weights existing in the graph:");
    print(Unquoted, "----------------------------------------");
    _edges := Dom::Graph::getEdges(G);
    _edgeWeights := Dom::Graph::getEdgeWeights(G);
    for _edge in _edges do
      print(NoNL, "Edge " . expr2text(_edge) . " has weight ");
      if (_edgeWeights = FAIL) or (contains(_edgeWeights, _edge) = FALSE)then
        print(Unquoted, "0");
      else
        print(Unquoted, expr2text(_edgeWeights[_edge]));
      end_if;
      //print(Unquoted, "Edge " . expr2text(_edge) . " has weight " . expr2text(_edgeWeight[_edge])); 
    end_for;
  end_if;
  null();
end_proc:

/*-
 selectEdge - checks if one ore more edges are already defined in a graph and returns a new table with edges
             that are not in the graph yet or that already are defined depending on "existing".
             If the optional parameter show is set to false, already defined edges will not be printed.
 selectEdge(G, Edge, existing)
 selectEdge(G, Edge, existing <,Show = TRUE/FALSE> <,OnlySpecifiedEdges=TRUE|FALSE>)
 selectEdge(G, [e1,..,en], existing)
 selectEdge(G, [e1,..,en], existing <,Show = TRUE/FALSE> <,OnlySpecifiedEdges=TRUE|FALSE>)
 Parameters:
   G    - Dom::Graph
   Edge - a list of one or more edges
   existing - TRUE returns all edges that already exist in the graph.
              FALSE returns all edges that are not in the graph.
   Show     - TRUE (standard) show if any vertices were found not matching the criteria.
              FALSE supress the output.

-*/
Dom::Graph::selectEdge := proc(G, Edge : Type::ListOf(DOM_LIST), existing)
local _currentEdges, _existingEdges, _newEdges, dummy, leftOperator, rightOperator,_show,_onlySE, i; 
begin
  if testargs() then
    if args(0) < 3 then
      error("Wrong number of arguments! See help for more details.");
    end_if;
  end_if;

  _show := TRUE; // predefinition
  _onlySE := FALSE;

  if args(0) > 3 then
    for i from 4 to args(0) do
      if testargs() then
        if not testtype(args(i), "_equal") then
          error("Wrong type of ".expr2text(args(i))." argument: equation expected!");      
        end_if;     
      end_if;
      leftOperator := op(args(i),1); 
      rightOperator := op(args(i),2);
      case leftOperator
        of hold(Show) do
          if testargs() then
            if not testtype(rightOperator, DOM_BOOL) then
              error("Wrong type of ".expr2text(args(i))." argument: TRUE or FALSE expected!");
            end_if;
          end_if;         
          _show := rightOperator;
          break;
        of hold(OnlySpecifiedEdges) do
          if testargs() then
            if not testtype(rightOperator, DOM_BOOL) then
              error("Wrong type of ".expr2text(args(i))." argument: TRUE or FALSE expected!");
            end_if;
          end_if;
          _onlySE := rightOperator;
          break;
        otherwise
          error("Wrong argument: ".expr2text(args(i))." !");      
      end_case;
    end_for;
  end_if;

  _currentEdges := Dom::Graph::getEdges(G);

  if existing = TRUE then
    if Dom::Graph::isDirected(G) or _onlySE then
      [_newEdges, _existingEdges, dummy] := split(_currentEdges, _edge -> contains(Edge, _edge) > 0);
    else
      [_newEdges, _existingEdges, dummy] := split(_currentEdges, (_edge -> contains(Edge, _edge) > 0)
                                                  or (_edge -> contains(Edge, revert(_edge)) > 0));
    end_if;
  else
    if Dom::Graph::isDirected(G) or _onlySE then
      [_existingEdges, _newEdges, dummy] := split(Edge, _edge -> contains(_currentEdges, _edge) > 0);
    else
      [_existingEdges, _newEdges, dummy] := split(Edge, (_edge -> contains(_currentEdges, _edge) > 0)
                                                  or (_edge -> contains(_currentEdges, revert(_edge)) > 0));
    end_if;
  end_if;

  if _show = TRUE then
    if _existingEdges <> [] then
      warning("The following edges were not used for operation: " . expr2text(_existingEdges) . "\n");
    end_if;
  end_if;

  return(_newEdges);
end_proc:

/*- 
 addEdge - adds one or several edges to a Dom::Graph
 addEdge(G, Edge)
 addEdge(G, Edge <,EdgeWeights=t> <,EdgeCosts=c> <,EdgeDescriptions=d> <,OnlySpecifiedEdges=TRUE|FALSE>)
 addEdge(G, [e1,..,en])
 addEdge(G, [e1,..,en] <,EdgeWeights=[t1,..,tn]> <,EdgeCosts=[c1,..,cn]> <,EdgeDescriptions=d>)
 Parameters:
   G               - Dom::Graph
   Edge            - a list of lists of one or more edges
   EdgeWeights    - a list of one or more edge weights
   EdgeCosts        - a list of one or more edge costs
   EdgeDescriptions - a list of one or more descriptions. 
                     If an empty string ("") is used, no table entry will be created for this edge.
                     Also any other object than DOM_STRING will be converted into one using expr2text.
-*/
Dom::Graph::addEdge := proc(G : Dom::Graph, Edge : Type::ListOf(DOM_LIST))
local _correctEdges, _edgeCosts, _edgeWeights, _edgesEntering, _edgesLeaving, _edgeDescriptions, _onlySE,
      arguments, leftOperator, rightOperator, i, j, _edge, _vertices, _edges, _directed, x, position, tmpList;
begin
   if testargs() then
     if args(0) < 2 then
       error("Wrong number of arguments! See help for more details.")
      end_if;
   end_if;

   _onlySE := FALSE;
   arguments := args();
   j := args(0);
   
   for i from 3 to j do
     if op(arguments[i],1) = hold(OnlySpecifiedEdges) then
       _onlySE := op(arguments[i],2);
       delete arguments[i];
       j := j - 1;
       break;
     end_if;
   end_for;

   _directed := Dom::Graph::isDirected(G);
   _correctEdges := Dom::Graph::selectEdge(G, Edge, FALSE, hold(OnlySpecifiedEdges)=_onlySE);       

   // If graph is undirected put also reverted edges to it.
   if not _directed and not _onlySE then
     tmpList := _correctEdges;
     for _edge in tmpList do
      _correctEdges := append(_correctEdges, revert(_edge));
     end_for;      
   end_if;

   _vertices := Dom::Graph::getVertices(G);
   _edgeCosts := Dom::Graph::getEdgeCosts(G);
   _edgeDescriptions := Dom::Graph::getEdgeDescriptions(G);
   _edgeWeights := Dom::Graph::getEdgeWeights(G);
   _edgesLeaving := Dom::Graph::getEdgesLeaving(G);
   _edgesEntering := Dom::Graph::getEdgesEntering(G);

   // Check if vertices exist !!!
    tmpList := Dom::Graph::checkForVertices(_correctEdges, Dom::Graph::getVertices(G));
    if not tmpList = [] then
      error("One or more edges contain vertices that are not in list : " . expr2text(tmpList) . "\n" );
    end_if;

   // first parameter (Edge) is mandatory, whereas all the rest is optional
   for i from 3 to j do
      if not testtype(arguments[i], "_equal") then
        error("Wrong argument type: equation expected!");
      end_if;
      // Store the description in leftOperator
      leftOperator := op(arguments[i],1); 
      // Store the edges in rightOperator
      rightOperator := op(arguments[i],2);
      
      if testargs() then
        if domtype(rightOperator) <> DOM_LIST then
          error("Wrong argument type for " . expr2text(leftOperator) . ": list expected!");
        end_if;
        if nops(rightOperator) < nops(Edge) then
          error("Number of " . expr2text(leftOperator) . " is smaller than number of edges.");
        elif nops(rightOperator) > nops(Edge) then
          error("Number of edges is smaller than number of " . expr2text(leftOperator) . "\n");
        end_if;
      end_if;

      case leftOperator
        of hold(EdgeWeights) do
          for x in _correctEdges do
            position := contains(Edge,x);
            _edgeWeights[x] := rightOperator[position];
          end_for;
          if _directed = FALSE then
            for x in _correctEdges do
              position := contains(Edge,x);
              _edgeWeights[revert(x)] := rightOperator[position];
            end_for;
          end_if;
          _edgeWeights := select(_edgeWeights, x-> not op(x,2)=hold(None));
          if nops(_edgeWeights) = 0 then
            _edgeWeights := FAIL;
          end_if;
          break;
        of hold(EdgeCosts) do
          for x in _correctEdges do
            position := contains(Edge,x);
            _edgeCosts[x] := rightOperator[position];
          end_for;
          if _directed = FALSE then
            for x in _correctEdges do
              position := contains(Edge,x);
              _edgeCosts[revert(x)] := rightOperator[position];
            end_for;
          end_if;
          _edgeCosts := select(_edgeCosts, x-> not op(x,2)=hold(None));
          if nops(_edgeCosts) = 0 then
            _edgeCosts := FAIL;
          end_if;
          break;
        of hold(EdgeDescriptions) do
          for x in _correctEdges do
            position := contains(Edge,x);
            case type(rightOperator[position])
              of DOM_STRING do
                if rightOperator[position] <> "" then
                  _edgeDescriptions[x] := rightOperator[position];                      
                  if _directed = FALSE then
                    _edgeDescriptions[revert(x)] := rightOperator[position];
                  end_if;
                end_if;
                break;
              otherwise
                _edgeDescriptions[x] := expr2text(rightOperator[position]);
                if _directed = FALSE then
                  _edgeDescriptions[revert(x)] := expr2text(rightOperator[position]);
                end_if;
                break;
            end_case;
          end_for;
          _edgeDescriptions := select(_edgeDescriptions, x-> not op(x,2)="None");
          if nops(_edgeDescriptions) = 0 then
            _edgeDescriptions := FAIL;
          end_if;
          break;
        otherwise
          error("Wrong argument type: " . expr2text(leftOperator) . ". Please consult help for more information.");
      end_case;
    end_for;
    
   /*-
    Generate the adjacency lists
   -*/

   _edges := _correctEdges;
   _edges := append(_edges, op(Dom::Graph::getEdges(G)));  
   
   _edgesEntering := table((_vertices[i]=sort( map(select(_edges, _edge->op(_edge,2)=_vertices[i]), op, 1) ) $ i=1..nops(_vertices))):
   _edgesLeaving := table((_vertices[i]=sort(map(select(_edges, _edge->op(_edge,1)=_vertices[i]), op, 2) ) $ i=1..nops(_vertices))):

   subsop(G, 2=_edges, 4=_edgeDescriptions, 5=_edgeWeights, 6=_edgeCosts, 7=_edgesEntering, 8=_edgesLeaving);
end_proc:

Dom::Graph::addEdges := Dom::Graph::addEdge:

/*- 
 removeEdge - removes one or several edges from a Dom::Graph
 removeEdge(G, Edge)
 removeEdge(G, [e1..en] <,OnlySpecifiedEdges=TRUE|FALSE>)
 Parameters:
   G    - Dom::Graph
   Edge - a list of one or more edges
-*/
Dom::Graph::removeEdge := proc(G : Dom::Graph, Edge : Type::ListOf(DOM_LIST))
local _edges, _correctEdges, _edgeWeights, _edgeDescriptions, _edgeCosts, _edgesEntering, _edgesLeaving, 
      x, position, operands, tmpList, _onlySE;
begin
   if testargs() then
     if args(0) < 2 then
       error("Wrong number of arguments! See help for more details.")
     end_if;
   end_if;

   _onlySE := FALSE;
   if args(0) = 3 then
     if op(args(3),1) = hold(OnlySpecifiedEdges) then
       _onlySE := op(args(3),2);
     end_if;
   end_if;
   // Get the already existing vertices
   _correctEdges := Dom::Graph::selectEdge(G, Edge, TRUE, Show=FALSE, OnlySpecifiedEdges=_onlySE);

   // If graph is undirected put also reverted edges to it.
   if Dom::Graph::isDirected(G) = FALSE and not _onlySE then
     tmpList := _correctEdges;
     for x in tmpList do
      _correctEdges := append(_correctEdges, revert(x));
     end_for;      
   end_if;

   _edges := Dom::Graph::getEdges(G);
   _edgeWeights := Dom::Graph::getEdgeWeights(G);
   _edgeDescriptions := Dom::Graph::getEdgeDescriptions(G);
   _edgeCosts := Dom::Graph::getEdgeCosts(G);
   _edgesEntering := Dom::Graph::getEdgesEntering(G);
   _edgesLeaving := Dom::Graph::getEdgesLeaving(G);

   // Check was already done and only not existing edges remain !
   // Also no check for Directed must be done, because if it is directed pair-edges will always be returned.
   for x in _correctEdges do
     position := contains(_edges,x);
     if position > 0 then
       delete _edges[position];
     end_if;

     if _edgeWeights <> FAIL then
       delete _edgeWeights[x];
       if nops(_edgeWeights) = 0 then
         _edgeWeights := FAIL;
       end_if;                  
     end_if;

     if _edgeDescriptions <> FAIL then
       delete _edgeDescriptions[x];
       if nops(_edgeDescriptions) = 0 then
         _edgeDescriptions := FAIL;
       end_if;  
     end_if;

     if _edgeCosts <> FAIL then
       delete _edgeCosts[x];
       if nops(_edgeCosts) = 0 then
         _edgeCosts := FAIL;
       end_if;  
     end_if;

     if _edgesEntering <> FAIL then
       operands := _edgesEntering[x[2]];
       position := contains(operands, x[1]);
       delete operands[position];          
       _edgesEntering[x[2]] := sort(operands);
     end_if;

     if _edgesLeaving <> FAIL then
       operands := _edgesLeaving[x[1]];
       position := contains(operands, x[2]);
       delete operands[position];          
       _edgesLeaving[x[1]] := sort(operands);
     end_if;
   end_for;

   subsop(G, 2=_edges, 4=_edgeDescriptions, 5=_edgeWeights, 6=_edgeCosts, 7=_edgesEntering, 8=_edgesLeaving);
end_proc:

/*-
 selectVertex - checks if one ore more Vertices are already defined in a graph and returns a new list with vertices
               that are either not in the graph so far or that already are defined depending on "existing".
               If the optional parameter Show is set to false, already defined vertices will not be printed.
 selectVertex(G, Vertex, existing)
 selectVertex(G, Vertex, existing <,Show = TRUE/FALSE>)
 selectVertex(G, [e1,..,en], existing)
 selectVertex(G, [e1,..,en], existing <,Show = TRUE/FALSE>)
 Parameters:
   G        - Dom::Graph
   Vertex   - a list of one or more vertices
   existing - TRUE returns all vertices that already exist in the graph.
              FALSE returns all vertices that are not in the graph.
   Show     - TRUE (standard) show if any vertices were found not matching the criteria.
              FALSE supress the output.
-*/
Dom::Graph::selectVertex := proc(G : Dom::Graph, Vertex : DOM_LIST, existing)
local _currentVertices, _existingVertices, _newVertices, dummy, leftOperator, rightOperator, _show; 
begin
  if testargs() then
    if args(0) < 3 then
      error("Wrong number of arguments! See help for more details.");
    end_if;
  end_if;

  _show := TRUE; // predefinition

  if args(0) = 4 then
    if testargs() then
      if not testtype(args(4), "_equal") then
        error("Wrong argument type: equation expected!");
      end_if;
    end_if;
    leftOperator := op(args(4),1); 
    rightOperator := op(args(4),2);
    case leftOperator
      of hold(Show) do
        _show := rightOperator;
        break;
      otherwise
        error("Wrong type of ".expr2text(args(4))." argument: Show expected!");
    end_case;
  end_if;

  _currentVertices := Dom::Graph::getVertices(G);

  if existing = TRUE then
    [_newVertices, _existingVertices,dummy] := split(_currentVertices, _edge -> contains(Vertex, _edge) > 0);
  else
    [_existingVertices, _newVertices,dummy] := split(Vertex, _edge -> contains(_currentVertices, _edge) > 0);
  end_if;

  if _show = TRUE then
    if _existingVertices <> [] then
      warning("The following vertices already exist: " . expr2text(_existingVertices) . "\n");
    end_if;
  end_if;

  return(_newVertices);

end_proc:

/*- 
 addVertex - adds one or several vertices to a Dom::Graph
 addVertex(G, Vertex)
 addVertex(G, Vertex, VertexWeights=n)
 addVertex(G, [v1,..,vn])
 addVertex(G, [v1,..,vn], VertexWeights=[c1,..cn])
 Parameters:
   G             - Dom::Graph
   Vertex        - a list of one or more vertices
   VertexWeights - a list of one or more vertice weights
                   If no Weight was specified, the Weight is set to 1 (in order to work with some algorithms)
-*/
Dom::Graph::addVertex := proc(G : Dom::Graph, Vertex : DOM_LIST)
local _vertices, _correctVertices, _vertexWeights, leftOperator, rightOperator, x, position, _edgesLeaving, _edgesEntering;
begin
  if testargs() then
    if args(0) < 2 or args(0) > 3 then
      error("Wrong number of arguments! See help for more details.")
     end_if;
  end_if;

  _correctVertices := Dom::Graph::selectVertex(G, Vertex, FALSE);
  _vertexWeights := FAIL;
  
  case args(0) 
    of 2 do
      _vertices := Dom::Graph::getVertices(G)._correctVertices;  // add list to existing vertices
      break;
    of 3 do
      if not testtype(args(3), "_equal") then
        error("Wrong argument type: equation expected!");
      end_if;

      // Should be VertexWeights.
      leftOperator := op(args(3),1); 
      // Should be the list.
      rightOperator := op(args(3),2);

      if leftOperator <> hold(VertexWeights) then
        error("Wrong argument type " .expr2text(args(4)) . ". VertexWeights expected!");
      end_if;

      if testtype(Vertex, DOM_LIST) then
         if nops(Vertex) <> nops(rightOperator) then
           error("Length of lists must be equal!");
         end_if;
         _vertices := Dom::Graph::getVertices(G)._correctVertices;
         _vertexWeights := Dom::Graph::getVertexWeights(G);
         // Check was already done and only not existing edges remain!
         for x in _correctVertices do
           position := contains(Vertex,x);
           if not op(rightOperator) = hold(None) then
             _vertexWeights[x] := rightOperator[position];              
           end_if;
         end_for;           
        else
          error("Vertice weights must be handed as a list.");
      end_if;
      break;
    otherwise
      error("Wrong argument type: " . expr2text(leftOperator) . ". Please consult help for more information.");
  end_case;

  _edgesLeaving := Dom::Graph::getEdgesLeaving(G);
  _edgesEntering := Dom::Graph::getEdgesEntering(G);

  for x in _correctVertices do
    _edgesLeaving[x] := [];
    _edgesEntering[x] := [];
  end_for;

  _vertices := sort(_vertices);

  subsop(G, 1=_vertices, 3=_vertexWeights, 7=_edgesEntering, 8=_edgesLeaving);
end_proc:
Dom::Graph::addVertices := Dom::Graph::addVertex:
// ccr: stopped here.

/*- 
 removeVertex - removes one or several vertices from a Dom::Graph. CAUTION: If a vertex is removed that is used within
                any edge, the edge will be deleted, too.
 removeVertex(G, Vertex)
 removeVertex(G, [v1,..,vn])
 Parameters:
   G              - Dom::Graph
   Vertex         - a list of one or more vertices

-*/
Dom::Graph::removeVertex := proc(G : Dom::Graph, Vertex:DOM_LIST)
local _vertices, _vertexWeights, _correctVertices, v, _edge, position, _edges, _edgesLeaving, _edgesEntering;
begin
  if testargs() then
    if args(0) < 2 then
      error("Wrong number of arguments! See help for more details.")
    end_if;
  end_if;
        
  // Get the already existing vertices
  _correctVertices := Dom::Graph::selectVertex(G, Vertex, TRUE, Show=FALSE);
  _vertices := Dom::Graph::getVertices(G);
  _vertexWeights := Dom::Graph::getVertexWeights(G);

  // Check was already done and only existing vertices remain !
  for v in _correctVertices do

    // Every time !
    _edges := Dom::Graph::getEdges(G);
    for _edge in _edges do
      if contains(_edge, v) > 0 then
        G := Dom::Graph::removeEdge(G, [_edge]);
      end_if;
    end_for;

    position := contains(_vertices, v);
    if position > 0 then
      delete _vertices[position];
    end_if;

    if _vertexWeights <> FAIL then
      delete _vertexWeights[v];           
    end_if;

    _edgesEntering := Dom::Graph::getEdgesEntering(G);
    delete _edgesEntering[op(v)];

    _edgesLeaving := Dom::Graph::getEdgesLeaving(G);
    delete _edgesLeaving[op(v)];

  end_for;           

  subsop(G, 1=_vertices, 3=_vertexWeights, 7=_edgesEntering, 8=_edgesLeaving);
end_proc:

/*
  createRandomEdgeCosts - creates random edge costs
  Parameters:
  G        - Dom::Graph
  range    - a range in between which the numbers have to be
  Real|Int - Either the numbers are integer or real.
*/
Dom::Graph::createRandomEdgeCosts := proc(G : Dom::Graph, range : "_range")
local _edges, edge, _edgeCosts, createNumber, i, domType;
begin
  createNumber := random(range);
  _edges := Dom::Graph::getEdges(G);
  domType := hold(Int);

  if args(0) = 3 then
    case args(3)
      of hold(Real) do
        domType := hold(Real);
        break;
      of hold(Int) do
        break;
      otherwise
        error("Wrong argument definition (try: Int or Real)!");
    end_case;  
  end_if;

  for edge in _edges do
     i := createNumber() $ 1;
     if domType = hold(Real) then
       i := i + frandom();
     end_if;
    _edgeCosts[edge] := i;
  end_for;
  subsop(G, 6=_edgeCosts);
end_proc:

/*
  createRandomEdgeWeights - creates random edge weights
  Parameters:
  G        - Dom::Graph
  range    - a range in between which the numbers have to be
  Real|Int - Either the numbers are integer or real.
*/
Dom::Graph::createRandomEdgeWeights := proc(G : Dom::Graph, range : "_range")
local _edges, edge, _edgeWeights, createNumber, i, domType;
begin
  createNumber := random(range);
  _edges := Dom::Graph::getEdges(G);
  domType := hold(Int);

  if args(0) = 3 then
    case args(3)
      of hold(Real) do
        domType := hold(Real);
        break;
      of hold(Int) do
        break;
      otherwise
        error("Wrong argument definition (try: Int or Real)!");
    end_case;  
  end_if;

  for edge in _edges do
     i := createNumber() $ 1;
     if domType = hold(Real) then
       i := i + frandom();
     end_if;
    _edgeWeights[edge] := i;
  end_for;
  subsop(G, 5=_edgeWeights);
end_proc:

/*
  createRandomVertexWeights - creates random edge weights
  Parameters:
  G        - Dom::Graph
  range    - a range in between which the numbers have to be
  Real|Int - Either the numbers are integer or real.
*/
Dom::Graph::createRandomVertexWeights := proc(G : Dom::Graph, range : "_range")
local _vertices, vertex, _vertexWeights, createNumber, i, domType;
begin
  createNumber := random(range);
  _vertices := Dom::Graph::getVertices(G);
  domType := hold(Int);

  if args(0) = 3 then
    case args(3)
      of hold(Real) do
        domType := hold(Real);
        break;
      of hold(Int) do
        break;
      otherwise
        error("Wrong argument definition (try: Int or Real)!");
    end_case;  
  end_if;

  for vertex in _vertices do
     i := createNumber() $ 1;
     if domType = hold(Real) then
       i := i + frandom();
     end_if;
    _vertexWeights[vertex] := i;
  end_for;
  subsop(G, 3=_vertexWeights);
end_proc:

/*-
 createCompleteGraph - generates a complete undirected Dom::Graph with n vertices (Clique)
 createCompleteGraph(n)
 Paramters:
   n - non-negative integer that defines the number of vertices to be used
-*/
Dom::Graph::createCompleteGraph := proc(n : Type::PosInt)
local i, j, _vertices, _edges, G, _edgesEntering, _edgesLeaving;
begin
  if testargs() then
    if args(0) <> 1 then
      error("Wrong number of arguments!");
    end_if;
  end_if;
  _vertices := [$1..n];
  _edges := (([i,j] $ j=1..i-1), ([i,j] $ j=i+1..n)) $ i=1..n;
  
  G := Dom::Graph([], []);

  // Generate the adjacency lists (without edges like (i, i))
  for i from 1 to nops(_vertices) do
    _edgesEntering[i] := [j $ j=1..(i-1), j $ j=(i+1)..nops(_vertices)];
  end_for;

  _edgesLeaving := _edgesEntering;

  subsop(G, 1=_vertices, 2=_edges, 7=_edgesEntering, 8=_edgesLeaving);
end_proc:

/* 
 createRandomGraph - generates a random graph with number of edges and number of vertices
 createRandomGraph(VertexNr, EdgeNr <, Undirected>)
 Parameters:
   VertexNr - positive integer that defines the number of vertices to be used
   EdgeNr   - positive integer that defines the number of edges to be used
   Undirected - if specified, the graph will be undirected
*/
Dom::Graph::createRandomGraph := proc(VertexNr : Type::PosInt, EdgeNr : Type::PosInt)
local _vertices, _edges, i, j, _isDirected, createNumber, G2;
begin
  if testargs() then
    if (args(0) < 1) or (args(0) > 3) then
      error("Wrong number of arguments!");
    end_if;
  end_if;

  _isDirected := TRUE;

  if args(0) = 3 then
    if args(3) = hold(Undirected) then
      _isDirected := FALSE;
    end_if;
  end_if;

  if (fact(VertexNr)/2) < EdgeNr then
  	 warning("Can not produce that many edges! Creating a complete graph instead.");
    return (Dom::Graph::createCompleteGraph(VertexNr));
  end_if;

  _edges := [];
  _vertices := [$1..VertexNr];
  // Two while loops with same break condition because edges can be defined multiple !
  createNumber := random(1..VertexNr);
  while nops(_edges) < EdgeNr do
    while nops(_edges) < EdgeNr do
      i := createNumber() $ 1;
      j := createNumber() $ 1;
      while i = j do
        j := createNumber() $ 1;      
      end_while;
      _edges := append(_edges, [i,j])
    end_while;
    if not _isDirected then
      for i from 1 to nops(_edges) do
        _edges[i] := sort(_edges[i]);
      end_for;
    end_if;
    // Remove duplicates:
    _edges:= [op({op(_edges)})];
  end_while;

  if _isDirected then
    G2 := Dom::Graph::new(_vertices, _edges, hold(Directed));
  else
    G2 := Dom::Graph::new(_vertices, _edges);
  end_if;

  G2;
end_proc:

/*- 
 createCircleGraph - generates a circleGraph
 createCircle([n|Vertices] <,Directed|Undirected>) Default = Directed.
 createCircle([v1,..,vn]) generates a cycle with the edges [v1,v2], [v2,v3], ... , [v(n-1),vn], [vn,v1].

 Parameters:
    n - Number of vertices to be used.
    L - list (of vertices)
-*/
Dom::Graph::createCircleGraph := proc(L)
local _edges, i, _vertices, _directed;
begin
  if testargs() then
    if args(0) < 1 or args(0) > 2 then
      error("Wrong number of arguments!");
    end_if;
    if (not testtype(L, DOM_LIST)) and (not testtype(L, DOM_INT)) then
      error("Wrong argument type: list or positive integer expected!");
    end_if;
  end_if;
  
  _directed := hold(Directed);

  if args(0) = 2 then
    if args(2) = hold(Directed) then
      _directed := hold(Directed);
    elif args(2) = hold(Undirected) then
      _directed := hold(Undirected);
    else
      error("Wrong argument! Try Directed or Undirected.");
    end_if;
  end_if;

  if domtype(L) = DOM_LIST then
    _vertices := L;
  else
    _vertices := [i $ i=1..L];
  end_if;
  _edges := [[_vertices[i],_vertices[i+1]] $ i=1..(nops(_vertices)-1)]:
  _edges := append(_edges, [_vertices[nops(_vertices)], _vertices[1]]);

  Dom::Graph::new(_vertices, _edges, _directed);
end_proc:

/*
  shortestPathSingleSource - finds a shortest path from a given vertex to a given or every vertex in a graph.
  shortestPathSingleSource(G, startVertex <, EndVertex = [n]> <, SearchWith = Dijkstra|Bellman> <, SearchFor = Weights|Costs> <,ReturnAsTable>)
  Parameters:
  G             - Dom::Graph
  StartVertex   - List containing one vertex from G.
  EndVertex     - List containing one vertex from G.
  SearchWith    - either Bellman or Dijkstra. Default = Bellman.
  SearchFor     - either Weights or Costs. Default = Weights.
  ReturnAsGraph - if TRUE a Graph is returned (only possible, if EndVertex is given !) a TABLE otherwise. 
*/
Dom::Graph::shortestPathSingleSource := proc(G : Dom::Graph)
local searchWith, searchFor, leftOperator, rightOperator, endVertex, _output, _solution, _edges, tmpVertex,
      _edgeWeights, _edgeCosts, _edgeDescriptions, _edgesLeaving, _edgesEntering, _vertices, _vertexWeights,
      wayTable, predTable, i, dummyTable, arguments;

begin 
  if testargs() then
    if (args(0) < 1) then
      error("Wrong number of arguments! See help for more details.")
    end_if;
  end_if;

  if Dom::Graph::isEmpty(G) then
  	 error("Graph must have at least one edge to make a search!");
  end_if;

  searchWith := hold(Bellman);
  searchFor := hold(Weights);
  _output := hold(Tables);
  endVertex := FAIL;

  arguments := args();
  for i from 2 to args(0) do
  	 if arguments[i] = hold(ReturnAsGraph) then
      _output := hold(Graph);
      delete arguments[i];
      break;
  	 end_if;
  end_for;

  for i from 2 to args(0) do
    if not testtype(arguments[i], "_equal") then
      error("Wrong type of " . expr2text(arguments[i]) . ": equation expected!");
    end_if;
    leftOperator := op(arguments[i],1); 
    rightOperator := op(arguments[i],2);

    case leftOperator
      of hold(SearchWith) do
        if rightOperator <> hold(Bellman) and rightOperator <> hold(Dijkstra) then
          error("Wrong argument type for " . expr2text(leftOperator) . ": Bellman or Dijkstra expected!");
        end_if;
        searchWith := rightOperator;        
        break;
      of hold(SearchFor) do
        if rightOperator <> hold(Weights) and rightOperator <> hold(Costs) then
          error("Wrong argument type for " . expr2text(leftOperator) . ": Weights or Costs expected!");
        end_if;
        searchFor := rightOperator;        
        break;
      of hold(StartVertex) do
        if domtype(rightOperator) <> DOM_LIST then
          error("Wrong argument type for " . expr2text(leftOperator) . ": list expected!");
        end_if;
        startVertex := rightOperator;
        break;
      of hold(EndVertex) do
        if domtype(rightOperator) <> DOM_LIST then
          error("Wrong argument type for " . expr2text(leftOperator) . ": list expected!");
        end_if;
        endVertex := op(rightOperator);
        break;
    end_case;
  end_for;

  if endVertex = FAIL then
    if _output = hold(Graph) then
      _output := hold(Tables); // Switch back to tables for output. (Definition)
    end_if;
  end_if;

  if searchWith = hold(Bellman) then    
    _solution := Dom::Graph::sssp_Bellman(G, startVertex, searchFor);
  else
    _solution := Dom::Graph::sssp_Dijkstra(G, startVertex, searchFor);
  end_if;

  if endVertex = FAIL then
    // just return the solution.
    return (_solution);
  end_if;

  startVertex := op(startVertex);
  tmpVertex := op(endVertex);
  _vertices := [tmpVertex];
  _edges := [];
  predTable := table();
  wayTable := table();

  if contains(op(_solution,2), tmpVertex) then
  // endVertex is accessible from startVertex
  // Gather information about vertices and edges existing in shortest path
  while not tmpVertex = startVertex do
    _edges := append(_edges, [op(_solution,2)[tmpVertex], tmpVertex]);
    wayTable[tmpVertex] := op(_solution,1)[tmpVertex];
    // For the table output create the way
    if not tmpVertex = startVertex then
      predTable[tmpVertex] := op(_solution,2)[tmpVertex];
    end_if;
    tmpVertex := op(_solution,2)[tmpVertex];
    _vertices := append(_vertices, tmpVertex);
  end_while;  
  wayTable[tmpVertex] := op(_solution,1)[tmpVertex];
  end_if;


  if _output = hold(Tables) then
    return([wayTable, predTable]);    
  end_if;

  dummyTable := Dom::Graph::getVertexWeights(G);
  if (not dummyTable = FAIL)  then
    for i in _vertices do
      _vertexWeights[i] := _vertexWeights[i];
    end_for;
  else
    _vertexWeights := FAIL;
  end_if;

  dummyTable := Dom::Graph::getEdgeCosts(G);
  if (not dummyTable = FAIL)  then
    for i in _edges do
      _edgeCosts[i] := dummyTable[i];
    end_for;
  else
    _edgeCosts := FAIL;
  end_if;

  dummyTable := Dom::Graph::getEdgeWeights(G);
  if (not dummyTable = FAIL)  then
    for i in _edges do
      _edgeWeights[i] := dummyTable[i];
    end_for;
  else
    _edgeWeights := FAIL;
  end_if;

  dummyTable := Dom::Graph::getEdgeDescriptions(G);
  if (not dummyTable = FAIL)  then
    for i in _edges do
      _edgeDescriptions[i] := dummyTable[i];
    end_for;
  else
    _edgeDescriptions := FAIL;
  end_if;

  _edgesEntering := table((_vertices[i]=sort( map(select(_edges, _edge->op(_edge,2)=_vertices[i]), op, 1) ) $ i=1..nops(_vertices))):
  _edgesLeaving  := table((_vertices[i]=sort( map(select(_edges, _edge->op(_edge,1)=_vertices[i]), op, 2) ) $ i=1..nops(_vertices))):

  subsop(G, 1=_vertices, 2=_edges, 3=_vertexWeights, 4=_edgeDescriptions, 5=_edgeWeights, 6=_edgeCosts, 7=_edgesEntering, 8=_edgesLeaving);

end_proc:

/*- 
 sssp_Dijkstra - returns the shortest path length (weights or costs are used) from one single node to every other
 Running time O(|Vertices|^2).
 Works correctly for non-negative weights only.
 Uses a subroutine 'choose' for choosing a vertex with minimum distance label.
 sssp_Dijkstra(G, startVertex <,Weights|Costs>)
 Parameters:
   G           - Dom::Graph
   startVertex - The vertex from which the shortest path to every other vertex is to be found
-*/
Dom::Graph::sssp_Dijkstra := proc(G : Dom::Graph, startVertex : DOM_LIST)
save PRETTYPRINT;
local setToFill, startSet, C, bestResults, bestVertex, j, bestPredecessor, 
      actWeight, choose, _edgeWeights, _edgesLeaving, _vertices, i, searchWith;
begin 
  if testargs() then
    if (args(0) < 2) or (args(0) > 3) then
      error("Wrong number of arguments! See help for more details.")
    end_if;
    if args(0) = 3 then    
      if not args(3) = hold(Costs) and not args(3) = hold(Weights) then
         error("Wrong argument type: " . expr2text(args(3)) . ". Please consult help for more information.");
      end_if;
    end_if;
  end_if;

  searchWith := hold(Weights);

  if args(0) = 3 then    
    case args(3)
      of hold(Costs) do
        searchWith := hold(Costs);
        break;
      of hold(Weights) do
        break;
      otherwise
        error("Wrong argument definition (try: Weights or Costs)!");
    end_case;
  end_if;
  
  extractMin := proc(startSet, bestResults)
  local i, tempBestVertex, bestVertex;
  begin
    bestVertex := op(startSet,1);
    tempBestVertex := bestResults[bestVertex];
    for i in startSet do
      if bestResults[i] < tempBestVertex then
        bestVertex := i; 
        tempBestVertex := bestResults[i];
      end_if;
    end_for;
    bestVertex;
  end_proc:
  
  _vertices := Dom::Graph::getVertices(G);
  _edgesLeaving := Dom::Graph::getEdgesLeaving(G);
  if searchWith = hold(Weights) then
    _edgeWeights := Dom::Graph::getEdgeWeights(G);
  else
    _edgeWeights := Dom::Graph::getEdgeCosts(G);
  end_if;

  // Otherwise an error occurs in contains(...)
  if _edgeWeights = FAIL then
    _edgeWeights := table();
  end_if;
  bestPredecessor := table();

  if contains(_vertices, op(startVertex)) = 0 then
     error("The starting vertex is not in the set of vertices !");
  end_if;   

  /*- constant C was chosen, because calculations with infinity are quite expensive!!  -*/
  C := (nops(_vertices)+1) * max(op(_edgeWeights, [1..nops(_edgeWeights),2]));
  // If edges have no weight, weight 1 is assumed for every edge.
  if C = FAIL then
    C := (nops(_vertices)+1);
  end_if;

  setToFill := {}; 
  startSet := {op(_vertices)};

  bestResults := table((_vertices[i] = C $ i=1..nops(_vertices)));
  bestResults[op(startVertex)] := 0;

  while nops(setToFill) < nops(_vertices) do 
    bestVertex := extractMin(startSet, bestResults);
    setToFill := setToFill union {bestVertex}; 
    startSet := startSet minus {bestVertex};
    for j in _edgesLeaving[bestVertex] do
      if contains(_edgeWeights, [bestVertex,j]) then
        actWeight := _edgeWeights[[bestVertex,j]];
      else
        actWeight := 1;
      end_if;
      if bestResults[j] > bestResults[bestVertex] + actWeight then
        bestResults[j] := bestResults[bestVertex] + actWeight;
        if bestResults[bestVertex] + actWeight > 0 then
          // Otherwise, the procedure would get the wrong predecessor !
          bestPredecessor[j] := bestVertex;         
        end_if;
      end_if;
    end_for;
  end_while;

  if C <> 0 then
    bestResults := subs(bestResults, C=infinity);
  end_if;

  userinfo(NoNL, 1, "Starting vertex was " . expr2text(op(startVertex)) ."\n" );
  for i in bestResults do
    if op(i,1) <> op(startVertex) then
      userinfo(NoNL, 1, "The shortest way to vertex " . expr2text(op(i,1)) . " had length " . expr2text(bestResults[op(i,1)]) . " and ");
	if not contains(bestPredecessor, op(i,1)) then
          userinfo(NoNL, 1, "no vertex as predecessor.\n");
        else
          userinfo(NoNL, 1, "its predecessor was " . bestPredecessor[op(i,1)] . "\n");
        end_if;        
    end_if;
  end_for;  
  // Return two tables
  [bestResults,bestPredecessor];
end_proc:

/*-
 sssp_Bellman - shortest path algorithm of Bellman
 Gives shortest pathes from one single node
 "On a routing problem", Quart. Appl. Math. 16 (1958)
 Running time O(|Vertices| |Edge|) (FIFO implementation)
 Taken from: Ahuja, Magnanti, Orlin: Dom::Graph Flows, Prentice-Hall, 1993
             Section 5.4

 sssp_Bellman(G, startVertex <,Weights|Costs>)
 Parameters:
   G           - Dom::Graph
   startVertex - The vertex from which the shortest path to every other vertex is to be found
-*/
Dom::Graph::sssp_Bellman := proc(G : Dom::Graph, startVertex : DOM_LIST)
save PRETTYPRINT;
local bestResults, j, C, LIST, i, bestPredecessor, actWeight, pass, _vertices, _edgesLeaving, _edgeWeights, bestVertex, searchWith;
begin
  if testargs() then
    if (args(0) < 2) or (args(0) > 3) then
      error("Wrong number of arguments (2 or 3 expected)! See help for more details.")
    end_if;
    if args(0) = 3 then    
      if not args(3) = hold(Costs) and not args(3) = hold(Weights) then
        error("Wrong argument type: " . expr2text(args(3)) . ". Please consult help for more information.");
      end_if;
    end_if;
  end_if;

  searchWith := hold(Weights);

  if args(0) = 3 then    
    case args(3)
      of hold(Costs) do
        searchWith := hold(Costs);
        break;
      of hold(Weights) do
        break;
      otherwise
        error("Wrong argument definition (try: Weights or Costs)!");
    end_case;
  end_if;

  _vertices := Dom::Graph::getVertices(G);
  _edgesLeaving := Dom::Graph::getEdgesLeaving(G);

  if searchWith = hold(Weights) then
    _edgeWeights := Dom::Graph::getEdgeWeights(G);
  else
    _edgeWeights := Dom::Graph::getEdgeCosts(G);
  end_if;

  // Otherwise an error occurs in contains(...)
  if _edgeWeights = FAIL then
    _edgeWeights := table();
  end_if;
  bestPredecessor := table();

  if contains(_vertices, op(startVertex)) = 0 then
    error("The starting vertex is not in the set of vertices !");
  end_if;   

  /*- constant C was chosen, because calculations with infinity are quite expensive!!  -*/
  if nops(_edgeWeights) = 0 then
    C := (nops(_vertices)+1);
  else
    C := (nops(_vertices)+1) * max(op(_edgeWeights, [1..nops(_edgeWeights),2]));  	
  end_if;

  bestResults := table((_vertices[i] = C $ i=1..nops(_vertices)));
  bestResults[op(startVertex)] := 0;
  LIST := startVertex;
  pass := table((_vertices[i] = 0 $ i=1..nops(_vertices)));
  while LIST <> [] do
    bestVertex := LIST[1];
    delete LIST[1];
    pass[bestVertex] := pass[bestVertex] + 1;
    if pass[bestVertex] > nops(_vertices) then
       error("Graph contains negative cycle(s)!");
    end_if;
    for j in _edgesLeaving[bestVertex] do
      if contains(_edgeWeights, [bestVertex,j]) then
        actWeight := _edgeWeights[[bestVertex,j]];
      else
        actWeight := 1;
      end_if;
      if bestResults[j] > bestResults[bestVertex] + actWeight then
        bestResults[j] := bestResults[bestVertex] + actWeight;
        bestPredecessor[j] := bestVertex;
        if contains(LIST, j) = 0 then
          LIST := append(LIST, j);
        end_if;
      end_if;
    end_for;
  end_while;

  if C <> 0 then
    bestResults := subs(bestResults, C=infinity);
  end_if;

  userinfo(NoNL, 1, "Starting vertex was " . expr2text(op(startVertex)) ."\n" );
  for i in bestResults do
    if op(i,1) <> op(startVertex) then
      userinfo(NoNL, 1, "The shortest way to vertex " . expr2text(op(i,1)) . " had length " . expr2text(bestResults[op(i,1)]) . " and ");
	if not contains(bestPredecessor, op(i,1)) then
	    userinfo(NoNL, 1, "no vertex as predecessor.\n");
        else
          userinfo(NoNL, 1, "its predecessor was " . bestPredecessor[op(i,1)] . "\n");
        end_if;        
    end_if;
  end_for;

  // Return two tables
  [bestResults,bestPredecessor];
end_proc:

/*
  shortestPathAllPairs - finds a shortest path from every vertex to every other vertex.
  shortestPathSingleSource(G <, SearchFor = Weights|Costs>)
  Parameters:
  G          - Dom::Graph
  SearchFor  - either Weights or Costs. Default = Weights.
*/
Dom::Graph::shortestPathAllPairs := proc(G : Dom::Graph)
local searchFor, leftOperator, rightOperator, i;
begin 
  if testargs() then
    if (args(0) < 1) then
      error("Wrong number of arguments! See help for more details.")
    end_if;
  end_if;

  if Dom::Graph::isEmpty(G) then
  	 error("Graph must have at least one edge to make a search!");
  end_if;

  searchFor := hold(Weights);

  for i from 2 to args(0) do
    if not testtype(args(i), "_equal") then
      error("Wrong type of " . expr2text(i) . ": equation expected!");
    end_if;
    leftOperator := op(args(i),1); 
    rightOperator := op(args(i),2);

    case leftOperator
      of hold(SearchFor) do
        if rightOperator <> hold(Weights) and rightOperator <> hold(Costs) then
          error("Wrong argument type for " . expr2text(leftOperator) . ": Weights or Costs expected!");
        end_if;
        searchFor := rightOperator;        
        break;
    end_case;
  end_for;

  Dom::Graph::apsp_Floyd(G, searchFor);

end_proc:

/*-
 Algorithm for determining all pair shortest paths.
 Algorithm of Floyd, "Algorithm 97, Shortest Path",
 Comm. ACM, Vol. 5, 1962
 Running time O(|Vertices|^3)
 Taken from: Ahuja, Magnanti, Orlin: Dom::Graph Flows, Prentice-Hall, 1993
       Section 5.6

 The first table returned shows the sum of the way
 The second table shows the predecessor of the second vertex (i._edge. [a,5] = 4 => 4 is the predecessor vertex of 5)
 Using a constant C was not possible, because it would be reduced and in the end not being detected anymore.
 apsp_Floyd(G)

 Parameters:
   G       - Dom::Graph
-*/
Dom::Graph::apsp_Floyd := proc(G : Dom::Graph)
save PRETTYPRINT;
local i, j, k, bestResults, actWeight, bestPredecessor, _vertices, _edges, _edgeWeights, searchWith;
begin
  if testargs() then
    if args(0) < 1 then
      error("Wrong number of arguments! See help for more details.")
    end_if;
  end_if;

  searchWith := hold(Weights);

  if args(0) = 3 then    
    case args(3)
      of hold(Costs) do
        searchWith := hold(Costs);
        break;
      of hold(Weights) do
        break;
      otherwise
        error("Wrong argument definition (try: Weights or Costs)!");
    end_case;
  end_if;

  _vertices := Dom::Graph::getVertices(G);
  _edges := Dom::Graph::getEdges(G);

  if searchWith = hold(Weights) then
    _edgeWeights := Dom::Graph::getEdgeWeights(G);
  else
    _edgeWeights := Dom::Graph::getEdgeCosts(G);
  end_if;

  // Otherwise an error occurs in constains(...)
  if _edgeWeights = FAIL then
    _edgeWeights := table();
  end_if;

  for i in _vertices do
    for j in _vertices do 
      if i=j then 
        bestResults[i,j] := 0;
      elif contains(_edges, [i,j]) > 0 then
        if contains(_edgeWeights, [i,j]) then
          actWeight := _edgeWeights[[i,j]];
        else
          actWeight := 1;
        end_if;        
        bestResults[i,j] := actWeight;
        bestPredecessor[i,j] := i;
      else
        bestResults[i,j] := infinity;
        // It't the predecessor and it will not be used fopr calculations !
        // Therefore infinity means, no predecessor is available.
        bestPredecessor[i,j] := infinity; 
      end_if
    end_for;
  end_for;

  for k in _vertices do      
    for i in _vertices do
      if bestResults[i,k] + bestResults[k,i] < 0 then 
        error("Graph contains one or more negative cycles!");
      end_if;
      for j in _vertices do
        if bestResults[i,j] > bestResults[i,k] + bestResults[k,j] then
         bestResults[i,j] := bestResults[i,k] + bestResults[k,j];
         bestPredecessor[i,j] := bestPredecessor[k,j];
        end_if;
      end_for;
    end_for;
  end_for;

  // Return two tables
  [bestResults,bestPredecessor];
end_proc:

/*-
 longestPath - changes the Description of one or more edges of a Dom::Graph
 longestPath(G <, startVertex> <, Depth>)
 longestPath(G, [startVertex = [n]], [endVertex = [n]], [Depth = n])

 // start with the endVertex and build the tree backwards until the startVertex was found.

 Parameters:
   G           - Dom::Graph
   startVertex - a list of one Vertex.
   endVertex   - a list of one Vertex.
   Depth       - a positive integer defining the recursion depth for searching (Default = 30)
-*/
Dom::Graph::longestPath := proc(G : Dom::Graph)
save PRETTYPRINT;
local i, C, bestResults, bestPredecessor, _startVertex, goDeeper,
      _vertices, _edges, _edgeWeights, depth, leftOperator, rightOperator;

begin
  if testargs() then
    if (args(0) < 1) or (args(0) > 3) then
      error("Wrong number of arguments! See help for more details.")
    end_if;
  end_if;

  for i from 2 to args(0) do
    if not testtype(args(i), "_equal") then
      error("Wrong type of " . expr2text(i) . ": equation expected!");
    end_if;
    leftOperator := op(args(i),1); 
    rightOperator := op(args(i),2);

    case leftOperator
      of hold(Depth) do
        depth := rightOperator;
        break;
      of hold(StartVertex) do
        if domtype(rightOperator) <> DOM_LIST then
          error("Wrong argument type for " . expr2text(leftOperator) . ": list expected!");
        end_if;
        _startVertex := op(rightOperator);
        break;
    end_case;
  end_for;

  _vertices := Dom::Graph::getVertices(G);
  _edges := Dom::Graph::getEdges(G);
  _edgeWeights := Dom::Graph::getEdgeWeights(G);

  // Otherwise an error occurs in constains(...)
  if _edgeWeights = FAIL then
    _edgeWeights := table();
  end_if;

  /*- constant C was chosen, because calculations with infinity are quite expensive!!  -*/
  C := -(nops(_vertices)+1) * max(op(_edgeWeights, [1..nops(_edgeWeights),2]));

  // If edges have no weight, weight 1 is assumed for every edge.
  if C = FAIL then
    C := -(nops(_vertices)+1);
  end_if;

  // Here comes the algorithm
  goDeeper := proc(G, vertex)
  begin
    
  end_proc:


  if C <> 0 then
    bestResults := subs(bestResults, C=infinity);
  end_if;

  // Return two tables
  [bestResults,bestPredecessor];
end_proc:


/*-
 inDegree - vertex indegrees of a Graph. Graph must be directed !
 inDegree(G)
 inDegree(G, [v])
 inDegree(G, [v1,..,vn])
 Parameters:
   G - Dom::Graph
   v - vertices of G
-*/
Dom::Graph::inDegree := proc(G : Dom::Graph)
local i, _result, _vertices , _edgesEntering;
begin
  if testargs() then
    if args(0) < 1 then
      error("Wrong number of arguments!"):
    end_if;
  end_if;

  _vertices := Dom::Graph::getVertices(G);
  _edgesEntering := Dom::Graph::getEdgesEntering(G);
   
  case args(0)
    of 1 do
      return(table(_vertices[i] = nops(_edgesEntering[_vertices[i]]) $ i=1..nops(_vertices)));
    of 2 do
       if not testtype(args(2), DOM_LIST) then
         error("Wrong argument type: List expected!");
       end_if;

       _result := table();
       for i in args(2) do 
         if contains(_vertices, i) > 0 then
           _result[i] := nops(_edgesEntering[i]);
         else
           warning("Dom::Graph doesn't contain specified vertex " . i . "\n");
         end_if;
       end_for;

       _result;
       break;
    otherwise
      error("Wrong number of arguments! See help for more details.")
  end_case;
end_proc:

/*-
 outDegree - vertex outdegrees of a Graph. Graph must be directed !
 outDegree(G)
 outDegree(G, v)
 outDegree(G, [v1,..,vn])
 Parameters:
   G - Dom::Graph
   v - vertices of G
-*/
Dom::Graph::outDegree := proc(G : Dom::Graph)
local i, _result, _vertices , _edgesLeaving;
begin
  if testargs() then
    if args(0) < 1 then
      error("Wrong number of arguments!"):
    end_if;
  end_if;

  _vertices := Dom::Graph::getVertices(G);
  _edgesLeaving := Dom::Graph::getEdgesLeaving(G);
   
  case args(0)
    of 1 do
      return(table(_vertices[i] = nops(_edgesLeaving[_vertices[i]]) $ i=1..nops(_vertices)));
    of 2 do
       if not testtype(args(2), DOM_LIST) then
         error("Wrong argument type: List expected!");
       end_if;

       _result := table();
       for i in args(2) do 
         if contains(_vertices, i) > 0 then
           _result[i] := nops(_edgesLeaving[i]);
         else
           warning("Dom::Graph doesn't contain specified node " . i . "\n");
         end_if;
       end_for;
       return(_result);
       break;
    otherwise
      error("Wrong number of arguments! See help for more details.")
  end_case;
end_proc:

 /*-
 topSort - topological sorting of a Dom::Graph
 topSort(G)
 Parameters:
   G - Dom::Graph
-*/
Dom::Graph::topSort := proc(G)
local indeg, outdeg, i, j, _vertices, _edgesEntering, _edgesLeaving, 
      _number, Q, _result, _parent, predecessor;
begin
  if testargs() then
    Dom::Graph::graphCheck(G);
  end_if;

  if not Dom::Graph::isDirected(G) then
    error("The graph must be directed and acyclic to create a topological sorting.")
  end_if;

  _vertices := Dom::Graph::getVertices(G);
  _edgesEntering := Dom::Graph::getEdgesEntering(G);
  _edgesLeaving := Dom::Graph::getEdgesLeaving(G);

  indeg := Dom::Graph::inDegree(G);
  outdeg := Dom::Graph::outDegree(G);

  _number := 0;
  Q := [];
  // First all single vertices
  for i in _vertices do
    _parent[i] := infinity;
    if (indeg[i] = 0) and (outdeg[i] = 0) then
      _number := _number + 1;
      _result[_number] := i;
    end_if;
  end_for;

  // Now all the vertices with no indegree (circles are not allowed !)
  for i in _vertices do
    if (indeg[i] = 0) and (outdeg[i] > 0) then
      Q := append(Q,i);
    end_if;   
  end_for;

  /*-
   G contains cycle if there isn't any node with indeg[i] = 0 and outdeg[i] > 0
  -*/
  if nops(Q) = 0 then
    error("The graph contains at least one cycle or is empty.");
  end_if;

//  predecessor := Q[1];

  repeat
    i := Q[1];
    predecessor := i;
    delete Q[1];
    _number := _number + 1;
    _result[_number] := i;
    if _parent[i] = infinity then
      _parent[i] := predecessor;    	
    end_if;
    for j in _edgesLeaving[i] do
      indeg[j] := indeg[j] - 1;
      if _parent[j] = infinity then
        _parent[j] := i;      	
      end_if;
      if indeg[j] = 0 then        
        Q := append(Q, j);
      end_if;
    end_for;
  until Q = [] end_repeat;

  if _number < nops(_vertices) then
    error("The graph contains at least one cycle.");
  end_if;
  
  for i in _result do
    userinfo(NoNL, 1, "Vertex " . expr2text(rhs(i)) . " is used at timestamp " . expr2text(lhs(i)) . "\n" );    
  end_for;
  [_result, _parent];
end_proc:

/* subGraph - extract a subgraph from a Dom::Graph
   subGraph(G, v)
   subGraph(G, [v1,..vn])
   Parameters:
   G - Dom::Graph
   v - vertices of G
*/
Dom::Graph::getSubGraph := proc(G : Dom::Graph, Vertices : DOM_LIST)
  local _vertices, _edges, G2, i;
begin
  _vertices := select(Dom::Graph::getVertices(G), v -> contains(Vertices, v) <> 0);
  _edges := select(Dom::Graph::getEdges(G), _edge -> contains(Vertices, _edge[1]) <> 0 and contains(Vertices, _edge[2]) <> 0);

  if Dom::Graph::isDirected(G) then
    G2 := Dom::Graph::new(_vertices, _edges, hold(Directed));
  else  
    for i from 1 to nops(_edges) do
      _edges[i] := sort(_edges[i]);
    end_for;
    _edges:= [op({op(_edges)})];
    G2 := Dom::Graph::new(_vertices, _edges);
  end_if;

  if not Dom::Graph::getVertexWeights(G) = FAIL then
    G2 := Dom::Graph::setVertexWeights(G2, _vertices, select(Dom::Graph::getVertexWeights(G), vw -> contains(_vertices, lhs(vw)) <> 0));
  end_if;

  if not Dom::Graph::getEdgeWeights(G) = FAIL then
    G2 := Dom::Graph::setEdgeWeights(G2, _edges, select(Dom::Graph::getEdgeWeights(G), ew -> contains(_edges, lhs(ew)) <> 0));
  end_if;

  if not Dom::Graph::getEdgeCosts(G) = FAIL then
    G2 := Dom::Graph::setEdgeCosts(G2, _edges, select(Dom::Graph::getEdgeCosts(G),  ec -> contains(_edges, lhs(ec)) <> 0));
  end_if;

  if not Dom::Graph::getEdgeDescriptions(G) = FAIL then
    G2 := Dom::Graph::setEdgeDescriptions(G2, _edges, select(Dom::Graph::getEdgeDescriptions(G),  ec -> contains(_edges, lhs(ec)) <> 0));
  end_if;

  G2;
end_proc:

/*
 returns the amount of vertices in G.
*/
Dom::Graph::getVertexNumber := proc(G : Dom::Graph)
begin
  nops( Dom::Graph::getVertices(G) );
end_proc:

/*
  returns the amount of edges in G. (Remember to divide the number by 2 if the graph is undirected !)
*/
Dom::Graph::getEdgeNumber := proc(G : Dom::Graph)
begin
  nops( Dom::Graph::getEdges(G) );
end_proc:

/*-
 plotCircleGraph - graphical plot where vertices are arranged in a circle of a Dom::Graph
                   The first point is drawn in blue and the second in green (if not redefined) so the direction can be found out easily.
                   The vertices can be assigned easily too if after the plot(%) command getVertices Information is invoked.

 plotCircleGraph(G <, PointWidth=n> <, SpecialEdges=[e1,..,en]> <, EdgeColor=RGB>  <, SpecialEdgeColor=RGB> 
                   <, VertexColor=RGB> <, Vertex1Color=RGB> <, Vertex2Color=RGB>)
 Parameters:
   G                - Dom::Graph
   pointWidth       - n is a positive integer. Default = 40
   SpecialEdges     - a list of edges that have to be drawn in blue or the given color respectively.
   EdgeColor        - edges will be drawn in this color. Default = RGB::Red
   SpecialEdgeColor - If SpecialEdges is set then the selected edges will be drawn in this color. Default = RGB::Red
   VertexColor      - The color in which the vertices are drawn. Default = RGB::Red
   Vertex1Color     - The color in which the first vertex is drawn. Default = RGB::Blue
   Vertex2Color     - The color in which the second vertex is drawn. Default = RGB::Green
-*/
Dom::Graph::plotCircleGraph := proc(G : Dom::Graph)
local alpha, i, pointList, edgeList, selfList, _vertices, _edges, edge, numberEdges, 
      v, w, numberVertices, _isDirected, _pointWidth, _specialList, _edgeColor, _specialEdgeColor, 
      leftOperator, rightOperator, _vertexColor, _v1Color, _v2Color, drawColor;
begin
  if testargs() then
     if (args(0) < 1) then
       error("Wrong number of arguments!");
     end_if;
  end_if;

  _pointWidth := 40;
  _edgeColor := RGB::Red;
  _specialEdgeColor := RGB::Blue;
  _vertexColor := RGB::Red;
  _v1Color := RGB::Blue;
  _v2Color := RGB::Green;

  _specialList := [];
  _vertices := Dom::Graph::getVertices(G);
  _edges := Dom::Graph::getEdges(G);
  _isDirected := Dom::Graph::isDirected(G); // otherwise the function would have to be called n-times during the for loop.

  for i from 2 to args(0) do

    leftOperator := op(args(i),1);
    rightOperator := op(args(i),2);

    case leftOperator 
      of hold(PointWidth) do
        _pointWidth := rightOperator;
        break;
      of hold(SpecialEdges) do
        _specialList := rightOperator;
        // make sure the edges are correct for undirected case.
        if not _isDirected then
          for v from 1 to nops(_specialList) do
            _specialList[v] := sort(_specialList[v]);
          end_for;
          _specialList:= [op({op(_specialList)})];
        end_if;
        break;
      of hold(EdgeColor) do
        if testargs() then
          if not testtype(rightOperator, Type::ListOf(Type::Real, 3, 3)) then
            error("Wrong argument type: Color expected!");
          end_if
        end_if;
        _edgeColor := rightOperator;
        break;
      of hold(SpecialEdgeColor) do
        if testargs() then
          if not testtype(rightOperator, Type::ListOf(Type::Real, 3, 3)) then
            error("Wrong argument type: Color expected!");
          end_if
        end_if;
        _specialEdgeColor := rightOperator;
        break;
      of hold(VertexColor) do
        if testargs() then
          if not testtype(rightOperator, Type::ListOf(Type::Real, 3, 3)) then
            error("Wrong argument type: Color expected!");
          end_if
        end_if;
        _vertexColor := rightOperator;
        break;
      of hold(Vertex1Color) do
        if testargs() then
          if not testtype(rightOperator, Type::ListOf(Type::Real, 3, 3)) then
            error("Wrong argument type: Color expected!");
          end_if
        end_if;
        _v1Color := rightOperator;
        break;
      of hold(Vertex2Color) do
        if testargs() then
          if not testtype(rightOperator, Type::ListOf(Type::Real, 3, 3)) then
            error("Wrong argument type: Color expected!");
          end_if
        end_if;
        _v2Color := rightOperator;
        break;
      otherwise
        error("Wrong argument : " . expr2text(leftOperator));
    end_case;

  end_for;

  if nops(_vertices) < 1 then
    error("Graph must contain at least one vertex !");
  end_if;

  // Sorts out multiple edges (saves time during ploting !!!)
  if not _isDirected then
    for i from 1 to nops(_edges) do
      _edges[i] := sort(_edges[i]);
    end_for;
    _edges:= [op({op(_edges)})];
  end_if;

  numberVertices := nops(_vertices);
  numberEdges := nops(_edges);
  alpha := float(2*PI / (numberVertices));
  // Starting point
  pointList := [plot::Point(cos(0), sin(0), Color = _v1Color)];
  if nops(_vertices) > 1 then
    pointList := append(pointList, plot::Point(cos(alpha), sin(alpha), Color = _v2Color) );
  end_if;
  if nops(_vertices) > 2 then
    pointList := append(pointList, plot::Point(cos(i*alpha), sin(i*alpha), Color = _vertexColor) $ i=2..(numberVertices-1));
  end_if;
  edgeList := [];
  selfList := [];
  for i from 1 to numberEdges do
    edge := _edges[i];
    v := contains(_vertices, edge[1]); 
    w := contains(_vertices, edge[2]);
    if v = w then
      selfList := append(selfList, plot::Ellipse2d(pointList[v] * 1.05, 0.05, 0.05, Color = _edgeColor));
    else
      if (contains(_specialList, edge) > 0) then
        drawColor := _specialEdgeColor
      else
        drawColor := _edgeColor
      end_if;
      if _isDirected then
        edgeList := append(edgeList, plot::vector2d(pointList[v],pointList[w], PI/6.0, 0.05, Color = drawColor));
      else
        edgeList := append(edgeList, plot::line(pointList[v],pointList[w], Color = drawColor));
      end_if;
    end_if;
  end_for;

  plot::Scene(op(edgeList), op(pointList), op(selfList), Axes = None, PointWidth = _pointWidth,
              PointStyle = FilledCircles, LineWidth = 1);
end_proc:

/*-
 plotBipartiteGraph - graphical plot where vertices are arranged left and right of a box and are connected respectively

 plotBipartiteGraph(G)
 plotBipartiteGraph(G <,PointWidth=n> <,SpecialEdges=[e1,..,en]> <, EdgeColor=RGB> 
                      <, SpecialEdgeColor=RGB> <, VertexColor=RGB>, <Vertex1Color=RGB>, <Vertex2Color=RGB>)
 Parameters:
   G                - Dom::Graph
   PointWidth       - n is a positive integer. Default = 40
   SpecialEdges     - a list of edges that have to be drawn in blue or the given color respectively.
   EdgeColor        - edges will be drawn in this color. Default = RGB::Red
   SpecialEdgeColor - If SpecialEdges is set then the selected edges will be drawn in this color. Default = RGB::Red
   VertexColor      - The color in which the vertices are drawn. Default = RGB::Red
   Vertex1Color     - The color in which the first vertex is drawn. Default = RGB::Blue
   Vertex2Color     - The color in which the second vertex is drawn. Default = RGB::Green
-*/
Dom::Graph::plotBipartiteGraph := proc(G : Dom::Graph)
local i, pointList, edgeList, _edges, edge, numberEdges, v, w, 
      _isDirected, _pointWidth, sets, setOne, setTwo, _vertices, leftOperator, rightOperator,
      _edgeColor, _specialEdgeColor, _vertexColor, _v1Color, _v2Color, _specialList, drawColor;
begin
  if testargs() then
     if (args(0) < 1) then
       error("Wrong number of arguments!");
     end_if;
  end_if;

  _vertices := Dom::Graph::getVertices(G);
  _edges := Dom::Graph::getEdges(G);
  _isDirected := Dom::Graph::isDirected(G); // otherwise the function would have to be called n-times during the for loop.
  sets := Dom::Graph::bipartite(G, hold(Lists)); // To get either FAIL or the two sets in a list.

  if sets = FAIL then
    error("Graph is not bipartite !");
  else
    setOne := op(sets, 1);
    setTwo := op(sets, 2);
  end_if;

  _pointWidth := 40;
  _specialList := [];
  _edgeColor := RGB::Red;
  _specialEdgeColor := RGB::Blue;
  _vertexColor := RGB::Red;
  _v1Color := RGB::Blue;
  _v2Color := RGB::Green;

  for i from 2 to args(0) do
    leftOperator := op(args(i),1);
    rightOperator := op(args(i),2);

    case leftOperator 
      of hold(PointWidth) do
        _pointWidth := rightOperator;
        break;
      of hold(SpecialEdges) do
        _specialList := rightOperator;
        // make sure the edges are correct for undirected case.
        if not _isDirected then
          for v from 1 to nops(_specialList) do
            _specialList[v] := sort(_specialList[v]);
          end_for;
          _specialList:= [op({op(_specialList)})];
        end_if;
        break;
      of hold(EdgeColor) do
        if testargs() then
          if not testtype(rightOperator, Type::ListOf(Type::Real, 3, 3)) then
            error("Wrong argument type: Color expected!");
          end_if
        end_if;
        _edgeColor := rightOperator;
        break;
      of hold(SpecialEdgeColor) do
        if testargs() then
          if not testtype(rightOperator, Type::ListOf(Type::Real, 3, 3)) then
            error("Wrong argument type: Color expected!");
          end_if
        end_if;
        _specialEdgeColor := rightOperator;
        break;
      of hold(VertexColor) do
        if testargs() then
          if not testtype(rightOperator, Type::ListOf(Type::Real, 3, 3)) then
            error("Wrong argument type: Color expected!");
          end_if
        end_if;
        _vertexColor := rightOperator;
        break;
      of hold(Vertex1Color) do
        if testargs() then
          if not testtype(rightOperator, Type::ListOf(Type::Real, 3, 3)) then
            error("Wrong argument type: Color expected!");
          end_if
        end_if;
        _v1Color := rightOperator;
        break;
      of hold(Vertex2Color) do
        if testargs() then
          if not testtype(rightOperator, Type::ListOf(Type::Real, 3, 3)) then
            error("Wrong argument type: Color expected!");
          end_if
        end_if;
        _v2Color := rightOperator;
        break;
      otherwise
        error("Wrong argument : " . expr2text(leftOperator));
    end_case;

  end_for;

  if nops(_vertices) < 1 then
    error("Graph must contain at least one vertex !");
  end_if;

  // Sorts out multiple edges (saves time during ploting !!!)
  if not _isDirected then
    for i from 1 to nops(_edges) do
      _edges[i] := sort(_edges[i]);
    end_for;
    _edges:= [op({op(_edges)})];
  end_if;

  numberEdges := nops(_edges);
  pointList := [];

  // Set 1
  if nops(setOne) > 0 then
    pointList := append(pointList, plot::Point(0, 0, Color = _v1Color));
  end_if;
  if nops(setOne) > 1 then
    pointList := append(pointList, plot::Point(0, (i/10), Color = _vertexColor) $ i=1..(nops(setOne)-1));
  end_if;

  // Set 2
  if nops(setTwo) > 0 then
    pointList := append(pointList, plot::Point(10, 0, Color = _v2Color) );   
  end_if;
  if nops(setTwo) > 1 then
    pointList := append(pointList, plot::Point(10, (i/10), Color = _vertexColor) $ i=1..(nops(setTwo)-1));
  end_if;

  edgeList := [];

  for i from 1 to numberEdges do
    edge := _edges[i];
    v := contains(setOne, edge[1]);
    if v = 0 then
      // For directed graphs the direction is important !
      v := contains(setTwo, edge[1]) + nops(setOne);
      w := contains(setOne, edge[2]);
    else
      w := contains(setTwo, edge[2]) + nops(setOne);
    end_if;

    if (contains(_specialList, edge) > 0) then
      drawColor := _specialEdgeColor;
    else
      drawColor := _edgeColor;
    end_if;

    if _isDirected then
      edgeList := append(edgeList, plot::vector2d(pointList[v],pointList[w], PI/6.0, 0.05, Color = drawColor));
    else
      edgeList := append(edgeList, plot::line(pointList[v],pointList[w], Color = drawColor));
    end_if;
  end_for;

  plot::Scene(op(edgeList), op(pointList), Axes = None, PointWidth = _pointWidth,
              PointStyle = FilledCircles, LineWidth = 1);
end_proc:

/*-
 plotGridGraph - graphical plot where vertices are arranged left and right of a box and are connected respectively

 plotGridGraph(G)
 plotGridGraph(G <, PointWidth=n> <, SpecialEdges=[e1,..,en]> <, EdgeColor=RGB> <, VerticesPerLine=n|[n1..nm]> 
                 <, VertexOrder=[v1,..vn]> <, SpecialEdgeColor=RGB> <, SpecialVertices=[e1,..en]> <, SpecialVertexColor=RGB>
                 <, VertexColor=RGB> <, Vertex1Color=RGB> <, Vertex2Color=RGB>)
 Parameters:
   G                - Dom::Graph
   VerticesPerLine  - n is a positive integer. Default is the ceilling of the sqrt ot the number of vertices.
                      OR n1..nm define how many vertices are to be placed per line.
   VertexOrder      - defines in which order the vertices are to be arranged. None is defined as filler for 
                      vertices that should not be drawn.
   PointWidth       - n is a positive integer. Default = 40
   SpecialEdges     - a list of edges that have to be drawn in blue or the given color respectively.
   SpecialVertices  - a list of vertices that have to be drawn in blue or the given color respectively.
   EdgeColor        - edges will be drawn in this color. Default = RGB::Red
   SpecialEdgeColor - If SpecialEdges is set then the selected edges will be drawn in this color. Default = RGB::Red
   VertexColor      - The color in which the vertices are drawn. Default = RGB::Red
   Vertex1Color     - The color in which the first vertex is drawn. Default = RGB::Blue
   Vertex2Color     - The color in which the second vertex is drawn. Default = RGB::Green
-*/
Dom::Graph::plotGridGraph := proc(G : Dom::Graph)
local i, pointList, edgeList, selfList, _edges, edge, numberEdges, v, w, _vpl, rowCounter, _vplList,
      _isDirected, _pointWidth, _vertices, leftOperator, rightOperator, maxWidth, j, _height, tmpList, _arguments,
      _edgeColor, _specialEdgeColor, _specialVertexColor, _vertexColor, _v1Color, _v2Color, _specialEdgeList, _specialVertexList, 
      drawColor, alpha;
begin
  if testargs() then
     if (args(0) < 1) then
       error("Wrong number of arguments!");
     end_if;
  end_if;

  _vertices := Dom::Graph::getVertices(G);
  _edges := Dom::Graph::getEdges(G);
  _isDirected := Dom::Graph::isDirected(G); // otherwise the function would have to be called n-times during the for loop.

  _vpl := FAIL;
  _pointWidth := 40;
  _specialEdgeList := [];
  _specialVertexList := [];
  _edgeColor := RGB::Red;
  _specialEdgeColor := RGB::Blue;
  _specialVertexColor := RGB::Blue;
  _vertexColor := RGB::Red;
  _v1Color := RGB::Blue;
  _v2Color := RGB::Green;
  maxWidth := 0;
  rowCounter := 0;
  _arguments := [args()];     

  // Two parameters have to be checked first and then to be removed from the list.
  for i from 2 to nops(_arguments) do
    leftOperator := op(_arguments[i],1);
    rightOperator := op(_arguments[i],2);

    if leftOperator = hold(VertexOrder) then
      tmpList := {op(rightOperator)} minus {hold(None)};
      if nops(_vertices) <> nops(tmpList) then
        error("Number of defined vertices does not equal number of vertices in graph");
      end_if;
      _vertices := rightOperator;
      delete _arguments[i];
      break;
    end_if;
  end_for;

  for i from 2 to nops(_arguments) do
    leftOperator := op(_arguments[i],1);
    rightOperator := op(_arguments[i],2);
    if leftOperator = hold(VertexColor) then
      if testargs() then
        if not testtype(rightOperator, Type::ListOf(Type::Real, 3, 3)) then
          error("Wrong argument type for VertexColor: Color expected!");
        end_if;
      end_if;
      _vertexColor := rightOperator;
      _v1Color := rightOperator;
      _v2Color := rightOperator;
      delete _arguments[i];
      break;
    end_if;
  end_for;

  // Now the "usual" checking can be done.
  for i from 2 to nops(_arguments) do
    leftOperator := op(_arguments[i],1);
    rightOperator := op(_arguments[i],2);

    case leftOperator 
      of hold(VerticesPerLine) do
        if testargs() then
          if not domtype(rightOperator) = DOM_LIST and not testtype(rightOperator, Type::NonNegInt) then
            error("Wrong argument type for VerticesPerLine: List or positiveInteger expected!");
          end_if
        end_if;
        _vpl := rightOperator;
        if domtype(rightOperator) = DOM_LIST then
          _vplList := rightOperator;
          for j from 1 to nops(_vplList) do
            rowCounter := rowCounter + op(_vplList,j);
            if maxWidth < op(_vplList,j) then
              maxWidth := op(_vplList,j);
            end_if;
          end_for;
          if rowCounter <> nops(_vertices) then
            error("Number of vertices in list VerticesPerLine (" . expr2text(rowCounter) . ") must equal number of vertices (" . expr2text(nops(_vertices)) . ") !");
          end_if;
        else
          if _vpl < 1 then
            error("VerticesPerLine must be greater than 0!");
          end_if;
          _vplList := [];
          maxWidth := _vpl;
          for j from 1 to floor(nops(_vertices) / _vpl) do
            _vplList := append(_vplList, _vpl);
          end_for;
          _vplList := append(_vplList, (nops(_vertices) - (_vpl*j)) );
        end_if;
        break;
      of hold(PointWidth) do
        if testargs() then
          if not testtype(rightOperator, Type::NonNegInt) then
            error("Wrong argument type for PointWidth: positive Integer expected!");
          end_if
        end_if;
        _pointWidth := rightOperator;
        break;
      of hold(SpecialVertices) do
        if testargs() then
          if not domtype(rightOperator) = DOM_LIST then
            error("Wrong argument type for SpecialVertices: List expected!");
          end_if
        end_if;
        _specialVertexList := rightOperator;
        break;
      of hold(SpecialEdges) do
        if testargs() then
          if not testtype(rightOperator, Type::ListOf(DOM_LIST)) then
            error("Wrong argument type for SpecialEdges: List of edges expected!");
          end_if
        end_if;
        _specialEdgeList := rightOperator;
        // make sure the edges are correct for undirected case.
        if not _isDirected then
          for v from 1 to nops(_specialEdgeList) do
            _specialEdgeList[v] := sort(_specialEdgeList[v]);
          end_for;
          _specialEdgeList:= [op({op(_specialEdgeList)})];
        end_if;
        break;
      of hold(EdgeColor) do
        if testargs() then
          if not testtype(rightOperator, Type::ListOf(Type::Real, 3, 3)) then
            error("Wrong argument type for EdgeColor: Color expected!");
          end_if;
        end_if;
        _edgeColor := rightOperator;
        break;
      of hold(SpecialEdgeColor) do
        if testargs() then
          if not testtype(rightOperator, Type::ListOf(Type::Real, 3, 3)) then
            error("Wrong argument type for SpecialEdgeColor: Color expected!");
          end_if;
        end_if;
        _specialEdgeColor := rightOperator;
        break;
      of hold(SpecialVertexColor) do
        if testargs() then
          if not testtype(rightOperator, Type::ListOf(Type::Real, 3, 3)) then
            error("Wrong argument type for SpecialVertexColor: Color expected!");
          end_if
        end_if;
        _specialVertexColor := rightOperator;
        break;
      of hold(Vertex1Color) do
        if testargs() then
          if not testtype(rightOperator, Type::ListOf(Type::Real, 3, 3)) then
            error("Wrong argument type for Vertex1Color: Color expected!");
          end_if;
        end_if;
        _v1Color := rightOperator;
        break;
      of hold(Vertex2Color) do
        if testargs() then
          if not testtype(rightOperator, Type::ListOf(Type::Real, 3, 3)) then
            error("Wrong argument type for Vertex2Color: Color expected!");
          end_if;
        end_if;
        _v2Color := rightOperator;
        break;
      otherwise
        error("Wrong argument : " . expr2text(leftOperator));
    end_case;
  end_for;

  // No parameter for vpl was given...
  if _vpl = FAIL then
    _vpl := ceil(sqrt(nops(_vertices)));
    _vplList := [];
    maxWidth := _vpl;
    for i from 1 to _vpl-1 do
      _vplList := append(_vplList, _vpl);
    end_for;
    if nops(_vertices) - ((_vpl-1) * _vpl) > 0 then
      _vplList := append(_vplList, ( nops(_vertices) - ((_vpl-1) * _vpl) ) );
    end_if;
  end_if;

  if nops(_vertices) < 1 then
    error("Graph must contain at least one vertex !");
  end_if;

  // Sorts out multiple edges (saves time during ploting !!!)
  if not _isDirected then
    for i from 1 to nops(_edges) do
      _edges[i] := sort(_edges[i]);
    end_for;
    _edges:= [op({op(_edges)})];
    // JUST A TEST !!!!
    _edges := sort(_edges);
  end_if;

  if nops(_specialEdgeList) > 0  then
    // Sort out edges NOT existing in the graph
    numberEdges := Dom::Graph::selectEdge(G, _specialEdgeList, FALSE, Show=FALSE);
    if nops(numberEdges) > 0 then
      warning("The following edges in SpecialEdges do not exist in Graph : " . expr2text(numberEdges) . "\n");
    end_if;
  end_if;

  if nops(_specialVertexList) > 0 then
    numberEdges := Dom::Graph::selectVertex(G, _specialVertexList, FALSE, Show=FALSE);
    if nops(numberEdges) > 0 then
      warning("The following vertices in SpecialVertices do not exist in Graph : " . expr2text(numberEdges) . "\n");
    end_if;   
  end_if;

  numberEdges := nops(_edges);
  pointList := [];
  rowCounter := 0; //ceil(nops(_vertices) / _vpl);
  _height := nops(_vplList);

  // width between vertices = maxWidth / (VerticesPerLine - 2) + 1
  j := 0;
  _vpl := _vplList[rowCounter + 1];
  if _vpl = 1 then
    j := 1;
    alpha := maxWidth/2;
  else
    alpha := maxWidth / (_vpl - 1);
  end_if;
  if not _vertices[1] = hold(None) then
    if contains(_specialVertexList, _vertices[1]) > 0 then
      pointList := append(pointList, plot::Point(j * alpha, _height - rowCounter, Color = _specialVertexColor));    
    else
      pointList := append(pointList, plot::Point(j * alpha, _height - rowCounter, Color = _v1Color));   
    end_if;
  end_if;

  // More than a single point ?
  j := j + 1;
  if nops(_vertices) > 1 then
    if _vpl = 1 then
      rowCounter := rowCounter + 1;
      j := 0;  // because here it is reduced to 0 and therefore the first point goes to 0,y
      _vpl := _vplList[rowCounter + 1];
      if _vpl = 1 then
        j := 1;
        alpha := maxWidth / 2;
      else
        alpha := maxWidth / (_vpl - 1);
      end_if;
    end_if;
    if not _vertices[2] = hold(None) then
      if contains(_specialVertexList, _vertices[2]) > 0 then
        pointList := append(pointList, plot::Point(j * alpha, _height - rowCounter, Color = _specialVertexColor));    
      else
        pointList := append(pointList, plot::Point(j * alpha, _height - rowCounter, Color = _v2Color));   
      end_if;
    end_if;
  end_if;

  // More than 2 points ? Then every vertex is drawn in same color from now on.
  j := j + 1;
  if nops(_vertices) > 2 then
    for i from 3 to nops(_vertices) do
      if j >= _vpl then
        j := 0;
        rowCounter := rowCounter + 1;
        _vpl := _vplList[rowCounter+1];
        if _vpl = 1 then
          j := 1;
          alpha := maxWidth/2;
        else
          alpha := maxWidth / (_vpl - 1);   
        end_if;
      end_if;
      if not _vertices[i] = hold(None) then
        if contains(_specialVertexList, _vertices[i]) > 0 then
          pointList := append(pointList, plot::Point(j * alpha, _height - rowCounter, Color = _specialVertexColor));    
        else
          pointList := append(pointList, plot::Point(j * alpha, _height - rowCounter, Color = _vertexColor));   
        end_if;
      end_if;
      j := j + 1;
    end_for;
  end_if;

  // Now kill all the None't...
  tmpList := listlib::removeDuplicates(_vertices, KeepOrder);
  j := contains(_vertices, hold(None));
  if j > 0 then
    delete tmpList[j];
  end_if;
  _vertices := tmpList;

  edgeList := [];
  selfList := [];
  for i from 1 to numberEdges do
    edge := _edges[i];
    v := contains(_vertices, edge[1]); 
    w := contains(_vertices, edge[2]);
    if v = w then
      selfList := append(selfList, plot::Ellipse2d(pointList[v] * (1 + ((_vpl/2)/100)), ((_vpl/2)/10), ((_vpl/4)/10), Color = _edgeColor));
    else
      if (contains(_specialEdgeList, edge) > 0) then
        drawColor := _specialEdgeColor
      else
        drawColor := _edgeColor
      end_if;
      if _isDirected then
        edgeList := append(edgeList, plot::vector2d(pointList[v],pointList[w], PI/6.0, 0.05, Color = drawColor));
      else
        edgeList := append(edgeList, plot::line(pointList[v],pointList[w], Color = drawColor));
      end_if;
    end_if;
  end_for;

  plot::Scene(op(selfList), op(edgeList), op(pointList), Axes = None, PointWidth = _pointWidth,
              PointStyle = FilledCircles, LineWidth = 1);
end_proc:



/*
 plotTreeGraph - graphical plot where vertices are arranged in a circle of a Dom::Graph
 plotTreeGraph(G)
 Parameters:
   G - Dom::Graph
-*/
/*
Dom::Graph::plotTreeGraph := proc(G : Dom::Graph, startVertex : DOM_LIST)
local _pointWidth, _vertices, _edges, numberVertices, numberEdges, edgeList, pointList, n,
       height, xcoord, ycoord, shift, leftchild, rightchild, nextpos, modifier, i, lc, rc;
       
begin
  // Small procedure to check y - coordinates
  assycoor :=  proc(node, level)
  begin
    if node > -1 then
      ycoord[node] := level;
      if level > height then
        height := level;
        assycoor(leftchild[node] + 1);
        assycoor(rightchild[node] + 1);
      end_if;
    end_if;  
  end_proc;

  // Small procedure to check x - coordinates
  asspxcoor :=  proc(node)
  local leftchildexists, rightchildexists, hasonlyleftchild, hasonlyrightchild,
        isleaf, diff, place, h;
  begin
    leftchildexists := FALSE;
    if leftchild[node] > -1 then
      leftchildexists := TRUE;
      asspxcoor(leftchild[node]);
    end_if;

    rightchildexists := FALSE;
    if rightchild[node] > -1 then
      rightchildexists := TRUE;
      asspxcoor(rightchild[node]);
    end_if;

    hasonlyleftchild := leftchildexists and not rightchildexists;
    hasonlyrightchild := rightchildexists and not leftchildexists;

    ifleaf := not leftchildrenexists and not rightchildrenexists;

    h := ycoord[node];

    if isleaf then
      place := nextpos[h];
    elif hasonlyleftchild then
      place := xcoord[leftchild[node]] + i;
    elif hasonlyrightchild then
      place := xcoord[rightchild[node]] + i;
    else // it has two children
      place := (xcoord[leftchild[node]] + xcoord[rightchild[node]]) / 2
    end_if;

    diff := nextpos[h] - place;

    if diff > modifier[b] then
      modifier[h] := diff;    
    end_if;

    if isleaf then
      xcoord[node] := place;
    else
      xcoord[node] := place + modifier[h];
    end_if;
    
    nextpos[h] := xcoord[node] + 2;
    shift[node] := modifier[h];
  end_proc;

  // shifts the positions
  makeshift :=  proc(node, amount)
  begin
    xcoord[node] := amount;
    if leftchild[node] > -1 then
      makeshift(leftchild[node], amount + shift[node]);
    end_if;
    if rightchild[node] > -1 then
      makeshift(rightchild[node], amount + shift[node]);
    end_if;
  end_proc;

  printthetree :=  proc()
  local i;
  begin
    for i in Dom::Graph::getVertices(G) do
      userinfo(NoNL, 1, "Vertex " . expr2text(i) . " (" . leftchild[i] . "," . rightchild[i] . ") : x = " . xcoord[i] . ", y = " . ycoord[i] . ", shift = " . shift[i] . "\n");         
    end_for;
    userinfo(NoNL, 1, "Height: " . height);
  end_proc;

  printthemodifier :=  proc()
  local i;
  begin
    for i from 0 to height do
      userinfo(NoNL, 1, "Level " . expr2text(i) . ": " . modifier[i] . "\n");         
    end_for;
  end_proc;

  // Start the main procedure
  if testargs() then
     if (args(0) < 2) then
       error("Wrong number of arguments!");
     end_if;
  end_if;
  
  if args(0) > 2 then
      _pointWidth := args(3);
  else
      _pointWidth := 40;
  end_if;

  // Find Vertice with no indegree
  //_adjacentEnteringEdges := Dom::Graph::getAdjacentEnteringEdges;
  //for i in _adjacentEnteringVertices do
  //  if nops(i) = 0 then
  //    startVertex := _adjacentEnteringVertices[i];
  //  end_if;
  //end_for;  
  
  _vertices := Dom::Graph::getVertices(G);
  _edges := Dom::Graph::getEdges(G);
  numberVertices := nops(_vertices);
  numberEdges := nops(_edges);

  // assign y-coordinates
  height := 0;
  assycoor(op(startVertex), 0);
  height := height + 1;

  for i in _vertices do
    nextpos[i] := 0;
    modifier[i] := 0;
  end_for;

  // assign provisional x-coordinates and shifts
  assxcoor(op(startVertex));

  userinfo(NoNL, 1, "Tree after phase 1\n");
  printthetree();
  userinfo(NoNL, 1, "Modifier after phase 1\n");
  printthemodiefier();
  
  makeshift(op(startVertex), 0);
  userinfo(NoNL, 1, "Tree after phase 2\n");
  printthetree();

  //plot::Scene(op(edgeList), op(pointList), Axes = None, PointWidth = _pointWidth,
  //            PointStyle = FilledCircles, LineWidth = 1);
end_proc:
*/

/* plotDiGraph - proposal from Wolfgang Lindner (basically the same as the Circle but the vertices that are
                 not used with edges are not drawn)
   plotDiGraph(G)
   Parameters:
   G - Dom::Graph
*/
/*
Dom::Graph::plotDiGraph := proc(G : Dom::Graph)
local _isOriented, niceVec, i, v, w, pointList, selfList, edgeList, numpoints, _edges, edge, _isDirected;

begin
   if testargs() then
      if (args(0) < 1) or (args(0) > 2) then
      error("Wrong number of arguments!");
      end_if;
      Dom::Graph::graphCheck(G);
   end_if;

  _isOriented := hold(positive);
  _edges := Dom::Graph::getEdges(G);

  if args(0) = 2 then
    if lhs(args(2)) = hold(Orientation) then
      _isOriented := rhs(args(2));
    end_if:
  end_if;

  numpoints := Dom::Graph::getVertexNumbers(Dom::Graph::getVertices(G));

  if _isOriented = hold(positive) then
    pointList := (plot::Point(sin(i/numpoints*2*float(PI)),
      cos(i/numpoints*2*float(PI)), PointWidth = 10, Color = RGB::Red, PointStyle = FilledCircles) $ i = 0..numpoints);
  elif _isOriented = hold(negative) then
    pointList := (plot::Point(-sin(i/numpoints*2*float(PI)),
      cos(i/numpoints*2*float(PI)), PointWidth = 10, Color = RGB::Red, PointStyle = FilledCircles) $ i = 0..numpoints);
  else
    error("Illegale Orientierung, erwarte 'Oriented = positive' oder 'Oriented = negative'");
  end_if:

  if Dom::Graph::isDirected(G) then 
    //definiere die lokale Funktion niceVec als  Vektor mit Spitze oder als simple Linie    
    niceVec := (p1, p2) -> plot::vector(p1, p2, 10*3.14/180, 0.1, args(3..args(0)));
  else
    niceVec := (p1, p2) -> plot::line(p1, p2, args(3..args(0)));
  end_if:

  _edges := Dom::Graph::getEdges(G);

  edgeList := [];
  selfList := [];
  _isDirected := Dom::Graph::isDirected(G); // otherwise the function would have to be called n-times during the for loop.
  for i from 1 to nops(_edges) do
    edge := _edges[i];
    v := edge[1]; 
    w := edge[2];
    if v = w then
      selfList := append(selfList, plot::Ellipse2d(pointList[v] * 1.05, 0.05, 0.05));
    end;
    if not v = w then
      if _isDirected then
          edgeList := append(edgeList, niceVec(pointList[v], pointList[w], Color = RGB::Black));
          //edgeList := append(edgeList, plot::vector(pointList[v],pointList[w], PI/6.0, 0.05));
        else          
          edgeList := append(edgeList, plot::line(pointList[v],pointList[w]));
          //edgeList := append(edgeList, plot::line(pointList[v],pointList[w]));
        end_if;
      end_if;
   end_for;
   plot::Scene(op(edgeList), op(pointList), op(selfList), Axes = None, PointWidth = 40,
              PointStyle = FilledSquares, LineWidth = 1);
//  plot::Scene(plot::Group(pointList, selfList op(edgeList, 1..nops(edgeList))), Axes = None, Scaling = Constrained);
end_proc:
*/

/* checkForVertices: returns an empty list if all vertices from all edges exist in the vertice-list, 
                     the list with the wrongly used vertices otherwise.
   checkForVertices(Edge, Vertex)
   checkForVertices([e1..en], [v1..vn])

   Parameters:
   Edge - List of one or more edges
   Vertex - List of one or more vertices
*/
Dom::Graph::checkForVertices := proc(Edge : Type::ListOf(DOM_LIST), Vertex : DOM_LIST)
local tmpList, _newVertices, _existingVertices, _dummy;
begin

  if nops(Edge) = 0 then
    return ([]);
  end_if;

  // Get all vertices in one list
  tmpList:= op({_concat(op(Edge))});

  [_existingVertices, _newVertices,  _dummy] := split(tmpList, v -> contains(Vertex, v) > 0);

  if nops(_newVertices) > 0 then
    return (_newVertices);
  else
    return ([]);
  end_if;
end_proc:

/* revertGraph: revert all edges of a network (makes only sense for a directed graph !)
   revertGraph(G)

   Parameters:
   G - Dom::Graph
*/

Dom::Graph::revertGraph := proc(G : Dom::Graph)
  local _vertices, _edges, _edgeWeights, _edgeCosts, _edgeDescriptions, G2;
begin

  _vertices := Dom::Graph::getVertices(G);
  _edges := map(Dom::Graph::getEdges(G), revert);

  if Dom::Graph::isDirected(G) then
    G2 := Dom::Graph::new(_vertices, _edges, hold(Directed));
  else  
    return(G);  // because there is nothing to revert !
  end_if;
  
  if not Dom::Graph::getVertexWeights(G) = FAIL then
    G2 := Dom::Graph::setVertexWeights(G2, _vertices, Dom::Graph::getVertexWeights(G));
  end_if;

  if not Dom::Graph::getEdgeWeights(G) = FAIL then
    _edgeWeights := table(op(map([op(Dom::Graph::getEdgeWeights(G))], ew -> revert(lhs(ew))=rhs(ew))));
    G2 := Dom::Graph::setEdgeWeights(G2, _edges, _edgeWeights);
  end_if;

  if not Dom::Graph::getEdgeCosts(G) = FAIL then
    _edgeCosts := table(op(map([op(Dom::Graph::getEdgeCosts(G))], ec -> revert(lhs(ec))=rhs(ec))));
    G2 := Dom::Graph::setEdgeCosts(G2, _edges, _edgeCosts);
  end_if;

  if not Dom::Graph::getEdgeDescriptions(G) = FAIL then
    _edgeDescriptions := table(op(map([op(Dom::Graph::getEdgeDescriptions(G))], ec -> revert(lhs(ec))=rhs(ec))));
    G2 := Dom::Graph::setEdgeDescriptions(G2, _edges, _edgeDescriptions);
  end_if;

  G2;
end_proc:

/*
Dom::Graph::chromaticPolynomial(G, x)

Parameters:
G - Dom::Graph
x - Identifier

computes the chromatic polynomial of the graph G; 
x is used as the indeterminate of the polynomial.
G must be undirected, i._edge., for every edge [a, b], the edge [b, a] must
exist, too.
*/

Dom::Graph::chromaticPolynomial:= proc(G: Dom::Graph, x: DOM_IDENT): DOM_POLY
  local
  equ: "_equal", 
  delEdge: DOM_PROC,
  joinVertices: DOM_PROC,
  crpoly: DOM_PROC;
begin

  if Dom::Graph::isDirected(G) then
    error("Graph must be undirected!");
  end_if;

  // local method delVertex
  //
  // T - adjacency table
  // a - vertex
  // b - vertex
  //
  // returns a new graph T, with the edge between a and b removed
  
  delEdge:= proc(T: DOM_TABLE, a, b): DOM_TABLE
  begin
    T[a]:= subs(T[a], b = null());
    T[b]:= subs(T[b], a = null());
    T
  end_proc;

  // local method joinVertices
  //
  // T - adjacency table
  // a - vertex
  // b - vertex
  //
  // returns the equivalence graph, with a and b now being identified
  // the edge between a and b is *not* turned into a loop
  // a is used as name for the vertex ab
  
  joinVertices:= proc(T: DOM_TABLE, a, b)
  begin
    // join adjacency lists, remove duplicates
    T[a]:= [op({op(T[a].T[b])} minus {a, b})];
    delete T[b];
    subs(T, b=a);
  end_proc;
  
  // local method crpoly
  // 
  // crpoly(T)
  //
  // T - adjacency table

  //
  // returns chromatic polynomial of the graph reprsented
  // by T, an element of Z[x], where x is the arg. of the outer procedure

  crpoly :=  proc(T: DOM_TABLE): DOM_POLY
    local equ, a, b;
  begin
    if nops(T) = 1 then
      return(poly(x))
    end_if;

    equ:= op(T, 1);
    // now equ is an equation vertex = [adjacent vertices]
    a:= op(equ, 1);
    
    if op(equ, 2) = [] then
      // isolated vertex: remove
      delete T[a];
      poly(x) * crpoly(T)
    else
      // simply continue with first edge
      b:= op(equ, [2, 1]);
      crpoly(delEdge(T, a, b)) - crpoly(joinVertices(T, a, b))
    end_if;
      
  end_proc;

  for equ in Dom::Graph::getEdgesLeaving(G) do
    if contains(op(equ,2), op(equ, 1)) > 0 then
      // the graph G contains a loop and cannot be coloured
      return(poly(0, [x]))
    end_if
  end_for;
  
  crpoly(Dom::Graph::getEdgesLeaving(G))
  
end_proc:

/*
  getAdjacentEdgesEntering(Vertex)
  Returns the vertices to which the edges lead from the given vertex
  Parameters:
  G      - Dom::Graph
  Vertex - The vertex which entering edges should be returned
*/
Dom::Graph::getAdjacentEdgesEntering := proc(G : Dom::Graph, Vertex: DOM_LIST)
local _edgesEntering;
begin
  if nops(Vertex) > 1 then
  	 error("Only one specified vertex allowed.");
  end_if;
  _edgesEntering := Dom::Graph::getEdgesEntering(G);
  _edgesEntering[op(Vertex)];
end_proc:

/*
  getAdjacentEdgesLeaving(Vertex)
  Parameters:
  G      - Dom::Graph
  Vertex - The vertex which leaving edges should be returned
*/
Dom::Graph::getAdjacentEdgesLeaving := proc(G : Dom::Graph, Vertex: DOM_LIST)
local _edgesLeaving;
begin
  if nops(Vertex) > 1 then
  	 error("Only one specified vertex allowed.");
  end_if;
  _edgesLeaving := Dom::Graph::getEdgesLeaving(G);
  _edgesLeaving[op(Vertex)];
end_proc:

/*
 depthFirstSearch(G <, StartVertex = v> <, VertexList = [v1,...,vn]>)
 Parameters:
 G           - Graph
 StartVertex - a vertex from which to start
 VertexList  - if the List is given, the order of this list is used (for SCC).
*/
Dom::Graph::depthFirstSearch := proc(G : Dom::Graph)
local timestamp, _color, _vertices, v, i, _startVertex, _startTime, 
      _finishTime, _parent, dfsVisit;
begin

  dfsVisit :=  proc(_vertex)
  local _adjacentEdges, v2;
  begin
    timestamp := timestamp + 1;
    _color[_vertex] := 1; // gray
    _startTime[_vertex] := timestamp;

    _adjacentEdges := [op(Dom::Graph::getAdjacentEdgesLeaving(G, [_vertex]))];
    if (nops(_adjacentEdges) > 0) and (_parent[_vertex] = infinity) then
      _parent[_vertex] := _vertex;
    end_if;
    for v2 in _adjacentEdges do
      if _color[v2] = 0 then
        _parent[v2] := _vertex;
        dfsVisit(v2);
      end_if;       
    end_for;
    _color[_vertex] := 2;
    timestamp := timestamp + 1;
    _finishTime[_vertex] := timestamp;
  end_proc;

  _vertices := Dom::Graph::getVertices(G);
  _startVertex := FAIL;
  // For empty graphs:
  _parent := table();
  // here it goes on....

  if args(0) > 1 then    
    for i from 2 to args(0) do
      case op(args(i),1)
        of hold(StartVertex) do
          _startVertex := op(op(args(i),2));
          break;
        of hold(VertexList) do
          _vertices := op(args(i),2);
          break;
        otherwise
          error("Wrong argument definition!");
      end_case;
    end_for;
  end_if;
  
  if _startVertex = FAIL then
    if nops(_vertices) > 0 then
      _startVertex := op(_vertices,1);    	
    else
      return ([table(), table(), table()])
    end_if;
  end_if;

  timestamp := 0;
  // Initialize with visiting-time 0
  for v in _vertices do
    _parent[v] := infinity;
    _color[v] := 0;
  end_for;

  dfsVisit(op(_startVertex));
  // If the graph was not connected there are still vertices not visited !
  for v in _vertices do
    if _color[v] = 0 then
      dfsVisit(op(v));
    end_if;
  end_for;

  for v in _startTime do
    userinfo(NoNL, 1, "Vertex " . expr2text(lhs(v)) . " is found at timestamp " . expr2text(rhs(v)) . "\n" );    
  end_for;

  [_startTime, _finishTime, _parent];
end_proc:

/*
 breadthFirstSearch(G <,startVertex>)
 Parameters:
 G           - Graph
 startVertex - a vertex from which to start
*/
Dom::Graph::breadthFirstSearch := proc(G : Dom::Graph)
local timestamp, _color, _level, _vertices, v, _startVertex, _startTime, _levelTable, _parent, bfsVisit;
begin

  bfsVisit :=  proc(_vertex)
    local _adjacentEdges, visitedList, lastVertex;
  begin
    visitedList := [_vertex];
    lastVertex := _vertex; // because it is deleted immediately (and thus can do no harm).
    timestamp := timestamp + 1;
    _startTime[_vertex] := timestamp;
    _levelTable[_vertex] := _level;
    _level := _level + 1;
    repeat      
      _color[_vertex] := 1;
      delete visitedList[contains(visitedList,_vertex)];
      _adjacentEdges := Dom::Graph::getAdjacentEdgesLeaving(G, [_vertex]);
      if (nops(_adjacentEdges) > 0) and (_parent[_vertex] = infinity) then
        _parent[_vertex] := _vertex;
      end_if;
      if (nops(_adjacentEdges) > 0) and contains(visitedList, lastVertex) = 0 then
        lastVertex := _adjacentEdges[ nops(_adjacentEdges) ];
      end_if;
      while not nops(_adjacentEdges) = 0 do
        if _color[op(_adjacentEdges,1)] = 0 then
          timestamp := timestamp + 1;
          _startTime[op(_adjacentEdges,1)] := timestamp;
          _levelTable[op(_adjacentEdges,1)] := _level;
          _parent[op(_adjacentEdges,1)] := _vertex;            
          visitedList := append(visitedList, op(_adjacentEdges,1));
        end_if;
        if _adjacentEdges[1] = lastVertex then
         _level := _level + 1;
        end_if;
        delete _adjacentEdges[1];
      end_while;
      _color[_vertex] := 2;
      _vertex := op(visitedList, 1);
    until nops(visitedList) = 0 end_repeat;
  end_proc;

  _vertices := Dom::Graph::getVertices(G);
  // For empty graphs:
  _startTime := table();
  _levelTable := table();
  _parent := table();
  // here it goes on....

  timestamp := 0;
  // Initialize with visiting-time 0
  for v in _vertices do
    _parent[v] := infinity;
    _color[v] := 0;
  end_for;

  _level := 0;
  if args(0) = 2 then
    if op(args(2),1) = hold(StartVertex) then
      _startVertex := op(op(args(2),2));
    else
      error("Wrong argument : " . expr2text(op(args(2),1)) .".");
    end_if;
    bfsVisit(op(_startVertex));
  end_if;

  // timestamp was set so complicated, because there might exist several forests.
  for v in _vertices do
    if _color[v] = 0 then
      _level := 0;
      bfsVisit(op(v));
    end_if;
  end_for;

  for v in _startTime do
    userinfo(NoNL, 1, "Vertex " . expr2text(lhs(v)) . " is found at timestamp " . expr2text(rhs(v)) . "\n" );    
  end_for;

  [_startTime, _levelTable, _parent];
end_proc:

/*
 biConnectedComponents(G <,startVertex>)
 returns [the biconnected components of the graph, bridges, singleVertices).
 
 Parameters:
 G           - Graph
 StartVertex - a vertex from which to start
 Detailed    - gives detailed information (articulation points, connected components)
*/
Dom::Graph::biConnectedComponents := proc(G : Dom::Graph)
local timestamp, _color, _vertices, v, i, _startVertex, _counter, _directed, createSCC, _high, _articulationPoints,
      _scc, detailed, _startTimes, _finishTimes, _parents, nrVertices,
      bfsVisitUndirected, dfsVisitDirected, findBiconnectedComponents;
begin

 /*
   Inner method for finding the articulation points for undirected graphs
  */
  findBiconnectedComponents :=  proc(_vertex)
  local _adjacentEdges, v;
  begin
    _color[_vertex] := nrVertices;
    nrVertices := nrVertices - 1;
    _high[_vertex] := _color[_vertex];
    _adjacentEdges := op(Dom::Graph::getAdjacentEdgesLeaving(G, [_vertex]));
    for v in _adjacentEdges do
      if not _parents[_vertex] = v then
        if _color[v] = 0 then
          _parents[v] := _vertex;
          findBiconnectedComponents(v);
          if _high[v] <= _color[_vertex] then
            if not _parents[_vertex] = infinity then
              _articulationPoints := append(_articulationPoints, _vertex);              
            end_if;
          end_if;
          if _high[_vertex] < _color[v] then
            _high[_vertex] := _color[v];
          end_if;
        else // (_vertex, v) is a back edge or a forward edge
          if _high[_vertex] < _color[v] then
            _high[_vertex] := _color[v];
          end_if;
        end_if;
      end_if;
    end_for;
  end_proc;  

  /*
    creates the return value
  */
  createSCC := proc(showDetails : DOM_BOOL)
  local u, v, v2, i,j, _edges, bridges, components, forefathers, nrComponents, 
        G2, tempVertices, ready, count, bridgeList;
  begin
    G2 := [];

    if _directed then
      nrComponents := 0;
      forefathers := [];
      components := table();      
      // Now put the information together...
      for v in _parents do
        // Check for singleVertex
        if op(v,2) = infinity or op(v,1) = op(v,2) then
          // Is it a forefather?
          if not op(v,2) = v then
            forefathers := append(forefathers,op(v,1));          
          end_if;
          nrComponents := nrComponents + 1;
        else
          components[op(v,1)] := op(v,2);
        end_if;
      end_for;
      // and use them to create the graphs...
      // for one vertex from each component
      for u in forefathers do
        tempVertices := [u];
        ready := FALSE;
        count := 0;
        // take the much smaller table to search for the vertices of each component
        while not ready do
          for v in components do
            // u is the forefather of the edge found. So take the vertex and try to find its predecessor.
            count := count + 1;
            if op(v,2) = u then
              u := op(v,1);
              tempVertices := append(tempVertices, u); // add the predecessor and make it new forefather.
              count := 0; // Just to make sure !
              break; // make it faster (the rest does not need to be checked.)
            end_if;
          end_for;
          // if v ran through without any more forefather the set is full with each member of the component
          if count = nops(components) then
            ready := TRUE;            
          end_if;
        end_while;
        // Now all vertices from one component are stored in tempVertices.
        // Add the graph to G2 (this may take time !).
        G2 := append(G2, Dom::Graph::getSubGraph(G, tempVertices));
      end_for;
    else
      nrComponents := 0;
      while contains(_scc, nrComponents) do
        G2 := append(G2, Dom::Graph::getSubGraph(G, _scc[nrComponents]));
        nrComponents := nrComponents + 1;
      end_while;
    end_if;
   
    if showDetails then
      if _directed then
     // Now get the bridges (and the articulation points...)
        v := [];
        // Takes O(1) because its just an extract !
        for u in G2 do
          v := append(v, op(Dom::Graph::getEdges(u)));
        end_for;
        v := {op(v)};
        u := {op(Dom::Graph::getEdges(G))};
        bridges := [op(u minus v)];
        for j from 1 to nops(bridges) do
          bridges[j] := sort(bridges[j]);
        end_for;
        bridges := sort(bridges, (x, y) -> sysorder(op(x,1), op(y,1)));

        _scc := [G2, bridges];        
      else
        bridgeList := [];
        for u in G2 do
          v2 := Dom::Graph::getVertices(u);
          _edges := Dom::Graph::getEdges(u);

          // Initialize again !
          nrVertices := nops(v2);
          for v in _vertices do
            _color[v] := 0;
            _parents[v] := infinity;
          end_for;
          delete (_startTimes);
          timestamp := 0;
          _articulationPoints := [];
          // Because the starting vertex can make trouble it has to be run twice !
          // once with the first vertex and once with the last.
          findBiconnectedComponents(op(v2,1),nrVertices);
          // Remove double vertices
          _articulationPoints := [op({op(_articulationPoints)})];
          // Store _articulationPoints in i
          i := {op(_articulationPoints)};

          // And initializing again...
          nrVertices := nops(v2);
          for v in _vertices do
            _color[v] := 0;
            _parents[v] := infinity;
          end_for;
          delete (_startTimes);
          
          timestamp := 0;
          _articulationPoints := [];
          findBiconnectedComponents(op(v2,nops(v2)),nrVertices);
          // Remove double vertices
          _articulationPoints := [op({op(_articulationPoints)})];

          // Store _articulationPoints in j
          j := {op(_articulationPoints)};
 
          // Now put them together and ALL bridges are found :)         
          _articulationPoints := [op(i union j)];
          bridges := [];
          // Now get the edges that are bridges...
          for i in _articulationPoints do
            for j in _articulationPoints do
              // Does edge [i, j] exist ?
              if contains(_edges, [i,j]) > 0 then
                bridges := append(bridges, [i, j]);
              end_if;
            end_for;
          end_for;
          if nops(bridges) = 0 then
            bridgeList := append(bridgeList, []);   // empty list for checking.
          else
            for j from 1 to nops(bridgeList) do
              bridgeList[j] := sort(bridgeList[j]);
            end_for;
            bridgeList := sort(bridgeList, (x, y) -> sysorder(op(x,1), op(y,1)));
            bridgeList := sort(bridgeList);
            bridgeList := append(bridgeList, op(bridges));
          end_if;
        end_for;
        _scc := [G2, bridgeList];
      end_if;
    else
      _scc := G2;
    end_if;
  end_proc;

 /*
   Inner method for finding the connected components in an undirected graph
  */
  bfsVisitUndirected :=  proc(_vertex, counter)
  local _adjacentEdges, v, tmpList;
  begin
    if contains(_scc, counter) then
      tmpList := _scc[counter];
      tmpList := append(tmpList, _vertex);
      _scc[counter] := tmpList;
    else      
      _scc[counter] := [_vertex];
    end_if;
    _color[_vertex] := 1;
    timestamp := timestamp + 1;
    // Assign timestamp to _startTimes and minG
    _startTimes[_vertex] := timestamp;
    _adjacentEdges := op(Dom::Graph::getAdjacentEdgesLeaving(G, [_vertex]));
    if nops(_adjacentEdges) = 0 then
      if nops(op(Dom::Graph::getAdjacentEdgesEntering(G, [_vertex]))) = 0 then
        _parents[_vertex] := infinity;
      end_if;
    end_if;
    for v in _adjacentEdges do
      if _color[v] = 0 then
        _parents[v] := _vertex;
        bfsVisitUndirected(v, counter);
      end_if;
    end_for;
    _color[_vertex] := 2;
    timestamp := timestamp + 1;
    _finishTimes[_vertex] := timestamp;
  end_proc;

  dfsVisitDirected := proc(_vertex)
  local firstDFS, secondDFS, revertG, _vertices;
  begin
    firstDFS := Dom::Graph::depthFirstSearch(G, hold(StartVertex)=_vertex);
    // Now sort vertices after finishing time
    _vertices := [op(op(firstDFS,2))];
    _vertices := map(sort([op(op(firstDFS,2))], () -> bool(rhs(args(1)) > rhs(args(2)))), () -> lhs(args(1)));
    
    revertG := Dom::Graph::revertGraph(G);
    secondDFS := Dom::Graph::depthFirstSearch(revertG, hold(StartVertex)=op(_vertices,1), hold(VertexList) = _vertices);
    _startTimes := op(secondDFS, 1);
    _finishTimes := op(secondDFS, 2);
    _parents := op(secondDFS, 3);
    createSCC;     
  end_proc;

  _vertices := {op(Dom::Graph::getVertices(G))};
  _directed := Dom::Graph::isDirected(G);
  timestamp := 0;
  _counter := 0;
  _scc := table();
  // Initialize with visiting-time 0
  for v in _vertices do
    _color[v] := 0;
    _parents[v] := infinity;
  end_for;

  _startVertex := FAIL;
  detailed := FALSE;
 
  if args(0) > 1 then
    for i from 2 to args(0) do
      if args(i) = hold(Detailed) then
        detailed := TRUE;
      elif op(args(i),1) = hold(StartVertex) then
        _startVertex := op(op(args(i),2));
      else
        error("Wrong argument : " . expr2text(op(args(i),1)));
      end_if;
    end_for;    
  end_if;

  if _startVertex = FAIL then
    _startVertex := op(_vertices,1);
  end_if;

  if _directed then
    dfsVisitDirected(op(_startVertex));
  else
    _parents[op(_startVertex)] := op(_startVertex);
    bfsVisitUndirected(op(_startVertex), _counter);
  end_if;

  for v in _vertices do
    if _color[v] = 0 then
      _counter := _counter + 1;
      if _directed then
        dfsVisitDirected(op(v));
      else
        _parents[op(v)] := op(v);
        bfsVisitUndirected(op(v), _counter);
      end_if;
    end_if;
  end_for;

  createSCC(detailed);

  _scc;
end_proc:


/* createMinSort - creates a List that is sorted either with Costs or Weights in ascending order.
Dom::Graph::createMinSort := proc(T : DOM_TABLE)
Parameters:
 Weights|Costs - extracts 
*/
Dom::Graph::createMinSort := proc(T : DOM_TABLE)
begin
  // Return a sorted list of edges
  // Example:
  // table([a,d]=4,[a,b]=3,[a,c]=2,[a,a]=1)
  // returns
  // [[a, a], [a, c], [a, b], [a, d]]
  //sort([op(T)], () -> bool(rhs(args(1)) < rhs(args(2))));
  map(sort([op(T)], () -> bool(rhs(args(1)) < rhs(args(2)))), () -> lhs(args(1)));
end_proc:

/*
 minimumSpanningTree - returns a new Graph that only consist of the minimum spanning edges.
 minimumSpanningTree(G <,SearchFor = Weights|Costs> < , ReturnAsTable>) 
 Paramters:
 G             - Graph
 Weights|Costs - defines wether to use the weight or the costs of the edges. Default = Weights
 ReturnAsTable - defines wether the output will be a new Graph or a tabular form. Default = Graph
*/
Dom::Graph::minimumSpanningTree := proc(G : Dom::Graph)
local _arguments, searchFor, output, leftOperator, rightOperator, i, j; //searchWith

begin

 // Check for connectivity
 if not Dom::Graph::isConnected(G) then
 	 error("Graph is not connected. Can not create a minimum spanning tree.");
 end_if;

 //  searchWith := hold(Kruskal);
  searchFor := hold(Weights);
  output := hold(Graph);  

  _arguments := args();
  j := args(0);

  for i from 2 to j do
    if _arguments[i] = hold(ReturnAsTable) then
      output := hold(Table);
      delete _arguments[i];
      j := j - 1;
      break;
    end_if;
  end_for;

  for i from 2 to j do
    leftOperator := op(_arguments[i],1);
    rightOperator := op(_arguments[i],2);
    case leftOperator
//      of hold(SearchWith)
//        if rightOperator = hold(Prim) then
//          searchFor := hold(Prim);  
//        elif rightOperator = hold(Kruskal) then
//        else
//          error("Wrong argument type for " . leftOperator . ": Kruskal or Prim expected!");
//       end_if;
//      of hold(StartVertex) do
//        break;
      of hold(SearchFor) do
        if rightOperator = hold(Costs) then
          searchFor := hold(Costs);  
        elif rightOperator = hold(Weights) then
        else
          error("Wrong argument type for " . leftOperator . ": Weights or Costs expected!");
        end_if;
        break;
      otherwise
        error("Wrong argument definition!");
    end_case;
  end_for;

  // If empty, return empty graph
  if Dom::Graph::isEmpty(G) then
    if output=hold(Table) then
      return (table());
    else
      return (G);
    end_if;
  end_if;

  if output = hold(Graph) then
    Dom::Graph::mst_Kruskal(G, searchFor);
  else
    Dom::Graph::mst_Kruskal(G, searchFor, hold(ReturnAsTable));  	
  end_if;
end_proc:

/*
 mst_Kruskal - returns a new Graph or table that only consist of the minimum spanning edges.
 minimumSpanningTree(G <,Weights|Costs> <,Graph|Tables>)
 Paramters:
 G             - Graph
 Weights|Costs - defines wether to use the weight or the costs of the edges. Default = Weights
 Graph|Tables  - defines wehter the output will be a new Graph or a tabular form. Default = Graph
*/
Dom::Graph::mst_Kruskal := proc(G : Dom::Graph)
local _vertices, _edgeNumbers, C, u, v, parentTable, rankTable, i, _edges, _kruskal, sumOfMSP, output,
      makeSet, unionSet, linkSet, findSet,  // For the set.
      _edgeWeights, _edgeCosts, _edgeDescriptions, _edgesEntering, _edgesLeaving, dummyTable, _isDirected;
     

begin
  _edgeNumbers := Dom::Graph::getEdgeWeights(G);
  output := hold(Graph);

  if args(0) > 1 then
    for i from 2 to args(0) do
      case args(i)
        of hold(Costs) do
          _edgeNumbers := Dom::Graph::getEdgeCosts(G);
          break;
        of hold(Weights) do
          break;
        of hold(Graph) do
          break;
        of hold(ReturnAsTable) do
          output := hold(Tables);
          break;
        otherwise
          error("Wrong argument definition.");
      end_case;     
    end_for;
  end_if;

  _edges := Dom::Graph::getEdges(G);

  // If the graph is unweighted
  if _edgeNumbers = FAIL then
    for i in _edges do
      _edgeNumbers[i] := 1;
    end_for;
  end_if;

  // There are no edges
  if nops(_edges) = 0 then
    error("No edges exist to create a minimum spanning tree.");
  end_if;

  // local method makeSet
  // v - vertex
  // creates the connection to the parent vertex and sets its rank to 0.
  makeSet:= proc(v)
  begin
    parentTable[v] := v;
    rankTable[v] := 0;
  end_proc;

  // local method unionSet
  // u,v - vertex
  // creates the connection to the parent vertex and sets its rank to 0.
  unionSet := proc(u, v)
  begin
    linkSet(findSet(u), findSet(v));
  end_proc;

  // local method linkSet
  // u,v - vertex
  // sets the rank of the better vertex higher.
  linkSet := proc(u, v)
  begin
    if rankTable[u] > rankTable[v] then
      parentTable[v] := u;
    else
      parentTable[u] := findSet(parentTable[v]);
      if rankTable[u] = rankTable[v] then
        rankTable[v] := rankTable[v] + 1;
      end_if;
    end_if;
  end_proc;

  // local method findSet
  // v - vertex
  // finds a given vertex if it exist.
  findSet := proc(v)
  begin
    if (not v = parentTable[v]) then
      parentTable[v] := findSet(parentTable[v]);
    end_if;
    parentTable[v];
  end_proc;

  _isDirected := Dom::Graph::isDirected(G);
  _vertices := Dom::Graph::getVertices(G);
  parentTable := table();
  rankTable := table();
  
  // Now a sorting of edges with the predefined order is created.
  C := (nops(_edgeNumbers)+1) * max(op(_edgeNumbers, [1..nops(_edgeNumbers),2]));
  // If edges have no weight, weight 1 is assumed for every edge.
  if C = FAIL then
    C := (nops(_edgeNumbers)+1);
  end_if;
  sumOfMSP := 0;

  for u in _vertices do
    makeSet(u);
  end_for; 

  /* Since only the edge values are interesting, reverted edges have to be removed... */
  //if (not _isDirected) then
  //  for i from 1 to nops(_edges) do
  //    _edges[i] := sort(_edges[i]);
  //  end_for;
  //  _edges := sort(_edges, (x, y) -> (if op(x,1) = op(y,1) then	sysorder(op(x,2), op(y,2)); else sysorder(op(x,1), op(y,1)); end_if;));
    //_edges := sort(_edges);
  //  _edges := listlib::removeDuplicates(_edges, KeepOrder);   
  //end_if;

  _edges := Dom::Graph::createMinSort(_edgeNumbers);

  for i from 1 to nops(_edges) do
    u := op(_edges[i],1);
    v := op(_edges[i],2);
    if not (findSet(u) = findSet(v)) then      
      unionSet(u, v);
      if not Dom::Graph::isDirected(G) and output = hold(Tables) then
        _kruskal[[v, u]] := _edgeNumbers[[v, u]];	
      end_if;
      _kruskal[[u, v]] := _edgeNumbers[[u,v]];
      sumOfMSP := sumOfMSP + _edgeNumbers[[u,v]];
    end_if;
  end_for;
  
  if output = hold(Tables) then
    [_kruskal, sumOfMSP];
  else
    _edges := [];
    dummyTable := op(_kruskal);
    for i in dummyTable do
      _edges := append(_edges,op(i,1));     
    end_for;
    // _edges consist only of |V|-1 edges !!!
    _edges := sort(_edges, (x, y) -> (if op(x,1) = op(y,1) then	sysorder(op(x,2), op(y,2)); else sysorder(op(x,1), op(y,1)); end_if;));

    dummyTable := Dom::Graph::getEdgeCosts(G);
    if (not dummyTable = FAIL)  then
      for i in _edges do
        _edgeCosts[i] := dummyTable[i];
        if not _isDirected then
          _edgeCosts[revert(i)] := dummyTable[i];        	
        end_if;
      end_for;
    else
      _edgeCosts := FAIL;
    end_if;

    dummyTable := Dom::Graph::getEdgeWeights(G);
    if (not dummyTable = FAIL)  then
      for i in _edges do
        _edgeWeights[i] := dummyTable[i];
        if not _isDirected then
          _edgeWeights[revert(i)] := dummyTable[i];        	
        end_if;
      end_for;
    else
      _edgeWeights := FAIL;
    end_if;

    dummyTable := Dom::Graph::getEdgeDescriptions(G);
    if (not dummyTable = FAIL)  then
      for i in _edges do
        _edgeDescriptions[i] := dummyTable[i];
        if not _isDirected then
          _edgeDescriptions[revert(i)] := dummyTable[i];        	
        end_if;
      end_for;
    else
      _edgeDescriptions := FAIL;
    end_if;

    if not _isDirected then
      _edges := _edges . map(_edges, revert);    	
    end_if;

    _edgesEntering := table((_vertices[i]=sort( map(select(_edges, _edge->op(_edge,2)=_vertices[i]), op, 1) ) $ i=1..nops(_vertices))):
    // Speedup:
    if _isDirected then
      _edgesLeaving  := table((_vertices[i]=sort( map(select(_edges, _edge->op(_edge,1)=_vertices[i]), op, 2) ) $ i=1..nops(_vertices))):
    else
      _edgesLeaving := _edgesEntering;
    end_if;

    subsop(G, 2=_edges, 4=_edgeDescriptions, 5=_edgeWeights, 6=_edgeCosts, 7=_edgesEntering, 8=_edgesLeaving);
  end_if;

end_proc:


/*
  getBestAdjacentEdge(G, Vertex, Vertices <, Weights|Costs> <, Min|Max> )
  returns the shortest/longest way from the given vertex to any of the adjacent ones. 
  If no adjacent edge exist, FAIL will be returned
  Parameters:
  G        - Graph
  Vertex   - The vertex which neighbors have to be searched.
  Vertices - The vertices that are adjacent and can be used.
  Min/Max  - Either the minimum or the maximum
  Costs/Weights - Depending either on costs or weights. 
*/
Dom::Graph::getBestAdjacentEdge := proc(G : Dom::Graph, Vertex : DOM_LIST, Vertices : DOM_LIST)
local _adjacent, searchWay, searchWith, _vertex, best, i, _edgeNumbers, retVertex;
begin

  // Store the vertices Vertex can reach.
  _adjacent := Dom::Graph::getAdjacentEdgesLeaving(G, Vertex);
  searchWay := hold(Min);
  searchWith := hold(Weight);
  best := FAIL;
  retVertex := FAIL;

  if nops(Vertices) = 0 then
    return (best);
  end_if;

  for i from 4 to args(0) do
    case args(i)
      of hold(Min) do        
        break;
      of hold(Max) do
        searchWay := hold(Max);
        break;
      of hold(Costs) do
        searchWith := hold(Costs);
        break;
      of hold(Weights) do        
        break;
      otherwise
        error("Wrong argument definition!");
    end_case;
  end_for;  

  Vertex := op(Vertex);

  if (searchWith = hold(Weight)) then
    _edgeNumbers := Dom::Graph::getEdgeWeights(G);
  else
    _edgeNumbers := Dom::Graph::getEdgeCosts(G);
  end_if;

  for _vertex in _adjacent do
    if (searchWay = hold(Min)) then
      if (contains(Vertices, _vertex) > 0) then
        if not best = FAIL then
          if (best > _edgeNumbers[[Vertex, _vertex]]) then
            best := _edgeNumbers[[Vertex, _vertex]];
            retVertex := _vertex;
          end_if;           
        else
          best := _edgeNumbers[[Vertex, _vertex]];
          retVertex := _vertex;
        end_if;
      end_if;
    else
      if (contains(Vertices, _vertex) > 0) then
        if not best = FAIL then
          if (best < _edgeNumbers[[Vertex, _vertex]]) then
            best := _edgeNumbers[[Vertex, _vertex]];
            retVertex := _vertex;
          end_if;           
        else
          best := _edgeNumbers[[Vertex, _vertex]];
          retVertex := _vertex;
        end_if;
      end_if;
    end_if;
  end_for;
  
  retVertex;  
end_proc:

/*
  bipartite(G, startVertex)
  finds out if a graph is bipartite or not.
  (Algorithm taken from Jungnickel / Graphen, Netzwerke und Algorithmen p.346)

  Parameters:
  G           - Graph
  Lists/Bool  - returns either two lists in a list (or FAIL) with the two sets of vertices or just TRUE/FALSE. (Default = Bool) 
*/
Dom::Graph::bipartite := proc(G : Dom::Graph)
local visitedSet, partOne, partTwo, bipart, _color, vertex, v, _adjacentEdges, output, 
      _vertices, toSearch, setToDo, setOne, setTwo;
begin
  output := hold(Bool);
  if args(0) > 1 then
    case args(2)
      of hold(Bool) do
        break;
      of hold(Lists) do
        output := hold(Lists);
        break;
      otherwise
        error("Wrong argument definition.");
    end_case;     
  end_if;

  bipart := TRUE;
  _vertices := Dom::Graph::getVertices(G);
  if nops(_vertices) = 0 then
    error("Graph has no vertices !");
  end_if;

  vertex := op(_vertices[1]);
  toSearch := {op(_vertices)} minus {vertex};
  visitedSet := {vertex};
  _color[vertex] := 1;
  setOne := 1;
  setTwo := 0;
  setToDo := {};

  while not nops(visitedSet) = 0 do
    visitedSet := visitedSet minus {vertex};
    // Check the adjacent edges
    _adjacentEdges := { op(Dom::Graph::getAdjacentEdgesLeaving(G, [vertex])) };

    // no incoming and no outgoing edge !
    if nops(_adjacentEdges) = 0 then
      if (contains(_color, vertex) = FALSE) then      
        setToDo := setToDo union {vertex};
      end_if;
    else
      if (contains(_color, vertex) = FALSE) then
        if setTwo > setOne then
          _color[vertex] := 1;
          setOne := setOne + 1;
        else
          _color[vertex] := 2;
          setTwo := setTwo + 1;
        end_if;
      end_if;     
    end_if;

    for v in _adjacentEdges do
      if (contains(_color, v) = FALSE) then
        _color[v] := _color[vertex] + 1;
        if _color[v] mod 2 = 1 then
          setOne := setOne + 1;
        else
          setTwo := setTwo + 1;
        end_if;

        visitedSet := visitedSet union {v};
        toSearch := toSearch minus {v}
      else
        if _color[v] = _color[vertex] then
          bipart := FALSE;
        end_if;
      end_if;
    end_for;
    if nops(visitedSet) = 0 then
      if nops(toSearch) > 0 then
        vertex := op(toSearch, 1);
        visitedSet := {vertex};
        toSearch := toSearch minus {vertex};
      end_if;
    else
      vertex := op(visitedSet, 1);      
    end_if;
  end_while;

  // Now add the vertices that are left (not connected to the graph) to the set that has lesser operators.
  for v in setToDo do
    if (contains(_color, v) = FALSE) then // For directed graphs necessary (vertices may be connected!)
      if setTwo > setOne then
        _color[op(v)] := 1;
        setOne := setOne + 1;
      else
        _color[op(v)] := 2;
        setTwo := setTwo + 1;
      end_if;
    end_if;
  end_for;

  if (output = hold(Bool)) then
    bipart;
  else
    if (bipart = TRUE) then
      partOne := {};

      for v in _vertices do
        if _color[v] mod 2 = 0 then
          partOne := partOne union {v}
        end_if;
      end_for;
      partTwo := {op(_vertices)} minus partOne;

      [sort([op(partOne)]), sort([op(partTwo)])];
    else
      FAIL; // output = Table and graph is not bipartite.
    end_if;
  end_if;
end_proc:

/*
  isConnected(G, startVertex)
  finds out if a graph is connected or not. 

  Parameters:
  G           - Graph
  
*/
Dom::Graph::isConnected := proc(G : Dom::Graph)
local parents, counter, singleVertex, i;
begin
  parents := Dom::Graph::depthFirstSearch(G);
  parents := op(parents,3);
  // If in the third list the parent is infinity, or more than once vertex=vertex then there exist a single vertex 
  counter := 0;
  singleVertex := 0;
  for i in parents do
    if op(i,1) = op(i,2) then
      counter := counter + 1;
    elif op(i,2) = infinity then
      singleVertex := singleVertex + 1;
    end_if;
  end_for;
  userinfo(NoNL, 1, "Components in the graph: " . expr2text(counter) . "\n" );    
  userinfo(NoNL, 1, "Single vertices in the graph:  " . expr2text(singleVertex) . "\n" );    
  if (singleVertex = 0) and (counter < 2) then
    TRUE;
  else
    FALSE;
  end_if;
end_proc:

/*
 Dom::Graph::stronglyConnectedComponents

  Decompose a network into strongly connected components.

  Input: A graph G

  Output: A list of graphs, which are the strongly connected
          components of G.  

  Algorithm: Chapter 23.5 of "Introduction to Algorithms", Cormen, Leiserson, Rivest.
  This proc exists only to have a file-scope static variable DFS
  which stores a helper function
*/

Dom::Graph::stronglyConnectedComponents := proc(G : Dom::Graph)
  //option escape;
  local DFS, dSCC;

begin
  // depth-first-search returning
  // the discovery and finish times and the respective predecessors
  DFS :=  proc(G : Dom::Graph, backwards = FALSE : DOM_BOOL)
    local DFS_visit, vertices, u, d, flow, pre, now, get_children, components;
  begin

    if backwards then
      get_children := Dom::Graph::getEdgesEntering(G);
    else
      get_children := Dom::Graph::getEdgesLeaving(G);
    end_if;
    
    DFS_visit := proc(u, vertices_in_component)
      local v;
    begin
      now := now + 1;
      d[u] := now;
      vertices_in_component := vertices_in_component.[u];
      for v in get_children[u] do
        if iszero(d[v]) then
          pre[v] := u;
          vertices_in_component := DFS_visit(v, vertices_in_component);
        end_if;
      end_for;
      now := now + 1;
      flow[u] := now;
      vertices_in_component;
    end_proc;
    
    if args(0) > 2 then
      vertices := args(3);
    else
      vertices := Dom::Graph::getVertices(G);
    end_if;
    d := table((u=0)$u in vertices);
    flow := table((u=0)$u in vertices);
    pre := table((u=FAIL)$u in vertices);
    now := 1;
    components := [];
    for u in vertices do
      if iszero(d[u]) then // vertex not visited yet
        components := components.[DFS_visit(u, [])];
      end_if;
    end_for;
    [d, flow, pre, components];
  end_proc;

  dSCC := proc(G : Dom::Graph)
    local fu, components;
  begin
    // call DFS(G) to compute finishing times flow[u] for each vertex u
    fu := DFS(G, FALSE)[2];
    
    // call DFS(Nt), but in the main loop of DFS, consider the vertices
    // in order of decreasing fu (as computed above)
    fu := sort([op(fu)], (a, b) -> rhs(a) > rhs(b));
    fu := map(fu, lhs);
    components := DFS(G, TRUE, fu)[4];

    // discard one-element components
    // components := select(components, c -> nops(c) > 1);

    // output the vertices of each tree in the depth-first forest
    // of the preceding step as a strongly connected component
    map(components, c -> Dom::Graph::getSubGraph(G, c));
  end_proc;

  dSCC(G);

end_proc:

/*-
 contract - Contracts vertices

-*/
Dom::Graph::contract := proc(G : Dom::Graph, Components : DOM_TABLE)
local _edgesLeaving, _edgesEntering, _edges, _edge, _vertices, _set1, _set2, found1, found2,i,position,
      leftOperator1, rightOperator1, leftOperator2, rightOperator2, vertex1, vertex2, _newEdges, useVertices;
  
begin
  if not Dom::Graph::getEdgeWeights(G) = FAIL then
  	 error("Graph is weighted! To contract, it has to be unweighted.");
  end_if;
  if not Dom::Graph::getEdgeCosts(G) = FAIL then
  	 error("Graph has edge costs! To contract, edges must not have costs.");
  end_if;
  if not Dom::Graph::getEdgeDescriptions(G) = FAIL then
  	 error("Graph has edge descriptions! To contract, edges must not have descriptions.");
  end_if;

  _edges := Dom::Graph::getEdges(G);
  _vertices := Dom::Graph::getVertices(G);
  _newEdges := [];
  // to check the vertices truly used
  useVertices := table();

  for _edge in _edges do    

    found1 := FALSE;
  	 for _set1 in Components do
      leftOperator1 := op(_set1, 1);
      rightOperator1 := op(_set1, 2);
  	 	 if contains(rightOperator1, op(_edge,1)) > 0 then
        vertex1 := leftOperator1;
        useVertices[leftOperator1] := 0; // rhs does not matter. Only lhs !
  	 	 	 found1 := TRUE;
        break;
  	 	 end_if;
  	 end_for;

    found2 := FALSE;
  	 for _set2 in Components do
      leftOperator2 := op(_set2, 1);
      rightOperator2 := op(_set2, 2);
  	 	 if contains(rightOperator2, op(_edge,2)) > 0 then
        vertex2 := leftOperator2;
        useVertices[leftOperator2] := 0; // rhs does not matter. Only lhs !
  	 	 	 found2 := TRUE;
        break;
  	 	 end_if;
  	 end_for;

    if found1 then
      // Vertex1 was found in sets     
    	 if found2 then
    	 	 // Vertex2 was found in sets
        _newEdges := append(_newEdges, [vertex1, vertex2]);
      else
        // Vertex2 was not found
        _newEdges := append(_newEdges, [vertex1, op(_edge,2)]);
    	 end_if;
    else
      // Vertex1 was not found in sets
    	 if found2 then
    	 	 // Vertex2 was found in sets
        _newEdges := append(_newEdges, [op(_edge,1), vertex2]);
      else
        // Vertex2 was not found
        _newEdges := append(_newEdges, [op(_edge,1), op(_edge,2)]);
    	 end_if;
    end_if;

  end_for;

  for _set1 in Components do
    // add name of set 1 to vertices id it was truly used...
    if contains(useVertices, op(_set1, 1)) then
    	 _vertices := append(_vertices, op(_set1, 1));    	
    end_if;
    // Remove vertices no longer needed from sets:
    _set1 := op(_set1, 2);
    for _set2 in _set1 do
      position := contains(_vertices, _set2);
      if position > 0 then
        delete _vertices[position];
      end_if;
    end_for;
  end_for;

  if not Dom::Graph::isDirected(G) then
    for i from 1 to nops(_newEdges) do
      _newEdges[i] := sort(_newEdges[i]);
    end_for;  	
  end_if;

  // remove Duplicate edges
  _newEdges := [op({op(_newEdges)})];

  if Dom::Graph::isDirected(G) then
    Dom::Graph::new(_vertices, _newEdges, hold(Directed));
  else
    Dom::Graph::new(_vertices, _newEdges);
  end_if;
end_proc:

/***************************************************************************************************************
 *** Here starts everything that has to do with networks                                                     ***
 ***************************************************************************************************************/

/*-
 Dom::Graph::convertSSQ - converts a Dom::Graph to a single source, single sink Dom::Graph
 convertSSQ(G,s,t)
 Parameters:
   G - Dom::Graph
   s - source (new node)
   t - sink (new node)
 Converts a Dom::Graph to a single source, single sink Dom::Graph. The given nodes s and t 
 are inserted in the Dom::Graph. The are connected to the other nodes in the following way:
   [s,i] for every i with Dom::Graph::getVertexWeight[i] > 0
   [i,t] for every i with Dom::Graph::getVertexWeight[i] < 0
   _edgeWeights[[s,i]] := getEdgeWeight[i]   getEdgeWeight[[s,i]] := 0
   _edgeWeights[[i,t]] := getEdgeWeight[i]   getEdgeWeight[[i,t]] := 0
-*/
Dom::Graph::convertSSQ := proc(G : Dom::Graph,s : DOM_LIST,t : DOM_LIST)
local _vertices,_vertexWeights,_edges,_edgeWeights,_edgeCosts,f,Q,S,i,new_Eds,new_Edt, G2;
begin

  _vertices := Dom::Graph::getVertices(G);

  if testargs() then
     if contains(_vertices, s) <> 0 or contains(_vertices, t) <> 0 then
     error("Specified nodes are already nodes of the Dom::Graph!");
     end_if;
  end_if;

  // Shortcut if the graph has no edges at all.
  if Dom::Graph::getEdgeNumber(G) = 0 then
    _vertices := _vertices.[s,t];
  	 return (Dom::Graph::new(_vertices, [[]], hold(Directed)));
  end_if;

  _vertexWeights := Dom::Graph::getVertexWeights(G);
  _edges := Dom::Graph::getEdges(G);
  _edgeWeights := Dom::Graph::getEdgeWeights(G);
  _edgeCosts := Dom::Graph::getEdgeCosts(G);
  s := op(s);
  t := op(t);

  if _vertexWeights = FAIL then
    for i in _edges do
      _vertexWeights[i] := 0;
    end_for;
  end_if;

  // Set to 1 because of predefinition
  if _edgeWeights = FAIL then
    for i in _edges do
      _edgeWeights[i] := 1;
    end_for;
  end_if;

  if _edgeCosts = FAIL then
    for i in _edges do
      _edgeCosts[i] := 0;
    end_for;
  end_if;

  f := proc() begin bool(_vertexWeights[args(1)] > 0) end_proc ;
  Q := select(_vertices, f);
  f := proc() begin bool(_vertexWeights[args(1)] < 0) end_proc ;
  S := select(_vertices, f);

  new_Eds := [[s,Q[i]] $ i=1..nops(Q)];
  new_Edt := [[S[i],t] $ i=1..nops(S)];
  _vertexWeights[s] := 0;
  _vertexWeights[t] := 0;
  for i in Q do 
    _edgeCosts[[s,i]] := 0;
    _edgeWeights[[s,i]] := _vertexWeights[i];
    _vertexWeights[s] := _vertexWeights[s] + _vertexWeights[i];
    _vertexWeights[i] := 0;
  end_for;
  for i in S do 
    _edgeCosts[[i,t]] := 0;
    _edgeWeights[[i,t]] := -_vertexWeights[i];
    _vertexWeights[t] := _vertexWeights[t] + _vertexWeights[i];
    _vertexWeights[i] := 0;
  end_for;
  _edges := _edges.new_Eds.new_Edt;
  _vertices := _vertices.[s,t];

  G2 := Dom::Graph(_vertices, _edges, Directed);
  subsop(G2, 3 = _vertexWeights, 5 = _edgeWeights, 6 = _edgeCosts);
end_proc:

/*- 
 admissibleFlow - checks a flow for admissibility
 admissibleFlow(G, flow)
 
 Costs will be filled up with 1 if not existing !

 Parameters:
    G    - Dom::Graph
    flow - flow in G (a table with flow[e] = c, when c units flow over the edge e)
 Returns TRUE, if flow is an admissible flow in G. FALSE otherwise.
-*/
Dom::Graph::admissibleFlow := proc(G : Dom::Graph, flow)
local _edges, _edgeWeights, Eset, item, edge, value;
begin 
  if testargs() then
    if args(0) <> 2 then
       error("Wrong number of arguments!");
    end_if;
    if not testtype(flow, DOM_TABLE) then
       error("Illegal argument (flow must be a DOM_TABLE)!");
    end_if;
    if not testtype(flow, Type::TableOfIndex(DOM_LIST/*,2,2*/)) then
       error("Illegal argument (flow must be a DOM_TABLE with edges as indices)!");   
    end_if;
    if not testtype(flow, Type::TableOfEntry(Type::Numeric)) then
       error("Illegal argument (flow values must be numeric values)!");
    end_if;
  end_if;
  
  _edges := Dom::Graph::getEdges(G);
  _edgeWeights := Dom::Graph::getEdgeWeights(G); 

  if _edgeWeights = FAIL then
    for edge in _edges do
      _edgeWeights[edge] := 1;
    end_for;
  end_if;

  Eset := [];
  for item in flow do 
    edge := op(item,1);
    Eset := append(Eset, edge);
    value := op(item, 2);
    if not Dom::Graph::isEdge(G, [edge]) then
       return(FALSE);
    end_if;
    if value < 0 or value > _edgeWeights[edge] then
       return(FALSE);
    end_if;
  end_for;
  if {op(Eset)} <> {op(_edges)} then
     return(FALSE);
  end_if;
  TRUE
end_proc:

/*-
  pp_fifo - computes maximal flow
  FIFO-Preflow-Push-Algorithm of Goldberg&Tarjan.
  "A new approach to the maximum-flow problem", Journal ACM 35(4), 1988
  Running time: O(n^3)
  Uses an exact distance labelling.
-*/
Dom::Graph::pp_fifo := proc(G : Dom::Graph, s : DOM_LIST, t : DOM_LIST)
local Vd,d,inflow,outflow,preflow,i,summe,Q,relabel,dist_sink,schritt,
      _vertices, _edges, _edgeWeights, _edgesEntering, _edgesLeaving;
begin

  _vertices := Dom::Graph::getVertices(G);
  _edges := Dom::Graph::getEdges(G);
  _edgeWeights := Dom::Graph::getEdgeWeights(G);
  _edgesEntering := Dom::Graph::getEdgesEntering(G);
  _edgesLeaving := Dom::Graph::getEdgesLeaving(G);

  // If no "Capacities" are predefined, 1 will be used.
  if _edgeWeights = FAIL then
    for i in _edges do
      _edgeWeights[i] := 1;
    end_for;
  elif nops(_edgeWeights) < nops(_edges) then
    for i in _edges do
      if not contains(_edgeWeights, i) then
        _edgeWeights[i] := 1;        	
      end_if;
    end_for;    	
  end_if;

  s := op(s);
  t := op(t);

    /*
      relabel
    */
    relabel := proc(v, flow, d, _edgeWeights, _edgesLeaving, _edgesEntering)
    local i,e,mini;
    begin
      mini := 3 * nops(_edgesLeaving); //- Upper bound for distance labels: 3 * |_vertices| -
      for i in _edgesLeaving[v] do 
        e := [v,i];
        if _edgeWeights[e]-flow[e] > 0 and d[v] <= d[i] then
          mini := min(mini, d[i]+1);
        end_if;
      end_for;
      for i in _edgesEntering[v] do 
        e := [i,v];
        if flow[e] > 0 and d[v] <= d[i] then
          mini := min(mini, d[i]+1);
        end_if;
      end_for;
      d[v] := mini;
      d;
    end_proc:

    /*--
     dist_sink - Berechnet den Abstand der Knoten zur Senke
     Es werden die k"urzesten Wege von der Senke t zu allen "ubrigen Knoten
     im Netzwerk berechnet. Das Ergebnis ist eine Tabelle d, mit
            d[v] = L"ange des k"urzesten Weges von v nach t
    --*/
    dist_sink := proc(_vertices,_edgesEntering,t)
    local d,j,LIST,i,n;
    begin
      n := nops(_vertices);
      d := table(_vertices[i] = n $ i=1..nops(_vertices));
      d[t] := 0;
      LIST := [t];
      while LIST <> [] do
        i := LIST[1];
        delete LIST[1];
        for j in _edgesEntering[i] do
          if d[j] > d[i] + 1 then
            d[j] := d[i] + 1;
            if contains(LIST, j) = 0 then
              LIST := append(LIST, j);
            end_if;
          end_if;
        end_for;
      end_while;
      d;
    end_proc:

    schritt := proc(QQ)
    local Q,i,j,e,f,PP,v,abgabe;
    begin
      Q:=QQ; // to avoid warnings 
      for i in Q do
        PP := inflow[i] - outflow[i];
        j := 1;
        v := op(_edgesLeaving[i],j);
        /*--  Abgabe an Nachfolgeknoten --*/
        while v <> FAIL and 0 < PP do
          e := [i,v];
          if d[i] = d[v]+1 then
            abgabe := max(min(_edgeWeights[e] - preflow[e], PP), 0);
            preflow[e] := preflow[e] + abgabe;
            PP := PP-abgabe; 
            outflow[i] := outflow[i] + abgabe;
            inflow[v] := inflow[v] + abgabe;
          end_if;
          j := j+1;
          v := op(_edgesLeaving[i],j);
        end_while;
        j := 1;
        v := op(_edgesEntering[i],j);
        /*-- Rueckgabe an Vorg"angerknoten --*/
        while v <> FAIL and 0 < PP do 
          e := [v,i];
          if d[i] = d[v]+1 then
            abgabe := max(min(preflow[e], PP),0);
            preflow[e] := preflow[e] - abgabe;
            PP := PP - abgabe;
            inflow[i] := inflow[i] - abgabe;
            outflow[v] := outflow[v] - abgabe;
          end_if;
          j := j+1;
          v := op(_edgesEntering[i],j);
        end_while;
        /*- Es konnte nicht der ganze "Uberschu"t verteilt werden -> Relabel -*/
        if PP > 0 then 
          d := relabel(i, preflow, d, _edgeWeights, _edgesLeaving, _edgesEntering); 
        end_if;
      end_for;

      f := proc() begin bool(inflow[args(1)] <> outflow[args(1)]) end_proc ;
      Q := select(Vd, f);
    end_proc:

  Vd := [op({op(_vertices)} minus {s,t})];
  d := dist_sink(_vertices, _edgesEntering,t);
  d[s] := nops(_vertices);
  d[FAIL] := 0;
  preflow := table(_edges[i] = 0 $ i=1..nops(_edges));
  inflow := table(_vertices[i] = 0 $ i=1..nops(_vertices));
  outflow := inflow;
   
  Q := []:
  for i in _edgesLeaving[s] do
    preflow[[s,i]] := _edgeWeights[[s,i]];
    outflow[s] := outflow[s] + _edgeWeights[[s,i]];
    inflow[i] := _edgeWeights[[s,i]];
    Q := append(Q, i);
  end_for;

  Q := schritt(Q);
  while Q <> [] do
    Q := schritt(Q);
  end_while;

  delete i;
  summe := _plus(preflow[[_edgesEntering[t][i],t]] $ i=1..nops(_edgesEntering[t]));

  summe, preflow;
end_proc:

/*- 
 minCut - computes a minimal cut
 minCut(G,s,t)
 Parameters:
   G - network
   s,t - vertices in G
 Computes a minimal cut seperating s and t in G.
 Uses Graph::pp_fifo.
-*/

Dom::Graph::minCut := proc(G : Dom::Graph,s : DOM_LIST,t : DOM_LIST)
local _vertices, _edges, _edgeWeights, _edgesLeaving, _edgesEntering,flow,cutset,S,Q,i,j,e,vc;
begin
  if testargs() then
    if args(0) <> 3 then
     error("Wrong number of arguments!");
    end_if;
    if not Dom::Graph::isVertex(G,s) and not Dom::Graph::isVertex(G,t) then
      error("Specified vertices are not contained in network!");
    end_if;
  end_if;

  _vertices := Dom::Graph::getVertices(G);
  _edges := Dom::Graph::getEdges(G);
  _edgeWeights := Dom::Graph::getEdgeWeights(G);
  _edgesLeaving := Dom::Graph::getEdgesLeaving(G);
  _edgesEntering := Dom::Graph::getEdgesEntering(G);

  if _edgeWeights = FAIL then
    for i in _edges do
      _edgeWeights[i] := 1;
    end_for;
  elif nops(_edgeWeights) < nops(_edges) then
    for i in _edges do
      if not contains(_edgeWeights, i) then
        _edgeWeights[i] := 1;        	
      end_if;
    end_for;    	
  end_if;

  flow := Dom::Graph::pp_fifo(G, s,t);
  vc := op(flow,1);
  flow := op(flow,2);
  s := op(s);
  t := op(t);

  /*- 
   Compute a minimal cut [S,T] with s in S and t in T.
   S is the set of all nodes, for which there exists a path
   from s in the reduced network.
  -*/
  S := {s}:
  Q := [s]:
  while Q <> [] do
     i := Q[1]:
     delete Q[1];
     for j in _edgesLeaving[i] do
        e := [i,j];
        if _edgeWeights[e] - flow[e] > 0 then
           if contains(Q,j) = 0 then
              Q := append(Q,j);
              S := S union {j};
           end_if;
        end_if;
     end_for;
   end_while:
   cutset := [];
   for i in S do 
      for j in _edgesLeaving[i] do
         if contains(S,j) then
      next;
   else
      cutset := append(cutset,[i,j]);
   end_if;
      end_for;
   end_for;
   vc, cutset;
end_proc:

 /*--
 minCost - minimal cost flow
 minCost(G)
 Parameters:
   G - network
 Uses Dom::Graph::relaxation
--*/
Dom::Graph::minCost := proc(G : Dom::Graph)
local flow, i, _edges, _edgeCosts, costTable; 
begin
   if testargs() then
     if args(0) <> 1 then
       error("Wrong number of parameters!");
     end_if;
   end_if;
   _edges := Dom::Graph::getEdges(G);
   _edgeCosts := Dom::Graph::getEdgeCosts(G);

  // If no "Capacities" are predefined, 0 will be used.
  if _edgeCosts = FAIL then
    for i in _edges do
      _edgeCosts[i] := 0;
    end_for;
  elif nops(_edgeCosts) < nops(_edges) then
    for i in _edges do
      if not contains(_edgeCost, i) then
        _edgeCosts[i] := 0;        	
      end_if;
    end_for;    	
  end_if;

  flow := Dom::Graph::relaxation(G);
  costTable := table(_edges[i] = _edgeCosts[_edges[i]] * flow[1][_edges[i]] $ i=1..nops(_edges));
  flow[1], costTable, _plus(_edgeCosts[_edges[i]] * flow[1][_edges[i]] $ i=1..nops(_edges)), flow[2];
end_proc:

/*- 
 maxFlow - computes a maximal flow
 maxFlow(G, t, t)
 Parameters:
   G    - Dom::Graph
   s, t - vertices in G
 Computes a maximal flow from t to t in G.
 Uses Dom::Graph::pp_fifo.
-*/
Dom::Graph::maxFlow := proc(G : Dom::Graph, s : DOM_LIST, t : DOM_LIST)
begin
  if testargs() then
    if args(0) <> 3 then
      error("Wrong number of arguments!");
    end_if;
    if not Dom::Graph::isVertex(G,s) and not Dom::Graph::isVertex(G,t) then
      error("Specified vertices are not contained in graph!");
    end_if;
  end_if;

  Dom::Graph::pp_fifo(G, s, t);
end_proc:

/*-
 residualGraph - computes the residual network
 residualGraph(G, flow <, Extended>)
 Parameters:
    G - Dom::Graph
    flow - flow in G (a table with flow[_edge] = c, when c units flow over the edge _edge)
    Extended - if Extended is stated, also those edges with a zero
               residual capacity are contained. Otherwise those edges are omitted.
-*/
Dom::Graph::residualGraph := proc(G : Dom::Graph, flow)
local _edges,_edge,_edgeCosts,de,der,dec,dew,ce,cec,cew,G1,_edgeWeights,reverse;
begin
  if testargs() then
    if args(0) < 2 or args(0) > 3 then
       error("Wrong number of arguments!");
    end_if;
    if not Dom::Graph::admissibleFlow(G, flow) then
       error("Specified flow is inadmissible in the given network!");
    end_if;   
    if args(0) = 3 then 
       if args(3) <> hold(Extended) then
          error("Illegal argument!");
       end_if;
    end_if; 
  end_if; 
        
  reverse := proc() begin [args(1)[2],args(1)[1]] end_proc ;
  
  _edges := Dom::Graph::getEdges(G);
  _edgeCosts := Dom::Graph::getEdgeCosts(G);
  _edgeWeights := Dom::Graph::getEdgeWeights(G);

  // If no "Capacities" are predefined, 1 will be used.
  if _edgeWeights = FAIL then
    for _edge in _edges do
      _edgeWeights[_edge] := 1;
    end_for;
    G := Dom::Graph::setEdgeWeights(G, _edges, _edgeWeights, OnlySpecifiedEdges=TRUE);
  elif nops(_edgeWeights) < nops(_edges) then
    for _edge in _edges do
      if not contains(_edgeWeights, _edge) then
        _edgeWeights[_edge] := 1;      	
      end_if;
    end_for;
    G := Dom::Graph::setEdgeWeights(G, _edges, _edgeWeights, OnlySpecifiedEdges=TRUE);
  end_if;

  if _edgeCosts = FAIL then
    for _edge in _edges do
      _edgeCosts[_edge] := 0;
    end_for;
    G := Dom::Graph::setEdgeCosts(G, _edges, _edgeCosts, OnlySpecifiedEdges=TRUE);
  elif nops(_edgeCosts) < nops(_edges) then
    for _edge in _edges do
      if not contains(_edgeCosts, _edge) then
        _edgeWeights[_edge] := 1;      	
      end_if;
    end_for;
    G := Dom::Graph::setEdgeWeights(G, _edges, _edgeWeights, OnlySpecifiedEdges=TRUE);
  end_if;

  if args(0) = 3 then   // Extended network: don't omit empty edges 
    der := []; dec := []; dew := [];
    ce := []; cec := []; cew := [];
    for _edge in _edges do
      if flow[_edge] > 0 then
        ce := append(ce,_edge);
        cec := append(cec, _edgeWeights[_edge]-flow[_edge]);
        if not Dom::Graph::isEdge(G, [reverse(_edge)]) then
          der := append(der, reverse(_edge));
          dec := append(dec, flow[_edge]);
          dew := append(dew, -_edgeCosts[_edge]); 
        end_if;
      end_if;
    end_for;
    G1 := G;
    if nops(der) > 0 then
      G1 := Dom::Graph::addEdge(G1, der, hold(EdgeWeights)=dec, hold(EdgeCosts)=dew, OnlySpecifiedEdges=TRUE);
    end_if;
    if nops(ce) > 0 then
      G1 := Dom::Graph::setEdgeWeights(G1, ce, cec, OnlySpecifiedEdges=TRUE);
    end_if;
  else // omit empty edges 
    de := []; der := []; dec := []; dew := [];
    ce := []; cec := []; cew := [];
    for _edge in _edges do
      if flow[_edge] > 0 then
        if flow[_edge] = _edgeWeights[_edge] then
          de := append(de,_edge);
        else
          ce := append(ce,_edge);
          cec := append(cec, _edgeWeights[_edge]-flow[_edge]);
        end_if;
        if not Dom::Graph::isEdge(G, [reverse(_edge)]) then
          der := append(der, reverse(_edge));
          dec := append(dec, flow[_edge]);
          dew := append(dew, -_edgeCosts[_edge]); 
        end_if;
      end_if;
    end_for;
    G1 := G;
    if nops(de) > 0 then
       G1 := Dom::Graph::removeEdge(G1, de, OnlySpecifiedEdges=TRUE);
    end_if;
    if nops(der) > 0 then
      G1 := Dom::Graph::addEdge(G1, der, hold(EdgeWeights)=dec, hold(EdgeCosts)=dew, OnlySpecifiedEdges=TRUE);
    end_if;
    if nops(ce) > 0 then
      G1 := Dom::Graph::setEdgeWeights(G1, ce, cec, OnlySpecifiedEdges=TRUE);
    end_if;
  end_if;
  G1;
end_proc:

/*-
 relaxation - Relaxation Algorithm for solving minimal cost flow problem
    D.P. Bertsekas: 
    Linear Dom::Graph Optimization, 
    MIT Press, Cambridge(Mass.)-London,1991
-*/
Dom::Graph::relaxation := proc(G : Dom::Graph)
local e,i,flow,price,deficit,DD,a,f, has_out_capacity, update_dd, flow_augmentation, price_change, relax_step,
      _vertices, _edges, _vertexWeights, _edgeCosts, _edgeWeights, _edgesLeaving, _edgesEntering;

begin
  _vertices := Dom::Graph::getVertices(G);
  _edges := Dom::Graph::getEdges(G);
  _vertexWeights := Dom::Graph::getVertexWeights(G);
  _edgeCosts := Dom::Graph::getEdgeCosts(G);
  _edgeWeights := Dom::Graph::getEdgeWeights(G);
  _edgesLeaving := Dom::Graph::getEdgesLeaving(G);
  _edgesEntering := Dom::Graph::getEdgesEntering(G);
  flow := table();

  if _vertexWeights = FAIL then
    for i in _edges do
      _vertexWeights[i] := 0;
    end_for;
  end_if;

  if _edgeWeights = FAIL then
    for i in _edges do
      _edgeWeights[i] := 0;
    end_for;
  end_if;

  // Set to 1 because of predefinition
  if _edgeCosts = FAIL then
    for i in _edges do
      _edgeCosts[i] := 1;
    end_for;
  end_if;

    /*- update_dd - computation of directional derivative  -*/
    update_dd := proc(S,_edgesLeaving,_edgesEntering,_edgeCosts,_edgeWeights,flow,price,deficit)
    local summe, j,e, v, pp;
    begin
      summe := _plus((deficit[op(S,j)] $ j=1..nops(S)));
      for v in S do
        for j in _edgesLeaving[v] do 
          e := [v,j];
          pp := price[e[1]] - (_edgeCosts[e] + price[e[2]]);
          if pp = 0 then
            if not contains(S,j) then
              summe := summe - (_edgeWeights[e] - flow[e])
            end_if
          end_if
        end_for;
        for j in _edgesEntering[v] do 
          e := [j,v];
          pp := price[e[1]] - (_edgeCosts[e] + price[e[2]]);
          if pp = 0 then
            if not contains(S,j) then
              summe := summe - flow[e]; //- (flow[e] - low[e]) -
            end_if
          end_if
        end_for;
      end_for;
      summe;
    end_proc:

    /*-
     flow_augmentation - increases flow along an augmenting path from i to j 
    -*/
    flow_augmentation := proc(label,i,j,defic,flo,_edgeWeights)
    local Pp, Pm, oplist, l,w,d,v,deficit,flow;
    // global label (local variable from relax_step) 
    // removed 
    begin
      deficit:=defic; flow:=flo; // to avoid warnings 
      Pp := {}; Pm := {};
      oplist := [deficit[i], -deficit[j]];
      w := j;
      while TRUE do
        l := label[w];
        if w = l[1] then
          // R"uckw"artskante, Fluss wird reduziert 
          Pm := Pm union {l};
          oplist := oplist.[flow[l]];  //- flow[l] - low[l] -
          v := l[2];
        else
          // Vorw"artskante, Fluss wird erh"oht 
          Pp := Pp union {l};
          oplist := oplist.[_edgeWeights[l] - flow[l]];
          v := l[1];
        end_if;
        if v = i then
          break
        end_if;
        w := v;
      end_while;

      if nops(oplist) > 0 then
        d := min(op(oplist));
      else
        d := 0;
      end_if;

      userinfo(2, d,Pp,Pm);

      for e in Pp do
        flow[e] := flow[e] + d;
        deficit[e[1]] := deficit[e[1]] - d;
        deficit[e[2]] := deficit[e[2]] + d
      end_for;

      for e in Pm do
        flow[e] := flow[e] - d;
        deficit[e[1]] := deficit[e[1]] + d;
        deficit[e[2]] := deficit[e[2]] - d
      end_for;
      flow,deficit;
    end_proc:

    /*++ price_change - price adaption  ++*/
    price_change := proc(S,_edgesLeaving,_edgesEntering,_edgeCosts,_edgeWeights,pric,defic,flo)
    local g,i,e,oplist,pp,j, deficit, flow, price;
    begin
      deficit:=defic; flow:=flo; price:=pric; // to avoid warnings 
      oplist := [];
      for i in S do 
        for j in _edgesLeaving[i] do
          e := [i,j];
          if not contains(S,j) then
            pp := price[e[1]] - (_edgeCosts[e] + price[e[2]]);
            if pp = 0 then
              deficit[e[1]] := deficit[e[1]] + flow[e];
              deficit[e[2]] := deficit[e[2]] - flow[e];
              flow[e] := _edgeWeights[e];
              deficit[e[1]] := deficit[e[1]] - flow[e];
              deficit[e[2]] := deficit[e[2]] + flow[e];
            elif flow[e] < _edgeWeights[e] then
              oplist := oplist.[price[j] + _edgeCosts[e] - price[i]]
            end_if;
          end_if;
        end_for;
        for j in _edgesEntering[i] do
          e := [j,i];
          if not contains(S,j) then
            pp := price[e[1]] - (_edgeCosts[e] + price[e[2]]);
            if pp = 0 then
              deficit[e[1]] := deficit[e[1]] + flow[e];
              deficit[e[2]] := deficit[e[2]] - flow[e];
              flow[e] := 0; //- := low[e] -
              deficit[e[1]] := deficit[e[1]] - flow[e];
              deficit[e[2]] := deficit[e[2]] + flow[e];
            elif 0 < flow[e] then //- low[e] < flow[e] -
              oplist := oplist.[price[j] - _edgeCosts[e] - price[i]]
            end_if;
           end_if;
        end_for;
      end_for;

      if nops(oplist) > 0 then
        g := min(op(oplist));
      else
        g := 0;
      end_if;

      for i in S do 
        price[i] := price[i] + g;
      end_for;
      flow,deficit,price;
    end_proc:

    /* relax step */
    relax_step := proc(v,_edges,_edgeWeights,_edgeCosts,_edgesLeaving,_edgesEntering,flo,defic,pric,DDD)
    local e,i,L,S,Lh,j,pp,label, price,deficit, flow, DD;
    begin
      L := {v}; S := {};
      DD:=DDD;
      price:=pric; deficit:=defic; flow:=flo; // to avoid warnings 
      while TRUE do
        if S = L then
          price_change(S,_edgesLeaving,_edgesEntering,_edgeCosts,_edgeWeights,price,deficit,flow);
          flow := %[1];
          deficit := %2[2];
          price := %3[3];
          return(flow,deficit,price,DD);
        else
          i := op(L minus S, 1);
          S := S union {i};
          DD := update_dd(S,_edgesLeaving,_edgesEntering,_edgeCosts,_edgeWeights,flow,price,deficit);
        end_if;
        if DD > 0 then 
          price_change(S,_edgesLeaving,_edgesEntering,_edgeCosts,_edgeWeights,price,deficit,flow);
          flow := %[1];
          deficit := %2[2];
          price := %3[3];
          return(flow,deficit,price,DD);
        else
          Lh := {};
          for j in _edgesLeaving[i] do
            if not contains(L,j) then
              e := [i,j];
              pp := price[e[1]] - (_edgeCosts[e] + price[e[2]]);
              if pp = 0 and flow[e] < _edgeWeights[e] then
                Lh := Lh union {j};
                label[j] := e;
              end_if;
            end_if;
          end_for;
          for j in _edgesEntering[i] do
            if not contains(L,j) then
              e := [j,i];
              pp := price[e[1]] - (_edgeCosts[e] + price[e[2]]);
              if pp = 0 and 0 < flow[e] then //- low_capacity[e] < flow[e] -
                Lh := Lh union {j};
                label[j] := e;
              end_if;
            end_if;
          end_for;
          L := L union Lh;
          for j in Lh do
            if deficit[j] < 0 then
              flow_augmentation(label,v,j,deficit,flow,_edgeWeights);
              flow := %[1];
              deficit := %2[2];
              return(flow,deficit,price,DD);
            end_if;
          end_for;
        end_if;
      end_while;
    end_proc:

   /*- Initialisation of 
        flow - network flow
        price - local prices
        deficit - supply/demand of nodes
        DD directional derivative
   -*/
   price := table(_vertices[i] = 0 $ i=1..nops(_vertices));
   deficit := table(_vertices[i] = _vertexWeights[_vertices[i]] $ i=1..nops(_vertices));
   for e in _edges do 
     i := price[e[1]] - (_edgeCosts[e] + price[e[2]]);
     if i <= 0 then
       flow[e] := 0;  //- lower_capacity[e] -
       deficit[e[1]] := deficit[e[1]] - flow[e];
       deficit[e[2]] := deficit[e[2]] + flow[e];
     else
       flow[e] := _edgeWeights[e];
       deficit[e[1]] := deficit[e[1]] - flow[e];
       deficit[e[2]] := deficit[e[2]] + flow[e];
     end_if;
   end_for;
   DD := 0;
   f := v -> bool(deficit[v] > 0);

   /* has_out_capacity */
   has_out_capacity := proc(v)
   local other, edge;
   begin
     for other in _edgesLeaving[v] do
       edge := [v, other];
       if not contains(flow, edge)
         or flow[edge] < _edgeWeights[edge] then
         return(TRUE);
       end_if
     end_for;
     return(FALSE);
   end_proc;
   
   repeat
     a := select(_vertices, f);
     a := select(a, has_out_capacity);
     if nops(a) = 0 then
      break; //- Termination: there are no more deficit nodes -
     end_if;
     a := relax_step(a[1],_edges,_edgeWeights,_edgeCosts,_edgesLeaving,_edgesEntering,flow,deficit,price,DD);
     flow := a[1];
     deficit := a[2];
     price := a[3];
     userinfo(1,"price=",price);
     DD := a[4];
   until FALSE end_repeat;
   
   flow,price;
end_proc:

// -------------------------------------------------------------------------------------------------------------
// Other additions:
// -------------------------------------------------------------------------------------------------------------

/*****************************************************************************
  $Id: Graph.mu,v 1.1.1.1 2003/09/15 16:37:42 nthiery Exp $
  File:         lib/DOMAINS/DOMAIN/GraphAdd.mu
  Authors:      Nicolas M. Thiry <nthiery@users.sourceforge.net>
                Florent Hivert <florent.hivert@univ-mlv.fr>
  Created:      30/09/2002
  License:      LGPL
*****************************************************************************/

Dom::Graph::levelDecomposition := 
/* Compute the level decomposition of a directed acyclic graph (DAG).
   
*/
proc(G : Dom::Graph) : Type::Union(Type::ListOf(DOM_LIST), DOM_FAIL)
    local _vertices, _inDegree, _edgesLeaving, res,
	  minimums,   // vertices that have no parents
	  newMins,    // new minimums
	  m, c; // remainings vertices
begin  
    if testargs() then
	if not Dom::Graph::isDirected(G) then
	    error("The graph must be directed.");
	end_if;
    end_if;
    _vertices := Dom::Graph::getVertices(G);
    _inDegree := Dom::Graph::inDegree(G);
    _edgesLeaving := Dom::Graph::getEdgesLeaving(G);
    minimums := select(_vertices, z -> (_inDegree[z]=0));
    res := [];
    while (minimums <> []) do
	newMins := [];
	for m in minimums do
	    delete(_inDegree[m]);
	    for c in _edgesLeaving[m] do
		_inDegree[c] := _inDegree[c] - 1;
		if _inDegree[c]= 0 then 
		    newMins := append(newMins, c);
		end_if;
	    end_for;
	end_for;
	res := append(res, minimums);
	minimums := newMins;
    end_while;
    if nops(_inDegree) = 0 then res else FAIL; end_if;
end_proc:


Dom::Graph::isAcyclic := 
// Test whether a directed graph is acyclic or not.
proc(G : Dom::Graph) : DOM_BOOL
begin
    if testargs() then
	if not Dom::Graph::isDirected(G) then
	    error("The graph must be directed.");
	end_if;
    end_if;
    if Dom::Graph::levelDecomposition(G) <> FAIL then TRUE else FALSE end_if;
end_proc:



// End of file
null():

